@@ -574,9 +574,12 @@ order for the change to take effect."
   ("lbag" . ("⟅"))
   ("rbag" . ("⟆"))
 
-  ("(|" . ("⦇"))
+  ("(|" . ("⦇"))  ;; Idiom brackets
   ("|)" . ("⦈"))
 
+  ("((" . ("⦅"))  ;; Banana brackets
+  ("))" . ("⦆"))
+
   ;; Primes.
 
   ("'" . ,(agda-input-to-string-list "′″‴⁗"))
@@ -774,15 +777,23 @@ order for the change to take effect."
   ("MIy" . ("𝒚"))
   ("MIz" . ("𝒛"))
   ("McA" . ("𝒜"))
+  ("McB" . ("ℬ"))
   ("McC" . ("𝒞"))
   ("McD" . ("𝒟"))
+  ("McE" . ("ℰ"))
+  ("McF" . ("ℱ"))
   ("McG" . ("𝒢"))
+  ("McH" . ("ℋ"))
+  ("McI" . ("ℐ"))
   ("McJ" . ("𝒥"))
   ("McK" . ("𝒦"))
+  ("McL" . ("ℒ"))
+  ("McM" . ("ℳ"))
   ("McN" . ("𝒩"))
   ("McO" . ("𝒪"))
   ("McP" . ("𝒫"))
   ("McQ" . ("𝒬"))
+  ("McR" . ("ℛ"))
   ("McS" . ("𝒮"))
   ("McT" . ("𝒯"))
   ("McU" . ("𝒰"))
@@ -795,7 +806,9 @@ order for the change to take effect."
   ("Mcb" . ("𝒷"))
   ("Mcc" . ("𝒸"))
   ("Mcd" . ("𝒹"))
+  ("Mce" . ("ℯ"))
   ("Mcf" . ("𝒻"))
+  ("Mcg" . ("ℊ"))
   ("Mch" . ("𝒽"))
   ("Mci" . ("𝒾"))
   ("Mcj" . ("𝒿"))
@@ -803,6 +816,7 @@ order for the change to take effect."
   ("Mcl" . ("𝓁"))
   ("Mcm" . ("𝓂"))
   ("Mcn" . ("𝓃"))
+  ("Mco" . ("ℴ"))
   ("Mcp" . ("𝓅"))
   ("Mcq" . ("𝓆"))
   ("Mcr" . ("𝓇"))
