@@ -1,2 +1,2 @@
-(define-package "agda2-mode" "2.5.3"
+(define-package "agda2-mode" "2.5.4.1"
   "interactive development for Agda, a dependently typed functional programming language")
