@@ -10,7 +10,7 @@
 
 ;;; Code:
 
-(defvar agda2-version "2.5.3"
+(defvar agda2-version "2.5.4.1"
   "The version of the Agda mode.
 Note that the same version of the Agda executable must be used.")
 
@@ -93,7 +93,7 @@ argument, and does not need to be listed here."
   :type '(repeat string)
   :group 'agda2)
 
-(defvar agda2-backends '("GHC" "JS" "LaTeX" "QuickLaTeX")
+(defvar agda2-backends '("GHC" "GHCNoMain" "JS" "LaTeX" "QuickLaTeX")
   "Compilation backends.")
 
 (defcustom agda2-backend
@@ -205,15 +205,15 @@ to this variable to take effect."
                     (standard-syntax-table))
     ;; Then override the remaining special cases.
     (dolist (cs '((?{ . "(}1n") (?} . "){4n") (?- . "w 123b") (?\n . "> b")
-                  (?. . ".") (?\; . ".") (?_ . ".") (?! . ".")))
+                  (?. . ".") (?\; . ".") (?! . ".")))
       (modify-syntax-entry (car cs) (cdr cs) tbl))
     tbl)
   "Syntax table used by the Agda mode:
 
-{}   | Comment characters, matching parentheses.
--    | Comment character, word constituent.
-\n   | Comment ender.
-.;_! | Punctuation.
+{}  | Comment characters, matching parentheses.
+-   | Comment character, word constituent.
+\n  | Comment ender.
+.;! | Punctuation.
 
 Remaining characters inherit their syntax classes from the
 standard syntax table if that table treats them as matching
@@ -434,6 +434,8 @@ agda2-include-dirs is not bound." :warning))
  ;; If Agda is not running syntax highlighting does not work properly.
  (unless (eq 'run (agda2-process-status))
    (agda2-restart))
+ ;; Make sure that Font Lock mode is not used.
+ (font-lock-mode 0)
  (agda2-highlight-setup)
  (condition-case err
      (agda2-highlight-reload)
@@ -967,7 +969,8 @@ Assumes that <clause> = {!<variables>!} is on one line."
   "Display the string STATUS in the current buffer's mode line.
 \(precondition: the current buffer has to use the Agda mode as the
 major mode)."
-  (setq agda2-buffer-external-status status))
+  (setq agda2-buffer-external-status status)
+  (force-mode-line-update))
 
 (defmacro agda2-warning-or-info-buffer (buffer kind title)
   "Used to define the functions agda2-warning-buffer and agda2-info-buffer."
@@ -1009,14 +1012,6 @@ The buffer is returned.")
 (agda2-warning-or-info-buffer
  agda2-info-buffer "info" "*Agda information*")
 
-(defun agda2-font-syntactic-face (state)
-  (cond ((nth 4 state)
-         ( save-excursion
-           (goto-char (nth 8 state))
-           (cond ((looking-at "--[[:space:]\n]") 'font-lock-comment-face)
-                 ((looking-at "{-[^#]") 'font-lock-comment-face)
-          )))))
-
 (defun agda2-info-action (name text &optional append)
   "Insert TEXT into the Agda info buffer and display it.
 NAME is displayed in the buffer's mode line.
@@ -1217,18 +1212,18 @@ is inserted, and point is placed before this text."
   "Quit and clean up after agda2."
   (interactive)
   (remove-hook 'first-change-hook 'agda2-abort-highlighting 'local)
+  (remove-hook 'after-save-hook 'agda2-highlight-tokens 'local)
   (agda2-remove-annotations)
   (agda2-term))
 
 (defun agda2-term ()
-  "Send a SIGTERM signal to the Agda2 process, then kill its buffer."
+  "Interrupt the Agda process and kill its buffer."
   (interactive)
+  (when (and agda2-process
+             (process-status agda2-process))
+    (interrupt-process agda2-process))
   (when (buffer-live-p agda2-process-buffer)
-    (with-current-buffer agda2-process-buffer
-      (condition-case nil
-          (signal-process agda2-process 'SIGTERM)
-        (error nil))
-      (kill-buffer))))
+    (kill-buffer agda2-process-buffer)))
 
 (defmacro agda2-maybe-normalised (name comment cmd want)
   "This macro constructs a function NAME which runs CMD.
@@ -1536,7 +1531,7 @@ Only if the buffer is unmodified, and only if there is anything to load."
 
 (defun agda2-literate-p ()
   "Is the current buffer a literate Agda buffer?"
-  (equal (file-name-extension (buffer-name)) "lagda"))
+  (not (equal (file-name-extension (buffer-file-name)) "agda")))
 
 (defmacro agda2--case (exp &rest branches) ;FIXME: Use `pcase' instead!
   (declare (debug t) (indent 1))
@@ -1829,18 +1824,7 @@ a file is loaded."
 ;; Comments and paragraphs
 
 (defun agda2-comments-and-paragraphs-setup nil
-  "Set up comment and paragraph handling for Agda mode."
-
-  ;; Syntax table setup for comments is done elsewhere.
-
-  ;; Enable highlighting of comments via Font Lock mode (which uses
-  ;; the syntax table).
-  (set (make-local-variable 'font-lock-defaults)
-       '(nil nil nil nil nil (font-lock-syntactic-face-function . agda2-font-syntactic-face)))
-  ;; If the following s-expression is removed, then highlighting of
-  ;; comments stops working.
-  (when font-lock-mode
-    (font-lock-mode t))
+  "Set up comment and paragraph handling for the Agda mode."
 
   ;; Empty lines (all white space according to Emacs) delimit
   ;; paragraphs.
@@ -1850,6 +1834,13 @@ a file is loaded."
   ;; Support for adding/removing comments.
   (set (make-local-variable 'comment-start) "-- ")
 
+  ;; Use the syntax table to locate comments (and possibly other
+  ;; things). Syntax table setup for comments is done elsewhere.
+  (set (make-local-variable 'comment-use-syntax) t)
+
+  ;; Update token-based highlighting after the buffer has been saved.
+  (add-hook 'after-save-hook 'agda2-highlight-tokens nil 'local)
+
   ;; Support for proper filling of text in comments (requires that
   ;; Filladapt is activated).
   (when (featurep 'filladapt)
@@ -1874,6 +1865,18 @@ From the beginning of the current line to the end of the buffer."
           (comment-dwim nil))
       (pop-mark))))
 
+(defun agda2-highlight-tokens nil
+  "Compute token-based highlighting information.
+
+Unless `agda2-highlight-level' is `none' or the Agda process is
+busy with something. This command might save the buffer."
+  (unless (or agda2-in-progress
+              (equal agda2-highlight-level 'none))
+    (agda2-go 'save nil t t
+              "Cmd_tokenHighlighting"
+              (agda2-string-quote (buffer-file-name))
+              "Keep")))
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Go to definition site
 
@@ -1925,9 +1928,9 @@ configuration and the selected window are not changed."
               (if windows
                   (dolist (window windows)
                     (with-selected-window window
-                      (annotation-goto-position (cdr filepos))))
+                      (goto-char (cdr filepos))))
                 (with-current-buffer buffer
-                  (annotation-goto-position (cdr filepos)))))))))))
+                  (goto-char (cdr filepos)))))))))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Implicit arguments
@@ -2024,12 +2027,10 @@ VERSION is empty, then agda and agda-mode are used instead.)"
     (unless (file-readable-p agda-mode-path)
       (error "%s" (concat "Could not read " agda-mode-path)))
 
-    ;; Kill some processes/buffers related to Agda.
-    (when (and agda2-process
-               (process-status agda2-process))
-      (kill-process agda2-process))
-    (when (buffer-live-p agda2-process-buffer)
-      (kill-buffer agda2-process-buffer))
+    ;; Turn off the Agda mode.
+    (agda2-quit)
+
+    ;; Kill some buffers related to Agda.
     (when (buffer-live-p agda2-info-buffer)
       (kill-buffer agda2-info-buffer))
     (when (and agda2-debug-buffer-name
