@@ -22,7 +22,7 @@ import qualified Agda.TypeChecking.Monad.Base as MB
 import Agda.TypeChecking.Monad.Signature (getConstInfo, getDefFreeVars, ignoreAbstractMode)
 import Agda.Utils.Permutation (Permutation(Perm), permute, takeP, compactP)
 import Agda.TypeChecking.Level (reallyUnLevelView)
-import Agda.TypeChecking.Monad.Base (mvJudgement, mvPermutation, getMetaInfo, ctxEntry, envContext, clEnv)
+import Agda.TypeChecking.Monad.Base (mvJudgement, mvPermutation, getMetaInfo, envContext, clEnv)
 import Agda.TypeChecking.Monad.MetaVars (lookupMeta, withMetaInfo, lookupInteractionPoint)
 import Agda.TypeChecking.Monad.Context (getContextArgs)
 import Agda.TypeChecking.Monad.Constraints (getAllConstraints)
@@ -125,7 +125,6 @@ tomy imi icns typs = do
       MB.Record {MB.recFields = fields, MB.recTel = tel} -> do -- the value of recPars seems unreliable or don't know what it signifies
        let pars n (I.El _ (I.Pi it typ)) = Cm.Arg (Cm.domInfo it) (I.var n) :
                                            pars (n - 1) (I.unAbs typ)
-           pars n (I.El s (I.Shared p))  = pars n (I.El s (I.derefPtr p))
            pars _ (I.El _ _) = []
            contyp npar I.EmptyTel = I.El (I.mkType 0 {- arbitrary -}) $
                                       I.Def cn $ map I.Apply $ pars (npar - 1) typ
@@ -173,7 +172,7 @@ tomy imi icns typs = do
          modify $ \s -> s {sEqs = (Map.insert (Map.size (fst $ sEqs s)) (Just (False, Meta m, sol')) (fst $ sEqs s), snd $ sEqs s)}
        let tt = MB.jMetaType $ mvJudgement mv
            minfo = getMetaInfo mv
-           localVars = map (snd . Cm.unDom . ctxEntry) . envContext . clEnv $ minfo
+           localVars = map (snd . Cm.unDom) . envContext . clEnv $ minfo
        (targettype, localVars) <- lift $ withMetaInfo minfo $ do
         vs <- getContextArgs
         targettype <- tt `piApplyM` permute (takeP (length vs) $ mvPermutation mv) vs
@@ -323,10 +322,9 @@ instance Conversion TOM I.Clause (Maybe ([Pat O], MExp O)) where
 
 instance Conversion TOM (Cm.Arg I.Pattern) (Pat O) where
   convert p = case Cm.unArg p of
-    I.VarP n    -> return $ PatVar (show n)
-    I.DotP _    -> return $ PatVar "_"
+    I.VarP _ n  -> return $ PatVar (show n)
+    I.DotP _ _  -> return $ PatVar "_"
       -- because Agda includes these when referring to variables in the body
-    I.AbsurdP{} -> return $ PatVar I.absurdPatternName
     I.ConP con _ pats -> do
       let n = I.conName con
       c     <- getConst True n TMAll
@@ -364,7 +362,8 @@ instance Conversion TOM I.Term (MExp O) where
         c   <- getConst False name TMAll
         as' <- convert as
         return $ NotM $ App Nothing (NotM OKVal) (Const c) as'
-      I.Con con ci as -> do
+      I.Con con ci es -> do
+        let Just as = I.allApplyElims es
         let name = I.conName con
         c   <- getConst True name TMAll
         as' <- convert as
@@ -395,7 +394,6 @@ instance Conversion TOM I.Term (MExp O) where
             return $ Meta m
           _ -> convert t
       I.DontCare _ -> return $ NotM dontCare
-      I.Shared p -> convert $ I.derefPtr p
 
 instance Conversion TOM a b => Conversion TOM (Cm.Arg a) (Hiding, b) where
   convert (Cm.Arg info a) = (getHiding info,) <$> convert a
@@ -419,12 +417,11 @@ fmExp m (I.Lam _ b) = fmExp m (I.unAbs b)
 fmExp m (I.Lit _) = False
 fmExp m (I.Level (I.Max as)) = any (fmLevel m) as
 fmExp m (I.Def _ as) = fmExps m $ I.argsFromElims as
-fmExp m (I.Con _ ci as) = fmExps m as
+fmExp m (I.Con _ ci as) = fmExps m $ I.argsFromElims as
 fmExp m (I.Pi x y)  = fmType m (Cm.unDom x) || fmType m (I.unAbs y)
 fmExp m (I.Sort _) = False
 fmExp m (I.MetaV mid _) = mid == m
 fmExp m (I.DontCare _) = False
-fmExp m (I.Shared p) = fmExp m $ I.derefPtr p
 
 fmExps :: I.MetaId -> I.Args -> Bool
 fmExps m [] = False
@@ -488,7 +485,7 @@ instance Conversion MOT (Exp O) I.Term where
 -}
           (ndrop, h) = case iscon of
                          Just n -> (n, \ q -> I.Con (I.ConHead q Cm.Inductive []) Cm.ConOSystem) -- TODO: restore fields
-                         Nothing -> (0, \ f vs -> I.Def f $ map I.Apply vs)
+                         Nothing -> (0, \ f vs -> I.Def f vs)
       frommyExps ndrop as (h name [])
     Lam hid t -> I.Lam (icnvh hid) <$> convert t
     Pi _ hid _ x y -> I.Pi . Cm.Dom (icnvh hid) <$> convert x <*> convert y
@@ -529,9 +526,8 @@ frommyExps ndrop (NotM as) trm =
   ALConPar _ -> __IMPOSSIBLE__
  where
   addend x (I.Var h xs) = I.Var h (xs ++ [I.Apply x])
-  addend x (I.Con h ci xs) = I.Con h ci (xs ++ [x])
+  addend x (I.Con h ci xs) = I.Con h ci (xs ++ [I.Apply x])
   addend x (I.Def h xs) = I.Def h (xs ++ [I.Apply x])
-  addend x (I.Shared p) = addend x (I.derefPtr p)
   addend _ _ = __IMPOSSIBLE__
 
 -- --------------------------------
@@ -544,7 +540,7 @@ modifyAbstractExpr = f
  where
   f (A.App i e1 (Cm.Arg info (Cm.Named n e2))) =
         A.App i (f e1) (Cm.Arg info (Cm.Named n (f e2)))
-  f (A.Lam i (A.DomainFree info n) _)
+  f (A.Lam i (A.DomainFree info (A.BindName n)) _)
      | prettyShow (A.nameConcrete n) == abslamvarname =
         A.AbsurdLam i $ Cm.argInfoHiding info
   f (A.Lam i b e) = A.Lam i b (f e)
@@ -554,8 +550,8 @@ modifyAbstractExpr = f
   f e = e
 
 modifyAbstractClause :: A.Clause -> A.Clause
-modifyAbstractClause (A.Clause lhs dots sdots (A.RHS e mc) decls catchall) =
-  A.Clause lhs dots sdots (A.RHS (modifyAbstractExpr e) mc) decls catchall
+modifyAbstractClause (A.Clause lhs spats (A.RHS e mc) decls catchall) =
+  A.Clause lhs spats (A.RHS (modifyAbstractExpr e) mc) decls catchall
 modifyAbstractClause cl = cl
 
 -- ---------------------------------
@@ -571,7 +567,7 @@ constructPats cmap mainm clause = do
      cnvp ns p =
       let hid = getHiding $ Cm.argInfo p
       in case Cm.namedArg p of
-       I.VarP n -> return ((hid, Id n) : ns, HI hid (CSPatVar $ length ns))
+       I.VarP _ n -> return ((hid, Id n) : ns, HI hid (CSPatVar $ length ns))
        I.ConP con _ ps -> do
         let c = I.conName con
         (c2, _) <- runStateT (getConst True c TMAll) (S {sConsts = (cmap, []), sMetas = initMapS, sEqs = initMapS, sCurMeta = Nothing, sMainMeta = mainm})
@@ -579,10 +575,9 @@ constructPats cmap mainm clause = do
         cc <- liftIO $ readIORef c2
         let Just npar = fst $ cdorigin cc
         return (ns', HI hid (CSPatConApp c2 (replicate npar (HI Hidden CSOmittedArg) ++ ps')))
-       I.DotP t -> do
+       I.DotP _ t -> do
         (t2, _) <- runStateT (convert t) (S {sConsts = (cmap, []), sMetas = initMapS, sEqs = initMapS, sCurMeta = Nothing, sMainMeta = mainm})
         return (ns, HI hid (CSPatExp t2))
-       I.AbsurdP{} -> return ((hid, Id I.absurdPatternName) : ns, HI hid (CSPatVar $ length ns))
        I.ProjP{} -> copatternsNotImplemented
        I.LitP{} -> literalsNotImplemented
  (names, pats) <- cnvps [] (IP.unnumberPatVars $ I.namedClausePats clause)
@@ -641,7 +636,7 @@ frommyClause (ids, pats, mrhs) = do
         return (I.ConP con I.noConPatternInfo ps')
        CSPatExp e -> do
         e' <- convert e {- renm e -} -- renaming before adding to clause below
-        return (I.DotP e')
+        return (I.dotP e')
        CSAbsurd -> __IMPOSSIBLE__ -- CSAbsurd not used
        _ -> __IMPOSSIBLE__
       return $ Cm.Arg (icnvh hid) $ Cm.unnamed p'   -- TODO: recover names
@@ -715,7 +710,7 @@ findClauseDeep ii = ignoreAbstractMode $ do  -- Andreas, 2016-09-04, issue #2162
       return $ Just (f, c, maybe __IMPOSSIBLE__ toplevel $ I.clauseBody c)
   where
     toplevel e =
-     case I.ignoreSharing e of
+     case e of
       I.MetaV{} -> True
       _ -> False
 
@@ -725,19 +720,19 @@ matchType :: Int -> Int -> I.Type -> I.Type -> Maybe (Nat, Nat) -- Nat is deffre
 matchType cdfv tctx ctyp ttyp = trmodps cdfv ctyp
  where
   trmodps 0 ctyp = tr 0 0 ctyp
-  trmodps n ctyp = case I.ignoreSharing $ I.unEl ctyp of
+  trmodps n ctyp = case I.unEl ctyp of
    I.Pi _ ot -> trmodps (n - 1) (I.absBody ot)
    _ -> __IMPOSSIBLE__
   tr narg na ctyp =
    case ft 0 0 Just ctyp ttyp of
     Just n -> Just (n, narg)
-    Nothing -> case I.ignoreSharing $ I.unEl ctyp of
+    Nothing -> case I.unEl ctyp of
      I.Pi _ (I.Abs _ ot) -> tr (narg + 1) (na + 1) ot
      I.Pi _ (I.NoAbs _ ot) -> tr (narg + 1) na ot
      _ -> Nothing
    where
     ft nl n c (I.El _ e1) (I.El _ e2) = f nl n c e1 e2
-    f nl n c e1 e2 = case I.ignoreSharing e1 of
+    f nl n c e1 e2 = case e1 of
      I.Var v1 as1 | v1 < nl -> case e2 of
       I.Var v2 as2 | v1 == v2 -> fes nl (n + 1) c as1 as2
       _ -> Nothing
@@ -745,7 +740,7 @@ matchType cdfv tctx ctyp ttyp = trmodps cdfv ctyp
      I.Var v1 as1 -> case e2 of
       I.Var v2 as2 | cdfv + na + nl - v1 == tctx + nl - v2 -> fes nl (n + 1) c as1 as2
       _ -> Nothing
-     _ -> case (I.ignoreSharing e1, I.ignoreSharing e2) of
+     _ -> case (e1, e2) of
       (I.MetaV{}, _) -> c n
       (_, I.MetaV{}) -> c n
       (I.Lam hid1 b1, I.Lam hid2 b2) | hid1 == hid2 -> f (nl + 1) n c (I.absBody b1) (I.absBody b2)
@@ -757,7 +752,7 @@ matchType cdfv tctx ctyp ttyp = trmodps cdfv ctyp
       _ -> Nothing
     fs nl n c es1 es2 = case (es1, es2) of
      ([], []) -> c n
-     (Cm.Arg info1 e1 : es1, Cm.Arg info2 e2 : es2) | Cm.argInfoHiding info1 == Cm.argInfoHiding info2 -> f nl n (\n -> fs nl n c es1 es2) e1 e2
+     (I.Apply (Cm.Arg info1 e1) : es1, I.Apply (Cm.Arg info2 e2) : es2) | Cm.argInfoHiding info1 == Cm.argInfoHiding info2 -> f nl n (\n -> fs nl n c es1 es2) e1 e2
      _ -> Nothing
     fes nl n c es1 es2 = case (es1, es2) of
      ([], []) -> c n
