@@ -3,7 +3,6 @@
 module Agda.Compiler.JS.Compiler where
 
 import Prelude hiding ( null, writeFile )
-import Control.Applicative
 import Control.Monad.Reader ( liftIO )
 import Control.Monad.Trans
 import Data.Char ( isSpace )
@@ -31,12 +30,12 @@ import Agda.Syntax.Internal
     conName,
     toTopLevelModuleName, arity, unEl, unAbs, nameFixity )
 import Agda.Syntax.Position
-import Agda.Syntax.Literal ( Literal(LitNat,LitFloat,LitString,LitChar,LitQName,LitMeta) )
+import Agda.Syntax.Literal ( Literal(..) )
 import Agda.Syntax.Fixity
 import qualified Agda.Syntax.Treeless as T
 import Agda.TypeChecking.Substitute ( absBody )
 import Agda.TypeChecking.Level ( reallyUnLevelView )
-import Agda.TypeChecking.Monad hiding (Global, Local)
+import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad.Debug ( reportSLn )
 import Agda.TypeChecking.Monad.Options ( setCommandLineOptions )
@@ -384,7 +383,7 @@ compileTerm' kit t = go t
       T.TCon q -> do
         d <- getConstInfo q
         qname q
-      T.TCase sc (T.CTData dt) def alts -> do
+      T.TCase sc ct def alts | T.CTData dt <- T.caseType ct -> do
         dt <- getConstInfo dt
         alts' <- traverse compileAlt alts
         let obj = Object $ Map.fromList alts'
@@ -404,6 +403,7 @@ compileTerm' kit t = go t
       T.TSort -> unit
       T.TErased -> unit
       T.TError T.TUnreachable -> return Undefined
+      T.TCoerce t -> go t
 
     unit = return $ Integer 0
 
@@ -414,7 +414,8 @@ compilePrim p =
     T.PEqI -> binOp "agdaRTS.uprimIntegerEqual"
     T.PEqF -> binOp "agdaRTS.uprimFloatEquality"
     T.PEqQ -> binOp "agdaRTS.uprimQNameEquality"
-    p | T.isPrimEq p -> curriedLambda 2 $ BinOp (local 1) "===" (local 0)
+    T.PEqS -> primEq
+    T.PEqC -> primEq
     T.PGeq -> binOp "agdaRTS.uprimIntegerGreaterOrEqualThan"
     T.PLt -> binOp "agdaRTS.uprimIntegerLessThan"
     T.PAdd -> binOp "agdaRTS.uprimIntegerPlus"
@@ -422,9 +423,19 @@ compilePrim p =
     T.PMul -> binOp "agdaRTS.uprimIntegerMultiply"
     T.PRem -> binOp "agdaRTS.uprimIntegerRem"
     T.PQuot -> binOp "agdaRTS.uprimIntegerQuot"
+    T.PAdd64 -> binOp "agdaRTS.uprimWord64Plus"
+    T.PSub64 -> binOp "agdaRTS.uprimWord64Minus"
+    T.PMul64 -> binOp "agdaRTS.uprimWord64Multiply"
+    T.PRem64 -> binOp "agdaRTS.uprimIntegerRem"     -- -|
+    T.PQuot64 -> binOp "agdaRTS.uprimIntegerQuot"   --  > These can use the integer functions
+    T.PEq64 -> binOp "agdaRTS.uprimIntegerEqual"    --  |
+    T.PLt64 -> binOp "agdaRTS.uprimIntegerLessThan" -- -|
+    T.PITo64 -> unOp "agdaRTS.primWord64FromNat"
+    T.P64ToI -> unOp "agdaRTS.primWord64ToNat"
     T.PSeq -> binOp "agdaRTS.primSeq"
-    _ -> __IMPOSSIBLE__
   where binOp js = curriedLambda 2 $ apply (PlainJS js) [local 1, local 0]
+        unOp js  = curriedLambda 1 $ apply (PlainJS js) [local 0]
+        primEq   = curriedLambda 2 $ BinOp (local 1) "===" (local 0)
 
 
 compileAlt :: T.TAlt -> TCM (MemberId, Exp)
@@ -452,6 +463,7 @@ qname q = do
 literal :: Literal -> Exp
 literal l = case l of
   (LitNat    _ x) -> Integer x
+  (LitWord64 _ x) -> Integer (fromIntegral x)
   (LitFloat  _ x) -> Double  x
   (LitString _ x) -> String  x
   (LitChar   _ x) -> Char    x
@@ -519,6 +531,7 @@ primitives = Set.fromList
   [ "primExp"
   , "primFloatDiv"
   , "primFloatEquality"
+  , "primFloatLess"
   , "primFloatNumericalEquality"
   , "primFloatNumericalLess"
   , "primFloatNegate"
@@ -540,4 +553,6 @@ primitives = Set.fromList
   , "primQNameEquality"
   , "primQNameLess"
   , "primQNameFixity"
+  , "primWord64ToNat"
+  , "primWord64FromNat"
   ]
