@@ -2,11 +2,10 @@
 
 module Agda.Compiler.MAlonzo.Compiler where
 
-#if __GLASGOW_HASKELL__ <= 708
-import Prelude hiding (foldl, mapM_, mapM, sequence, concat)
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ((<>))
 #endif
 
-import Control.Applicative
 import Control.Monad.Reader hiding (mapM_, forM_, mapM, forM, sequence)
 import Control.Monad.State  hiding (mapM_, forM_, mapM, forM, sequence)
 
@@ -31,6 +30,7 @@ import System.FilePath hiding (normalise)
 
 import Agda.Compiler.CallCompiler
 import Agda.Compiler.Common
+import Agda.Compiler.MAlonzo.Coerce
 import Agda.Compiler.MAlonzo.Misc
 import Agda.Compiler.MAlonzo.Pretty
 import Agda.Compiler.MAlonzo.Primitives
@@ -57,12 +57,12 @@ import Agda.Syntax.Literal
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Datatypes
+import Agda.TypeChecking.Primitive (getBuiltinName)
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
-import Agda.TypeChecking.Level (reallyUnLevelView)
 import Agda.TypeChecking.Warnings
 
 import Agda.TypeChecking.CompiledClause
@@ -79,6 +79,7 @@ import qualified Agda.Utils.IO.UTF8 as UTF8
 import qualified Agda.Utils.HashMap as HMap
 import Agda.Utils.Singleton
 import Agda.Utils.Size
+import Agda.Utils.String
 import Agda.Utils.Tuple
 
 import Paths_Agda
@@ -148,7 +149,9 @@ ghcPostCompile opts isMain mods = copyRTEModules >> callGHC opts isMain mods
 
 --- Module compilation ---
 
-type GHCModuleEnv = Maybe CoinductionKit
+-- | This environment is no longer used for anything.
+
+type GHCModuleEnv = ()
 
 ghcPreModule :: GHCOptions -> ModuleName -> FilePath -> TCM (Recompile GHCModuleEnv IsMain)
 ghcPreModule _ m ifile = ifM uptodate noComp yesComp
@@ -164,7 +167,7 @@ ghcPreModule _ m ifile = ifM uptodate noComp yesComp
       out <- outFile_
       reportSLn "compile.ghc" 1 $ repl [m, ifile, out] "Compiling <<0>> in <<1>> to <<2>>"
       stImportedModules .= Set.empty  -- we use stImportedModules to accumulate the required Haskell imports
-      Recompile <$> coinductionKit
+      return (Recompile ())
 
 ghcPostModule :: GHCOptions -> GHCModuleEnv -> IsMain -> ModuleName -> [[HS.Decl]] -> TCM IsMain
 ghcPostModule _ _ _ _ defs = do
@@ -197,8 +200,12 @@ imports = (hsImps ++) <$> imps where
   unqualRTE :: HS.ImportDecl
   unqualRTE = HS.ImportDecl mazRTE False $ Just $
               (False, [ HS.IVar $ HS.Ident x
-                      | x <- [mazCoerceName, mazErasedName] ++
-                             map treelessPrimName [T.PAdd, T.PSub, T.PMul, T.PQuot, T.PRem, T.PGeq, T.PLt, T.PEqI, T.PEqF] ])
+                      | x <- [mazCoerceName, mazErasedName, mazAnyTypeName] ++
+                             map treelessPrimName rtePrims ])
+
+  rtePrims = [T.PAdd, T.PSub, T.PMul, T.PQuot, T.PRem, T.PGeq, T.PLt, T.PEqI, T.PEqF,
+              T.PAdd64, T.PSub64, T.PMul64, T.PQuot64, T.PRem64, T.PLt64, T.PEq64,
+              T.PITo64, T.P64ToI]
 
   imps :: TCM [HS.ImportDecl]
   imps = List.map decl . uniq <$>
@@ -217,39 +224,32 @@ imports = (hsImps ++) <$> imps where
 -- Main compiling clauses
 --------------------------------------------------
 
--- | Note that the INFINITY, SHARP and FLAT builtins are translated as
--- follows (if a 'CoinductionKit' is given):
---
--- @
---   type Infinity a b = b
---
---   sharp :: a -> a
---   sharp x = x
---
---   flat :: a -> a
---   flat x = x
--- @
-
-definition :: Maybe CoinductionKit -> Definition -> TCM [HS.Decl]
+definition :: GHCModuleEnv -> Definition -> TCM [HS.Decl]
 -- ignore irrelevant definitions
 {- Andreas, 2012-10-02: Invariant no longer holds
-definition kit (Defn Forced{}  _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
-definition kit (Defn UnusedArg _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
 definition kit (Defn NonStrict _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
 -}
-definition kit Defn{defArgInfo = info, defName = q} | isIrrelevant info = do
+definition env Defn{defArgInfo = info, defName = q} | isIrrelevant info = do
   reportSDoc "compile.ghc.definition" 10 $
     text "Not compiling" <+> prettyTCM q <> text "."
   return []
-definition kit Defn{defName = q, defType = ty, theDef = d} = do
+definition env Defn{defName = q, defType = ty, theDef = d} = do
   reportSDoc "compile.ghc.definition" 10 $ vcat
     [ text "Compiling" <+> prettyTCM q <> text ":"
     , nest 2 $ text (show d)
     ]
   pragma <- getHaskellPragma q
+  mbool  <- getBuiltinName builtinBool
+  mlist  <- getBuiltinName builtinList
+  minf   <- getBuiltinName builtinInf
+  mflat  <- getBuiltinName builtinFlat
   checkTypeOfMain q ty $ do
     infodecl q <$> case d of
 
+      _ | Just HsDefn{} <- pragma, Just q == mflat ->
+        genericError
+          "\"COMPILE GHC\" pragmas are not allowed for the FLAT builtin."
+
       _ | Just (HsDefn r hs) <- pragma -> setCurrentRange r $ do
         -- Make sure we have imports for all names mentioned in the type.
         hsty <- haskellType q
@@ -263,43 +263,45 @@ definition kit Defn{defName = q, defType = ty, theDef = d} = do
 
         return $ fbWithType hsty (fakeExp hs)
 
-      -- Special treatment of coinductive builtins.
-      Datatype{} | Just q == (nameOfInf <$> kit) -> do
-        let infT = unqhname "T" q
-            infV = unqhname "d" q
-            a    = ihname "a" 0
-            b    = ihname "a" 1
-            vars = [a, b]
-        return [ HS.TypeDecl infT
-                             (List.map HS.UnkindedVar vars)
-                             (HS.TyVar b)
-               , HS.FunBind [HS.Match infV
-                                      (List.map HS.PVar vars)
-                                      (HS.UnGuardedRhs HS.unit_con)
-                                      emptyBinds]
-               ]
-      Constructor{} | Just q == (nameOfSharp <$> kit) -> do
-        let sharp = unqhname "d" q
-            x     = ihname "x" 0
-        return $
-          [ HS.TypeSig [sharp] $ fakeType $
-              "forall a. a -> a"
-          , HS.FunBind [HS.Match sharp
-                                 [HS.PVar x]
-                                 (HS.UnGuardedRhs (HS.Var (HS.UnQual x)))
-                                 emptyBinds]
-          ]
-      Function{} | Just q == (nameOfFlat <$> kit) -> do
-        let flat = unqhname "d" q
-            x    = ihname "x" 0
-        return $
-          [ HS.TypeSig [flat] $ fakeType $
-              "forall a. a -> a"
-          , HS.FunBind [HS.Match flat
-                                 [HS.PVar x]
-                                 (HS.UnGuardedRhs (HS.Var (HS.UnQual x)))
-                                 emptyBinds]
-          ]
+      -- Compiling Bool
+      Datatype{} | Just q == mbool -> do
+        _ <- sequence_ [primTrue, primFalse] -- Just to get the proper error for missing TRUE/FALSE
+        let d = unqhname "d" q
+        Just true  <- getBuiltinName builtinTrue
+        Just false <- getBuiltinName builtinFalse
+        cs <- mapM compiledcondecl [false, true]
+        return $ [ compiledTypeSynonym q "Bool" 0
+                 , HS.FunBind [HS.Match d [] (HS.UnGuardedRhs HS.unit_con) emptyBinds] ] ++
+                 cs
+
+      -- Compiling List
+      Datatype{ dataPars = np } | Just q == mlist -> do
+        _ <- sequence_ [primNil, primCons] -- Just to get the proper error for missing NIL/CONS
+        caseMaybe pragma (return ()) $ \ p -> setCurrentRange p $ warning . GenericWarning =<< do
+          fsep $ pwords "Ignoring GHC pragma for builtin lists; they always compile to Haskell lists."
+        let d = unqhname "d" q
+            t = unqhname "T" q
+        Just nil  <- getBuiltinName builtinNil
+        Just cons <- getBuiltinName builtinCons
+        let vars f n = map (f . ihname "a") [0 .. n - 1]
+        cs <- mapM compiledcondecl [nil, cons]
+        return $ [ HS.TypeDecl t (vars HS.UnkindedVar (np - 1)) (HS.FakeType "[]")
+                 , HS.FunBind [HS.Match d (vars HS.PVar np) (HS.UnGuardedRhs HS.unit_con) emptyBinds] ] ++
+                 cs
+
+      -- Compiling Inf
+      _ | Just q == minf -> do
+        _ <- primSharp -- To get a proper error for missing SHARP.
+        Just sharp <- getBuiltinName builtinSharp
+        sharpC     <- compiledcondecl sharp
+        let d   = unqhname "d" q
+            err = "No term-level implementation of the INFINITY builtin."
+        return $ [ compiledTypeSynonym q "MAlonzo.RTE.Infinity" 2
+                 , HS.FunBind [HS.Match d [HS.PVar (ihname "a" 0)]
+                     (HS.UnGuardedRhs (HS.FakeExp ("error " ++ show err)))
+                     emptyBinds]
+                 , sharpC
+                 ]
 
       Axiom{} -> do
         ar <- typeArity ty
@@ -314,41 +316,49 @@ definition kit Defn{defName = q, defType = ty, theDef = d} = do
         computeErasedConstructorArgs q
         ccscov <- constructorCoverageCode q (np + ni) cs ty hsCons
         cds <- mapM compiledcondecl cs
-        return $ tvaldecl q (dataInduction d) 0 (np + ni) [] (Just __IMPOSSIBLE__) ++
+        return $ tvaldecl q (dataInduction d) (np + ni) [] (Just __IMPOSSIBLE__) ++
                  [compiledTypeSynonym q ty np] ++ cds ++ ccscov
-      Datatype{ dataPars = np, dataIxs = ni, dataClause = cl, dataCons = cs } -> do
+      Datatype{ dataPars = np, dataIxs = ni, dataClause = cl,
+                dataCons = cs, dataInduction = ind } -> do
         computeErasedConstructorArgs q
-        (ars, cds) <- unzip <$> mapM condecl cs
-        return $ tvaldecl q (dataInduction d) (List.maximum (np:ars) - np) (np + ni) cds cl
+        cds <- mapM (flip condecl ind) cs
+        return $ tvaldecl q ind (np + ni) cds cl
       Constructor{} -> return []
-      Record{ recPars = np, recClause = cl, recConHead = con }
-        | Just (HsData r ty hsCons) <- pragma -> setCurrentRange r $ do
-        let cs = [conName con]
-        computeErasedConstructorArgs q
-        ccscov <- constructorCoverageCode q np cs ty hsCons
-        cds <- mapM compiledcondecl cs
-        return $ tvaldecl q Inductive 0 np [] (Just __IMPOSSIBLE__) ++
-                 [compiledTypeSynonym q ty np] ++ cds ++ ccscov
-      Record{ recClause = cl, recConHead = con, recFields = flds } -> do
-        computeErasedConstructorArgs q
-        let c = conName con
-        let noFields = length flds
-        let ar = I.arity ty
-        cd <- snd <$> condecl c
-        return $ tvaldecl q Inductive noFields ar [cd] cl
+      Record{ recPars = np, recClause = cl, recConHead = con,
+              recInduction = ind } ->
+        let -- Non-recursive record types are treated as being
+            -- inductive.
+            inductionKind = fromMaybe Inductive ind
+        in case pragma of
+          Just (HsData r ty hsCons) -> setCurrentRange r $ do
+            let cs = [conName con]
+            computeErasedConstructorArgs q
+            ccscov <- constructorCoverageCode q np cs ty hsCons
+            cds <- mapM compiledcondecl cs
+            return $
+              tvaldecl q inductionKind np [] (Just __IMPOSSIBLE__) ++
+              [compiledTypeSynonym q ty np] ++ cds ++ ccscov
+          _ -> do
+            computeErasedConstructorArgs q
+            cd <- condecl (conName con) inductionKind
+            return $ tvaldecl q inductionKind (I.arity ty) [cd] cl
       AbstractDefn{} -> __IMPOSSIBLE__
   where
   function :: Maybe HaskellPragma -> TCM [HS.Decl] -> TCM [HS.Decl]
   function mhe fun = do
-    ccls <- mkwhere <$> fun
+    ccls  <- mkwhere <$> fun
+    mflat <- getBuiltinName builtinFlat
     case mhe of
+      Just HsExport{} | Just q == mflat ->
+        genericError
+          "\"COMPILE GHC as\" pragmas are not allowed for the FLAT builtin."
       Just (HsExport r name) -> do
         t <- setCurrentRange r $ haskellType q
         let tsig :: HS.Decl
-            tsig = HS.TypeSig [HS.Ident name] (fakeType t)
+            tsig = HS.TypeSig [HS.Ident name] t
 
             def :: HS.Decl
-            def = HS.FunBind [HS.Match (HS.Ident name) [] (HS.UnGuardedRhs (hsVarUQ $ dname q)) emptyBinds]
+            def = HS.FunBind [HS.Match (HS.Ident name) [] (HS.UnGuardedRhs (hsCoerce $ hsVarUQ $ dname q)) emptyBinds]
         return ([tsig,def] ++ ccls)
       _ -> return ccls
 
@@ -358,29 +368,37 @@ definition kit Defn{defName = q, defType = ty, theDef = d} = do
     used <- getCompiledArgUse q
     let dostrip = any not used
 
+    -- Compute the type approximation
+    def <- getConstInfo q
+    (argTypes0, resType) <- hsTelApproximation $ defType def
+    let pars = case theDef def of
+                 Function{ funProjection = Just Projection{ projIndex = i } } | i > 0 -> i - 1
+                 _ -> 0
+        argTypes  = drop pars argTypes0
+        argTypesS = [ t | (t, True) <- zip argTypes (used ++ repeat True) ]
+
     e <- if dostrip then closedTerm (stripUnusedArguments used treeless)
                     else closedTerm treeless
     let (ps, b) = lamView e
         lamView e =
-          case stripTopCoerce e of
+          case e of
             HS.Lambda ps b -> (ps, b)
-            b                -> ([], b)
-        stripTopCoerce (HS.Lambda ps b) = HS.Lambda ps $ stripTopCoerce b
-        stripTopCoerce e =
-          case hsAppView e of
-            [c,  e] | c == mazCoerce -> e
-            _                        -> e
+            b              -> ([], b)
 
+        tydecl  f ts t = HS.TypeSig [f] (foldr HS.TyFun t ts)
         funbind f ps b = HS.FunBind [HS.Match f ps (HS.UnGuardedRhs b) emptyBinds]
+        tyfunbind f ts t ps b =
+          let ts' = ts ++ (replicate (length ps - length ts) mazAnyType)
+          in [tydecl f ts' t, funbind f ps b]
 
     -- The definition of the non-stripped function
     (ps0, _) <- lamView <$> closedTerm (foldr ($) T.TErased $ replicate (length used) T.TLam)
     let b0 = foldl HS.App (hsVarUQ $ duname q) [ hsVarUQ x | (~(HS.PVar x), True) <- zip ps0 used ]
 
     return $ if dostrip
-      then [ funbind (dname q) ps0 b0
-           , funbind (duname q) ps b ]
-      else [ funbind (dname q) ps b ]
+      then tyfunbind (dname q) argTypes resType ps0 b0 ++
+           tyfunbind (duname q) argTypesS resType ps b
+      else tyfunbind (dname q) argTypes resType ps b
 
   mkwhere :: [HS.Decl] -> [HS.Decl]
   mkwhere (HS.FunBind [m0, HS.Match dn ps rhs emptyBinds] : fbs@(_:_)) =
@@ -391,9 +409,9 @@ definition kit Defn{defName = q, defType = ty, theDef = d} = do
 
   mkwhere fbs = fbs
 
-  fbWithType :: HaskellType -> HS.Exp -> [HS.Decl]
+  fbWithType :: HS.Type -> HS.Exp -> [HS.Decl]
   fbWithType ty e =
-    [ HS.TypeSig [unqhname "d" q] $ fakeType ty ] ++ fb e
+    [ HS.TypeSig [unqhname "d" q] ty ] ++ fb e
 
   fb :: HS.Exp -> [HS.Decl]
   fb e  = [HS.FunBind [HS.Match (unqhname "d" q) []
@@ -413,24 +431,24 @@ constructorCoverageCode q np cs hsTy hsCons = do
 -- | Environment for naming of local variables.
 --   Invariant: @reverse ccCxt ++ ccNameSupply@
 data CCEnv = CCEnv
-  { ccNameSupply :: NameSupply  -- ^ Supply of fresh names
-  , ccCxt        :: CCContext   -- ^ Names currently in scope
+  { _ccNameSupply :: NameSupply  -- ^ Supply of fresh names
+  , _ccContext    :: CCContext   -- ^ Names currently in scope
   }
 
 type NameSupply = [HS.Name]
 type CCContext  = [HS.Name]
 
-mapNameSupply :: (NameSupply -> NameSupply) -> CCEnv -> CCEnv
-mapNameSupply f e = e { ccNameSupply = f (ccNameSupply e) }
+ccNameSupply :: Lens' NameSupply CCEnv
+ccNameSupply f e =  (\ ns' -> e { _ccNameSupply = ns' }) <$> f (_ccNameSupply e)
 
-mapContext :: (CCContext -> CCContext) -> CCEnv -> CCEnv
-mapContext f e = e { ccCxt = f (ccCxt e) }
+ccContext :: Lens' CCContext CCEnv
+ccContext f e = (\ cxt -> e { _ccContext = cxt }) <$> f (_ccContext e)
 
 -- | Initial environment for expression generation.
 initCCEnv :: CCEnv
 initCCEnv = CCEnv
-  { ccNameSupply = map (ihname "v") [0..]  -- DON'T CHANGE THESE NAMES!
-  , ccCxt        = []
+  { _ccNameSupply = map (ihname "v") [0..]  -- DON'T CHANGE THESE NAMES!
+  , _ccContext    = []
   }
 
 -- | Term variables are de Bruijn indices.
@@ -442,18 +460,18 @@ type CC = ReaderT CCEnv TCM
 freshNames :: Int -> ([HS.Name] -> CC a) -> CC a
 freshNames n _ | n < 0 = __IMPOSSIBLE__
 freshNames n cont = do
-  (xs, rest) <- splitAt n <$> asks ccNameSupply
-  local (mapNameSupply (const rest)) $ cont xs
+  (xs, rest) <- splitAt n <$> view ccNameSupply
+  local (over ccNameSupply (const rest)) $ cont xs
 
 -- | Introduce n variables into the context.
 intros :: Int -> ([HS.Name] -> CC a) -> CC a
 intros n cont = freshNames n $ \xs ->
-  local (mapContext (reverse xs ++)) $ cont xs
+  local (over ccContext (reverse xs ++)) $ cont xs
 
 checkConstructorType :: QName -> HaskellCode -> TCM [HS.Decl]
 checkConstructorType q hs = do
   ty <- haskellType q
-  return [ HS.TypeSig [unqhname "check" q] $ fakeType ty
+  return [ HS.TypeSig [unqhname "check" q] ty
          , HS.FunBind [HS.Match (unqhname "check" q) []
                                 (HS.UnGuardedRhs $ fakeExp hs) emptyBinds]
          ]
@@ -477,17 +495,34 @@ checkCover q ty n cs hsCons = do
          ]
 
 closedTerm :: T.TTerm -> TCM HS.Exp
-closedTerm v = hsCast <$> term v `runReaderT` initCCEnv
+closedTerm v = do
+  v <- addCoercions v
+  term v `runReaderT` initCCEnv
+
+-- Translate case on bool to if
+mkIf :: T.TTerm -> CC T.TTerm
+mkIf t@(TCase e _ d [TACon c1 0 b1, TACon c2 0 b2]) | T.isUnreachable d = do
+  mTrue  <- lift $ getBuiltinName builtinTrue
+  mFalse <- lift $ getBuiltinName builtinFalse
+  let isTrue  c = Just c == mTrue
+      isFalse c = Just c == mFalse
+  if | isTrue c1, isFalse c2 -> return $ T.tIfThenElse (TCoerce $ TVar e) b1 b2
+     | isTrue c2, isFalse c1 -> return $ T.tIfThenElse (TCoerce $ TVar e) b2 b1
+     | otherwise             -> return t
+mkIf t = return t
 
 -- | Extract Agda term to Haskell expression.
 --   Erased arguments are extracted as @()@.
 --   Types are extracted as @()@.
 term :: T.TTerm -> CC HS.Exp
-term tm0 = case tm0 of
+term tm0 = mkIf tm0 >>= \ tm0 -> case tm0 of
   T.TVar i -> do
-    x <- lookupIndex i <$> asks ccCxt
+    x <- lookupIndex i <$> view ccContext
     return $ hsVarUQ x
-  T.TApp (T.TDef f) ts -> do
+  T.TApp (T.TPrim T.PIf) [c, x, y] -> HS.If <$> term c
+                                            <*> term x
+                                            <*> term y
+  T.TApp t ts | Just (coe, f) <- coerceView t -> do
     used <- lift $ getCompiledArgUse f
     isCompiled <- lift $ isJust <$> getHaskellPragma f  -- #2248: no unused argument pruning for COMPILE'd functions
     let given   = length ts
@@ -495,39 +530,36 @@ term tm0 = case tm0 of
         missing = drop given used
     if not isCompiled && any not used
       then if any not missing then term (etaExpand (needed - given) tm0) else do
-        f <- lift $ HS.Var <$> xhqn "du" f  -- used stripped function
+        f <- lift $ coe . HS.Var <$> xhqn "du" f  -- used stripped function
         f `apps` [ t | (t, True) <- zip ts $ used ++ repeat True ]
       else do
         t' <- term (T.TDef f)
-        t' `apps` ts
-  T.TApp (T.TCon c) ts -> do
-    kit <- lift coinductionKit
-    if Just c == (nameOfSharp <$> kit)
-      then do
-        t' <- HS.Var <$> lift (xhqn "d" c)
-        apps t' ts
-      else do
-        (ar, _) <- lift $ conArityAndPars c
-        erased  <- lift $ getErasedConArgs c
-        let missing = drop (length ts) erased
-            notErased = not
-        case all notErased missing of
-          False -> term $ etaExpand (length missing) tm0
-          True  -> do
-            f <- lift $ HS.Con <$> conhqn c
-            f `apps` [ t | (t, False) <- zip ts erased ]
+        coe t' `apps` ts
+    where coerceView (T.TCoerce (T.TDef f)) = Just (hsCoerce, f)
+          coerceView (T.TDef f)             = Just (id, f)
+          coerceView _                      = Nothing
+  T.TApp (T.TCon c) ts -> do  -- Note that constructors are never coerced
+    (ar, _) <- lift $ conArityAndPars c
+    erased  <- lift $ getErasedConArgs c
+    let missing = drop (length ts) erased
+        notErased = not
+    case all notErased missing of
+      False -> term $ etaExpand (length missing) tm0
+      True  -> do
+        f <- lift $ HS.Con <$> conhqn c
+        f `apps` [ t | (t, False) <- zip ts erased ]
   T.TApp t ts -> do
     t' <- term t
     t' `apps` ts
   T.TLam at -> do
-    (nm:_) <- asks ccNameSupply
+    (nm:_) <- view ccNameSupply
     intros 1 $ \ [x] ->
       hsLambda [HS.PVar x] <$> term at
   T.TLet t1 t2 -> do
     t1' <- term t1
     intros 1 $ \[x] -> do
       t2' <- term t2
-      return $ hsLet x (hsCast t1') t2'
+      return $ hsLet x t1' t2'
 
   T.TCase sc ct def alts -> do
     sc' <- term (T.TVar sc)
@@ -535,7 +567,7 @@ term tm0 = case tm0 of
     def' <- term def
     let defAlt = HS.Alt HS.PWildCard (HS.UnGuardedRhs def') emptyBinds
 
-    return $ HS.Case (hsCast sc') (alts' ++ [defAlt])
+    return $ HS.Case (hsCoerce sc') (alts' ++ [defAlt])
 
   T.TLit l -> return $ literal l
   T.TDef q -> do
@@ -544,6 +576,7 @@ term tm0 = case tm0 of
   T.TPrim p  -> return $ compilePrim p
   T.TUnit    -> return HS.unit_con
   T.TSort    -> return HS.unit_con
+  T.TCoerce e -> hsCoerce <$> term e
   T.TErased  -> return $ hsVarUQ $ HS.Ident mazErasedName
   T.TError e -> return $ case e of
     T.TUnreachable ->  rtmUnreachableError
@@ -553,6 +586,8 @@ term tm0 = case tm0 of
                 (T.mkTApp (raise n t) [T.TVar i | i <- [n - 1, n - 2..0]])
                 (replicate n ())
 
+hsCoerce :: HS.Exp -> HS.Exp
+hsCoerce t = HS.App mazCoerce t
 
 compilePrim :: T.TPrim -> HS.Exp
 compilePrim s = HS.Var $ hsName $ treelessPrimName s
@@ -563,7 +598,12 @@ alt sc a = do
     T.TACon {T.aCon = c} -> do
       intros (T.aArity a) $ \ xs -> do
         erased <- lift $ getErasedConArgs c
-        hConNm <- lift $ conhqn c
+        nil  <- lift $ getBuiltinName builtinNil
+        cons <- lift $ getBuiltinName builtinCons
+        hConNm <-
+          if | Just c == nil  -> return $ HS.UnQual $ HS.Ident "[]"
+             | Just c == cons -> return $ HS.UnQual $ HS.Symbol ":"
+             | otherwise      -> lift $ conhqn c
         mkAlt (HS.PApp hConNm $ map HS.PVar [ x | (x, False) <- zip xs erased ])
     T.TAGuard g b -> do
       g <- term g
@@ -578,22 +618,25 @@ alt sc a = do
   where
     mkGuarded eq lit b = do
       b  <- term b
-      sc <- term (T.TVar sc)
-      let guard =
+      let varName = HS.Ident "l" -- only used locally in the guard
+          pv = HS.PVar varName
+          v  = hsVarUQ varName
+          guard =
             HS.Var (HS.UnQual (HS.Ident eq)) `HS.App`
-              sc `HS.App` lit
-      return $ HS.Alt HS.PWildCard
+              v `HS.App` lit
+      return $ HS.Alt pv
                       (HS.GuardedRhss [HS.GuardedRhs [HS.Qualifier guard] b])
                       emptyBinds
 
     mkAlt :: HS.Pat -> CC HS.Alt
     mkAlt pat = do
         body' <- term $ T.aBody a
-        return $ HS.Alt pat (HS.UnGuardedRhs $ hsCast body') emptyBinds
+        return $ HS.Alt pat (HS.UnGuardedRhs body') emptyBinds
 
 literal :: Literal -> HS.Exp
 literal l = case l of
   LitNat    _ _   -> typed "Integer"
+  LitWord64 _ _   -> typed "MAlonzo.RTE.Word64"
   LitFloat  _ x   -> floatExp x "Double"
   LitQName  _ x   -> litqname x
   LitString _ s   -> litString s
@@ -620,6 +663,7 @@ literal l = case l of
 
 hslit :: Literal -> HS.Literal
 hslit l = case l of LitNat    _ x -> HS.Int    x
+                    LitWord64 _ x -> HS.Int    (fromIntegral x)
                     LitFloat  _ x -> HS.Frac   (toRational x)
                     LitChar   _ x -> HS.Char   x
                     LitQName  _ x -> __IMPOSSIBLE__
@@ -668,20 +712,21 @@ erasedArity q = do
   erased  <- length . filter id <$> getErasedConArgs q
   return (ar - erased)
 
-condecl :: QName -> TCM (Nat, HS.ConDecl)
-condecl q = do
-  (ar, np) <- conArityAndPars q
-  erased   <- length . filter id <$> getErasedConArgs q
-  let ar' = ar - erased
-  return $ (ar' + np, cdecl q ar')
-
-cdecl :: QName -> Nat -> HS.ConDecl
-cdecl q n = HS.ConDecl (unqhname "C" q)
-            [ HS.TyVar $ ihname "a" i | i <- [0 .. n - 1] ]
+condecl :: QName -> Induction -> TCM HS.ConDecl
+condecl q _ind = do
+  def <- getConstInfo q
+  let Constructor{ conPars = np, conErased = erased } = theDef def
+  (argTypes0, _) <- hsTelApproximation (defType def)
+  let argTypes   = [ (Just HS.Lazy, t)
+                     -- Currently all constructors are lazy.
+                   | (t, False) <- zip (drop np argTypes0)
+                                       (erased ++ repeat False)
+                   ]
+  return $ HS.ConDecl (unqhname "C" q) argTypes
 
 compiledcondecl :: QName -> TCM HS.Decl
 compiledcondecl q = do
-  (ar, np) <- conArityAndPars q
+  ar <- erasedArity q
   hsCon <- fromMaybe __IMPOSSIBLE__ <$> getHaskellConstructor q
   let patVars = map (HS.PVar . ihname "a") [0 .. ar - 1]
   return $ HS.PatSyn (HS.PApp (HS.UnQual $ unqhname "C" q) patVars) (HS.PApp (hsName hsCon) patVars)
@@ -696,75 +741,28 @@ compiledTypeSynonym q hsT arity =
 tvaldecl :: QName
          -> Induction
             -- ^ Is the type inductive or coinductive?
-         -> Nat -> Nat -> [HS.ConDecl] -> Maybe Clause -> [HS.Decl]
-tvaldecl q ind ntv npar cds cl =
+         -> Nat -> [HS.ConDecl] -> Maybe Clause -> [HS.Decl]
+tvaldecl q ind npar cds cl =
   HS.FunBind [HS.Match vn pvs (HS.UnGuardedRhs HS.unit_con) emptyBinds] :
-  maybe [HS.DataDecl kind tn tvs cds []]
+  maybe [HS.DataDecl kind tn [] cds' []]
         (const []) cl
   where
   (tn, vn) = (unqhname "T" q, unqhname "d" q)
-  tvs = [ HS.UnkindedVar $ ihname "a" i | i <- [0 .. ntv  - 1]]
   pvs = [ HS.PVar        $ ihname "a" i | i <- [0 .. npar - 1]]
 
   -- Inductive data types consisting of a single constructor with a
   -- single argument are translated into newtypes.
-  kind = case (ind, cds) of
-    (Inductive, [HS.ConDecl _ [_]]) -> HS.NewType
-    _                               -> HS.DataType
+  (kind, cds') = case (ind, cds) of
+    (Inductive, [HS.ConDecl c [(_, t)]]) ->
+      (HS.NewType, [HS.ConDecl c [(Nothing, t)]])
+      -- The strictness annotations are removed for newtype
+      -- constructors.
+    _ -> (HS.DataType, cds)
 
 infodecl :: QName -> [HS.Decl] -> [HS.Decl]
 infodecl _ [] = []
-infodecl q ds = fakeD (unqhname "name" q) (show $ prettyShow q) : ds
-
---------------------------------------------------
--- Inserting unsafeCoerce
---------------------------------------------------
-
-hsCast :: HS.Exp -> HS.Exp
-{-
-hsCast = addcast . go where
-  addcast [e@(HS.Var(HS.UnQual(HS.Ident(c:ns))))] | c == 'v' && all isDigit ns = e
-  addcast es = foldl HS.App mazCoerce es
-  -- this need to be extended if you generate other kinds of exps.
-  go (HS.App e1 e2    ) = go e1 ++ [hsCast e2]
-  go (HS.Lambda _ ps e) = [ HS.Lambda ps (hsCast e) ]
-  go e = [e]
--}
-
--- TODO: what's the specification for hsCast, hsCast' and hsCoerce???
-hsCast e = hsCoerce (hsCast' e)
-
-hsCast' :: HS.Exp -> HS.Exp
-hsCast' (HS.InfixApp e1 op e2) = hsCoerce $ HS.InfixApp (hsCast' e1) op (hsCast' e2)
-hsCast' (HS.Lambda ps e)       = HS.Lambda ps $ hsCast' e
-hsCast' (HS.Let bs e)          = HS.Let bs $ hsCast' e
-hsCast' (HS.Case sc alts)      = HS.Case (hsCast' sc) (map (hsMapAlt hsCast') alts)
-hsCast' e =
-  case hsAppView e of
-    f : es -> foldl HS.App (hsCoerce f) (map hsCastApp es)
-    _      -> __IMPOSSIBLE__
-
--- We still have to coerce function applications in arguments to coerced
--- functions.
-hsCastApp :: HS.Exp -> HS.Exp
-hsCastApp (HS.Lambda ps b) = HS.Lambda ps (hsCastApp b)
-hsCastApp (HS.Let bs e) = HS.Let bs $ hsCastApp e
-hsCastApp (HS.Case sc bs) = HS.Case (hsCastApp sc) (map (hsMapAlt hsCastApp) bs)
-hsCastApp (HS.InfixApp e1 op e2) = HS.InfixApp (hsCastApp e1) op (hsCastApp e2)
-hsCastApp e =
-  case hsAppView e of
-    f : es@(_:_) -> foldl HS.App (hsCoerce f) $ map hsCastApp es
-    _ -> e
-
--- No coercion for literal integers
-hsCoerce :: HS.Exp -> HS.Exp
-hsCoerce e@(HS.ExpTypeSig (HS.Lit (HS.Int{})) _) = e
-hsCoerce (HS.Case sc alts) = HS.Case sc (map (hsMapAlt hsCoerce) alts)
-hsCoerce (HS.Let bs e) = HS.Let bs $ hsCoerce e
-hsCoerce e =
-  case hsAppView e of
-    c : _ | c == mazCoerce || c == mazIncompleteMatch -> e
-    _ -> mazCoerce `HS.App` e
+infodecl q ds =
+  fakeD (unqhname "name" q) (haskellStringLiteral $ prettyShow q) : ds
 
 --------------------------------------------------
 -- Writing out a haskell module
@@ -780,7 +778,7 @@ writeModule :: HS.Module -> TCM ()
 writeModule (HS.Module m ps imp ds) = do
   -- Note that GHC assumes that sources use ASCII or UTF-8.
   out <- outFile m
-  liftIO $ UTF8.writeFile out $ prettyPrint $
+  liftIO $ UTF8.writeFile out $ (++ "\n") $ prettyPrint $
     HS.Module m (p : ps) imp ds
   where
   p = HS.LanguagePragma $ List.map HS.Ident $
