@@ -5,7 +5,7 @@ module Agda.Compiler.ToTreeless
   , closedTermToTreeless
   ) where
 
-import Control.Applicative
+import Control.Arrow (first, second)
 import Control.Monad.Reader
 import Control.Monad.State
 import Data.Maybe
@@ -14,14 +14,16 @@ import qualified Data.Map as Map
 import Data.Traversable (traverse)
 
 import Agda.Syntax.Common
-import Agda.Syntax.Internal (QName)
+import Agda.Syntax.Internal as I
 import qualified Agda.Syntax.Treeless as C
-import qualified Agda.Syntax.Internal as I
+import Agda.Syntax.Treeless (TTerm)
 import Agda.Syntax.Literal
 import qualified Agda.TypeChecking.CompiledClause as CC
+import qualified Agda.TypeChecking.CompiledClause.Compile as CC
 import Agda.TypeChecking.Records (getRecordConstructor)
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.CompiledClause
+import Agda.TypeChecking.Telescope
 
 import Agda.Compiler.Treeless.Builtin
 import Agda.Compiler.Treeless.Simplify
@@ -52,6 +54,18 @@ import Agda.Utils.Impossible
 prettyPure :: P.Pretty a => a -> TCM Doc
 prettyPure = return . P.pretty
 
+-- | Recompile clauses with forcing translation turned on.
+getCompiledClauses :: QName -> TCM CC.CompiledClauses
+getCompiledClauses q = do
+  def <- getConstInfo q
+  let cs = defClauses def
+      isProj | Function{ funProjection = proj } <- theDef def = isJust (projProper =<< proj)
+             | otherwise = False
+      translate | isProj    = CC.DontRunRecordPatternTranslation
+                | otherwise = CC.RunRecordPatternTranslation
+  reportSDoc "treeless.convert" 40 $ text "-- before clause compiler" $$ (pretty q <+> text "=") <?> vcat (map pretty cs)
+  CC.compileClauses' translate cs
+
 -- | Converts compiled clauses to treeless syntax.
 --
 -- Note: Do not use any of the concrete names in the returned
@@ -64,7 +78,7 @@ toTreeless q = ifM (alwaysInline q) (pure Nothing) $ Just <$> toTreeless' q
 toTreeless' :: QName -> TCM C.TTerm
 toTreeless' q =
   flip fromMaybeM (getTreeless q) $ verboseBracket "treeless.convert" 20 ("compiling " ++ prettyShow q) $ do
-    Just cc <- defCompiled <$> getConstInfo q
+    cc <- getCompiledClauses q
     unlessM (alwaysInline q) $ setTreeless q (C.TDef q)
       -- so recursive inlining doesn't loop, but not for always inlined
       -- functions, since that would risk inlining to fail.
@@ -86,24 +100,7 @@ ccToTreeless q cc = do
   reportSDoc "treeless.convert" (30 + v) $ text "-- compiled clauses of" <+> prettyTCM q $$ nest 2 (prettyPure cc)
   body <- casetreeTop cc
   reportSDoc "treeless.opt.converted" (30 + v) $ text "-- converted" $$ pbody body
-  body <- simplifyTTerm body
-  reportSDoc "treeless.opt.simpl" (35 + v) $ text "-- after first simplification"  $$ pbody body
-  body <- translateBuiltins body
-  reportSDoc "treeless.opt.builtin" (30 + v) $ text "-- after builtin translation" $$ pbody body
-  body <- simplifyTTerm body
-  reportSDoc "treeless.opt.simpl" (30 + v) $ text "-- after second simplification"  $$ pbody body
-  body <- eraseTerms q body
-  reportSDoc "treeless.opt.erase" (30 + v) $ text "-- after erasure"  $$ pbody body
-  body <- caseToSeq body
-  reportSDoc "treeless.opt.uncase" (30 + v) $ text "-- after uncase"  $$ pbody body
-  body <- recoverAsPatterns body
-  reportSDoc "treeless.opt.aspat" (30 + v) $ text "-- after @-pattern recovery"  $$ pbody body
-  body <- detectIdentityFunctions q body
-  reportSDoc "treeless.opt.id" (30 + v) $ text "-- after identity function detection"  $$ pbody body
-  body <- simplifyTTerm body
-  reportSDoc "treeless.opt.simpl" (30 + v) $ text "-- after third simplification"  $$ pbody body
-  body <- eraseTerms q body
-  reportSDoc "treeless.opt.erase" (30 + v) $ text "-- after second erasure"  $$ pbody body
+  body <- runPipeline q (compilerPipeline v q) body
   used <- usedArguments q body
   when (any not used) $
     reportSDoc "treeless.opt.unused" (30 + v) $
@@ -114,6 +111,63 @@ ccToTreeless q cc = do
   setCompiledArgUse q used
   return body
 
+data Pipeline = FixedPoint Int Pipeline
+              | Sequential [Pipeline]
+              | SinglePass CompilerPass
+
+data CompilerPass = CompilerPass
+  { passTag       :: String
+  , passVerbosity :: Int
+  , passName      :: String
+  , passCode      :: TTerm -> TCM TTerm
+  }
+
+compilerPass :: String -> Int -> String -> (TTerm -> TCM TTerm) -> Pipeline
+compilerPass tag v name code = SinglePass (CompilerPass tag v name code)
+
+compilerPipeline :: Int -> QName -> Pipeline
+compilerPipeline v q =
+  Sequential
+    [ compilerPass "simpl"   (35 + v) "simplification"      simplifyTTerm
+    , compilerPass "builtin" (30 + v) "builtin translation" translateBuiltins
+    , FixedPoint 5 $ Sequential
+      [ compilerPass "simpl"  (30 + v) "simplification"                simplifyTTerm
+      , compilerPass "erase"  (30 + v) "erasure" $                     eraseTerms q
+      , compilerPass "uncase" (30 + v) "uncase"                        caseToSeq
+      , compilerPass "aspat"  (30 + v) "@-pattern recovery"            recoverAsPatterns
+      ]
+    , compilerPass "id" (30 + v) "identity function detection" $ detectIdentityFunctions q
+    ]
+
+runPipeline :: QName -> Pipeline -> TTerm -> TCM TTerm
+runPipeline q pipeline t = case pipeline of
+  SinglePass p   -> runCompilerPass q p t
+  Sequential ps  -> foldM (flip $ runPipeline q) t ps
+  FixedPoint n p -> runFixedPoint n q p t
+
+runCompilerPass :: QName -> CompilerPass -> TTerm -> TCM TTerm
+runCompilerPass q p t = do
+  t' <- passCode p t
+  let dbg f   = reportSDoc ("treeless.opt." ++ passTag p) (passVerbosity p) $ f $ text ("-- " ++ passName p)
+      pbody b = sep [ text (prettyShow q) <+> text "=", nest 2 $ prettyPure b ]
+  dbg $ if | t == t'   -> (<+> text "(No effect)")
+           | otherwise -> ($$ pbody t')
+  return t'
+
+runFixedPoint :: Int -> QName -> Pipeline -> TTerm -> TCM TTerm
+runFixedPoint n q pipeline = go 1
+  where
+    go i t | i > n = do
+      reportSLn "treeless.opt.loop" 20 $ "++ Optimisation loop reached maximum iterations (" ++ show n ++ ")"
+      return t
+    go i t = do
+      reportSLn "treeless.opt.loop" 30 $ "++ Optimisation loop iteration " ++ show i
+      t' <- runPipeline q pipeline t
+      if | t == t'   -> do
+            reportSLn "treeless.opt.loop" 30 $ "++ Optimisation loop terminating after " ++ show i ++ " iterations"
+            return t'
+         | otherwise -> go (i + 1) t'
+
 closedTermToTreeless :: I.Term -> TCM C.TTerm
 closedTermToTreeless t = do
   substTerm t `runReaderT` initCCEnv
@@ -173,19 +227,21 @@ casetree cc = do
       -- if some arguments are not used in the body.
       v <- lift (putAllowedReductions [ProjectionReductions, CopatternReductions] $ normalise v)
       substTerm v
-    CC.Case _ (CC.Branches True _ _ Just{}) -> __IMPOSSIBLE__
+    CC.Case _ (CC.Branches True _ _ _ Just{} _) -> __IMPOSSIBLE__
       -- Andreas, 2016-06-03, issue #1986: Ulf: "no catch-all for copatterns!"
       -- lift $ do
       --   typeError . GenericDocError =<< do
       --     text "Not yet implemented: compilation of copattern matching with catch-all clause"
-    CC.Case (Arg _ n) (CC.Branches True conBrs _ Nothing) -> lambdasUpTo n $ do
+    CC.Case (Arg _ n) (CC.Branches True conBrs _ _ Nothing _) -> lambdasUpTo n $ do
       mkRecord =<< traverse casetree (CC.content <$> conBrs)
-    CC.Case (Arg _ n) (CC.Branches False conBrs litBrs catchAll) -> lambdasUpTo (n + 1) $ do
-      if Map.null conBrs && Map.null litBrs then do
+    CC.Case (Arg _ n) (CC.Branches False conBrs etaBr litBrs catchAll lazy) -> lambdasUpTo (n + 1) $ do
+                    -- We can treat eta-matches as regular matches here.
+      let conBrs' = Map.union conBrs $ Map.fromList $ map (first conName) $ maybeToList etaBr
+      if Map.null conBrs' && Map.null litBrs then do
         -- there are no branches, just return default
-        fromCatchAll
+        updateCatchAll catchAll fromCatchAll
       else do
-        caseTy <- case (Map.keys conBrs, Map.keys litBrs) of
+        caseTy <- case (Map.keys conBrs', Map.keys litBrs) of
               ((c:_), []) -> do
                 c' <- lift (canonicalName c)
                 dtNm <- conData . theDef <$> lift (getConstInfo c')
@@ -198,8 +254,9 @@ casetree cc = do
         updateCatchAll catchAll $ do
           x <- lookupLevel n <$> asks ccCxt
           def <- fromCatchAll
-          C.TCase x caseTy def <$> do
-            br1 <- conAlts n conBrs
+          let caseInfo = C.CaseInfo { caseType = caseTy, caseLazy = lazy }
+          C.TCase x caseInfo def <$> do
+            br1 <- conAlts n conBrs'
             br2 <- litAlts n litBrs
             return (br1 ++ br2)
   where
@@ -214,12 +271,13 @@ commonArity cc =
     [] -> 0
     as -> minimum as
   where
-    arities cxt (Case (Arg _ x) (Branches False cons lits def)) =
+    arities cxt (Case (Arg _ x) (Branches False cons eta lits def _)) =
       concatMap (wArities cxt') (Map.elems cons) ++
+      concatMap (wArities cxt') (map snd $ maybeToList eta) ++
       concatMap (wArities cxt' . WithArity 0) (Map.elems lits) ++
       concat [ arities cxt' c | Just c <- [def] ] -- ??
       where cxt' = max (x + 1) cxt
-    arities cxt (Case _ (Branches True _ _ _)) = [cxt]
+    arities cxt (Case _ Branches{projPatterns = True}) = [cxt]
     arities cxt (Done xs _) = [max cxt (length xs)]
     arities _   Fail        = []
 
@@ -333,7 +391,7 @@ recConFromProj q = do
 --   indices in the environment as well.
 substTerm :: I.Term -> CC C.TTerm
 substTerm term = normaliseStatic term >>= \ term ->
-  case I.ignoreSharing $ I.unSpine term of
+  case I.unSpine term of
     I.Var ind es -> do
       ind' <- lookupIndex ind <$> asks ccCxt
       let args = fromMaybe __IMPOSSIBLE__ $ I.allApplyElims es
@@ -347,10 +405,10 @@ substTerm term = normaliseStatic term >>= \ term ->
     I.Def q es -> do
       let args = fromMaybe __IMPOSSIBLE__ $ I.allApplyElims es
       maybeInlineDef q args
-    I.Con c ci args -> do
+    I.Con c ci es -> do
+        let args = fromMaybe __IMPOSSIBLE__ $ I.allApplyElims es
         c' <- lift $ canonicalName $ I.conName c
         C.mkTApp (C.TCon c') <$> substArgs args
-    I.Shared _ -> __IMPOSSIBLE__ -- the ignoreSharing fun should already take care of this
     I.Pi _ _ -> return C.TUnit
     I.Sort _  -> return C.TSort
     I.MetaV _ _ -> __IMPOSSIBLE__   -- we don't compiled if unsolved metas
@@ -391,5 +449,4 @@ substArg x | erasable x = return C.TErased
       case getRelevance x of
         Irrelevant -> True
         NonStrict  -> True
-        Forced{}   -> False -- TODO: would like this to be True
         Relevant   -> False
