@@ -4,11 +4,12 @@
 module Agda.Compiler.Treeless.Pretty () where
 
 import Control.Arrow ((&&&), (***), first, second)
-import Control.Applicative
 import Control.Monad.Reader
 import Data.Maybe
+import qualified Data.Map as Map
 
 import Agda.Syntax.Treeless
+import Agda.Compiler.Treeless.Subst
 import Agda.Utils.Pretty
 
 data PEnv = PEnv { pPrec :: Int
@@ -25,6 +26,19 @@ withNames n k = do
   (xs, ys) <- asks $ splitAt n . pFresh
   local (\ e -> e { pFresh = ys }) (k xs)
 
+-- | Don't generate fresh names for unused variables.
+withNames' :: HasFree a => Int -> a -> ([String] -> P b) -> P b
+withNames' n tm k = withNames n' $ k . insBlanks
+  where
+    fv = freeVars tm
+    n'  = length $ filter (< n) $ Map.keys fv
+    insBlanks = go n
+      where
+        go 0 _ = []
+        go i xs0@(~(x : xs))
+          | Map.member (i - 1) fv = x   : go (i - 1) xs
+          | otherwise             = "_" : go (i - 1) xs0
+
 bindName :: String -> P a -> P a
 bindName x = local $ \ e -> e { pBound = x : pBound e }
 
@@ -59,12 +73,22 @@ opName PRem = "rem"
 opName PGeq = ">="
 opName PLt  = "<"
 opName PEqI = "==I"
+opName PAdd64 = "+64"
+opName PSub64 = "-64"
+opName PMul64 = "*64"
+opName PQuot64 = "quot64"
+opName PRem64 = "rem64"
+opName PLt64  = "<64"
+opName PEq64 = "==64"
 opName PEqF = "==F"
 opName PEqS = "==S"
 opName PEqC = "==C"
 opName PEqQ = "==Q"
 opName PIf  = "if_then_else_"
 opName PSeq = "seq"
+opName PITo64 = "toWord64"
+opName P64ToI = "fromWord64"
+
 
 isInfix :: TPrim -> Maybe (Int, Int, Int)
 isInfix op =
@@ -74,6 +98,10 @@ isInfix op =
     PSub -> l 6
     PGeq -> non 4
     PLt  -> non 4
+    PMul64 -> l 7
+    PAdd64 -> l 6
+    PSub64 -> l 6
+    PLt64  -> non 4
     p | isPrimEq p -> non 4
     _    -> Nothing
   where
@@ -107,22 +135,18 @@ pTerm t = case t of
     paren 9 $ (\a bs -> sep [a, nest 2 $ fsep bs])
               <$> pTerm' 9 f
               <*> mapM (pTerm' 10) es
-  TLam _ -> paren 0 $ withNames n $ \xs -> bindNames xs $
+  TLam _ -> paren 0 $ withNames' n b $ \ xs -> bindNames xs $
     (\b -> sep [ text ("λ " ++ unwords xs ++ " →")
                , nest 2 b ]) <$> pTerm' 0 b
     where
-      (n, b) = lamV t
-      lamV (TLam b) = first succ $ lamV b
-      lamV t        = (0, t)
+      (n, b) = tLamView t
   TLet{} -> paren 0 $ withNames (length es) $ \ xs ->
     (\ (binds, b) -> sep [ text "let" <+> vcat [ sep [ text x <+> text "="
                                                      , nest 2 e ] | (x, e) <- binds ]
                               <+> text "in", b ])
       <$> pLets (zip xs es) b
     where
-      (es, b) = letV t
-      letV (TLet e b) = first (e :) $ letV b
-      letV t          = ([], t)
+      (es, b) = tLetView t
 
       pLets [] b = ([],) <$> pTerm' 0 b
       pLets ((x, e) : bs) b = do
@@ -142,7 +166,7 @@ pTerm t = case t of
         pAlt' <$> ((text "_" <+> text "|" <+>) <$> pTerm' 0 g)
               <*> (pTerm' 0 b)
       pAlt (TACon c a b) =
-        withNames a $ \ xs -> bindNames xs $
+        withNames' a b $ \ xs -> bindNames xs $
         pAlt' <$> pTerm' 0 (TApp (TCon c) [TVar i | i <- reverse [0..a - 1]])
               <*> pTerm' 0 b
       pAlt' p b = sep [p <+> text "→", nest 2 b]
@@ -151,3 +175,4 @@ pTerm t = case t of
   TSort -> pure $ text "Set"
   TErased -> pure $ text "_"
   TError err -> paren 9 $ pure $ text "error" <+> text (show (show err))
+  TCoerce t -> paren 9 $ (text "coe" <+>) <$> pTerm' 10 t
