@@ -26,7 +26,7 @@ import Agda.Syntax.Abstract as A hiding (Open, Apply, Assign)
 import Agda.Syntax.Abstract.Views as A
 import Agda.Syntax.Abstract.Pretty
 import Agda.Syntax.Common
-import Agda.Syntax.Info (ExprInfo(..),MetaInfo(..),emptyMetaInfo,exprNoRange)
+import Agda.Syntax.Info (ExprInfo(..),MetaInfo(..),emptyMetaInfo,exprNoRange,defaultAppInfo_,defaultAppInfo)
 import qualified Agda.Syntax.Info as Info
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Literal
@@ -35,7 +35,7 @@ import Agda.Syntax.Translation.AbstractToConcrete
 import Agda.Syntax.Translation.ConcreteToAbstract
 import Agda.Syntax.Scope.Base
 import Agda.Syntax.Scope.Monad
-import Agda.Syntax.Fixity(Precedence(..))
+import Agda.Syntax.Fixity(Precedence(..), argumentCtx_)
 import Agda.Syntax.Parser
 
 import Agda.TheTypeChecker
@@ -186,7 +186,7 @@ give force ii mr e = liftTCM $ do
      giveExpr force (Just ii) mi e
     `catchError` \ case
       -- Turn PatternErr into proper error:
-      PatternErr{} -> typeError . GenericDocError =<< do
+      PatternErr -> typeError . GenericDocError =<< do
         withInteractionId ii $ TP.text "Failed to give" TP.<+> prettyTCM e
       err -> throwError err
   removeInteractionPoint ii
@@ -230,8 +230,8 @@ refine force ii mr e = do
           ii <- registerInteractionPoint False rng Nothing
           let info = Info.MetaInfo
                 { Info.metaRange = rng
-                , Info.metaScope = scope { scopePrecedence = [ArgumentCtx] }
-                    -- Ulf, 2017-09-07: The `ArgumentCtx` above is causing #737.
+                , Info.metaScope = scope { scopePrecedence = [argumentCtx_] }
+                    -- Ulf, 2017-09-07: The `argumentCtx_` above is causing #737.
                     -- If we're building an operator application the precedence
                     -- should be something else.
                 , metaNumber = Nothing -- in order to print just as ?, not ?n
@@ -252,13 +252,25 @@ refine force ii mr e = do
               -- reduce beta-redexes where the argument is used at most once
               smartApp i e arg =
                 case lamView $ unScope e of
-                  Just (x, e) | count x e < 2 -> mapExpr subX e
+                  Just (A.BindName x, e) | count x e < 2 -> mapExpr subX e
                     where subX (A.Var y) | x == y = namedArg arg
                           subX e = e
                   _ -> App i e arg
-          return $ smartApp (ExprRange r) e $ defaultNamedArg metaVar
-          --ToDo: The position of metaVar is not correct
-          --ToDo: The fixity of metavars is not correct -- fixed? MT
+          return $ smartApp (defaultAppInfo r) e $ defaultNamedArg metaVar
+
+-- Andreas, 2017-12-16:
+-- Ulf, your attempt to fix #737 introduced regression #2873.
+-- Going through concrete syntax does some arbitrary disambiguation
+-- of constructors, which subsequently makes refine fail.
+-- I am not convinced of the printing-parsing shortcut to address problems.
+-- (Unless you prove the roundtrip property.)
+--
+--           rescopeExpr scope $ smartApp (defaultAppInfo r) e $ defaultNamedArg metaVar
+-- -- | Turn an abstract expression into concrete syntax and then back into
+-- --   abstract. This ensures that context precedences are set correctly for
+-- --   abstract expressions built by hand. Used by refine above.
+-- rescopeExpr :: ScopeInfo -> Expr -> TCM Expr
+-- rescopeExpr scope = withScope_ scope . (concreteToAbstract_ <=< runAbsToCon . preserveInteractionIds . toConcrete)
 
 {-| Evaluate the given expression in the current environment -}
 evalInCurrent :: Expr -> TCM Expr
@@ -334,6 +346,7 @@ data OutputConstraint a b
       | IsEmptyType a
       | SizeLtSat a
       | FindInScopeOF b a [(a,a)]
+      | PTSInstance b b
   deriving (Functor)
 
 -- | A subset of 'OutputConstraint'.
@@ -363,6 +376,7 @@ outputFormId (OutputForm _ _ o) = out o
       IsEmptyType _              -> __IMPOSSIBLE__   -- Should never be used on IsEmpty constraints
       SizeLtSat{}                -> __IMPOSSIBLE__
       FindInScopeOF _ _ _        -> __IMPOSSIBLE__
+      PTSInstance i _            -> i
 
 instance Reify ProblemConstraint (Closure (OutputForm Expr Expr)) where
   reify (PConstr pids cl) = enterClosure cl $ \c -> buildClosure =<< (OutputForm (getRange c) (Set.toList pids) <$> reify c)
@@ -373,11 +387,11 @@ reifyElimToExpr e = case e of
     I.Proj _o f -> appl "proj" <$> reify ((defaultArg $ I.Def f []) :: Arg Term)
   where
     appl :: String -> Arg Expr -> Expr
-    appl s v = A.App exprNoRange (A.Lit (LitString noRange s)) $ fmap unnamed v
+    appl s v = A.App defaultAppInfo_ (A.Lit (LitString noRange s)) $ fmap unnamed v
 
 instance Reify Constraint (OutputConstraint Expr Expr) where
     reify (ValueCmp cmp t u v)   = CmpInType cmp <$> reify t <*> reify u <*> reify v
-    reify (ElimCmp cmp t v es1 es2) =
+    reify (ElimCmp cmp _ t v es1 es2) =
       CmpElim cmp <$> reify t <*> mapM reifyElimToExpr es1
                               <*> mapM reifyElimToExpr es2
     reify (LevelCmp cmp t t')    = CmpLevels cmp <$> reify t <*> reify t'
@@ -402,15 +416,15 @@ instance Reify Constraint (OutputConstraint Expr Expr) where
               domType <- maybe (return underscore) reify mt
               target  <- reify target
               let bs = TypedBindings noRange $ Arg ai $
-                       TBind noRange xs domType
-                  e  = A.Lam Info.defaultLamInfo_ (DomainFull bs) body
+                       TBind noRange (map (fmap A.BindName) xs) domType
+                  e  = A.Lam Info.exprNoRange (DomainFull bs) body
               return $ TypedAssign m' e target
             CheckArgs _ _ args t0 t1 _ -> do
               t0 <- reify t0
               t1 <- reify t1
               return $ PostponedCheckArgs m' (map (namedThing . unArg) args) t0 t1
             UnquoteTactic tac _ goal -> do
-              tac <- A.App exprNoRange (A.Unquote exprNoRange) . defaultNamedArg <$> reify tac
+              tac <- A.App defaultAppInfo_ (A.Unquote exprNoRange) . defaultNamedArg <$> reify tac
               OfType tac <$> reify goal
           Open{}  -> __IMPOSSIBLE__
           OpenIFS{}  -> __IMPOSSIBLE__
@@ -422,6 +436,11 @@ instance Reify Constraint (OutputConstraint Expr Expr) where
             (,) <$> reify tm <*> reify ty)
     reify (IsEmpty r a) = IsEmptyType <$> reify a
     reify (CheckSizeLtSat a) = SizeLtSat  <$> reify a
+    reify (CheckFunDef d i q cs) = __IMPOSSIBLE__
+    reify (HasBiggerSort a) = OfType <$> reify a <*> reify (UnivSort a)
+    reify (HasPTSRule a b) = do
+      (a,(x,b)) <- reify (a,b)
+      return $ PTSInstance a b
 
 -- ASR TODO (28 December 2014): This function will be unnecessary when
 -- using a Pretty instance for OutputConstraint instead of the Show
@@ -463,6 +482,7 @@ instance (Show a,Show b) => Show (OutputConstraint a b) where
       where
       showCand (tm,ty) = indent 6 $ show tm ++ " : " ++ show ty
       indent n s = List.intercalate ("\n" ++ replicate n ' ') $ lines s
+    show (PTSInstance a b)      = "PTS instance for " ++ show (a,b)
 
 instance (ToConcrete a c, ToConcrete b d) =>
          ToConcrete (OutputForm a b) (OutputForm c d) where
@@ -474,17 +494,17 @@ instance (ToConcrete a c, ToConcrete b d) =>
     toConcrete (JustType e) = JustType <$> toConcrete e
     toConcrete (JustSort e) = JustSort <$> toConcrete e
     toConcrete (CmpInType cmp t e e') =
-      CmpInType cmp <$> toConcreteCtx TopCtx t <*> toConcreteCtx ArgumentCtx e
-                                               <*> toConcreteCtx ArgumentCtx e'
+      CmpInType cmp <$> toConcreteCtx TopCtx t <*> toConcreteCtx argumentCtx_ e
+                                               <*> toConcreteCtx argumentCtx_ e'
     toConcrete (CmpElim cmp t e e') =
       CmpElim cmp <$> toConcreteCtx TopCtx t <*> toConcreteCtx TopCtx e <*> toConcreteCtx TopCtx e'
-    toConcrete (CmpTypes cmp e e') = CmpTypes cmp <$> toConcreteCtx ArgumentCtx e
-                                                  <*> toConcreteCtx ArgumentCtx e'
-    toConcrete (CmpLevels cmp e e') = CmpLevels cmp <$> toConcreteCtx ArgumentCtx e
-                                                    <*> toConcreteCtx ArgumentCtx e'
+    toConcrete (CmpTypes cmp e e') = CmpTypes cmp <$> toConcreteCtx argumentCtx_ e
+                                                  <*> toConcreteCtx argumentCtx_ e'
+    toConcrete (CmpLevels cmp e e') = CmpLevels cmp <$> toConcreteCtx argumentCtx_ e
+                                                    <*> toConcreteCtx argumentCtx_ e'
     toConcrete (CmpTeles cmp e e') = CmpTeles cmp <$> toConcrete e <*> toConcrete e'
-    toConcrete (CmpSorts cmp e e') = CmpSorts cmp <$> toConcreteCtx ArgumentCtx e
-                                                  <*> toConcreteCtx ArgumentCtx e'
+    toConcrete (CmpSorts cmp e e') = CmpSorts cmp <$> toConcreteCtx argumentCtx_ e
+                                                  <*> toConcreteCtx argumentCtx_ e'
     toConcrete (Guard o pid) = Guard <$> toConcrete o <*> pure pid
     toConcrete (Assign m e) = noTakenNames $ Assign <$> toConcrete m <*> toConcreteCtx TopCtx e
     toConcrete (TypedAssign m e a) = TypedAssign <$> toConcrete m <*> toConcreteCtx TopCtx e
@@ -496,6 +516,7 @@ instance (ToConcrete a c, ToConcrete b d) =>
     toConcrete (FindInScopeOF s t cs) =
       FindInScopeOF <$> toConcrete s <*> toConcrete t
                     <*> mapM (\(tm,ty) -> (,) <$> toConcrete tm <*> toConcrete ty) cs
+    toConcrete (PTSInstance a b) = PTSInstance <$> toConcrete a <*> toConcrete b
 
 instance (Pretty a, Pretty b) => Pretty (OutputConstraint' a b) where
   pretty (OfType' e t) = pretty e <+> text ":" <+> pretty t
@@ -536,7 +557,7 @@ getSolvedInteractionPoints all norm = concat <$> do
         scope <- getScope
         let sol v = do
               -- Andreas, 2014-02-17 exclude metas solved by metas
-              v <- ignoreSharing <$> instantiate v
+              v <- instantiate v
               let isMeta = case v of MetaV{} -> True; _ -> False
               if isMeta && not all then return [] else do
                 e <- reify =<< normalForm norm v
@@ -698,7 +719,6 @@ metaHelperType norm ii rng s = case words s of
       I.Sort{}     -> pure v
       I.Level{}    -> pure v
       I.MetaV{}    -> pure v
-      I.Shared{}   -> pure v
     onNamesElims f = traverse $ traverse $ onNamesTm f
     onNamesArgs f  = traverse $ traverse $ onNamesTm f
     onNamesAbs f   = onNamesAbs' f (stringToArgName <.> f . argNameToString)
@@ -799,7 +819,7 @@ introTactic pmLambda ii = do
                 (EmptyTel, EmptyTel) -> return []
                 _ -> introFun (telToList tel' ++ telToList tel)
 
-        case ignoreSharing $ unEl t of
+        case unEl t of
           I.Def d _ -> do
             def <- getConstInfo d
             case theDef def of
@@ -884,7 +904,7 @@ atTopLevel m = inConcreteMode $ do
       -- Unfortunately, referring to let-bound variables
       -- from the top level module telescope will for now result in a not-in-scope error.
       let names :: [A.Name]
-          names = map localVar $ filter ((False ==) . localLetBound) $ map snd $ reverse $ scopeLocals scope
+          names = map localVar $ filter ((LetBound /=) . localBinder) $ map snd $ reverse $ scopeLocals scope
       -- Andreas, 2016-12-31, issue #2371
       -- The following is an unnecessary complication, as shadowed locals
       -- are not in scope anyway (they are ambiguous).
@@ -904,7 +924,11 @@ atTopLevel m = inConcreteMode $ do
         ]
       M.withCurrentModule current $
         withScope_ scope $
-          addContext gamma $
+          addContext gamma $ do
+            -- We're going inside the top-level module, so we have to set the
+            -- checkpoint for it and all its submodules to the new checkpoint.
+            cp <- view eCurrentCheckpoint
+            stModuleCheckpoints %= fmap (const cp)
             m
 
 -- | Parse a name.
