@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 
 ------------------------------------------------------------------------
 -- | Code for instructing Emacs to do things
@@ -14,6 +15,10 @@ module Agda.Interaction.EmacsCommand
   , displayRunningInfo
   ) where
 
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ((<>))
+#endif
+
 import qualified Data.List as List
 
 import Agda.Utils.Pretty
@@ -30,6 +35,7 @@ data Lisp a
     -- ^ List.
   | Q (Lisp a)
     -- Quoted expression.
+  deriving Eq
 
 instance Pretty a => Pretty (Lisp a) where
   pretty (A a )     = pretty a
