@@ -11,6 +11,7 @@ module Agda.Interaction.Highlighting.Generate
   , printUnsolvedInfo
   , printHighlightingInfo
   , highlightAsTypeChecked
+  , warningHighlighting
   , computeUnsolvedMetaWarnings
   , computeUnsolvedConstraints
   , storeDisambiguatedName
@@ -21,7 +22,6 @@ import Prelude hiding (null)
 import Control.Monad
 import Control.Monad.Trans
 import Control.Monad.Reader
-import Control.Applicative
 import Control.Arrow (second)
 
 import Data.Monoid
@@ -29,13 +29,15 @@ import Data.Generics.Geniplate
 import qualified Data.Map as Map
 import Data.Maybe
 import Data.List ((\\), isPrefixOf)
-import qualified Data.Foldable as Fold (fold, foldMap)
+import qualified Data.Foldable as Fold (fold, foldMap, toList)
 import qualified Data.IntMap as IntMap
 import Data.Void
 
 import Agda.Interaction.Response (Response(Resp_HighlightingInfo))
 import Agda.Interaction.Highlighting.Precise
 import Agda.Interaction.Highlighting.Range
+import Agda.Interaction.Response
+  (RemoveTokenBasedHighlighting(KeepHighlighting))
 
 import qualified Agda.TypeChecking.Errors as E
 import Agda.TypeChecking.MetaVars (isBlockedTerm)
@@ -45,6 +47,7 @@ import qualified Agda.TypeChecking.Monad as M
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Warnings (runPM)
 
+import Agda.Syntax.Abstract (IsProjP(..))
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Concrete (FieldAssignment'(..))
 import qualified Agda.Syntax.Common as Common
@@ -57,6 +60,7 @@ import qualified Agda.Syntax.Literal as L
 import qualified Agda.Syntax.Parser as Pa
 import qualified Agda.Syntax.Parser.Tokens as T
 import qualified Agda.Syntax.Position as P
+import Agda.Syntax.Position (getRange)
 
 import Agda.Utils.FileName
 import Agda.Utils.Function
@@ -99,7 +103,7 @@ highlightAsTypeChecked rPre r m
   r'        = rToR (P.continuousPerLine r)
   delta     = rPre' `minus` r'
   clear     = mempty
-  highlight = mempty { otherAspects = [TypeChecks] }
+  highlight = parserBased { otherAspects = [TypeChecks] }
 
   wrap rs x y = do
     p rs x
@@ -107,12 +111,16 @@ highlightAsTypeChecked rPre r m
     p rs y
     return v
     where
-    p rs x = printHighlightingInfo (singletonC rs x)
+    p rs x = printHighlightingInfo KeepHighlighting (singletonC rs x)
 
 -- | Lispify and print the given highlighting information.
 
-printHighlightingInfo :: MonadTCM tcm => HighlightingInfo -> tcm ()
-printHighlightingInfo info = do
+printHighlightingInfo ::
+  MonadTCM tcm =>
+  RemoveTokenBasedHighlighting ->
+  HighlightingInfo ->
+  tcm ()
+printHighlightingInfo remove info = do
   modToSrc <- use stModuleToSource
   method   <- view eHighlightingMethod
   liftTCM $ reportSLn "highlighting" 50 $ unlines
@@ -122,7 +130,7 @@ printHighlightingInfo info = do
     ]
   unless (null $ ranges info) $ do
     liftTCM $ appInteractionOutputCallback $
-        Resp_HighlightingInfo info method modToSrc
+        Resp_HighlightingInfo info remove method modToSrc
 
 -- | Highlighting levels.
 
@@ -187,7 +195,7 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
           Just i  -> ( fromIntegral $ P.posPos $ P.iStart i
                      , fromIntegral $ P.posPos $ P.iEnd i)
     (prevTokens, (curTokens, postTokens)) <-
-      (second (splitAtC to)) . splitAtC from <$> use stTokens
+      second (splitAtC to) . splitAtC from <$> use stTokens
 
     -- theRest needs to be placed before nameInfo here since record
     -- field declarations contain QNames. constructorInfo also needs
@@ -208,12 +216,12 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
       stTokens     .= prevTokens `mappend` postTokens
 
     ifTopLevelAndHighlightingLevelIs NonInteractive $
-      printHighlightingInfo syntaxInfo
+      printHighlightingInfo KeepHighlighting syntaxInfo
   where
   -- All names mentioned in the syntax tree (not bound variables).
   names :: [A.AmbiguousQName]
   names =
-    (map (A.AmbQ . (:[])) $
+    (map I.unambiguous $
      filter (not . extendedLambda) $
      universeBi decl) ++
     universeBi decl
@@ -238,22 +246,27 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     , Fold.foldMap getNamedArg    $ universeBi decl
     ]
     where
-    bound n = nameToFile modMap file [] (A.nameConcrete n) P.noRange
-                         (\isOp -> mempty { aspect = Just $ Name (Just Bound) isOp })
+    bound (A.BindName n) = nameToFile modMap file [] (A.nameConcrete n) P.noRange
+                         (\isOp -> parserBased { aspect =
+                                     Just $ Name (Just Bound) isOp })
                          (Just $ A.nameBindingSite n)
 
-    patsyn n = nameToFileA modMap file n True $ \isOp ->
-                  mempty { aspect = Just $ Name (Just $ Constructor Common.Inductive) isOp }
+    patsyn n =               -- TODO: resolve overloading
+              nameToFileA modMap file (I.headAmbQ n) True $ \isOp ->
+                  parserBased { aspect =
+                    Just $ Name (Just $ Constructor Common.Inductive) isOp }
 
     macro n = nameToFileA modMap file n True $ \isOp ->
-                  mempty { aspect = Just $ Name (Just Macro) isOp }
+                  parserBased { aspect = Just $ Name (Just Macro) isOp }
 
     field m n = nameToFile modMap file m n P.noRange
-                           (\isOp -> mempty { aspect = Just $ Name (Just Field) isOp })
+                           (\isOp -> parserBased { aspect =
+                                       Just $ Name (Just Field) isOp })
                            Nothing
     asName n = nameToFile modMap file []
                           n P.noRange
-                          (\isOp -> mempty { aspect = Just $ Name (Just Module) isOp })
+                          (\isOp -> parserBased { aspect =
+                                      Just $ Name (Just Module) isOp })
                           Nothing
 
     -- For top level modules, we set the binding site to the beginning of the file
@@ -262,11 +275,13 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     mod isTopLevelModule n =
       nameToFile modMap file []
                  (A.nameConcrete n) P.noRange
-                 (\isOp -> mempty { aspect = Just $ Name (Just Module) isOp })
-                 (Just $ applyWhen isTopLevelModule P.beginningOfFile $ A.nameBindingSite n)
+                 (\isOp -> parserBased { aspect =
+                             Just $ Name (Just Module) isOp })
+                 (Just $ applyWhen isTopLevelModule P.beginningOfFile $
+                           A.nameBindingSite n)
 
     getVarAndField :: A.Expr -> File
-    getVarAndField (A.Var x)            = bound x
+    getVarAndField (A.Var x)            = bound $ A.BindName x
     getVarAndField (A.Rec       _ fs)   = mconcat [ field [] x | Left (FieldAssignment x _) <- fs ]
     getVarAndField (A.RecUpdate _ _ fs) = mconcat [ field [] x |      (FieldAssignment x _) <- fs ]
     getVarAndField _                    = mempty
@@ -274,7 +289,9 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     -- Ulf, 2014-04-09: It would be nicer to have it on Named_ a, but
     -- you can't have polymorphic functions in universeBi.
     getNamedArg :: Common.RString -> File
-    getNamedArg x = singleton (rToR $ P.getRange x) mempty{ aspect = Just $ Name (Just Argument) False }
+    getNamedArg x = singleton (rToR $ P.getRange x) $
+                       parserBased { aspect =
+                         Just $ Name (Just Argument) False }
 
     getLet :: A.LetBinding -> File
     getLet (A.LetBind _ _ x _ _)     = bound x
@@ -292,15 +309,17 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     getTyped A.TLet{}         = mempty
 
     getPatSynArgs :: A.Declaration -> File
-    getPatSynArgs (A.PatternSynDef _ xs _) = mconcat $ map (bound . Common.unArg) xs
+    getPatSynArgs (A.PatternSynDef _ xs _) = mconcat $ map (bound . A.BindName . Common.unArg) xs
     getPatSynArgs _                        = mempty
 
-    getPattern' :: A.Pattern' e -> File
+    getPattern' :: IsProjP e => A.Pattern' e -> File
     getPattern' (A.VarP x)    = bound x
     getPattern' (A.AsP _ x _) = bound x
-    getPattern' (A.DotP pi _ _) =
-      singleton (rToR $ P.getRange pi)
-                (mempty { otherAspects = [DottedPattern] })
+    getPattern' (A.DotP pi e)
+      | Just _ <- isProjP e = mempty
+      | otherwise =
+          singleton (rToR $ P.getRange pi)
+                (parserBased { otherAspects = [DottedPattern] })
     getPattern' (A.PatternSynP _ q _) = patsyn q
     getPattern' (A.RecP _ fs) = mconcat [ field [] x | FieldAssignment x _ <- fs ]
     getPattern' _             = mempty
@@ -342,7 +361,7 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     getModuleInfo :: SI.ModuleInfo -> File
     getModuleInfo (SI.ModuleInfo { SI.minfoAsTo   = asTo
                                  , SI.minfoAsName = name }) =
-      singleton (rToR asTo) (mempty { aspect = Just Symbol })
+      singleton (rToR asTo) (parserBased { aspect = Just Symbol })
         `mappend`
       maybe mempty asName name
 
@@ -381,6 +400,7 @@ tokenHighlighting = merge . map tokenToCFile
   tokenToCFile (T.TokKeyword _ i)               = aToF Keyword (P.getRange i)
   tokenToCFile (T.TokSymbol  _ i)               = aToF Symbol (P.getRange i)
   tokenToCFile (T.TokLiteral (L.LitNat    r _)) = aToF Number r
+  tokenToCFile (T.TokLiteral (L.LitWord64 r _)) = aToF Number r
   tokenToCFile (T.TokLiteral (L.LitFloat  r _)) = aToF Number r
   tokenToCFile (T.TokLiteral (L.LitString r _)) = aToF String r
   tokenToCFile (T.TokLiteral (L.LitChar   r _)) = aToF String r
@@ -506,20 +526,21 @@ generateConstructorInfo modMap file kinds decl = do
         constrs = IntMap.elems m2
 
     -- Return suitable syntax highlighting information.
-    let files = for constrs $ \ q -> generate modMap file kinds $ A.AmbQ [q]
+    let files = for constrs $ \ q -> generate modMap file kinds $ I.unambiguous q
     return $ Fold.fold files
 
 printSyntaxInfo :: P.Range -> TCM ()
 printSyntaxInfo r = do
   syntaxInfo <- use stSyntaxInfo
   ifTopLevelAndHighlightingLevelIs NonInteractive $
-      printHighlightingInfo (selectC r syntaxInfo)
-
+      printHighlightingInfo KeepHighlighting (selectC r syntaxInfo)
 
 -- | Prints syntax highlighting info for an error.
 
 printErrorInfo :: TCErr -> TCM ()
-printErrorInfo e = printHighlightingInfo . compress =<< errorHighlighting e
+printErrorInfo e =
+  printHighlightingInfo KeepHighlighting . compress =<<
+    errorHighlighting e
 
 -- | Generate highlighting for error.
 --   Does something special for termination errors.
@@ -540,9 +561,9 @@ errorHighlighting e = do
   -- Print new highlighting.
   s <- E.prettyError e
   let error = singleton (rToR r)
-         $ mempty { otherAspects = [Error]
-                  , note         = Just s
-                  }
+         $ parserBased { otherAspects = [Error]
+                       , note         = Just s
+                       }
   return $ mconcat [ erase, error ]
 
 -- | Generate syntax highlighting for warnings.
@@ -554,15 +575,16 @@ warningHighlighting w = case tcWarning w of
   UnreachableClauses{}       -> unreachableErrorHighlighting $ P.getRange w
   CoverageIssue{}            -> coverageErrorHighlighting $ P.getRange w
   CoverageNoExactSplit{}     -> catchallHighlighting $ P.getRange w
+  UnsolvedConstraints cs     -> constraintsHighlighting cs
+  UnsolvedMetaVariables rs   -> metasHighlighting rs
   -- expanded catch-all case to get a warning for new constructors
-  UnsolvedMetaVariables{}    -> mempty
   UnsolvedInteractionMetas{} -> mempty
-  UnsolvedConstraints{}      -> mempty
   OldBuiltin{}               -> mempty
   EmptyRewritePragma{}       -> mempty
   UselessPublic{}            -> mempty
   UselessInline{}            -> mempty
   ParseWarning{}             -> mempty
+  InversionDepthReached{}    -> mempty
   GenericWarning{}           -> mempty
   GenericNonFatalError{}     -> mempty
   SafeFlagPostulate{}        -> mempty
@@ -574,13 +596,14 @@ warningHighlighting w = case tcWarning w of
   SafeFlagPolarity           -> mempty
   DeprecationWarning{}       -> mempty
   NicifierIssue{}            -> mempty
+  UserWarning{}              -> mempty
 
 -- | Generate syntax highlighting for termination errors.
 
 terminationErrorHighlighting :: [TerminationError] -> File
 terminationErrorHighlighting termErrs = functionDefs `mappend` callSites
   where
-    m            = mempty { otherAspects = [TerminationProblem] }
+    m            = parserBased { otherAspects = [TerminationProblem] }
     functionDefs = Fold.foldMap (\x -> singleton (rToR $ bindingSite x) m) $
                    concatMap M.termErrFunctions termErrs
     callSites    = Fold.foldMap (\r -> singleton (rToR r) m) $
@@ -594,20 +617,20 @@ positivityErrorHighlighting :: I.QName -> OccursWhere -> File
 positivityErrorHighlighting q o = several (rToR <$> P.getRange q : rs) m
   where
     rs = case o of Unknown -> []; Known r _ -> [r]
-    m  = mempty { otherAspects = [PositivityProblem] }
+    m  = parserBased { otherAspects = [PositivityProblem] }
 
 unreachableErrorHighlighting :: P.Range -> File
 unreachableErrorHighlighting r = singleton (rToR $ P.continuousPerLine r) m
-  where m = mempty { otherAspects = [ReachabilityProblem] }
+  where m = parserBased { otherAspects = [ReachabilityProblem] }
 
 coverageErrorHighlighting :: P.Range -> File
 coverageErrorHighlighting r = singleton (rToR $ P.continuousPerLine r) m
-  where m = mempty { otherAspects = [CoverageProblem] }
+  where m = parserBased { otherAspects = [CoverageProblem] }
 
 
 catchallHighlighting :: P.Range -> File
 catchallHighlighting r = singleton (rToR $ P.continuousPerLine r) m
-  where m = mempty { otherAspects = [CatchallClause] }
+  where m = parserBased { otherAspects = [CatchallClause] }
 
 
 -- | Generates and prints syntax highlighting information for unsolved
@@ -618,7 +641,7 @@ printUnsolvedInfo = do
   metaInfo       <- computeUnsolvedMetaWarnings
   constraintInfo <- computeUnsolvedConstraints
 
-  printHighlightingInfo
+  printHighlightingInfo KeepHighlighting
     (compress $ metaInfo `mappend` constraintInfo)
 
 -- | Generates syntax highlighting information for unsolved meta
@@ -635,19 +658,34 @@ computeUnsolvedMetaWarnings = do
   ms <- filterM notBlocked =<< getOpenMetas
 
   rs <- mapM getMetaRange (ms \\ is)
-  return $ several (map (rToR . P.continuousPerLine) rs)
-                   (mempty { otherAspects = [UnsolvedMeta] })
+  return $ metasHighlighting rs
+
+metasHighlighting :: [P.Range] -> File
+metasHighlighting rs = several (map (rToR . P.continuousPerLine) rs)
+                     $ parserBased { otherAspects = [UnsolvedMeta] }
 
 -- | Generates syntax highlighting information for unsolved constraints
--- that are not connected to a meta variable.
+--   (ideally: that are not connected to a meta variable).
 
 computeUnsolvedConstraints :: TCM File
-computeUnsolvedConstraints = do
-  cs <- getAllConstraints
-  -- get ranges of emptyness constraints
-  let rs = [ r | PConstr{ theConstraint = Closure{ clValue = IsEmpty r t }} <- cs ]
-  return $ several (map (rToR . P.continuousPerLine) rs)
-                   (mempty { otherAspects = [UnsolvedConstraint] })
+computeUnsolvedConstraints = constraintsHighlighting <$> getAllConstraints
+
+constraintsHighlighting :: Constraints -> File
+constraintsHighlighting cs =
+  several (map (rToR . P.continuousPerLine) rs)
+          (parserBased { otherAspects = [UnsolvedConstraint] })
+  where
+  -- get ranges of interesting unsolved constraints
+  rs = (`mapMaybe` (map theConstraint cs)) $ \case
+    Closure{ clValue = IsEmpty r t           } -> Just r
+    Closure{ clEnv = e, clValue = ValueCmp{} } -> Just $ getRange (envRange e)
+    Closure{ clEnv = e, clValue = ElimCmp{}  } -> Just $ getRange (envRange e)
+    Closure{ clEnv = e, clValue = TypeCmp{}  } -> Just $ getRange (envRange e)
+    Closure{ clEnv = e, clValue = TelCmp{}   } -> Just $ getRange (envRange e)
+    Closure{ clEnv = e, clValue = SortCmp{}  } -> Just $ getRange (envRange e)
+    Closure{ clEnv = e, clValue = LevelCmp{} } -> Just $ getRange (envRange e)
+    Closure{ clEnv = e, clValue = CheckSizeLtSat{} } -> Just $ getRange (envRange e)
+    _ -> Nothing
 
 -- | Generates a suitable file for a possibly ambiguous name.
 
@@ -659,9 +697,9 @@ generate :: SourceToModule
          -> A.AmbiguousQName
          -> File
 generate modMap file kinds (A.AmbQ qs) =
-  mconcat $ map (\q -> nameToFileA modMap file q include m) qs
+  Fold.foldMap (\ q -> nameToFileA modMap file q include m) qs
   where
-    ks   = map kinds qs
+    ks   = map kinds (Fold.toList qs)
     -- Ulf, 2014-06-03: [issue1064] It's better to pick the first rather
     -- than doing no highlighting if there's an ambiguity between an
     -- inductive and coinductive constructor.
@@ -674,8 +712,8 @@ generate modMap file kinds (A.AmbQ qs) =
     -- Note that all names in an AmbiguousQName should have the same
     -- concrete name, so either they are all operators, or none of
     -- them are.
-    m isOp  = mempty { aspect = Just $ Name kind isOp }
-    include = allEqual (map bindingSite qs)
+    m isOp  = parserBased { aspect = Just $ Name kind isOp }
+    include = allEqual (map bindingSite $ Fold.toList qs)
 
 -- | Converts names to suitable 'File's.
 
