@@ -13,8 +13,8 @@ import Prelude hiding (log)
 import Data.Char
 import Data.Maybe
 import Data.Function
+import Data.Foldable (toList)
 import Control.Monad.RWS.Strict
-import Control.Applicative
 import Control.Arrow (second)
 import System.Directory
 import System.Exit
@@ -48,8 +48,7 @@ import qualified Agda.Interaction.FindFile as Find
 import Agda.Interaction.Highlighting.Precise
 import Agda.TypeChecking.Monad (TCM, Interface(..))
 import qualified Agda.TypeChecking.Monad as TCM
-import Agda.Interaction.Options
-  (optGHCiInteraction, optLaTeXDir, optCountClusters)
+import qualified Agda.Interaction.Options as O
 import Agda.Compiler.CallCompiler
 import qualified Agda.Utils.IO.UTF8 as UTF8
 import Agda.Utils.FileName (filePath, AbsolutePath, mkAbsolute)
@@ -218,10 +217,8 @@ replaceSpaces :: Text -> Text
 replaceSpaces = T.map (\c -> if isSpaceNotNewline c then ' ' else c)
 
 
--- | Yields the next token, taking special care to begin/end code
--- blocks. Junk occuring before and after the code blocks is separated
--- into separate tokens, this makes it easier to keep track of whether
--- we are in a code block or not.
+-- | If the `Token` consists of spaces, the internal column counter is advanced
+--   by the length of the token. Otherwise, `moveColumnForToken` is a no-op.
 moveColumnForToken :: Token -> LaTeX ()
 moveColumnForToken t = do
   unless (isSpaces (text t)) $ do
@@ -295,12 +292,6 @@ registerColumnZero = do
 -- | Changes to the state that are performed at the start of a code
 -- block.
 
--- This code is based on the assumption that there are no
--- non-whitespace characters following \begin{code}. For occurrences
--- of \begin{code} which start a code block this is true. However, the
--- LaTeX backend does not identify code blocks correctly, see Issue
--- #2400.
-
 enterCode :: LaTeX ()
 enterCode = do
   resetColumn
@@ -420,10 +411,8 @@ cmdArg :: Text -> Text
 cmdArg x = T.singleton '{' <+> x <+> T.singleton '}'
 
 ------------------------------------------------------------------------
--- * Automaton.
+-- * Output generation from a stream of labelled tokens.
 
--- | The start state, @nonCode@, prints non-code (the LaTeX part of
---   literate Agda) until it sees a @beginBlock@.
 processLayers :: [(LayerRole, Tokens)] -> LaTeX ()
 processLayers = mapM_ $ \(layerRole,toks) -> do
   case layerRole of
@@ -432,8 +421,11 @@ processLayers = mapM_ $ \(layerRole,toks) -> do
     L.Code    -> processCode    toks
 
 processMarkup, processComment, processCode :: Tokens -> LaTeX ()
--- | Deals with markup
-processMarkup  = mapM_ moveColumnForToken
+
+-- | Deals with markup, which is output verbatim.
+processMarkup = mapM_ $ \t -> do
+  moveColumnForToken t
+  output (Text (text t))
 
 -- | Deals with literate text, which is output verbatim
 processComment = mapM_ $ \t -> do
@@ -444,11 +436,11 @@ processComment = mapM_ $ \t -> do
 -- | Deals with code blocks. Every token, except spaces, is pretty
 -- printed as a LaTeX command.
 processCode toks' = do
-  output $ Text $ beginCode <+> nl
+  output $ Text nl
   enterCode
   mapM_ go toks'
   ptOpenWhenColumnZero =<< gets column
-  output $ Text $ ptClose <+> nl <+> endCode
+  output $ Text $ ptClose <+> nl
   leaveCode
 
   where
@@ -467,10 +459,13 @@ processCode toks' = do
             spaces $ T.group $ replaceSpaces tok
         else do
           ptOpenWhenColumnZero col
-          case aspect (info tok') of
-            Nothing -> output $ Text $ escape tok
-            Just a  -> output $ Text $
-                         cmdPrefix <+> T.pack (cmd a) <+> cmdArg (escape tok)
+          output $ Text $
+            -- we return the escaped token wrapped in commands corresponding
+            -- to its aspect (if any) and other aspects (e.g. error, unsolved meta)
+            foldr (\c t -> cmdPrefix <+> T.pack c <+> cmdArg t)
+                  (escape tok)
+                  $ concatMap fromAspect (toList $ aspect $ info tok')
+                    ++ map fromOtherAspect (otherAspects $ info tok')
 
     -- Non-whitespace tokens at the start of a line trigger an
     -- alignment column.
@@ -479,8 +474,13 @@ processCode toks' = do
           registerColumnZero
           output . Text . ptOpen =<< columnZero
 
-    cmd :: Aspect -> String
-    cmd a = let s = show a in case a of
+    -- Translation from OtherAspect to command strings. So far it happens
+    -- to correspond to @show@ but it does not have to (cf. fromAspect)
+    fromOtherAspect :: OtherAspect -> String
+    fromOtherAspect = show
+
+    fromAspect :: Aspect -> [String]
+    fromAspect a = let s = [show a] in case a of
       Comment           -> s
       Option            -> s
       Keyword           -> s
@@ -488,7 +488,7 @@ processCode toks' = do
       Number            -> s
       Symbol            -> s
       PrimitiveType     -> s
-      Name Nothing isOp -> cmd (Name (Just Postulate) isOp)
+      Name Nothing isOp -> fromAspect (Name (Just Postulate) isOp)
         -- At the time of writing the case above can be encountered in
         -- --only-scope-checking mode, for instance for the token "Size"
         -- in the following code:
@@ -497,30 +497,32 @@ processCode toks' = do
         --
         -- The choice of "Postulate" works for this example, but might
         -- be less appropriate for others.
-      Name (Just kind) _ -> case kind of
-        Bound                     -> s
-        Constructor Inductive     -> "InductiveConstructor"
-        Constructor CoInductive   -> "CoinductiveConstructor"
-        Datatype                  -> s
-        Field                     -> s
-        Function                  -> s
-        Module                    -> s
-        Postulate                 -> s
-        Primitive                 -> s
-        Record                    -> s
-        Argument                  -> s
-        Macro                     -> s
+      Name (Just kind) isOp ->
+        (\c -> if isOp then ["Operator", c] else [c]) $
+        case kind of
+          Bound                     -> s
+          Constructor Inductive     -> "InductiveConstructor"
+          Constructor CoInductive   -> "CoinductiveConstructor"
+          Datatype                  -> s
+          Field                     -> s
+          Function                  -> s
+          Module                    -> s
+          Postulate                 -> s
+          Primitive                 -> s
+          Record                    -> s
+          Argument                  -> s
+          Macro                     -> s
         where
         s = show kind
 
--- Escapes special characters.
+-- | Escapes special characters.
 escape :: Text -> Text
 escape (T.uncons -> Nothing)     = T.empty
 escape (T.uncons -> Just (c, s)) = T.pack (replace c) <+> escape s
   where
   replace :: Char -> String
   replace c = case c of
-    '_'  -> "\\_"
+    '_'  -> "\\AgdaUnderscore{}"
     '{'  -> "\\{"
     '}'  -> "\\}"
     '#'  -> "\\#"
@@ -592,7 +594,7 @@ spaces [ s ] = do
 
   output $ MaybeColumn column
 
--- Split multi-lines string literals into multiple string literals
+-- | Split multi-lines string literals into multiple string literals
 -- Isolating leading spaces for the alignment machinery to work
 -- properly
 stringLiteral :: Token -> Tokens
@@ -626,12 +628,12 @@ generateLaTeX i = do
 
   options <- TCM.commandLineOptions
 
-  dir <- case optGHCiInteraction options of
-    False -> return $ optLaTeXDir options
+  dir <- case O.optGHCiInteraction options of
+    False -> return $ O.optLaTeXDir options
     True  -> do
       sourceFile <- Find.findFile mod
       return $ filePath (projectRoot sourceFile mod)
-                 </> optLaTeXDir options
+                 </> O.optLaTeXDir options
   liftIO $ createDirectoryIfMissing True dir
 
   (code, _, _) <- liftIO $ readProcessWithExitCode
@@ -654,7 +656,7 @@ generateLaTeX i = do
   liftIO $ do
     source <- UTF8.readTextFile inAbsPath
     latex <- E.encodeUtf8 `fmap`
-               toLaTeX (optCountClusters options)
+               toLaTeX (O.optCountClusters $ O.optPragmaOptions options)
                        (mkAbsolute inAbsPath) source hi
     createDirectoryIfMissing True $ dir </> takeDirectory outPath
     BS.writeFile (dir </> outPath) latex
