@@ -14,11 +14,6 @@ import Control.Monad.State
 import Control.Monad.Trans.Maybe
 import qualified Control.Exception as E
 
-#if __GLASGOW_HASKELL__ <= 708
-import Data.Foldable ( Foldable )
-import Data.Traversable ( Traversable, traverse )
-#endif
-
 import Data.Function (on)
 import qualified Data.Map as Map
 import qualified Data.List as List
@@ -64,9 +59,12 @@ import Agda.Interaction.FindFile
 import {-# SOURCE #-} Agda.Interaction.InteractionTop (showOpenMetas)
 import Agda.Interaction.Options
 import qualified Agda.Interaction.Options.Lenses as Lens
-import Agda.Interaction.Highlighting.Precise (HighlightingInfo)
+import Agda.Interaction.Highlighting.Precise
+  (HighlightingInfo, compress)
 import Agda.Interaction.Highlighting.Generate
 import Agda.Interaction.Highlighting.Vim
+import Agda.Interaction.Response
+  (RemoveTokenBasedHighlighting(KeepHighlighting))
 
 import Agda.Utils.Except ( MonadError(catchError, throwError) )
 import Agda.Utils.FileName
@@ -134,7 +132,7 @@ mergeInterface i = do
             Just b1 = Map.lookup b bs
             Just b2 = Map.lookup b bi
     mapM_ check (map fst $ Map.toList $ Map.intersection bs bi)
-    addImportedThings sig bi (iPatternSyns i) (iDisplayForms i) warns
+    addImportedThings sig bi (iPatternSyns i) (iDisplayForms i) (iUserWarnings i) warns
     reportSLn "import.iface.merge" 20 $
       "  Rebinding primitives " ++ show prim
     mapM_ rebind prim
@@ -145,12 +143,13 @@ mergeInterface i = do
 
 addImportedThings ::
   Signature -> BuiltinThings PrimFun ->
-  A.PatternSynDefns -> DisplayForms -> [TCWarning] -> TCM ()
-addImportedThings isig ibuiltin patsyns display warnings = do
+  A.PatternSynDefns -> DisplayForms -> Map A.QName String -> [TCWarning] -> TCM ()
+addImportedThings isig ibuiltin patsyns display userwarn warnings = do
   stImports              %= \ imp -> unionSignatures [imp, isig]
   stImportedBuiltins     %= \ imp -> Map.union imp ibuiltin
   stPatternSynImports    %= \ imp -> Map.union imp patsyns
   stImportedDisplayForms %= \ imp -> HMap.unionWith (++) imp display
+  stUserWarnings         %= \ imp -> Map.union imp userwarn
   stTCWarnings           %= \ imp -> List.union imp warnings
   addImportedInstances isig
 
@@ -298,7 +297,7 @@ getInterface' x isMain = do
     -- Preserve the pragma options unless we are checking the main
     -- interface.
     bracket_ (use stPragmaOptions)
-             (unless (includeStateChanges isMain) . setPragmaOptions) $ do
+             (unless (includeStateChanges isMain) . (stPragmaOptions .=)) $ do
      -- Forget the pragma options (locally).
      setCommandLineOptions . stPersistentOptions . stPersistentState =<< get
 
@@ -490,18 +489,8 @@ typeCheck x file isMain = do
   case isMain of
     MainInterface _ -> do
       r <- withMsgs $ createInterface file x isMain
-
-      -- Merge the signature with the signature for imported
-      -- things.
-      reportSLn "import.iface" 40 $ "Merging with state changes included."
-      sig     <- getSignature
-      patsyns <- getPatternSyns
-      display <- use stImportsDisplayForms
-      addImportedThings sig Map.empty patsyns display []
-      setSignature emptySignature
-      setPatternSyns Map.empty
-
       return (True, r)
+
     NotMainInterface -> do
       ms       <- getImportPath
       nesting  <- asks envModuleNestingLevel
@@ -514,6 +503,7 @@ typeCheck x file isMain = do
       isig     <- use stImports
       ibuiltin <- use stImportedBuiltins
       display  <- use stImportsDisplayForms
+      userwarn <- use stUserWarnings
       ipatsyns <- getPatternSynImports
       ho       <- getInteractionOutputCallback
       -- Every interface is treated in isolation. Note: Some changes to
@@ -536,7 +526,7 @@ typeCheck x file isMain = do
                setInteractionOutputCallback ho
                stModuleToSource .= mf
                setVisitedModules vs
-               addImportedThings isig ibuiltin ipatsyns display []
+               addImportedThings isig ibuiltin ipatsyns display userwarn []
 
                r  <- withMsgs $ createInterface file x isMain
                mf <- use stModuleToSource
@@ -575,7 +565,9 @@ chaseMsg
 chaseMsg kind x file = do
   indentation <- (`replicate` ' ') <$> asks envModuleNestingLevel
   let maybeFile = caseMaybe file "." $ \ f -> " (" ++ f ++ ")."
-  reportSLn "import.chase" 1 $ concat $
+      vLvl | kind == "Checking" = 1
+           | otherwise          = 2
+  reportSLn "import.chase" vLvl $ concat $
     [ indentation, kind, " ", prettyShow x, maybeFile ]
 
 
@@ -590,8 +582,7 @@ highlightFromInterface i file = do
   reportSLn "import.iface" 5 $
     "Generating syntax info for " ++ filePath file ++
     " (read from interface)."
-  printHighlightingInfo (iHighlighting i)
-
+  printHighlightingInfo KeepHighlighting (iHighlighting i)
 
 readInterface :: FilePath -> TCM (Maybe Interface)
 readInterface file = do
@@ -701,7 +692,7 @@ createInterface file mname isMain = Bench.billTo [Bench.TopModule mname] $
     Bench.billTo [Bench.Highlighting] $ do
       -- Generate and print approximate syntax highlighting info.
       ifTopLevelAndHighlightingLevelIs NonInteractive $
-        printHighlightingInfo fileTokenInfo
+        printHighlightingInfo KeepHighlighting fileTokenInfo
       let onlyScope = isMain == MainInterface ScopeCheck
       ifTopLevelAndHighlightingLevelIsOr NonInteractive onlyScope $
         mapM_ (\ d -> generateAndPrintSyntaxInfo d Partial onlyScope) ds
@@ -710,6 +701,10 @@ createInterface file mname isMain = Bench.billTo [Bench.TopModule mname] $
 
     -- Type checking.
 
+    -- Now that all the options are in we can check if caching should
+    -- be on.
+    activateLoadedFileCache
+
     -- invalidate cache if pragmas change, TODO move
     cachingStarts
     opts <- use stPragmaOptions
@@ -752,9 +747,20 @@ createInterface file mname isMain = Bench.billTo [Bench.TopModule mname] $
 
       -- Move any remaining token highlighting to stSyntaxInfo.
       toks <- use stTokens
-      ifTopLevelAndHighlightingLevelIs NonInteractive $ printHighlightingInfo toks
+      ifTopLevelAndHighlightingLevelIs NonInteractive $
+        printHighlightingInfo KeepHighlighting toks
       stTokens .= mempty
-      stSyntaxInfo %= \inf -> inf `mappend` toks
+
+      -- Grabbing warnings and unsolved metas to highlight them
+      warnings <- getAllWarnings' AllWarnings RespectFlags
+      unless (null warnings) $ reportSDoc "import.iface.create" 20 $
+        P.text "collected warnings: " P.<> prettyTCM warnings
+      unsolved <- getAllUnsolved
+      unless (null unsolved) $ reportSDoc "import.iface.create" 20 $
+        P.text "collected unsolved: " P.<> prettyTCM unsolved
+      let warningInfo = compress $ foldMap warningHighlighting $ unsolved ++ warnings
+
+      stSyntaxInfo %= \inf -> (inf `mappend` toks) `mappend` warningInfo
 
       whenM (optGenerateVimFile <$> commandLineOptions) $
         -- Generate Vim file.
@@ -791,9 +797,8 @@ createInterface file mname isMain = Bench.billTo [Bench.TopModule mname] $
 
     -- Serialization.
     reportSLn "import.iface.create" 7 $ "Starting serialization."
-    syntaxInfo <- use stSyntaxInfo
     i <- Bench.billTo [Bench.Serialization, Bench.BuildInterface] $ do
-      buildInterface file topLevel syntaxInfo options
+      buildInterface file topLevel options
 
     reportSLn "tc.top" 101 $ unlines $
       "Signature:" :
@@ -844,22 +849,37 @@ createInterface file mname isMain = Bench.billTo [Bench.TopModule mname] $
 -- in by the user, or not (for instance when deciding if we are
 -- writing an interface file or not)
 
-getAllWarnings' :: WhichWarnings -> IgnoreFlags -> TCM [TCWarning]
-getAllWarnings' ww ifs = do
+getUniqueMetasRanges :: [MetaId] -> TCM [Range]
+getUniqueMetasRanges = fmap List.nub . mapM getMetaRange
+
+getUnsolvedMetas :: TCM [Range]
+getUnsolvedMetas = do
   openMetas            <- getOpenMetas
   interactionMetas     <- getInteractionMetas
-  let getUniqueMetas = fmap List.nub . mapM getMetaRange
-  unsolvedInteractions <- getUniqueMetas interactionMetas
-  unsolvedMetas        <- getUniqueMetas (openMetas List.\\ interactionMetas)
+  getUniqueMetasRanges (openMetas List.\\ interactionMetas)
+
+getAllUnsolved :: TCM [TCWarning]
+getAllUnsolved = do
+  unsolvedInteractions <- getUniqueMetasRanges =<< getInteractionMetas
   unsolvedConstraints  <- getAllConstraints
-  collectedTCWarnings  <- use stTCWarnings
+  unsolvedMetas        <- getUnsolvedMetas
 
-  unsolved <- mapM warning_
-                   [ UnsolvedInteractionMetas unsolvedInteractions
-                   , UnsolvedMetaVariables    unsolvedMetas
-                   , UnsolvedConstraints      unsolvedConstraints ]
+  let checkNonEmpty c rs = c rs <$ guard (not $ null rs)
 
-  fmap (filter ((<= ww) . classifyWarning . tcWarning))
+  mapM warning_ $ catMaybes
+                [ checkNonEmpty UnsolvedInteractionMetas unsolvedInteractions
+                , checkNonEmpty UnsolvedMetaVariables    unsolvedMetas
+                , checkNonEmpty UnsolvedConstraints      unsolvedConstraints ]
+
+getAllWarnings' :: WhichWarnings -> IgnoreFlags -> TCM [TCWarning]
+getAllWarnings' ww ifs = do
+  unsolved            <- getAllUnsolved
+  collectedTCWarnings <- use stTCWarnings
+
+  let showWarn w = classifyWarning w <= ww &&
+                    not (null unsolved && onlyShowIfUnsolved w)
+
+  fmap (filter (showWarn . tcWarning))
     $ applyFlagsToTCWarnings ifs $ reverse
     $ unsolved ++ collectedTCWarnings
 
@@ -898,12 +918,10 @@ buildInterface
   :: AbsolutePath
   -> TopLevelInfo
      -- ^ 'TopLevelInfo' for the current module.
-  -> HighlightingInfo
-     -- ^ Syntax highlighting info for the module.
   -> [OptionsPragma]
      -- ^ Options set in @OPTIONS@ pragmas.
   -> TCM Interface
-buildInterface file topLevel syntaxInfo pragmas = do
+buildInterface file topLevel pragmas = do
     reportSLn "import.iface" 5 "Building interface..."
     let m = topLevelModuleName topLevel
     scope'  <- getScope
@@ -925,9 +943,11 @@ buildInterface file topLevel syntaxInfo pragmas = do
     -- Ulf, 2016-04-12:
     -- Non-closed display forms are not applicable outside the module anyway,
     -- and should be dead-code eliminated (#1928).
-    display <- HMap.filter (not . null) . HMap.map (filter isGlobal) <$> use stImportsDisplayForms
+    display <- HMap.filter (not . null) . HMap.map (filter isClosed) <$> use stImportsDisplayForms
     -- TODO: Kill some ranges?
     (display, sig) <- eliminateDeadCode display =<< getSignature
+    userwarns <- use stUserWarnings
+    syntaxInfo <- use stSyntaxInfo
     -- Andreas, 2015-02-09 kill ranges in pattern synonyms before
     -- serialization to avoid error locations pointing to external files
     -- when expanding a pattern synoym.
@@ -944,6 +964,7 @@ buildInterface file topLevel syntaxInfo pragmas = do
       , iInsideScope     = topLevelScope topLevel
       , iSignature       = sig
       , iDisplayForms    = display
+      , iUserWarnings    = userwarns
       , iBuiltin         = builtin'
       , iForeignCode     = foreignCode
       , iHighlighting    = syntaxInfo
