@@ -52,6 +52,7 @@ import Agda.Syntax.Concrete.Generic as C
 import Agda.Syntax.Concrete.Pretty ()
 import Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Pretty
+import qualified Agda.Syntax.Internal as I
 import Agda.Syntax.Info (mkDefInfo)
 import Agda.Syntax.Translation.ConcreteToAbstract
 import Agda.Syntax.Translation.AbstractToConcrete hiding (withScope)
@@ -85,6 +86,7 @@ import Agda.Utils.Except
   , runExceptT
   )
 
+import Agda.Utils.Either
 import Agda.Utils.FileName
 import Agda.Utils.Function
 import Agda.Utils.Hash
@@ -285,8 +287,7 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
             Right <$>
               (toIO $ handleErr $ Exception noRange $ text $ show e)
 
-          asyncHandler e@E.ThreadKilled = return (Left e)
-          asyncHandler e                = handle e
+          asyncHandler e@AsyncCancelled = return (Left e)
 
           generalHandler (e :: E.SomeException) = handle e
 
@@ -316,7 +317,8 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
         unless (null s1) $ mapM_ putResponse $
             [ Resp_DisplayInfo $ Info_Error s ] ++
             tellEmacsToJumpToError (getRange e) ++
-            [ Resp_HighlightingInfo info method modFile ] ++
+            [ Resp_HighlightingInfo info KeepHighlighting
+                                    method modFile ] ++
             [ Resp_Status $ Status { sChecked = False
                                    , sShowImplicitArguments = x
                                    } ]
@@ -340,7 +342,9 @@ runInteraction (IOTCM current highlighting highlightingMethod cmd) =
     withCurrentFile $ interpret cmd
 
     cf <- gets theCurrentFile
-    when (Just current == (fst <$> cf)) $
+    when (updateInteractionPointsAfter cmd
+            &&
+          Just current == (fst <$> cf)) $
         putResponse . Resp_InteractionPoints =<< gets theInteractionPoints
 
   where
@@ -462,9 +466,6 @@ data Interaction' range
     -- show those instead.
   | Cmd_metas
 
-    -- | Display all warnings.
-  | Cmd_warnings
-
     -- | Shows all the top-level names in the given module, along with
     -- their types. Uses the top-level scope.
   | Cmd_show_module_contents_toplevel
@@ -512,6 +513,20 @@ data Interaction' range
   | Cmd_load_highlighting_info
                         FilePath
 
+    -- | Tells Agda to compute token-based highlighting information
+    -- for the file.
+    --
+    -- This command works even if the file's module name does not
+    -- match its location in the file system, or if the file is not
+    -- scope-correct. Furthermore no file names are put in the
+    -- generated output. Thus it is fine to put source code into a
+    -- temporary file before calling this command. However, the file
+    -- extension should be correct.
+    --
+    -- If the second argument is 'Remove', then the (presumably
+    -- temporary) file is removed after it has been read.
+  | Cmd_tokenHighlighting FilePath Remove
+
     -- | Tells Agda to compute highlighting information for the expression just
     --   spliced into an interaction point.
   | Cmd_highlight InteractionId range String
@@ -596,6 +611,13 @@ data IOTCM' range
          -- -^ What to do
             deriving (Show, Read, Functor, Foldable, Traversable)
 
+-- | Used to indicate whether something should be removed or not.
+
+data Remove
+  = Remove
+  | Keep
+  deriving (Show, Read)
+
 ---------------------------------------------------------
 -- Read instances
 
@@ -678,9 +700,49 @@ independent :: Interaction -> Bool
 independent (Cmd_load {})                   = True
 independent (Cmd_compile {})                = True
 independent (Cmd_load_highlighting_info {}) = True
+independent Cmd_tokenHighlighting {}        = True
 independent Cmd_show_version                = True
 independent _                               = False
 
+-- | Should 'Resp_InteractionPoints' be issued after the command has
+-- run?
+
+updateInteractionPointsAfter :: Interaction -> Bool
+updateInteractionPointsAfter Cmd_load{}                          = True
+updateInteractionPointsAfter Cmd_compile{}                       = True
+updateInteractionPointsAfter Cmd_constraints{}                   = False
+updateInteractionPointsAfter Cmd_metas{}                         = False
+updateInteractionPointsAfter Cmd_show_module_contents_toplevel{} = False
+updateInteractionPointsAfter Cmd_search_about_toplevel{}         = False
+updateInteractionPointsAfter Cmd_solveAll{}                      = True
+updateInteractionPointsAfter Cmd_solveOne{}                      = True
+updateInteractionPointsAfter Cmd_infer_toplevel{}                = False
+updateInteractionPointsAfter Cmd_compute_toplevel{}              = False
+updateInteractionPointsAfter Cmd_load_highlighting_info{}        = False
+updateInteractionPointsAfter Cmd_tokenHighlighting{}             = False
+updateInteractionPointsAfter Cmd_highlight{}                     = True
+updateInteractionPointsAfter ShowImplicitArgs{}                  = False
+updateInteractionPointsAfter ToggleImplicitArgs{}                = False
+updateInteractionPointsAfter Cmd_give{}                          = True
+updateInteractionPointsAfter Cmd_refine{}                        = True
+updateInteractionPointsAfter Cmd_intro{}                         = True
+updateInteractionPointsAfter Cmd_refine_or_intro{}               = True
+updateInteractionPointsAfter Cmd_auto{}                          = True
+updateInteractionPointsAfter Cmd_context{}                       = False
+updateInteractionPointsAfter Cmd_helper_function{}               = False
+updateInteractionPointsAfter Cmd_infer{}                         = False
+updateInteractionPointsAfter Cmd_goal_type{}                     = False
+updateInteractionPointsAfter Cmd_goal_type_context{}             = False
+updateInteractionPointsAfter Cmd_goal_type_context_infer{}       = False
+updateInteractionPointsAfter Cmd_goal_type_context_check{}       = False
+updateInteractionPointsAfter Cmd_show_module_contents{}          = False
+updateInteractionPointsAfter Cmd_make_case{}                     = True
+updateInteractionPointsAfter Cmd_compute{}                       = False
+updateInteractionPointsAfter Cmd_why_in_scope{}                  = False
+updateInteractionPointsAfter Cmd_why_in_scope_toplevel{}         = False
+updateInteractionPointsAfter Cmd_show_version{}                  = False
+updateInteractionPointsAfter Cmd_abort{}                         = False
+
 -- | Interpret an interaction
 
 interpret :: Interaction -> CommandM ()
@@ -721,13 +783,6 @@ interpret Cmd_metas = do -- CL.showMetas []
   (pwe, pwa) <- interpretWarnings
   display_info $ Info_AllGoalsWarnings (unlines ms) pwa pwe
 
-interpret Cmd_warnings = do
-  -- Ulf, 2016-08-09: Warnings are now printed in the info buffer by Cmd_metas.
-  -- pws <- interpretWarnings
-  -- unless (null pwd) $ display_info $ Info_Warning pws
-  return ()
-
-
 interpret (Cmd_show_module_contents_toplevel norm s) =
   liftCommandMT B.atTopLevel $ showModuleContents norm noRange s
 
@@ -746,10 +801,10 @@ interpret (Cmd_solveOne norm ii _ _) = solveInstantiatedGoals norm' (Just ii)
                   _            -> norm
 
 interpret (Cmd_infer_toplevel norm s) =
-  parseAndDoAtToplevel (B.typeInCurrent norm) Info_InferredType s
+  parseAndDoAtToplevel (prettyA <=< B.typeInCurrent norm) Info_InferredType s
 
 interpret (Cmd_compute_toplevel cmode s) =
-  parseAndDoAtToplevel' action Info_NormalForm $ computeWrapInput cmode s
+  parseAndDoAtToplevel action Info_NormalForm $ computeWrapInput cmode s
   where
   action = allowNonTerminatingReductions
          . (if computeIgnoreAbstract cmode then ignoreAbstractMode else inConcreteMode)
@@ -770,6 +825,8 @@ interpret ToggleImplicitArgs = do
              ps { optShowImplicit = not $ optShowImplicit ps } }
 
 interpret (Cmd_load_highlighting_info source) = do
+  l <- envHighlightingLevel <$> ask
+  when (l /= None) $ do
     -- Make sure that the include directories have
     -- been set.
     setCommandLineOpts =<< lift commandLineOptions
@@ -797,16 +854,36 @@ interpret (Cmd_load_highlighting_info source) = do
                 return Nothing
     mapM_ putResponse resp
 
+interpret (Cmd_tokenHighlighting source remove) = do
+  info <- do l <- envHighlightingLevel <$> ask
+             if l == None
+               then return Nothing
+               else do
+                 source <- liftIO (absolute source)
+                 lift $ (Just <$> generateTokenInfo source)
+                           `catchError` \_ ->
+                        return Nothing
+      `finally`
+    case remove of
+      Remove -> liftIO $ removeFile source
+      Keep   -> return ()
+  case info of
+    Just info -> lift $ printHighlightingInfo RemoveHighlighting info
+    Nothing   -> return ()
+
 interpret (Cmd_highlight ii rng s) = do
-  scope <- getOldInteractionScope ii
-  removeOldInteractionScope ii
-  handle $ do
-    e <- try ("Highlighting failed to parse expression in " ++ show ii) $
-           B.parseExpr rng s
-    e <- try ("Highlighting failed to scope check expression in " ++ show ii) $
-           concreteToAbstract scope e
-    lift $ printHighlightingInfo =<< generateTokenInfoFromString rng s
-    lift $ highlightExpr e
+  l <- envHighlightingLevel <$> ask
+  when (l /= None) $ do
+    scope <- getOldInteractionScope ii
+    removeOldInteractionScope ii
+    handle $ do
+      e <- try ("Highlighting failed to parse expression in " ++ show ii) $
+             B.parseExpr rng s
+      e <- try ("Highlighting failed to scope check expression in " ++ show ii) $
+             concreteToAbstract scope e
+      lift $ printHighlightingInfo KeepHighlighting =<<
+               generateTokenInfoFromString rng s
+      lift $ highlightExpr e
   where
     handle :: ExceptT String TCM () -> CommandM ()
     handle m = do
@@ -816,7 +893,8 @@ interpret (Cmd_highlight ii rng s) = do
         Right _ -> return ()
     try :: String -> TCM a -> ExceptT String TCM a
     try err m = mkExceptT $ do
-      (Right <$> m) `catchError` \ _ -> return (Left err)
+      (mapLeft (const err) <$> freshTCM m) `catchError` \ _ -> return (Left err)
+      -- freshTCM to avoid scope checking creating new interaction points
 
 interpret (Cmd_give   force ii rng s) = give_gen force ii rng s Give
 interpret (Cmd_refine ii rng s) = give_gen WithoutForce ii rng s Refine
@@ -926,14 +1004,12 @@ interpret (Cmd_make_case ii rng s) = do
   liftCommandMT (B.withInteractionId ii) $ do
     hidden <- lift $ showImplicitArguments
     tel <- lift $ lookupSection (qnameModule f) -- don't shadow the names in this telescope
-    let cs'  :: [A.Clause] = List.map (extlam_dropLLifted casectxt hidden) cs
-    pcs      :: [Doc]     <- lift $ inTopContext $ addContext tel $ mapM prettyA cs'
-    let pcs' :: [String]   = List.map (extlam_dropName casectxt . render) pcs
+    pcs      :: [Doc]      <- lift $ inTopContext $ addContext tel $ mapM prettyA cs
+    let pcs' :: [String]    = List.map (extlam_dropName casectxt . render) pcs
     lift $ reportSDoc "interaction.case" 60 $ TCP.vcat
       [ TCP.text "InteractionTop.Cmd_make_case"
       , TCP.nest 2 $ TCP.vcat
         [ TCP.text "cs   = " TCP.<+> TCP.vcat (map prettyA cs)
-        , TCP.text "cs'  = " TCP.<+> TCP.vcat (map prettyA cs')
         , TCP.text "pcs  = " TCP.<+> TCP.vcat (map return pcs)
         , TCP.text "pcs' = " TCP.<+> TCP.vcat (map TCP.text pcs')
         ]
@@ -942,7 +1018,6 @@ interpret (Cmd_make_case ii rng s) = do
       [ TCP.text "InteractionTop.Cmd_make_case"
       , TCP.nest 2 $ TCP.vcat
         [ TCP.text "cs   = " TCP.<+> TCP.text (show cs)
-        , TCP.text "cs'  = " TCP.<+> TCP.text (show cs')
         ]
       ]
     putResponse $ Resp_MakeCase (makeCaseVariant casectxt) pcs'
@@ -965,14 +1040,6 @@ interpret (Cmd_make_case ii rng s) = do
         replEquals (w   : ws) = w : replEquals ws
         replEquals []         = []
 
-    -- Drops pattern added to extended lambda functions when lambda lifting them
-    extlam_dropLLifted :: CaseContext -> Bool -> A.Clause -> A.Clause
-    extlam_dropLLifted Nothing _ x = x
-    extlam_dropLLifted _ _ (A.Clause (A.LHS _ A.LHSProj{} _) _ _ _ _ _) = __IMPOSSIBLE__
-    extlam_dropLLifted (Just (ExtLamInfo h nh)) hidden cl@A.Clause{ A.clauseLHS = A.LHS info (A.LHSHead name nps) ps }
-      = let n = if hidden then h + nh else nh
-        in cl{ A.clauseLHS = A.LHS info (A.LHSHead name (drop n nps)) ps }
-
 interpret (Cmd_compute cmode ii rng s) = display_info . Info_NormalForm =<< do
   liftLocalState $ do
     e <- B.parseExprIn ii rng $ computeWrapInput cmode s
@@ -1092,10 +1159,8 @@ cmd_load' file argv unsolvedOK mode cmd = do
     putResponse Resp_ClearRunningInfo
 
     -- Remove any prior syntax highlighting.
-    putResponse Resp_ClearHighlighting
+    putResponse (Resp_ClearHighlighting NotOnlyTokenBased)
 
-    -- We activate the cache only when agda is used interactively
-    lift activateLoadedFileCache
 
     ok <- lift $ Imp.typeCheckMain f mode
 
@@ -1184,7 +1249,7 @@ give_gen force ii rng s0 giveRefine = do
     iis       <- lift $ sortInteractionPoints iis
     modifyTheInteractionPoints $ replace ii iis
     -- print abstract expr
-    ce        <- lift $ abstractToConcreteEnv (makeEnv scope) ae
+    ce        <- lift $ abstractToConcreteScope scope ae
     lift $ reportSLn "interaction.give" 30 $ unlines
       [ "ce = " ++ show ce
       , "scopePrecedence = " ++ show (scopePrecedence scope)
@@ -1204,8 +1269,11 @@ give_gen force ii rng s0 giveRefine = do
     -- the highlighting is moved together with the text when the hole goes away.
     -- To make it work for refine we'd have to adjust the ranges.
     when literally $ lift $ do
-      printHighlightingInfo =<< generateTokenInfoFromString rng s
-      highlightExpr ae
+      l <- envHighlightingLevel <$> ask
+      when (l /= None) $ do
+        printHighlightingInfo KeepHighlighting =<<
+          generateTokenInfoFromString rng s
+        highlightExpr ae
     putResponse $ Resp_GiveAction ii $ mkNewTxt literally ce
     lift $ reportSLn "interaction.give" 30 $ "putResponse GiveAction passed"
     -- display new goal set (if not measuring time)
@@ -1465,7 +1533,7 @@ nameModifiers = "" : "'" : "''" : [show i | i <-[3..]]
 -- as the scope), performs the given command with the expression as
 -- input, and displays the result.
 
-parseAndDoAtToplevel'
+parseAndDoAtToplevel
   :: (A.Expr -> TCM Doc)
      -- ^ The command to perform.
   -> (Doc -> DisplayInfo)
@@ -1473,15 +1541,12 @@ parseAndDoAtToplevel'
   -> String
      -- ^ The expression to parse.
   -> CommandM ()
-parseAndDoAtToplevel' cmd title s = do
+parseAndDoAtToplevel cmd title s = do
   (time, res) <- localStateCommandM $ do
     e <- lift $ runPM $ parse exprParser s
-    maybeTimed (lift $ B.atTopLevel $
-                cmd =<< concreteToAbstract_ e)
-  display_info (title $ fromMaybe empty time $$ res)
-
-parseAndDoAtToplevel :: (A.Expr -> TCM A.Expr) -> (Doc -> DisplayInfo) -> String -> CommandM ()
-parseAndDoAtToplevel cmd = parseAndDoAtToplevel' (prettyA <=< cmd)
+    maybeTimed $ lift $ B.atTopLevel $ do
+      cmd =<< concreteToAbstract_ e
+  display_info $ title $ fromMaybe empty time $$ res
 
 maybeTimed :: CommandM a -> CommandM (Maybe Doc, a)
 maybeTimed work = do
@@ -1497,7 +1562,7 @@ tellToUpdateHighlighting
   :: Maybe (HighlightingInfo, HighlightingMethod, ModuleToSource) -> IO [Response]
 tellToUpdateHighlighting Nothing                = return []
 tellToUpdateHighlighting (Just (info, method, modFile)) =
-  return [Resp_HighlightingInfo info method modFile]
+  return [Resp_HighlightingInfo info KeepHighlighting method modFile]
 
 -- | Tells the Emacs mode to go to the first error position (if any).
 
