@@ -16,15 +16,18 @@ import Agda.Syntax.Common
 import Agda.Syntax.Position
 import qualified Agda.Syntax.Concrete as C
 import qualified Agda.Syntax.Abstract as A
+import qualified Agda.Syntax.Abstract.Views as A
 import qualified Agda.Syntax.Info as A
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
-import Agda.Syntax.Scope.Monad (resolveName, ResolvedName(..))
+import Agda.Syntax.Scope.Base  ( ResolvedName(..), Binder(..) )
+import Agda.Syntax.Scope.Monad ( resolveName )
 import Agda.Syntax.Translation.ConcreteToAbstract
 import Agda.Syntax.Translation.InternalToAbstract
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Coverage
+import Agda.TypeChecking.Coverage.Match ( SplitPatVar(..) , SplitPattern , applySplitPSubst , fromSplitPatterns )
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.RecordPatterns
 import Agda.TypeChecking.Reduce
@@ -92,26 +95,21 @@ parseVariables f tel ii rng ss = do
        , text $ "number of locals= " ++ show nlocals
        ]
 
-    -- Compute which variables correspond to module parameters. These cannot be split on.
-    -- Note: these are not necessarily the outer-most bound variables, since
-    -- module parameter refinement may have instantiated them, or
-    -- with-abstraction might have reshuffled the variables (#2181).
-    pars <- freeVarsToApply f
-    let nonSplittableVars = [ i | Var i [] <- map unArg pars ]
-
     -- Resolve each string to a variable.
     forM ss $ \ s -> do
-      let failNotVar = typeError $ GenericError $ "Not a variable: " ++ s
-          done i
-            | i < 0                    = typeError $ GenericError $
-               "Cannot split on local variable " ++ s
-               -- See issue #2239
-
-            | elem i nonSplittableVars = typeError $ GenericError $
-               "Cannot split on variable " ++ s ++ ". It is either a module parameter " ++
-               "or already instantiated by a dot pattern"
-
-            | otherwise                = return i
+      let failNotVar      = typeError $ GenericError $ "Not a variable: " ++ s
+          failUnbound     = typeError $ GenericError $ "Unbound variable " ++ s
+          failAmbiguous   = typeError $ GenericError $ "Ambiguous variable " ++ s
+          failLocal       = typeError $ GenericError $
+            "Cannot split on local variable " ++ s
+          failModuleBound = typeError $ GenericError $
+            "Cannot split on module parameter " ++ s
+          failLetBound v  = typeError . GenericError $
+            "cannot split on let-bound variable " ++ s
+          failInstantiatedVar v = typeError . GenericDocError =<< sep
+              [ text $ "Cannot split on variable " ++ s ++ ", because it is bound to"
+              , prettyTCM v
+              ]
 
       -- Note: the range in the concrete name is only approximate.
       resName <- resolveName $ C.QName $ C.Name r $ C.stringNameParts s
@@ -125,23 +123,31 @@ parseVariables f tel ii rng ss = do
 
         -- If s is a variable name in scope, get its de Bruijn index
         -- via the type checker.
-        VarName x _ -> do
+        VarName x b -> do
           (v, _) <- getVarInfo x
-          case ignoreSharing v of
-            Var i [] -> done $ i - nlocals
-            _        -> typeError . GenericDocError =<< sep
-              [ text $ "Cannot split on variable " ++ s ++ ", because it is bound to"
-              , prettyTCM v
-              ]
+          case (v , b) of
+            -- Slightly dangerous: the pattern variable `x` may be
+            -- refined to the module parameter `var i`. But in this
+            -- case the instantiation could as well be the other way
+            -- around, so the new clauses will still make sense.
+            (Var i [] , PatternBound) -> do
+              when (i < nlocals) __IMPOSSIBLE__
+              return $ i - nlocals
+            (Var i [] , LambdaBound)
+              | i < nlocals -> failLocal
+              | otherwise   -> failModuleBound
+            (Var i [] , LetBound) -> failLetBound v
+            (_        , _       ) -> failInstantiatedVar v
 
         -- If s is not a name, compare it to the printed variable representation.
         -- This fallback is to enable splitting on hidden variables.
         UnknownName -> do
           case filter ((s ==) . fst) xs of
-            []      -> typeError $ GenericError $ "Unbound variable " ++ s
-            [(_,i)] -> done $ i - nlocals
+            []                    -> failUnbound
+            [(_,i)] | i < nlocals -> failLocal
+                    | otherwise   -> return $ i - nlocals
             -- Issue 1325: Variable names in context can be ambiguous.
-            _       -> typeError $ GenericError $ "Ambiguous variable " ++ s
+            _                     -> failAmbiguous
 
 -- | Lookup the clause for an interaction point in the signature.
 --   Returns the CaseContext, the clause itself, and a list of previous clauses
@@ -226,8 +232,25 @@ makeCase hole rng s = withInteractionId hole $ do
     scs <- if newPats then return [sc] else postProjInExtLam $ do
       res <- splitResult f sc
       case res of
-        Nothing  -> typeError $ GenericError $ "Cannot split on result here"
-        Just cov -> ifNotM (optCopatterns <$> pragmaOptions) failNoCop $ {-else-} do
+
+        Left err  -> do
+          -- Andreas, 2017-12-16, issue #2871
+          -- If there is nothing to split, introduce trailing hidden arguments.
+
+          -- Get trailing hidden pattern variables
+          let trailingPatVars :: [NamedArg DBPatVar]
+              trailingPatVars = takeWhileJust isVarP $ reverse ps
+              isVarP (Arg ai (Named n (VarP _ x))) = Just $ Arg ai $ Named n x
+              isVarP _ = Nothing
+          -- If all are already coming from the user, there is really nothing todo!
+          when (all ((UserWritten ==) . getOrigin) trailingPatVars) $ do
+            typeError . GenericDocError =<< do
+              text "Cannot split on result here, because" <+> prettyTCM err
+          -- Otherwise, we make these user-written
+          let xs = map (dbPatVarIndex . namedArg) trailingPatVars
+          return [makePatternVarsVisible xs sc]
+
+        Right cov -> ifNotM (optCopatterns <$> pragmaOptions) failNoCop $ {-else-} do
           -- Andreas, 2016-05-03: do not introduce function arguments after projection.
           -- This is sometimes annoying and can anyway be done by another C-c C-c.
           -- mapM (snd <.> fixTarget) $ splitClauses cov
@@ -249,7 +272,7 @@ makeCase hole rng s = withInteractionId hole $ do
             if isAbsurd then makeAbsurdClause f sc else makeAbstractClause f rhs sc
     reportSDoc "interaction.case" 65 $ vcat
       [ text "split result:"
-      , nest 2 $ vcat $ map (text . show) cs
+      , nest 2 $ vcat $ map (text . show . A.deepUnscope) cs
       ]
     checkClauseIsClean ipCl
     return (f, casectxt, cs)
@@ -273,7 +296,7 @@ makeCase hole rng s = withInteractionId hole $ do
 
   -- Finds the new variable corresponding to an old one, if any.
   newVar :: SplitClause -> Nat -> Maybe Nat
-  newVar c x = case ignoreSharing $ applyPatSubst (scSubst c) (var x) of
+  newVar c x = case applySplitPSubst (scSubst c) (var x) of
     Var y [] -> Just y
     _        -> Nothing
 
@@ -285,26 +308,27 @@ makeCase hole rng s = withInteractionId hole $ do
     when (List.any ((== ipCl) . ipClause) sips) $
       typeError $ GenericError $ "Cannot split as clause rhs has been refined.  Please reload"
 
--- | Mark the variables given by the list of deBruijn indices as 'UserWritten'
---   in the 'SplitClause'.
+-- | Make the given pattern variables visible by marking their origin as
+--   'CaseSplit' and pattern origin as 'PatOSplit' in the 'SplitClause'.
 makePatternVarsVisible :: [Nat] -> SplitClause -> SplitClause
 makePatternVarsVisible [] sc = sc
 makePatternVarsVisible is sc@SClause{ scPats = ps } =
-  sc{ scPats = map (mapNamedArg mkVis) ps }
+  sc{ scPats = mapNamedArgPattern mkVis ps }
   where
-  mkVis :: NamedArg DBPatVar -> NamedArg DBPatVar
-  mkVis nx@(Arg ai (Named n (DBPatVar x i)))
+  mkVis :: NamedArg SplitPattern -> NamedArg SplitPattern
+  mkVis (Arg ai (Named n (VarP o (SplitPatVar x i ls))))
     | i `elem` is =
       -- We could introduce extra consistency checks, like
       -- if visible ai then __IMPOSSIBLE__ else
       -- or passing the parsed name along and comparing it with @x@
-      setOrigin CaseSplit nx
-    | otherwise = nx
+      Arg (setOrigin CaseSplit ai) $ Named n $ VarP PatOSplit $ SplitPatVar x i ls
+  mkVis np = np
 
 -- | Make clause with no rhs (because of absurd match).
 
 makeAbsurdClause :: QName -> SplitClause -> TCM A.Clause
-makeAbsurdClause f (SClause tel ps _ _ t) = do
+makeAbsurdClause f (SClause tel sps _ _ t) = do
+  let ps = fromSplitPatterns sps
   reportSDoc "interaction.case" 10 $ vcat
     [ text "Interaction.MakeCase.makeAbsurdClause: split clause:"
     , nest 2 $ vcat
@@ -332,7 +356,7 @@ makeAbstractClause f rhs cl = do
 
   lhs <- A.clauseLHS <$> makeAbsurdClause f cl
   reportSDoc "interaction.case" 60 $ text "reified lhs: " <+> text (show lhs)
-  return $ A.Clause lhs [] [] rhs [] False
+  return $ A.Clause lhs [] rhs A.noWhereDecls False
   -- let ii = InteractionId (-1)  -- Dummy interaction point since we never type check this.
   --                              -- Can end up in verbose output though (#1842), hence not __IMPOSSIBLE__.
   -- let info = A.emptyMetaInfo   -- metaNumber = Nothing in order to print as ?, not ?n
