@@ -1,9 +1,5 @@
 {-# LANGUAGE CPP #-}
 
-#if __GLASGOW_HASKELL__ <= 708
-{-# LANGUAGE DeriveDataTypeable #-}
-#endif
-
 module Agda.Interaction.Options
     ( CommandLineOptions(..)
     , IgnoreFlags(..)
@@ -33,17 +29,15 @@ module Agda.Interaction.Options
     , getOptSimple
     ) where
 
-import Control.Applicative
 import Control.Monad            ( (>=>), when )
 import Control.Monad.Trans
 
+import Data.IORef
 import Data.Either
 import Data.Maybe
 import Data.List                ( isSuffixOf , intercalate )
-
-#if __GLASGOW_HASKELL__ <= 708
-import Data.Typeable            ( Typeable )
-#endif
+import Data.Set                 ( Set )
+import qualified Data.Set as Set
 
 import System.Console.GetOpt    ( getOpt', usageInfo, ArgOrder(ReturnInOrder)
                                 , OptDescr(..), ArgDescr(..)
@@ -55,6 +49,9 @@ import Text.EditDistance
 import Agda.Termination.CutOff  ( CutOff(..) )
 
 import Agda.Interaction.Library
+import Agda.Interaction.Options.Help
+import Agda.Interaction.Options.IORefs
+import Agda.Interaction.Options.Warnings
 
 import Agda.Utils.Except
   ( ExceptT
@@ -74,6 +71,8 @@ import Agda.Version
 -- Paths_Agda.hs is in $(BUILD_DIR)/build/autogen/.
 import Paths_Agda ( getDataFileName )
 
+import qualified System.IO.Unsafe as UNSAFE (unsafePerformIO)
+
 -- | This should probably go somewhere else.
 isLiterate :: FilePath -> Bool
 isLiterate file = any (`isSuffixOf` file) literateExts
@@ -87,22 +86,6 @@ type Verbosity = Trie String Int
 data IgnoreFlags = IgnoreFlags | RespectFlags
   deriving Eq
 
--- Potentially turn harmless warnings into nothing, or errors
--- (does not apply to non-fatal errors)
-data WarningMode = LeaveAlone | TurnIntoErrors | IgnoreAllWarnings
-  deriving (Show, Eq
-#if __GLASGOW_HASKELL__ <= 708
-           , Typeable
-#endif
-           )
-
-warningModes :: [ (String, WarningMode) ]
-warningModes =
-  [ (defaultWarningMode, LeaveAlone)
-  , ("ignore"          , IgnoreAllWarnings)
-  , ("error"           , TurnIntoErrors)
-  ]
-
 -- Don't forget to update
 --   doc/user-manual/tools/command-line-options.rst
 -- if you make changes to the command-line options!
@@ -120,7 +103,7 @@ data CommandLineOptions = Options
   , optUseLibs          :: Bool
   -- ^ look for .agda-lib files
   , optShowVersion      :: Bool
-  , optShowHelp         :: Bool
+  , optShowHelp         :: Maybe Help
   , optInteractive      :: Bool
   , optGHCiInteraction  :: Bool
   , optOptimSmashing    :: Bool
@@ -131,16 +114,11 @@ data CommandLineOptions = Options
   , optGenerateHTML     :: Bool
   , optDependencyGraph  :: Maybe FilePath
   , optLaTeXDir         :: FilePath
-  , optCountClusters    :: Bool
-    -- ^ Count extended grapheme clusters rather than code points when
-    -- generating LaTeX.
   , optHTMLDir          :: FilePath
   , optCSSFile          :: Maybe FilePath
   , optIgnoreInterfaces :: Bool
   , optForcing          :: Bool
   , optPragmaOptions    :: PragmaOptions
-  , optSharing          :: Bool
-  , optCaching          :: Bool
   , optOnlyScopeChecking :: Bool
     -- ^ Should the top-level module only be scope-checked, and not
     --   type-checked?
@@ -152,6 +130,7 @@ data CommandLineOptions = Options
 data PragmaOptions = PragmaOptions
   { optShowImplicit              :: Bool
   , optShowIrrelevant            :: Bool
+  , optUseUnicode                :: Bool
   , optVerbose                   :: Verbosity
   , optProofIrrelevance          :: Bool
   , optAllowUnsolved             :: Bool
@@ -177,16 +156,20 @@ data PragmaOptions = PragmaOptions
       -- ^ Should system generated projections 'ProjSystem' be printed
       --   postfix (True) or prefix (False).
   , optInstanceSearchDepth       :: Int
+  , optInversionMaxDepth         :: Int
   , optSafe                      :: Bool
   , optWarningMode               :: WarningMode
   , optCompileNoMain             :: Bool
+  , optCaching                   :: Bool
+  , optCountClusters             :: Bool
+    -- ^ Count extended grapheme clusters rather than code points when
+    -- generating LaTeX.
+  , optAutoInline                :: Bool
+    -- ^ Automatic compile-time inlining for simple definitions (unless marked
+    --   NOINLINE).
+  , optPrintPatternSynonyms      :: Bool
   }
-  deriving ( Show
-           , Eq
-#if __GLASGOW_HASKELL__ <= 708
-           , Typeable
-#endif
-           )
+  deriving (Show, Eq)
 
 -- | The options from an @OPTIONS@ pragma.
 --
@@ -217,7 +200,7 @@ defaultOptions = Options
   , optDefaultLibs      = True
   , optUseLibs          = True
   , optShowVersion      = False
-  , optShowHelp         = False
+  , optShowHelp         = Nothing
   , optInteractive      = False
   , optGHCiInteraction  = False
   , optOptimSmashing    = True
@@ -227,14 +210,11 @@ defaultOptions = Options
   , optGenerateHTML     = False
   , optDependencyGraph  = Nothing
   , optLaTeXDir         = defaultLaTeXDir
-  , optCountClusters    = False
   , optHTMLDir          = defaultHTMLDir
   , optCSSFile          = Nothing
   , optIgnoreInterfaces = False
   , optForcing          = True
   , optPragmaOptions    = defaultPragmaOptions
-  , optSharing          = False
-  , optCaching          = False
   , optOnlyScopeChecking = False
   }
 
@@ -242,6 +222,7 @@ defaultPragmaOptions :: PragmaOptions
 defaultPragmaOptions = PragmaOptions
   { optShowImplicit              = False
   , optShowIrrelevant            = False
+  , optUseUnicode                = True
   , optVerbose                   = defaultVerbosity
   , optProofIrrelevance          = False
   , optExperimentalIrrelevance   = False
@@ -264,9 +245,14 @@ defaultPragmaOptions = PragmaOptions
   , optRewriting                 = False
   , optPostfixProjections        = False
   , optInstanceSearchDepth       = 500
+  , optInversionMaxDepth         = 50
   , optSafe                      = False
-  , optWarningMode               = fromJust $ lookup defaultWarningMode warningModes
+  , optWarningMode               = defaultWarningMode
   , optCompileNoMain             = False
+  , optCaching                   = True
+  , optCountClusters             = False
+  , optAutoInline                = True
+  , optPrintPatternSynonyms      = True
   }
 
 -- | The default termination depth.
@@ -284,12 +270,6 @@ defaultLaTeXDir = "latex"
 defaultHTMLDir :: String
 defaultHTMLDir = "html"
 
--- | The default warning mode.
-
-defaultWarningMode :: String
-defaultWarningMode = "warn"
-
-
 type OptM = ExceptT String IO
 
 runOptM :: OptM a -> IO (Either String a)
@@ -366,16 +346,20 @@ inputFlag f o =
 versionFlag :: Flag CommandLineOptions
 versionFlag o = return $ o { optShowVersion = True }
 
-helpFlag :: Flag CommandLineOptions
-helpFlag o = return $ o { optShowHelp = True }
+helpFlag :: Maybe String -> Flag CommandLineOptions
+helpFlag Nothing    o = return $ o { optShowHelp = Just GeneralHelp }
+helpFlag (Just str) o = case string2HelpTopic str of
+  Just hpt -> return $ o { optShowHelp = Just (HelpFor hpt) }
+  Nothing -> throwError $ "unknown help topic " ++ str ++ " (available: " ++
+                           intercalate ", " (map fst allHelpTopics) ++ ")"
 
 safeFlag :: Flag PragmaOptions
 safeFlag o = return $ o { optSafe = True }
 
 sharingFlag :: Bool -> Flag CommandLineOptions
-sharingFlag b o = return $ o { optSharing = b }
+sharingFlag _ o = return o
 
-cachingFlag :: Bool -> Flag CommandLineOptions
+cachingFlag :: Bool -> Flag PragmaOptions
 cachingFlag b o = return $ o { optCaching = b }
 
 proofIrrelevanceFlag :: Flag PragmaOptions
@@ -399,6 +383,11 @@ showImplicitFlag o = return $ o { optShowImplicit = True }
 showIrrelevantFlag :: Flag PragmaOptions
 showIrrelevantFlag o = return $ o { optShowIrrelevant = True }
 
+asciiOnlyFlag :: Flag PragmaOptions
+asciiOnlyFlag o = do
+  lift $ writeIORef unicodeOrAscii AsciiOnly
+  return $ o { optUseUnicode = False }
+
 ghciInteractionFlag :: Flag CommandLineOptions
 ghciInteractionFlag o = return $ o { optGHCiInteraction = True }
 
@@ -411,7 +400,7 @@ latexFlag o = return $ o { optGenerateLaTeX = True }
 onlyScopeCheckingFlag :: Flag CommandLineOptions
 onlyScopeCheckingFlag o = return $ o { optOnlyScopeChecking = True }
 
-countClustersFlag :: Flag CommandLineOptions
+countClustersFlag :: Flag PragmaOptions
 countClustersFlag o =
 #ifdef COUNT_CLUSTERS
   return $ o { optCountClusters = True }
@@ -420,6 +409,12 @@ countClustersFlag o =
     "Cluster counting has not been enabled in this build of Agda."
 #endif
 
+noAutoInlineFlag :: Flag PragmaOptions
+noAutoInlineFlag o = return $ o { optAutoInline = False }
+
+noPrintPatSynFlag :: Flag PragmaOptions
+noPrintPatSynFlag o = return $ o { optPrintPatternSynonyms = False }
+
 latexDirFlag :: FilePath -> Flag CommandLineOptions
 latexDirFlag d o = return $ o { optLaTeXDir = d }
 
@@ -496,6 +491,11 @@ instanceDepthFlag s o = do
   d <- integerArgument "--instance-search-depth" s
   return $ o { optInstanceSearchDepth = d }
 
+inversionMaxDepthFlag :: String -> Flag PragmaOptions
+inversionMaxDepthFlag s o = do
+  d <- integerArgument "--inversion-max-depth" s
+  return $ o { optInversionMaxDepth = d }
+
 interactiveFlag :: Flag CommandLineOptions
 interactiveFlag  o = return $ o { optInteractive    = True
                                 , optPragmaOptions  = (optPragmaOptions o)
@@ -551,13 +551,9 @@ verboseFlag s o =
     usage = throwError "argument to verbose should be on the form x.y.z:N or N"
 
 warningModeFlag :: String -> Flag PragmaOptions
-warningModeFlag s o =
-    case lookup s warningModes of
-      Just m -> return $ o { optWarningMode = m }
-      Nothing -> usage
-  where
-    usage = throwError $ "unknown warning mode (available: " ++
-                           intercalate ", " (map fst warningModes) ++ ")"
+warningModeFlag s o = case warningModeUpdate s of
+  Just upd -> return $ o { optWarningMode = upd (optWarningMode o) }
+  Nothing  -> throwError $ "unknown warning flag " ++ s ++ ". See --help=warning."
 
 terminationDepthFlag :: String -> Flag PragmaOptions
 terminationDepthFlag s o =
@@ -573,8 +569,11 @@ integerArgument flag s =
 
 standardOptions :: [OptDescr (Flag CommandLineOptions)]
 standardOptions =
-    [ Option ['V']  ["version"] (NoArg versionFlag) "show version number"
-    , Option ['?']  ["help"]    (NoArg helpFlag)    "show this help"
+    [ Option ['V']  ["version"] (NoArg versionFlag)       "show version number"
+    , Option ['?']  ["help"]    (OptArg helpFlag "TOPIC")
+                      ("show help for TOPIC (available: "
+                       ++ intercalate ", " (map fst allHelpTopics)
+                       ++ ")")
     , Option ['I']  ["interactive"] (NoArg interactiveFlag)
                     "start in interactive mode"
     , Option []     ["interaction"] (NoArg ghciInteractionFlag)
@@ -590,15 +589,6 @@ standardOptions =
     , Option []     ["latex-dir"] (ReqArg latexDirFlag "DIR")
                     ("directory in which LaTeX files are placed (default: " ++
                      defaultLaTeXDir ++ ")")
-    , Option []     ["count-clusters"] (NoArg countClustersFlag)
-                    ("count extended grapheme clusters when " ++
-                     "generating LaTeX (note that this flag " ++
-#ifdef COUNT_CLUSTERS
-                     "is not enabled in all builds of Agda)"
-#else
-                     "has not been enabled in this build of Agda)"
-#endif
-                    )
     , Option []     ["html"] (NoArg htmlFlag)
                     "generate HTML files with highlighted source code"
     , Option []     ["html-dir"] (ReqArg htmlDirFlag "DIR")
@@ -623,13 +613,9 @@ standardOptions =
     , Option []     ["no-forcing"] (NoArg noForcingFlag)
                     "disable the forcing optimisation"
     , Option []     ["sharing"] (NoArg $ sharingFlag True)
-                    "enable sharing and call-by-need evaluation (experimental) (default: OFF)"
+                    "DEPRECATED: does nothing"
     , Option []     ["no-sharing"] (NoArg $ sharingFlag False)
-                    "disable sharing and call-by-need evaluation"
-    , Option []     ["caching"] (NoArg $ cachingFlag True)
-                    "enable caching of typechecking (experimental) (default: OFF)"
-    , Option []     ["no-caching"] (NoArg $ cachingFlag False)
-                    "disable caching of typechecking"
+                    "DEPRECATED: does nothing"
     , Option []     ["only-scope-checking"] (NoArg onlyScopeCheckingFlag)
                     "only scope-check the top-level module, do not type-check it"
     ] ++ map (fmap lift) pragmaOptions
@@ -645,6 +631,8 @@ pragmaOptions =
                     "show implicit arguments when printing"
     , Option []     ["show-irrelevant"] (NoArg showIrrelevantFlag)
                     "show irrelevant arguments when printing"
+    , Option []     ["no-unicode"] (NoArg asciiOnlyFlag)
+                    "don't use unicode characters when printing terms"
     , Option ['v']  ["verbose"] (ReqArg verboseFlag "N")
                     "set verbosity level to N"
     -- , Option []          ["proof-irrelevance"] (NoArg proofIrrelevanceFlag)
@@ -699,14 +687,32 @@ pragmaOptions =
                     "make postfix projection notation the default"
     , Option []     ["instance-search-depth"] (ReqArg instanceDepthFlag "N")
                     "set instance search depth to N (default: 500)"
+    , Option []     ["inversion-max-depth"] (ReqArg inversionMaxDepthFlag "N")
+                    "set maximum depth for pattern match inversion to N (default: 50)"
     , Option []     ["safe"] (NoArg safeFlag)
                     "disable postulates, unsafe OPTION pragmas and primTrustMe"
-    , Option ['W']  ["warning"] (ReqArg warningModeFlag "MODE")
-                    ("set warning mode to MODE (available: "
-                       ++ intercalate ", " (map fst warningModes)
-                       ++ ". Default: " ++  defaultWarningMode ++ ")")
+    , Option ['W']  ["warning"] (ReqArg warningModeFlag "FLAG")
+                    ("set warning flags. See --help=warning.")
     , Option []     ["no-main"] (NoArg compileFlagNoMain)
                     "do not treat the requested module as the main module of a program when compiling"
+    , Option []     ["caching"] (NoArg $ cachingFlag True)
+                    "enable caching of typechecking (default)"
+    , Option []     ["no-caching"] (NoArg $ cachingFlag False)
+                    "disable caching of typechecking"
+    , Option []     ["count-clusters"] (NoArg countClustersFlag)
+                    ("count extended grapheme clusters when " ++
+                     "generating LaTeX (note that this flag " ++
+#ifdef COUNT_CLUSTERS
+                     "is not enabled in all builds of Agda)"
+#else
+                     "has not been enabled in this build of Agda)"
+#endif
+                    )
+    , Option []     ["no-auto-inline"] (NoArg noAutoInlineFlag)
+                    ("disable automatic compile-time inlining " ++
+                     "(only definitions marked INLINE will be inlined)")
+    , Option []     ["no-print-pattern-synonyms"] (NoArg noPrintPatSynFlag)
+                    "expand pattern synonyms when printing terms"
     ]
 
 -- | Used for printing usage info.
@@ -793,12 +799,14 @@ parsePluginOptions argv opts =
 
 -- | The usage info message. The argument is the program name (probably
 --   agda).
-usage :: [OptDescr ()] -> String -> String
-usage options progName = usageInfo (header progName) options
+usage :: [OptDescr ()] -> String -> Help -> String
+usage options progName GeneralHelp = usageInfo (header progName) options
     where
         header progName = unlines [ "Agda version " ++ version, ""
                                   , "Usage: " ++ progName ++ " [OPTIONS...] [FILE]" ]
 
+usage options progName (HelpFor topic) = helpTopicUsage topic
+
 -- Remove +RTS .. -RTS from arguments
 stripRTS :: [String] -> [String]
 stripRTS [] = []
