@@ -1,6 +1,7 @@
 {-# LANGUAGE CPP                   #-}
 {-# LANGUAGE DeriveDataTypeable    #-}
 {-# LANGUAGE TemplateHaskell       #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
 {-| The abstract syntax. This is what you get after desugaring and scope
     analysis of the concrete syntax. The type checker works on abstract syntax,
@@ -12,11 +13,11 @@ module Agda.Syntax.Abstract
     ) where
 
 import Prelude
-import Control.Arrow (first)
-import Control.Applicative
+import Control.Arrow (first, second)
 
 import Data.Foldable (Foldable)
 import qualified Data.Foldable as Fold
+import Data.Function (on)
 import Data.Map (Map)
 import Data.Maybe
 import Data.Sequence (Seq, (<|), (><))
@@ -25,10 +26,9 @@ import Data.Traversable
 import Data.Void
 
 import Data.Data (Data)
-import Data.Typeable (Typeable)
 
 import Agda.Syntax.Concrete.Name (NumHoles(..))
-import Agda.Syntax.Concrete (FieldAssignment'(..), exprFieldA)
+import Agda.Syntax.Concrete (FieldAssignment'(..), exprFieldA, HoleContent'(..))
 import qualified Agda.Syntax.Concrete as C
 import Agda.Syntax.Concrete.Pretty ()
 import Agda.Syntax.Info
@@ -45,11 +45,32 @@ import Agda.TypeChecking.Positivity.Occurrence
 import Agda.Utils.Functor
 import Agda.Utils.Geniplate
 import Agda.Utils.Lens
+import Agda.Utils.NonemptyList
 import Agda.Utils.Pretty
 
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+-- A name in a binding position: we also compare the nameConcrete
+-- when comparing the binders for equality.
+
+-- With --caching on we compare abstract syntax to determine if we can
+-- reuse previous typechecking results: during that comparison two
+-- names can have the same nameId but be semantically different,
+-- e.g. in "{_ : A} -> .." vs "{r : A} -> ..".
+newtype BindName = BindName { unBind :: Name }
+  deriving (Show,Data,HasRange,SetRange,KillRange)
+
+instance Eq BindName where
+  (BindName n) == (BindName m)
+    = ((==) `on` nameId) n m
+      && ((==) `on` nameConcrete) n m
+
+instance Ord BindName where
+  (BindName n) `compare` (BindName m)
+    = (compare `on` nameId) n m
+      `mappend` (compare `on` nameConcrete) n m
+
 type Args = [NamedArg Expr]
 
 -- | Expressions after scope checking (operators parsed, names resolved).
@@ -58,7 +79,7 @@ data Expr
   | Def  QName                         -- ^ Constant: axiom, function, data or record type.
   | Proj ProjOrigin AmbiguousQName     -- ^ Projection (overloaded).
   | Con  AmbiguousQName                -- ^ Constructor (overloaded).
-  | PatternSyn QName                   -- ^ Pattern synonym.
+  | PatternSyn AmbiguousQName          -- ^ Pattern synonym.
   | Macro QName                        -- ^ Macro.
   | Lit Literal                        -- ^ Literal.
   | QuestionMark MetaInfo InteractionId
@@ -71,11 +92,11 @@ data Expr
   | Underscore   MetaInfo
     -- ^ Meta variable for hidden argument (must be inferred locally).
   | Dot ExprInfo Expr                  -- ^ @.e@, for postfix projection.
-  | App  ExprInfo Expr (NamedArg Expr) -- ^ Ordinary (binary) application.
+  | App  AppInfo Expr (NamedArg Expr)  -- ^ Ordinary (binary) application.
   | WithApp ExprInfo Expr [Expr]       -- ^ With application.
-  | Lam  LamInfo LamBinding Expr      -- ^ @λ bs → e@.
-  | AbsurdLam LamInfo Hiding          -- ^ @λ()@ or @λ{}@.
-  | ExtendedLam LamInfo DefInfo QName [Clause]
+  | Lam  ExprInfo LamBinding Expr      -- ^ @λ bs → e@.
+  | AbsurdLam ExprInfo Hiding          -- ^ @λ()@ or @λ{}@.
+  | ExtendedLam ExprInfo DefInfo QName [Clause]
   | Pi   ExprInfo Telescope Expr       -- ^ Dependent function space @Γ → A@.
   | Fun  ExprInfo (Arg Expr) Expr      -- ^ Non-dependent function space.
   | Set  ExprInfo Integer              -- ^ @Set@, @Set1@, @Set2@, ...
@@ -93,7 +114,7 @@ data Expr
   | Tactic ExprInfo Expr [NamedArg Expr] [NamedArg Expr]
                                        -- ^ @tactic e x1 .. xn | y1 | .. | yn@
   | DontCare Expr                      -- ^ For printing @DontCare@ from @Syntax.Internal@.
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 -- | Record field assignment @f = e@.
 type Assign  = FieldAssignment' Expr
@@ -106,7 +127,7 @@ data Axiom
   = FunSig    -- ^ A function signature.
   | NoFunSig  -- ^ Not a function signature, i.e., a postulate (in user input)
               --   or another (e.g. data/record) type signature (internally).
-  deriving (Typeable, Data, Eq, Ord, Show)
+  deriving (Data, Eq, Ord, Show)
 
 -- | Renaming (generic).
 type Ren a = [(a, a)]
@@ -114,7 +135,7 @@ type Ren a = [(a, a)]
 data ScopeCopyInfo = ScopeCopyInfo
   { renModules :: Ren ModuleName
   , renNames   :: Ren QName }
-  deriving (Eq, Show, Typeable, Data)
+  deriving (Eq, Show, Data)
 
 initCopyInfo :: ScopeCopyInfo
 initCopyInfo = ScopeCopyInfo
@@ -151,7 +172,7 @@ data Declaration
   | DataDef    DefInfo QName [LamBinding] [Constructor]
       -- ^ the 'LamBinding's are 'DomainFree' and bind the parameters of the datatype.
   | RecSig     DefInfo QName Telescope Expr -- ^ lone record signature
-  | RecDef     DefInfo QName (Maybe (Ranged Induction)) (Maybe Bool) (Maybe QName) [LamBinding] Expr [Declaration]
+  | RecDef     DefInfo QName (Maybe (Ranged Induction)) (Maybe HasEta) (Maybe QName) [LamBinding] Expr [Declaration]
       -- ^ The 'LamBinding's are 'DomainFree' and bind the parameters of the datatype.
       --   The 'Expr' gives the constructor type telescope, @(x1 : A1)..(xn : An) -> Prop@,
       --   and the optional name is the constructor's name.
@@ -160,7 +181,7 @@ data Declaration
   | UnquoteDecl MutualInfo [DefInfo] [QName] Expr
   | UnquoteDef  [DefInfo] [QName] Expr
   | ScopedDecl ScopeInfo [Declaration]  -- ^ scope annotation
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 class GetDefInfo a where
   getDefInfo :: a -> Maybe DefInfo
@@ -186,11 +207,13 @@ data ModuleApplication
       -- ^ @tel. M args@:  applies @M@ to @args@ and abstracts @tel@.
     | RecordModuleIFS ModuleName
       -- ^ @M {{...}}@
-  deriving (Typeable, Data, Show, Eq)
+  deriving (Data, Show, Eq)
 
 data Pragma
   = OptionsPragma [String]
-  | BuiltinPragma String Expr
+  | BuiltinPragma String ResolvedName
+    -- ^ 'ResolvedName' is not 'UnknownName'.
+    --   Name can be ambiguous e.g. for built-in constructors.
   | BuiltinNoDefPragma String QName
     -- ^ Builtins that do not come with a definition,
     --   but declare a name for an Agda concept.
@@ -208,13 +231,13 @@ data Pragma
     -- ^ For coinductive records, use pragma instead of regular
     --   @eta-equality@ definition (as it is might make Agda loop).
   | InjectivePragma QName
-  | InlinePragma QName
+  | InlinePragma Bool QName -- INLINE or NOINLINE
   | DisplayPragma QName [NamedArg Pattern] Expr
-  deriving (Typeable, Data, Show, Eq)
+  deriving (Data, Show, Eq)
 
 -- | Bindings that are valid in a @let@.
 data LetBinding
-  = LetBind LetInfo ArgInfo Name Expr Expr
+  = LetBind LetInfo ArgInfo BindName Expr Expr
     -- ^ @LetBind info rel name type defn@
   | LetPatBind LetInfo Pattern Expr
     -- ^ Irrefutable pattern binding.
@@ -223,10 +246,11 @@ data LetBinding
     -- The @ImportDirective@ is for highlighting purposes.
   | LetOpen ModuleInfo ModuleName ImportDirective
     -- ^ only for highlighting and abstractToConcrete
-  | LetDeclaredVariable Name
+  | LetDeclaredVariable BindName
     -- ^ Only used for highlighting. Refers to the first occurrence of
     -- @x@ in @let x : A; x = e@.
-  deriving (Typeable, Data, Show, Eq)
+  deriving (Data, Show, Eq)
+
 
 -- | Only 'Axiom's.
 type TypeSignature  = Declaration
@@ -235,14 +259,15 @@ type Field          = TypeSignature
 
 -- | A lambda binding is either domain free or typed.
 data LamBinding
-  = DomainFree ArgInfo Name   -- ^ . @x@ or @{x}@ or @.x@ or @.{x}@
+  = DomainFree ArgInfo BindName   -- ^ . @x@ or @{x}@ or @.x@ or @.{x}@
   | DomainFull TypedBindings  -- ^ . @(xs:e)@ or @{xs:e}@ or @(let Ds)@
-  deriving (Typeable, Data, Show, Eq)
+  deriving (Data, Show, Eq)
+
 
 -- | Typed bindings with hiding information.
 data TypedBindings = TypedBindings Range (Arg TypedBinding)
             -- ^ . @(xs : e)@ or @{xs : e}@
-  deriving (Typeable, Data, Show, Eq)
+  deriving (Data, Show, Eq)
 
 -- | A typed binding.  Appears in dependent function spaces, typed lambdas, and
 --   telescopes.  It might be tempting to simplify this to only bind a single
@@ -259,40 +284,55 @@ data TypedBindings = TypedBindings Range (Arg TypedBinding)
 --   that the metas of the copy are aliases of the metas of the original.
 
 data TypedBinding
-  = TBind Range [WithHiding Name] Expr
+  = TBind Range [WithHiding BindName] Expr
     -- ^ As in telescope @(x y z : A)@ or type @(x y z : A) -> B@.
   | TLet Range [LetBinding]
     -- ^ E.g. @(let x = e)@ or @(let open M)@.
-  deriving (Typeable, Data, Show, Eq)
+  deriving (Data, Show, Eq)
+
 
 type Telescope  = [TypedBindings]
 
-data NamedDotPattern = NamedDot Name I.Term I.Type
-  deriving (Typeable, Data, Show)
-
-data StrippedDotPattern = StrippedDot Expr I.Term I.Type
-  deriving (Typeable, Data, Show)
+-- | A user pattern together with an internal term that it should be equal to
+--   after splitting is complete.
+--   Special cases:
+--    * User pattern is a variable but internal term isn't:
+--      this will be turned into an as pattern.
+--    * User pattern is a dot pattern:
+--      this pattern won't trigger any splitting but will be checked
+--      for equality after all splitting is complete and as patterns have
+--      been bound.
+--    * User pattern is an absurd pattern:
+--      emptiness of the type will be checked after splitting is complete.
+data ProblemEq = ProblemEq
+  { problemInPat :: Pattern
+  , problemInst  :: I.Term
+  , problemType  :: Dom I.Type
+  } deriving (Data, Show)
 
 -- These are not relevant for caching purposes
-instance Eq NamedDotPattern    where _ == _ = True
-instance Eq StrippedDotPattern where _ == _ = True
+instance Eq ProblemEq where _ == _ = True
 
 -- | We could throw away @where@ clauses at this point and translate them to
 --   @let@. It's not obvious how to remember that the @let@ was really a
 --   @where@ clause though, so for the time being we keep it here.
 data Clause' lhs = Clause
   { clauseLHS        :: lhs
-  , clauseNamedDots  :: [NamedDotPattern]
-      -- ^ Only in with-clauses where we inherit some already checked dot patterns from the parent.
+  , clauseStrippedPats :: [ProblemEq]
+      -- ^ Only in with-clauses where we inherit some already checked patterns from the parent.
       --   These live in the context of the parent clause left-hand side.
-  , clauseStrippedDots :: [StrippedDotPattern]
-      -- ^ In with-clauses where a dot pattern from the parent clause is
-      --   repeated in the with-clause. In this case it's not actually part of
-      --   the clause, but it still needs to be checked (Issue 142).
   , clauseRHS        :: RHS
-  , clauseWhereDecls :: [Declaration]
+  , clauseWhereDecls :: WhereDeclarations
   , clauseCatchall   :: Bool
-  } deriving (Typeable, Data, Show, Functor, Foldable, Traversable, Eq)
+  } deriving (Data, Show, Functor, Foldable, Traversable, Eq)
+
+data WhereDeclarations = WhereDecls
+  { whereModule :: Maybe ModuleName
+  , whereDecls  :: [Declaration]
+  } deriving (Data, Show, Eq)
+
+noWhereDecls :: WhereDeclarations
+noWhereDecls = WhereDecls Nothing []
 
 type Clause = Clause' LHS
 type SpineClause = Clause' SpineLHS
@@ -312,19 +352,23 @@ data RHS
     { rewriteExprs      :: [(QName, Expr)]
       -- ^ The 'QName's are the names of the generated with functions,
       --   one for each 'Expr'.
+    , rewriteStrippedPats :: [ProblemEq]
+      -- ^ The patterns stripped by with-desugaring. These are only present
+      --   if this rewrite follows a with.
     , rewriteRHS        :: RHS
       -- ^ The RHS should not be another @RewriteRHS@.
-    , rewriteWhereDecls :: [Declaration]
+    , rewriteWhereDecls :: WhereDeclarations
       -- ^ The where clauses are attached to the @RewriteRHS@ by
       ---  the scope checker (instead of to the clause).
     }
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
+-- | Ignore 'rhsConcrete' when comparing 'RHS's.
 instance Eq RHS where
   RHS e _          == RHS e' _            = e == e'
   AbsurdRHS        == AbsurdRHS           = True
   WithRHS a b c    == WithRHS a' b' c'    = and [ a == a', b == b', c == c' ]
-  RewriteRHS a b c == RewriteRHS a' b' c' = and [ a == a', b == b', c == c' ]
+  RewriteRHS a b c d == RewriteRHS a' b' c' d' = and [ a == a', b == b', c == c' , d == d' ]
   _                == _                   = False
 
 -- | The lhs of a clause in spine view (inside-out).
@@ -333,139 +377,80 @@ instance Eq RHS where
 data SpineLHS = SpineLHS
   { spLhsInfo     :: LHSInfo             -- ^ Range.
   , spLhsDefName  :: QName               -- ^ Name of function we are defining.
-  , spLhsPats     :: [NamedArg Pattern]  -- ^ Function parameters (patterns).
-  , spLhsWithPats :: [Pattern]           -- ^ @with@ patterns (after @|@).
+  , spLhsPats     :: [NamedArg Pattern]  -- ^ Elimination by pattern, projections, with-patterns.
   }
-  deriving (Typeable, Data, Show, Eq)
-
+  deriving (Data, Show, Eq)
 
+-- | Ignore 'Range' when comparing 'LHS's.
 instance Eq LHS where
-  (LHS _ core wps) == (LHS _ core' wps') = core == core' && wps == wps'
+  LHS _ core == LHS _ core' = core == core'
 
 -- | The lhs of a clause in focused (projection-application) view (outside-in).
 --   Projection patters are represented as 'LHSProj's.
 data LHS = LHS
   { lhsInfo     :: LHSInfo               -- ^ Range.
   , lhsCore     :: LHSCore               -- ^ Copatterns.
-  , lhsWithPats :: [Pattern]             -- ^ @with@ patterns (after @|@).
   }
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
--- | The lhs minus @with@-patterns in projection-application view.
+-- | The lhs in projection-application and with-pattern view.
 --   Parameterised over the type @e@ of dot patterns.
 data LHSCore' e
     -- | The head applied to ordinary patterns.
-  = LHSHead  { lhsDefName  :: QName                    -- ^ Head @f@.
-             , lhsPats     :: [NamedArg (Pattern' e)]  -- ^ Applied to patterns @ps@.
+  = LHSHead  { lhsDefName  :: QName
+                 -- ^ Head @f@.
+             , lhsPats     :: [NamedArg (Pattern' e)]
+                 -- ^ Applied to patterns @ps@.
              }
-    -- | Projection
+    -- | Projection.
   | LHSProj  { lhsDestructor :: AmbiguousQName
-               -- ^ Record projection identifier.
+                 -- ^ Record projection identifier.
              , lhsFocus      :: NamedArg (LHSCore' e)
-               -- ^ Main branch.
-             , lhsPatsRight  :: [NamedArg (Pattern' e)]
-               -- ^ Further applied to patterns.
+                 -- ^ Main argument of projection.
+             , lhsPats       :: [NamedArg (Pattern' e)]
+                 -- ^ Further applied to patterns.
              }
-  deriving (Typeable, Data, Show, Functor, Foldable, Traversable, Eq)
+    -- | With patterns.
+  | LHSWith  { lhsHead         :: LHSCore' e
+                 -- ^ E.g. the 'LHSHead'.
+             , lhsWithPatterns :: [Pattern' e]
+                 -- ^ Applied to with patterns @| p1 | ... | pn@.
+                 --   These patterns are not prefixed with @WithP@!
+             , lhsPats         :: [NamedArg (Pattern' e)]
+                 -- ^ Further applied to patterns.
+             }
+  deriving (Data, Show, Functor, Foldable, Traversable, Eq)
 
 type LHSCore = LHSCore' Expr
 
--- | Convert a focused lhs to spine view and back.
-class LHSToSpine a b where
-  lhsToSpine :: a -> b
-  spineToLhs :: b -> a
-
--- | Clause instance.
-instance LHSToSpine Clause SpineClause where
-  lhsToSpine = fmap lhsToSpine
-  spineToLhs = fmap spineToLhs
-
--- | List instance (for clauses).
-instance LHSToSpine a b => LHSToSpine [a] [b] where
-  lhsToSpine = map lhsToSpine
-  spineToLhs = map spineToLhs
-
--- | LHS instance.
-instance LHSToSpine LHS SpineLHS where
-  lhsToSpine (LHS i core wps) = SpineLHS i f ps wps
-    where QNamed f ps = lhsCoreToSpine core
-  spineToLhs (SpineLHS i f ps wps) = LHS i (spineToLhsCore $ QNamed f ps) wps
-
-lhsCoreToSpine :: LHSCore' e -> A.QNamed [NamedArg (Pattern' e)]
-lhsCoreToSpine (LHSHead f ps) = QNamed f ps
-lhsCoreToSpine (LHSProj d h ps) = (++ (p : ps)) <$> lhsCoreToSpine (namedArg h)
-  where p = updateNamedArg (const $ ProjP patNoRange ProjPrefix d) h
-
-spineToLhsCore :: IsProjP e => QNamed [NamedArg (Pattern' e)] -> LHSCore' e
-spineToLhsCore (QNamed f ps) = lhsCoreAddSpine (LHSHead f []) ps
-
--- | Add applicative patterns (non-projection patterns) to the right.
-lhsCoreApp :: IsProjP e => LHSCore' e -> [NamedArg (Pattern' e)] -> LHSCore' e
-lhsCoreApp (LHSHead f ps)   ps' = LHSHead f   $ ps ++ ps'
-lhsCoreApp (LHSProj d h ps) ps' = LHSProj d h $ ps ++ ps'
-
--- | Add projection and applicative patterns to the right.
-lhsCoreAddSpine :: IsProjP e => LHSCore' e -> [NamedArg (Pattern' e)] -> LHSCore' e
-lhsCoreAddSpine core ps = case ps2 of
-    [] -> lhsCoreApp core ps
-    p@(Arg info (Named n (ProjP i o d))) : ps2' | let nh = numHoles d->
-      -- Andreas, 2016-06-13
-      -- If the projection was written prefix by the user
-      -- or it is fully applied an operator
-      -- we turn it to prefix projection form.
-      (if o == ProjPrefix || nh > 0 && nh <= 1 + length ps2' then
-        LHSProj d (Arg info $ Named n $ lhsCoreApp core ps1) []
-      else lhsCoreApp core $ ps1 ++ [p])
-        `lhsCoreAddSpine` ps2'
-    _ -> __IMPOSSIBLE__
-  where
-    (ps1, ps2) = break (isJust . isProjP) ps
-
--- | Used for checking pattern linearity.
-lhsCoreAllPatterns :: LHSCore' e -> [Pattern' e]
-lhsCoreAllPatterns = map namedArg . qnamed . lhsCoreToSpine
-
--- | Used in AbstractToConcrete.
-lhsCoreToPattern :: LHSCore -> Pattern
-lhsCoreToPattern lc =
-  case lc of
-    LHSHead f aps -> DefP noInfo (AmbQ [f]) aps
-    LHSProj d lhscore aps -> DefP noInfo d $
-      fmap (fmap lhsCoreToPattern) lhscore : aps
-  where noInfo = patNoRange -- TODO, preserve range!
-
-mapLHSHead :: (QName -> [NamedArg Pattern] -> LHSCore) -> LHSCore -> LHSCore
-mapLHSHead f (LHSHead x ps)   = f x ps
-mapLHSHead f (LHSProj d l ps) = LHSProj d (fmap (fmap (mapLHSHead f)) l) ps
-
 ---------------------------------------------------------------------------
 -- * Patterns
 ---------------------------------------------------------------------------
 
 -- | Parameterised over the type of dot patterns.
 data Pattern' e
-  = VarP Name
-  | ConP ConPatInfo AmbiguousQName [NamedArg (Pattern' e)]
+  = VarP BindName
+  | ConP ConPatInfo AmbiguousQName (NAPs e)
   | ProjP PatInfo ProjOrigin AmbiguousQName
     -- ^ Destructor pattern @d@.
-  | DefP PatInfo AmbiguousQName [NamedArg (Pattern' e)]
+  | DefP PatInfo AmbiguousQName (NAPs e)
     -- ^ Defined pattern: function definition @f ps@.
     --   It is also abused to convert destructor patterns into concrete syntax
     --   thus, we put AmbiguousQName here as well.
   | WildP PatInfo
     -- ^ Underscore pattern entered by user.
     --   Or generated at type checking for implicit arguments.
-  | AsP PatInfo Name (Pattern' e)
-  | DotP PatInfo Origin e
-    -- ^ Dot pattern @.e@: the Origin keeps track whether this dot pattern was
-    --   written by the user or inserted by the system (e.g. while expanding
-    --   the ellipsis in a with clause).
+  | AsP PatInfo BindName (Pattern' e)
+  | DotP PatInfo e
+    -- ^ Dot pattern @.e@
   | AbsurdP PatInfo
   | LitP Literal
-  | PatternSynP PatInfo QName [NamedArg (Pattern' e)]
+  | PatternSynP PatInfo AmbiguousQName (NAPs e)
   | RecP PatInfo [FieldAssignment' (Pattern' e)]
-  deriving (Typeable, Data, Show, Functor, Foldable, Traversable, Eq)
+  | WithP PatInfo (Pattern' e)  -- ^ @| p@, for with-patterns.
+  deriving (Data, Show, Functor, Foldable, Traversable, Eq)
 
+type NAPs e   = [NamedArg (Pattern' e)]
 type Pattern  = Pattern' Expr
 type Patterns = [NamedArg Pattern]
 
@@ -482,7 +467,7 @@ class MaybePostfixProjP a where
   maybePostfixProjP :: a -> Maybe (ProjOrigin, AmbiguousQName)
 
 instance IsProjP e => MaybePostfixProjP (Pattern' e) where
-  maybePostfixProjP (DotP _ _ e)  = isProjP e <&> \ (_o, d) -> (ProjPostfix, d)
+  maybePostfixProjP (DotP _ e)    = isProjP e <&> \ (_o, d) -> (ProjPostfix, d)
   maybePostfixProjP (ProjP _ o d) = Just (o, d)
   maybePostfixProjP _ = Nothing
 
@@ -492,6 +477,12 @@ instance MaybePostfixProjP a => MaybePostfixProjP (Arg a) where
 instance MaybePostfixProjP a => MaybePostfixProjP (Named n a) where
   maybePostfixProjP = maybePostfixProjP . namedThing
 
+{--------------------------------------------------------------------------
+    Things we parse but are not part of the Agda file syntax
+ --------------------------------------------------------------------------}
+
+type HoleContent = C.HoleContent' Expr
+
 {--------------------------------------------------------------------------
     Instances
  --------------------------------------------------------------------------}
@@ -639,57 +630,63 @@ instance HasRange Declaration where
     getRange (UnquoteDef i _ _)       = getRange i
 
 instance HasRange (Pattern' e) where
-    getRange (VarP x)            = getRange x
+    getRange (VarP x)           = getRange x
     getRange (ConP i _ _)        = getRange i
     getRange (ProjP i _ _)       = getRange i
     getRange (DefP i _ _)        = getRange i
     getRange (WildP i)           = getRange i
     getRange (AsP i _ _)         = getRange i
-    getRange (DotP i _ _)        = getRange i
+    getRange (DotP i _)          = getRange i
     getRange (AbsurdP i)         = getRange i
     getRange (LitP l)            = getRange l
     getRange (PatternSynP i _ _) = getRange i
     getRange (RecP i _)          = getRange i
+    getRange (WithP i _)         = getRange i
 
 instance HasRange SpineLHS where
-    getRange (SpineLHS i _ _ _)  = getRange i
+    getRange (SpineLHS i _ _)  = getRange i
 
 instance HasRange LHS where
-    getRange (LHS i _ _)   = getRange i
+    getRange (LHS i _)   = getRange i
 
 instance HasRange (LHSCore' e) where
-    getRange (LHSHead f ps) = fuseRange f ps
+    getRange (LHSHead f ps)         = fuseRange f ps
     getRange (LHSProj d lhscore ps) = d `fuseRange` lhscore `fuseRange` ps
+    getRange (LHSWith h wps ps)     = h `fuseRange` wps `fuseRange` ps
 
 instance HasRange a => HasRange (Clause' a) where
-    getRange (Clause lhs _ _ rhs ds catchall) = getRange (lhs, rhs, ds)
+    getRange (Clause lhs _ rhs ds catchall) = getRange (lhs, rhs, ds)
 
 instance HasRange RHS where
     getRange AbsurdRHS                = noRange
     getRange (RHS e _)                = getRange e
     getRange (WithRHS _ e cs)         = fuseRange e cs
-    getRange (RewriteRHS xes rhs wh)  = getRange (map snd xes, rhs, wh)
+    getRange (RewriteRHS xes _ rhs wh) = getRange (map snd xes, rhs, wh)
+
+instance HasRange WhereDeclarations where
+  getRange (WhereDecls _ ds) = getRange ds
 
 instance HasRange LetBinding where
-    getRange (LetBind  i _ _ _ _     ) = getRange i
+    getRange (LetBind i _ _ _ _     ) = getRange i
     getRange (LetPatBind  i _ _      ) = getRange i
     getRange (LetApply i _ _ _ _     ) = getRange i
     getRange (LetOpen  i _ _         ) = getRange i
-    getRange (LetDeclaredVariable x)   = getRange x
+    getRange (LetDeclaredVariable x)  = getRange x
 
 -- setRange for patterns applies the range to the outermost pattern constructor
 instance SetRange (Pattern' a) where
-    setRange r (VarP x)             = VarP (setRange r x)
+    setRange r (VarP x)            = VarP (setRange r x)
     setRange r (ConP i ns as)       = ConP (setRange r i) ns as
     setRange r (ProjP _ o ns)       = ProjP (PatRange r) o ns
     setRange r (DefP _ ns as)       = DefP (PatRange r) ns as -- (setRange r n) as
     setRange r (WildP _)            = WildP (PatRange r)
     setRange r (AsP _ n p)          = AsP (PatRange r) (setRange r n) p
-    setRange r (DotP _ o e)         = DotP (PatRange r) o e
+    setRange r (DotP _ e)           = DotP (PatRange r) e
     setRange r (AbsurdP _)          = AbsurdP (PatRange r)
     setRange r (LitP l)             = LitP (setRange r l)
-    setRange r (PatternSynP _ n as) = PatternSynP (PatRange r) (setRange r n) as
+    setRange r (PatternSynP _ n as) = PatternSynP (PatRange r) n as
     setRange r (RecP i as)          = RecP (PatRange r) as
+    setRange r (WithP i p)          = WithP (setRange r i) p
 
 instance KillRange LamBinding where
   killRange (DomainFree info x) = killRange1 (DomainFree info) x
@@ -763,49 +760,51 @@ instance KillRange ScopeCopyInfo where
   killRange (ScopeCopyInfo a b) = killRange2 ScopeCopyInfo a b
 
 instance KillRange e => KillRange (Pattern' e) where
-  killRange (VarP x)            = killRange1 VarP x
+  killRange (VarP x)           = killRange1 VarP x
   killRange (ConP i a b)        = killRange3 ConP i a b
   killRange (ProjP i o a)       = killRange3 ProjP i o a
   killRange (DefP i a b)        = killRange3 DefP i a b
   killRange (WildP i)           = killRange1 WildP i
   killRange (AsP i a b)         = killRange3 AsP i a b
-  killRange (DotP i o a)        = killRange3 DotP i o a
+  killRange (DotP i a)          = killRange2 DotP i a
   killRange (AbsurdP i)         = killRange1 AbsurdP i
   killRange (LitP l)            = killRange1 LitP l
   killRange (PatternSynP i a p) = killRange3 PatternSynP i a p
   killRange (RecP i as)         = killRange2 RecP i as
+  killRange (WithP i p)         = killRange2 WithP i p
 
 instance KillRange SpineLHS where
-  killRange (SpineLHS i a b c)  = killRange4 SpineLHS i a b c
+  killRange (SpineLHS i a b)  = killRange3 SpineLHS i a b
 
 instance KillRange LHS where
-  killRange (LHS i a b)   = killRange3 LHS i a b
+  killRange (LHS i a)   = killRange2 LHS i a
 
 instance KillRange e => KillRange (LHSCore' e) where
   killRange (LHSHead a b)   = killRange2 LHSHead a b
   killRange (LHSProj a b c) = killRange3 LHSProj a b c
+  killRange (LHSWith a b c) = killRange3 LHSWith a b c
 
 instance KillRange a => KillRange (Clause' a) where
-  killRange (Clause lhs dots sdots rhs ds catchall) = killRange6 Clause lhs dots sdots rhs ds catchall
+  killRange (Clause lhs spats rhs ds catchall) = killRange5 Clause lhs spats rhs ds catchall
 
-instance KillRange NamedDotPattern where
-  killRange (NamedDot a b c) = killRange3 NamedDot a b c
-
-instance KillRange StrippedDotPattern where
-  killRange (StrippedDot a b c) = killRange3 StrippedDot a b c
+instance KillRange ProblemEq where
+  killRange (ProblemEq p v a) = killRange3 ProblemEq p v a
 
 instance KillRange RHS where
   killRange AbsurdRHS                = AbsurdRHS
   killRange (RHS e c)                = killRange2 RHS e c
   killRange (WithRHS q e cs)         = killRange3 WithRHS q e cs
-  killRange (RewriteRHS xes rhs wh)  = killRange3 RewriteRHS xes rhs wh
+  killRange (RewriteRHS xes spats rhs wh) = killRange4 RewriteRHS xes spats rhs wh
+
+instance KillRange WhereDeclarations where
+  killRange (WhereDecls a b) = killRange2 WhereDecls a b
 
 instance KillRange LetBinding where
-  killRange (LetBind    i info a b c) = killRange5 LetBind  i info a b c
+  killRange (LetBind   i info a b c) = killRange5 LetBind i info a b c
   killRange (LetPatBind i a b       ) = killRange3 LetPatBind i a b
   killRange (LetApply   i a b c d   ) = killRange5 LetApply i a b c d
   killRange (LetOpen    i x dir     ) = killRange3 LetOpen  i x dir
-  killRange (LetDeclaredVariable x)   = killRange1 LetDeclaredVariable x
+  killRange (LetDeclaredVariable x)  = killRange1 LetDeclaredVariable x
 
 -- See Agda.Utils.GeniPlate:
 -- Does not descend into ScopeInfo and renaming maps, for instance.
@@ -881,7 +880,10 @@ instance AllNames RHS where
   allNames (RHS e _)                 = allNames e
   allNames AbsurdRHS{}               = Seq.empty
   allNames (WithRHS q _ cls)         = q <| allNames cls
-  allNames (RewriteRHS qes rhs cls) = Seq.fromList (map fst qes) >< allNames rhs >< allNames cls
+  allNames (RewriteRHS qes _ rhs cls) = Seq.fromList (map fst qes) >< allNames rhs >< allNames cls
+
+instance AllNames WhereDeclarations where
+  allNames (WhereDecls _ ds) = allNames ds
 
 instance AllNames Expr where
   allNames Var{}                   = Seq.empty
@@ -928,7 +930,7 @@ instance AllNames TypedBinding where
   allNames (TLet _ lbs)  = allNames lbs
 
 instance AllNames LetBinding where
-  allNames (LetBind _ _ _ e1 e2)   = allNames e1 >< allNames e2
+  allNames (LetBind _ _ _ e1 e2)  = allNames e1 >< allNames e2
   allNames (LetPatBind _ _ e)      = allNames e
   allNames (LetApply _ _ app _ _)  = allNames app
   allNames LetOpen{}               = Seq.empty
@@ -969,62 +971,63 @@ instance AnyAbstract Declaration where
   anyAbstract (RecSig i _ _ _)       = defAbstract i == AbstractDef
   anyAbstract _                      = __IMPOSSIBLE__
 
+class NameToExpr a where
+  nameExpr :: a -> Expr
+
 -- | Turn an 'AbstractName' to an expression.
-nameExpr :: AbstractName -> Expr
-nameExpr d = mk (anameKind d) $ anameName d
-  where
+instance NameToExpr AbstractName where
+  nameExpr d = mk (anameKind d) $ anameName d
+    where
     mk DefName        x = Def x
-    mk FldName        x = Proj ProjSystem $ AmbQ [x]
-    mk ConName        x = Con $ AmbQ [x]
-    mk PatternSynName x = PatternSyn x
+    mk FldName        x = Proj ProjSystem $ unambiguous x
+    mk ConName        x = Con $ unambiguous x
+    mk PatternSynName x = PatternSyn $ unambiguous x
     mk MacroName      x = Macro x
-    mk QuotableName   x = App i (Quote i) (defaultNamedArg $ Def x)
-      where i = ExprRange (getRange x)
+    mk QuotableName   x = App (defaultAppInfo r) (Quote i) (defaultNamedArg $ Def x)
+      where i = ExprRange r
+            r = getRange x
+
+-- | Assumes name is not 'UnknownName'.
+instance NameToExpr ResolvedName where
+  nameExpr = \case
+    VarName x _          -> Var x
+    DefinedName _ x      -> nameExpr x  -- Can be 'DefName', 'MacroName', 'QuotableName'.
+    FieldName xs         -> Proj ProjSystem . AmbQ . fmap anameName $ xs
+    ConstructorName xs   -> Con . AmbQ . fmap anameName $ xs
+    PatternSynResName xs -> PatternSyn . AmbQ . fmap anameName $ xs
+    UnknownName          -> __IMPOSSIBLE__
 
 app :: Expr -> [NamedArg Expr] -> Expr
-app = foldl (App (ExprRange noRange))
+app = foldl (App defaultAppInfo_)
 
 mkLet :: ExprInfo -> [LetBinding] -> Expr -> Expr
 mkLet i [] e = e
 mkLet i ds e = Let i ds e
 
 patternToExpr :: Pattern -> Expr
-patternToExpr (VarP x)            = Var x
+patternToExpr (VarP x)           = Var (unBind x)
 patternToExpr (ConP _ c ps)       =
   Con c `app` map (fmap (fmap patternToExpr)) ps
 patternToExpr (ProjP _ o ds)      = Proj o ds
-patternToExpr (DefP _ (AmbQ [f]) ps) =
-  Def f `app` map (fmap (fmap patternToExpr)) ps
-patternToExpr (DefP _ (AmbQ _) ps) = __IMPOSSIBLE__
+patternToExpr (DefP _ fs ps) =
+  Def (headAmbQ fs) `app` map (fmap (fmap patternToExpr)) ps
 patternToExpr (WildP _)           = Underscore emptyMetaInfo
 patternToExpr (AsP _ _ p)         = patternToExpr p
-patternToExpr (DotP _ _ e)        = e
+patternToExpr (DotP _ e)          = e
 patternToExpr (AbsurdP _)         = Underscore emptyMetaInfo  -- TODO: could this happen?
 patternToExpr (LitP l)            = Lit l
-patternToExpr (PatternSynP _ _ _) = __IMPOSSIBLE__
+patternToExpr (PatternSynP _ c ps) = PatternSyn c `app` (map . fmap . fmap) patternToExpr ps
 patternToExpr (RecP _ as)         = Rec exprNoRange $ map (Left . fmap patternToExpr) as
+patternToExpr (WithP r p)         = __IMPOSSIBLE__
 
 type PatternSynDefn = ([Arg Name], Pattern' Void)
 type PatternSynDefns = Map QName PatternSynDefn
 
 lambdaLiftExpr :: [Name] -> Expr -> Expr
 lambdaLiftExpr []     e = e
-lambdaLiftExpr (n:ns) e = Lam defaultLamInfo_ (DomainFree defaultArgInfo n) $
-                                     lambdaLiftExpr ns e
+lambdaLiftExpr (n:ns) e = Lam exprNoRange (DomainFree defaultArgInfo $ BindName n) $
+                            lambdaLiftExpr ns e
 
-substPattern :: [(Name, Pattern)] -> Pattern -> Pattern
-substPattern s p = case p of
-  VarP z        -> fromMaybe p (lookup z s)
-  ConP i q ps   -> ConP i q (map (fmap (fmap (substPattern s))) ps)
-  RecP i ps     -> RecP i (map (fmap (substPattern s)) ps)
-  ProjP{}       -> p
-  WildP i       -> p
-  DotP i o e    -> DotP i o (substExpr (map (fmap patternToExpr) s) e)
-  AbsurdP i     -> p
-  LitP l        -> p
-  DefP{}        -> p              -- destructor pattern
-  AsP i x p     -> AsP i x (substPattern s p) -- Note: cannot substitute into as-variable
-  PatternSynP{} -> __IMPOSSIBLE__ -- pattern synonyms (already gone)
 
 class SubstExpr a where
   substExpr :: [(Name, Expr)] -> a -> a
