@@ -9,15 +9,19 @@ module Agda.Syntax.Abstract.Name
   , IsNoName(..)
   ) where
 
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ((<>))
+#endif
+
 import Control.DeepSeq
 
-import Data.Foldable (Foldable)
+import Data.Foldable (Foldable, toList)
 import Data.Traversable (Traversable)
 import Data.Data (Data)
-import Data.Typeable (Typeable)
 import Data.List
 import Data.Function
 import Data.Hashable (Hashable(..))
+import Data.Void
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
@@ -28,6 +32,7 @@ import qualified Agda.Syntax.Concrete.Name as C
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.NonemptyList
 import Agda.Utils.Pretty
 import Agda.Utils.Size
 import Agda.Utils.Suffix
@@ -43,7 +48,7 @@ data Name = Name { nameId          :: !NameId
                  , nameBindingSite :: Range
                  , nameFixity      :: Fixity'
                  }
-    deriving (Typeable, Data)
+    deriving Data
 
 -- | Qualified names are non-empty lists of names. Equality on qualified names
 --   are just equality on the last name, i.e. the module part is just
@@ -54,28 +59,45 @@ data Name = Name { nameId          :: !NameId
 data QName = QName { qnameModule :: ModuleName
                    , qnameName   :: Name
                    }
-    deriving (Typeable, Data)
+    deriving Data
 
 -- | Something preceeded by a qualified name.
 data QNamed a = QNamed
   { qname  :: QName
   , qnamed :: a
   }
-  deriving (Typeable, Functor, Foldable, Traversable)
+  deriving (Functor, Foldable, Traversable)
 
 -- | A module name is just a qualified name.
 --
 -- The 'SetRange' instance for module names sets all individual ranges
 -- to the given one.
 newtype ModuleName = MName { mnameToList :: [Name] }
-  deriving (Eq, Ord, Typeable, Data)
+  deriving (Eq, Ord, Data)
 
 -- | Ambiguous qualified names. Used for overloaded constructors.
 --
 -- Invariant: All the names in the list must have the same concrete,
 -- unqualified name.  (This implies that they all have the same 'Range').
-newtype AmbiguousQName = AmbQ { unAmbQ :: [QName] }
-  deriving (Eq, Ord, Typeable, Data)
+newtype AmbiguousQName = AmbQ { unAmbQ :: NonemptyList QName }
+  deriving (Eq, Ord, Data)
+
+-- | A singleton "ambiguous" name.
+unambiguous :: QName -> AmbiguousQName
+unambiguous x = AmbQ (x :! [])
+
+-- | Get the first of the ambiguous names.
+headAmbQ :: AmbiguousQName -> QName
+headAmbQ (AmbQ xs) = headNe xs
+
+-- | Is a name ambiguous.
+isAmbiguous :: AmbiguousQName -> Bool
+isAmbiguous (AmbQ (_ :! xs)) = not (null xs)
+
+-- | Get the name if unambiguous.
+getUnambiguous :: AmbiguousQName -> Maybe QName
+getUnambiguous (AmbQ (x :! [])) = Just x
+getUnambiguous _                = Nothing
 
 -- | Check whether we are a projection pattern.
 class IsProjP a where
@@ -87,6 +109,9 @@ instance IsProjP a => IsProjP (Arg a) where
 instance IsProjP a => IsProjP (Named n a) where
   isProjP = isProjP . namedThing
 
+instance IsProjP Void where
+  isProjP _ = __IMPOSSIBLE__
+
 -- | A module is anonymous if the qualification path ends in an underscore.
 isAnonymousModuleName :: ModuleName -> Bool
 isAnonymousModuleName (MName []) = False
@@ -268,7 +293,7 @@ instance NumHoles QName where
 
 -- | We can have an instance for ambiguous names as all share a common concrete name.
 instance NumHoles AmbiguousQName where
-  numHoles (AmbQ qs) = numHoles $ fromMaybe __IMPOSSIBLE__ $ headMaybe qs
+  numHoles = numHoles . headAmbQ
 
 ------------------------------------------------------------------------
 -- * Show instances (only for debug printing!)
@@ -315,7 +340,7 @@ instance Pretty QName where
   pretty = hcat . punctuate (text ".") . map pretty . qnameToList
 
 instance Pretty AmbiguousQName where
-  pretty (AmbQ qs) = hcat $ punctuate (text " | ") $ map pretty qs
+  pretty (AmbQ qs) = hcat $ punctuate (text " | ") $ map pretty (toList qs)
 
 instance Pretty a => Pretty (QNamed a) where
   pretty (QNamed a b) = pretty a <> text "." <> pretty b
@@ -339,8 +364,7 @@ instance HasRange QName where
 -- | The range of an @AmbiguousQName@ is the range of any of its
 --   disambiguations (they are the same concrete name).
 instance HasRange AmbiguousQName where
-  getRange (AmbQ [])    = noRange
-  getRange (AmbQ (c:_)) = getRange c
+  getRange (AmbQ (c :! _)) = getRange c
 
 -- ** SetRange
 
