@@ -5,7 +5,7 @@
 
 module Agda.Syntax.Abstract.Views where
 
-import Control.Applicative
+import Control.Applicative ( Const(Const), getConst )
 import Control.Arrow (first)
 import Control.Monad.Identity
 
@@ -24,22 +24,28 @@ import Agda.Syntax.Scope.Base (emptyScopeInfo)
 import Agda.Utils.Either
 import Agda.Utils.Lens
 
-data AppView = Application Expr [NamedArg Expr]
+data AppView' arg = Application Expr [NamedArg arg]
+  deriving (Functor)
+
+type AppView = AppView' Expr
 
 -- | Gather applications to expose head and spine.
 --
 --   Note: everything is an application, possibly of itself to 0 arguments
 appView :: Expr -> AppView
-appView e =
+appView = fmap snd . appView'
+
+appView' :: Expr -> AppView' (AppInfo, Expr)
+appView' e =
   case e of
-    App _ e1 e2
+    App i e1 e2
       | Dot _ e2' <- unScope $ namedArg e2
       , Just f <- maybeProjTurnPostfix e2'
-                   -> Application f [defaultNamedArg e1]
+                   -> Application f [defaultNamedArg (i, e1)]
     App i e1 arg
-      | Application hd es <- appView e1
-                   -> Application hd $ es ++ [arg]
-    ScopedExpr _ e -> appView e
+      | Application hd es <- appView' e1
+                   -> Application hd $ es ++ [(fmap . fmap) (i,) arg]
+    ScopedExpr _ e -> appView' e
     _              -> Application e []
 
 maybeProjTurnPostfix :: Expr -> Maybe Expr
@@ -51,20 +57,20 @@ maybeProjTurnPostfix e =
 
 unAppView :: AppView -> Expr
 unAppView (Application h es) =
-  foldl (App (ExprRange noRange)) h es
+  foldl (App defaultAppInfo_) h es
 
 -- | Collects plain lambdas.
-data LamView = LamView [(LamInfo, LamBinding)] Expr
+data LamView = LamView [LamBinding] Expr
 
 lamView :: Expr -> LamView
-lamView (Lam i b e) = cons (i, b) $ lamView e
+lamView (Lam i b e) = cons b $ lamView e
   where cons b (LamView bs e) = LamView (b : bs) e
 lamView (ScopedExpr _ e) = lamView e
 lamView e = LamView [] e
 
 -- | Gather top-level 'AsP'atterns to expose underlying pattern.
 asView :: A.Pattern -> ([Name], A.Pattern)
-asView (A.AsP _ x p) = first (x :) $ asView p
+asView (A.AsP _ x p) = first (unBind x :) $ asView p
 asView p             = ([], p)
 
 -- | Check whether we are dealing with a universe.
@@ -113,13 +119,7 @@ class ExprLike a where
   foldExpr :: Monoid m => (Expr -> m) -> a -> m
   foldExpr f = getConst . recurseExpr (\ pre post -> Const (f pre) <* post)
 
-  traverseExpr
-#if __GLASGOW_HASKELL__ <= 708
-    :: (Applicative m, Monad m)
-#else
-    :: Monad m
-#endif
-    => (Expr -> m Expr) -> a -> m a
+  traverseExpr :: (Applicative m, Monad m) => (Expr -> m Expr) -> a -> m a
   traverseExpr f = recurseExpr (\ pre post -> f =<< post)
 
   mapExpr :: (Expr -> Expr) -> (a -> a)
@@ -320,7 +320,7 @@ instance ExprLike LetBinding where
 instance ExprLike a => ExprLike (Pattern' a) where
 
 instance ExprLike a => ExprLike (Clause' a) where
-  recurseExpr f (Clause lhs dots sdots rhs ds ca) = Clause <$> rec lhs <*> pure dots <*> pure sdots <*> rec rhs <*> rec ds <*> pure ca
+  recurseExpr f (Clause lhs spats rhs ds ca) = Clause <$> rec lhs <*> pure spats <*> rec rhs <*> rec ds <*> pure ca
     where rec = recurseExpr f
 
 instance ExprLike RHS where
@@ -329,9 +329,12 @@ instance ExprLike RHS where
       RHS e c                 -> RHS <$> rec e <*> pure c
       AbsurdRHS{}             -> pure rhs
       WithRHS x es cs         -> WithRHS x <$> rec es <*> rec cs
-      RewriteRHS xes rhs ds   -> RewriteRHS <$> rec xes <*> rec rhs <*> rec ds
+      RewriteRHS xes spats rhs ds -> RewriteRHS <$> rec xes <*> pure spats <*> rec rhs <*> rec ds
     where rec e = recurseExpr f e
 
+instance ExprLike WhereDeclarations where
+  recurseExpr f (WhereDecls a b) = WhereDecls a <$> recurseExpr f b
+
 instance ExprLike ModuleApplication where
   recurseExpr f a =
     case a of
@@ -342,7 +345,7 @@ instance ExprLike ModuleApplication where
 instance ExprLike Pragma where
   recurseExpr f p =
     case p of
-      BuiltinPragma s e           -> BuiltinPragma s <$> rec e
+      BuiltinPragma s x           -> pure p
       OptionsPragma{}             -> pure p
       BuiltinNoDefPragma{}        -> pure p
       RewritePragma{}             -> pure p
@@ -362,12 +365,12 @@ instance ExprLike Pragma where
     where rec e = recurseExpr f e
 
 instance ExprLike LHS where
-  recurseExpr f (LHS i c ps) = LHS i <$> recurseExpr f c <*> recurseExpr f ps
+  recurseExpr f (LHS i p) = LHS i <$> recurseExpr f p
 
 instance ExprLike a => ExprLike (LHSCore' a) where
 
 instance ExprLike SpineLHS where
-  recurseExpr f (SpineLHS i x ps wps) = SpineLHS i x <$> recurseExpr f ps <*> recurseExpr f wps
+  recurseExpr f (SpineLHS i x ps) = SpineLHS i x <$> recurseExpr f ps
 
 instance ExprLike Declaration where
   recurseExpr f d =
