@@ -7,7 +7,6 @@
 -}
 module Agda.Syntax.Common where
 
-import Control.Applicative
 import Control.DeepSeq
 
 import Data.ByteString.Char8 (ByteString)
@@ -18,14 +17,18 @@ import qualified Data.Strict.Maybe as Strict
 import Data.Semigroup hiding (Arg)
 import Data.Traversable
 import Data.Data (Data)
-import Data.Typeable (Typeable)
 import Data.Word
+import Data.IntSet (IntSet)
+import qualified Data.IntSet as IntSet
 
 import GHC.Generics (Generic)
 
 import Agda.Syntax.Position
 
 import Agda.Utils.Functor
+import Agda.Utils.Lens
+import Agda.Utils.PartialOrd
+import Agda.Utils.POMonoid
 import Agda.Utils.Pretty hiding ((<>))
 
 #include "undefined.h"
@@ -37,17 +40,34 @@ import Agda.Utils.Impossible
 
 -- | Used to specify whether something should be delayed.
 data Delayed = Delayed | NotDelayed
-  deriving (Typeable, Data, Show, Eq, Ord)
+  deriving (Data, Show, Eq, Ord)
 
 instance KillRange Delayed where
   killRange = id
 
+---------------------------------------------------------------------------
+-- * Eta-equality
+---------------------------------------------------------------------------
+
+data HasEta = NoEta | YesEta
+  deriving (Data, Show, Eq, Ord)
+
+instance HasRange HasEta where
+  getRange _ = noRange
+
+instance KillRange HasEta where
+  killRange = id
+
+instance NFData HasEta where
+  rnf NoEta  = ()
+  rnf YesEta = ()
+
 ---------------------------------------------------------------------------
 -- * Induction
 ---------------------------------------------------------------------------
 
 data Induction = Inductive | CoInductive
-  deriving (Typeable, Data, Eq, Ord)
+  deriving (Data, Eq, Ord)
 
 instance Show Induction where
   show Inductive   = "inductive"
@@ -68,10 +88,10 @@ instance NFData Induction where
 ---------------------------------------------------------------------------
 
 data Overlappable = YesOverlap | NoOverlap
-  deriving (Typeable, Data, Show, Eq, Ord)
+  deriving (Data, Show, Eq, Ord)
 
 data Hiding  = Hidden | Instance Overlappable | NotHidden
-  deriving (Typeable, Data, Show, Eq, Ord)
+  deriving (Data, Show, Eq, Ord)
 
 -- | Just for the 'Hiding' instance. Should never combine different
 --   overlapping.
@@ -114,7 +134,7 @@ data WithHiding a = WithHiding
   { whHiding :: !Hiding
   , whThing  :: a
   }
-  deriving (Typeable, Data, Eq, Ord, Show, Functor, Foldable, Traversable)
+  deriving (Data, Eq, Ord, Show, Functor, Foldable, Traversable)
 
 instance Decoration WithHiding where
   traverseF f (WithHiding h a) = WithHiding h <$> f a
@@ -209,28 +229,163 @@ sameHiding x y =
     (hx, hy)                 -> hx == hy
 
 ---------------------------------------------------------------------------
--- * Relevance
+-- * Modalities
 ---------------------------------------------------------------------------
 
--- | An constructor argument is big if the sort of its type is bigger than
---   the sort of the data type.  Only parameters (and maybe forced arguments)
---   are allowed to be big.
---   @
---      List : Set -> Set
---      nil  : (A : Set) -> List A
---   @
---   @A@ is big in constructor @nil@ as the sort @Set1@ of its type @Set@
---   is bigger than the sort @Set@ of the data type @List@.
-data Big = Big | Small
-  deriving (Typeable, Data, Show, Eq, Enum, Bounded)
+-- | We have a tuple of modalities, which might not be fully orthogonal.
+--   For instance, irrelevant stuff is also run-time irrelevant.
+data Modality = Modality
+  { modRelevance :: Relevance
+      -- ^ Legacy irrelevance.
+      --   See Pfenning, LiCS 2001; Abel/Vezzosi/Winterhalter, ICFP 2017.
+  , modQuantity  :: Quantity
+      -- ^ Cardinality / runtime erasure.
+      --   See Conor McBride, I got plenty o' nutting, Wadlerfest 2016.
+  } deriving (Data, Eq, Ord, Show, Generic)
 
-instance Ord Big where
-  Big <= Small = False
-  _   <= _     = True
+defaultModality :: Modality
+defaultModality = Modality defaultRelevance defaultQuantity
 
-instance NFData Big where
-  rnf Big   = ()
-  rnf Small = ()
+-- | Pointwise composition.
+instance Semigroup Modality where
+  Modality r q <> Modality r' q' = Modality (r <> r') (q <> q')
+
+-- | Pointwise unit.
+instance Monoid Modality where
+  mempty = Modality mempty mempty
+  mappend = (<>)
+
+-- | Dominance ordering.
+instance PartialOrd Modality where
+  comparable (Modality r q) (Modality r' q') = comparable (r, q) (r', q')
+
+instance POSemigroup Modality where
+instance POMonoid Modality where
+
+-- boilerplate instances
+
+instance KillRange Modality where
+  killRange = id
+
+instance NFData Modality where
+
+class LensModality a where
+
+  getModality :: a -> Modality
+
+  setModality :: Modality -> a -> a
+  setModality = mapModality . const
+
+  mapModality :: (Modality -> Modality) -> a -> a
+  mapModality f a = setModality (f $ getModality a) a
+
+instance LensModality Modality where
+  getModality = id
+  setModality = const
+  mapModality = id
+
+instance LensRelevance Modality where
+  getRelevance = modRelevance
+  setRelevance h m = m { modRelevance = h }
+  mapRelevance f m = m { modRelevance = f (modRelevance m) }
+
+instance LensQuantity Modality where
+  getQuantity = modQuantity
+  setQuantity h m = m { modQuantity = h }
+  mapQuantity f m = m { modQuantity = f (modQuantity m) }
+
+-- default accessors for Relevance
+
+getRelevanceMod :: LensModality a => LensGet Relevance a
+getRelevanceMod = getRelevance . getModality
+
+setRelevanceMod :: LensModality a => LensSet Relevance a
+setRelevanceMod = mapModality . setRelevance
+
+mapRelevanceMod :: LensModality a => LensMap Relevance a
+mapRelevanceMod = mapModality . mapRelevance
+
+-- default accessors for Quantity
+
+getQuantityMod :: LensModality a => LensGet Quantity a
+getQuantityMod = getQuantity . getModality
+
+setQuantityMod :: LensModality a => LensSet Quantity a
+setQuantityMod = mapModality . setQuantity
+
+mapQuantityMod :: LensModality a => LensMap Quantity a
+mapQuantityMod = mapModality . mapQuantity
+
+---------------------------------------------------------------------------
+-- * Quantities
+---------------------------------------------------------------------------
+
+-- | Quantity for linearity.
+data Quantity
+  = Quantity0  -- ^ Zero uses, erased at runtime.
+  -- TODO: | Quantity1  -- ^ Linear use (could be updated destructively).
+  -- (needs postponable constraints between quantities to compute uses).
+  | Quantityω  -- ^ Unrestricted use.
+  deriving (Data, Show, Generic, Eq, Enum, Bounded)
+
+defaultQuantity :: Quantity
+defaultQuantity = Quantityω
+
+-- | Composition of quantities (multiplication).
+--
+-- 'Quantity0' is dominant.
+instance Semigroup Quantity where
+  Quantity0 <> _ = Quantity0
+  _ <> Quantity0 = Quantity0
+  Quantityω <> _ = Quantityω
+  -- _ <> Quantityω = Quantityω  -- redundant
+
+-- | In the absense of finite quantities besides 0, ω is the unit.
+instance Monoid Quantity where
+  mempty = Quantityω
+  mappend = (<>)
+
+-- | Note that the order is @ω ≤ 0@, more relevant is smaller.
+instance Ord Quantity where
+  compare = curry $ \case
+    (Quantityω, Quantityω) -> EQ
+    (Quantityω, Quantity0) -> LT
+    (Quantity0, Quantityω) -> GT
+    (Quantity0, Quantity0) -> EQ
+
+instance PartialOrd Quantity where
+  comparable = comparableOrd
+
+instance POSemigroup Quantity where
+instance POMonoid Quantity where
+
+-- boilerplate instances
+
+class LensQuantity a where
+
+  getQuantity :: a -> Quantity
+
+  setQuantity :: Quantity -> a -> a
+  setQuantity = mapQuantity . const
+
+  mapQuantity :: (Quantity -> Quantity) -> a -> a
+  mapQuantity f a = setQuantity (f $ getQuantity a) a
+
+instance LensQuantity Quantity where
+  getQuantity = id
+  setQuantity = const
+  mapQuantity = id
+
+instance KillRange Quantity where
+  killRange = id
+
+instance NFData Quantity where
+  rnf Quantity0 = ()
+  rnf Quantityω = ()
+
+---------------------------------------------------------------------------
+-- * Relevance
+---------------------------------------------------------------------------
 
 -- | A function argument can be relevant or irrelevant.
 --   See "Agda.TypeChecking.Irrelevance".
@@ -240,32 +395,21 @@ data Relevance
                 --   Therefore, it is irrelevant at run-time.
                 --   It is treated relevantly during equality checking.
   | Irrelevant  -- ^ The argument is irrelevant at compile- and runtime.
-  | Forced Big  -- ^ The argument can be skipped during equality checking
-                --   because its value is already determined by the type.
-                --   If a constructor argument is big, it has to be regarded
-                --   absent, otherwise we get into paradoxes.
-    deriving (Typeable, Data, Show, Eq)
+    deriving (Data, Show, Eq, Enum, Bounded, Generic)
 
 allRelevances :: [Relevance]
-allRelevances =
-  [ Relevant
-  , NonStrict
-  , Irrelevant
-  , Forced Small
-  , Forced Big
-  ]
+allRelevances = [minBound..maxBound]
+
+defaultRelevance :: Relevance
+defaultRelevance = Relevant
 
 instance KillRange Relevance where
   killRange rel = rel -- no range to kill
 
-instance Ord Relevance where
-  (<=) = moreRelevant
-
 instance NFData Relevance where
   rnf Relevant   = ()
   rnf NonStrict  = ()
   rnf Irrelevant = ()
-  rnf (Forced a) = rnf a
 
 -- | A lens to access the 'Relevance' attribute in data structures.
 --   Minimal implementation: @getRelevance@ and one of @setRelevance@ or @mapRelevance@.
@@ -290,40 +434,38 @@ isRelevant a = getRelevance a == Relevant
 isIrrelevant :: LensRelevance a => a -> Bool
 isIrrelevant a = getRelevance a == Irrelevant
 
+isNonStrict :: LensRelevance a => a -> Bool
+isNonStrict a = getRelevance a == NonStrict
+
 -- | Information ordering.
 -- @Relevant  \`moreRelevant\`
---  Forced    \`moreRelevant\`
 --  NonStrict \`moreRelevant\`
 --  Irrelevant@
 moreRelevant :: Relevance -> Relevance -> Bool
-moreRelevant r r' =
-  case (r, r') of
-    -- top
-    (_, Irrelevant) -> True
-    (Irrelevant, _) -> False
-    -- bottom
-    (Relevant, _)   -> True
-    (_, Relevant)   -> False
-    -- second bottom
-    (Forced{}, _)   -> True
-    (_, Forced{})   -> False
-    -- remaining case
-    (NonStrict,NonStrict) -> True
+moreRelevant = (<=)
 
-irrelevant :: Relevance -> Bool
-irrelevant r =
-  case r of
-    Irrelevant -> True
-    NonStrict  -> False
-    Relevant   -> False
-    Forced{}   -> False
+-- | More relevant is smaller.
+instance Ord Relevance where
+  compare = curry $ \case
+    (r, r') | r == r' -> EQ
+    -- top
+    (_, Irrelevant) -> LT
+    (Irrelevant, _) -> GT
+    -- bottom
+    (Relevant, _) -> LT
+    (_, Relevant) -> GT
+    -- redundant case
+    (NonStrict,NonStrict) -> EQ
+
+-- | More relevant is smaller.
+instance PartialOrd Relevance where
+  comparable = comparableOrd
 
 -- | @unusableRelevance rel == True@ iff we cannot use a variable of @rel@.
 unusableRelevance :: LensRelevance a => a -> Bool
 unusableRelevance a = case getRelevance a of
   Irrelevant -> True
   NonStrict  -> True
-  Forced{}   -> False -- @Forced@ has no semantic relevance
   Relevant   -> False
 
 -- | 'Relevance' composition.
@@ -335,9 +477,6 @@ composeRelevance r r' =
     (_, Irrelevant) -> Irrelevant
     (NonStrict, _)  -> NonStrict
     (_, NonStrict)  -> NonStrict
-    (Forced b, Forced b') -> Forced (max b b')  -- prefer Big over Small
-    (Forced b, _)     -> Forced b
-    (_, Forced b)     -> Forced b
     (Relevant, Relevant) -> Relevant
 
 -- | @inverseComposeRelevance r x@ returns the most irrelevant @y@
@@ -350,18 +489,24 @@ inverseComposeRelevance r x =
   case (r, x) of
     (Relevant, x)        -> x          -- going to relevant arg.: nothing changes
     _ | r == x           -> Relevant   -- because Relevant is comp.-neutral
-    (Forced{}, Forced{}) -> Relevant   -- same, but (==) does not ignore Big
-    (Forced{}, x)          -> x
     (Irrelevant, x)      -> Relevant   -- going irrelevant: every thing usable
     (_, Irrelevant)      -> Irrelevant -- otherwise: irrelevant things remain unusable
     (NonStrict, _)       -> Relevant   -- but @NonStrict@s become usable
 
--- | For comparing @Relevance@ ignoring @Forced@.
-ignoreForced :: Relevance -> Relevance
-ignoreForced Forced{}   = Relevant
-ignoreForced Relevant   = Relevant
-ignoreForced NonStrict  = NonStrict
-ignoreForced Irrelevant = Irrelevant
+-- | 'Relevance' forms a semigroup under composition.
+instance Semigroup Relevance where
+  (<>) = composeRelevance
+
+-- | 'Relevant' is the unit.
+instance Monoid Relevance where
+  mempty  = Relevant
+  mappend = (<>)
+
+instance POSemigroup Relevance where
+instance POMonoid Relevance where
+
+instance LeftClosedPOMonoid Relevance where
+  inverseCompose = inverseComposeRelevance
 
 -- | Irrelevant function arguments may appear non-strictly in the codomain type.
 irrToNonStrict :: Relevance -> Relevance
@@ -388,7 +533,7 @@ data Origin
   | Inserted     -- ^ E.g. inserted hidden arguments.
   | Reflected    -- ^ Produced by the reflection machinery.
   | CaseSplit    -- ^ Produced by an interactive case split.
-  deriving (Typeable, Data, Show, Eq, Ord)
+  deriving (Data, Show, Eq, Ord)
 
 instance KillRange Origin where
   killRange = id
@@ -404,7 +549,7 @@ data WithOrigin a = WithOrigin
   { woOrigin :: !Origin
   , woThing  :: a
   }
-  deriving (Typeable, Data, Eq, Ord, Show, Functor, Foldable, Traversable)
+  deriving (Data, Eq, Ord, Show, Functor, Foldable, Traversable)
 
 instance Decoration WithOrigin where
   traverseF f (WithOrigin h a) = WithOrigin h <$> f a
@@ -444,6 +589,61 @@ instance LensOrigin (WithOrigin a) where
   setOrigin h (WithOrigin _ a) = WithOrigin h a
   mapOrigin f (WithOrigin h a) = WithOrigin (f h) a
 
+-----------------------------------------------------------------------------
+-- * Free variable annotations
+-----------------------------------------------------------------------------
+
+data FreeVariables = UnknownFVs | KnownFVs IntSet
+  deriving (Data, Eq, Ord, Show)
+
+instance Semigroup FreeVariables where
+  UnknownFVs   <> _            = UnknownFVs
+  _            <> UnknownFVs   = UnknownFVs
+  KnownFVs vs1 <> KnownFVs vs2 = KnownFVs (IntSet.union vs1 vs2)
+
+instance Monoid FreeVariables where
+  mempty  = KnownFVs IntSet.empty
+  mappend = (<>)
+
+instance NFData FreeVariables where
+  rnf UnknownFVs    = ()
+  rnf (KnownFVs fv) = rnf fv
+
+unknownFreeVariables :: FreeVariables
+unknownFreeVariables = UnknownFVs
+
+noFreeVariables :: FreeVariables
+noFreeVariables = mempty
+
+oneFreeVariable :: Int -> FreeVariables
+oneFreeVariable = KnownFVs . IntSet.singleton
+
+freeVariablesFromList :: [Int] -> FreeVariables
+freeVariablesFromList = mconcat . map oneFreeVariable
+
+-- | A lens to access the 'FreeVariables' attribute in data structures.
+--   Minimal implementation: @getFreeVariables@ and one of @setFreeVariables@ or @mapFreeVariables@.
+class LensFreeVariables a where
+
+  getFreeVariables :: a -> FreeVariables
+
+  setFreeVariables :: FreeVariables -> a -> a
+  setFreeVariables o = mapFreeVariables (const o)
+
+  mapFreeVariables :: (FreeVariables -> FreeVariables) -> a -> a
+  mapFreeVariables f a = setFreeVariables (f $ getFreeVariables a) a
+
+instance LensFreeVariables FreeVariables where
+  getFreeVariables = id
+  setFreeVariables = const
+  mapFreeVariables = id
+
+hasNoFreeVariables :: LensFreeVariables a => a -> Bool
+hasNoFreeVariables x =
+  case getFreeVariables x of
+    UnknownFVs  -> False
+    KnownFVs fv -> IntSet.null fv
+
 ---------------------------------------------------------------------------
 -- * Argument decoration
 ---------------------------------------------------------------------------
@@ -451,13 +651,14 @@ instance LensOrigin (WithOrigin a) where
 -- | A function argument can be hidden and/or irrelevant.
 
 data ArgInfo = ArgInfo
-  { argInfoHiding       :: Hiding
-  , argInfoRelevance    :: Relevance
-  , argInfoOrigin       :: Origin
-  } deriving (Typeable, Data, Eq, Ord, Show)
+  { argInfoHiding        :: Hiding
+  , argInfoModality      :: Modality
+  , argInfoOrigin        :: Origin
+  , argInfoFreeVariables :: FreeVariables
+  } deriving (Data, Eq, Ord, Show)
 
 instance KillRange ArgInfo where
-  killRange (ArgInfo h r o) = killRange3 ArgInfo h r o
+  killRange i = i -- There are no ranges in ArgInfo's
 
 class LensArgInfo a where
   getArgInfo :: a -> ArgInfo
@@ -472,27 +673,93 @@ instance LensArgInfo ArgInfo where
   mapArgInfo = id
 
 instance NFData ArgInfo where
-  rnf (ArgInfo a b c) = rnf a `seq` rnf b `seq` rnf c
+  rnf (ArgInfo a b c d) = rnf a `seq` rnf b `seq` rnf c `seq` rnf d
 
 instance LensHiding ArgInfo where
   getHiding = argInfoHiding
   setHiding h ai = ai { argInfoHiding = h }
   mapHiding f ai = ai { argInfoHiding = f (argInfoHiding ai) }
 
-instance LensRelevance ArgInfo where
-  getRelevance = argInfoRelevance
-  setRelevance h ai = ai { argInfoRelevance = h }
-  mapRelevance f ai = ai { argInfoRelevance = f (argInfoRelevance ai) }
+instance LensModality ArgInfo where
+  getModality = argInfoModality
+  setModality m ai = ai { argInfoModality = m }
+  mapModality f ai = ai { argInfoModality = f (argInfoModality ai) }
 
 instance LensOrigin ArgInfo where
   getOrigin = argInfoOrigin
   setOrigin o ai = ai { argInfoOrigin = o }
   mapOrigin f ai = ai { argInfoOrigin = f (argInfoOrigin ai) }
 
+instance LensFreeVariables ArgInfo where
+  getFreeVariables = argInfoFreeVariables
+  setFreeVariables o ai = ai { argInfoFreeVariables = o }
+  mapFreeVariables f ai = ai { argInfoFreeVariables = f (argInfoFreeVariables ai) }
+
+-- inherited instances
+
+instance LensRelevance ArgInfo where
+  getRelevance = getRelevanceMod
+  setRelevance = setRelevanceMod
+  mapRelevance = mapRelevanceMod
+
+instance LensQuantity ArgInfo where
+  getQuantity = getQuantityMod
+  setQuantity = setQuantityMod
+  mapQuantity = mapQuantityMod
+
 defaultArgInfo :: ArgInfo
-defaultArgInfo =  ArgInfo { argInfoHiding       = NotHidden
-                          , argInfoRelevance    = Relevant
-                          , argInfoOrigin       = UserWritten }
+defaultArgInfo =  ArgInfo
+  { argInfoHiding        = NotHidden
+  , argInfoModality      = defaultModality
+  , argInfoOrigin        = UserWritten
+  , argInfoFreeVariables = UnknownFVs
+  }
+
+-- Accessing through ArgInfo
+
+-- default accessors for Hiding
+
+getHidingArgInfo :: LensArgInfo a => LensGet Hiding a
+getHidingArgInfo = getHiding . getArgInfo
+
+setHidingArgInfo :: LensArgInfo a => LensSet Hiding a
+setHidingArgInfo = mapArgInfo . setHiding
+
+mapHidingArgInfo :: LensArgInfo a => LensMap Hiding a
+mapHidingArgInfo = mapArgInfo . mapHiding
+
+-- default accessors for Modality
+
+getModalityArgInfo :: LensArgInfo a => LensGet Modality a
+getModalityArgInfo = getModality . getArgInfo
+
+setModalityArgInfo :: LensArgInfo a => LensSet Modality a
+setModalityArgInfo = mapArgInfo . setModality
+
+mapModalityArgInfo :: LensArgInfo a => LensMap Modality a
+mapModalityArgInfo = mapArgInfo . mapModality
+
+-- default accessors for Origin
+
+getOriginArgInfo :: LensArgInfo a => LensGet Origin a
+getOriginArgInfo = getOrigin . getArgInfo
+
+setOriginArgInfo :: LensArgInfo a => LensSet Origin a
+setOriginArgInfo = mapArgInfo . setOrigin
+
+mapOriginArgInfo :: LensArgInfo a => LensMap Origin a
+mapOriginArgInfo = mapArgInfo . mapOrigin
+
+-- default accessors for FreeVariables
+
+getFreeVariablesArgInfo :: LensArgInfo a => LensGet FreeVariables a
+getFreeVariablesArgInfo = getFreeVariables . getArgInfo
+
+setFreeVariablesArgInfo :: LensArgInfo a => LensSet FreeVariables a
+setFreeVariablesArgInfo = mapArgInfo . setFreeVariables
+
+mapFreeVariablesArgInfo :: LensArgInfo a => LensMap FreeVariables a
+mapFreeVariablesArgInfo = mapArgInfo . mapFreeVariables
 
 
 ---------------------------------------------------------------------------
@@ -502,7 +769,7 @@ defaultArgInfo =  ArgInfo { argInfoHiding       = NotHidden
 data Arg e  = Arg
   { argInfo :: ArgInfo
   , unArg :: e
-  } deriving (Data, Typeable, Ord, Functor, Foldable, Traversable)
+  } deriving (Data, Ord, Functor, Foldable, Traversable)
 
 instance Decoration Arg where
   traverseF f (Arg ai a) = Arg ai <$> f a
@@ -516,11 +783,19 @@ instance SetRange a => SetRange (Arg a) where
 instance KillRange a => KillRange (Arg a) where
   killRange (Arg info a) = killRange2 Arg info a
 
+-- | Ignores 'Quantity', 'Relevance', 'Origin', and 'FreeVariables'.
+--   Ignores content of argument if 'Irrelevant'.
+--
 instance Eq a => Eq (Arg a) where
-  Arg (ArgInfo h1 _ _) x1 == Arg (ArgInfo h2 _ _) x2 = (h1, x1) == (h2, x2)
+  Arg (ArgInfo h1 m1 _ _) x1 == Arg (ArgInfo h2 m2 _ _) x2 =
+    h1 == h2 && (isIrrelevant m1 || isIrrelevant m2 || x1 == x2)
+    -- Andreas, 2017-10-04, issue #2775, ignore irrelevant arguments during with-abstraction.
+    -- This is a hack, we should not use '(==)' in with-abstraction
+    -- and more generally not use it on Syntax.
+    -- Andrea: except for caching.
 
 instance Show a => Show (Arg a) where
-    show (Arg (ArgInfo h r o) a) = showR r $ showO o $ showH h $ show a
+    show (Arg (ArgInfo h (Modality r q) o fv) a) = showFVs fv $ showQ q $ showR r $ showO o $ showH h $ show a
       where
         showH Hidden       s = "{" ++ s ++ "}"
         showH NotHidden    s = "(" ++ s ++ ")"
@@ -530,34 +805,60 @@ instance Show a => Show (Arg a) where
         showR r s = case r of
           Irrelevant   -> "." ++ s
           NonStrict    -> "?" ++ s
-          Forced Big   -> "!b" ++ s
-          Forced Small -> "!" ++ s
           Relevant     -> "r" ++ s -- Andreas: I want to see it explicitly
+        showQ q s = case q of
+          Quantity0   -> "0" ++ s
+          Quantityω   -> "ω" ++ s
         showO o s = case o of
           UserWritten -> "u" ++ s
           Inserted    -> "i" ++ s
           Reflected   -> "g" ++ s -- generated by reflection
           CaseSplit   -> "c" ++ s -- generated by case split
+        showFVs UnknownFVs    s = s
+        showFVs (KnownFVs fv) s = "fv" ++ show (IntSet.toList fv) ++ s
 
 instance NFData e => NFData (Arg e) where
   rnf (Arg a b) = rnf a `seq` rnf b
 
-instance LensHiding (Arg e) where
-  getHiding = getHiding . argInfo
-  mapHiding = mapArgInfo . mapHiding
-
-instance LensRelevance (Arg e) where
-  getRelevance = getRelevance . argInfo
-  mapRelevance = mapArgInfo . mapRelevance
-
-instance LensOrigin (Arg e) where
-  getOrigin = getOrigin . argInfo
-  mapOrigin = mapArgInfo . mapOrigin
-
 instance LensArgInfo (Arg a) where
   getArgInfo        = argInfo
+  setArgInfo ai arg = arg { argInfo = ai }
   mapArgInfo f arg  = arg { argInfo = f $ argInfo arg }
 
+-- The other lenses are defined through LensArgInfo
+
+instance LensHiding (Arg e) where
+  getHiding = getHidingArgInfo
+  setHiding = setHidingArgInfo
+  mapHiding = mapHidingArgInfo
+
+instance LensModality (Arg e) where
+  getModality = getModalityArgInfo
+  setModality = setModalityArgInfo
+  mapModality = mapModalityArgInfo
+
+instance LensOrigin (Arg e) where
+  getOrigin = getOriginArgInfo
+  setOrigin = setOriginArgInfo
+  mapOrigin = mapOriginArgInfo
+
+instance LensFreeVariables (Arg e) where
+  getFreeVariables = getFreeVariablesArgInfo
+  setFreeVariables = setFreeVariablesArgInfo
+  mapFreeVariables = mapFreeVariablesArgInfo
+
+-- Since we have LensModality, we get relevance and quantity by default
+
+instance LensRelevance (Arg e) where
+  getRelevance = getRelevanceMod
+  setRelevance = setRelevanceMod
+  mapRelevance = mapRelevanceMod
+
+instance LensQuantity (Arg e) where
+  getQuantity = getQuantityMod
+  setQuantity = setQuantityMod
+  mapQuantity = mapQuantityMod
+
 defaultArg :: a -> Arg a
 defaultArg = Arg defaultArgInfo
 
@@ -607,7 +908,7 @@ instance Underscore Doc where
 data Dom e = Dom
   { domInfo   :: ArgInfo
   , unDom     :: e
-  } deriving (Typeable, Data, Ord, Functor, Foldable, Traversable)
+  } deriving (Data, Ord, Functor, Foldable, Traversable)
 
 instance Decoration Dom where
   traverseF f (Dom ai a) = Dom ai <$> f a
@@ -618,28 +919,52 @@ instance HasRange a => HasRange (Dom a) where
 instance KillRange a => KillRange (Dom a) where
   killRange (Dom info a) = killRange2 Dom info a
 
+-- | Ignores 'Origin' and 'FreeVariables'.
 instance Eq a => Eq (Dom a) where
-  Dom (ArgInfo h1 r1 _) x1 == Dom (ArgInfo h2 r2 _) x2 =
-    (h1, ignoreForced r1, x1) == (h2, ignoreForced r2, x2)
+  Dom (ArgInfo h1 m1 _ _) x1 == Dom (ArgInfo h2 m2 _ _) x2 =
+    (h1, m1, x1) == (h2, m2, x2)
 
 instance Show a => Show (Dom a) where
   show = show . argFromDom
 
-instance LensHiding (Dom e) where
-  getHiding = getHiding . domInfo
-  mapHiding = mapArgInfo . mapHiding
-
-instance LensRelevance (Dom e) where
-  getRelevance = getRelevance . domInfo
-  mapRelevance = mapArgInfo . mapRelevance
-
 instance LensArgInfo (Dom e) where
-  getArgInfo = domInfo
-  mapArgInfo f arg = arg { domInfo = f $ domInfo arg }
+  getArgInfo        = domInfo
+  setArgInfo ai dom = dom { domInfo = ai }
+  mapArgInfo f  dom = dom { domInfo = f $ domInfo dom }
+
+-- The other lenses are defined through LensArgInfo
+
+instance LensHiding (Dom e) where
+  getHiding = getHidingArgInfo
+  setHiding = setHidingArgInfo
+  mapHiding = mapHidingArgInfo
+
+instance LensModality (Dom e) where
+  getModality = getModalityArgInfo
+  setModality = setModalityArgInfo
+  mapModality = mapModalityArgInfo
 
 instance LensOrigin (Dom e) where
-  getOrigin = getOrigin . getArgInfo
-  mapOrigin = mapArgInfo . mapOrigin
+  getOrigin = getOriginArgInfo
+  setOrigin = setOriginArgInfo
+  mapOrigin = mapOriginArgInfo
+
+instance LensFreeVariables (Dom e) where
+  getFreeVariables = getFreeVariablesArgInfo
+  setFreeVariables = setFreeVariablesArgInfo
+  mapFreeVariables = mapFreeVariablesArgInfo
+
+-- Since we have LensModality, we get relevance and quantity by default
+
+instance LensRelevance (Dom e) where
+  getRelevance = getRelevanceMod
+  setRelevance = setRelevanceMod
+  mapRelevance = mapRelevanceMod
+
+instance LensQuantity (Dom e) where
+  getQuantity = getQuantityMod
+  setQuantity = setQuantityMod
+  mapQuantity = mapQuantityMod
 
 argFromDom :: Dom a -> Arg a
 argFromDom (Dom i a) = Arg i a
@@ -659,7 +984,7 @@ data Named name a =
     Named { nameOf     :: Maybe name
           , namedThing :: a
           }
-    deriving (Eq, Ord, Typeable, Data, Functor, Foldable, Traversable)
+    deriving (Eq, Ord, Data, Functor, Foldable, Traversable)
 
 -- | Standard naming.
 type Named_ = Named RString
@@ -717,7 +1042,7 @@ data Ranged a = Ranged
   { rangeOf     :: Range
   , rangedThing :: a
   }
-  deriving (Typeable, Data, Functor, Foldable, Traversable)
+  deriving (Data, Functor, Foldable, Traversable)
 
 -- | Thing with no range info.
 unranged :: a -> Ranged a
@@ -772,7 +1097,7 @@ data ConOrigin
   | ConOCon     -- ^ User wrote a constructor (pattern).
   | ConORec     -- ^ User wrote a record (pattern).
   | ConOSplit   -- ^ Generated by interactive case splitting.
-  deriving (Typeable, Data, Show, Eq, Ord, Enum, Bounded)
+  deriving (Data, Show, Eq, Ord, Enum, Bounded)
 
 instance KillRange ConOrigin where
   killRange = id
@@ -787,13 +1112,13 @@ data ProjOrigin
   = ProjPrefix    -- ^ User wrote a prefix projection.
   | ProjPostfix   -- ^ User wrote a postfix projection.
   | ProjSystem    -- ^ Projection was generated by the system.
-  deriving (Typeable, Data, Show, Eq, Ord, Enum, Bounded)
+  deriving (Data, Show, Eq, Ord, Enum, Bounded)
 
 instance KillRange ProjOrigin where
   killRange = id
 
 data DataOrRecord = IsData | IsRecord
-  deriving (Typeable, Data, Eq, Ord, Show)
+  deriving (Data, Eq, Ord, Show)
 
 ---------------------------------------------------------------------------
 -- * Infixity, access, abstract, etc.
@@ -802,7 +1127,7 @@ data DataOrRecord = IsData | IsRecord
 -- | Functions can be defined in both infix and prefix style. See
 --   'Agda.Syntax.Concrete.LHS'.
 data IsInfix = InfixDef | PrefixDef
-    deriving (Typeable, Data, Show, Eq, Ord)
+    deriving (Data, Show, Eq, Ord)
 
 -- | Access modifier.
 data Access
@@ -812,7 +1137,7 @@ data Access
   | PublicAccess
   | OnlyQualified  -- ^ Visible from outside, but not exported when opening the module
                              --   Used for qualified constructors.
-    deriving (Typeable, Data, Show, Eq, Ord)
+    deriving (Data, Show, Eq, Ord)
 
 instance Pretty Access where
   pretty = text . \case
@@ -831,14 +1156,14 @@ instance KillRange Access where
 
 -- | Abstract or concrete
 data IsAbstract = AbstractDef | ConcreteDef
-    deriving (Typeable, Data, Show, Eq, Ord)
+    deriving (Data, Show, Eq, Ord)
 
 instance KillRange IsAbstract where
   killRange = id
 
 -- | Is this definition eligible for instance search?
 data IsInstance = InstanceDef | NotInstanceDef
-    deriving (Typeable, Data, Show, Eq, Ord)
+    deriving (Data, Show, Eq, Ord)
 
 instance KillRange IsInstance where
   killRange = id
@@ -852,7 +1177,7 @@ instance NFData IsInstance where
 
 -- | Is this a macro definition?
 data IsMacro = MacroDef | NotMacroDef
-  deriving (Typeable, Data, Show, Eq, Ord)
+  deriving (Data, Show, Eq, Ord)
 
 instance KillRange IsMacro where killRange = id
 instance HasRange  IsMacro where getRange _ = noRange
@@ -867,7 +1192,7 @@ type Arity  = Nat
 -- | The unique identifier of a name. Second argument is the top-level module
 --   identifier.
 data NameId = NameId {-# UNPACK #-} !Word64 {-# UNPACK #-} !Word64
-    deriving (Eq, Ord, Typeable, Data, Generic)
+    deriving (Eq, Ord, Data, Generic)
 
 instance KillRange NameId where
   killRange = id
@@ -895,7 +1220,7 @@ instance Hashable NameId where
 -- | A meta variable identifier is just a natural number.
 --
 newtype MetaId = MetaId { metaId :: Nat }
-    deriving (Eq, Ord, Num, Real, Enum, Integral, Typeable, Data)
+    deriving (Eq, Ord, Num, Real, Enum, Integral, Data)
 
 instance Pretty MetaId where
   pretty (MetaId n) = text $ "_" ++ show n
@@ -925,7 +1250,7 @@ data PositionInName
     -- @foo_bar@.
   | End
     -- ^ The following underscore is at the end of the name: @foo_@.
-  deriving (Show, Eq, Ord, Typeable, Data)
+  deriving (Show, Eq, Ord, Data)
 
 -- | Placeholders are used to represent the underscores in a section.
 
@@ -934,7 +1259,7 @@ data MaybePlaceholder e
   | NoPlaceholder !(Strict.Maybe PositionInName) e
     -- ^ The second argument is used only (but not always) for name
     -- parts other than underscores.
-  deriving (Typeable, Data, Eq, Ord, Functor, Foldable, Traversable, Show)
+  deriving (Data, Eq, Ord, Functor, Foldable, Traversable, Show)
 
 -- | An abbreviation: @noPlaceholder = 'NoPlaceholder'
 -- 'Strict.Nothing'@.
@@ -966,9 +1291,6 @@ newtype InteractionId = InteractionId { interactionId :: Nat }
              , Real
              , Enum
              , Data
-#if __GLASGOW_HASKELL__ <= 708
-             , Typeable
-#endif
              )
 
 instance Show InteractionId where
@@ -989,10 +1311,10 @@ data ImportDirective' a b = ImportDirective
   , impRenaming    :: [Renaming' a b]
   , publicOpen     :: Bool -- ^ Only for @open@. Exports the opened names from the current module.
   }
-  deriving (Typeable, Data, Eq)
+  deriving (Data, Eq)
 
 data Using' a b = UseEverything | Using [ImportedName' a b]
-  deriving (Typeable, Data, Eq)
+  deriving (Data, Eq)
 
 instance Semigroup (Using' a b) where
   UseEverything <> u             = u
@@ -1015,7 +1337,7 @@ isDefaultImportDir _                                             = False
 data ImportedName' a b
   = ImportedModule  b
   | ImportedName    a
-  deriving (Typeable, Data, Eq, Ord)
+  deriving (Data, Eq, Ord)
 
 setImportedName :: ImportedName' a a -> a -> ImportedName' a a
 setImportedName (ImportedName   x) y = ImportedName   y
@@ -1033,7 +1355,7 @@ data Renaming' a b = Renaming
   , renToRange :: Range
     -- ^ The range of the \"to\" keyword.  Retained for highlighting purposes.
   }
-  deriving (Typeable, Data, Eq)
+  deriving (Data, Eq)
 
 -- ** HasRange instances
 
@@ -1104,7 +1426,7 @@ data TerminationCheck m
     -- ^ Treat as terminating (unsafe).  Same effect as 'NoTerminationCheck'.
   | TerminationMeasure Range m
     -- ^ Skip termination checking but use measure instead.
-    deriving (Typeable, Data, Show, Eq, Functor)
+    deriving (Data, Show, Eq, Functor)
 
 instance KillRange m => KillRange (TerminationCheck m) where
   killRange (TerminationMeasure _ m) = TerminationMeasure noRange (killRange m)
