@@ -1,5 +1,8 @@
 {-# LANGUAGE CPP                #-}
 {-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFunctor      #-}
+{-# LANGUAGE DeriveFoldable     #-}
+{-# LANGUAGE DeriveTraversable  #-}
 
 {-| The concrete syntax is a raw representation of the program text
     without any desugaring at all.  This is what the parser produces.
@@ -36,16 +39,15 @@ module Agda.Syntax.Concrete
   , AsName(..)
   , OpenShortHand(..), RewriteEqn, WithExpr
   , LHS(..), Pattern(..), LHSCore(..)
+  , LamClause(..)
   , RHS, RHS'(..), WhereClause, WhereClause'(..), ExprWhere(..)
+  , DoStmt(..)
   , Pragma(..)
   , Module
   , ThingWithFixity(..)
+  , HoleContent, HoleContent'(..)
   , topLevelModuleName
   , spanAllowedBeforeModule
-    -- * Pattern tools
-  , patternNames, patternQNames
-    -- * Lenses
-  , mapLhsOriginalPattern
   )
   where
 
@@ -58,7 +60,6 @@ import Data.List hiding (null)
 import Data.Set (Set)
 import Data.Monoid
 
-import Data.Typeable (Typeable)
 import Data.Data (Data)
 
 import Agda.Syntax.Position
@@ -83,14 +84,14 @@ data OpApp e
     -- ^ An abstraction inside a special syntax declaration
     --   (see Issue 358 why we introduce this).
   | Ordinary e
-  deriving (Typeable, Data, Functor, Foldable, Traversable)
+  deriving (Data, Functor, Foldable, Traversable)
 
 fromOrdinary :: e -> OpApp e -> e
 fromOrdinary d (Ordinary e) = e
 fromOrdinary d _            = d
 
 data FieldAssignment' a = FieldAssignment { _nameFieldA :: Name, _exprFieldA :: a }
-  deriving (Typeable, Data, Functor, Foldable, Traversable, Show, Eq)
+  deriving (Data, Functor, Foldable, Traversable, Show, Eq)
 
 type FieldAssignment = FieldAssignment' Expr
 
@@ -99,7 +100,7 @@ data ModuleAssignment  = ModuleAssignment
                            , _exprModA      :: [Expr]
                            , _importDirModA :: ImportDirective
                            }
-  deriving (Typeable, Data)
+  deriving Data
 type RecordAssignment  = Either FieldAssignment ModuleAssignment
 type RecordAssignments = [RecordAssignment]
 
@@ -140,7 +141,7 @@ data Expr
   | InstanceArg Range (Named_ Expr)            -- ^ ex: @{{e}}@ or @{{x=e}}@
   | Lam Range [LamBinding] Expr                -- ^ ex: @\\x {y} -> e@ or @\\(x:A){y:B} -> e@
   | AbsurdLam Range Hiding                     -- ^ ex: @\\ ()@
-  | ExtendedLam Range [(LHS,RHS,WhereClause,Bool)]  -- ^ ex: @\\ { p11 .. p1a -> e1 ; .. ; pn1 .. pnz -> en }@
+  | ExtendedLam Range [LamClause]              -- ^ ex: @\\ { p11 .. p1a -> e1 ; .. ; pn1 .. pnz -> en }@
   | Fun Range Expr Expr                        -- ^ ex: @e -> e@ or @.e -> e@ (NYI: @{e} -> e@)
   | Pi Telescope Expr                          -- ^ ex: @(xs:e) -> e@ or @{xs:e} -> e@
   | Set Range                                  -- ^ ex: @Set@
@@ -148,9 +149,10 @@ data Expr
   | SetN Range Integer                         -- ^ ex: @Set0, Set1, ..@
   | Rec Range RecordAssignments                -- ^ ex: @record {x = a; y = b}@, or @record { x = a; M1; M2 }@
   | RecUpdate Range Expr [FieldAssignment]     -- ^ ex: @record e {x = a; y = b}@
-  | Let Range [Declaration] Expr               -- ^ ex: @let Ds in e@
+  | Let Range [Declaration] (Maybe Expr)       -- ^ ex: @let Ds in e@, missing body when parsing do-notation let
   | Paren Range Expr                           -- ^ ex: @(e)@
   | IdiomBrackets Range Expr                   -- ^ ex: @(| e |)@
+  | DoBlock Range [DoStmt]                     -- ^ ex: @do x <- m1; m2@
   | Absurd Range                               -- ^ ex: @()@ or @{}@, only in patterns
   | As Range Name Expr                         -- ^ ex: @x\@p@, only in patterns
   | Dot Range Expr                             -- ^ ex: @.p@, only in patterns
@@ -163,7 +165,8 @@ data Expr
   | Unquote Range                              -- ^ ex: @unquote@, should be applied to a term of type @Term@
   | DontCare Expr                              -- ^ to print irrelevant things
   | Equal Range Expr Expr                      -- ^ ex: @a = b@, used internally in the parser
-  deriving (Typeable, Data)
+  | Ellipsis Range                             -- ^ @...@, used internally to parse patterns.
+  deriving Data
 
 -- | Concrete patterns. No literals in patterns at the moment.
 data Pattern
@@ -183,20 +186,25 @@ data Pattern
   | WildP Range                            -- ^ @_@
   | AbsurdP Range                          -- ^ @()@
   | AsP Range Name Pattern                 -- ^ @x\@p@ unused
-  | DotP Range Origin Expr                 -- ^ @.e@ (the Origin keeps track
-                                           -- whether this dot pattern was
-                                           -- written by the user or inserted
-                                           -- by the system)
+  | DotP Range Expr                        -- ^ @.e@
   | LitP Literal                           -- ^ @0@, @1@, etc.
   | RecP Range [FieldAssignment' Pattern]  -- ^ @record {x = p; y = q}@
-  deriving (Typeable, Data)
+  | EllipsisP Range                        -- ^ @...@, only as left-most pattern.
+  | WithP Range Pattern                    -- ^ @| p@, for with-patterns.
+  deriving Data
+
+data DoStmt
+  = DoBind Range Pattern Expr [LamClause]   -- ^ @p ← e where cs@
+  | DoThen Expr
+  | DoLet Range [Declaration]
+  deriving Data
 
 -- | A lambda binding is either domain free or typed.
 type LamBinding = LamBinding' TypedBindings
 data LamBinding' a
   = DomainFree ArgInfo BoundName  -- ^ . @x@ or @{x}@ or @.x@ or @.{x}@ or @{.x}@
   | DomainFull a                  -- ^ . @(xs : e)@ or @{xs : e}@
-  deriving (Typeable, Data, Functor, Foldable, Traversable)
+  deriving (Data, Functor, Foldable, Traversable)
 
 
 -- | A sequence of typed bindings with hiding information. Appears in dependent
@@ -209,14 +217,14 @@ type TypedBindings = TypedBindings' TypedBinding
 
 data TypedBindings' a = TypedBindings Range (Arg a)
      -- ^ . @(xs : e)@ or @{xs : e}@ or something like @(x {y} _ : e)@.
-  deriving (Typeable, Data, Functor, Foldable, Traversable)
+  deriving (Data, Functor, Foldable, Traversable)
 
 data BoundName = BName
   { boundName   :: Name
   , boundLabel  :: Name    -- ^ for implicit function types the label matters and can't be alpha-renamed
   , bnameFixity :: Fixity'
   }
-  deriving (Typeable, Data, Eq, Show)
+  deriving (Data, Eq, Show)
 
 mkBoundName_ :: Name -> BoundName
 mkBoundName_ x = mkBoundName x noFixity'
@@ -231,7 +239,7 @@ type TypedBinding = TypedBinding' Expr
 data TypedBinding' e
   = TBind Range [WithHiding BoundName] e  -- ^ Binding @(x1 ... xn : A)@.
   | TLet  Range [Declaration]  -- ^ Let binding @(let Ds)@ or @(open M args)@.
-  deriving (Typeable, Data, Functor, Foldable, Traversable)
+  deriving (Data, Functor, Foldable, Traversable)
 
 -- | A telescope is a sequence of typed bindings. Bound variables are in scope
 --   in later types.
@@ -251,38 +259,37 @@ countTelVars tel =
 
    We use fixity information to see which name is actually defined.
 -}
-data LHS
-  = LHS { lhsOriginalPattern :: Pattern       -- ^ @f ps@
-        , lhsWithPattern     :: [Pattern]     -- ^ @| p@ (many)
-        , lhsRewriteEqn      :: [RewriteEqn]  -- ^ @rewrite e@ (many)
-        , lhsWithExpr        :: [WithExpr]    -- ^ @with e@ (many)
-        }
-    -- ^ original pattern, with-patterns, rewrite equations and with-expressions
-  | Ellipsis Range [Pattern] [RewriteEqn] [WithExpr]
-    -- ^ new with-patterns, rewrite equations and with-expressions
-  deriving (Typeable, Data)
+data LHS = LHS
+  { lhsOriginalPattern :: Pattern       -- ^ e.g. @f ps | wps@
+  , lhsRewriteEqn      :: [RewriteEqn]  -- ^ @rewrite e@ (many)
+  , lhsWithExpr        :: [WithExpr]    -- ^ @with e@ (many)
+  } -- ^ Original pattern (including with-patterns), rewrite equations and with-expressions.
+  deriving Data
 
 type RewriteEqn = Expr
 type WithExpr   = Expr
 
--- | Processed (scope-checked) intermediate form of the core @f ps@ of 'LHS'.
+-- | Processed (operator-parsed) intermediate form of the core @f ps@ of 'LHS'.
 --   Corresponds to 'lhsOriginalPattern'.
 data LHSCore
-  = LHSHead  { lhsDefName  :: QName               -- ^ @f@
-             , lhsPats     :: [NamedArg Pattern]  -- ^ @ps@
+  = LHSHead  { lhsDefName      :: QName               -- ^ @f@
+             , lhsPats         :: [NamedArg Pattern]  -- ^ @ps@
              }
-  | LHSProj  { lhsDestructor :: QName      -- ^ record projection identifier
-             , lhsPatsLeft   :: [NamedArg Pattern]  -- ^ side patterns
-             , lhsFocus      :: NamedArg LHSCore    -- ^ main branch
-             , lhsPatsRight  :: [NamedArg Pattern]  -- ^ side patterns
+  | LHSProj  { lhsDestructor   :: QName               -- ^ Record projection.
+             , lhsPatsLeft     :: [NamedArg Pattern]  -- ^ Patterns for record indices (currently none).
+             , lhsFocus        :: NamedArg LHSCore    -- ^ Main argument.
+             , lhsPats         :: [NamedArg Pattern]  -- ^ More application patterns.
+             }
+  | LHSWith  { lhsHead         :: LHSCore
+             , lhsWithPatterns :: [Pattern]          -- ^ Non-empty; at least one @(| p)@.
+             , lhsPats         :: [NamedArg Pattern] -- ^ More application patterns.
              }
-  deriving (Typeable)
 
 type RHS = RHS' Expr
 data RHS' e
   = AbsurdRHS -- ^ No right hand side because of absurd match.
   | RHS e
-  deriving (Typeable, Data, Functor, Foldable, Traversable)
+  deriving (Data, Functor, Foldable, Traversable)
 
 
 type WhereClause = WhereClause' [Declaration]
@@ -293,7 +300,13 @@ data WhereClause' decls
     -- ^ Named where: @module M where@.
     --   The 'Access' flag applies to the 'Name' (not the module contents!)
     --   and is propagated from the parent function.
-  deriving (Typeable, Data, Functor, Foldable, Traversable)
+  deriving (Data, Functor, Foldable, Traversable)
+
+data LamClause = LamClause { lamLHS      :: LHS
+                           , lamRHS      :: RHS
+                           , lamWhere    :: WhereClause -- ^ always 'NoWhere' (see parser)
+                           , lamCatchAll :: Bool }
+  deriving Data
 
 -- | An expression followed by a where clause.
 --   Currently only used to give better a better error message in interaction.
@@ -314,7 +327,7 @@ data AsName = AsName
   , asRange :: Range
     -- ^ The range of the \"as\" keyword.  Retained for highlighting purposes.
   }
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 {--------------------------------------------------------------------------
     Declarations
@@ -338,7 +351,7 @@ data Declaration
   | DataSig     Range Induction Name [LamBinding] Expr -- ^ lone data signature in mutual block
   | Data        Range Induction Name [LamBinding] (Maybe Expr) [TypeSignatureOrInstanceBlock]
   | RecordSig   Range Name [LamBinding] Expr -- ^ lone record signature in mutual block
-  | Record      Range Name (Maybe (Ranged Induction)) (Maybe Bool) (Maybe (Name, IsInstance)) [LamBinding] (Maybe Expr) [Declaration]
+  | Record      Range Name (Maybe (Ranged Induction)) (Maybe HasEta) (Maybe (Name, IsInstance)) [LamBinding] (Maybe Expr) [Declaration]
     -- ^ The optional name is a name for the record constructor.
   | Infix Fixity [Name]
   | Syntax      Name Notation -- ^ notation declaration for a name
@@ -360,23 +373,23 @@ data Declaration
   | UnquoteDecl Range [Name] Expr
   | UnquoteDef  Range [Name] Expr
   | Pragma      Pragma
-  deriving (Typeable, Data)
+  deriving Data
 
 data ModuleApplication
   = SectionApp Range [TypedBindings] Expr
     -- ^ @tel. M args@
   | RecordModuleIFS Range QName
     -- ^ @M {{...}}@
-  deriving (Typeable, Data)
+  deriving Data
 
 data OpenShortHand = DoOpen | DontOpen
-  deriving (Typeable, Data, Eq, Show)
+  deriving (Data, Eq, Show)
 
 -- Pragmas ----------------------------------------------------------------
 
 data Pragma
   = OptionsPragma             Range [String]
-  | BuiltinPragma             Range String Expr
+  | BuiltinPragma             Range String QName
   | RewritePragma             Range [QName]
   | CompiledDataPragma        Range QName String [String]
   | CompiledTypePragma        Range QName String
@@ -390,21 +403,28 @@ data Pragma
   | CompilePragma             Range String QName String -- ^ first string is backend name
   | StaticPragma              Range QName
   | InjectivePragma           Range QName
-  | InlinePragma              Range QName
+  | InlinePragma              Range Bool QName  -- ^ INLINE or NOINLINE
   | ImportPragma              Range String
     -- ^ Invariant: The string must be a valid Haskell module name.
   | ImportUHCPragma           Range String
     -- ^ same as above, but for the UHC backend
   | ImpossiblePragma          Range
+    -- ^ Throws an internal error in the scope checker.
   | EtaPragma                 Range QName
     -- ^ For coinductive records, use pragma instead of regular
     --   @eta-equality@ definition (as it is might make Agda loop).
   | TerminationCheckPragma    Range (TerminationCheck Name)
+    -- ^ Applies to the following function (and all that are mutually recursive with it)
+    --   or to the functions in the following mutual block.
+  | WarningOnUsage            Range QName String
+    -- ^ Applies to the named function
   | CatchallPragma            Range
+    -- ^ Applies to the following function clause.
   | DisplayPragma             Range Pattern Expr
   | NoPositivityCheckPragma   Range
+    -- ^ Applies to the following data/record type or mutual block.
   | PolarityPragma            Range Name [Occurrence]
-  deriving (Typeable, Data)
+  deriving Data
 
 ---------------------------------------------------------------------------
 
@@ -433,6 +453,7 @@ spanAllowedBeforeModule :: [Declaration] -> ([Declaration], [Declaration])
 spanAllowedBeforeModule = span isAllowedBeforeModule
   where
     isAllowedBeforeModule (Pragma OptionsPragma{}) = True
+    isAllowedBeforeModule (Pragma BuiltinPragma{}) = True
     isAllowedBeforeModule (Private _ _ ds) = all isAllowedBeforeModule ds
     isAllowedBeforeModule Import{}       = True
     isAllowedBeforeModule ModuleMacro{}  = True
@@ -440,13 +461,16 @@ spanAllowedBeforeModule = span isAllowedBeforeModule
     isAllowedBeforeModule _              = False
 
 {--------------------------------------------------------------------------
-    Lenses
+    Things we parse but are not part of the Agda file syntax
  --------------------------------------------------------------------------}
 
-mapLhsOriginalPattern :: (Pattern -> Pattern) -> LHS -> LHS
-mapLhsOriginalPattern f lhs@Ellipsis{}                    = lhs
-mapLhsOriginalPattern f lhs@LHS{ lhsOriginalPattern = p } =
-  lhs { lhsOriginalPattern = f p }
+-- | Extended content of an interaction hole.
+data HoleContent' e
+  = HoleContentExpr    e   -- ^ @e@
+  | HoleContentRewrite [e] -- ^ @rewrite e0 | ... | en@
+  deriving (Functor, Foldable, Traversable)
+
+type HoleContent = HoleContent' Expr
 
 {--------------------------------------------------------------------------
     Views
@@ -468,33 +492,6 @@ appView e =
     arg (InstanceArg _ e) = makeInstance $ defaultArg e
     arg e                 = defaultArg (unnamed e)
 
-{--------------------------------------------------------------------------
-    Patterns
- --------------------------------------------------------------------------}
-
--- | Get all the identifiers in a pattern in left-to-right order.
-patternQNames :: Pattern -> [QName]
-patternQNames p =
-  case p of
-    IdentP x               -> [x]
-    AppP p p'              -> concatMap patternQNames [p, namedArg p']
-    RawAppP _ ps           -> concatMap patternQNames  ps
-    OpAppP _ x _ ps        -> x : concatMap (patternQNames . namedArg) ps
-    HiddenP _ (namedPat)   -> patternQNames (namedThing namedPat)
-    ParenP _ p             -> patternQNames p
-    WildP _                -> []
-    AbsurdP _              -> []
-    AsP _ x p              -> patternQNames p
-    DotP{}                 -> []
-    LitP _                 -> []
-    QuoteP _               -> []
-    InstanceP _ (namedPat) -> patternQNames (namedThing namedPat)
-    RecP _ fs              -> concatMap (patternQNames . (^. exprFieldA)) fs
-
--- | Get all the identifiers in a pattern in left-to-right order.
-patternNames :: Pattern -> [Name]
-patternNames = map unqualify . patternQNames
-
 {--------------------------------------------------------------------------
     Instances
  --------------------------------------------------------------------------}
@@ -558,6 +555,7 @@ instance HasRange Expr where
       Let r _ _          -> r
       Paren r _          -> r
       IdiomBrackets r _  -> r
+      DoBlock r _        -> r
       As r _ _           -> r
       Dot r _            -> r
       Absurd r           -> r
@@ -574,6 +572,7 @@ instance HasRange Expr where
       Tactic r _ _       -> r
       DontCare{}         -> noRange
       Equal r _ _        -> r
+      Ellipsis r         -> r
 
 -- instance HasRange Telescope where
 --     getRange (TeleBind bs) = getRange bs
@@ -635,17 +634,25 @@ instance HasRange Declaration where
   getRange (Pragma p)              = getRange p
 
 instance HasRange LHS where
-  getRange (LHS p ps eqns ws) = fuseRange p (fuseRange ps (eqns ++ ws))
-  getRange (Ellipsis r _ _ _) = r
+  getRange (LHS p eqns ws) = fuseRange p (eqns ++ ws)
 
 instance HasRange LHSCore where
   getRange (LHSHead f ps)              = fuseRange f ps
   getRange (LHSProj d ps1 lhscore ps2) = d `fuseRange` ps1 `fuseRange` lhscore `fuseRange` ps2
+  getRange (LHSWith f wps ps)          = f `fuseRange` wps `fuseRange` ps
 
 instance HasRange RHS where
   getRange AbsurdRHS = noRange
   getRange (RHS e)   = getRange e
 
+instance HasRange LamClause where
+  getRange (LamClause lhs rhs wh _) = getRange (lhs, rhs, wh)
+
+instance HasRange DoStmt where
+  getRange (DoBind r _ _ _) = r
+  getRange (DoThen e)       = getRange e
+  getRange (DoLet r _)      = r
+
 instance HasRange Pragma where
   getRange (OptionsPragma r _)               = r
   getRange (BuiltinPragma r _ _)             = r
@@ -662,12 +669,13 @@ instance HasRange Pragma where
   getRange (ForeignPragma r _ _)             = r
   getRange (StaticPragma r _)                = r
   getRange (InjectivePragma r _)             = r
-  getRange (InlinePragma r _)                = r
+  getRange (InlinePragma r _ _)              = r
   getRange (ImportPragma r _)                = r
   getRange (ImportUHCPragma r _)             = r
   getRange (ImpossiblePragma r)              = r
   getRange (EtaPragma r _)                   = r
   getRange (TerminationCheckPragma r _)      = r
+  getRange (WarningOnUsage r _ _)            = r
   getRange (CatchallPragma r)                = r
   getRange (DisplayPragma r _ _)             = r
   getRange (NoPositivityCheckPragma r)       = r
@@ -689,8 +697,10 @@ instance HasRange Pattern where
   getRange (QuoteP r)         = r
   getRange (HiddenP r _)      = r
   getRange (InstanceP r _)    = r
-  getRange (DotP r _ _)       = r
+  getRange (DotP r _)         = r
   getRange (RecP r _)         = r
+  getRange (EllipsisP r)      = r
+  getRange (WithP r _)        = r
 
 -- SetRange instances
 ------------------------------------------------------------------------
@@ -711,8 +721,10 @@ instance SetRange Pattern where
   setRange r (QuoteP _)         = QuoteP r
   setRange r (HiddenP _ p)      = HiddenP r p
   setRange r (InstanceP _ p)    = InstanceP r p
-  setRange r (DotP _ o e)       = DotP r o e
+  setRange r (DotP _ e)         = DotP r e
   setRange r (RecP _ fs)        = RecP r fs
+  setRange r (EllipsisP _)      = EllipsisP r
+  setRange r (WithP _ p)        = WithP r p
 
 -- KillRange instances
 ------------------------------------------------------------------------
@@ -779,6 +791,7 @@ instance KillRange Expr where
   killRange (Let _ d e)          = killRange2 (Let noRange) d e
   killRange (Paren _ e)          = killRange1 (Paren noRange) e
   killRange (IdiomBrackets _ e)  = killRange1 (IdiomBrackets noRange) e
+  killRange (DoBlock _ ss)       = killRange1 (DoBlock noRange) ss
   killRange (Absurd _)           = Absurd noRange
   killRange (As _ n e)           = killRange2 (As noRange) n e
   killRange (Dot _ e)            = killRange1 (Dot noRange) e
@@ -791,14 +804,22 @@ instance KillRange Expr where
   killRange (Tactic _ t es)      = killRange2 (Tactic noRange) t es
   killRange (DontCare e)         = killRange1 DontCare e
   killRange (Equal _ x y)        = Equal noRange x y
+  killRange (Ellipsis _)         = Ellipsis noRange
 
 instance KillRange LamBinding where
   killRange (DomainFree i b) = killRange2 DomainFree i b
   killRange (DomainFull t)   = killRange1 DomainFull t
 
 instance KillRange LHS where
-  killRange (LHS p ps r w)     = killRange4 LHS p ps r w
-  killRange (Ellipsis _ p r w) = killRange3 (Ellipsis noRange) p r w
+  killRange (LHS p r w)     = killRange3 LHS p r w
+
+instance KillRange LamClause where
+  killRange (LamClause a b c d) = killRange4 LamClause a b c d
+
+instance KillRange DoStmt where
+  killRange (DoBind r p e w) = killRange4 DoBind r p e w
+  killRange (DoThen e)       = killRange1 DoThen e
+  killRange (DoLet r ds)     = killRange2 DoLet r ds
 
 instance KillRange ModuleApplication where
   killRange (SectionApp _ t e)    = killRange2 (SectionApp noRange) t e
@@ -810,7 +831,7 @@ instance KillRange e => KillRange (OpApp e) where
 
 instance KillRange Pattern where
   killRange (IdentP q)        = killRange1 IdentP q
-  killRange (AppP p n)        = killRange2 AppP p n
+  killRange (AppP p ps)       = killRange2 AppP p ps
   killRange (RawAppP _ p)     = killRange1 (RawAppP noRange) p
   killRange (OpAppP _ n ns p) = killRange3 (OpAppP noRange) n ns p
   killRange (HiddenP _ n)     = killRange1 (HiddenP noRange) n
@@ -819,10 +840,12 @@ instance KillRange Pattern where
   killRange (WildP _)         = WildP noRange
   killRange (AbsurdP _)       = AbsurdP noRange
   killRange (AsP _ n p)       = killRange2 (AsP noRange) n p
-  killRange (DotP _ o e)      = killRange1 (DotP noRange) o e
+  killRange (DotP _ e)        = killRange1 (DotP noRange) e
   killRange (LitP l)          = killRange1 LitP l
   killRange (QuoteP _)        = QuoteP noRange
   killRange (RecP _ fs)       = killRange1 (RecP noRange) fs
+  killRange (EllipsisP _)     = EllipsisP noRange
+  killRange (WithP _ p)       = killRange1 (WithP noRange) p
 
 instance KillRange Pragma where
   killRange (OptionsPragma _ s)               = OptionsPragma noRange s
@@ -838,13 +861,14 @@ instance KillRange Pragma where
   killRange (HaskellCodePragma _ s)           = HaskellCodePragma noRange s
   killRange (StaticPragma _ q)                = killRange1 (StaticPragma noRange) q
   killRange (InjectivePragma _ q)             = killRange1 (InjectivePragma noRange) q
-  killRange (InlinePragma _ q)                = killRange1 (InlinePragma noRange) q
+  killRange (InlinePragma _ b q)              = killRange1 (InlinePragma noRange b) q
   killRange (ImportPragma _ s)                = ImportPragma noRange s
   killRange (ImportUHCPragma _ s)             = ImportUHCPragma noRange s
   killRange (CompilePragma _ b q s)           = killRange1 (\ q -> CompilePragma noRange b q s) q
   killRange (ForeignPragma _ b s)             = ForeignPragma noRange b s
   killRange (ImpossiblePragma _)              = ImpossiblePragma noRange
   killRange (TerminationCheckPragma _ t)      = TerminationCheckPragma noRange (killRange t)
+  killRange (WarningOnUsage _ nm str)         = WarningOnUsage noRange nm str
   killRange (CatchallPragma _)                = CatchallPragma noRange
   killRange (DisplayPragma _ lhs rhs)         = killRange2 (DisplayPragma noRange) lhs rhs
   killRange (EtaPragma _ q)                   = killRange1 (EtaPragma noRange) q
@@ -896,6 +920,7 @@ instance NFData Expr where
   rnf (Let _ a b)        = rnf a `seq` rnf b
   rnf (Paren _ a)        = rnf a
   rnf (IdiomBrackets _ a)= rnf a
+  rnf (DoBlock _ a)      = rnf a
   rnf (Absurd _)         = ()
   rnf (As _ a b)         = rnf a `seq` rnf b
   rnf (Dot _ a)          = rnf a
@@ -908,6 +933,7 @@ instance NFData Expr where
   rnf (Unquote _)        = ()
   rnf (DontCare a)       = rnf a
   rnf (Equal _ a b)      = rnf a `seq` rnf b
+  rnf (Ellipsis _)       = ()
 
 -- | Ranges are not forced.
 
@@ -923,9 +949,11 @@ instance NFData Pattern where
   rnf (WildP _) = ()
   rnf (AbsurdP _) = ()
   rnf (AsP _ a b) = rnf a `seq` rnf b
-  rnf (DotP _ a b) = rnf a `seq` rnf b
+  rnf (DotP _ a) = rnf a
   rnf (LitP a) = rnf a
   rnf (RecP _ a) = rnf a
+  rnf (EllipsisP _) = ()
+  rnf (WithP _ a) = rnf a
 
 -- | Ranges are not forced.
 
@@ -973,12 +1001,13 @@ instance NFData Pragma where
   rnf (ForeignPragma _ b s)             = rnf b `seq` rnf s
   rnf (StaticPragma _ a)                = rnf a
   rnf (InjectivePragma _ a)             = rnf a
-  rnf (InlinePragma _ a)                = rnf a
+  rnf (InlinePragma _ _ a)              = rnf a
   rnf (ImportPragma _ a)                = rnf a
   rnf (ImportUHCPragma _ a)             = rnf a
   rnf (ImpossiblePragma _)              = ()
   rnf (EtaPragma _ a)                   = rnf a
   rnf (TerminationCheckPragma _ a)      = rnf a
+  rnf (WarningOnUsage _ a b)            = rnf a `seq` rnf b
   rnf (CatchallPragma _)                = ()
   rnf (DisplayPragma _ a b)             = rnf a `seq` rnf b
   rnf (NoPositivityCheckPragma _)       = ()
@@ -1015,8 +1044,7 @@ instance NFData a => NFData (OpApp a) where
 -- | Ranges are not forced.
 
 instance NFData LHS where
-  rnf (LHS a b c d)      = rnf a `seq` rnf b `seq` rnf c `seq` rnf d
-  rnf (Ellipsis _ a b c) = rnf a `seq` rnf b `seq` rnf c
+  rnf (LHS a b c) = rnf a `seq` rnf b `seq` rnf c
 
 instance NFData a => NFData (FieldAssignment' a) where
   rnf (FieldAssignment a b) = rnf a `seq` rnf b
@@ -1029,6 +1057,9 @@ instance NFData a => NFData (WhereClause' a) where
   rnf (AnyWhere a)    = rnf a
   rnf (SomeWhere a b c) = rnf a `seq` rnf b `seq` rnf c
 
+instance NFData LamClause where
+  rnf (LamClause a b c d) = rnf (a, b, c, d)
+
 instance NFData a => NFData (LamBinding' a) where
   rnf (DomainFree a b) = rnf a `seq` rnf b
   rnf (DomainFull a)   = rnf a
@@ -1039,3 +1070,8 @@ instance NFData BoundName where
 instance NFData a => NFData (RHS' a) where
   rnf AbsurdRHS = ()
   rnf (RHS a)   = rnf a
+
+instance NFData DoStmt where
+  rnf (DoBind _ p e w) = rnf (p, e, w)
+  rnf (DoThen e)       = rnf e
+  rnf (DoLet _ ds)     = rnf ds
