@@ -24,6 +24,10 @@
 --
 --   * Report basic well-formedness error,
 --     when one of the above transformation fails.
+--     When possible, errors should be deferred to the scope checking phase
+--     (ConcreteToAbstract), where we are in the TCM and can produce more
+--     informative error messages.
+
 
 module Agda.Syntax.Concrete.Definitions
     ( NiceDeclaration(..)
@@ -36,6 +40,7 @@ module Agda.Syntax.Concrete.Definitions
     , notSoNiceDeclarations
     , niceHasAbstract
     , Measure
+    , declarationWarningName
     ) where
 
 import Prelude hiding (null)
@@ -44,23 +49,20 @@ import Control.Arrow ((***), first, second)
 import Control.Applicative hiding (empty)
 import Control.Monad.State
 
-#if __GLASGOW_HASKELL__ <= 708
-import Data.Foldable ( foldMap )
-#endif
-
 import qualified Data.Map as Map
 import Data.Map (Map)
 import Data.Maybe
-import Data.Semigroup ( Semigroup, Monoid, (<>), mempty, mappend )
+import Data.Monoid ( Monoid, mempty, mappend )
+import Data.Semigroup ( Semigroup, (<>) )
 import qualified Data.List as List
 import qualified Data.Set as Set
 import Data.Traversable (Traversable, traverse)
 import qualified Data.Traversable as Trav
 
 import Data.Data (Data)
-import Data.Typeable (Typeable)
 
 import Agda.Syntax.Concrete
+import Agda.Syntax.Concrete.Pattern
 import Agda.Syntax.Common hiding (TerminationCheck())
 import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Position
@@ -68,8 +70,11 @@ import Agda.Syntax.Fixity
 import Agda.Syntax.Notation
 import Agda.Syntax.Concrete.Pretty ()
 
+import Agda.Interaction.Options.Warnings
+
 import Agda.TypeChecking.Positivity.Occurrence
 
+import Agda.Utils.AffineHole
 import Agda.Utils.Except ( MonadError(throwError,catchError) )
 import Agda.Utils.Function
 import Agda.Utils.Functor
@@ -142,11 +147,11 @@ data NiceDeclaration
       --   Andreas, 2017-01-01: Because of issue #2372, we add 'IsInstance' here.
       --   An alias should know that it is an instance.
   | DataDef Range Fixity' IsAbstract PositivityCheck Name [LamBinding] [NiceConstructor]
-  | RecDef Range Fixity' IsAbstract PositivityCheck Name (Maybe (Ranged Induction)) (Maybe Bool) (Maybe (ThingWithFixity Name, IsInstance)) [LamBinding] [NiceDeclaration]
+  | RecDef Range Fixity' IsAbstract PositivityCheck Name (Maybe (Ranged Induction)) (Maybe HasEta) (Maybe (ThingWithFixity Name, IsInstance)) [LamBinding] [NiceDeclaration]
   | NicePatternSyn Range Fixity' Name [Arg Name] Pattern
   | NiceUnquoteDecl Range [Fixity'] Access IsAbstract IsInstance TerminationCheck [Name] Expr
   | NiceUnquoteDef Range [Fixity'] Access IsAbstract TerminationCheck [Name] Expr
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 type TerminationCheck = Common.TerminationCheck Measure
 
@@ -164,18 +169,17 @@ type NiceTypeSignature  = NiceDeclaration
 -- | One clause in a function definition. There is no guarantee that the 'LHS'
 --   actually declares the 'Name'. We will have to check that later.
 data Clause = Clause Name Catchall LHS RHS WhereClause [Clause]
-    deriving (Typeable, Data, Show)
+    deriving (Data, Show)
 
 -- | The exception type.
 data DeclarationException
         = MultipleFixityDecls [(Name, [Fixity'])]
         | MultiplePolarityPragmas [Name]
+        | MultipleEllipses Pattern
         | InvalidName Name
         | DuplicateDefinition Name
         | MissingDefinition Name
         | MissingWithClauses Name
-        | MissingTypeSignature LHS -- Andreas 2012-06-02: currently unused, remove after a while -- Fredrik 2012-09-20: now used, can we keep it?
-        | MissingDataSignature Name
         | WrongDefinition Name DataRecOrFun DataRecOrFun
         | WrongParameters Name Params Params
           -- ^ 'Name' of symbol, 'Params' of signature, 'Params' of definition.
@@ -184,29 +188,60 @@ data DeclarationException
         | DeclarationPanic String
         | WrongContentBlock KindOfBlock Range
         | AmbiguousFunClauses LHS [Name] -- ^ in a mutual block, a clause could belong to any of the @[Name]@ type signatures
-        | InvalidTerminationCheckPragma Range
         | InvalidMeasureMutual Range
           -- ^ In a mutual block, all or none need a MEASURE pragma.
           --   Range is of mutual block.
         | PragmaNoTerminationCheck Range
           -- ^ Pragma @{-# NO_TERMINATION_CHECK #-}@ has been replaced
           --   by {-# TERMINATING #-} and {-# NON_TERMINATING #-}.
-        | InvalidCatchallPragma Range
         | UnquoteDefRequiresSignature [Name]
         | BadMacroDef NiceDeclaration
-        | InvalidNoPositivityCheckPragma Range
-
-    deriving (Typeable, Data, Show)
+    deriving (Data, Show)
 
 -- | Non-fatal errors encountered in the Nicifier
 data DeclarationWarning
   = UnknownNamesInFixityDecl [Name]
+  | UnknownFixityInMixfixDecl [Name]
   | UnknownNamesInPolarityPragmas [Name]
   | PolarityPragmasButNotPostulates [Name]
   | UselessPrivate Range
   | UselessAbstract Range
   | UselessInstance Range
-  deriving (Typeable, Data, Show)
+  | EmptyMutual Range     -- ^ Empty @mutual@    block.
+  | EmptyAbstract Range   -- ^ Empty @abstract@  block.
+  | EmptyPrivate Range    -- ^ Empty @private@   block.
+  | EmptyInstance Range   -- ^ Empty @instance@  block
+  | EmptyMacro Range      -- ^ Empty @macro@     block.
+  | EmptyPostulate Range  -- ^ Empty @postulate@ block.
+  | InvalidTerminationCheckPragma Range
+      -- ^ A {-# TERMINATING #-} and {-# NON_TERMINATING #-} pragma
+      --   that does not apply to any function.
+  | InvalidNoPositivityCheckPragma Range
+      -- ^ A {-# NO_POSITIVITY_CHECK #-} pragma
+      --   that does not apply to any data or record type.
+  | InvalidCatchallPragma Range
+      -- ^ A {-# CATCHALL #-} pragma
+      --   that does not precede a function clause.
+  deriving (Data, Show)
+
+declarationWarningName :: DeclarationWarning -> WarningName
+declarationWarningName dw = case dw of
+  UnknownNamesInFixityDecl{}        -> UnknownNamesInFixityDecl_
+  UnknownFixityInMixfixDecl{}       -> UnknownFixityInMixfixDecl_
+  UnknownNamesInPolarityPragmas{}   -> UnknownNamesInPolarityPragmas_
+  PolarityPragmasButNotPostulates{} -> PolarityPragmasButNotPostulates_
+  UselessPrivate{}                  -> UselessPrivate_
+  UselessAbstract{}                 -> UselessAbstract_
+  UselessInstance{}                 -> UselessInstance_
+  EmptyMutual{}                     -> EmptyMutual_
+  EmptyAbstract{}                   -> EmptyAbstract_
+  EmptyPrivate{}                    -> EmptyPrivate_
+  EmptyInstance{}                   -> EmptyInstance_
+  EmptyMacro{}                      -> EmptyMacro_
+  EmptyPostulate{}                  -> EmptyPostulate_
+  InvalidTerminationCheckPragma{}   -> InvalidTerminationCheckPragma_
+  InvalidNoPositivityCheckPragma{}  -> InvalidNoPositivityCheckPragma_
+  InvalidCatchallPragma{}           -> InvalidCatchallPragma_
 
 -- | Several declarations expect only type signatures as sub-declarations.  These are:
 data KindOfBlock
@@ -215,18 +250,17 @@ data KindOfBlock
   | InstanceBlock   -- ^ @instance@.  Actually, here all kinds of sub-declarations are allowed a priori.
   | FieldBlock      -- ^ @field@.  Ensured by parser.
   | DataBlock       -- ^ @data ... where@.  Here we got a bad error message for Agda-2.5 (Issue 1698).
-  deriving (Typeable, Data, Eq, Ord, Show)
+  deriving (Data, Eq, Ord, Show)
 
 
 instance HasRange DeclarationException where
   getRange (MultipleFixityDecls xs)             = getRange (fst $ head xs)
   getRange (MultiplePolarityPragmas xs)         = getRange (head xs)
+  getRange (MultipleEllipses d)                 = getRange d
   getRange (InvalidName x)                      = getRange x
   getRange (DuplicateDefinition x)              = getRange x
   getRange (MissingDefinition x)                = getRange x
   getRange (MissingWithClauses x)               = getRange x
-  getRange (MissingTypeSignature x)             = getRange x
-  getRange (MissingDataSignature x)             = getRange x
   getRange (WrongDefinition x k k')             = getRange x
   getRange (WrongParameters x _ _)              = getRange x
   getRange (AmbiguousFunClauses lhs xs)         = getRange lhs
@@ -234,21 +268,28 @@ instance HasRange DeclarationException where
   getRange (Codata r)                           = r
   getRange (DeclarationPanic _)                 = noRange
   getRange (WrongContentBlock _ r)              = r
-  getRange (InvalidTerminationCheckPragma r)    = r
   getRange (InvalidMeasureMutual r)             = r
   getRange (PragmaNoTerminationCheck r)         = r
-  getRange (InvalidCatchallPragma r)            = r
   getRange (UnquoteDefRequiresSignature x)      = getRange x
   getRange (BadMacroDef d)                      = getRange d
-  getRange (InvalidNoPositivityCheckPragma r)   = r
 
 instance HasRange DeclarationWarning where
   getRange (UnknownNamesInFixityDecl xs)        = getRange . head $ xs
+  getRange (UnknownFixityInMixfixDecl xs)       = getRange . head $ xs
   getRange (UnknownNamesInPolarityPragmas xs)   = getRange . head $ xs
   getRange (PolarityPragmasButNotPostulates xs) = getRange . head $ xs
   getRange (UselessPrivate r)                   = r
   getRange (UselessAbstract r)                  = r
   getRange (UselessInstance r)                  = r
+  getRange (EmptyMutual r)                      = r
+  getRange (EmptyAbstract r)                    = r
+  getRange (EmptyPrivate r)                     = r
+  getRange (EmptyInstance r)                    = r
+  getRange (EmptyMacro r)                       = r
+  getRange (EmptyPostulate r)                   = r
+  getRange (InvalidTerminationCheckPragma r)    = r
+  getRange (InvalidNoPositivityCheckPragma r)   = r
+  getRange (InvalidCatchallPragma r)            = r
 
 instance HasRange NiceDeclaration where
   getRange (Axiom r _ _ _ _ _ _ _ _)         = r
@@ -282,6 +323,8 @@ instance Pretty DeclarationException where
         f (x, fs) = pretty x Pretty.<> text ": " <+> fsep (map pretty fs)
   pretty (MultiplePolarityPragmas xs) = fsep $
     pwords "Multiple polarity pragmas for" ++ map pretty xs
+  pretty (MultipleEllipses p) = fsep $
+    pwords "Multiple ellipses in left-hand side" ++ [pretty p]
   pretty (InvalidName x) = fsep $
     pwords "Invalid name:" ++ [pretty x]
   pretty (DuplicateDefinition x) = fsep $
@@ -290,10 +333,7 @@ instance Pretty DeclarationException where
     pwords "Missing definition for" ++ [pretty x]
   pretty (MissingWithClauses x) = fsep $
     pwords "Missing with-clauses for function" ++ [pretty x]
-  pretty (MissingTypeSignature x) = fsep $
-    pwords "Missing type signature for left hand side" ++ [pretty x]
-  pretty (MissingDataSignature x) = fsep $
-    pwords "Missing type signature for " ++ [pretty x]
+
   pretty (WrongDefinition x k k') = fsep $ pretty x :
     pwords ("has been declared as a " ++ show k ++
       ", but is being defined as a " ++ show k')
@@ -314,12 +354,8 @@ instance Pretty DeclarationException where
       _ -> "Unexpected declaration"
   pretty (PragmaNoTerminationCheck _) = fsep $
     pwords "Pragma {-# NO_TERMINATION_CHECK #-} has been removed.  To skip the termination check, label your definitions either as {-# TERMINATING #-} or {-# NON_TERMINATING #-}."
-  pretty (InvalidTerminationCheckPragma _) = fsep $
-    pwords "Termination checking pragmas can only precede a mutual block or a function definition."
   pretty (InvalidMeasureMutual _) = fsep $
     pwords "In a mutual block, either all functions must have the same (or no) termination checking pragma."
-  pretty (InvalidCatchallPragma _) = fsep $
-    pwords "The CATCHALL pragma can only preceed a function clause."
   pretty (UnquoteDefRequiresSignature xs) = fsep $
     pwords "Missing type signatures for unquoteDef" ++ map pretty xs
   pretty (BadMacroDef nd) = fsep $
@@ -330,12 +366,12 @@ instance Pretty DeclarationException where
     "The codata construction has been removed. " ++
     "Use the INFINITY builtin instead."
   pretty (DeclarationPanic s) = text s
-  pretty (InvalidNoPositivityCheckPragma _) = fsep $
-    pwords "No positivity checking pragmas can only precede a mutual block or a data/record definition."
 
 instance Pretty DeclarationWarning where
   pretty (UnknownNamesInFixityDecl xs) = fsep $
     pwords "The following names are not declared in the same scope as their syntax or fixity declaration (i.e., either not in scope at all, imported from another module, or declared in a super module):" ++ map pretty xs
+  pretty (UnknownFixityInMixfixDecl xs) = fsep $
+    pwords "The following mixfix names do not have an associated fixity declaration:" ++ map pretty xs
   pretty (UnknownNamesInPolarityPragmas xs) = fsep $
     pwords "The following names are not declared in the same scope as their polarity pragmas (they could for instance be out of scope, imported from another module, or declared in a super module):" ++ map pretty xs
   pretty (PolarityPragmasButNotPostulates xs) = fsep $
@@ -346,6 +382,18 @@ instance Pretty DeclarationWarning where
     pwords "Using abstract here has no effect. Abstract applies to only definitions like data definitions, record type definitions and function clauses."
   pretty (UselessInstance _)      = fsep $
     pwords "Using instance here has no effect. Instance applies only to declarations that introduce new identifiers into the module, like type signatures and axioms."
+  pretty (EmptyMutual    _) = fsep $ pwords "Empty mutual block."
+  pretty (EmptyAbstract  _) = fsep $ pwords "Empty abstract block."
+  pretty (EmptyPrivate   _) = fsep $ pwords "Empty private block."
+  pretty (EmptyInstance  _) = fsep $ pwords "Empty instance block."
+  pretty (EmptyMacro     _) = fsep $ pwords "Empty macro block."
+  pretty (EmptyPostulate _) = fsep $ pwords "Empty postulate block."
+  pretty (InvalidTerminationCheckPragma _) = fsep $
+    pwords "Termination checking pragmas can only precede a function definition or a mutual block (that contains a function definition)."
+  pretty (InvalidNoPositivityCheckPragma _) = fsep $
+    pwords "No positivity checking pragmas can only precede a data/record definition or a mutual block (that contains a data/record definition)."
+  pretty (InvalidCatchallPragma _) = fsep $
+    pwords "The CATCHALL pragma can only precede a function clause."
 
 declName :: NiceDeclaration -> String
 declName Axiom{}             = "Postulates"
@@ -386,7 +434,7 @@ data DataRecOrFun
     -- ^ Name of a record type with parameters.
   | FunName  TerminationCheck
     -- ^ Name of a function.
-  deriving (Typeable, Data)
+  deriving Data
 
 -- Ignore pragmas when checking equality
 instance Eq DataRecOrFun where
@@ -719,6 +767,11 @@ niceDeclarations ds = do
       -- return $ Map.restrictKeys polarities declared
       return $ Map.filterWithKey (\ k _ -> Set.member k declared) polarities
 
+  -- If we have mixfix identifiers without a corresponding fixity
+  -- declaration, we raise a warning
+  ifNull (Set.filter isOpenMixfix declared Set.\\ Map.keysSet fixs) (return ()) $
+    niceWarning . UnknownFixityInMixfixDecl . Set.toList
+
   -- Run the nicifier in an initial environment of fixity decls
   -- and polarities.  But keep the warnings.
   st <- get
@@ -748,7 +801,7 @@ niceDeclarations ds = do
     declaredNames d = case d of
       TypeSig _ x _        -> [x]
       Field _ x _          -> [x]
-      FunClause (LHS p [] [] []) _ _ _
+      FunClause (LHS p [] []) _ _ _
         | IdentP (QName x) <- removeSingletonRawAppP p
                            -> [x]
       FunClause{}          -> []
@@ -779,25 +832,12 @@ niceDeclarations ds = do
     inferMutualBlocks (d : ds) =
       case declKind d of
         OtherDecl    -> (d :) <$> inferMutualBlocks ds
-        LoneDefs _ xs -> __IMPOSSIBLE__
+        LoneDefs{}   -> (d :) <$> inferMutualBlocks ds  -- Andreas, 2017-10-09, issue #2576: report error in ConcreteToAbstract
         LoneSig k x  -> do
           addLoneSig x k
           ((tcs, pcs), (ds0, ds1)) <- untilAllDefined ([terminationCheck k], [positivityCheck k]) ds
           tc <- combineTermChecks (getRange d) tcs
-
-          -- Record modules are, for performance reasons, not always
-          -- placed in mutual blocks.
-
-          -- ASR (01 January 2016): If the record module has a
-          -- NO_POSITIVITY_CHECK pragma, it is placed in a mutual
-          -- block. See Issue 1760.
-          let prefix :: [NiceDeclaration] -> [NiceDeclaration]
-              prefix = case (d, ds0) of
-                (NiceRecSig{}, [r@(RecDef _ _ _ True _ _ _ _ _ _)]) -> ([d, r] ++)
-                _                                                   ->
-                  (NiceMutual (getRange (d : ds0)) tc (and pcs) (d : ds0) :)
-
-          prefix <$> inferMutualBlocks ds1
+          (NiceMutual (getRange (d : ds0)) tc (and pcs) (d : ds0) :) <$> inferMutualBlocks ds1
       where
         untilAllDefined :: ([TerminationCheck], [PositivityCheck])
                         -> [NiceDeclaration]
@@ -808,8 +848,9 @@ niceDeclarations ds = do
             case ds of
               []     -> __IMPOSSIBLE__ <$ (checkLoneSigs . Map.toList =<< use loneSigs)
               d : ds -> case declKind d of
-                LoneSig k x ->
-                  addLoneSig x k >> cons d (untilAllDefined (terminationCheck k : tc, positivityCheck k : pc) ds)
+                LoneSig k x -> do
+                  addLoneSig x k
+                  cons d (untilAllDefined (terminationCheck k : tc, positivityCheck k : pc) ds)
                 LoneDefs k xs -> do
                   mapM_ removeLoneSig xs
                   cons d (untilAllDefined (terminationCheck k : tc, positivityCheck k : pc) ds)
@@ -832,8 +873,11 @@ niceDeclarations ds = do
       (xs ++) <$> nice ys
 
     nice1 :: [Declaration] -> Nice ([NiceDeclaration], [Declaration])
-    nice1 []     = __IMPOSSIBLE__
-    nice1 (d:ds) = case d of
+    nice1 []     = return ([], []) -- Andreas, 2017-09-16, issue #2759: no longer __IMPOSSIBLE__
+    nice1 (d:ds) = do
+      let justWarning w = do niceWarning w; nice1 ds
+
+      case d of
 
         (TypeSig info x t)            -> do
           termCheck <- use terminationCheckPragma
@@ -862,7 +906,7 @@ niceDeclarations ds = do
             [] -> case lhs of
               -- Subcase: The lhs is single identifier (potentially anonymous).
               -- Treat it as a function clause without a type signature.
-              LHS p [] [] [] | Just x <- isSingleIdentifierP p -> do
+              LHS p [] [] | Just x <- isSingleIdentifierP p -> do
                 d  <- mkFunDef defaultArgInfo termCheck x Nothing [d] -- fun def without type signature is relevant
                 return (d , ds)
               -- Subcase: The lhs is a proper pattern.
@@ -874,7 +918,8 @@ niceDeclarations ds = do
             -- case: clauses match exactly one of the sigs
             [(x,(fits,rest))] -> do
                removeLoneSig x
-               cs  <- mkClauses x (expandEllipsis fits) False
+               ds  <- expandEllipsis fits
+               cs  <- mkClauses x ds False
                fx  <- getFixity x
                return ([FunDef (getRange fits) fits fx ConcreteDef NotInstanceDef termCheck x cs] , rest)
 
@@ -911,21 +956,27 @@ niceDeclarations ds = do
                     niceDeclarations r x ((tel,) <$> mt) (Just (tel, cs))
               <*> return ds
 
+        Mutual r []  -> justWarning $ EmptyMutual r
         Mutual r ds' ->
           (,ds) <$> (singleton <$> (mkOldMutual r =<< nice ds'))
 
+        Abstract r []  -> justWarning $ EmptyAbstract r
         Abstract r ds' ->
           (,ds) <$> (abstractBlock r =<< nice ds')
 
+        Private r UserWritten []  -> justWarning $ EmptyPrivate r
         Private r o ds' ->
           (,ds) <$> (privateBlock r o =<< nice ds')
 
+        InstanceB r []  -> justWarning $ EmptyInstance r
         InstanceB r ds' ->
           (,ds) <$> (instanceBlock r =<< nice ds')
 
+        Macro r []  -> justWarning $ EmptyMacro r
         Macro r ds' ->
           (,ds) <$> (macroBlock r =<< nice ds')
 
+        Postulate r []  -> justWarning $ EmptyPostulate r
         Postulate _ ds' ->
           (,ds) <$> (mapM setPolarity =<< niceAxioms PostulateBlock ds')
           where
@@ -976,8 +1027,9 @@ niceDeclarations ds = do
     nicePragma (TerminationCheckPragma r (TerminationMeasure _ x)) ds =
       if canHaveTerminationMeasure ds then
         withTerminationCheckPragma (TerminationMeasure r x) $ nice1 ds
-      else
-        throwError $ InvalidTerminationCheckPragma r
+      else do
+        niceWarning $ InvalidTerminationCheckPragma r
+        nice1 ds
 
     nicePragma (TerminationCheckPragma r NoTerminationCheck) ds =
       throwError $ PragmaNoTerminationCheck r
@@ -985,20 +1037,23 @@ niceDeclarations ds = do
     nicePragma (TerminationCheckPragma r tc) ds =
       if canHaveTerminationCheckPragma ds then
         withTerminationCheckPragma tc $ nice1 ds
-      else
-        throwError $ InvalidTerminationCheckPragma r
+      else do
+        niceWarning $ InvalidTerminationCheckPragma r
+        nice1 ds
 
     nicePragma (CatchallPragma r) ds =
       if canHaveCatchallPragma ds then
         withCatchallPragma True $ nice1 ds
-      else
-        throwError $ InvalidCatchallPragma r
+      else do
+        niceWarning $ InvalidCatchallPragma r
+        nice1 ds
 
     nicePragma (NoPositivityCheckPragma r) ds =
       if canHaveNoPositivityCheckPragma ds then
         withPositivityCheckPragma False $ nice1 ds
-      else
-        throwError $ InvalidNoPositivityCheckPragma r
+      else do
+        niceWarning $ InvalidNoPositivityCheckPragma r
+        nice1 ds
 
     nicePragma (PolarityPragma{}) ds = return ([], ds)
 
@@ -1014,7 +1069,7 @@ niceDeclarations ds = do
     canHaveTerminationCheckPragma :: [Declaration] -> Bool
     canHaveTerminationCheckPragma []     = False
     canHaveTerminationCheckPragma (d:ds) = case d of
-      Mutual{}      -> True
+      Mutual _ ds   -> any (canHaveTerminationCheckPragma . singleton) ds
       TypeSig{}     -> True
       FunClause{}   -> True
       UnquoteDecl{} -> True
@@ -1031,7 +1086,7 @@ niceDeclarations ds = do
     canHaveNoPositivityCheckPragma :: [Declaration] -> Bool
     canHaveNoPositivityCheckPragma []     = False
     canHaveNoPositivityCheckPragma (d:ds) = case d of
-      Mutual{}                    -> True
+      Mutual _ ds                 -> any (canHaveNoPositivityCheckPragma . singleton) ds
       (Data _ Inductive _ _ _ _)  -> True
       (DataSig _ Inductive _ _ _) -> True
       Record{}                    -> True
@@ -1039,6 +1094,7 @@ niceDeclarations ds = do
       (Pragma p) | isAttachedPragma p -> canHaveNoPositivityCheckPragma ds
       _                           -> False
 
+    -- Pragma that attaches to the following declaration.
     isAttachedPragma :: Pragma -> Bool
     isAttachedPragma p = case p of
       TerminationCheckPragma{}  -> True
@@ -1051,7 +1107,7 @@ niceDeclarations ds = do
     defaultTypeSig :: DataRecOrFun -> Name -> Maybe Expr -> Nice (Maybe Expr)
     defaultTypeSig k x t@Just{} = return t
     defaultTypeSig k x Nothing  = do
-      caseMaybeM (getSig x) (throwError $ MissingDataSignature x) $ \ k' -> do
+      caseMaybeM (getSig x) (return Nothing) $ \ k' -> do
         unless (sameKind k k') $ throwError $ WrongDefinition x k' k
         unless (k == k') $ matchParameters x k' k
         Nothing <$ removeLoneSig x
@@ -1111,7 +1167,8 @@ niceDeclarations ds = do
 
     -- Create a function definition.
     mkFunDef info termCheck x mt ds0 = do
-      cs <- mkClauses x (expandEllipsis ds0) False
+      ds <- expandEllipsis ds0
+      cs <- mkClauses x ds False
       f  <- getFixity x
       return [ FunSig (fuseRange x t) f PublicAccess ConcreteDef NotInstanceDef NotMacroDef info termCheck x t
              , FunDef (getRange ds0) ds0 f ConcreteDef NotInstanceDef termCheck x cs ]
@@ -1123,93 +1180,81 @@ niceDeclarations ds = do
     underscore r = Underscore r Nothing
 
 
-    expandEllipsis :: [Declaration] -> [Declaration]
-    expandEllipsis [] = []
-    expandEllipsis (d@(FunClause Ellipsis{} _ _ _) : ds) =
-      d : expandEllipsis ds
-    expandEllipsis (d@(FunClause lhs@(LHS p ps _ _) _ _ _) : ds) =
-      d : expand (wipe p) (map wipe ps) ds
+    expandEllipsis :: [Declaration] -> Nice [Declaration]
+    expandEllipsis [] = return []
+    expandEllipsis (d@(FunClause lhs@(LHS p _ _) _ _ _) : ds)
+      | hasEllipsis p = (d :) <$> expandEllipsis ds
+      | otherwise     = (d :) <$> expand (killRange p) ds
       where
-        expand _ _ [] = []
-        expand p ps (d@(Pragma (CatchallPragma r)) : ds) = d : expand p ps ds
-        expand p ps (FunClause (Ellipsis r ps' eqs es) rhs wh ca : ds) =
-          FunClause (LHS (setRange r p) ((setRange r ps) ++ ps') eqs es) rhs wh ca
-            : expand p (applyUnless (null es) (++ (map wipe ps')) ps) ds
-                       -- If we have with-expressions (es /= []) then the following
-                       -- ellipses also get the additional with patterns ps'
-        -- We can have ellipses after a fun clause.
-        -- They refer to the last clause that introduced new with-expressions.
-        expand p ps (d@(FunClause (LHS _ _ _ []) _ _ _) : ds) =
-          d : expand p ps ds
-        -- Same here: If we have new with-expressions, the next ellipses will
-        -- refer to us.
-        expand _ _ (d@(FunClause (LHS p' ps' _ (_ : _)) _ _ _) : ds) =
-          d : expand (wipe p') (map wipe ps') ds
-          -- Andreas, Jesper, 2017-05-13, issue #2578
-          -- Need to update the range also on the next with-patterns.
-        expand _ _ (_ : ds) = __IMPOSSIBLE__
-    expandEllipsis (_ : ds) = __IMPOSSIBLE__
-
-    -- Before copying a pattern, remove traces to its origin.
-    wipe :: Pattern -> Pattern
-    wipe = killRange . setInserted
-
-    setInserted :: Pattern -> Pattern
-    setInserted p = case p of
-      IdentP{} -> p
-      QuoteP{} -> p
-      AppP p q -> AppP (setInserted p) (fmap (fmap setInserted) q)
-      RawAppP r ps -> RawAppP r (map setInserted ps)
-      OpAppP r c ns ps -> OpAppP r c ns (map (fmap $ fmap setInserted) ps)
-      HiddenP r p -> HiddenP r (fmap setInserted p)
-      InstanceP r p -> InstanceP r (fmap setInserted p)
-      ParenP r p -> ParenP r (setInserted p)
-      WildP{} -> p
-      AbsurdP{} -> p
-      AsP r n p -> AsP r n (setInserted p)
-      DotP r _ e -> DotP r Inserted e
-      LitP{} -> p
-      RecP r fs -> RecP r (map (fmap setInserted) fs)
+        expand :: Pattern -> [Declaration] -> Nice [Declaration]
+        expand _ [] = return []
+        expand p (d : ds) = do
+          case d of
+            Pragma (CatchallPragma _) -> do
+                  (d :) <$> expand p ds
+            FunClause (LHS p0 eqs es) rhs wh ca -> do
+              case hasEllipsis' p0 of
+                ManyHoles -> throwError $ MultipleEllipses p0
+                OneHole cxt -> do
+                  -- Replace the ellipsis by @p@.
+                  let p1 = cxt p
+                  let d' = FunClause (LHS p1 eqs es) rhs wh ca
+                  -- If we have with-expressions (es /= []) then the following
+                  -- ellipses also get the additional patterns in p0.
+                  (d' :) <$> expand (if null es then p else killRange p1) ds
+                ZeroHoles _ -> do
+                  -- We can have ellipses after a fun clause without.
+                  -- They refer to the last clause that introduced new with-expressions.
+                  -- Same here: If we have new with-expressions, the next ellipses will
+                  -- refer to us.
+                  -- Andreas, Jesper, 2017-05-13, issue #2578
+                  -- Need to update the range also on the next with-patterns.
+                  (d :) <$> expand (if null es then p else killRange p0) ds
+            _ -> __IMPOSSIBLE__
+    expandEllipsis _ = __IMPOSSIBLE__
 
     -- Turn function clauses into nice function clauses.
     mkClauses :: Name -> [Declaration] -> Catchall -> Nice [Clause]
     mkClauses _ [] _ = return []
-    mkClauses x (Pragma (CatchallPragma r) : cs) True  = throwError $ InvalidCatchallPragma r
-    mkClauses x (Pragma (CatchallPragma r) : cs) False = do
-      when (null cs) $ throwError $ InvalidCatchallPragma r
+    mkClauses x (Pragma (CatchallPragma r) : cs) True  = do
+      niceWarning $ InvalidCatchallPragma r
       mkClauses x cs True
-    mkClauses x (FunClause lhs@(LHS _ _ _ []) rhs wh ca : cs) catchall =
-      (Clause x (ca || catchall) lhs rhs wh [] :) <$> mkClauses x cs False
-    mkClauses x (FunClause lhs@(LHS _ ps _ es) rhs wh ca : cs) catchall = do
-      when (null with) $ throwError $ MissingWithClauses x
-      wcs <- mkClauses x with False
+    mkClauses x (Pragma (CatchallPragma r) : cs) False = do
+      when (null cs) $ niceWarning $ InvalidCatchallPragma r
+      mkClauses x cs True
+
+    mkClauses x (FunClause lhs rhs wh ca : cs) catchall
+      | null (lhsWithExpr lhs) || hasEllipsis lhs  =
+      (Clause x (ca || catchall) lhs rhs wh [] :) <$> mkClauses x cs False   -- Will result in an error later.
+
+    mkClauses x (FunClause lhs rhs wh ca : cs) catchall = do
+      when (null withClauses) $ throwError $ MissingWithClauses x
+      wcs <- mkClauses x withClauses False
       (Clause x (ca || catchall) lhs rhs wh wcs :) <$> mkClauses x cs' False
       where
-        (with, cs') = subClauses cs
+        (withClauses, cs') = subClauses cs
 
         -- A clause is a subclause if the number of with-patterns is
         -- greater or equal to the current number of with-patterns plus the
         -- number of with arguments.
+        numWith = numberOfWithPatterns p + length es where LHS p _ es = lhs
+
         subClauses :: [Declaration] -> ([Declaration],[Declaration])
-        subClauses (c@(FunClause (LHS _ ps' _ _) _ _ _) : cs)
-         | length ps' >= length ps + length es = mapFst (c:) (subClauses cs)
+        subClauses (c@(FunClause (LHS p0 _ _) _ _ _) : cs)
+         | isEllipsis p0 ||
+           numberOfWithPatterns p0 >= numWith = mapFst (c:) (subClauses cs)
          | otherwise                           = ([], c:cs)
-        subClauses (c@(FunClause (Ellipsis _ ps' _ _) _ _ _) : cs)
-         = mapFst (c:) (subClauses cs)
         subClauses (c@(Pragma (CatchallPragma r)) : cs) = case subClauses cs of
           ([], cs') -> ([], c:cs')
           (cs, cs') -> (c:cs, cs')
         subClauses [] = ([],[])
         subClauses _  = __IMPOSSIBLE__
-    mkClauses x (FunClause lhs@Ellipsis{} rhs wh ca : cs) catchall =
-      (Clause x (ca || catchall) lhs rhs wh [] :) <$> mkClauses x cs False   -- Will result in an error later.
     mkClauses _ _ _ = __IMPOSSIBLE__
 
     -- for finding clauses for a type sig in mutual blocks
     couldBeFunClauseOf :: Maybe Fixity' -> Name -> Declaration -> Bool
     couldBeFunClauseOf mFixity x (Pragma (CatchallPragma{})) = True
-    couldBeFunClauseOf mFixity x (FunClause Ellipsis{} _ _ _) = True
-    couldBeFunClauseOf mFixity x (FunClause (LHS p _ _ _) _ _ _) =
+    couldBeFunClauseOf mFixity x (FunClause (LHS p _ _) _ _ _) = hasEllipsis p ||
       let
       pns        = patternNames p
       xStrings   = nameStringParts x
@@ -1235,20 +1280,6 @@ niceDeclarations ds = do
         _ -> False -- trace ("couldBe not (case default)") $ False
     couldBeFunClauseOf _ _ _ = False -- trace ("couldBe not (fun default)") $ False
 
-    -- ASR (27 May 2014). Commented out unused code.
-    -- @isFunClauseOf@ is for non-mutual blocks where clauses must follow the
-    -- type sig immediately
-    -- isFunClauseOf :: Name -> Declaration -> Bool
-    -- isFunClauseOf x (FunClause Ellipsis{} _ _) = True
-    -- isFunClauseOf x (FunClause (LHS p _ _ _) _ _) =
-    --  -- p is the whole left hand side, excluding "with" patterns and clauses
-    --   case removeSingletonRawAppP p of
-    --     IdentP (QName q)    -> x == q  -- lhs is just an identifier
-    --     _                   -> True
-    --         -- more complicated lhss must come with type signatures, so we just assume
-    --         -- it's part of the current definition
-    -- isFunClauseOf _ _ = False
-
     isSingleIdentifierP :: Pattern -> Maybe Name
     isSingleIdentifierP p = case removeSingletonRawAppP p of
       IdentP (QName x) -> Just x
@@ -1266,24 +1297,31 @@ niceDeclarations ds = do
     mkOldMutual r ds = do
         -- Check that there aren't any missing definitions
         checkLoneSigs loneNames
+
         -- Check that there are no declarations that aren't allowed in old style mutual blocks
-        case filter notAllowedInMutual ds of
-          []  -> return ()
-          (NiceFunClause _ _ _ _ s_ (FunClause lhs _ _ _)):_ -> throwError $ MissingTypeSignature lhs
-          d:_ -> throwError $ NotAllowedInMutual d
+        forM_ ds $ \case
+          -- Andreas, 2013-11-23 allow postulates in mutual blocks
+          Axiom{} -> return ()
+          -- Andreas, 2017-10-09, issue #2576, raise error about missing type signature
+          -- in ConcreteToAbstract rather than here.
+          NiceFunClause{} -> return ()
+          -- Andreas, 2018-05-11, issue #3052, allow pat.syn.s in mutual blocks
+          NicePatternSyn{} -> return ()
+          -- Otherwise, only categorized signatures and definitions are allowed (data/record/fun)
+          d -> when (declKind d == OtherDecl) $ throwError $ NotAllowedInMutual d
+
+        -- Compute termination checking flag for mutual block
         tc0 <- use terminationCheckPragma
         let tcs = map termCheck ds
         tc <- combineTermChecks r (tc0:tcs)
 
+        -- Compute positivity checking flag for mutual block
         pc0 <- use positivityCheckPragma
         let pc :: PositivityCheck
             pc = pc0 && all positivityCheckOldMutual ds
 
         return $ NiceMutual r tc pc $ sigs ++ other
       where
-        -- Andreas, 2013-11-23 allow postulates in mutual blocks
-        notAllowedInMutual Axiom{} = False
-        notAllowedInMutual d       = declKind d == OtherDecl
         -- Pull type signatures to the top
         (sigs, other) = List.partition isTypeSig ds
         isTypeSig Axiom{}                     = True
@@ -1363,7 +1401,7 @@ niceDeclarations ds = do
         Axiom r f p a i rel mp x e       -> (\ i -> Axiom r f p a i rel mp x e) <$> setInstance i
         FunSig r f p a i m rel tc x e    -> (\ i -> FunSig r f p a i m rel tc x e) <$> setInstance i
         NiceUnquoteDecl r f p a i tc x e -> (\ i -> NiceUnquoteDecl r f p a i tc x e) <$> setInstance i
-        NiceMutual{}                     -> return d
+        NiceMutual r termCheck pc ds     -> NiceMutual r termCheck pc <$> mapM mkInstance ds
         NiceFunClause{}                  -> return d
         FunDef r ds f a i tc x cs        -> (\ i -> FunDef r ds f a i tc x cs) <$> setInstance i
         NiceField{}                      -> return d  -- Field instance are handled by the parser
@@ -1626,7 +1664,7 @@ fixitiesAndPolarities = foldMap $ \ d -> case d of
 -- | (Approximately) convert a 'NiceDeclaration' back to a list of
 -- 'Declaration's.
 notSoNiceDeclarations :: NiceDeclaration -> [Declaration]
-notSoNiceDeclarations d =
+notSoNiceDeclarations d = fixityDecl d ++
   case d of
     Axiom _ _ _ _ i rel mp x e       -> (case mp of
                                            Nothing   -> []
@@ -1655,6 +1693,32 @@ notSoNiceDeclarations d =
     inst InstanceDef    ds = [InstanceB (getRange ds) ds]
     inst NotInstanceDef ds = ds
 
+    fixityDecl d = nameAndFixity >>= \ (x, f) -> infixDecl (theFixity f) x ++ syntaxDecl (theNotation f) x
+      where
+        nameAndFixity = case d of
+          Axiom _ f _ _ _ _ _ x _           -> [(x, f)]
+          NiceField _ f _ _ _ x _           -> [(x, f)]
+          PrimitiveFunction _ f _ _ x _     -> [(x, f)]
+          NiceRecSig _ f _ _ _ x _ _        -> [(x, f)]
+          NiceDataSig _ f _ _ _ x _ _       -> [(x, f)]
+          FunSig _ f _ _ _ _ _ _ x _        -> [(x, f)]
+          NicePatternSyn _ f x _ _          -> [(x, f)]
+          NiceUnquoteDecl _ fs _ _ _ _ xs _ -> zip xs fs
+          NiceUnquoteDef _ fs _ _ _ xs _    -> zip xs fs
+          NiceMutual{}      -> []
+          NiceModule{}      -> []
+          NiceModuleMacro{} -> []
+          NiceOpen{}        -> []
+          NiceImport{}      -> []
+          NicePragma{}      -> []
+          NiceFunClause{}   -> []
+          FunDef{}          -> []  -- use the fixity from the FunSig/DataSig/RecSig
+          DataDef{}         -> []
+          RecDef{}          -> []
+
+    infixDecl  f x = [Infix f [x] | notElem f [defaultFixity, noFixity]]
+    syntaxDecl n x = [Syntax x n  | not (null n) ]
+
 -- | Has the 'NiceDeclaration' a field of type 'IsAbstract'?
 niceHasAbstract :: NiceDeclaration -> Maybe IsAbstract
 niceHasAbstract d =
