@@ -8,15 +8,17 @@
 -}
 module Agda.Syntax.Concrete.Name where
 
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ((<>))
+#endif
+
 import Control.DeepSeq
-import Control.Applicative
 
 import Data.ByteString.Char8 (ByteString)
 import Data.Function
 import qualified Data.List as List
 
 import Data.Data (Data)
-import Data.Typeable (Typeable)
 
 import GHC.Generics (Generic)
 
@@ -43,7 +45,15 @@ import Agda.Utils.Impossible
 data Name
   = Name Range [NamePart]  -- ^ A (mixfix) identifier.
   | NoName Range NameId    -- ^ @_@.
-  deriving (Typeable, Data)
+  deriving Data
+
+-- | An open mixfix identifier is either prefix, infix, or suffix.
+--   That is to say: at least one of its extremities is a @Hole@
+
+isOpenMixfix :: Name -> Bool
+isOpenMixfix n = case n of
+  Name _ (x : xs@(_:_)) -> x == Hole || last xs == Hole
+  _                     -> False
 
 instance Underscore Name where
   underscore = NoName noRange __IMPOSSIBLE__
@@ -56,7 +66,7 @@ instance Underscore Name where
 data NamePart
   = Hole       -- ^ @_@ part.
   | Id RawName  -- ^ Identifier part.
-  deriving (Typeable, Data, Generic)
+  deriving (Data, Generic)
 
 -- | Define equality on @Name@ to ignore range so same names in different
 --   locations are equal.
@@ -99,7 +109,7 @@ instance Ord NamePart where
 data QName
   = Qual  Name QName -- ^ @A.rest@.
   | QName Name       -- ^ @x@.
-  deriving (Typeable, Data, Eq, Ord)
+  deriving (Data, Eq, Ord)
 
 instance Underscore QName where
   underscore = QName underscore
@@ -114,7 +124,7 @@ data TopLevelModuleName = TopLevelModuleName
   { moduleNameRange :: Range
   , moduleNameParts :: [String]
   }
-  deriving (Show, Typeable, Data)
+  deriving (Show, Data)
 
 instance Eq    TopLevelModuleName where (==)    = (==)    `on` moduleNameParts
 instance Ord   TopLevelModuleName where compare = compare `on` moduleNameParts
