@@ -17,9 +17,9 @@ module Agda.Syntax.Concrete.Operators
     , parsePatternSyn
     ) where
 
+import Control.Applicative ( Alternative((<|>)), liftA2 )
 import Control.Arrow ((***), (&&&), first, second)
 import Control.DeepSeq
-import Control.Applicative
 import Control.Monad
 
 import Data.Either (partitionEithers)
@@ -37,8 +37,10 @@ import qualified Data.Traversable as Trav
 import Agda.Syntax.Concrete.Pretty ()
 import Agda.Syntax.Common
 import Agda.Syntax.Concrete hiding (appView)
+import qualified Agda.Syntax.Concrete as C
 import Agda.Syntax.Concrete.Operators.Parser
 import Agda.Syntax.Concrete.Operators.Parser.Monad hiding (parse)
+import Agda.Syntax.Concrete.Pattern
 import qualified Agda.Syntax.Abstract.Name as A
 import Agda.Syntax.Position
 import Agda.Syntax.Fixity
@@ -532,7 +534,7 @@ parsePat prs p = case p of
     HiddenP _ _      -> fail "bad hidden argument"
     InstanceP _ _    -> fail "bad instance argument"
     AsP r x p        -> AsP r x <$> parsePat prs p
-    DotP r o e       -> return $ DotP r o e
+    DotP r e         -> return $ DotP r e
     ParenP r p       -> fullParen' <$> parsePat prs p
     WildP _          -> return p
     AbsurdP _        -> return p
@@ -540,6 +542,8 @@ parsePat prs p = case p of
     QuoteP _         -> return p
     IdentP _         -> return p
     RecP r fs        -> RecP r <$> mapM (traverse (parsePat prs)) fs
+    EllipsisP _      -> fail "bad ellipsis"
+    WithP r p        -> WithP r <$> parsePat prs p
 
 
 {- Implement parsing of copattern left hand sides, e.g.
@@ -634,7 +638,7 @@ classifyPattern conf p =
     -- case @f ps@
     Arg _ (Named _ (IdentP x)) : ps | Just x == topName conf -> do
       guard $ all validPat ps
-      return $ Right (x, LHSHead x ps)
+      return $ Right (x, lhsCoreAddSpine (LHSHead x []) ps)
 
     -- case @d ps@
     Arg _ (Named _ (IdentP x)) : ps | x `elem` fldNames conf -> do
@@ -645,7 +649,7 @@ classifyPattern conf p =
       guard $ all (isLeft . namedArg) ps3
       let (f, lhs)      = fromR p2
           (ps', _:ps'') = splitAt (length ps1) ps
-      return $ Right (f, LHSProj x ps' lhs ps'')
+      return $ Right (f, lhsCoreAddSpine (LHSProj x ps' lhs []) ps'')
 
     -- case: ordinary pattern
     _ -> do
@@ -689,9 +693,11 @@ parsePatternOrSyn lhsOrPatSyn p = billToParser IsPattern $ do
 -- | Helper function for 'parseLHS' and 'parsePattern'.
 validConPattern :: [QName] -> Pattern -> Bool
 validConPattern cons p = case appView p of
+    [WithP _ p]   -> validConPattern cons p
     [_]           -> True
     IdentP x : ps -> elem x cons && all (validConPattern cons) ps
     [QuoteP _, _] -> True
+    DotP _ e : ps -> all (validConPattern cons) ps
     _             -> False
 -- Andreas, 2012-06-04: I do not know why the following line was
 -- the catch-all case.  It seems that the new catch-all works also
