@@ -6,8 +6,13 @@
 -}
 module Agda.Syntax.Concrete.Pretty where
 
-import Prelude hiding (null)
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ( (<>), null )
+#else
+import Prelude hiding ( null )
+#endif
 
+import Data.IORef
 import Data.Functor
 import Data.Maybe
 import qualified Data.Strict.Maybe as Strict
@@ -20,6 +25,8 @@ import Agda.Syntax.Position
 
 import Agda.TypeChecking.Positivity.Occurrence
 
+import Agda.Interaction.Options.IORefs (UnicodeOrAscii(..), unicodeOrAscii)
+
 import Agda.Utils.Function
 import Agda.Utils.Functor
 import Agda.Utils.Null
@@ -29,6 +36,28 @@ import Agda.Utils.String
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+import qualified System.IO.Unsafe as UNSAFE (unsafePerformIO)
+
+-- Andreas, 2017-10-02, TODO: restore Show to its original purpose
+--
+-- deriving instance Show Expr
+-- deriving instance (Show a) => Show (OpApp a)
+-- deriving instance Show Declaration
+-- deriving instance Show Pattern
+-- deriving instance Show TypedBinding
+-- deriving instance Show TypedBindings
+-- deriving instance Show LamBinding
+-- deriving instance Show ModuleAssignment
+-- deriving instance (Show a, Show b) => Show (ImportDirective' a b)
+-- deriving instance (Show a, Show b) => Show (Using' a b)
+-- deriving instance (Show a, Show b) => Show (Renaming' a b)
+-- deriving instance Show Pragma
+-- deriving instance Show RHS
+-- deriving instance Show LHS
+-- deriving instance Show LHSCore
+-- deriving instance Show WhereClause
+-- deriving instance Show ModuleApplication
+
 instance Show Expr            where show = show . pretty
 instance Show Declaration     where show = show . pretty
 instance Show Pattern         where show = show . pretty
@@ -39,6 +68,38 @@ instance (Pretty a, Pretty b) => Show (ImportDirective' a b)
                               where show = show . pretty
 instance Show Pragma          where show = show . pretty
 instance Show RHS             where show = show . pretty
+instance Show LHS where show = show . pretty
+instance Show LHSCore where show = show . pretty
+instance Show WhereClause where show = show . pretty
+instance Show ModuleApplication where show = show . pretty
+
+
+-- | Picking the appropriate set of special characters depending on
+-- whether we are allowed to use unicode or have to limit ourselves
+-- to ascii.
+
+data SpecialCharacters = SpecialCharacters
+  { _dbraces :: Doc -> Doc
+  , _lambda  :: Doc
+  , _arrow   :: Doc
+  , _forallQ :: Doc
+  }
+
+{-# NOINLINE specialCharacters #-}
+specialCharacters :: SpecialCharacters
+specialCharacters =
+  let opt = UNSAFE.unsafePerformIO (readIORef unicodeOrAscii) in
+  case opt of
+    UnicodeOk -> SpecialCharacters { _dbraces = ((text "\x2983 " <>) . (<> text " \x2984"))
+                                   , _lambda  = text "\x03bb"
+                                   , _arrow   = text "\x2192"
+                                   , _forallQ = text "\x2200"
+                                   }
+    AsciiOnly -> SpecialCharacters { _dbraces = braces . braces'
+                                   , _lambda  = text "\\"
+                                   , _arrow   = text "->"
+                                   , _forallQ = text "forall"
+                                   }
 
 braces' :: Doc -> Doc
 braces' d = case render d of
@@ -48,7 +109,11 @@ braces' d = case render d of
 
 -- double braces...
 dbraces :: Doc -> Doc
-dbraces = braces . braces'
+dbraces = _dbraces specialCharacters
+
+-- forall quantifier
+forallQ :: Doc
+forallQ = _forallQ specialCharacters
 
 -- Lays out a list of documents [d₁, d₂, …] in the following way:
 -- @
@@ -65,8 +130,8 @@ bracesAndSemicolons (d : ds) =
   sep ([text "{" <+> d] ++ map (text ";" <+>) ds ++ [text "}"])
 
 arrow, lambda :: Doc
-arrow  = text "\x2192"
-lambda = text "\x03bb"
+arrow  = _arrow specialCharacters
+lambda = _lambda specialCharacters
 
 -- | @prettyHiding info visible doc@ puts the correct braces
 --   around @doc@ according to info @info@ and returns
@@ -92,7 +157,6 @@ instance Pretty a => Pretty (WithHiding a) where
   pretty w = prettyHiding w id $ pretty $ dget w
 
 instance Pretty Relevance where
-  pretty Forced{}   = empty
   pretty Relevant   = empty
   pretty Irrelevant = text "."
   pretty NonStrict  = text ".."
@@ -132,20 +196,13 @@ instance Pretty Expr where
 
             HiddenArg _ e -> braces' $ pretty e
             InstanceArg _ e -> dbraces $ pretty e
+            Lam _ bs (AbsurdLam _ h) -> lambda <+> fsep (map pretty bs) <+> absurd h
             Lam _ bs e ->
                 sep [ lambda <+> fsep (map pretty bs) <+> arrow
                     , nest 2 $ pretty e
                     ]
-            AbsurdLam _ NotHidden  -> lambda <+> text "()"
-            AbsurdLam _ Instance{} -> lambda <+> text "{{}}"
-            AbsurdLam _ Hidden     -> lambda <+> text "{}"
-            ExtendedLam _ pes ->
-              lambda <+> bracesAndSemicolons (map (\(x,y,z,_) -> prettyClause x y z) pes)
-                   where prettyClause lhs rhs wh = sep [ pretty lhs
-                                                       , nest 2 $ pretty' rhs
-                                                       ] $$ nest 2 (pretty wh)
-                         pretty' (RHS e)   = arrow <+> pretty e
-                         pretty' AbsurdRHS = empty
+            AbsurdLam _ h -> lambda <+> absurd h
+            ExtendedLam _ pes -> lambda <+> bracesAndSemicolons (map pretty pes)
             Fun _ e1 e2 ->
                 sep [ pretty e1 <+> arrow
                     , pretty e2
@@ -157,12 +214,13 @@ instance Pretty Expr where
             Set _   -> text "Set"
             Prop _  -> text "Prop"
             SetN _ n    -> text "Set" <> text (showIndex n)
-            Let _ ds e  ->
+            Let _ ds me  ->
                 sep [ text "let" <+> vcat (map pretty ds)
-                    , text "in" <+> pretty e
+                    , maybe empty (\ e -> text "in" <+> pretty e) me
                     ]
             Paren _ e -> parens $ pretty e
             IdiomBrackets _ e -> text "(|" <+> pretty e <+> text "|)"
+            DoBlock _ ss -> text "do" <+> vcat (map pretty ss)
             As _ x e  -> pretty x <> text "@" <> pretty e
             Dot _ e   -> text "." <> pretty e
             Absurd _  -> text "()"
@@ -184,6 +242,11 @@ instance Pretty Expr where
             -- Andreas, 2011-10-03 print irrelevant things as .(e)
             DontCare e -> text "." <> parens (pretty e)
             Equal _ a b -> pretty a <+> text "=" <+> pretty b
+            Ellipsis _  -> text "..."
+        where
+          absurd NotHidden  = text "()"
+          absurd Instance{} = text "{{}}"
+          absurd Hidden     = text "{}"
 
 instance (Pretty a, Pretty b) => Pretty (Either a b) where
   pretty = either pretty pretty
@@ -194,6 +257,15 @@ instance Pretty a => Pretty (FieldAssignment' a) where
 instance Pretty ModuleAssignment where
   pretty (ModuleAssignment m es i) = (fsep $ pretty m : map pretty es) <+> pretty i
 
+instance Pretty LamClause where
+  pretty (LamClause lhs rhs wh _) =
+    sep [ pretty lhs
+        , nest 2 $ pretty' rhs
+        ] $$ nest 2 (pretty wh)
+    where
+      pretty' (RHS e)   = arrow <+> pretty e
+      pretty' AbsurdRHS = empty
+
 instance Pretty BoundName where
   pretty BName{ boundName = x, boundLabel = l }
     | x == l    = pretty x
@@ -218,7 +290,7 @@ newtype Tel = Tel Telescope
 
 instance Pretty Tel where
     pretty (Tel tel)
-      | any isMeta tel = text "∀" <+> fsep (map pretty tel)
+      | any isMeta tel = forallQ <+> fsep (map pretty tel)
       | otherwise      = fsep (map pretty tel)
       where
         isMeta (TypedBindings _ (Arg _ (TBind _ _ (Underscore _ Nothing)))) = True
@@ -250,7 +322,6 @@ instance Pretty RHS where
     pretty (RHS e)   = text "=" <+> pretty e
     pretty AbsurdRHS = empty
 
-instance Show WhereClause where show = show . pretty
 instance Pretty WhereClause where
   pretty  NoWhere = empty
   pretty (AnyWhere [Module _ x [] ds]) | isNoName (unqualify x)
@@ -262,15 +333,12 @@ instance Pretty WhereClause where
          , nest 2 (vcat $ map pretty ds)
          ]
 
-instance Show LHS where show = show . pretty
 instance Pretty LHS where
   pretty lhs = case lhs of
-    LHS p ps eqs es  -> pr (pretty p) ps eqs es
-    Ellipsis _ ps eqs es -> pr (text "...") ps eqs es
+    LHS p eqs es  -> pr (pretty p) eqs es
     where
-      pr d ps eqs es =
+      pr d eqs es =
         sep [ d
-            , nest 2 $ fsep $ map ((text "|" <+>) . pretty) ps
             , nest 2 $ pThing "rewrite" eqs
             , nest 2 $ pThing "with" es
             ]
@@ -278,18 +346,29 @@ instance Pretty LHS where
       pThing thing (e : es) = fsep $ (text thing <+> pretty e)
                                    : map ((text "|" <+>) . pretty) es
 
-instance Show LHSCore where show = show . pretty
 instance Pretty LHSCore where
   pretty (LHSHead f ps) = sep $ pretty f : map (parens . pretty) ps
   pretty (LHSProj d ps lhscore ps') = sep $
     pretty d : map (parens . pretty) ps ++
     parens (pretty lhscore) : map (parens . pretty) ps'
+  pretty (LHSWith h wps ps) = if null ps then doc else
+      sep $ parens doc : map (parens . pretty) ps
+    where
+    doc = sep $ pretty h : map ((text "|" <+>) . pretty) wps
 
-instance Show ModuleApplication where show = show . pretty
 instance Pretty ModuleApplication where
   pretty (SectionApp _ bs e) = fsep (map pretty bs) <+> text "=" <+> pretty e
   pretty (RecordModuleIFS _ rec) = text "=" <+> pretty rec <+> text "{{...}}"
 
+instance Pretty DoStmt where
+  pretty (DoBind _ p e cs) =
+    ((pretty p <+> text "←") <?> pretty e) <?> prCs cs
+    where
+      prCs [] = empty
+      prCs cs = text "where" <?> vcat (map pretty cs)
+  pretty (DoThen e) = pretty e
+  pretty (DoLet _ ds) = text "let" <+> vcat (map pretty ds)
+
 instance Pretty Declaration where
     prettyList = vcat . map pretty
     pretty d =
@@ -302,7 +381,7 @@ instance Pretty Declaration where
                 sep [ text "field"
                     , nest 2 $ mkInst inst $ mkOverlap i $
                       prettyRelevance i $ prettyHiding i id $
-                        pretty $ TypeSig (i {argInfoRelevance = Relevant}) x e
+                        pretty $ TypeSig (setRelevance Relevant i) x e
                     ]
                 where
                   mkInst InstanceDef    d = sep [ text "instance", nest 2 d ]
@@ -370,7 +449,9 @@ instance Pretty Declaration where
                     pType Nothing  =
                               text "where"
                     pInd = maybeToList $ text . show . rangedThing <$> ind
-                    pEta = maybeToList $ (\x -> if x then text "eta-equality" else text "no-eta-equality") <$> eta
+                    pEta = maybeToList $ eta <&> \case
+                      YesEta -> text "eta-equality"
+                      NoEta  -> text "no-eta-equality"
                     pCon = maybeToList $ (text "constructor" <+>) . pretty <$> fst <$> con
             Infix f xs  ->
                 pretty f <+> (fsep $ punctuate comma $ map pretty xs)
@@ -453,8 +534,10 @@ instance Pretty Pragma where
       hsep $ [text "STATIC", pretty i]
     pretty (InjectivePragma _ i) =
       hsep $ [text "INJECTIVE", pretty i]
-    pretty (InlinePragma _ i) =
+    pretty (InlinePragma _ True i) =
       hsep $ [text "INLINE", pretty i]
+    pretty (InlinePragma _ False i) =
+      hsep $ [text "NOINLINE", pretty i]
     pretty (ImportPragma _ i) =
       hsep $ [text "IMPORT", text i]
     pretty (ImportUHCPragma _ i) =
@@ -470,6 +553,7 @@ instance Pretty Pragma where
         NonTerminating         -> text "NON_TERMINATING"
         Terminating            -> text "TERMINATING"
         TerminationMeasure _ x -> hsep $ [text "MEASURE", pretty x]
+    pretty (WarningOnUsage _ nm str) = hsep [ text "WARNING_ON_USAGE", pretty nm, text str ]
     pretty (CatchallPragma _) = text "CATCHALL"
     pretty (DisplayPragma _ lhs rhs) = text "DISPLAY" <+> sep [ pretty lhs <+> text "=", nest 2 $ pretty rhs ]
     pretty (NoPositivityCheckPragma _) = text "NO_POSITIVITY_CHECK"
@@ -511,8 +595,7 @@ instance Pretty e => Pretty (Named_ e) where
 
 instance Pretty Pattern where
     prettyList = fsep . map pretty
-    pretty p =
-        case p of
+    pretty = \case
             IdentP x        -> pretty x
             AppP p1 p2      -> sep [ pretty p1, nest 2 $ pretty p2 ]
             RawAppP _ ps    -> fsep $ map pretty ps
@@ -522,11 +605,13 @@ instance Pretty Pattern where
             ParenP _ p      -> parens $ pretty p
             WildP _         -> underscore
             AsP _ x p       -> pretty x <> text "@" <> pretty p
-            DotP _ _ p      -> text "." <> pretty p
+            DotP _ p        -> text "." <> pretty p
             AbsurdP _       -> text "()"
             LitP l          -> pretty l
             QuoteP _        -> text "quote"
             RecP _ fs       -> sep [ text "record", bracesAndSemicolons (map pretty fs) ]
+            EllipsisP _     -> text "..."
+            WithP _ p       -> text "|" <+> pretty p
 
 prettyOpApp :: forall a .
   Pretty a => QName -> [NamedArg (MaybePlaceholder a)] -> [Doc]
@@ -540,12 +625,13 @@ prettyOpApp q es = merge [] $ prOp ms xs es
            NoName{}  -> __IMPOSSIBLE__
 
     prOp :: [Name] -> [NamePart] -> [NamedArg (MaybePlaceholder a)] -> [(Doc, Maybe PositionInName)]
-    prOp ms (Hole : xs) (e : es) = (pretty e, case namedArg e of
-                                                Placeholder p -> Just p
-                                                _             -> Nothing) :
-                                   prOp ms xs es
+    prOp ms (Hole : xs) (e : es) =
+      case namedArg e of
+        Placeholder p   -> (qual ms $ pretty e, Just p) : prOp [] xs es
+        NoPlaceholder{} -> (pretty e, Nothing) : prOp ms xs es
+          -- Module qualifier needs to go on section holes (#3072)
     prOp _  (Hole : _)  []       = __IMPOSSIBLE__
-    prOp ms (Id x : xs) es       = ( pretty (foldr Qual (QName (Name noRange $ [Id x])) ms)
+    prOp ms (Id x : xs) es       = ( qual ms $ pretty $ Name noRange $ [Id x]
                                    , Nothing
                                    ) : prOp [] xs es
       -- Qualify the name part with the module.
@@ -553,6 +639,8 @@ prettyOpApp q es = merge [] $ prOp ms xs es
 
     prOp _  []       es          = map (\e -> (pretty e, Nothing)) es
 
+    qual ms doc = hcat $ punctuate (text ".") $ map pretty ms ++ [doc]
+
     -- Section underscores should be printed without surrounding
     -- whitespace. This function takes care of that.
     merge :: [Doc] -> [(Doc, Maybe PositionInName)] -> [Doc]
