@@ -18,7 +18,6 @@ import qualified Data.Set as Set
 import Data.Traversable
 
 import Data.Data (Data)
-import Data.Typeable (Typeable)
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
@@ -44,14 +43,14 @@ data Fixity' = Fixity'
       -- ^ Range of the name in the fixity declaration
       --   (used for correct highlighting, see issue #2140).
     }
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 instance Eq Fixity' where
   Fixity' f n _ == Fixity' f' n' _ = f == f' && n == n'
 
 -- | Decorating something with @Fixity'@.
 data ThingWithFixity x = ThingWithFixity x Fixity'
-  deriving (Functor, Foldable, Traversable, Typeable, Data, Show)
+  deriving (Functor, Foldable, Traversable, Data, Show)
 
 -- | All the notation information related to a name.
 data NewNotation = NewNotation
@@ -68,7 +67,7 @@ data NewNotation = NewNotation
   , notaIsOperator :: Bool
     -- ^ True if the notation comes from an operator (rather than a
     -- syntax declaration).
-  } deriving (Typeable, Show)
+  } deriving Show
 
 -- | If an operator has no specific notation, then it is computed from
 -- its name.
@@ -218,12 +217,12 @@ data PrecedenceLevel
     -- ^ No fixity declared.
   | Related !Integer
     -- ^ Fixity level declared as the @Integer@.
-  deriving (Eq, Ord, Show, Typeable, Data)
+  deriving (Eq, Ord, Show, Data)
 
 -- | Associativity.
 
 data Associativity = NonAssoc | LeftAssoc | RightAssoc
-   deriving (Eq, Ord, Show, Typeable, Data)
+   deriving (Eq, Ord, Show, Data)
 
 -- | Fixity of operators.
 
@@ -233,7 +232,7 @@ data Fixity = Fixity
   , fixityLevel :: !PrecedenceLevel
   , fixityAssoc :: !Associativity
   }
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 instance Eq Fixity where
   f1 == f2 = compare f1 f2 == EQ
@@ -249,14 +248,25 @@ noFixity = Fixity noRange Unrelated NonAssoc
 defaultFixity :: Fixity
 defaultFixity = Fixity noRange (Related 20) NonAssoc
 
+-- | Do we prefer parens around arguments like @λ x → x@ or not?
+--   See 'lamBrackets'.
+data ParenPreference = PreferParen | PreferParenless
+  deriving (Eq, Ord, Show, Data)
+
+preferParen :: ParenPreference -> Bool
+preferParen p = PreferParen == p
+
+preferParenless :: ParenPreference -> Bool
+preferParenless p = PreferParenless == p
+
 -- * Precendence
 
 -- | Precedence is associated with a context.
 data Precedence = TopCtx | FunctionSpaceDomainCtx
-                | LeftOperandCtx Fixity | RightOperandCtx Fixity
-                | FunctionCtx | ArgumentCtx | InsideOperandCtx
+                | LeftOperandCtx Fixity | RightOperandCtx Fixity ParenPreference
+                | FunctionCtx | ArgumentCtx ParenPreference | InsideOperandCtx
                 | WithFunCtx | WithArgCtx | DotPatternCtx
-    deriving (Show, Typeable, Data, Eq)
+    deriving (Show, Data, Eq)
 
 instance Pretty Precedence where
   pretty = text . show
@@ -274,11 +284,9 @@ headPrecedence :: PrecedenceStack -> Precedence
 headPrecedence []      = TopCtx
 headPrecedence (p : _) = p
 
--- | The precedence corresponding to a possibly hidden argument.
-hiddenArgumentCtx :: Hiding -> Precedence
-hiddenArgumentCtx NotHidden  = ArgumentCtx
-hiddenArgumentCtx Hidden     = TopCtx
-hiddenArgumentCtx Instance{} = TopCtx
+-- | Argument context preferring parens.
+argumentCtx_ :: Precedence
+argumentCtx_ = ArgumentCtx PreferParen
 
 -- | Do we need to bracket an operator application of the given fixity
 --   in a context with the given precedence.
@@ -291,15 +299,15 @@ opBrackets' :: Bool ->   -- Is the last argument a parenless lambda?
                Fixity -> PrecedenceStack -> Bool
 opBrackets' isLam f ps = brack f (headPrecedence ps)
   where
-    false = isLam && lamBrackets' False ps -- require more parens for `e >>= λ x → e₁` than `e >>= e₁`
+    false = isLam && lamBrackets ps -- require more parens for `e >>= λ x → e₁` than `e >>= e₁`
     brack                        (Fixity _ (Related n1) LeftAssoc)
                (LeftOperandCtx   (Fixity _ (Related n2) LeftAssoc))  | n1 >= n2 = false
     brack                        (Fixity _ (Related n1) RightAssoc)
-               (RightOperandCtx  (Fixity _ (Related n2) RightAssoc)) | n1 >= n2 = false
+               (RightOperandCtx  (Fixity _ (Related n2) RightAssoc) _) | n1 >= n2 = false
     brack f1   (LeftOperandCtx  f2) | Related f1 <- fixityLevel f1
                                     , Related f2 <- fixityLevel f2
                                     , f1 > f2 = false
-    brack f1   (RightOperandCtx f2) | Related f1 <- fixityLevel f1
+    brack f1   (RightOperandCtx f2 _) | Related f1 <- fixityLevel f1
                                     , Related f2 <- fixityLevel f2
                                     , f1 > f2 = false
     brack _ TopCtx                 = false
@@ -314,15 +322,12 @@ opBrackets' isLam f ps = brack f (headPrecedence ps)
 -- need brackets in certain right operand contexts. To decide we need to look
 -- at the stack of precedences and not just the current precedence.
 -- Example: @m₁ >>= (λ x → x) >>= m₂@ (for @_>>=_@ left associative).
--- The first argument is the parenthesis preference. If True, we add
--- parentheses also in right operand contexts where they aren't strictly
--- needed.
-lamBrackets' :: Bool -> PrecedenceStack -> Bool
-lamBrackets' _      []       = False
-lamBrackets' strict (p : ps) = case p of
+lamBrackets :: PrecedenceStack -> Bool
+lamBrackets []       = False
+lamBrackets (p : ps) = case p of
   TopCtx                 -> __IMPOSSIBLE__
-  ArgumentCtx            -> strict || lamBrackets' strict ps
-  RightOperandCtx{}      -> strict || lamBrackets' strict ps
+  ArgumentCtx pref       -> preferParen pref || lamBrackets ps
+  RightOperandCtx _ pref -> preferParen pref || lamBrackets ps
   FunctionSpaceDomainCtx -> True
   LeftOperandCtx{}       -> True
   FunctionCtx            -> True
@@ -331,10 +336,6 @@ lamBrackets' strict (p : ps) = case p of
   WithArgCtx             -> True
   DotPatternCtx          -> True
 
--- | Preference for adding parentheses.
-lamBrackets :: PrecedenceStack -> Bool
-lamBrackets = lamBrackets' True
-
 -- | Does a function application need brackets?
 appBrackets :: PrecedenceStack -> Bool
 appBrackets = appBrackets' False
@@ -344,9 +345,9 @@ appBrackets' :: Bool ->   -- Is the argument of the application a parenless lamb
                 PrecedenceStack -> Bool
 appBrackets' isLam ps = brack (headPrecedence ps)
   where
-    brack ArgumentCtx   = True
+    brack ArgumentCtx{} = True
     brack DotPatternCtx = True
-    brack _             = isLam && lamBrackets' False ps -- allow e + e₁ λ x → e₂
+    brack _             = isLam && lamBrackets ps -- allow e + e₁ λ x → e₂
 
 -- | Does a with application need brackets?
 withAppBrackets :: PrecedenceStack -> Bool
