@@ -1,6 +1,5 @@
 module Agda.Syntax.IdiomBrackets (parseIdiomBrackets) where
 
-import Control.Applicative
 import Control.Monad
 
 import Agda.Syntax.Common
@@ -8,6 +7,7 @@ import Agda.Syntax.Position
 import Agda.Syntax.Concrete
 import Agda.Syntax.Concrete.Operators
 
+import Agda.Syntax.Scope.Base
 import Agda.Syntax.Scope.Monad
 import Agda.TypeChecking.Monad
 
