@@ -12,7 +12,6 @@ module Agda.Syntax.Info where
 import Prelude hiding (null)
 
 import Data.Data (Data)
-import Data.Typeable (Typeable)
 
 import qualified Agda.Syntax.Concrete.Name as C
 import Agda.Syntax.Common
@@ -34,7 +33,7 @@ data MetaInfo = MetaInfo
   , metaNumber         :: Maybe MetaId
   , metaNameSuggestion :: String
   }
-  deriving (Typeable, Data, Show, Eq)
+  deriving (Data, Show, Eq)
 
 emptyMetaInfo :: MetaInfo
 emptyMetaInfo = MetaInfo
@@ -55,7 +54,7 @@ instance KillRange MetaInfo where
  --------------------------------------------------------------------------}
 
 newtype ExprInfo = ExprRange Range
-  deriving (Typeable, Data, Show, Eq, Null)
+  deriving (Data, Show, Eq, Null)
 
 exprNoRange :: ExprInfo
 exprNoRange = ExprRange noRange
@@ -67,33 +66,33 @@ instance KillRange ExprInfo where
   killRange (ExprRange r) = exprNoRange
 
 {--------------------------------------------------------------------------
-    Lambda information
+    Application information
  --------------------------------------------------------------------------}
 
--- | Information about lambdas.
-data LamInfo = LamInfo { lamRange  :: Range
-                       , lamOrigin :: Origin
-                       , lamParens :: Bool    -- ^ Do we prefer the lambda with or without parens?
+-- | Information about application
+data AppInfo = AppInfo { appRange  :: Range
+                       , appOrigin :: Origin
+                       , appParens :: ParenPreference -- ^ Do we prefer a appbda argument with or without parens?
                        }
-  deriving (Typeable, Data, Show, Eq, Ord)
+  deriving (Data, Show, Eq, Ord)
 
 -- | Default is system inserted and prefer parens.
-defaultLamInfo :: Range -> LamInfo
-defaultLamInfo r = LamInfo{ lamRange = r, lamOrigin = Inserted, lamParens = True }
+defaultAppInfo :: Range -> AppInfo
+defaultAppInfo r = AppInfo{ appRange = r, appOrigin = Inserted, appParens = PreferParen }
 
--- | `LamInfo` with no range information.
-defaultLamInfo_ :: LamInfo
-defaultLamInfo_ = defaultLamInfo noRange
+-- | `AppInfo` with no range information.
+defaultAppInfo_ :: AppInfo
+defaultAppInfo_ = defaultAppInfo noRange
 
-instance HasRange LamInfo where
-  getRange = lamRange
+instance HasRange AppInfo where
+  getRange = appRange
 
-instance KillRange LamInfo where
-  killRange (LamInfo r o p) = LamInfo (killRange r) o p
+instance KillRange AppInfo where
+  killRange (AppInfo r o p) = AppInfo (killRange r) o p
 
-instance LensOrigin LamInfo where
-  getOrigin = lamOrigin
-  setOrigin o i = i { lamOrigin = o }
+instance LensOrigin AppInfo where
+  getOrigin = appOrigin
+  setOrigin o i = i { appOrigin = o }
 
 {--------------------------------------------------------------------------
     Module information
@@ -110,7 +109,7 @@ data ModuleInfo = ModuleInfo
   , minfoDirective :: Maybe ImportDirective
     -- ^ Retained for @abstractToConcrete@ of 'ModuleMacro'.
   }
-  deriving (Typeable, Data, Eq)
+  deriving (Data, Eq)
 
 deriving instance Show ModuleInfo
 
@@ -128,7 +127,7 @@ instance KillRange ModuleInfo where
 ---------------------------------------------------------------------------
 
 newtype LetInfo = LetRange Range
-  deriving (Typeable, Data, Show, Eq, Null)
+  deriving (Data, Show, Eq, Null)
 
 instance HasRange LetInfo where
   getRange (LetRange r)   = r
@@ -148,7 +147,7 @@ data DefInfo = DefInfo
   , defMacro    :: IsMacro
   , defInfo     :: DeclInfo
   }
-  deriving (Typeable, Data, Show, Eq)
+  deriving (Data, Show, Eq)
 
 mkDefInfo :: Name -> Fixity' -> Access -> IsAbstract -> Range -> DefInfo
 mkDefInfo x f a ab r = DefInfo f a ab NotInstanceDef NotMacroDef (DeclInfo x r)
@@ -174,7 +173,7 @@ data DeclInfo = DeclInfo
   { declName  :: Name
   , declRange :: Range
   }
-  deriving (Typeable, Data, Show, Eq)
+  deriving (Data, Show, Eq)
 
 instance HasRange DeclInfo where
   getRange = declRange
@@ -194,7 +193,7 @@ data MutualInfo = MutualInfo
   , mutualPositivityCheck :: PositivityCheck
   , mutualRange           :: Range
   }
-  deriving (Typeable, Data, Show, Eq)
+  deriving (Data, Show, Eq)
 
 -- | Default value for 'MutualInfo'.
 instance Null MutualInfo where
@@ -211,7 +210,7 @@ instance KillRange MutualInfo where
  --------------------------------------------------------------------------}
 
 newtype LHSInfo = LHSRange Range
-  deriving (Typeable, Data, Show, Eq, Null)
+  deriving (Data, Show, Eq, Null)
 
 instance HasRange LHSInfo where
   getRange (LHSRange r) = r
@@ -226,7 +225,7 @@ instance KillRange LHSInfo where
 -- | For a general pattern we remember the source code position.
 newtype PatInfo
   = PatRange Range
-  deriving (Typeable, Data, Eq, Null, Show, HasRange, KillRange)
+  deriving (Data, Eq, Null, Show, SetRange, HasRange, KillRange)
 
 -- | Empty range for patterns.
 patNoRange :: PatInfo
@@ -238,17 +237,20 @@ data ConPatInfo = ConPatInfo
     -- ^ Does this pattern come form the eta-expansion of an implicit pattern?
     ---  Or from a user written constructor or record pattern?
   , patInfo     :: PatInfo
+  , patLazy     :: Bool
   }
-  deriving (Typeable, Data, Eq)
+  deriving (Data, Eq)
 
 instance Show ConPatInfo where
-  show (ConPatInfo po i) = applyWhen (po == ConOSystem) ("implicit " ++) $ show i
+  show (ConPatInfo po i l) =
+    applyWhen l ("lazy " ++) $
+    applyWhen (po == ConOSystem) ("implicit " ++) $ show i
 
 instance HasRange ConPatInfo where
   getRange = getRange . patInfo
 
 instance KillRange ConPatInfo where
-  killRange (ConPatInfo b i) = ConPatInfo b $ killRange i
+  killRange (ConPatInfo b i l) = ConPatInfo b (killRange i) l
 
 instance SetRange ConPatInfo where
-  setRange r (ConPatInfo b i) = ConPatInfo b $ PatRange r
+  setRange r (ConPatInfo b i l) = ConPatInfo b (PatRange r) l
