@@ -5,14 +5,9 @@
 {-# LANGUAGE TemplateHaskell            #-}
 {-# LANGUAGE UndecidableInstances       #-}  -- because of shortcomings of FunctionalDependencies
 
-#if __GLASGOW_HASKELL__ <= 708
-{-# LANGUAGE OverlappingInstances #-}
-#endif
-
 module Agda.Syntax.Internal
     ( module Agda.Syntax.Internal
     , module Agda.Syntax.Abstract.Name
-    , module Agda.Utils.Pointer
     , MetaId(..)
     ) where
 
@@ -26,11 +21,11 @@ import Data.Foldable ( Foldable, foldMap )
 import Data.Function
 import qualified Data.List as List
 import Data.Maybe
-import Data.Semigroup (Semigroup, Monoid, (<>), mempty, mappend, Sum(..))
+import Data.Monoid ( Monoid, mempty, mappend )
+import Data.Semigroup ( Semigroup, (<>), Sum(..) )
 
 import Data.Traversable
 import Data.Data (Data)
-import Data.Typeable (Typeable)
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
@@ -45,9 +40,9 @@ import Agda.Utils.Geniplate
 import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.Maybe
+import Agda.Utils.NonemptyList
 import Agda.Utils.Null
 import Agda.Utils.Permutation
-import Agda.Utils.Pointer
 import Agda.Utils.Size
 import qualified Agda.Utils.Pretty as P
 import Agda.Utils.Pretty hiding ((<>))
@@ -71,7 +66,7 @@ data ConHead = ConHead
                               --   Empty list for data constructors.
                               --   'Arg' is not needed here since it
                               --   is stored in the constructor args.
-  } deriving (Typeable, Data, Show)
+  } deriving (Data, Show)
 
 instance Eq ConHead where
   (==) = (==) `on` conName
@@ -111,7 +106,10 @@ data Term = Var {-# UNPACK #-} !Int Elims -- ^ @x es@ neutral
           | Lam ArgInfo (Abs Term)        -- ^ Terms are beta normal. Relevance is ignored
           | Lit Literal
           | Def QName Elims               -- ^ @f es@, possibly a delta/iota-redex
-          | Con ConHead ConInfo Args      -- ^ @c vs@ or @record { fs = vs }@
+          | Con ConHead ConInfo Elims
+          -- ^ @c es@ or @record { fs = es }@
+          --   @es@ allows only Apply and IApply eliminations,
+          --   and IApply only for data constructors.
           | Pi (Dom Type) (Abs Type)      -- ^ dependent or non-dependent function space
           | Sort Sort
           | Level Level
@@ -120,9 +118,7 @@ data Term = Var {-# UNPACK #-} !Int Elims -- ^ @x es@ neutral
             -- ^ Irrelevant stuff in relevant position, but created
             --   in an irrelevant context.  Basically, an internal
             --   version of the irrelevance axiom @.irrAx : .A -> A@.
-          | Shared !(Ptr Term)
-            -- ^ Explicit sharing
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 type ConInfo = ConOrigin
 
@@ -131,7 +127,7 @@ type ConInfo = ConOrigin
 data Elim' a
   = Apply (Arg a)         -- ^ Application.
   | Proj ProjOrigin QName -- ^ Projection.  'QName' is name of a record projection.
-  deriving (Typeable, Data, Show, Functor, Foldable, Traversable)
+  deriving (Data, Show, Functor, Foldable, Traversable)
 
 type Elim = Elim' Term
 type Elims = [Elim]  -- ^ eliminations ordered left-to-right.
@@ -167,7 +163,7 @@ data Abs a = Abs   { absName :: ArgName, unAbs :: a }
                -- ^ The body has (at least) one free variable.
                --   Danger: 'unAbs' doesn't shift variables properly
            | NoAbs { absName :: ArgName, unAbs :: a }
-  deriving (Typeable, Data, Functor, Foldable, Traversable)
+  deriving (Data, Functor, Foldable, Traversable)
 
 instance Decoration Abs where
   traverseF f (Abs   x a) = Abs   x <$> f a
@@ -176,7 +172,7 @@ instance Decoration Abs where
 -- | Types are terms with a sort annotation.
 --
 data Type' a = El { _getSort :: Sort, unEl :: a }
-  deriving (Typeable, Data, Show, Functor, Foldable, Traversable)
+  deriving (Data, Show, Functor, Foldable, Traversable)
 
 type Type = Type' Term
 
@@ -203,7 +199,7 @@ instance LensSort a => LensSort (Abs a) where
 --   and so on.
 data Tele a = EmptyTel
             | ExtendTel a (Abs (Tele a))  -- ^ 'Abs' is never 'NoAbs'.
-  deriving (Typeable, Data, Show, Functor, Foldable, Traversable)
+  deriving (Data, Show, Functor, Foldable, Traversable)
 
 type Telescope = Tele (Dom Type)
 
@@ -214,24 +210,22 @@ data Sort
   | Prop        -- ^ Dummy sort.
   | Inf         -- ^ @Setω@.
   | SizeUniv    -- ^ @SizeUniv@, a sort inhabited by type @Size@.
-  | DLub Sort (Abs Sort)
-    -- ^ Dependent least upper bound.
-    --   If the free variable occurs in the second sort,
-    --   the whole thing should reduce to Inf,
-    --   otherwise it's the normal lub.
-  deriving (Typeable, Data, Show)
+  | PiSort Sort (Abs Sort) -- ^ Sort of the pi type.
+  | UnivSort Sort -- ^ Sort of another sort.
+  | MetaS {-# UNPACK #-} !MetaId Elims
+  deriving (Data, Show)
 
 -- | A level is a maximum expression of 0..n 'PlusLevel' expressions
 --   each of which is a number or an atom plus a number.
 --
 --   The empty maximum is the canonical representation for level 0.
 newtype Level = Max [PlusLevel]
-  deriving (Show, Typeable, Data)
+  deriving (Show, Data)
 
 data PlusLevel
   = ClosedLevel Integer     -- ^ @n@, to represent @Setₙ@.
   | Plus Integer LevelAtom  -- ^ @n + ℓ@.
-  deriving (Show, Typeable, Data)
+  deriving (Show, Data)
 
 -- | An atomic term of type @Level@.
 data LevelAtom
@@ -243,7 +237,7 @@ data LevelAtom
     -- ^ A neutral term of type @Level@.
   | UnreducedLevel Term
     -- ^ Introduced by 'instantiate', removed by 'reduce'.
-  deriving (Show, Typeable, Data)
+  deriving (Show, Data)
 
 ---------------------------------------------------------------------------
 -- * Blocked Terms
@@ -268,7 +262,7 @@ data NotBlocked
   | ReallyNotBlocked
     -- ^ Reduction was not blocked, we reached a whnf
     --   which can be anything but a stuck @'Def'@.
-  deriving (Show, Typeable, Data)
+  deriving (Show, Data)
 
 -- | 'ReallyNotBlocked' is the unit.
 --   'MissingClauses' is dominant.
@@ -292,8 +286,8 @@ instance Monoid NotBlocked where
 data Blocked t
   = Blocked    { theBlockingMeta :: MetaId    , ignoreBlocking :: t }
   | NotBlocked { blockingStatus  :: NotBlocked, ignoreBlocking :: t }
-  deriving (Typeable, Show, Functor, Foldable, Traversable)
-  -- deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable)
+  deriving (Show, Functor, Foldable, Traversable)
+  -- deriving (Eq, Ord, Functor, Foldable, Traversable)
 
 -- | Blocking by a meta is dominant.
 instance Applicative Blocked where
@@ -393,7 +387,7 @@ data Clause = Clause
       --   @Just False@ means clause is not unreachable.
       --   @Just True@ means clause is unreachable.
     }
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 clausePats :: Clause -> [Arg DeBruijnPattern]
 clausePats = map (fmap namedThing) . namedClausePats
@@ -410,6 +404,19 @@ patVarNameToString = argNameToString
 nameToPatVarName :: Name -> PatVarName
 nameToPatVarName = nameToArgName
 
+-- | Origin of the pattern: what did the user write in this position?
+data PatOrigin
+  = PatOSystem         -- ^ Pattern inserted by the system
+  | PatOSplit          -- ^ Pattern generated by case split
+  | PatOVar Name       -- ^ User wrote a variable pattern
+  | PatODot            -- ^ User wrote a dot pattern
+  | PatOWild           -- ^ User wrote a wildcard pattern
+  | PatOCon            -- ^ User wrote a constructor pattern
+  | PatORec            -- ^ User wrote a record pattern
+  | PatOLit            -- ^ User wrote a literal pattern
+  | PatOAbsurd         -- ^ User wrote an absurd pattern
+  deriving (Data, Show, Eq)
+
 -- | Patterns are variables, constructors, or wildcards.
 --   @QName@ is used in @ConP@ rather than @Name@ since
 --     a constructor might come from a particular namespace.
@@ -417,34 +424,33 @@ nameToPatVarName = nameToArgName
 --     the arguments we are matching with) use @QName@.
 --
 data Pattern' x
-  = VarP x
+  = VarP PatOrigin x
     -- ^ @x@
-  | DotP Term
+  | DotP PatOrigin Term
     -- ^ @.t@
   | ConP ConHead ConPatternInfo [NamedArg (Pattern' x)]
     -- ^ @c ps@
     --   The subpatterns do not contain any projection copatterns.
-  | AbsurdP (Pattern' x)
-    -- ^ @()@
-    --   The argument is to keep track of the original pattern
-    --   (before the absurd match).
   | LitP Literal
     -- ^ E.g. @5@, @"hello"@.
   | ProjP ProjOrigin QName
     -- ^ Projection copattern.  Can only appear by itself.
-  deriving (Typeable, Data, Show, Functor, Foldable, Traversable)
+  deriving (Data, Show, Functor, Foldable, Traversable)
 
 type Pattern = Pattern' PatVarName
     -- ^ The @PatVarName@ is a name suggestion.
 
-varP :: ArgName -> Pattern
-varP = VarP
+varP :: a -> Pattern' a
+varP = VarP PatOSystem
+
+dotP :: Term -> Pattern' a
+dotP = DotP PatOSystem
 
 -- | Type used when numbering pattern variables.
 data DBPatVar = DBPatVar
   { dbPatVarName  :: PatVarName
   , dbPatVarIndex :: Int
-  } deriving (Typeable, Data, Show)
+  } deriving (Data, Show, Eq)
 
 type DeBruijnPattern = Pattern' DBPatVar
 
@@ -457,13 +463,13 @@ namedDBVarP m = (fmap . fmap) (\x -> DBPatVar x m) . namedVarP
 
 -- | The @ConPatternInfo@ states whether the constructor belongs to
 --   a record type (@Just@) or data type (@Nothing@).
---   In the former case, the @Bool@ says whether the record pattern
+--   In the former case, the @PatOrigin@ says whether the record pattern
 --   orginates from the expansion of an implicit pattern.
 --   The @Type@ is the type of the whole record pattern.
 --   The scope used for the type is given by any outer scope
 --   plus the clause's telescope ('clauseTel').
 data ConPatternInfo = ConPatternInfo
-  { conPRecord :: Maybe ConOrigin
+  { conPRecord :: Maybe PatOrigin
     -- ^ @Nothing@ if data constructor.
     --   @Just@ if record constructor.
   , conPType   :: Maybe (Arg Type)
@@ -473,20 +479,39 @@ data ConPatternInfo = ConPatternInfo
     --   Could be absent (@Nothing@) if pattern comes from some
     --   plugin (like Agsy).
     --   Needed e.g. for with-clause stripping.
+  , conPLazy :: Bool
+    -- ^ Lazy patterns are generated by the forcing translation
+    --   ('Agda.TypeChecking.Forcing.forcingTranslation') and are dropped by
+    --   the clause compiler (TODO: not yet)
+    --   ('Agda.TypeChecking.CompiledClause.Compile.compileClauses') when the
+    --   variables they bind are unused. The GHC backend compiles lazy matches
+    --   to lazy patterns in Haskell (TODO: not yet).
   }
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 noConPatternInfo :: ConPatternInfo
-noConPatternInfo = ConPatternInfo Nothing Nothing
+noConPatternInfo = ConPatternInfo Nothing Nothing False
 
 -- | Build partial 'ConPatternInfo' from 'ConInfo'
 toConPatternInfo :: ConInfo -> ConPatternInfo
-toConPatternInfo ConORec = ConPatternInfo (Just ConORec) Nothing
+toConPatternInfo ConORec = noConPatternInfo{ conPRecord = Just PatORec }
 toConPatternInfo _ = noConPatternInfo
 
 -- | Build 'ConInfo' from 'ConPatternInfo'.
 fromConPatternInfo :: ConPatternInfo -> ConInfo
-fromConPatternInfo = fromMaybe ConOSystem . conPRecord
+fromConPatternInfo = fromMaybe ConOCon . fmap patToConO . conPRecord
+  where
+    patToConO :: PatOrigin -> ConOrigin
+    patToConO = \case
+      PatOSystem -> ConOSystem
+      PatOSplit  -> ConOSplit
+      PatOVar{}  -> ConOSystem
+      PatODot    -> ConOSystem
+      PatOWild   -> ConOSystem
+      PatOCon    -> ConOCon
+      PatORec    -> ConORec
+      PatOLit    -> __IMPOSSIBLE__
+      PatOAbsurd -> ConOSystem
 
 -- | Extract pattern variables in left-to-right order.
 --   A 'DotP' is also treated as variable (see docu for 'Clause').
@@ -495,9 +520,8 @@ class PatternVars a b | b -> a where
 
 instance PatternVars a (Arg (Pattern' a)) where
   -- patternVars :: Arg (Pattern' a) -> [Arg (Either a Term)]
-  patternVars (Arg i (VarP x)     ) = [Arg i $ Left x]
-  patternVars (Arg i (DotP t)     ) = [Arg i $ Right t]
-  patternVars (Arg i (AbsurdP p)  ) = patternVars (Arg i p)
+  patternVars (Arg i (VarP _ x)   ) = [Arg i $ Left x]
+  patternVars (Arg i (DotP _ t)   ) = [Arg i $ Right t]
   patternVars (Arg _ (ConP _ _ ps)) = patternVars ps
   patternVars (Arg _ (LitP _)     ) = []
   patternVars (Arg _ ProjP{}      ) = []
@@ -508,18 +532,28 @@ instance PatternVars a (NamedArg (Pattern' a)) where
 instance PatternVars a b => PatternVars a [b] where
   patternVars = concatMap patternVars
 
+
+-- | Retrieve the origin of a pattern
+patternOrigin :: Pattern' x -> Maybe PatOrigin
+patternOrigin (VarP o _) = Just o
+patternOrigin (DotP o _) = Just o
+patternOrigin LitP{}     = Nothing
+patternOrigin (ConP _ ci _) = conPRecord ci
+patternOrigin ProjP{}    = Nothing
+
 -- | Does the pattern perform a match that could fail?
 properlyMatching :: DeBruijnPattern -> Bool
-properlyMatching (VarP x) = False
+properlyMatching p
+  | patternOrigin p == Just PatOAbsurd = True
+properlyMatching VarP{} = False
 properlyMatching DotP{} = False
-properlyMatching AbsurdP{} = True
 properlyMatching LitP{} = True
 properlyMatching (ConP _ ci ps) = isNothing (conPRecord ci) || -- not a record cons
   List.any (properlyMatching . namedArg) ps  -- or one of subpatterns is a proper m
 properlyMatching ProjP{} = True
 
 instance IsProjP (Pattern' a) where
-  isProjP (ProjP o d) = Just (o, AmbQ [d])
+  isProjP (ProjP o d) = Just (o, unambiguous d)
   isProjP _ = Nothing
 
 -----------------------------------------------------------------------------
@@ -580,9 +614,6 @@ data Substitution' a
            , Foldable
            , Traversable
            , Data
-#if __GLASGOW_HASKELL__ <= 708
-           , Typeable
-#endif
            )
 
 type Substitution = Substitution' Term
@@ -636,75 +667,6 @@ absurdPatternName = "()"
 isAbsurdPatternName :: PatVarName -> Bool
 isAbsurdPatternName x = x == absurdPatternName
 
----------------------------------------------------------------------------
--- * Pointers and Sharing
----------------------------------------------------------------------------
-
--- | Remove top-level @Shared@ data constructors.
-ignoreSharing :: Term -> Term
-ignoreSharing (Shared p) = ignoreSharing $ derefPtr p
-ignoreSharing v          = v
-
-ignoreSharingType :: Type -> Type
-ignoreSharingType (El s v) = El s (ignoreSharing v)
--- ignoreSharingType v = v
-
--- | Introduce sharing.
-shared_ :: Term -> Term
-shared_ v@Shared{}   = v
-shared_ v@(Var _ []) = v
-shared_ v@(Con _ _ []) = v -- Issue 1691: sharing (zero : Nat) destroys constructorForm
-shared_ v            = Shared (newPtr v)
-
--- | Typically m would be TCM and f would be Blocked.
-updateSharedFM
-#if __GLASGOW_HASKELL__ <= 708
-  :: (Applicative m, Monad m, Traversable f)
-#else
-  :: (Monad m, Traversable f)
-#endif
-  => (Term -> m (f Term)) -> Term -> m (f Term)
-updateSharedFM f v0@(Shared p) = do
-  fv <- f (derefPtr p)
-  flip traverse fv $ \v ->
-    case derefPtr (setPtr v p) of
-      Var _ [] -> return v
-      _        -> return $! compressPointerChain v0
-updateSharedFM f v = f v
-
-updateSharedM :: Monad m => (Term -> m Term) -> Term -> m Term
-updateSharedM f v0@(Shared p) = do
-  v <- f (derefPtr p)
-  case derefPtr (setPtr v p) of
-    Var _ [] -> return v
-    _        -> return $! compressPointerChain v0
-updateSharedM f v = f v
-
-updateShared :: (Term -> Term) -> Term -> Term
-updateShared f v0@(Shared p) =
-  case derefPtr (setPtr (f $ derefPtr p) p) of
-    v@(Var _ []) -> v
-    _            -> compressPointerChain v0
-updateShared f v = f v
-
-pointerChain :: Term -> [Ptr Term]
-pointerChain (Shared p) = p : pointerChain (derefPtr p)
-pointerChain _          = []
-
--- Redirect all top-level pointers to point to the last pointer. So, after
--- compression there are at most two top-level indirections. Then return the
--- inner-most pointer so we have only one pointer for the result.
-compressPointerChain :: Term -> Term
-compressPointerChain v =
-  case reverse $ pointerChain v of
-    p:_:ps@(_:_) -> setPointers (Shared p) ps
-    p:_:_        -> (Shared p)
-    _            -> v
-  where
-    setPointers u [] = u
-    setPointers u (p : ps) =
-      setPtr u p `seq` setPointers u ps
-
 ---------------------------------------------------------------------------
 -- * Smart constructors
 ---------------------------------------------------------------------------
@@ -717,32 +679,24 @@ var i | i >= 0    = Var i []
 -- | Add 'DontCare' is it is not already a @DontCare@.
 dontCare :: Term -> Term
 dontCare v =
-  case ignoreSharing v of
+  case v of
     DontCare{} -> v
     _          -> DontCare v
 
 -- | A dummy type.
 typeDontCare :: Type
-typeDontCare = El Prop (Sort Prop)
+typeDontCare = sort Prop
 
 -- | Top sort (Set\omega).
 topSort :: Type
-topSort = El Inf (Sort Inf)
+topSort = sort Inf
 
 sort :: Sort -> Type
-sort s = El (sSuc s) $ Sort s
+sort s = El (UnivSort s) $ Sort s
 
 varSort :: Int -> Sort
 varSort n = Type $ Max [Plus 0 $ NeutralLevel mempty $ var n]
 
--- | Get the next higher sort.
-sSuc :: Sort -> Sort
-sSuc Prop            = mkType 1
-sSuc Inf             = Inf
-sSuc SizeUniv        = SizeUniv
-sSuc (DLub a b)      = DLub (sSuc a) (fmap sSuc b)
-sSuc (Type l)        = Type $ levelSuc l
-
 levelSuc :: Level -> Level
 levelSuc (Max []) = Max [ClosedLevel 1]
 levelSuc (Max as) = Max $ map inc as
@@ -753,7 +707,7 @@ mkType :: Integer -> Sort
 mkType n = Type $ Max [ClosedLevel n | n > 0]
 
 isSort :: Term -> Maybe Sort
-isSort v = case ignoreSharing v of
+isSort v = case v of
   Sort s -> Just s
   _      -> Nothing
 
@@ -858,13 +812,13 @@ notBlocked = NotBlocked ReallyNotBlocked
 
 -- | Removing a topmost 'DontCare' constructor.
 stripDontCare :: Term -> Term
-stripDontCare v = case ignoreSharing v of
+stripDontCare v = case v of
   DontCare v -> v
   _          -> v
 
 -- | Doesn't do any reduction.
 arity :: Type -> Nat
-arity t = case ignoreSharing $ unEl t of
+arity t = case unEl t of
   Pi  _ b -> 1 + arity (unAbs b)
   _       -> 0
 
@@ -872,6 +826,13 @@ arity t = case ignoreSharing $ unEl t of
 notInScopeName :: ArgName -> ArgName
 notInScopeName = stringToArgName . ("." ++) . argNameToString
 
+-- | Remove the dot from a notInScopeName. This is used when printing function
+--   types that have abstracted over not-in-scope names.
+unNotInScopeName :: ArgName -> ArgName
+unNotInScopeName = stringToArgName . undot . argNameToString
+  where undot ('.':s) = s
+        undot s       = s
+
 -- | Pick the better name suggestion, i.e., the one that is not just underscore.
 class Suggest a b where
   suggest :: a -> b -> String
@@ -918,14 +879,14 @@ unSpine' p v =
 --   can be projected.
 hasElims :: Term -> Maybe (Elims -> Term, Elims)
 hasElims v =
-  case ignoreSharing v of
+  case v of
     Var   i es -> Just (Var   i, es)
     Def   f es -> Just (Def   f, es)
     MetaV x es -> Just (MetaV x, es)
     Con{}      -> Nothing
     Lit{}      -> Nothing
     -- Andreas, 2016-04-13, Issue 1932: We convert λ x → x .f  into f
-    Lam _ (Abs _ v)  -> case ignoreSharing v of
+    Lam _ (Abs _ v)  -> case v of
       Var 0 [Proj _o f] -> Just (Def f, [])
       _ -> Nothing
     Lam{}      -> Nothing
@@ -933,7 +894,6 @@ hasElims v =
     Sort{}     -> Nothing
     Level{}    -> Nothing
     DontCare{} -> Nothing
-    Shared{}   -> __IMPOSSIBLE__
 
 -- | Drop 'Apply' constructor. (Unsafe!)
 argFromElim :: Elim' a -> Arg a
@@ -945,6 +905,9 @@ isApplyElim :: Elim' a -> Maybe (Arg a)
 isApplyElim (Apply u) = Just u
 isApplyElim Proj{}    = Nothing
 
+isApplyElim' :: Empty -> Elim' a -> Arg a
+isApplyElim' e = fromMaybe (absurd e) . isApplyElim
+
 -- | Drop 'Apply' constructors. (Safe)
 allApplyElims :: [Elim' a] -> Maybe [Arg a]
 allApplyElims = mapM isApplyElim
@@ -1035,11 +998,7 @@ class TermSize a where
 
   tsize :: a -> Sum Int
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPABLE #-} (Foldable t, TermSize a) => TermSize (t a) where
-#else
-instance (Foldable t, TermSize a) => TermSize (t a) where
-#endif
   tsize = foldMap tsize
 
 instance TermSize Term where
@@ -1054,7 +1013,6 @@ instance TermSize Term where
     Pi a b      -> 1 + tsize a + tsize b
     Sort s      -> tsize s
     DontCare mv -> tsize mv
-    Shared p    -> tsize (derefPtr p)
 
 instance TermSize Sort where
   tsize s = case s of
@@ -1062,7 +1020,9 @@ instance TermSize Sort where
     Prop      -> 1
     Inf       -> 1
     SizeUniv  -> 1
-    DLub s s' -> 1 + tsize s + tsize s'
+    PiSort s s' -> 1 + tsize s + tsize s'
+    UnivSort s -> 1 + tsize s
+    MetaS _ es -> 1 + tsize es
 
 instance TermSize Level where
   tsize (Max as) = 1 + tsize as
@@ -1104,7 +1064,6 @@ instance KillRange Term where
     Pi a b      -> killRange2 Pi a b
     Sort s      -> killRange1 Sort s
     DontCare mv -> killRange1 DontCare mv
-    Shared p    -> Shared $ updatePtr killRange p
 
 instance KillRange Level where
   killRange (Max as) = killRange1 Max as
@@ -1128,7 +1087,9 @@ instance KillRange Sort where
     Inf        -> Inf
     SizeUniv   -> SizeUniv
     Type a     -> killRange1 Type a
-    DLub s1 s2 -> killRange2 DLub s1 s2
+    PiSort s1 s2 -> killRange2 PiSort s1 s2
+    UnivSort s -> killRange1 UnivSort s
+    MetaS x es -> killRange1 (MetaS x) es
 
 instance KillRange Substitution where
   killRange IdS                  = IdS
@@ -1138,8 +1099,11 @@ instance KillRange Substitution where
   killRange (Strengthen err rho) = killRange1 (Strengthen err) rho
   killRange (Lift n rho)         = killRange1 (Lift n) rho
 
+instance KillRange PatOrigin where
+  killRange = id
+
 instance KillRange ConPatternInfo where
-  killRange (ConPatternInfo mr mt) = killRange1 (ConPatternInfo mr) mt
+  killRange (ConPatternInfo mr mt lz) = killRange1 (ConPatternInfo mr) mt lz
 
 instance KillRange DBPatVar where
   killRange (DBPatVar x i) = killRange2 DBPatVar x i
@@ -1147,9 +1111,8 @@ instance KillRange DBPatVar where
 instance KillRange a => KillRange (Pattern' a) where
   killRange p =
     case p of
-      VarP x           -> killRange1 VarP x
-      DotP v           -> killRange1 DotP v
-      AbsurdP p        -> killRange1 AbsurdP p
+      VarP o x         -> killRange2 VarP o x
+      DotP o v         -> killRange2 DotP o v
       ConP con info ps -> killRange3 ConP con info ps
       LitP l           -> killRange1 LitP l
       ProjP o q        -> killRange1 (ProjP o) q
@@ -1199,7 +1162,7 @@ instance Pretty a => Pretty (Substitution' a) where
 
 instance Pretty Term where
   prettyPrec p v =
-    case ignoreSharing v of
+    case v of
       Var x els -> text ("@" ++ show x) `pApp` els
       Lam ai b   ->
         mparens (p > 0) $
@@ -1207,7 +1170,7 @@ instance Pretty Term where
             , nest 2 $ pretty (unAbs b) ]
       Lit l                -> pretty l
       Def q els            -> pretty q `pApp` els
-      Con c ci vs          -> pretty (conName c) `pApp` map Apply vs
+      Con c ci vs          -> pretty (conName c) `pApp` vs
       Pi a (NoAbs _ b)     -> mparens (p > 0) $
         sep [ prettyPrec 1 (unDom a) <+> text "->"
             , nest 2 $ pretty b ]
@@ -1218,7 +1181,6 @@ instance Pretty Term where
       Level l     -> prettyPrec p l
       MetaV x els -> pretty x `pApp` els
       DontCare v  -> prettyPrec p v
-      Shared{}    -> __IMPOSSIBLE__
     where
       pApp d els = mparens (not (null els) && p > 9) $
                    sep [d, nest 2 $ fsep (map (prettyPrec 10) els)]
@@ -1279,10 +1241,12 @@ instance Pretty Sort where
       Prop -> text "Prop"
       Inf -> text "Setω"
       SizeUniv -> text "SizeUniv"
-      DLub s b -> mparens (p > 9) $
-        text "dlub" <+> prettyPrec 10 s
-                    <+> parens (sep [ text ("λ " ++ absName b ++ " ->")
-                                    , nest 2 $ pretty (unAbs b) ])
+      PiSort s b -> mparens (p > 9) $
+        text "piSort" <+> prettyPrec 10 s
+                      <+> parens (sep [ text ("λ " ++ absName b ++ " ->")
+                                      , nest 2 $ pretty (unAbs b) ])
+      UnivSort s -> mparens (p > 9) $ text "univSort" <+> prettyPrec 10 s
+      MetaS x es -> prettyPrec p $ MetaV x es
 
 instance Pretty Type where
   prettyPrec p (El _ a) = prettyPrec p a
@@ -1295,11 +1259,10 @@ instance Pretty DBPatVar where
   prettyPrec _ x = text $ patVarNameToString (dbPatVarName x) ++ "@" ++ show (dbPatVarIndex x)
 
 instance Pretty a => Pretty (Pattern' a) where
-  prettyPrec n (VarP x)      = prettyPrec n x
-  prettyPrec _ (DotP t)      = text "." P.<> prettyPrec 10 t
-  prettyPrec _ (AbsurdP _)   = text "()"
-  prettyPrec n (ConP c i nps)= mparens (n > 0) $
-    pretty (conName c) <+> fsep (map pretty ps)
+  prettyPrec n (VarP _o x)   = prettyPrec n x
+  prettyPrec _ (DotP _o t)   = text "." P.<> prettyPrec 10 t
+  prettyPrec n (ConP c i nps)= mparens (n > 0 && not (null nps)) $
+    pretty (conName c) <+> fsep (map (prettyPrec 10) ps)
     where ps = map (fmap namedThing) nps
   -- -- Version with printing record type:
   -- prettyPrec _ (ConP c i ps) = (if b then braces else parens) $ prTy $
@@ -1328,7 +1291,6 @@ instance NFData Term where
     Level l    -> rnf l
     MetaV _ es -> rnf es
     DontCare v -> rnf v
-    Shared{}   -> ()
 
 instance NFData Type where
   rnf (El s v) = rnf (s, v)
@@ -1339,7 +1301,9 @@ instance NFData Sort where
     Prop     -> ()
     Inf      -> ()
     SizeUniv -> ()
-    DLub a b -> rnf (a, unAbs b)
+    PiSort a b -> rnf (a, unAbs b)
+    UnivSort a -> rnf a
+    MetaS _ es -> rnf es
 
 instance NFData Level where
   rnf (Max as) = rnf as
