@@ -67,7 +67,6 @@ instance GetDefs Term where
     Level l    -> getDefs l
     MetaV x vs -> getDefs x >> getDefs vs
     DontCare v -> getDefs v
-    Shared p   -> getDefs $ derefPtr p  -- TODO: exploit sharing!
 
 instance GetDefs MetaId where
   getDefs x = doMeta x
@@ -81,7 +80,9 @@ instance GetDefs Sort where
     Prop      -> return ()
     Inf       -> return ()
     SizeUniv  -> return ()
-    DLub s s' -> getDefs s >> getDefs s'
+    PiSort s s' -> getDefs s >> getDefs s'
+    UnivSort s  -> getDefs s
+    MetaS x es  -> getDefs x >> getDefs es
 
 instance GetDefs Level where
   getDefs (Max ls) = getDefs ls
