@@ -11,8 +11,6 @@ import qualified Data.Map as Map
 import Data.Set (Set)
 import qualified Data.Set as Set
 
-import Control.Applicative
-
 import Agda.Syntax.Common
 import Agda.Syntax.Literal
 import Agda.Syntax.Internal
@@ -23,6 +21,8 @@ import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.CompiledClause
 
 import Agda.Utils.Functor
+import Agda.Utils.NonemptyList
+
 import Agda.Utils.Impossible
 #include "undefined.h"
 
@@ -34,6 +34,7 @@ class NamesIn a where
 
 instance NamesIn a => NamesIn (Maybe a)              where
 instance NamesIn a => NamesIn [a]                    where
+instance NamesIn a => NamesIn (NonemptyList a)       where
 instance NamesIn a => NamesIn (Arg a)                where
 instance NamesIn a => NamesIn (Dom a)                where
 instance NamesIn a => NamesIn (Named n a)            where
@@ -57,7 +58,7 @@ instance (NamesIn a, NamesIn b, NamesIn c) => NamesIn (a, b, c) where
 -- If someone adds a field containing names, this would go unnoticed.
 
 instance NamesIn Definition where
-  namesIn def = namesIn (defType def, theDef def)
+  namesIn def = namesIn (defType def, theDef def, defDisplay def)
 
 instance NamesIn Defn where
   namesIn def = case def of
@@ -90,8 +91,7 @@ instance NamesIn (Pattern' a) where
   namesIn p = case p of
     VarP{}        -> Set.empty
     LitP l        -> namesIn l
-    DotP v        -> namesIn v
-    AbsurdP p     -> namesIn p
+    DotP _ v      -> namesIn v
     ConP c _ args -> namesIn (c, args)
     ProjP _ f     -> namesIn f
 
@@ -104,10 +104,12 @@ instance NamesIn Sort where
     Prop     -> Set.empty
     Inf      -> Set.empty
     SizeUniv -> Set.empty
-    DLub a b -> namesIn (a, b)
+    PiSort a b -> namesIn (a, b)
+    UnivSort a -> namesIn a
+    MetaS _ es -> namesIn es
 
 instance NamesIn Term where
-  namesIn v = case ignoreSharing v of
+  namesIn v = case v of
     Var _ args   -> namesIn args
     Lam _ b      -> namesIn b
     Lit l        -> namesIn l
@@ -118,7 +120,6 @@ instance NamesIn Term where
     Level l      -> namesIn l
     MetaV _ args -> namesIn args
     DontCare v   -> namesIn v
-    Shared{}     -> __IMPOSSIBLE__
 
 instance NamesIn Level where
   namesIn (Max ls) = namesIn ls
@@ -138,6 +139,7 @@ instance NamesIn LevelAtom where
 instance NamesIn Literal where
   namesIn l = case l of
     LitNat{}      -> Set.empty
+    LitWord64{}   -> Set.empty
     LitString{}   -> Set.empty
     LitChar{}     -> Set.empty
     LitFloat{}    -> Set.empty
@@ -152,7 +154,6 @@ instance NamesIn QName   where namesIn x = Set.singleton x  -- interesting case
 instance NamesIn ConHead where namesIn h = namesIn (conName h)
 
 instance NamesIn a => NamesIn (Open a)  where
-instance NamesIn a => NamesIn (Local a) where
 
 instance NamesIn DisplayForm where
   namesIn (Display _ ps v) = namesIn (ps, v)
@@ -184,6 +185,7 @@ instance NamesIn (A.Pattern' a) where
     A.PatternSynP _ c args -> namesIn (c, args)
     A.RecP _ fs            -> namesIn fs
     A.DotP{}               -> __IMPOSSIBLE__    -- Dot patterns are not allowed in pattern synonyms
+    A.WithP _ p            -> namesIn p
 
 instance NamesIn AmbiguousQName where
   namesIn (AmbQ cs) = namesIn cs
