@@ -14,6 +14,7 @@ module Agda.Syntax.Parser
     , moduleNameParser
     , exprParser
     , exprWhereParser
+    , holeContentParser
     , tokensParser
       -- * Parse errors
     , ParseError(..)
@@ -30,7 +31,6 @@ import Control.Monad.Reader
 import Control.Monad.Writer hiding ((<>))
 
 import qualified Data.List as List
-import Data.Typeable ( Typeable )
 
 import Agda.Syntax.Position
 import Agda.Syntax.Parser.Monad as M hiding (Parser, parseFlags)
@@ -52,13 +52,6 @@ import Agda.Utils.IO.UTF8 (readTextFile)
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Pretty
 
-
-
-
-#if __GLASGOW_HASKELL__ <= 708
-import Control.Applicative ((<$>), Applicative)
-#endif
-
 #include "undefined.h"
 import Agda.Utils.Impossible
 
@@ -196,6 +189,15 @@ exprWhereParser = Parser
   , parseLiterate = parseLiterateWithoutComments
   }
 
+-- | Parses an expression or some other content of an interaction hole.
+
+holeContentParser :: Parser HoleContent
+holeContentParser = Parser
+  { parser        = P.holeContentParser
+  , parseFlags    = withoutComments
+  , parseLiterate = parseLiterateWithoutComments
+  }
+
 -- | Gives the parsed token stream (including comments).
 
 tokensParser :: Parser [Token]
