@@ -9,7 +9,7 @@ import Agda.Syntax.Literal (Literal)
 import Agda.Syntax.Position
 
 data Keyword
-        = KwLet | KwIn | KwWhere | KwData | KwCoData
+        = KwLet | KwIn | KwWhere | KwData | KwCoData | KwDo
         | KwPostulate | KwMutual | KwAbstract | KwPrivate | KwInstance
         | KwOverlap
         | KwOpen | KwImport | KwModule | KwPrimitive | KwMacro
@@ -22,9 +22,10 @@ data Keyword
         | KwCOMPILED_DATA | KwCOMPILED_TYPE | KwCOMPILED | KwCOMPILED_EXPORT
         | KwHASKELL | KwFOREIGN | KwCOMPILE
         | KwCOMPILED_JS | KwCOMPILED_UHC | KwCOMPILED_DATA_UHC
-        | KwIMPORT | KwIMPORT_UHC | KwIMPOSSIBLE | KwSTATIC | KwINJECTIVE | KwINLINE
+        | KwIMPORT | KwIMPORT_UHC | KwIMPOSSIBLE | KwSTATIC | KwINJECTIVE | KwINLINE | KwNOINLINE
         | KwETA
         | KwNO_TERMINATION_CHECK | KwTERMINATING | KwNON_TERMINATING
+        | KwWARNING_ON_USAGE
         | KwMEASURE | KwDISPLAY
         | KwREWRITE
         | KwQuoteGoal | KwQuoteContext | KwQuote | KwQuoteTerm
@@ -36,7 +37,7 @@ data Keyword
 
 layoutKeywords :: [Keyword]
 layoutKeywords =
-    [ KwLet, KwWhere, KwPostulate, KwMutual, KwAbstract, KwPrivate, KwInstance, KwMacro, KwPrimitive, KwField ]
+    [ KwLet, KwWhere, KwDo, KwPostulate, KwMutual, KwAbstract, KwPrivate, KwInstance, KwMacro, KwPrimitive, KwField ]
 
 data Symbol
         = SymDot | SymSemi | SymVirtualSemi | SymBar
