@@ -4,10 +4,6 @@
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE NoMonomorphismRestriction  #-}
 
-#if __GLASGOW_HASKELL__ <= 708
-{-# LANGUAGE OverlappingInstances #-}
-#endif
-
 {-| Position information for syntax. Crucial for giving good error messages.
 -}
 
@@ -72,7 +68,11 @@ module Agda.Syntax.Position
   , interleaveRanges
   ) where
 
-import Prelude hiding (null)
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ( (<>), null )
+#else
+import Prelude hiding ( null )
+#endif
 
 import Control.Applicative hiding (empty)
 import Control.Monad
@@ -89,7 +89,6 @@ import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Traversable (Traversable)
 import Data.Data (Data)
-import Data.Typeable (Typeable)
 import Data.Sequence (Seq)
 import qualified Data.Sequence as Seq
 import Data.Void
@@ -99,6 +98,7 @@ import GHC.Generics (Generic)
 import Agda.Utils.FileName
 import Agda.Utils.List
 import qualified Agda.Utils.Maybe.Strict as Strict
+import Agda.Utils.NonemptyList
 import Agda.Utils.Null
 import Agda.Utils.Pretty
 
@@ -128,7 +128,7 @@ data Position' a = Pn
   , posCol  :: !Int32
     -- ^ Column number, counting from 1.
   }
-  deriving (Data, Typeable, Functor, Foldable, Traversable, Generic)
+  deriving (Data, Functor, Foldable, Traversable, Generic)
 
 positionInvariant :: Position' a -> Bool
 positionInvariant p =
@@ -152,7 +152,7 @@ type PositionWithoutFile = Position' ()
 --
 -- Note the invariant which intervals have to satisfy: 'intervalInvariant'.
 data Interval' a = Interval { iStart, iEnd :: !(Position' a) }
-  deriving (Data, Typeable, Eq, Ord, Functor, Foldable, Traversable, Generic)
+  deriving (Data, Eq, Ord, Functor, Foldable, Traversable, Generic)
 
 type Interval            = Interval' SrcFile
 type IntervalWithoutFile = Interval' ()
@@ -197,7 +197,7 @@ data Range' a
   = NoRange
   | Range !a (Seq IntervalWithoutFile)
   deriving
-    (Typeable, Data, Eq, Ord, Functor, Foldable, Traversable, Generic)
+    (Data, Eq, Ord, Functor, Foldable, Traversable, Generic)
 
 type Range = Range' SrcFile
 
@@ -278,6 +278,11 @@ instance HasRange Bool where
 instance HasRange a => HasRange [a] where
     getRange = foldr fuseRange noRange
 
+-- | Precondition: The ranges of the list elements must point to the
+-- same file (or be empty).
+instance HasRange a => HasRange (NonemptyList a) where
+    getRange = Fold.foldr fuseRange noRange
+
 -- | Precondition: The ranges of the tuple elements must point to the
 -- same file (or be empty).
 instance (HasRange a, HasRange b) => HasRange (a,b) where
@@ -478,33 +483,20 @@ instance KillRange Int where
 instance KillRange Integer where
   killRange = id
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPABLE #-} KillRange a => KillRange [a] where
-#else
-instance KillRange a => KillRange [a] where
-#endif
   killRange = map killRange
 
+instance KillRange a => KillRange (NonemptyList a) where
+  killRange = fmap killRange
+
 -- | Overlaps with @KillRange [a]@.
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPING #-} KillRange String where
-#else
-instance KillRange String where
-#endif
   killRange = id
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPABLE #-} KillRange a => KillRange (Map k a) where
-#else
-instance KillRange a => KillRange (Map k a) where
-#endif
   killRange = fmap killRange
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPABLE #-} (Ord a, KillRange a) => KillRange (Set a) where
-#else
-instance (Ord a, KillRange a) => KillRange (Set a) where
-#endif
   killRange = Set.map killRange
 
 instance (KillRange a, KillRange b) => KillRange (a, b) where
