@@ -6,7 +6,11 @@
 -}
 module Agda.Syntax.Scope.Base where
 
-import Prelude hiding (null)
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ( (<>), null )
+#else
+import Prelude hiding ( null )
+#endif
 
 import Control.Arrow (first, second, (***))
 import Control.Applicative hiding (empty)
@@ -22,7 +26,6 @@ import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Maybe
 
-import Data.Typeable (Typeable)
 import Data.Data (Data)
 
 import Agda.Benchmarking
@@ -39,6 +42,7 @@ import qualified Agda.Utils.AssocList as AssocList
 import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.List
+import Agda.Utils.NonemptyList
 import Agda.Utils.Null
 import Agda.Utils.Pretty
 import qualified Agda.Utils.Map as Map
@@ -57,7 +61,7 @@ data Scope = Scope
       , scopeImports        :: Map C.QName A.ModuleName
       , scopeDatatypeModule :: Maybe DataOrRecord
       }
-  deriving (Typeable, Data, Eq, Show)
+  deriving (Data, Eq, Show)
 
 -- | See 'Agda.Syntax.Common.Access'.
 data NameSpaceId
@@ -67,7 +71,7 @@ data NameSpaceId
   | OnlyQualifiedNS  -- ^ Visible (as qualified) from outside,
                      --   but not exported when opening the module.
                      --   Used for qualified constructors.
-  deriving (Typeable, Data, Eq, Bounded, Enum, Show)
+  deriving (Data, Eq, Bounded, Enum, Show)
 
 type ScopeNameSpaces = [(NameSpaceId, NameSpace)]
 
@@ -99,34 +103,43 @@ updateScopeNameSpacesM f s = for (f $ scopeNameSpaces s) $ \ x ->
 data ScopeInfo = ScopeInfo
       { scopeCurrent       :: A.ModuleName
       , scopeModules       :: Map A.ModuleName Scope
+      , scopeVarsToBind    :: LocalVars
       , scopeLocals        :: LocalVars
       , scopePrecedence    :: PrecedenceStack
       , scopeInverseName   :: Map A.QName [C.QName]
       , scopeInverseModule :: Map A.ModuleName [C.QName]
       , scopeInScope       :: InScopeSet
       }
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 instance Eq ScopeInfo where
-  ScopeInfo c1 m1 l1 p1 _ _ _ == ScopeInfo c2 m2 l2 p2 _ _ _ =
-    c1 == c2 && m1 == m2 && l1 == l2 && p1 == p2
+  ScopeInfo c1 m1 v1 l1 p1 _ _ _ == ScopeInfo c2 m2 v2 l2 p2 _ _ _ =
+    c1 == c2 && m1 == m2 && v1 == v2 && l1 == l2 && p1 == p2
 
 -- | Local variables.
 type LocalVars = AssocList C.Name LocalVar
 
+-- | For each bound variable, we want to know whether it was bound by a
+--   λ, Π, module telescope, pattern, or @let@.
+data Binder
+  = LambdaBound  -- ^ @λ@ (currently also used for @Π@ and module parameters)
+  | PatternBound -- ^ @f ... =@
+  | LetBound     -- ^ @let ... in@
+  deriving (Data, Show, Eq)
+
 -- | A local variable can be shadowed by an import.
 --   In case of reference to a shadowed variable, we want to report
 --   a scope error.
 data LocalVar = LocalVar
   { localVar        :: A.Name
     -- ^ Unique ID of local variable.
-  , localLetBound   :: Bool
-    -- ^ Flag whether the variable is introduced by a @let@.
+  , localBinder     :: Binder
+    -- ^ Kind of binder used to introduce the variable (@λ@, @let@, ...).
   , localShadowedBy :: [AbstractName]
      -- ^ If this list is not empty, the local variable is
      --   shadowed by one or more imports.
   }
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 instance Eq LocalVar where
   (==) = (==) `on` localVar
@@ -153,6 +166,13 @@ notShadowedLocal _ = Nothing
 notShadowedLocals :: LocalVars -> AssocList C.Name A.Name
 notShadowedLocals = mapMaybe $ \ (c,x) -> (c,) <$> notShadowedLocal x
 
+-- | Lens for 'scopeVarsToBind'.
+updateVarsToBind :: (LocalVars -> LocalVars) -> ScopeInfo -> ScopeInfo
+updateVarsToBind f sc = sc { scopeVarsToBind = f (scopeVarsToBind sc) }
+
+setVarsToBind :: LocalVars -> ScopeInfo -> ScopeInfo
+setVarsToBind vars = updateVarsToBind (const vars)
+
 -- | Lens for 'scopeLocals'.
 updateScopeLocals :: (LocalVars -> LocalVars) -> ScopeInfo -> ScopeInfo
 updateScopeLocals f sc = sc { scopeLocals = f (scopeLocals sc) }
@@ -179,7 +199,7 @@ data NameSpace = NameSpace
         -- ^ Maps concrete module names to a list of abstract module names.
       , nsInScope :: InScopeSet
       }
-  deriving (Typeable, Data, Eq, Show)
+  deriving (Data, Eq, Show)
 
 type ThingsInScope a = Map C.Name [a]
 type NamesInScope    = ThingsInScope AbstractName
@@ -227,7 +247,7 @@ data KindOfName
   | PatternSynName -- ^ Name of a pattern synonym.
   | MacroName      -- ^ Name of a macro
   | QuotableName   -- ^ A name that can only be quoted.
-  deriving (Eq, Show, Typeable, Data, Enum, Bounded)
+  deriving (Eq, Show, Data, Enum, Bounded)
 
 -- | A list containing all name kinds.
 allKindsOfNames :: [KindOfName]
@@ -244,7 +264,7 @@ data WhyInScope
     -- ^ Imported from another module.
   | Applied C.QName WhyInScope
     -- ^ Imported by a module application.
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 -- | A decoration of 'Agda.Syntax.Abstract.Name.QName'.
 data AbstractName = AbsName
@@ -255,7 +275,7 @@ data AbstractName = AbsName
   , anameLineage :: WhyInScope
     -- ^ Explanation where this name came from.
   }
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 -- | A decoration of abstract syntax module names.
 data AbstractModule = AbsModule
@@ -264,7 +284,7 @@ data AbstractModule = AbsModule
   , amodLineage :: WhyInScope
     -- ^ Explanation where this name came from.
   }
-  deriving (Typeable, Data, Show)
+  deriving (Data, Show)
 
 instance Eq AbstractName where
   (==) = (==) `on` anameName
@@ -286,6 +306,39 @@ instance Ord AbstractModule where
 lensAmodName :: Functor m => (A.ModuleName -> m A.ModuleName) -> AbstractModule -> m AbstractModule
 lensAmodName f am = f (amodName am) <&> \ m -> am { amodName = m }
 
+
+data ResolvedName
+  = -- | Local variable bound by λ, Π, module telescope, pattern, @let@.
+    VarName
+    { resolvedVar      :: A.Name
+    , resolvedBinder   :: Binder    -- ^ What kind of binder?
+    }
+
+  | -- | Function, data/record type, postulate.
+    DefinedName Access AbstractName -- ^ 'anameKind' can be 'DefName', 'MacroName', 'QuotableName'.
+
+  | -- | Record field name.  Needs to be distinguished to parse copatterns.
+    FieldName (NonemptyList AbstractName)       -- ^ @('FldName' ==) . 'anameKind'@ for all names.
+
+  | -- | Data or record constructor name.
+    ConstructorName (NonemptyList AbstractName) -- ^ @('ConName' ==) . 'anameKind'@ for all names.
+
+  | -- | Name of pattern synonym.
+    PatternSynResName (NonemptyList AbstractName) -- ^ @('PatternSynName' ==) . 'anameKind'@ for all names.
+
+  | -- | Unbound name.
+    UnknownName
+  deriving (Data, Show, Eq)
+
+instance Pretty ResolvedName where
+  pretty = \case
+    VarName x _         -> text "variable"    <+> pretty x
+    DefinedName a x     -> pretty a           <+> pretty x
+    FieldName xs        -> text "field"       <+> pretty xs
+    ConstructorName xs  -> text "constructor" <+> pretty xs
+    PatternSynResName x -> text "pattern"     <+> pretty x
+    UnknownName         -> text "<unknown name>"
+
 -- * Operations on name and module maps.
 
 mergeNames :: Eq a => ThingsInScope a -> ThingsInScope a -> ThingsInScope a
@@ -351,6 +404,7 @@ emptyScopeInfo :: ScopeInfo
 emptyScopeInfo = ScopeInfo
   { scopeCurrent       = noModuleName
   , scopeModules       = Map.singleton noModuleName emptyScope
+  , scopeVarsToBind    = []
   , scopeLocals        = []
   , scopePrecedence    = []
   , scopeInverseName   = Map.empty
@@ -416,7 +470,7 @@ zipScope fd fm fs s1 s2 =
                zipWith' (,) (scopeNameSpaces s1) (scopeNameSpaces s2)
          , assert (nsid == nsid')
          ]
-     , scopeImports = Map.union (scopeImports s1) (scopeImports s2)
+     , scopeImports  = (Map.union `on` scopeImports)  s1 s2
      }
   where
     assert True  = True
@@ -802,7 +856,7 @@ data AllowAmbiguousNames
       -- ^ Used for instance arguments to check whether a name is in scope,
       --   but we do not care whether is is ambiguous
   | AmbiguousConProjs
-      -- ^ Ambiguous constructors or projections.
+      -- ^ Ambiguous constructors, projections, or pattern synonyms.
   | AmbiguousNothing
   deriving (Eq)
 
@@ -819,8 +873,6 @@ inverseScopeLookup = inverseScopeLookup' AmbiguousConProjs
 
 inverseScopeLookup' :: AllowAmbiguousNames -> Either A.ModuleName A.QName -> ScopeInfo -> [C.QName]
 inverseScopeLookup' amb name scope = billToPure [ Scoping , InverseScopeLookup ] $
-  -- trace ("importMap = " ++ show importMap) $
-  -- trace ("moduleMap = " ++ show moduleMap) $
   case name of
     Left  m -> best $ filter unambiguousModule $ findModule m
     Right q -> best $ filter unambiguousName   $ findName q
@@ -844,8 +896,7 @@ inverseScopeLookup' amb name scope = billToPure [ Scoping , InverseScopeLookup ]
     unambiguousName   q = amb == AmbiguousAnything
         || unique xs
         || amb == AmbiguousConProjs
-           && (all ((ConName ==) . anameKind) xs
-            || all ((FldName ==) . anameKind) xs)
+           && or [ all ((kind ==) . anameKind) xs | kind <- [ConName, FldName, PatternSynName] ]
       where xs = scopeLookup q scope
 
 recomputeInverseScopeMaps :: ScopeInfo -> ScopeInfo
@@ -972,10 +1023,11 @@ blockOfLines _  [] = []
 blockOfLines hd ss = hd : map (nest 2) ss
 
 instance Pretty ScopeInfo where
-  pretty (ScopeInfo this mods locals ctx _ _ _) = vcat $
+  pretty (ScopeInfo this mods toBind locals ctx _ _ _) = vcat $
     [ text "ScopeInfo"
     , text "  current = " <> pretty this
     ] ++
+    (if null toBind then [] else [ text "  toBind  = " <> pretty locals ]) ++
     (if null locals then [] else [ text "  locals  = " <> pretty locals ]) ++
     [ text "  context = " <> pretty ctx
     , text "  modules"
