@@ -6,9 +6,8 @@
 
 module Agda.Syntax.Scope.Monad where
 
-import Prelude hiding (mapM)
+import Prelude hiding (mapM, any, all)
 import Control.Arrow (first, second, (***))
-import Control.Applicative
 import Control.Monad hiding (mapM, forM)
 import Control.Monad.Writer hiding (mapM, forM)
 import Control.Monad.State hiding (mapM, forM)
@@ -19,6 +18,7 @@ import qualified Data.Map as Map
 import Data.Maybe
 import Data.Set (Set)
 import qualified Data.Set as Set
+import Data.Foldable (any, all)
 import Data.Traversable hiding (for)
 
 import Agda.Syntax.Common
@@ -40,7 +40,9 @@ import Agda.Utils.Function
 import Agda.Utils.Functor
 import Agda.Utils.List
 import Agda.Utils.Maybe
+import Agda.Utils.Monad
 import Agda.Utils.Null (unlessNull)
+import Agda.Utils.NonemptyList
 import Agda.Utils.Pretty
 import Agda.Utils.Size
 import Agda.Utils.Tuple
@@ -60,6 +62,15 @@ isDatatypeModule :: A.ModuleName -> ScopeM (Maybe DataOrRecord)
 isDatatypeModule m = do
    scopeDatatypeModule . Map.findWithDefault __IMPOSSIBLE__ m . scopeModules <$> getScope
 
+
+-- Debugging
+
+printLocals :: Int -> String -> ScopeM ()
+printLocals v s = verboseS "scope.top" v $ do
+  locals <- getLocalVars
+  reportSLn "scope.top" v $ s ++ " " ++ prettyShow locals
+
+
 -- * General operations
 
 getCurrentModule :: ScopeM A.ModuleName
@@ -140,18 +151,17 @@ modifyCurrentNameSpace :: NameSpaceId -> (NameSpace -> NameSpace) -> ScopeM ()
 modifyCurrentNameSpace acc f = modifyCurrentScope $ updateScopeNameSpaces $
   AssocList.updateAt acc f
 
-pushContextPrecedence :: Precedence -> ScopeM ()
-pushContextPrecedence p = modifyScope_ $ \ s -> s { scopePrecedence = pushPrecedence p $ scopePrecedence s }
+pushContextPrecedence :: Precedence -> ScopeM PrecedenceStack
+pushContextPrecedence p = do
+  old <- scopePrecedence <$> getScope
+  modifyScope_ $ \ s -> s { scopePrecedence = pushPrecedence p $ scopePrecedence s }
+  return old
 
-popContextPrecedence :: ScopeM ()
-popContextPrecedence = modifyScope_ $ \ s -> s { scopePrecedence = drop 1 $ scopePrecedence s }
+setContextPrecedence :: PrecedenceStack -> ScopeM ()
+setContextPrecedence ps = modifyScope_ $ \ s -> s { scopePrecedence = ps }
 
 withContextPrecedence :: Precedence -> ScopeM a -> ScopeM a
-withContextPrecedence p m = do
-  pushContextPrecedence p
-  x <- m
-  popContextPrecedence
-  return x
+withContextPrecedence p = bracket_ (pushContextPrecedence p) setContextPrecedence
 
 getLocalVars :: ScopeM LocalVars
 getLocalVars = scopeLocals <$> getScope
@@ -164,11 +174,22 @@ setLocalVars vars = modifyLocalVars $ const vars
 
 -- | Run a computation without changing the local variables.
 withLocalVars :: ScopeM a -> ScopeM a
-withLocalVars m = do
-  vars <- getLocalVars
-  x    <- m
-  setLocalVars vars
-  return x
+withLocalVars = bracket_ getLocalVars setLocalVars
+
+getVarsToBind :: ScopeM LocalVars
+getVarsToBind = scopeVarsToBind <$> getScope
+
+addVarToBind :: C.Name -> LocalVar -> ScopeM ()
+addVarToBind x y = modifyScope_ $ updateVarsToBind $ AssocList.insert x y
+
+-- | After collecting some variable names in the scopeVarsToBind,
+--   bind them all simultaneously.
+bindVarsToBind :: ScopeM ()
+bindVarsToBind = do
+  vars <- getVarsToBind
+  modifyLocalVars (vars++)
+  printLocals 10 "bound variables:"
+  modifyScope_ $ setVarsToBind []
 
 -- * Names
 
@@ -200,38 +221,6 @@ freshAbstractQName fx x = do
 
 -- * Resolving names
 
-data ResolvedName
-  =   -- | Local variable bound by λ, Π, module telescope, pattern, @let@.
-    VarName
-    { resolvedVar      :: A.Name
-    , resolvedLetBound :: Bool    -- ^ Variable bound by @let@?
-    }
-
-  |   -- | Function, data/record type, postulate.
-    DefinedName Access AbstractName
-
-  |   -- | Record field name.  Needs to be distinguished to parse copatterns.
-    FieldName [AbstractName]
-
-  |   -- | Data or record constructor name.
-    ConstructorName [AbstractName]
-
-  |   -- | Name of pattern synonym.
-    PatternSynResName AbstractName
-
-  |   -- | Unbound name.
-    UnknownName
-  deriving (Show, Eq)
-
-instance Pretty ResolvedName where
-  pretty = \case
-    VarName x _         -> text "variable"    <+> pretty x
-    DefinedName a x     -> pretty a           <+> pretty x
-    FieldName xs        -> text "field"       <+> pretty xs
-    ConstructorName xs  -> text "constructor" <+> pretty xs
-    PatternSynResName x -> text "pattern"     <+> pretty x
-    UnknownName         -> text "<unknown name>"
-
 -- | Look up the abstract name referred to by a given concrete name.
 resolveName :: C.QName -> ScopeM ResolvedName
 resolveName = resolveName' allKindsOfNames Nothing
@@ -260,29 +249,27 @@ resolveName' kinds names x = do
         ys -> shadowed ys
       where
       shadowed ys =
-        typeError $ AmbiguousName x $ A.qualify_ y : map anameName ys
+        typeError $ AmbiguousName x $ A.qualify_ y :! map anameName ys
     -- Case: we do not have a local variable x.
     Nothing -> do
       -- Consider only names of one of the given kinds
       let filtKind = filter $ \ y -> anameKind (fst y) `elem` kinds
       -- Consider only names in the given set of names
           filtName = filter $ \ y -> maybe True (Set.member (aName (fst y))) names
-      case filtKind $ filtName $ scopeLookup' x scope of
-        [] -> return UnknownName
+      caseListNe (filtKind $ filtName $ scopeLookup' x scope) (return UnknownName) $ \ case
+        ds       | all ((ConName ==) . anameKind . fst) ds ->
+          return $ ConstructorName $ fmap (upd . fst) ds
 
-        ds       | all ((==ConName) . anameKind . fst) ds ->
-          return $ ConstructorName $ map (upd . fst) ds
+        ds       | all ((FldName ==) . anameKind . fst) ds ->
+          return $ FieldName $ fmap (upd . fst) ds
 
-        ds       | all ((==FldName) . anameKind . fst) ds ->
-          return $ FieldName $ map (upd . fst) ds
+        ds       | all ((PatternSynName ==) . anameKind . fst) ds ->
+          return $ PatternSynResName $ fmap (upd . fst) ds
 
-        [(d, a)] | anameKind d == PatternSynName ->
-          return $ PatternSynResName $ upd d
-
-        [(d, a)] ->
+        (d, a) :! [] ->
           return $ DefinedName a $ upd d
 
-        ds -> typeError $ AmbiguousName x (map (anameName . fst) ds)
+        ds -> typeError $ AmbiguousName x (fmap (anameName . fst) ds)
   where
   upd d = updateConcreteName d $ unqualify x
   updateConcreteName :: AbstractName -> C.Name -> AbstractName
@@ -293,10 +280,9 @@ resolveName' kinds names x = do
 resolveModule :: C.QName -> ScopeM AbstractModule
 resolveModule x = do
   ms <- scopeLookup x <$> getScope
-  case ms of
-    [AbsModule m why] -> return $ AbsModule (m `withRangeOf` x) why
-    []                -> typeError $ NoSuchModule x
-    ms                -> typeError $ AmbiguousModule x (map amodName ms)
+  caseListNe ms (typeError $ NoSuchModule x) $ \ case
+    AbsModule m why :! [] -> return $ AbsModule (m `withRangeOf` x) why
+    ms                    -> typeError $ AmbiguousModule x (fmap amodName ms)
 
 -- | Get the notation of a name. The name is assumed to be in scope.
 getNotation
@@ -311,12 +297,12 @@ getNotation x ns = do
     DefinedName _ d     -> return $ notation d
     FieldName ds        -> return $ oneNotation ds
     ConstructorName ds  -> return $ oneNotation ds
-    PatternSynResName n -> return $ notation n
+    PatternSynResName n -> return $ oneNotation n
     UnknownName         -> __IMPOSSIBLE__
   where
     notation = namesToNotation x . qnameName . anameName
     oneNotation ds =
-      case mergeNotations $ map notation ds of
+      case mergeNotations $ map notation $ toList ds of
         [n] -> n
         _   -> __IMPOSSIBLE__
 
@@ -324,11 +310,15 @@ getNotation x ns = do
 
 -- | Bind a variable.
 bindVariable
-  :: Bool    -- ^ Let-bound variable?
+  :: Binder  -- ^ @λ@, @Π@, @let@, ...?
   -> C.Name  -- ^ Concrete name.
   -> A.Name  -- ^ Abstract name.
   -> ScopeM ()
-bindVariable b x y = modifyScope_ $ updateScopeLocals $ AssocList.insert x $ LocalVar y b []
+bindVariable b x y = modifyLocalVars $ AssocList.insert x $ LocalVar y b []
+
+-- | Temporarily unbind a variable. Used for non-recursive lets.
+unbindVariable :: C.Name -> ScopeM a -> ScopeM a
+unbindVariable x = bracket_ (getLocalVars <* modifyLocalVars (AssocList.delete x)) (modifyLocalVars . const)
 
 -- | Bind a defined name. Must not shadow anything.
 bindName :: Access -> KindOfName -> C.Name -> A.QName -> ScopeM ()
@@ -343,7 +333,7 @@ bindName acc kind x y = do
     VarName z _         -> clash $ A.qualify (mnameFromList []) z
     FieldName       ds  -> ambiguous FldName ds
     ConstructorName ds  -> ambiguous ConName ds
-    PatternSynResName n -> clash $ anameName n
+    PatternSynResName n -> ambiguous PatternSynName n
     UnknownName         -> success
   let ns = if isNoName x then PrivateNS else localNameSpace acc
   modifyCurrentScope $ addNamesToScope ns x ys
@@ -351,10 +341,9 @@ bindName acc kind x y = do
     success = return [ AbsName y kind Defined ]
     clash   = typeError . ClashingDefinition (C.QName x)
 
-    ambiguous k ds@(d:_) =
-      if kind == k && all ((==k) . anameKind) ds
-      then success else clash $ anameName d
-    ambiguous k [] = __IMPOSSIBLE__
+    ambiguous k ds =
+      if kind == k && all ((== k) . anameKind) ds
+      then success else clash $ anameName (headNe ds)
 
 -- | Rebind a name. Use with care!
 --   Ulf, 2014-06-29: Currently used to rebind the name defined by an
