@@ -22,7 +22,7 @@ module Agda.Syntax.Translation.InternalToAbstract
   ) where
 
 import Prelude hiding (mapM_, mapM, null)
-import Control.Applicative hiding (empty)
+import Control.Arrow ((&&&))
 import Control.Monad.State hiding (mapM_, mapM)
 import Control.Monad.Reader hiding (mapM_, mapM)
 
@@ -30,7 +30,8 @@ import Data.Foldable (Foldable, foldMap)
 import qualified Data.List as List
 import qualified Data.Map as Map
 import Data.Maybe
-import Data.Semigroup (Semigroup, Monoid, (<>), mempty, mappend)
+import Data.Monoid ( Monoid, mempty, mappend )
+import Data.Semigroup ( Semigroup, (<>) )
 import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Traversable (Traversable, traverse, mapM)
@@ -44,7 +45,7 @@ import qualified Agda.Syntax.Concrete.Name as C
 import Agda.Syntax.Concrete (FieldAssignment'(..), exprFieldA)
 import Agda.Syntax.Info as Info
 import Agda.Syntax.Abstract as A
-import Agda.Syntax.Abstract.Pattern ( foldAPattern )
+import Agda.Syntax.Abstract.Pattern
 import Agda.Syntax.Abstract.Pretty
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Internal.Pattern as I
@@ -102,12 +103,27 @@ nelims e (I.Apply arg : es) = do
   arg <- reify arg  -- This replaces the arg by _ if irrelevant
   dontShowImp <- not <$> showImplicitArguments
   let hd | notVisible arg && dontShowImp = e
-         | otherwise                     = A.App noExprInfo e arg
+         | otherwise                     = A.App defaultAppInfo_ e arg
   nelims hd es
-nelims e (I.Proj o@ProjPrefix d  : es) =
-  nelims (A.App noExprInfo (A.Proj o $ AmbQ [d]) $ defaultNamedArg e) es
-nelims e (I.Proj o d  : es) =
-  nelims (A.App noExprInfo e (defaultNamedArg $ A.Proj o $ AmbQ [d])) es
+nelims e (I.Proj ProjPrefix d : es)             = nelimsProjPrefix e d es
+nelims e (I.Proj o          d : es) | isSelf e  = nelimsProjPrefix e d es
+                                    | otherwise =
+  nelims (A.App defaultAppInfo_ e (defaultNamedArg $ A.Proj o $ unambiguous d)) es
+
+nelimsProjPrefix :: Expr -> QName -> [I.Elim' (Named_ Expr)] -> TCM Expr
+nelimsProjPrefix e d es =
+  nelims (A.App defaultAppInfo_ (A.Proj ProjPrefix $ unambiguous d) $ defaultNamedArg e) es
+
+-- | If we are inside a record definition, the record value (self)
+--   is a variable with name "".
+--   In this case, we need to use a prefix-projection.  (Issue #2868.)
+--   Prefix projections magically print correctly since the thing
+--   we are projecting from has a null name, so vanishes in the visible world.
+--   We love hacks, don't we?  Sigh.
+isSelf :: Expr -> Bool
+isSelf = \case
+  A.Var x -> null $ prettyShow x
+  _ -> False
 
 -- | Drops hidden arguments unless --show-implicit.
 elims :: Expr -> [I.Elim' Expr] -> TCM Expr
@@ -166,7 +182,7 @@ instance Reify DisplayTerm Expr where
   reify d = case d of
     DTerm v -> reifyTerm False v
     DDot  v -> reify v
-    DCon c ci vs -> apps (A.Con (AmbQ [conName c])) =<< reify vs
+    DCon c ci vs -> apps (A.Con (unambiguous (conName c))) =<< reify vs
     DDef f es -> elims (A.Def f) =<< reify es
     DWithApp u us es0 -> do
       (e, es) <- reify (u, us)
@@ -185,9 +201,14 @@ reifyDisplayForm f es fallback = do
 --   rewrite a lhs with a display form.
 --
 --   Note: we are not necessarily in the empty context upon entry!
-reifyDisplayFormP :: A.SpineLHS -> TCM A.SpineLHS
-reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
-  ifNotM displayFormsEnabled (return lhs) $ {- else -} do
+reifyDisplayFormP
+  :: QName         -- ^ LHS head symbol
+  -> A.Patterns    -- ^ Patterns to be taken into account to find display form.
+  -> A.Patterns    -- ^ Remaining trailing patterns ("with patterns").
+  -> TCM (QName, A.Patterns) -- ^ New head symbol and new patterns.
+reifyDisplayFormP f ps wps = do
+  let fallback = return (f, ps ++ wps)
+  ifNotM displayFormsEnabled fallback $ {- else -} do
     -- Try to rewrite @f 0 1 2 ... |ps|-1@ to a dt.
     -- Andreas, 2014-06-11  Issue 1177:
     -- I thought we need to add the placeholders for ps to the context,
@@ -205,17 +226,17 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
         -- for the @i@th pattern of @ps@.
         -- Andreas, 2014-06-11:
         -- Are we sure that @d@ did not use @var i@ otherwise?
-        lhs' <- displayLHS ps wps d
+        (f', ps', wps') <- displayLHS ps d
         reportSDoc "reify.display" 70 $ do
-          doc <- prettyA lhs'
+          doc <- prettyA $ SpineLHS empty f' (ps' ++ wps' ++ wps)
           return $ vcat
             [ text "rewritten lhs to"
             , text "  lhs' = " <+> doc
             ]
-        reifyDisplayFormP lhs'
+        reifyDisplayFormP f' ps' (wps' ++ wps)
       _ -> do
         reportSLn "reify.display" 70 $ "display form absent or not valid as lhs"
-        return lhs
+        fallback
   where
     -- Andreas, 2015-05-03: Ulf, please comment on what
     -- is the idea behind okDisplayForm.
@@ -250,7 +271,7 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
     okToDropE I.Proj{}    = False
 
     okToDrop :: Arg I.Term -> Bool
-    okToDrop arg = notVisible arg && case ignoreSharing $ unArg arg of
+    okToDrop arg = notVisible arg && case unArg arg of
       I.Var _ []   -> True
       I.DontCare{} -> True  -- no matching on irrelevant things.  __IMPOSSIBLE__ anyway?
       I.Level{}    -> True  -- no matching on levels. __IMPOSSIBLE__ anyway?
@@ -265,7 +286,7 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
 
     okTerm :: I.Term -> Bool
     okTerm (I.Var _ []) = True
-    okTerm (I.Con c ci vs) = all okArg vs
+    okTerm (I.Con c ci vs) = all okElim vs
     okTerm (I.Def x []) = isNoName $ qnameToConcrete x -- Handling wildcards in display forms
     okTerm _            = False
 
@@ -278,41 +299,40 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
     flattenWith (DTerm (I.Def f es)) = (f, map (fmap DTerm) es, [])
     flattenWith _ = __IMPOSSIBLE__
 
-    displayLHS :: [NamedArg A.Pattern] -> [A.Pattern] -> DisplayTerm -> TCM A.SpineLHS
-    displayLHS ps wps d = do
+    displayLHS
+      :: A.Patterns   -- ^ Patterns to substituted into display term.
+      -> DisplayTerm  -- ^ Display term.
+      -> TCM (QName, A.Patterns, A.Patterns)  -- ^ New head, patterns, with-patterns.
+    displayLHS ps d = do
         let (f, vs, es) = flattenWith d
-        ds <- mapM (namedArg <.> elimToPat) es
-        vs <- mapM elimToPat vs
-        return $ SpineLHS i f vs (ds ++ wps)
+        ps  <- mapM elimToPat vs
+        wps <- mapM (updateNamedArg (A.WithP empty) <.> elimToPat) es
+        return (f, ps, wps)
       where
         argToPat :: Arg DisplayTerm -> TCM (NamedArg A.Pattern)
         argToPat arg = traverse termToPat arg
 
         elimToPat :: I.Elim' DisplayTerm -> TCM (NamedArg A.Pattern)
         elimToPat (I.Apply arg) = argToPat arg
-        elimToPat (I.Proj o d)  = return $ defaultNamedArg $ A.ProjP patNoRange o $ AmbQ [d]
+        elimToPat (I.Proj o d)  = return $ defaultNamedArg $ A.ProjP patNoRange o $ unambiguous d
 
+        -- | Substitute variables in display term by patterns.
         termToPat :: DisplayTerm -> TCM (Named_ A.Pattern)
 
+        -- Main action HERE:
         termToPat (DTerm (I.Var n [])) = return $ unArg $ fromMaybe __IMPOSSIBLE__ $ ps !!! n
 
         termToPat (DCon c ci vs)          = fmap unnamed <$> tryRecPFromConP =<< do
-           A.ConP (ConPatInfo ci patNoRange) (AmbQ [conName c]) <$> mapM argToPat vs
+           A.ConP (ConPatInfo ci patNoRange False) (unambiguous (conName c)) <$> mapM argToPat vs
 
         termToPat (DTerm (I.Con c ci vs)) = fmap unnamed <$> tryRecPFromConP =<< do
-           A.ConP (ConPatInfo ci patNoRange) (AmbQ [conName c]) <$> mapM (argToPat . fmap DTerm) vs
+           A.ConP (ConPatInfo ci patNoRange False) (unambiguous (conName c)) <$> mapM (elimToPat . fmap DTerm) vs
 
         termToPat (DTerm (I.Def _ [])) = return $ unnamed $ A.WildP patNoRange
         termToPat (DDef _ [])          = return $ unnamed $ A.WildP patNoRange
 
-        -- Currently we don't keep track of the origin of a dot pattern in the internal syntax,
-        -- so here we give __IMPOSSIBLE__. This is only used for printing purposes, the origin
-        -- should not be used anyway after this point.
-        -- Andreas, 2017-02-14: This crashes with -v 100.
-        -- termToPat (DDot v)             = A.DotP patNoRange __IMPOSSIBLE__ <$> termToExpr v
-        -- termToPat v                    = A.DotP patNoRange __IMPOSSIBLE__ <$> reify v -- __IMPOSSIBLE__
-        termToPat (DDot v)             = unnamed . A.DotP patNoRange Inserted <$> termToExpr v
-        termToPat v                    = unnamed . A.DotP patNoRange Inserted <$> reify v
+        termToPat (DDot v)             = unnamed . A.DotP patNoRange <$> termToExpr v
+        termToPat v                    = unnamed . A.DotP patNoRange <$> reify v
 
         len = length ps
 
@@ -325,8 +345,9 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
           reportSLn "reify.display" 60 $ "termToExpr " ++ show v
           -- After unSpine, a Proj elimination is __IMPOSSIBLE__!
           case unSpine v of
-            I.Con c ci vs ->
-              apps (A.Con (AmbQ [conName c])) =<< argsToExpr vs
+            I.Con c ci es -> do
+              let vs = fromMaybe __IMPOSSIBLE__ $ mapM isApplyElim es
+              apps (A.Con (unambiguous (conName c))) =<< argsToExpr vs
             I.Def f es -> do
               let vs = fromMaybe __IMPOSSIBLE__ $ mapM isApplyElim es
               apps (A.Def f) =<< argsToExpr vs
@@ -360,12 +381,13 @@ reifyTerm expandAnonDefs0 v = do
   -- then we print system-generated projections as postfix, else prefix.
   havePfp <- optPostfixProjections <$> pragmaOptions
   let pred = if havePfp then (== ProjPrefix) else (/= ProjPostfix)
-  v <- ignoreSharing <$> instantiate v
+  v <- instantiate v
   case applyUnless metasBare (unSpine' pred) v of
     I.Var n es   -> do
         x  <- liftTCM $ nameOfBV n `catchError` \_ -> freshName_ ("@" ++ show n)
         elims (A.Var x) =<< reify es
     I.Def x es   -> do
+      reportSLn "reify.def" 100 $ "reifying def " ++ prettyShow x
       reifyDisplayForm x es $ reifyDef expandAnonDefs x es
     I.Con c ci vs -> do
       let x = conName c
@@ -376,9 +398,9 @@ reifyTerm expandAnonDefs0 v = do
           let keep (a, v) = showImp || visible a
           r  <- getConstructorData x
           xs <- getRecordFieldNames r
-          vs <- map unArg <$> reify vs
+          vs <- map unArg <$> reify (fromMaybe __IMPOSSIBLE__ $ allApplyElims vs)
           return $ A.Rec noExprInfo $ map (Left . uncurry FieldAssignment . mapFst unArg) $ filter keep $ zip xs vs
-        False -> reifyDisplayForm x (map I.Apply vs) $ do
+        False -> reifyDisplayForm x vs $ do
           def <- getConstInfo x
           let Constructor{conPars = np} = theDef def
           -- if we are the the module that defines constructor x
@@ -387,9 +409,9 @@ reifyTerm expandAnonDefs0 v = do
           -- the number of parameters is greater (if the data decl has
           -- extra parameters) or equal (if not) to n
           when (n > np) __IMPOSSIBLE__
-          let h = A.Con (AmbQ [x])
+          let h = A.Con (unambiguous x)
           if null vs then return h else do
-            es <- reify vs
+            es <- reify (fromMaybe __IMPOSSIBLE__ $ allApplyElims vs)
             -- Andreas, 2012-04-20: do not reify parameter arguments of constructor
             -- if the first regular constructor argument is hidden
             -- we turn it into a named argument, in order to avoid confusion
@@ -422,7 +444,7 @@ reifyTerm expandAnonDefs0 v = do
 --    I.Lam info b | isAbsurdBody b -> return $ A. AbsurdLam noExprInfo $ getHiding info
     I.Lam info b    -> do
       (x,e) <- reify b
-      return $ A.Lam (setOrigin (getOrigin info) defaultLamInfo_) (DomainFree info x) e
+      return $ A.Lam exprNoRange (DomainFree info $ BindName x) e
       -- Andreas, 2011-04-07 we do not need relevance information at internal Lambda
     I.Lit l        -> reify l
     I.Level l      -> reify l
@@ -440,8 +462,9 @@ reifyTerm expandAnonDefs0 v = do
             {- else -} (reify a)
       where
         mkPi b (Arg info a) = do
-          (x, b) <- reify b
-          return $ A.Pi noExprInfo [TypedBindings noRange $ Arg info (TBind noRange [pure x] a)] b
+          -- #2776: Out-of-scope dots are not helpful at this point.
+          (x, b) <- reify b{ absName = unNotInScopeName $ absName b }
+          return $ A.Pi noExprInfo [TypedBindings noRange $ Arg info (TBind noRange [pure $ BindName x] a)] b
         -- We can omit the domain type if it doesn't have any free variables
         -- and it's mentioned in the target type.
         domainFree a b = do
@@ -454,7 +477,6 @@ reifyTerm expandAnonDefs0 v = do
       ifM (asks envPrintMetasBare) {-then-} (return x') {-else-} $
         elims x' =<< reify es
     I.DontCare v -> A.DontCare <$> reifyTerm expandAnonDefs v
-    I.Shared p   -> reifyTerm expandAnonDefs $ derefPtr p
   where
     -- Andreas, 2012-10-20  expand a copy if not in scope
     -- to improve error messages.
@@ -499,7 +521,17 @@ reifyTerm expandAnonDefs0 v = do
                 | isAbsurdLambdaName x -> do
                   -- get hiding info from last pattern, which should be ()
                   let h = getHiding $ last $ namedClausePats cl
-                  elims (A.AbsurdLam defaultLamInfo_ h) =<< reify (drop n es)
+                      n = length (namedClausePats cl) - 1  -- drop all args before the absurd one
+                      absLam = A.AbsurdLam exprNoRange h
+                  if | n > length es -> do -- We don't have all arguments before the absurd one!
+                        let name (I.VarP _ x) = patVarNameToString $ dbPatVarName x
+                            name _            = __IMPOSSIBLE__  -- only variables before absurd pattern
+                            vars = map (getArgInfo &&& name . namedArg) $ drop (length es) $ init $ namedClausePats cl
+                            lam (i, s) = do
+                              x <- freshName_ s
+                              return $ A.Lam exprNoRange (A.DomainFree i $ A.BindName x)
+                        foldr ($) absLam <$> mapM lam vars
+                      | otherwise -> elims absLam =<< reify (drop n es)
 
       -- Otherwise (no absurd lambda):
        _ -> do
@@ -511,13 +543,18 @@ reifyTerm expandAnonDefs0 v = do
 
         -- Check whether we have an extended lambda and display forms are on.
         df <- displayFormsEnabled
-        toppars <- size <$> do lookupSection $ qnameModule x
-        let extLam = case def of
-             Function{ funExtLam = Just{}, funProjection = Just{} } -> __IMPOSSIBLE__
-             Function{ funExtLam = Just (ExtLamInfo h nh) } -> Just (toppars + h + nh)
-             _ -> Nothing
+
+        -- #3004: give up if we have to print a pattern lambda inside its own body!
+        alreadyPrinting <- view ePrintingPatternLambdas
+
+        extLam <- case def of
+          Function{ funExtLam = Just{}, funProjection = Just{} } -> __IMPOSSIBLE__
+          Function{ funExtLam = Just (ExtLamInfo m) } -> Just . size <$> lookupSection m
+          _ -> return Nothing
         case extLam of
-          Just pars | df -> reifyExtLam x pars (defClauses defn) es
+          Just pars | df, notElem x alreadyPrinting ->
+            locally ePrintingPatternLambdas (x :) $
+            reifyExtLam x pars (defClauses defn) es
 
         -- Otherwise (ordinary function call):
           _ -> do
@@ -585,7 +622,7 @@ reifyTerm expandAnonDefs0 v = do
              [ "  pad = " ++ show pad
              , "  nes = " ++ show nes
              ]
-           let hd = List.foldl' (A.App noExprInfo) (A.Def x) pad
+           let hd = List.foldl' (A.App defaultAppInfo_) (A.Def x) pad
            nelims hd =<< reify nes
 
     -- Andreas, 2016-07-06 Issue #2047
@@ -623,7 +660,7 @@ reifyTerm expandAnonDefs0 v = do
       cls <- mapM (reify . NamedClause x False . (`applyE` pars)) cls
       let cx    = nameConcrete $ qnameName x
           dInfo = mkDefInfo cx noFixity' PublicAccess ConcreteDef (getRange x)
-      elims (A.ExtendedLam defaultLamInfo_ dInfo x cls) =<< reify rest
+      elims (A.ExtendedLam exprNoRange dInfo x cls) =<< reify rest
 
 -- | @nameFirstIfHidden (x:a) ({e} es) = {x = e} es@
 nameFirstIfHidden :: Dom (ArgName, t) -> [Elim' a] -> [Elim' (Named_ a)]
@@ -641,7 +678,7 @@ instance Reify i a => Reify (Named n i) (Named n a) where
 instance (Reify i a) => Reify (Arg i) (Arg a) where
   reify (Arg info i) = Arg info <$> (flip reifyWhen i =<< condition)
     where condition = (return (argInfoHiding info /= Hidden) `or2M` showImplicitArguments)
-              `and2M` (return (argInfoRelevance info /= Irrelevant) `or2M` showIrrelevantArguments)
+              `and2M` (return (getRelevance info /= Irrelevant) `or2M` showIrrelevantArguments)
   reifyWhen b i = traverse (reifyWhen b) i
 
 
@@ -662,23 +699,19 @@ instance (Ord k, Monoid v) => Monoid (MonoidMap k v) where
 -- | Removes implicit arguments that are not needed, that is, that don't bind
 --   any variables that are actually used and doesn't do pattern matching.
 --   Doesn't strip any arguments that were written explicitly by the user.
-stripImplicits :: ([NamedArg A.Pattern], [A.Pattern]) ->
-                  TCM ([NamedArg A.Pattern], [A.Pattern])
-stripImplicits (ps, wps) = do          -- v if show-implicit we don't need the names
-  ifM showImplicitArguments (return (map (unnamed . namedThing <$>) ps, wps)) $ do
+stripImplicits :: A.Patterns -> TCM A.Patterns
+stripImplicits ps = do
+  -- if --show-implicit we don't need the names
+  ifM showImplicitArguments (return $ map (unnamed . namedThing <$>) ps) $ do
     reportSLn "reify.implicit" 30 $ unlines
       [ "stripping implicits"
       , "  ps   = " ++ show ps
-      , "  wps  = " ++ show wps
       ]
-    let allps       = ps ++ map defaultNamedArg wps
-        sps         = blankDots $ strip allps
-        (ps', wps') = splitAt (length sps - length wps) sps
+    let ps' = blankDots $ strip ps
     reportSLn "reify.implicit" 30 $ unlines
       [ "  ps'  = " ++ show ps'
-      , "  wps' = " ++ show (map namedArg wps')
       ]
-    return (ps', map namedArg wps')
+    return ps'
     where
       -- Replace variables in dot patterns by an underscore _ if they are hidden
       -- in the pattern. This is slightly nicer than making the implicts explicit.
@@ -688,9 +721,6 @@ stripImplicits (ps, wps) = do          -- v if show-implicit we don't need the n
         where
           stripArgs _ [] = []
           stripArgs fixedPos (a : as)
-            -- Andreas, 2017-01-18, issue #819: preserves _ when splitting:
-            -- An Inserted visible variable comes form a WildP and is restored as such.
-            | visible a, getOrigin a == Inserted, varOrDot (namedArg a) = goWild
             -- A hidden non-UserWritten variable is removed if not needed for
             -- correct position of the following hidden arguments.
             | canStrip a =
@@ -721,13 +751,14 @@ stripImplicits (ps, wps) = do          -- v if show-implicit we don't need the n
             A.ConP i c ps -> A.ConP i c $ stripArgs True ps
             A.ProjP{}     -> p
             A.DefP _ _ _  -> p
-            A.DotP _ _ e  -> p
+            A.DotP _ e    -> p
             A.WildP _     -> p
             A.AbsurdP _   -> p
             A.LitP _      -> p
             A.AsP i x p   -> A.AsP i x $ stripPat p
             A.PatternSynP _ _ _ -> __IMPOSSIBLE__ -- p
             A.RecP i fs   -> A.RecP i $ map (fmap stripPat) fs  -- TODO Andreas: is this right?
+            A.WithP i p   -> A.WithP i $ stripPat p -- TODO #2822: right?
 
           varOrDot A.VarP{}      = True
           varOrDot A.WildP{}     = True
@@ -760,27 +791,24 @@ instance (BlankVars a, BlankVars b) => BlankVars (Either a b) where
   blank bound (Left x)  = Left $ blank bound x
   blank bound (Right y) = Right $ blank bound y
 
-instance BlankVars A.NamedDotPattern where
-  blank bound = id
-
-instance BlankVars A.StrippedDotPattern where
+instance BlankVars A.ProblemEq where
   blank bound = id
 
 instance BlankVars A.Clause where
-  blank bound (A.Clause lhs namedDots strippedDots rhs [] ca) =
+  blank bound (A.Clause lhs strippedPats rhs (A.WhereDecls _ []) ca) =
     let bound' = varsBoundIn lhs `Set.union` bound
     in  A.Clause (blank bound' lhs)
-                 (blank bound' namedDots)
-                 (blank bound' strippedDots)
-                 (blank bound' rhs) [] ca
-  blank bound (A.Clause lhs namedDots strippedDots rhs (_:_) ca) = __IMPOSSIBLE__
+                 (blank bound' strippedPats)
+                 (blank bound' rhs) noWhereDecls ca
+  blank bound (A.Clause lhs strippedPats rhs _ ca) = __IMPOSSIBLE__
 
 instance BlankVars A.LHS where
-  blank bound (A.LHS i core wps) = uncurry (A.LHS i) $ blank bound (core, wps)
+  blank bound (A.LHS i core) = A.LHS i $ blank bound core
 
 instance BlankVars A.LHSCore where
   blank bound (A.LHSHead f ps) = A.LHSHead f $ blank bound ps
   blank bound (A.LHSProj p b ps) = uncurry (A.LHSProj p) $ blank bound (b, ps)
+  blank bound (A.LHSWith h wps ps) = uncurry (uncurry A.LHSWith) $ blank bound ((h, wps), ps)
 
 instance BlankVars A.Expr where
   blank bound e = case e of
@@ -826,7 +854,7 @@ instance BlankVars RHS where
   blank bound (RHS e mc)             = RHS (blank bound e) mc
   blank bound AbsurdRHS              = AbsurdRHS
   blank bound (WithRHS _ es clauses) = __IMPOSSIBLE__ -- NZ
-  blank bound (RewriteRHS xes rhs _) = __IMPOSSIBLE__ -- NZ
+  blank bound (RewriteRHS xes spats rhs _) = __IMPOSSIBLE__ -- NZ
 
 instance BlankVars A.LamBinding where
   blank bound b@A.DomainFree{} = b
@@ -849,15 +877,16 @@ class Binder a where
   varsBoundIn = foldMap varsBoundIn
 
 instance Binder A.LHS where
-  varsBoundIn (A.LHS _ core ps) = varsBoundIn (core, ps)
+  varsBoundIn (A.LHS _ core) = varsBoundIn core
 
 instance Binder A.LHSCore where
-  varsBoundIn (A.LHSHead _ ps)   = varsBoundIn ps
-  varsBoundIn (A.LHSProj _ b ps) = varsBoundIn (b, ps)
+  varsBoundIn (A.LHSHead _ ps)     = varsBoundIn ps
+  varsBoundIn (A.LHSProj _ b ps)   = varsBoundIn (b, ps)
+  varsBoundIn (A.LHSWith h wps ps) = varsBoundIn ((h, wps), ps)
 
 instance Binder A.Pattern where
   varsBoundIn = foldAPattern $ \case
-    A.VarP x            -> if prettyShow x == "()" then empty else singleton x -- TODO: get rid of this hack?
+    A.VarP x            -> singleton $ unBind x
     A.AsP _ x _         -> empty
     A.ConP _ _ _        -> empty
     A.ProjP{}           -> empty
@@ -868,9 +897,10 @@ instance Binder A.Pattern where
     A.LitP{}            -> empty
     A.PatternSynP _ _ _ -> empty
     A.RecP _ _          -> empty
+    A.WithP _ _         -> empty
 
 instance Binder A.LamBinding where
-  varsBoundIn (A.DomainFree _ x) = singleton x
+  varsBoundIn (A.DomainFree _ x) = singleton $ unBind x
   varsBoundIn (A.DomainFull b)   = varsBoundIn b
 
 instance Binder TypedBindings where
@@ -881,7 +911,7 @@ instance Binder TypedBinding where
   varsBoundIn (TLet _ bs)    = varsBoundIn bs
 
 instance Binder LetBinding where
-  varsBoundIn (LetBind _ _ x _ _) = singleton x
+  varsBoundIn (LetBind _ _ x _ _) = singleton $ unBind x
   varsBoundIn (LetPatBind _ p _)  = varsBoundIn p
   varsBoundIn LetApply{}          = empty
   varsBoundIn LetOpen{}           = empty
@@ -890,6 +920,9 @@ instance Binder LetBinding where
 instance Binder (WithHiding Name) where
   varsBoundIn (WithHiding _ x) = singleton x
 
+instance Binder (WithHiding BindName) where
+  varsBoundIn (WithHiding _ x) = singleton $ unBind x
+
 instance Binder a => Binder (FieldAssignment' a) where
 instance Binder a => Binder (Arg a)              where
 instance Binder a => Binder (Named x a)          where
@@ -913,33 +946,54 @@ reifyPatterns = mapM $ stripNameFromExplicit <.> traverse (traverse reifyPat)
     reifyPat p = do
      liftTCM $ reportSLn "reify.pat" 80 $ "reifying pattern " ++ show p
      case p of
-      I.VarP x -> do
-        n <- liftTCM $ nameOfBV $ dbPatVarIndex x
-        case dbPatVarName x of
-          "_"  -> return $ A.VarP n
-          -- Andreas, 2017-09-03: TODO for #2580
-          -- Patterns @VarP "()"@ should have been replaced by @AbsurdP@, but the
-          -- case splitter still produces them.
-          y    -> if prettyShow (nameConcrete n) == "()" then return $ A.VarP n else
-            -- Andreas, 2017-09-03, issue #2729
-            -- Restore original pattern name.  AbstractToConcrete picks unique names.
-            return $ A.VarP n { nameConcrete = C.Name noRange [ C.Id y ] }
-      I.DotP v -> do
-        t <- liftTCM $ reify v
-        -- This is only used for printing purposes, so the Origin shouldn't be
-        -- used after this point anyway.
-        return $ A.DotP patNoRange Inserted t
-        -- WAS: return $ A.DotP patNoRange __IMPOSSIBLE__ t
-        -- Crashes on -v 100.
-      I.AbsurdP p -> return $ A.AbsurdP patNoRange
+      I.VarP PatODot x -> reifyDotP $ var $ dbPatVarIndex x
+      I.VarP PatOWild _ -> return $ A.WildP patNoRange
+      I.VarP PatOAbsurd _ -> return $ A.AbsurdP patNoRange
+      I.VarP _ x -> reifyVarP x
+      I.DotP PatOWild _ -> return $ A.WildP patNoRange
+      I.DotP PatOAbsurd _ -> return $ A.AbsurdP patNoRange
+      -- If Agda turned a user variable @x@ into @.x@, print it back as @x@.
+      I.DotP (PatOVar x) v@(I.Var i []) -> do
+        x' <- nameOfBV i
+        if nameConcrete x == nameConcrete x' then
+          return $ A.VarP (BindName x')
+        else
+          reifyDotP v
+      I.DotP o v -> reifyDotP v
       I.LitP l  -> return $ A.LitP l
-      I.ProjP o d     -> return $ A.ProjP patNoRange o $ AmbQ [d]
-      I.ConP c cpi ps -> do
-        liftTCM $ reportSLn "reify.pat" 60 $ "reifying pattern " ++ show p
-        tryRecPFromConP =<< do A.ConP ci (AmbQ [conName c]) <$> reifyPatterns ps
-        where
-          ci = ConPatInfo origin patNoRange
-          origin = fromMaybe ConOCon $ I.conPRecord cpi
+      I.ProjP o d     -> return $ A.ProjP patNoRange o $ unambiguous d
+      I.ConP c cpi ps -> case conPRecord cpi of
+        Just PatOWild   -> return $ A.WildP patNoRange
+        Just PatOAbsurd -> return $ A.AbsurdP patNoRange
+        _               -> reifyConP c cpi ps
+
+    reifyVarP :: MonadTCM tcm => DBPatVar -> tcm A.Pattern
+    reifyVarP x = do
+      n <- liftTCM $ nameOfBV $ dbPatVarIndex x
+      case dbPatVarName x of
+        "_"  -> return $ A.VarP $ A.BindName n
+        -- Andreas, 2017-09-03: TODO for #2580
+        -- Patterns @VarP "()"@ should have been replaced by @AbsurdP@, but the
+        -- case splitter still produces them.
+        y    -> if prettyShow (nameConcrete n) == "()" then return $ A.VarP $ A.BindName n else
+          -- Andreas, 2017-09-03, issue #2729
+          -- Restore original pattern name.  AbstractToConcrete picks unique names.
+          return $ A.VarP $ A.BindName n { nameConcrete = C.Name noRange [ C.Id y ] }
+
+    reifyDotP :: MonadTCM tcm => Term -> tcm A.Pattern
+    reifyDotP v = do
+      t <- liftTCM $ reify v
+      return $ A.DotP patNoRange t
+
+    reifyConP :: MonadTCM tcm
+              => ConHead -> ConPatternInfo -> [NamedArg DeBruijnPattern]
+              -> tcm A.Pattern
+    reifyConP c cpi ps = do
+      tryRecPFromConP =<< do A.ConP ci (unambiguous (conName c)) <$> reifyPatterns ps
+      where
+        ci = ConPatInfo origin patNoRange False
+        origin = fromConPatternInfo cpi
+
 
 -- | If the record constructor is generated or the user wrote a record pattern,
 --   turn constructor pattern into record pattern.
@@ -948,8 +1002,8 @@ tryRecPFromConP :: MonadTCM tcm => A.Pattern -> tcm A.Pattern
 tryRecPFromConP p = do
   let fallback = return p
   case p of
-    A.ConP ci (AmbQ [c]) ps -> do
-        caseMaybeM (liftTCM $ isRecordConstructor c) fallback $ \ (r, def) -> do
+    A.ConP ci c ps -> do
+        caseMaybeM (liftTCM $ isRecordConstructor $ headAmbQ c) fallback $ \ (r, def) -> do
           -- If the record constructor is generated or the user wrote a record pattern,
           -- print record pattern.
           -- Otherwise, print constructor pattern.
@@ -971,28 +1025,26 @@ instance Reify NamedClause A.Clause where
       ++ "\n  toDrop = " ++ show toDrop
       ++ "\n  cl     = " ++ show cl
     ps  <- reifyPatterns $ namedClausePats cl
-    lhs <- liftTCM $ reifyDisplayFormP $ SpineLHS info f ps [] -- LHS info (LHSHead f ps) []
+    lhs <- uncurry (SpineLHS empty) <$> reifyDisplayFormP f ps []
     -- Unless @toDrop@ we have already dropped the module patterns from the clauses
-    -- (e.g. for extended lambdas).
+    -- (e.g. for extended lambdas). We still get here with toDrop = True and
+    -- pattern lambdas when doing make-case, so take care to drop the right
+    -- number of parameters.
     lhs <- if not toDrop then return lhs else do
-      nfv <- getDefFreeVars f `catchError` \_ -> return 0
+      nfv <- (size <.> lookupSection =<< getDefModule f) `catchError` \_ -> return 0
       return $ dropParams nfv lhs
     lhs <- stripImps lhs
     reportSLn "reify.clause" 60 $ "reifying NamedClause, lhs = " ++ show lhs
     rhs <- caseMaybe (clauseBody cl) (return AbsurdRHS) $ \ e -> do
        RHS <$> reify e <*> pure Nothing
     reportSLn "reify.clause" 60 $ "reifying NamedClause, rhs = " ++ show rhs
-    let result = A.Clause (spineToLhs lhs) [] [] rhs [] (I.clauseCatchall cl)
+    let result = A.Clause (spineToLhs lhs) [] rhs A.noWhereDecls (I.clauseCatchall cl)
     reportSLn "reify.clause" 60 $ "reified NamedClause, result = " ++ show result
     return result
     where
-      perm = fromMaybe __IMPOSSIBLE__ $ clausePerm cl
-      info = LHSRange noRange
-
-      dropParams n (SpineLHS i f ps wps) = SpineLHS i f (drop n ps) wps
-      stripImps (SpineLHS i f ps wps) = do
-        (ps, wps) <- stripImplicits (ps, wps)
-        return $ SpineLHS i f ps wps
+      dropParams n (SpineLHS i f ps) = SpineLHS i f $ drop n ps
+      stripImps :: SpineLHS -> TCM SpineLHS
+      stripImps (SpineLHS i f ps) =  SpineLHS i f <$> stripImplicits ps
 
 instance Reify Type Expr where
     reifyWhen = reifyWhenE
@@ -1007,21 +1059,30 @@ instance Reify Sort Expr where
         I.Type (I.Max [I.ClosedLevel n]) -> return $ A.Set noExprInfo n
         I.Type a -> do
           a <- reify a
-          return $ A.App noExprInfo (A.Set noExprInfo 0) (defaultNamedArg a)
+          return $ A.App defaultAppInfo_ (A.Set noExprInfo 0) (defaultNamedArg a)
         I.Prop       -> return $ A.Prop noExprInfo
         I.Inf       -> A.Var <$> freshName_ ("Setω" :: String)
         I.SizeUniv  -> do
           I.Def sizeU [] <- primSizeUniv
           return $ A.Def sizeU
-        I.DLub s1 s2 -> do
-          lub <- freshName_ ("dLub" :: String) -- TODO: hack
+        I.PiSort s1 s2 -> do
+          pis <- freshName_ ("piSort" :: String) -- TODO: hack
           (e1,e2) <- reify (s1, I.Lam defaultArgInfo $ fmap Sort s2)
-          let app x y = A.App noExprInfo x (defaultNamedArg y)
-          return $ A.Var lub `app` e1 `app` e2
+          let app x y = A.App defaultAppInfo_ x (defaultNamedArg y)
+          return $ A.Var pis `app` e1 `app` e2
+        I.UnivSort s -> do
+          univs <- freshName_ ("univSort" :: String) -- TODO: hack
+          e <- reify s
+          return $ A.App defaultAppInfo_ (A.Var univs) $ defaultNamedArg e
+        I.MetaS x es -> reify $ I.MetaV x es
 
 instance Reify Level Expr where
   reifyWhen = reifyWhenE
-  reify l = reify =<< reallyUnLevelView l
+  reify l   = ifM haveLevels (reify =<< reallyUnLevelView l) $ {-else-} do
+    -- Andreas, 2017-09-18, issue #2754
+    -- While type checking the level builtins, they are not
+    -- available for debug printing.  Thus, print some garbage instead.
+    A.Var <$> freshName_ (".#Lacking_Level_Builtins#" :: String)
 
 instance (Free i, Reify i a) => Reify (Abs i) (Name, a) where
   reify (NoAbs x v) = (,) <$> freshName_ x <*> reify v
@@ -1032,7 +1093,7 @@ instance (Free i, Reify i a) => Reify (Abs i) (Name, a) where
     s <- return $ if isUnderscore s && 0 `freeIn` v then "z" else s
 
     x <- freshName_ s
-    e <- addContext' x -- type doesn't matter
+    e <- addContext x -- type doesn't matter
          $ reify v
     return (x,e)
 
@@ -1042,7 +1103,7 @@ instance Reify I.Telescope A.Telescope where
     Arg info e <- reify arg
     (x,bs)  <- reify tel
     let r = getRange e
-    return $ TypedBindings r (Arg info (TBind r [pure x] e)) : bs
+    return $ TypedBindings r (Arg info (TBind r [pure $ BindName x] e)) : bs
 
 instance Reify i a => Reify (Dom i) (Arg a) where
     reify (Dom info i) = Arg info <$> reify i
