@@ -18,7 +18,8 @@ import Control.Monad.State
 
 import Data.Foldable (Foldable)
 import Data.Traversable (Traversable)
-import Data.Semigroup (Semigroup(..), Monoid(..))
+import Data.Monoid ( Monoid(..) )
+import Data.Semigroup ( Semigroup(..) )
 
 import Agda.Interaction.Options
 
@@ -458,7 +459,7 @@ isCoinductiveProjection mustBeRecursive q = liftTCM $ do
             reportSLn "term.guardedness" 40 $ prettyShow q ++ " is coinductive; record type is " ++ prettyShow r
             if not mustBeRecursive then return True else do
               reportSLn "term.guardedness" 40 $ prettyShow q ++ " must be recursive"
-              if not (recRecursive rdef) then return False else do
+              if not (safeRecRecursive rdef) then return False else do
                 reportSLn "term.guardedness" 40 $ prettyShow q ++ " has been declared recursive, doing actual check now..."
                 -- TODO: the following test for recursiveness of a projection should be cached.
                 -- E.g., it could be stored in the @Projection@ component.
@@ -487,6 +488,25 @@ isCoinductiveProjection mustBeRecursive q = liftTCM $ do
       _ -> do
         reportSLn "term.guardedness" 40 $ prettyShow q ++ " is not a proper projection"
         return False
+  where
+  -- Andreas, 2018-02-24, issue #2975, example:
+  -- @
+  -- record R : Set where
+  --   coinductive
+  --   field force : R
+
+  --   r : R
+  --   force r = r
+  -- @
+  -- The termination checker expects the positivity checker to have run on the
+  -- record declaration R to know whether R is recursive.
+  -- However, here, because the awkward processing of record declarations (see #434),
+  -- that has not happened.  To avoid crashing (as in Agda 2.5.3),
+  -- we rather give the possibly wrong answer here,
+  -- restoring the behavior of Agda 2.5.2.  TODO: fix record declaration checking.
+  safeRecRecursive :: Defn -> Bool
+  safeRecRecursive (Record { recMutual = Just qs }) = not $ null qs
+  safeRecRecursive _ = False
 
 -- * De Bruijn pattern stuff
 
@@ -515,12 +535,11 @@ class UsableSizeVars a where
 
 instance UsableSizeVars DeBruijnPattern where
   usableSizeVars = foldrPattern $ \case
-    VarP x    -> const $ ifM terGetUseSizeLt (return $ VarSet.singleton $ dbPatVarIndex x) $
+    VarP _ x   -> const $ ifM terGetUseSizeLt (return $ VarSet.singleton $ dbPatVarIndex x) $
                    {-else-} return mempty
     ConP c _ _ -> conUseSizeLt $ conName c
     LitP{}     -> none
     DotP{}     -> none
-    AbsurdP{}  -> none
     ProjP{}    -> none
     where none _ = return mempty
 
@@ -533,12 +552,11 @@ instance UsableSizeVars [DeBruijnPattern] where
 
 instance UsableSizeVars (Masked DeBruijnPattern) where
   usableSizeVars (Masked m p) = (`foldrPattern` p) $ \case
-    VarP x    -> const $ ifM terGetUseSizeLt (return $ VarSet.singleton $ dbPatVarIndex x) $
+    VarP _ x   -> const $ ifM terGetUseSizeLt (return $ VarSet.singleton $ dbPatVarIndex x) $
                    {-else-} return mempty
     ConP c _ _ -> if m then none else conUseSizeLt $ conName c
     LitP{}     -> none
     DotP{}     -> none
-    AbsurdP{}  -> none
     ProjP{}    -> none
     where none _ = return mempty
 
@@ -609,7 +627,7 @@ terSetSizeDepth tel cont = do
     forM (telToList tel) $ \ dom -> do
       a <- reduce $ snd $ unDom dom
       ifM (isJust <$> isSizeType a) (return 1) {- else -} $
-        case ignoreSharing $ unEl a of
+        case unEl a of
           MetaV{} -> return 1
           _       -> return 0
   terLocal (set terSizeDepth n) cont
