@@ -17,7 +17,11 @@ module Agda.Termination.TermCheck
     , Result
     ) where
 
-import Prelude hiding (null)
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ( (<>), null )
+#else
+import Prelude hiding ( null )
+#endif
 
 import Control.Applicative hiding (empty)
 import Control.Monad.Reader
@@ -194,17 +198,17 @@ termMutual names0 = ifNotM (optTerminationCheck <$> pragmaOptions) (return mempt
 
   -- NO_TERMINATION_CHECK
   if (Info.mutualTermCheck i `elem` [ NoTerminationCheck, Terminating ]) then do
-      reportSLn "term.warn.yes" 2 $ "Skipping termination check for " ++ prettyShow names
+      reportSLn "term.warn.yes" 10 $ "Skipping termination check for " ++ prettyShow names
       forM_ allNames $ \ q -> setTerminates q True -- considered terminating!
       return mempty
   -- NON_TERMINATING
     else if (Info.mutualTermCheck i == NonTerminating) then do
-      reportSLn "term.warn.yes" 2 $ "Considering as non-terminating: " ++ prettyShow names
+      reportSLn "term.warn.yes" 10 $ "Considering as non-terminating: " ++ prettyShow names
       forM_ allNames $ \ q -> setTerminates q False
       return mempty
   -- Trivially terminating (non-recursive)
     else ifM skip (do
-      reportSLn "term.warn.yes" 2 $ "Trivially terminating: " ++ prettyShow names
+      reportSLn "term.warn.yes" 10 $ "Trivially terminating: " ++ prettyShow names
       forM_ allNames $ \ q -> setTerminates q True
       return mempty)
    $ {- else -} do
@@ -408,7 +412,7 @@ termFunction name = do
 typeEndsInDef :: MonadTCM tcm => Type -> tcm (Maybe QName)
 typeEndsInDef t = liftTCM $ do
   TelV _ core <- telView t
-  case ignoreSharing $ unEl core of
+  case unEl core of
     Def d vs -> return $ Just d
     _        -> return Nothing
 
@@ -521,18 +525,18 @@ instance TermToPattern a b => TermToPattern (Named c a) (Named c b) where
 --   termToPattern t = unnamed <$> termToPattern t
 
 instance TermToPattern Term DeBruijnPattern where
-  termToPattern t = (liftTCM $ ignoreSharing <$> constructorForm t) >>= \case
+  termToPattern t = (liftTCM $ constructorForm t) >>= \case
     -- Constructors.
-    Con c _ args -> ConP c noConPatternInfo . map (fmap unnamed) <$> termToPattern args
+    Con c _ args -> ConP c noConPatternInfo . map (fmap unnamed) <$> termToPattern (fromMaybe __IMPOSSIBLE__ $ allApplyElims args)
     Def s [Apply arg] -> do
       suc <- terGetSizeSuc
       if Just s == suc then ConP (ConHead s Inductive []) noConPatternInfo . map (fmap unnamed) <$> termToPattern [arg]
-       else return $ DotP t
+       else return $ dotP t
     DontCare t  -> termToPattern t -- OR: __IMPOSSIBLE__  -- removed by stripAllProjections
     -- Leaves.
-    Var i []    -> VarP . (`DBPatVar` i) . prettyShow <$> nameOfBV i
+    Var i []    -> varP . (`DBPatVar` i) . prettyShow <$> nameOfBV i
     Lit l       -> return $ LitP l
-    t           -> return $ DotP t
+    t           -> return $ dotP t
 
 
 -- | Masks all non-data/record type patterns if --without-K.
@@ -590,8 +594,8 @@ termClause' clause = do
 
   where
     parseDotP = \case
-      DotP t -> termToDBP t
-      p      -> return p
+      DotP o t -> termToDBP t
+      p        -> return p
     stripCoCon p = case p of
       ConP (ConHead c _ _) _ _ -> do
         ifM ((Just c ==) <$> terGetSizeSuc) (return p) $ {- else -} do
@@ -655,7 +659,9 @@ instance ExtractCalls Sort where
       Inf        -> return empty
       SizeUniv   -> return empty
       Type t     -> terUnguarded $ extract t  -- no guarded levels
-      DLub s1 s2 -> extract (s1, s2)
+      PiSort s1 s2 -> extract (s1, s2)
+      UnivSort s -> extract s
+      MetaS x es -> return empty
 
 -- | Extract recursive calls from a type.
 
@@ -732,10 +738,9 @@ function g es0 = ifM (terGetInlineWithFunctions `and2M` do isJust <$> isWithFunc
 
     -- We have to reduce constructors in case they're reexported.
     -- Andreas, Issue 1530: constructors have to be reduced deep inside terms,
-    -- thus, we need to use traverseTermM.  Sharing is handled by traverseTermM,
-    -- so no ignoreSharing needed here.
+    -- thus, we need to use traverseTermM.
     let (reduceCon :: Term -> TCM Term) = traverseTermM $ \ t -> case t of
-           Con c ci vs -> (`apply` vs) <$> reduce (Con c ci [])  -- make sure we don't reduce the arguments
+           Con c ci vs -> (`applyE` vs) <$> reduce (Con c ci [])  -- make sure we don't reduce the arguments
            _ -> return t
 
     -- Reduce constructors only when this call is actually a recursive one.
@@ -776,7 +781,15 @@ function g es0 = ifM (terGetInlineWithFunctions `and2M` do isJust <$> isWithFunc
          -- We still need to reduce constructors, even when with-inlining happened.
          es <- -- ifM terGetHaveInlinedWith (return es0) {-else-} $
            liftTCM $ forM es0 $
-             etaContract <=< traverse reduceCon <=< instantiateFull
+             -- 2017-09-09, re issue #2732
+             -- The eta-contraction here does not seem necessary to make structural order
+             -- comparison not having to worry about eta.
+             -- Maybe we thought an eta redex could come from a meta instantiation.
+             -- However, eta-contraction is already performed by instantiateFull.
+             -- See test/Succeed/Issue2732-termination.agda.
+             -- etaContract <=<
+             traverse reduceCon <=< instantiateFull
+
            -- 2017-05-16, issue #2403: Argument normalization is too expensive,
            -- even if we only expand non-recursive functions.
            -- Argument normalization TURNED OFF.
@@ -850,11 +863,11 @@ instance ExtractCalls Term where
 
     -- Instantiate top-level MetaVar.
     t <- liftTCM $ instantiate t
-    case ignoreSharing t of
+    case t of
 
       -- Constructed value.
-      Con ConHead{conName = c} _ args -> do
-
+      Con ConHead{conName = c} _ es -> do
+        let args = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
         -- A constructor preserves the guardedness of all its arguments.
         let argsg = zip args $ repeat True
 
@@ -907,8 +920,6 @@ instance ExtractCalls Term where
         -- Andreas, 2014-03-26 Benchmark discontinued, < 0.3% spent on levels.
         extract l
 
-      Shared{} -> __IMPOSSIBLE__
-
 -- | Extract recursive calls from level expressions.
 
 deriving instance ExtractCalls Level
@@ -932,7 +943,7 @@ maskSizeLt dom@(Dom info a) = liftTCM $ do
     (Nothing, _)  -> __IMPOSSIBLE__
     (Just size, Just sizelt) -> do
       TelV tel c <- telView a
-      case ignoreSharingType a of
+      case a of
         El s (Def d [v]) | d == sizelt -> return $ Dom info $
           abstract tel $ El s $ Def size []
         _ -> return dom
@@ -961,7 +972,7 @@ compareArgs es = do
   -- Count the number of coinductive projection(pattern)s in caller and callee.
   -- Only recursive coinductive projections are eligible (Issue 1209).
   projsCaller <- length <$> do
-    filterM (isCoinductiveProjection True) $ mapMaybe (fmap (head . unAmbQ . snd) . isProjP . getMasked) pats
+    filterM (isCoinductiveProjection True) $ mapMaybe (fmap (headAmbQ . snd) . isProjP . getMasked) pats
   projsCallee <- length <$> do
     filterM (isCoinductiveProjection True) $ mapMaybe (fmap snd . isProjElim) es
   cutoff <- terGetCutOff
@@ -1085,10 +1096,9 @@ offsetFromConstructor c = maybe 1 (const 0) <$> do
 subPatterns :: DeBruijnPattern -> [DeBruijnPattern]
 subPatterns = foldPattern $ \case
   ConP _ _ ps -> map namedArg ps
-  VarP _      -> mempty
+  VarP _ _    -> mempty
   LitP _      -> mempty
-  DotP _      -> mempty
-  AbsurdP _   -> mempty
+  DotP _ _    -> mempty
   ProjP _ _   -> mempty
 
 
@@ -1130,7 +1140,7 @@ instance StripAllProjections Args where
 
 instance StripAllProjections Term where
   stripAllProjections t = do
-    case ignoreSharing t of
+    case t of
       Var i es   -> Var i <$> stripAllProjections es
       Con c ci ts -> Con c ci <$> stripAllProjections ts
       Def d es   -> Def d <$> stripAllProjections es
@@ -1138,15 +1148,13 @@ instance StripAllProjections Term where
       _ -> return t
 
 -- | @compareTerm' t dbpat@
---
---   Precondition: top meta variable resolved
 
 compareTerm' :: Term -> Masked DeBruijnPattern -> TerM Order
 compareTerm' v mp@(Masked m p) = do
   suc  <- terGetSizeSuc
   cutoff <- terGetCutOff
   let ?cutoff = cutoff
-  v <- return $ ignoreSharing v
+  v <- liftTCM (instantiate v)
   case (v, p) of
 
     -- Andreas, 2013-11-20 do not drop projections,
@@ -1189,7 +1197,7 @@ compareTerm' v mp@(Masked m p) = do
 
     (Lit l, _) -> do
       v <- liftTCM $ constructorForm v
-      case ignoreSharing v of
+      case v of
         Lit{}       -> return Order.unknown
         v           -> compareTerm' v mp
 
@@ -1198,14 +1206,16 @@ compareTerm' v mp@(Masked m p) = do
     (Con{}, ConP c _ ps) | any (isSubTerm v . namedArg) ps ->
       decr True <$> offsetFromConstructor (conName c)
 
-    (Con c _ ts, ConP c' _ ps) | conName c == conName c'->
+    (Con c _ es, ConP c' _ ps) | conName c == conName c'->
+      let ts = fromMaybe __IMPOSSIBLE__ $ allApplyElims es in
       compareConArgs ts ps
 
     (Con _ _ [], _) -> return Order.le
 
     -- new case for counting constructors / projections
     -- register also increase
-    (Con c _ ts, _) -> do
+    (Con c _ es, _) -> do
+      let ts = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
       increase <$> offsetFromConstructor (conName c)
                <*> (infimum <$> mapM (\ t -> compareTerm' (unArg t) mp) ts)
 
@@ -1215,18 +1225,18 @@ compareTerm' v mp@(Masked m p) = do
 subTerm :: (?cutoff :: CutOff) => Term -> DeBruijnPattern -> Order
 subTerm t p = if equal t p then Order.le else properSubTerm t p
   where
-    equal (Shared p) dbp = equal (derefPtr p) dbp
-    equal (Con c _ ts) (ConP c' _ ps) =
+    equal (Con c _ es) (ConP c' _ ps) =
+      let ts = fromMaybe __IMPOSSIBLE__ $ allApplyElims es in
       and $ (conName c == conName c')
           : (length ts == length ps)
           : zipWith (\ t p -> equal (unArg t) (namedArg p)) ts ps
-    equal (Var i []) (VarP x) = i == dbPatVarIndex x
+    equal (Var i []) (VarP _ x) = i == dbPatVarIndex x
     equal (Lit l)    (LitP l') = l == l'
     -- Terms.
     -- Checking for identity here is very fragile.
     -- However, we cannot do much more, as we are not allowed to normalize t.
     -- (It might diverge, and we are just in the process of termination checking.)
-    equal t         (DotP t') = t == t'
+    equal t         (DotP _ t') = t == t'
     equal _ _ = False
 
     properSubTerm t (ConP _ _ ps) =
@@ -1275,8 +1285,7 @@ compareVar i (Masked m p) = do
     ProjP{}   -> no
     LitP{}    -> no
     DotP{}   -> no
-    AbsurdP{} -> no
-    VarP x    -> compareVarVar i (Masked m x)
+    VarP _ x  -> compareVarVar i (Masked m x)
 
     ConP s _ [p] | Just (conName s) == suc ->
       setUsability True . decrease 1 <$> compareVar i (notMasked $ namedArg p)
@@ -1306,4 +1315,4 @@ compareVarVar i (Masked m x@(DBPatVar _ j))
       res <- isBounded i
       case res of
         BoundedNo  -> return Order.unknown
-        BoundedLt v -> setUsability u . decrease 1 <$> compareTerm' v (Masked m $ VarP x)
+        BoundedLt v -> setUsability u . decrease 1 <$> compareTerm' v (Masked m $ varP x)
