@@ -4,7 +4,6 @@
 -- | Functions for abstracting terms over other terms.
 module Agda.TypeChecking.Abstract where
 
-import Control.Applicative
 import Control.Monad
 import Control.Monad.State
 import Data.Function
@@ -45,6 +44,10 @@ piAbstractTerm v a b = do
     sep [ text "piAbstract" <+> sep [ prettyTCM v <+> text ":", nest 2 $ prettyTCM a ]
         , nest 2 $ text "from" <+> prettyTCM b
         , nest 2 $ text "-->" <+> prettyTCM fun ]
+  reportSDoc "tc.abstract" 70 $
+    sep [ text "piAbstract" <+> sep [ (text . show) v <+> text ":", nest 2 $ (text . show) a ]
+        , nest 2 $ text "from" <+> (text . show) b
+        , nest 2 $ text "-->" <+> (text . show) fun ]
   return fun
 
 -- | @piAbstract (v, a) b[v] = (w : a) -> b[w]@
@@ -86,7 +89,7 @@ instance IsPrefixOf Args where
 
 instance IsPrefixOf Term where
   isPrefixOf u v =
-    case (ignoreSharing u, ignoreSharing v) of
+    case (u, v) of
       (Var   i us, Var   j vs) | i == j  -> us `isPrefixOf` vs
       (Def   f us, Def   g vs) | f == g  -> us `isPrefixOf` vs
       (Con c _ us, Con d _ vs) | c == d  -> us `isPrefixOf` vs
@@ -101,6 +104,11 @@ abstractTerm a u@Con{} b v = do
         , nest 2 $ sep [ prettyTCM u <+> text ":", nest 2 $ prettyTCM a ]
         , text "over"
         , nest 2 $ sep [ prettyTCM v <+> text ":", nest 2 $ prettyTCM b ] ]
+  reportSDoc "tc.abstract" 70 $
+    sep [ text "Abstracting"
+        , nest 2 $ sep [ (text . show) u <+> text ":", nest 2 $ (text . show) a ]
+        , text "over"
+        , nest 2 $ sep [ (text . show) v <+> text ":", nest 2 $ (text . show) b ] ]
 
   hole <- qualify <$> currentModule <*> freshName_ "hole"
   noMutualBlock $ addConstant hole $ defaultDefn defaultArgInfo hole a Axiom
@@ -124,21 +132,18 @@ abstractTerm a u@Con{} b v = do
                       , nest 2 $ sep [ prettyTCM v <+> text ":", nest 2 $ prettyTCM b ] ]
                 return v
 
-  v <- catchError_ (checkInternal' (defaultAction { preAction = abstr }) v b) $ \ err -> do
-        reportSDoc "impossible" 10 $
-          vcat [ text "Type error in term to abstract"
-               , nest 2 $ (prettyTCM =<< getContextTelescope) <+> text "⊢"
-               , nest 2 $ sep [ prettyTCM v <+> text ":", nest 2 $ prettyTCM b ]
-               , nest 2 $ prettyTCM err ]
-        reportSDoc "impossible" 60 $
-          vcat [ text "Type error in term to abstract (raw)"
-               , nest 2 $ ((text . show) =<< getContextTelescope) <+> text "⊢"
-               , nest 2 $ sep [ (text . show) v <+> text ":", nest 2 $ (text . show) b ]
-               ]
-        __IMPOSSIBLE__
-  reportSDoc "tc.abstract" 50 $ sep [ text "Resulting abstraction", nest 2 $ prettyTCM v ]
+  -- #2763: This can fail if the user is with-abstracting incorrectly (for
+  -- instance, abstracting over a first component of a sigma without also
+  -- abstracting the second component). In this case we skip abstraction
+  -- altogether and let the type check of the final with-function type produce
+  -- the error message.
+  res <- catchError_ (checkInternal' (defaultAction { preAction = abstr }) v b) $ \ err -> do
+        reportSDoc "tc.abstract.ill-typed" 40 $
+          text "Skipping typed abstraction over ill-typed term" <?> (prettyTCM v <?> (text ":" <+> prettyTCM b))
+        return v
+  reportSDoc "tc.abstract" 50 $ text "Resulting abstraction" <?> prettyTCM res
   modifySignature $ updateDefinitions $ HMap.delete hole
-  return $ absTerm (Def hole args) v
+  return $ absTerm (Def hole args) res
 
 abstractTerm _ u _ v = return $ absTerm u v -- Non-constructors can use untyped abstraction
 
@@ -162,13 +167,9 @@ instance AbsTerm Term where
       Sort s      -> Sort $ absT s
       MetaV m vs  -> MetaV m $ absT vs
       DontCare mv -> DontCare $ absT mv
-      Shared p    -> Shared $ absT p
       where
         absT x = absTerm u x
 
-instance AbsTerm a => AbsTerm (Ptr a) where
-  absTerm u = fmap (absTerm u)
-
 instance AbsTerm Type where
   absTerm u (El s v) = El (absTerm u s) (absTerm u v)
 
@@ -178,7 +179,9 @@ instance AbsTerm Sort where
     Prop       -> Prop
     Inf        -> Inf
     SizeUniv   -> SizeUniv
-    DLub s1 s2 -> DLub (absS s1) (absS s2)
+    PiSort s1 s2 -> PiSort (absS s1) (absS s2)
+    UnivSort s -> UnivSort $ absS s
+    MetaS x es -> MetaS x $ absS es
     where absS x = absTerm u x
 
 instance AbsTerm Level where
