@@ -18,7 +18,6 @@ module Agda.TypeChecking.CheckInternal
   ) where
 
 import Control.Arrow ((&&&), (***), first, second)
-import Control.Applicative
 import Control.Monad
 
 import Agda.Syntax.Common
@@ -34,6 +33,7 @@ import Agda.TypeChecking.ProjectionLike (elimView)
 import Agda.TypeChecking.Records (getDefType)
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
+import Agda.TypeChecking.Sort
 import Agda.TypeChecking.Telescope
 
 import Agda.Utils.Functor (($>))
@@ -70,16 +70,17 @@ checkType' t = do
     , prettyTCM t
     ]
   v <- elimView True $ unEl t -- bring projection-like funs in post-fix form
-  case ignoreSharing v of
+  case v of
     Pi a b -> do
       s1 <- checkType' $ unDom a
       s2 <- (b $>) <$> do
-        addContext (absName b, a) $ do
-          checkType' $ absBody b
-      return $ dLub s1 s2
+        let goInside = case b of Abs{}   -> addContext (absName b, a)
+                                 NoAbs{} -> id
+        goInside $ checkType' $ unAbs b
+      inferPiSort s1 s2
     Sort s -> do
       _ <- checkSort defaultAction s
-      return $ sSuc s
+      inferUnivSort s
     Var i es   -> do
       a <- typeOfBV i
       checkTypeSpine a (Var i   []) es
@@ -94,7 +95,6 @@ checkType' t = do
     v@Lit{}    -> typeError $ InvalidType v
     v@Level{}  -> typeError $ InvalidType v
     DontCare v -> checkType' $ t $> v
-    Shared{}   -> __IMPOSSIBLE__
 
 checkTypeSpine :: Type -> Term -> Elims -> TCM Sort
 checkTypeSpine a self es = shouldBeSort =<< do snd <$> inferSpine a self es
@@ -125,7 +125,7 @@ eraseUnusedAction :: Action
 eraseUnusedAction = defaultAction { postAction = eraseUnused }
   where
     eraseUnused :: Type -> Term -> TCM Term
-    eraseUnused t v = case ignoreSharing v of
+    eraseUnused t v = case v of
       Def f es -> do
         pols <- getPolarity f
         return $ Def f $ eraseIfNonvariant pols es
@@ -150,7 +150,7 @@ checkInternal' action v t = do
   -- Bring projection-like funs in post-fix form,
   -- even lone ones (True).
   v <- elimView True =<< preAction action t v
-  postAction action t =<< case ignoreSharing v of
+  postAction action t =<< case v of
     Var i es   -> do
       a <- typeOfBV i
       checkSpine action a (Var i []) es t
@@ -163,7 +163,7 @@ checkInternal' action v t = do
     Con c ci vs -> do
       -- We need to fully apply the constructor to make getConType work!
       fullyApplyCon c vs t $ \ _d _dt _pars a vs' tel t -> do
-        Con c ci vs2 <- checkArgs action a (Con c ci []) vs' t
+        Con c ci vs2 <- checkSpine action a (Con c ci []) vs' t
         -- Strip away the extra arguments
         return $ applySubst (strengthenS __IMPOSSIBLE__ (size tel))
           $ Con c ci $ take (length vs) vs2
@@ -188,22 +188,21 @@ checkInternal' action v t = do
       -- TODO: checkPTS sa sb s
       goInside $ Pi a . mkRng <$> checkInternal' action (unEl $ unAbs b) (sort sb)
     Sort s     -> do
-      s <- checkSort action s  -- this ensures @s /= Inf@
-      Sort s <$ ((sSuc s `leqSort`) =<< shouldBeSort t)
+      s <- checkSort action s
+      Sort s <$ ((sortFitsIn s) =<< shouldBeSort t) -- sortFitsIn ensures @s /= Inf@
     Level l    -> do
       l <- checkLevel action l
       Level l <$ ((`subtype` t) =<< levelType)
     DontCare v -> DontCare <$> checkInternal' action v t
-    Shared{}   -> __IMPOSSIBLE__
 
 -- | Make sure a constructor is fully applied
 --   and infer the type of the constructor.
 --   Raises a type error if the constructor does not belong to the given type.
 fullyApplyCon
   :: ConHead -- ^ Constructor.
-  -> Args    -- ^ Constructor arguments.
+  -> Elims    -- ^ Constructor arguments.
   -> Type    -- ^ Type of the constructor application.
-  -> (QName -> Type -> Args -> Type -> Args -> Telescope -> Type -> TCM a)
+  -> (QName -> Type -> Args -> Type -> Elims -> Telescope -> Type -> TCM a)
        -- ^ Name of the data/record type,
        --   type of the data/record type,
        --   reconstructed parameters,
@@ -222,7 +221,7 @@ fullyApplyCon c vs t0 ret = do
       Nothing ->
         typeError $ DoesNotConstructAnElementOf (conName c) t
       Just ((d, dt, pars), a) ->
-        ret d dt pars a (raise (size tel) vs ++ teleArgs tel) tel t
+        ret d dt pars a (raise (size tel) vs ++ map Apply (teleArgs tel)) tel t
 
 checkSpine
   :: Action
@@ -268,22 +267,15 @@ checkHiding    h h' = unless (sameHiding h h') $ typeError $ HidingMismatch h h'
 -- | @checkRelevance action term type@.
 --
 --   The @term@ 'Relevance' can be updated by the @action@.
---   Note that the relevances might not match precisedly,
---   because of the non-semantic 'Forced' relevance.
 checkRelevance :: Action -> Relevance -> Relevance -> TCM Relevance
-checkRelevance action r0 r0' = do
+checkRelevance action r r' = do
   unless (r == r') $ typeError $ RelevanceMismatch r r'
-  return $ relevanceAction action r0' r0  -- Argument order for actions: @type@ @term@
-  where
-    r  = canon r0
-    r' = canon r0'
-    canon Forced{}  = Relevant
-    canon r         = r
+  return $ relevanceAction action r' r  -- Argument order for actions: @type@ @term@
 
 -- | Infer type of a neutral term.
 infer :: Term -> TCM Type
 infer v = do
-  case ignoreSharing v of
+  case v of
     Var i es   -> do
       a <- typeOfBV i
       snd <$> inferSpine a (Var i   []) es
@@ -292,7 +284,6 @@ infer v = do
     MetaV x es -> do -- we assume meta instantiations to be well-typed
       a <- metaType x
       snd <$> inferSpine a (MetaV x []) es
-    Shared{} -> __IMPOSSIBLE__
     _ -> __IMPOSSIBLE__
 
 -- | Infer ordinary function application.
@@ -363,7 +354,7 @@ shouldBeSort t = ifIsSort t return (typeError $ ShouldBeASort t)
 ifIsSort :: Type -> (Sort -> TCM a) -> TCM a -> TCM a
 ifIsSort t yes no = do
   t <- reduce t
-  case ignoreSharing $ unEl t of
+  case unEl t of
     Sort s -> yes s
     _      -> no
 
@@ -374,13 +365,21 @@ checkSort action s =
     Type l   -> Type <$> checkLevel action l
     Prop     -> __IMPOSSIBLE__
       -- the dummy Prop should not be part of a term we check
-    Inf      -> typeError $ SetOmegaNotValidType
-      -- we cannot have Setω on the lhs of the colon
-    SizeUniv -> typeError $ InvalidTypeSort s
-    DLub a b -> do
+    Inf      -> return Inf
+    SizeUniv -> return SizeUniv
+    PiSort a b -> do
       a <- checkSort action a
       addContext (absName b, defaultDom (sort a) :: Dom Type) $ do
-        DLub a . Abs (absName b) <$> checkSort action (absBody b)
+        PiSort a . Abs (absName b) <$> checkSort action (absBody b)
+    UnivSort s -> UnivSort <$> checkSort action s
+    MetaS x es -> do -- we assume sort meta instantiations to be well-formed
+      a <- metaType x
+      let self = Sort $ MetaS x []
+      ((_,v),_) <- inferSpine' action a self self es
+      case v of
+        Sort s     -> return s
+        MetaV x es -> return $ MetaS x es
+        _          -> __IMPOSSIBLE__
 
 -- | Check if level is well-formed.
 checkLevel :: Action -> Level -> TCM Level
@@ -413,7 +412,7 @@ subtype t1 t2 = do
 -- | Compute the sort of a type.
 
 inferSort :: Term -> TCM Sort
-inferSort t = case ignoreSharing t of
+inferSort t = case t of
     Var i es   -> do
       a <- typeOfBV i
       (_, s) <- eliminate (Var i []) a es
@@ -426,14 +425,13 @@ inferSort t = case ignoreSharing t of
       a <- metaType x
       (_, s) <- eliminate (MetaV x []) a es
       shouldBeSort s
-    Pi a b     -> return $ dLub (getSort a) (getSort <$> b)
-    Sort s     -> return $ sSuc s
+    Pi a b     -> inferPiSort (getSort a) (getSort <$> b)
+    Sort s     -> inferUnivSort s
     Con{}      -> __IMPOSSIBLE__
     Lit{}      -> __IMPOSSIBLE__
     Lam{}      -> __IMPOSSIBLE__
     Level{}    -> __IMPOSSIBLE__
     DontCare{} -> __IMPOSSIBLE__
-    Shared{}   -> __IMPOSSIBLE__
 
 -- | @eliminate t self es@ eliminates value @self@ of type @t@ by spine @es@
 --   and returns the remaining value and its type.
