@@ -13,12 +13,12 @@ import Prelude hiding (null)
 
 import qualified Data.Map as Map
 import Data.Map (Map)
-import Data.Semigroup (Semigroup, Monoid, (<>), mempty, mappend, Any(..))
-import Data.Foldable (Foldable, foldMap)
-import Data.Traversable (Traversable, traverse)
+import Data.Semigroup hiding (Arg(..))
+import Data.Foldable hiding (null)
+import Data.Traversable
+import Data.Maybe
 
 import Data.Data (Data)
-import Data.Typeable (Typeable)
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
@@ -33,7 +33,7 @@ import Agda.Utils.Pretty hiding ((<>))
 import Agda.Utils.Impossible
 
 data WithArity c = WithArity { arity :: Int, content :: c }
-  deriving (Typeable, Data, Functor, Foldable, Traversable, Show)
+  deriving (Data, Functor, Foldable, Traversable, Show)
 
 -- | Branches in a case tree.
 
@@ -44,12 +44,18 @@ data Case c = Branches
   , conBranches    :: Map QName (WithArity c)
     -- ^ Map from constructor (or projection) names to their arity
     --   and the case subtree.  (Projections have arity 0.)
+  , etaBranch      :: Maybe (ConHead, WithArity c)
+    -- ^ Eta-expand with the given (eta record) constructor. If this is
+    --   present, there should not be any conBranches or litBranches.
   , litBranches    :: Map Literal c
     -- ^ Map from literal to case subtree.
   , catchAllBranch :: Maybe c
     -- ^ (Possibly additional) catch-all clause.
+  , lazyMatch :: Bool
+    -- ^ Lazy pattern match. Requires single (non-copattern) branch with no lit
+    --   branches and no catch-all.
   }
-  deriving (Typeable, Data, Functor, Foldable, Traversable, Show)
+  deriving (Data, Functor, Foldable, Traversable, Show)
 
 -- | Case tree with bodies.
 
@@ -66,21 +72,35 @@ data CompiledClauses' a
     --   still reduce.
   | Fail
     -- ^ Absurd case.
-  deriving (Typeable, Data, Functor, Traversable, Foldable, Show)
+  deriving (Data, Functor, Traversable, Foldable, Show)
 
 type CompiledClauses = CompiledClauses' Term
 
 litCase :: Literal -> c -> Case c
-litCase l x = Branches False Map.empty (Map.singleton l x) Nothing
+litCase l x = Branches False Map.empty Nothing (Map.singleton l x) Nothing False
 
 conCase :: QName -> WithArity c -> Case c
-conCase c x = Branches False (Map.singleton c x) Map.empty Nothing
+conCase c x = Branches False (Map.singleton c x) Nothing Map.empty Nothing False
+
+etaCase :: ConHead -> WithArity c -> Case c
+etaCase c x = Branches False Map.empty (Just (c, x)) Map.empty Nothing True
 
 projCase :: QName -> c -> Case c
-projCase c x = Branches True (Map.singleton c $ WithArity 0 x) Map.empty Nothing
+projCase c x = Branches True (Map.singleton c $ WithArity 0 x) Nothing Map.empty Nothing False
 
 catchAll :: c -> Case c
-catchAll x = Branches False Map.empty Map.empty (Just x)
+catchAll x = Branches False Map.empty Nothing Map.empty (Just x) False
+
+-- | Check that the requirements on lazy matching (single inductive case) are
+--   met, and set lazy to False otherwise.
+checkLazyMatch :: Case c -> Case c
+checkLazyMatch b = b { lazyMatch = lazyMatch b && requirements }
+  where
+    requirements = and
+      [ null (catchAllBranch b)
+      , Map.size (conBranches b) <= 1
+      , null (litBranches b)
+      , not $ projPatterns b ]
 
 -- | Check whether a case tree has a catch-all clause.
 hasCatchAll :: CompiledClauses -> Bool
@@ -91,6 +111,15 @@ hasCatchAll = getAny . loop
     Done{}    -> mempty
     Case _ br -> maybe (foldMap loop br) (const $ Any True) $ catchAllBranch br
 
+-- | Check whether a case tree has any projection patterns
+hasProjectionPatterns :: CompiledClauses -> Bool
+hasProjectionPatterns = getAny . loop
+  where
+  loop cc = case cc of
+    Fail{}    -> mempty
+    Done{}    -> mempty
+    Case _ br -> Any (projPatterns br) <> foldMap loop br
+
 instance Semigroup c => Semigroup (WithArity c) where
   WithArity n1 c1 <> WithArity n2 c2
     | n1 == n2  = WithArity n1 (c1 <> c2)
@@ -101,19 +130,24 @@ instance (Semigroup c, Monoid c) => Monoid (WithArity c) where
   mappend = (<>)
 
 instance Semigroup m => Semigroup (Case m) where
-  Branches cop  cs  ls  m <> Branches cop' cs' ls' m' =
+  Branches cop cs eta ls m lazy <> Branches cop' cs' eta' ls' m' lazy' = checkLazyMatch $
     Branches (cop || cop') -- for @projCase <> mempty@
              (Map.unionWith (<>) cs cs')
+             (unionEta eta eta')
              (Map.unionWith (<>) ls ls')
              (m <> m')
+             (lazy && lazy')
+    where unionEta Nothing b = b
+          unionEta b Nothing = b
+          unionEta Just{} Just{} = __IMPOSSIBLE__
 
 instance (Semigroup m, Monoid m) => Monoid (Case m) where
   mempty  = empty
   mappend = (<>)
 
 instance Null (Case m) where
-  empty = Branches False Map.empty Map.empty Nothing
-  null (Branches _cop cs ls mcatch) = null cs && null ls && null mcatch
+  empty = Branches False Map.empty Nothing Map.empty Nothing True
+  null (Branches _cop cs eta ls mcatch _lazy) = null cs && null eta && null ls && null mcatch
 
 -- * Pretty instances.
 
@@ -121,12 +155,15 @@ instance Pretty a => Pretty (WithArity a) where
   pretty = pretty . content
 
 instance Pretty a => Pretty (Case a) where
-  prettyPrec p (Branches _cop cs ls m) =
-    mparens (p > 0) $ vcat $
-      prettyMap cs ++ prettyMap ls ++ prC m
+  prettyPrec p (Branches _cop cs eta ls m lazy) =
+    mparens (p > 0) $ prLazy lazy <+> vcat (prettyMap cs ++ prEta eta ++ prettyMap ls ++ prC m)
     where
+      prLazy True  = text "~"
+      prLazy False = empty
       prC Nothing = []
       prC (Just x) = [text "_ ->" <+> pretty x]
+      prEta Nothing = []
+      prEta (Just (c, cc)) = [(text "eta" <+> pretty c <+> text "->") <?> pretty cc]
 
 prettyMap :: (Pretty k, Pretty v) => Map k v -> [Doc]
 prettyMap m = [ sep [ pretty k <+> text "->"
@@ -134,16 +171,14 @@ prettyMap m = [ sep [ pretty k <+> text "->"
               | (k, v) <- Map.toList m ]
 
 instance Pretty CompiledClauses where
-  pretty (Done hs t) = text ("done" ++ show hs) <+> pretty t
+  pretty (Done hs t) = text ("done" ++ show hs) <?> pretty t
   pretty Fail        = text "fail"
   pretty (Case n bs) | projPatterns bs =
     sep [ text "record"
         , nest 2 $ pretty bs
         ]
   pretty (Case n bs) =
-    sep [ text ("case " ++ prettyShow n ++ " of")
-        , nest 2 $ pretty bs
-        ]
+    text ("case " ++ prettyShow n ++ " of") <?> pretty bs
 
 -- * KillRange instances.
 
@@ -151,10 +186,12 @@ instance KillRange c => KillRange (WithArity c) where
   killRange = fmap killRange
 
 instance KillRange c => KillRange (Case c) where
-  killRange (Branches cop con lit all) = Branches cop
+  killRange (Branches cop con eta lit all lazy) = Branches cop
     (killRangeMap con)
+    (killRange eta)
     (killRangeMap lit)
     (killRange all)
+    lazy
 
 instance KillRange CompiledClauses where
   killRange (Case i br) = killRange2 Case i br
