@@ -4,13 +4,17 @@ module Agda.TypeChecking.CompiledClause.Compile where
 
 import Prelude hiding (null)
 
+import Control.Applicative
+import Control.Arrow (first, second)
 import Control.Monad
 
 import Data.Maybe
 import Data.Monoid
 import qualified Data.Map as Map
-import Data.List (nubBy)
+import Data.List (nubBy, findIndex)
 import Data.Function
+import qualified Data.IntSet as IntSet
+import Data.Traversable (traverse)
 
 import Debug.Trace
 
@@ -20,21 +24,43 @@ import Agda.Syntax.Internal.Pattern
 import Agda.TypeChecking.CompiledClause
 import Agda.TypeChecking.Coverage
 import Agda.TypeChecking.Coverage.SplitTree
+import Agda.TypeChecking.Forcing
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.RecordPatterns
 import Agda.TypeChecking.Substitute
-import Agda.TypeChecking.Pretty (prettyTCM, nest, sep, text)
+import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Free
+import Agda.TypeChecking.Free.Precompute
+import Agda.TypeChecking.Reduce
 
 import Agda.Utils.Functor
 import Agda.Utils.Maybe
 import Agda.Utils.Null
 import Agda.Utils.List
-import Agda.Utils.Pretty (Pretty(..), prettyShow)
 import qualified Agda.Utils.Pretty as P
 
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+data RunRecordPatternTranslation = RunRecordPatternTranslation | DontRunRecordPatternTranslation
+  deriving (Eq)
+
+compileClauses' :: RunRecordPatternTranslation -> [Clause] -> TCM CompiledClauses
+compileClauses' recpat cs = do
+  -- Apply forcing translation. This only shuffles the deBruijn variables
+  -- so doesn't affect the right hand side.
+  cs <- sequence [ forcingTranslation ps <&> \ qs -> c{ namedClausePats = qs }
+                 | c@Clause{ namedClausePats = ps } <- cs ]
+
+  -- Throw away the unreachable clauses (#2723).
+  let notUnreachable = (Just True /=) . clauseUnreachable
+  cs <- map unBruijn <$> normaliseProjP (filter notUnreachable cs)
+
+  let translate | recpat == RunRecordPatternTranslation = translateCompiledClauses
+                | otherwise                             = return
+
+  translate $ compile cs
+
 -- | Process function clauses into case tree.
 --   This involves:
 --   1. Coverage checking, generating a split tree.
@@ -48,34 +74,30 @@ compileClauses ::
 compileClauses mt cs = do
   -- Construct clauses with pattern variables bound in left-to-right order.
   -- Discard de Bruijn indices in patterns.
-  let unBruijn cs = [ Cl (map (fmap (fmap dbPatVarName . namedThing)) $ namedClausePats c)
-                         (compiledClauseBody c) | c <- cs ]
-  shared <- sharedFun
   case mt of
-    Nothing -> compile shared . unBruijn <$> normaliseProjP cs
+    Nothing -> compile . map unBruijn <$> normaliseProjP cs
     Just (q, t)  -> do
       splitTree <- coverageCheck q t cs
 
       -- The coverage checker might have added some clauses (#2288)!
       -- Throw away the unreachable clauses (#2723).
       let notUnreachable = (Just True /=) . clauseUnreachable
-      cs <- normaliseProjP =<< filter notUnreachable . defClauses <$> getConstInfo q
-      let cls = unBruijn cs
+      cs <- normaliseProjP =<< instantiateFull =<< filter notUnreachable . defClauses <$> getConstInfo q
+
+      let cls = map unBruijn cs
 
       reportSDoc "tc.cc" 30 $ sep $ do
         (text "clauses patterns  before compilation") : do
           map (prettyTCM . map unArg . clPats) cls
-      reportSDoc "tc.cc" 50 $ do
-        sep [ text "clauses before compilation"
-            , (nest 2 . text . show) cs
-            ]
-      let cc = compileWithSplitTree shared splitTree cls
+      reportSDoc "tc.cc" 50 $
+        text "clauses before compilation" <?> pretty cs
+      let cc = compileWithSplitTree splitTree cls
       reportSDoc "tc.cc" 12 $ sep
         [ text "compiled clauses (still containing record splits)"
         , nest 2 $ return $ P.pretty cc
         ]
       cc <- translateCompiledClauses cc
-      return cc
+      return (fmap precomputeFreeVars_ cc)
 
 -- | Stripped-down version of 'Agda.Syntax.Internal.Clause'
 --   used in clause compiler.
@@ -85,70 +107,94 @@ data Cl = Cl
   , clBody :: Maybe Term
   } deriving (Show)
 
-instance Pretty Cl where
-  pretty (Cl ps b) = P.prettyList ps P.<+> P.text "->" P.<+> maybe (P.text "_|_") pretty b
+instance P.Pretty Cl where
+  pretty (Cl ps b) = P.prettyList ps P.<+> P.text "->" P.<+> maybe (P.text "_|_") P.pretty b
 
 type Cls = [Cl]
 
-compileWithSplitTree :: (Term -> Term) -> SplitTree -> Cls -> CompiledClauses
-compileWithSplitTree shared t cs = case t of
+-- | Strip down a clause. Don't forget to apply the substitution to the dot
+--   patterns!
+unBruijn :: Clause -> Cl
+unBruijn c = Cl (applySubst sub $ (map . fmap) (fmap dbPatVarName . namedThing) $ namedClausePats c)
+                (applySubst sub $ clauseBody c)
+  where
+    sub = renamingR $ fromMaybe __IMPOSSIBLE__ (clausePerm c)
+
+compileWithSplitTree :: SplitTree -> Cls -> CompiledClauses
+compileWithSplitTree t cs = case t of
   SplitAt i ts -> Case i $ compiles ts $ splitOn (length ts == 1) (unArg i) cs
         -- if there is just one case, we force expansion of catch-alls
         -- this is needed to generate a sound tree on which we can
         -- collapse record pattern splits
-  SplittingDone n -> compile shared cs
+  SplittingDone n -> compile cs
     -- after end of split tree, continue with left-to-right strategy
 
   where
     compiles :: SplitTrees -> Case Cls -> Case CompiledClauses
     compiles ts br@Branches{ projPatterns = cop
                            , conBranches = cons
+                           , etaBranch   = Nothing
                            , litBranches = lits
                            , catchAllBranch = catchAll }
-      = Branches
-          { projPatterns   = cop
-          , conBranches    = updCons cons
-          , litBranches    = compile shared <$> lits
-          , catchAllBranch = compile shared <$> catchAll
+      = br{ conBranches    = updCons cons
+          , etaBranch      = Nothing
+          , litBranches    = updLits lits
+          , catchAllBranch = updCatchall catchAll
           }
       where
         updCons = Map.mapWithKey $ \ c cl ->
-         caseMaybe (lookup c ts) (compile shared) (compileWithSplitTree shared) <$> cl
+         caseMaybe (lookup (SplitCon c) ts) compile compileWithSplitTree <$> cl
          -- When the split tree is finished, we continue with @compile@.
+        updLits = Map.mapWithKey $ \ l cl ->
+          caseMaybe (lookup (SplitLit l) ts) compile compileWithSplitTree cl
+        updCatchall = fmap $ caseMaybe (lookup SplitCatchall ts) compile compileWithSplitTree
+    compiles _ Branches{etaBranch = Just{}} = __IMPOSSIBLE__  -- we haven't inserted eta matches yet
 
-compile :: (Term -> Term) -> Cls -> CompiledClauses
-compile shared cs = case nextSplit cs of
-  Just (isRecP, n)-> Case n $ fmap (compile shared) $ splitOn isRecP (unArg n) cs
+compile :: Cls -> CompiledClauses
+compile [] = Fail
+compile cs = case nextSplit cs of
+  Just (isRecP, n) -> Case n $ fmap compile $ splitOn isRecP (unArg n) cs
   Nothing -> case clBody c of
     -- It's possible to get more than one clause here due to
     -- catch-all expansion.
-    Just t  -> Done (map (fmap name) $ clPats c) (shared t)
+    Just t  -> Done (map (fmap name) $ clPats c) t
     Nothing -> Fail
   where
     -- If there are more than one clauses, take the first one.
     c = headWithDefault __IMPOSSIBLE__ cs
-    name (VarP x) = x
-    name (DotP _) = underscore
-    name AbsurdP{} = absurdPatternName
+    name (VarP _ x) = x
+    name (DotP _ _) = underscore
     name ConP{}  = __IMPOSSIBLE__
     name LitP{}  = __IMPOSSIBLE__
     name ProjP{} = __IMPOSSIBLE__
 
 -- | Get the index of the next argument we need to split on.
---   This the number of the first pattern that does a match in the first clause.
+--   This the number of the first pattern that does a (non-lazy) match in the first clause.
+--   Or the first lazy match where all clauses agree on the constructor, if there are no
+--   non-lazy matches.
 nextSplit :: Cls -> Maybe (Bool, Arg Int)
-nextSplit []            = __IMPOSSIBLE__
-nextSplit (Cl ps _ : _) = headMaybe $ catMaybes $
-  zipWith (\ (Arg ai p) n -> (, Arg ai n) <$> properSplit p) ps [0..]
+nextSplit []             = __IMPOSSIBLE__
+nextSplit (Cl ps _ : cs) = findSplit nonLazy ps <|> findSplit allAgree ps
+  where
+    nonLazy _ (ConP _ cpi _) = not $ conPLazy cpi
+    nonLazy _ _              = True
+
+    findSplit okPat ps = headMaybe (catMaybes $
+      zipWith (\ (Arg ai p) n -> (, Arg ai n) <$> properSplit p <* guard (okPat n p)) ps [0..])
+
+    allAgree i (ConP c _ _) = all ((== Just (conName c)) . getCon . map unArg . drop i . clPats) cs
+    allAgree _ _            = False
+
+    getCon (ConP c _ _ : _) = Just $ conName c
+    getCon _                = Nothing
 
 -- | Is is not a variable pattern?
 --   And if yes, is it a record pattern?
 properSplit :: Pattern' a -> Maybe Bool
-properSplit (ConP _ cpi _) = Just $ isJust $ conPRecord cpi
+properSplit (ConP _ cpi _) = Just (Just PatORec == conPRecord cpi)
 properSplit LitP{}  = Just False
 properSplit ProjP{} = Just False
 properSplit VarP{}  = Nothing
-properSplit AbsurdP{} = Nothing -- for purposes of compilation
 properSplit DotP{}  = Nothing
 
 -- | Is this a variable pattern?
@@ -157,7 +203,6 @@ properSplit DotP{}  = Nothing
 isVar :: Pattern' a -> Bool
 isVar VarP{}  = True
 isVar DotP{}  = True
-isVar AbsurdP{} = True
 isVar ConP{}  = False
 isVar LitP{}  = False
 isVar ProjP{} = False
@@ -172,12 +217,11 @@ splitOn single n cs = mconcat $ map (fmap (:[]) . splitC n) $
 splitC :: Int -> Cl -> Case Cl
 splitC n (Cl ps b) = caseMaybe mp fallback $ \case
   ProjP _ d   -> projCase d $ Cl (ps0 ++ ps1) b
-  ConP c _ qs -> conCase (conName c) $ WithArity (length qs) $
-                   Cl (ps0 ++ map (fmap namedThing) qs ++ ps1) b
+  ConP c i qs -> (conCase (conName c) $ WithArity (length qs) $
+                   Cl (ps0 ++ map (fmap namedThing) qs ++ ps1) b) { lazyMatch = conPLazy i }
   LitP l      -> litCase l $ Cl (ps0 ++ ps1) b
   VarP{}      -> fallback
   DotP{}      -> fallback
-  AbsurdP{}   -> fallback
   where
     (ps0, rest) = splitAt n ps
     mp          = unArg <$> headMaybe rest
@@ -284,12 +328,12 @@ expandCatchAlls single n cs =
     expand cl (qs, q) =
       case unArg q of
         ConP c mt qs' -> Cl (ps0 ++ [q $> ConP c mt conPArgs] ++ ps1)
-                            (substBody n' m (Con c ci conArgs) b)
+                            (substBody n' m (Con c ci (map Apply conArgs)) b)
           where
             ci       = fromConPatternInfo mt
             m        = length qs'
             -- replace all direct subpatterns of q by _
-            conPArgs = map (fmap ($> VarP "_")) qs'
+            conPArgs = map (fmap ($> varP "_")) qs'
             conArgs  = zipWith (\ q' i -> q' $> var i) qs' $ downFrom m
         LitP l -> Cl (ps0 ++ [q $> LitP l] ++ ps1) (substBody n' 0 (Lit l) b)
         _ -> __IMPOSSIBLE__
@@ -301,13 +345,7 @@ expandCatchAlls single n cs =
         -- The following pattern match cannot fail (by construction of @ps@).
         (ps0, _:ps1) = splitAt n ps
 
-        n' = countVars ps1
-        countVars = sum . map (count . unArg)
-        count VarP{}        = 1
-        count (ConP _ _ ps) = countVars $ map (fmap namedThing) ps
-        count DotP{}        = 1   -- dot patterns are treated as variables in the clauses
-        count (AbsurdP p)   = count p
-        count _             = 0
+        n' = countPatternVars ps1
 
 -- | Make sure (by eta-expansion) that clause has arity at least @n@
 --   where @n@ is also the length of the provided list.
@@ -320,8 +358,10 @@ ensureNPatterns n ais0 cl@(Cl ps b)
   ais  = drop k ais0
   -- m = Number of arguments to add
   m    = n - k
-  ps'  = for ais $ \ ai -> Arg ai $ VarP "_"
+  ps'  = for ais $ \ ai -> Arg ai $ varP "_"
   args = zipWith (\ i ai -> Arg ai $ var i) (downFrom m) ais
 
 substBody :: (Subst t a) => Int -> Int -> t -> a -> a
 substBody n m v = applySubst $ liftS n $ v :# raiseS m
+
+instance PrecomputeFreeVars a => PrecomputeFreeVars (CompiledClauses' a) where
