@@ -2,7 +2,6 @@
 
 module Agda.TypeChecking.CompiledClause.Match where
 
-import Control.Applicative
 import Control.Monad.Reader (asks)
 
 import qualified Data.List as List
@@ -69,8 +68,6 @@ match' ((c, es, patch) : stack) = do
 
   do
 
-    shared <- sharedFun
-
     case c of
 
       -- impossible case
@@ -89,11 +86,24 @@ match' ((c, es, patch) : stack) = do
           -- at least the first @n@ elims must be @Apply@s, so we can
           -- turn them into a subsitution
           toSubst        = parallelS . reverse . map (unArg . argFromElim . ignoreReduced)
-          -- Andreas, 2013-05-21 why introduce sharing only here,
-          -- and not in underapplied case also?
-          (es0, es1)     = splitAt n $ map (fmap $ fmap shared) es
+          (es0, es1)     = splitAt n es
           lam x t        = Lam (argInfo x) (Abs (unArg x) t)
 
+      -- splitting on an eta-record constructor
+      Case (Arg _ n) Branches{etaBranch = Just (c, cc), catchAllBranch = ca} ->
+        case splitAt n es of
+          (_, []) -> no (NotBlocked Underapplied) es
+          (es0, MaybeRed _ e@(Apply (Arg _ v0)) : es1) ->
+              let projs = [ MaybeRed NotReduced $ Apply $ defaultArg $ v0 `applyE` [Proj ProjSystem f] | f <- fs ]
+                  catchAllFrame stack = maybe stack (\c -> (c, es, patch) : stack) ca in
+              match' $ (content cc, es0 ++ projs ++ es1, patchEta) : catchAllFrame stack
+            where
+              fs = conFields c
+              patchEta es = patch (es0 ++ [e] ++ es1)
+                where (es0, es') = splitAt n es
+                      (_, es1)   = splitAt (length fs) es'
+          _ -> __IMPOSSIBLE__
+
       -- splitting on the @n@th elimination
       Case (Arg _ n) bs -> do
         case splitAt n es of
@@ -122,7 +132,7 @@ match' ((c, es, patch) : stack) = do
                   case Map.lookup (conName c) (conBranches bs) of
                     Nothing -> stack
                     Just cc -> ( content cc
-                               , es0 ++ map (MaybeRed NotReduced . Apply) vs ++ es1
+                               , es0 ++ map (MaybeRed NotReduced) vs ++ es1
                                , patchCon c ci (length vs)
                                ) : stack
                 -- If our argument is @Proj p@, we push @projFrame p@ onto the stack.
@@ -139,18 +149,18 @@ match' ((c, es, patch) : stack) = do
                 patchCon c ci m es = patch (es0 ++ [Con c ci vs <$ e] ++ es2)
                   where (es0, rest) = splitAt n es
                         (es1, es2)  = splitAt m rest
-                        vs          = map argFromElim es1
+                        vs          = es1
 
             -- Now do the matching on the @n@ths argument:
             id $
-             case fmap ignoreSharing <$> eb of
+             case eb of
               Blocked x _            -> no (Blocked x) es'
               NotBlocked _ (Apply (Arg info (MetaV x _))) -> no (Blocked x) es'
 
               -- In case of a literal, try also its constructor form
               NotBlocked _ (Apply (Arg info v@(Lit l))) -> performedSimplification $ do
                 cv <- constructorForm v
-                let cFrame stack = case ignoreSharing cv of
+                let cFrame stack = case cv of
                       Con c ci vs -> conFrame c ci vs stack
                       _        -> stack
                 match' $ litFrame l $ cFrame $ catchAllFrame stack
