@@ -3,7 +3,6 @@
 
 module Agda.TypeChecking.Conversion where
 
-import Control.Applicative
 import Control.Monad
 import Control.Monad.Reader
 import Control.Monad.State
@@ -28,6 +27,7 @@ import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Constraints
 import {-# SOURCE #-} Agda.TypeChecking.CheckInternal (infer)
 import Agda.TypeChecking.Errors
+import Agda.TypeChecking.Forcing (isForced, nextIsForced)
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.Datatypes (getConType, getFullyAppliedConType)
 import Agda.TypeChecking.Records
@@ -116,19 +116,13 @@ compareTerm cmp a u v = do
     , nest 2 $ prettyTCM u <+> prettyTCM cmp <+> prettyTCM v
     , nest 2 $ text ":" <+> prettyTCM a
     ]
-  -- Check pointer equality first.
-  let checkPointerEquality def | not $ null $ List.intersect (pointerChain u) (pointerChain v) = do
-        verboseS "profile.sharing" 10 $ tick "pointer equality"
-        return ()
-      checkPointerEquality def = def
-  checkPointerEquality $ do
-    -- Check syntactic equality. This actually saves us quite a bit of work.
-    ((u, v), equal) <- runReduceM $ SynEq.checkSyntacticEquality u v
+  -- Check syntactic equality. This actually saves us quite a bit of work.
+  ((u, v), equal) <- runReduceM $ SynEq.checkSyntacticEquality u v
   -- OLD CODE, traverses the *full* terms u v at each step, even if they
   -- are different somewhere.  Leads to infeasibility in issue 854.
   -- (u, v) <- instantiateFull (u, v)
   -- let equal = u == v
-    unifyPointers cmp u v $ if equal then verboseS "profile.sharing" 20 $ tick "equal terms" else do
+  unifyPointers cmp u v $ if equal then verboseS "profile.sharing" 20 $ tick "equal terms" else do
       verboseS "profile.sharing" 20 $ tick "unequal terms"
       reportSDoc "tc.conv.term" 15 $ sep
         [ text "compareTerm (not syntactically equal)"
@@ -144,13 +138,13 @@ compareTerm cmp a u v = do
           unlessSubtyping cont =
               if cmp == CmpEq then cont else do
                 -- Andreas, 2014-04-12 do not short cut if type is blocked.
-                ifBlockedType a (\ _ _ -> fallback) {-else-} $ \ a -> do
+                ifBlockedType a (\ _ _ -> fallback) {-else-} $ \ _ a -> do
                   -- do not short circuit size comparison!
                   caseMaybeM (isSizeType a) cont (\ _ -> fallback)
 
           dir = fromCmp cmp
           rid = flipCmp dir     -- The reverse direction.  Bad name, I know.
-      case (ignoreSharing u, ignoreSharing v) of
+      case (u, v) of
         (MetaV x us, MetaV y vs)
           | x /= y    -> unlessSubtyping $ solve1 `orelse` solve2 `orelse` compareTerm' cmp a u v
           | otherwise -> fallback
@@ -237,17 +231,16 @@ compareTerm' cmp a m n =
     reportSDoc "tc.conv.level" 60 $ nest 2 $ sep
       [ text "a'   =" <+> pretty a'
       , text "mlvl =" <+> pretty mlvl
-      , text $ "(Just (ignoreSharing $ unEl a') == mlvl) = " ++ show (Just (ignoreSharing $ unEl a') == mlvl)
+      , text $ "(Just (unEl a') == mlvl) = " ++ show (Just (unEl a') == mlvl)
       ]
     case s of
       Prop | proofIrr -> return ()
       _    | isSize   -> compareSizes cmp m n
-      _               -> case ignoreSharing $ unEl a' of
+      _               -> case unEl a' of
         a | Just a == mlvl -> do
           a <- levelView m
           b <- levelView n
           equalLevel a b
--- OLD:        Pi dom _  -> equalFun (dom, a') m n
         a@Pi{}    -> equalFun a m n
         Lam _ _   -> __IMPOSSIBLE__
         Def r es  -> do
@@ -259,17 +252,15 @@ compareTerm' cmp a m n =
               -- Andreas, 2010-10-11: allowing neutrals to be blocked things does not seem
               -- to change Agda's behavior
               --    isNeutral Blocked{}          = False
-                  isNeutral = isNeutral' . fmap ignoreSharing
-                  isMeta    = isMeta'    . fmap ignoreSharing
-                  isNeutral' (NotBlocked _ Con{}) = return False
+                  isNeutral (NotBlocked _ Con{}) = return False
               -- Andreas, 2013-09-18 / 2015-06-29: a Def by copatterns is
               -- not neutral if it is blocked (there can be missing projections
               -- to trigger a reduction.
-                  isNeutral' (NotBlocked r (Def q _)) = do    -- Andreas, 2014-12-06 optimize this using r !!
+                  isNeutral (NotBlocked r (Def q _)) = do    -- Andreas, 2014-12-06 optimize this using r !!
                     not <$> usesCopatterns q -- a def by copattern can reduce if projected
-                  isNeutral' _                   = return True
-                  isMeta' (NotBlocked _ MetaV{}) = True
-                  isMeta' _                      = False
+                  isNeutral _                   = return True
+                  isMeta (NotBlocked _ MetaV{}) = True
+                  isMeta _                      = False
 
               reportSDoc "tc.conv.term" 30 $ prettyTCM a <+> text "is eta record type"
               m <- reduceB m
@@ -295,17 +286,16 @@ compareTerm' cmp a m n =
                   -- No subtyping on record terms
                   c <- getRecordConstructor r
                   -- Record constructors are covariant (see test/succeed/CovariantConstructors).
-                  compareArgs (repeat $ polFromCmp cmp) (telePi_ tel $ sort Prop) (Con c ConOSystem []) m' n'
+                  compareArgs (repeat $ polFromCmp cmp) [] (telePi_ tel $ sort Prop) (Con c ConOSystem []) m' n'
 
             else compareAtom cmp a' m n
         _ -> compareAtom cmp a' m n
   where
     -- equality at function type (accounts for eta)
     equalFun :: Term -> Term -> Term -> TCM ()
-    equalFun (Shared p) m n = equalFun (derefPtr p) m n
     equalFun (Pi dom@(Dom info _) b) m n = do
         name <- freshName_ $ suggest (absName b) "x"
-        addContext' (name, dom) $ compareTerm cmp (absBody b) m' n'
+        addContext (name, dom) $ compareTerm cmp (absBody b) m' n'
       where
         (m',n') = raise 1 (m,n) `apply` [Arg info $ var 0]
     equalFun _ _ _ = __IMPOSSIBLE__
@@ -324,16 +314,6 @@ compareTel t1 t2 cmp tel1 tel2 =
     (ExtendTel dom1@(Dom i1 a1) tel1, ExtendTel dom2@(Dom i2 a2) tel2) -> do
       compareDom cmp dom1 dom2 tel1 tel2 bad bad $
         compareTel t1 t2 cmp (absBody tel1) (absBody tel2)
-
-{- OLD, before 2013-05-15
-          let checkDom = escapeContext 1 $ compareType cmp a1 a2
-              c = TelCmp t1 t2 cmp (absBody tel1) (absBody tel2)
-
-          addContext (name, dom1) $
-            if dependent
-            then guardConstraint c checkDom
-            else checkDom >> solveConstraint_ c
--}
   where
     -- Andreas, 2011-05-10 better report message about types
     bad = typeError $ UnequalTypes cmp t2 t1
@@ -362,7 +342,7 @@ etaInequal cmp t m n = do
           ]
         patternViolation
   -- if type is not blocked, then we would have tried eta already
-  flip (ifBlockedType t) (\ _ -> inequal) $ \ _ _ -> do
+  flip (ifBlockedType t) (\ _ _ -> inequal) $ \ _ _ -> do
     -- type is blocked
     case (m, n) of
       (Con{}, _) -> dontKnow
@@ -419,7 +399,7 @@ compareAtom cmp t m n =
 
     -- constructorForm changes literal to constructors
     -- only needed if the other side is not a literal
-    (mb'', nb'') <- case (ignoreSharing $ ignoreBlocking mb', ignoreSharing $ ignoreBlocking nb') of
+    (mb'', nb'') <- case (ignoreBlocking mb', ignoreBlocking nb') of
       (Lit _, Lit _) -> return (mb', nb')
       _ -> (,) <$> traverse constructorForm mb'
                <*> traverse constructorForm nb'
@@ -450,7 +430,7 @@ compareAtom cmp t m n =
         text "compareAtom" <+> fsep [ prettyTCM mb <+> prettyTCM cmp
                                     , prettyTCM nb
                                     , text ":" <+> prettyTCM t ]
-      case (ignoreSharing <$> mb, ignoreSharing <$> nb) of
+      case (mb, nb) of
         -- equate two metas x and y.  if y is the younger meta,
         -- try first y := x and then x := y
         (NotBlocked _ (MetaV x xArgs), NotBlocked _ (MetaV y yArgs))
@@ -465,33 +445,35 @@ compareAtom cmp t m n =
                     PrunedEverything -> return ()
                     PrunedNothing    -> postpone
                     PrunedSomething  -> postpone
-                    -- OLD CODE: if killedAll then return () else checkSyntacticEquality
                 -- not all relevant arguments are variables
                 Nothing -> checkSyntacticEquality -- Check syntactic equality on meta-variables
                                 -- (same as for blocked terms)
             | otherwise -> do
                 [p1, p2] <- mapM getMetaPriority [x,y]
-                -- instantiate later meta variables first
+                -- First try the one with the highest priority. If that doesn't
+                -- work, try the low priority one.
                 let (solve1, solve2)
-                      | (p1,x) > (p2,y) = (l,r)
-                      | otherwise       = (r,l)
-                      where l = assign dir x xArgs n
-                            r = assign rid y yArgs m
+                      | (p1, x) > (p2, y) = (l1, r2)
+                      | otherwise         = (r1, l2)
+                      where l1 = assign dir x xArgs n
+                            r1 = assign rid y yArgs m
+                            -- Careful: the first attempt might prune the low
+                            -- priority meta! (Issue #2978)
+                            l2 = ifM (isInstantiatedMeta x) (compareTermDir dir t m n) l1
+                            r2 = ifM (isInstantiatedMeta y) (compareTermDir rid t n m) r1
 
                     try m h = m `catchError_` \err -> case err of
                       PatternErr{} -> h
                       _            -> throwError err
 
-                -- First try the one with the highest priority. If that doesn't
-                -- work, try the low priority one.
                 try solve1 solve2
 
         -- one side a meta, the other an unblocked term
         (NotBlocked _ (MetaV x es), _) -> assign dir x es n
         (_, NotBlocked _ (MetaV x es)) -> assign rid x es m
         (Blocked{}, Blocked{})  -> checkSyntacticEquality
-        (Blocked{}, _)    -> useInjectivity cmp t m n
-        (_,Blocked{})     -> useInjectivity cmp t m n
+        (Blocked{}, _)  -> useInjectivity (fromCmp cmp) t m n   -- The blocked term goes first
+        (_, Blocked{})  -> useInjectivity (flipCmp $ fromCmp cmp) t n m
         _ -> do
           -- -- Andreas, 2013-10-20 put projection-like function
           -- -- into the spine, to make compareElims work.
@@ -502,7 +484,7 @@ compareAtom cmp t m n =
           -- Andreas, 2015-07-01, actually, don't put them into the spine.
           -- Polarity cannot be communicated properly if projection-like
           -- functions are post-fix.
-          case (ignoreSharing m, ignoreSharing n) of
+          case (m, n) of
             (Pi{}, Pi{}) -> equalFun m n
 
             (Sort s1, Sort Inf) -> return ()
@@ -512,14 +494,14 @@ compareAtom cmp t m n =
             (Var i es, Var i' es') | i == i' -> do
                 a <- typeOfBV i
                 -- Variables are invariant in their arguments
-                compareElims [] a (var i) es es'
+                compareElims [] [] a (var i) es es'
             (Def f [], Def f' []) | f == f' -> return ()
             (Def f es, Def f' es') | f == f' -> do
                 a <- computeElimHeadType f es es'
                 -- The polarity vector of projection-like functions
                 -- does not include the parameters.
                 pol <- getPolarity' cmp f
-                compareElims pol a (Def f []) es es'
+                compareElims pol [] a (Def f []) es es'
             (Def f es, Def f' es') ->
               unlessM (bothAbsurd f f') $ do
                 trySizeUniv cmp t m n f es f' es'
@@ -528,13 +510,14 @@ compareAtom cmp t m n =
                 | x == y -> do
                     -- Get the type of the constructor instantiated to the datatype parameters.
                     a' <- conType x t
+                    forcedArgs <- getForcedArgs $ conName x
                     -- Constructors are covariant in their arguments
                     -- (see test/succeed/CovariantConstructors).
-                    compareArgs (repeat $ polFromCmp cmp) a' (Con x ci []) xArgs yArgs
+                    compareElims (repeat $ polFromCmp cmp) forcedArgs a' (Con x ci []) xArgs yArgs
             _ -> etaInequal cmp t m n -- fixes issue 856 (unsound conversion error)
     where
         -- Andreas, 2013-05-15 due to new postponement strategy, type can now be blocked
-        conType c t = ifBlockedType t (\ _ _ -> patternViolation) $ \ t -> do
+        conType c t = ifBlockedType t (\ _ _ -> patternViolation) $ \ _ t -> do
           let impossible = do
                 reportSDoc "impossible" 10 $
                   text "expected data/record type, found " <+> prettyTCM t
@@ -547,7 +530,7 @@ compareAtom cmp t m n =
                 -- Thus, instead of crashing, just give up gracefully.
                 patternViolation
           maybe impossible (return . snd) =<< getFullyAppliedConType c t
-        equalFun t1 t2 = case (ignoreSharing t1, ignoreSharing t2) of
+        equalFun t1 t2 = case (t1, t2) of
           (Pi dom1 b1, Pi dom2 b2) -> do
             verboseBracket "tc.conv.fun" 15 "compare function types" $ do
               reportSDoc "tc.conv.fun" 20 $ nest 2 $ vcat
@@ -558,18 +541,6 @@ compareAtom cmp t m n =
             where
             errH = typeError $ UnequalHiding t1 t2
             errR = typeError $ UnequalRelevance cmp t1 t2
-
-{- OLD, before 2013-05-15
-                let checkDom = escapeContext 1 $ compareType cmp a2 a1
-                    conCoDom = TypeCmp cmp (absBody b1) (absBody b2)
-                -- We only need to require a1 == a2 if t2 is a dependent function type.
-                -- If it's non-dependent it doesn't matter what we add to the context.
-                name <- freshName_ (suggest b1 b2)
-                addContext (name, dom1) $
-                  if isBinderUsed b2 -- dependent function type?
-                  then guardConstraint conCoDom checkDom
-                  else checkDom >> solveConstraint_ conCoDom
--}
           _ -> __IMPOSSIBLE__
 
 -- | Check whether @a1 `cmp` a2@ and continue in context extended by @a1@.
@@ -585,9 +556,7 @@ compareDom :: Free c
   -> TCM ()
 compareDom cmp dom1@(Dom i1 a1) dom2@(Dom i2 a2) b1 b2 errH errR cont
   | not (sameHiding dom1 dom2) = errH
-  -- Andreas 2010-09-21 compare r1 and r2, but ignore forcing annotations!
-  | not $ compareRelevance cmp (ignoreForced $ getRelevance dom1)
-                               (ignoreForced $ getRelevance dom2) = errR
+  | not $ compareRelevance cmp (getRelevance dom1) (getRelevance dom2) = errR
   | otherwise = do
       let r = max (getRelevance dom1) (getRelevance dom2)
               -- take "most irrelevant"
@@ -599,7 +568,7 @@ compareDom cmp dom1@(Dom i1 a1) dom2@(Dom i2 a2) b1 b2 errH errR cont
         -- We only need to require a1 == a2 if b2 is dependent
         -- If it's non-dependent it doesn't matter what we add to the context.
       name <- freshName_ $ suggest b1 b2
-      addContext' (name, dom) $ cont
+      addContext (name, dom) $ cont
       stealConstraints pid
         -- Andreas, 2013-05-15 Now, comparison of codomains is not
         -- blocked any more by getting stuck on domains.
@@ -632,14 +601,14 @@ antiUnify pid a u v = do
   ((u, v), eq) <- runReduceM (SynEq.checkSyntacticEquality u v)
   if eq then return u else do
   (u, v) <- reduce (u, v)
-  case (ignoreSharing u, ignoreSharing v) of
+  case (u, v) of
     (Pi ua ub, Pi va vb) -> do
       wa0 <- antiUnifyType pid (unDom ua) (unDom va)
       let wa = wa0 <$ ua
       wb <- addContext wa $ antiUnifyType pid (unAbs ub) (unAbs vb)
       return $ Pi wa (wb <$ ub)
     (Lam i u, Lam _ v) ->
-      case ignoreSharing $ unEl a of
+      case unEl a of
         Pi a b -> Lam i . (<$ u) <$> addContext a (antiUnify pid (unAbs b) (unAbs u) (unAbs v))
         _      -> fallback
     (Var i us, Var j vs) | i == j -> maybeGiveUp $ do
@@ -653,7 +622,7 @@ antiUnify pid a u v = do
     -- thus, we would not see clearly if we used @getFullyAppliedConType@ instead.)
     (Con x ci us, Con y _ vs) | x == y -> maybeGiveUp $ do
       a <- maybe patternViolation (return . snd) =<< getConType x a
-      antiUnifyElims pid a (Con x ci []) (map Apply us) (map Apply vs)
+      antiUnifyElims pid a (Con x ci []) us vs
     (Def f us, Def g vs) | f == g, length us == length vs -> maybeGiveUp $ do
       a <- computeElimHeadType f us vs
       antiUnifyElims pid a (Def f []) us vs
@@ -676,7 +645,7 @@ antiUnifyElims pid a self (Proj o f : es1) (Proj _ g : es2) | f == g = do
     Just (_, self, a) -> antiUnifyElims pid a self es1 es2
     Nothing -> patternViolation -- can fail for projection like
 antiUnifyElims pid a self (Apply u : es1) (Apply v : es2) = do
-  case ignoreSharing $ unEl a of
+  case unEl a of
     Pi a b -> do
       w <- antiUnify pid (unDom a) (unArg u) (unArg v)
       antiUnifyElims pid (b `lazyAbsApp` w) (apply self [w <$ u]) es1 es2
@@ -685,8 +654,8 @@ antiUnifyElims _ _ _ _ _ = patternViolation -- trigger maybeGiveUp in antiUnify
 
 -- | @compareElims pols a v els1 els2@ performs type-directed equality on eliminator spines.
 --   @t@ is the type of the head @v@.
-compareElims :: [Polarity] -> Type -> Term -> [Elim] -> [Elim] -> TCM ()
-compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 els02) $ do
+compareElims :: [Polarity] -> [IsForced] -> Type -> Term -> [Elim] -> [Elim] -> TCM ()
+compareElims pols0 fors0 a v els01 els02 = catchConstraint (ElimCmp pols0 fors0 a v els01 els02) $ do
   let v1 = applyE v els01
       v2 = applyE v els02
       failure = typeError $ UnequalTerms CmpEq v1 v2 a
@@ -697,6 +666,7 @@ compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 el
      nest 2 $ vcat
       [ text "a     =" <+> prettyTCM a
       , text "pols0 (truncated to 10) =" <+> sep (map prettyTCM $ take 10 pols0)
+      , text "fors0 (truncated to 10) =" <+> sep (map prettyTCM $ take 10 fors0)
       , text "v     =" <+> prettyTCM v
       , text "els01 =" <+> prettyTCM els01
       , text "els02 =" <+> prettyTCM els02
@@ -724,8 +694,9 @@ compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 el
         , text ""
         ]
       let (pol, pols) = nextPolarity pols0
-      ifBlockedType a (\ m t -> patternViolation) $ \ a -> do
-        case ignoreSharing . unEl $ a of
+          (for, fors) = nextIsForced fors0
+      ifBlockedType a (\ m t -> patternViolation) $ \ _ a -> do
+        case unEl a of
           (Pi (Dom info b) codom) -> do
             mlvl <- tryMaybe primLevel
             let freeInCoDom (Abs _ c) = 0 `freeInIgnoringSorts` c
@@ -743,12 +714,13 @@ compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 el
 
             -- compare arg1 and arg2
             pid <- newProblem_ $ applyRelevanceToContext r $
-                case r of
-                  Forced{}   -> return ()
-                  r | irrelevant r ->
-                                compareIrrelevant b (unArg arg1) (unArg arg2)
-                  _          -> compareWithPol pol (flip compareTerm b)
-                                  (unArg arg1) (unArg arg2)
+                if isForced for then
+                  return ()
+                else if isIrrelevant r then
+                  compareIrrelevant b (unArg arg1) (unArg arg2)
+                else
+                  compareWithPol pol (flip compareTerm b)
+                    (unArg arg1) (unArg arg2)
             -- if comparison got stuck and function type is dependent, block arg
             solved <- isProblemSolved pid
             arg <- if dependent && not solved
@@ -759,35 +731,16 @@ compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 el
                     return arg
                    else return arg1
             -- continue, possibly with blocked instantiation
-            compareElims pols (codom `lazyAbsApp` unArg arg) (apply v [arg]) els1 els2
+            compareElims pols fors (codom `lazyAbsApp` unArg arg) (apply v [arg]) els1 els2
             -- any left over constraints of arg are associatd to the comparison
             stealConstraints pid
+            {- Stealing solves this issue:
 
-{- Stealing solves this issue:
-
-   Does not create enough blocked tc-problems,
-   see test/fail/DontPrune.
-   (There are remaining problems which do not show up as yellow.)
-   Need to find a way to associate pid also to result of compareElims.
--}
-
-{- OLD, before 2013-05-15
-
-            let checkArg = applyRelevanceToContext r $
-                               case r of
-                  Forced     -> return ()
-                  r | irrelevant r ->
-                                compareIrrelevant b (unArg arg1) (unArg arg2)
-                  _          -> compareWithPol pol (flip compareTerm b)
-                                  (unArg arg1) (unArg arg2)
-
-                theRest = ElimCmp pols (piApply a [arg1]) (apply v [arg1]) els1 els2
-
-            if dependent
-              then guardConstraint theRest checkArg
-              else checkArg >> solveConstraint_ theRest
--}
-
+               Does not create enough blocked tc-problems,
+               see test/fail/DontPrune.
+               (There are remaining problems which do not show up as yellow.)
+               Need to find a way to associate pid also to result of compareElims.
+            -}
           a -> do
             reportSDoc "impossible" 10 $
               text "unexpected type when comparing apply eliminations " <+> prettyTCM a
@@ -801,7 +754,7 @@ compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 el
     -- case: f == f' are projections
     (Proj o f : els1, Proj _ f' : els2)
       | f /= f'   -> typeError . GenericError . show =<< prettyTCM f <+> text "/=" <+> prettyTCM f'
-      | otherwise -> ifBlockedType a (\ m t -> patternViolation) $ \ a -> do
+      | otherwise -> ifBlockedType a (\ m t -> patternViolation) $ \ _ a -> do
         res <- projectTyped v a o f -- fails only if f is proj.like but parameters cannot be retrieved
         case res of
           Just (_, u, t) -> do
@@ -809,7 +762,7 @@ compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 el
             -- The arguments following the principal argument of a projection
             -- are invariant.  (At least as long as we have no explicit polarity
             -- annotations.)
-            compareElims [] t u els1 els2
+            compareElims [] [] t u els1 els2
           Nothing -> do
             reportSDoc "tc.conv.elims" 30 $ sep
               [ text $ "projection " ++ show f
@@ -840,7 +793,6 @@ compareIrrelevant t v w = do
     ]
   try v w $ try w v $ return ()
   where
-    try (Shared p) w fallback = try (derefPtr p) w fallback
     try (MetaV x es) w fallback = do
       mv <- lookupMeta x
       let rel  = getMetaRelevance mv
@@ -851,7 +803,7 @@ compareIrrelevant t v w = do
         [ nest 2 $ text $ "rel  = " ++ show rel
         , nest 2 $ text "inst =" <+> pretty inst
         ]
-      if not (irrelevant rel) || inst
+      if not (isIrrelevant rel) || inst
         then fallback
         -- Andreas, 2016-08-08, issue #2131:
         -- Mining for solutions for irrelevant metas is not definite.
@@ -872,9 +824,9 @@ polFromCmp CmpEq  = Invariant
 
 -- | Type-directed equality on argument lists
 --
-compareArgs :: [Polarity] -> Type -> Term -> Args -> Args -> TCM ()
-compareArgs pol a v args1 args2 =
-  compareElims pol a v (map Apply args1) (map Apply args2)
+compareArgs :: [Polarity] -> [IsForced] -> Type -> Term -> Args -> Args -> TCM ()
+compareArgs pol for a v args1 args2 =
+  compareElims pol for a v (map Apply args1) (map Apply args2)
 
 ---------------------------------------------------------------------------
 -- * Types
@@ -963,11 +915,12 @@ coerce v t1 t2 = blockTerm t2 $ do
   -- If n  > 0 and b2 is not blocked, it is safe to
   -- insert n many hidden args
   if n <= 0 then fallback else do
-    ifBlockedType b2 (\ _ _ -> fallback) $ \ _ -> do
+    ifBlockedType b2 (\ _ _ -> fallback) $ \ _ _ -> do
       (args, t1') <- implicitArgs n notVisible t1
-      coerceSize leqType (v `apply` args) t1' t2
+      let v' = v `apply` args
+      v' <$ coerceSize leqType v' t1' t2
   where
-    fallback = coerceSize leqType v t1 t2
+    fallback = v <$ coerceSize leqType v t1 t2
 
 -- | Account for situations like @k : (Size< j) <= (Size< k + 1)@
 --
@@ -979,7 +932,7 @@ coerce v t1 t2 = blockTerm t2 $ do
 --   For now, we do a cheap heuristics.
 --
 --   Precondition: types are reduced.
-coerceSize :: (Type -> Type -> TCM ()) -> Term -> Type -> Type -> TCM Term
+coerceSize :: (Type -> Type -> TCM ()) -> Term -> Type -> Type -> TCM ()
 coerceSize leqType v t1 t2 = workOnTypes $ do
     reportSDoc "tc.conv.coerce" 70 $
       text "coerceSize" <+> vcat
@@ -987,8 +940,8 @@ coerceSize leqType v t1 t2 = workOnTypes $ do
         , text "from type t1 =" <+> pretty t1
         , text "to type   t2 =" <+> pretty t2
         ]
-    let fallback = v <$ leqType t1 t2
-        done = caseMaybeM (isSizeType t1) fallback $ \ b1 -> return v
+    let fallback = leqType t1 t2
+        done = caseMaybeM (isSizeType t1) fallback $ \ _ -> return ()
     -- Andreas, 2015-07-22, Issue 1615:
     -- If t1 is a meta and t2 a type like Size< v2, we need to make sure we do not miss
     -- the constraint v < v2!
@@ -999,7 +952,7 @@ coerceSize leqType v t1 t2 = workOnTypes $ do
       mv <- sizeMaxView v
       if any (\case{ DOtherSize{} -> True; _ -> False }) mv then fallback else do
       -- Andreas, 2015-02-11 do not instantiate metas here (triggers issue 1203).
-      ifM (tryConversion $ dontAssignMetas $ leqType t1 t2) (return v) $ {- else -} do
+      unlessM (tryConversion $ dontAssignMetas $ leqType t1 t2) $ do
         -- A (most probably weaker) alternative is to just check syn.eq.
         -- ifM (snd <$> checkSyntacticEquality t1 t2) (return v) $ {- else -} do
         reportSDoc "tc.conv.coerce" 20 $ text "coercing to a size type"
@@ -1052,25 +1005,40 @@ leqSort s1 s2 = catchConstraint (SortCmp CmpLeq s1 s2) $ do
                         , prettyTCM s2 ]
         ]
   case (s1, s2) of
-
-      (_       , Inf     ) -> yes
-
-      (SizeUniv, _       ) -> equalSort s1 s2
-      (_       , SizeUniv) -> equalSort s1 s2
-
+      -- The most basic rule: @Set l =< Set l'@ iff @l =< l'@
       (Type a  , Type b  ) -> leqLevel a b
 
-      (Prop    , Prop    ) -> yes
-      (Prop    , Type _  ) -> yes
-      (Type _  , Prop    ) -> no
-
-      -- (SizeUniv, SizeUniv) -> yes
-      -- (SizeUniv, _       ) -> no
-      -- (_       , SizeUniv) -> no
+      -- Prop is currently not supported
+      (Prop    , _       ) -> __IMPOSSIBLE__
+      (_       , Prop    ) -> __IMPOSSIBLE__
 
+      -- Setω is the top sort
+      (_       , Inf     ) -> yes
       (Inf     , _       ) -> equalSort s1 s2
-      (DLub{}  , _       ) -> postpone
-      (_       , DLub{}  ) -> postpone
+
+      -- Set0 and SizeUniv are bottom sorts
+      -- note: this ceases to hold if we add @Prop =< Set0@
+      (_       , Type (Max [])) -> equalSort s1 s2
+      (_       , SizeUniv) -> equalSort s1 s2
+
+      -- If we compare @univSort s@ against @Set (lsuc l)@,
+      -- we can simplify the comparison.
+      (UnivSort s, Type b    )
+        | Just b' <- levelSucView b -> leqSort s (Type b')
+      (Type a    , UnivSort s)
+        | Just a' <- levelSucView a -> leqSort (Type a') s
+
+      -- SizeUniv is unrelated to any Set l
+      (SizeUniv, Type{}  ) -> no
+
+      -- PiSort, UnivSort and MetaS might reduce once we instantiate
+      -- more metas, so we postpone.
+      (PiSort{}, _       ) -> postpone
+      (_       , PiSort{}) -> postpone
+      (UnivSort{}, _     ) -> postpone
+      (_     , UnivSort{}) -> postpone
+      (MetaS{} , _       ) -> postpone
+      (_       , MetaS{} ) -> postpone
 
 leqLevel :: Level -> Level -> TCM ()
 leqLevel a b = liftTCM $ do
@@ -1246,7 +1214,7 @@ equalLevel' a b = do
         _ | as == bs -> ok
           | any isBlocked (as ++ bs) -> do
               lvl <- levelType
-              liftTCM $ useInjectivity CmpEq lvl (Level a) (Level b)
+              liftTCM $ addConstraint $ ValueCmp CmpEq lvl (Level a) (Level b)
 
         -- closed == closed
         ([ClosedLevel n], [ClosedLevel m])
@@ -1370,28 +1338,6 @@ equalSort s1 s2 = do
             yes      = return ()
             no       = unlessM typeInType $ typeError $ UnequalSorts s1 s2
 
-            -- Test whether a level is infinity.
-            isInf ClosedLevel{}   = no
-            isInf (Plus _ l) = case l of
-              MetaLevel x es -> assignE DirEq x es (Sort Inf) $ equalAtom topSort
-                -- Andreas, 2015-02-14
-                -- This seems to be a hack, as a level meta is instantiated
-                -- by a sort.
-              NeutralLevel _ v -> case ignoreSharing v of
-                Sort Inf -> yes
-                _        -> no
-              _ -> no
-
-            -- Equate a level with SizeUniv.
-            eqSizeUniv l0 = case l0 of
-              Plus 0 l -> case l of
-                MetaLevel x es -> assignE DirEq x es (Sort SizeUniv) $ equalAtom topSort
-                NeutralLevel _ v -> case ignoreSharing v of
-                  Sort SizeUniv -> yes
-                  _ -> no
-                _ -> no
-              _ -> no
-
         reportSDoc "tc.conv.sort" 30 $ sep
           [ text "equalSort"
           , vcat [ nest 2 $ fsep [ prettyTCM s1 <+> text "=="
@@ -1403,40 +1349,76 @@ equalSort s1 s2 = do
 
         case (s1, s2) of
 
-            (Type a  , Type b  ) -> equalLevel a b
+            -- before anything else, try syntactic equality
+            _ | s1 == s2              -> return ()
 
-            (SizeUniv, SizeUniv) -> yes
-            (SizeUniv, Type (Max as@(_:_))) -> mapM_ eqSizeUniv as
-            (Type (Max as@(_:_)), SizeUniv) -> mapM_ eqSizeUniv as
-            (SizeUniv, _       ) -> no
-            (_       , SizeUniv) -> no
+            -- one side is a meta sort: try to instantiate
+            -- In case both sides are meta sorts, instantiate the
+            -- bigger (i.e. more recent) one.
+            (MetaS x es , MetaS y es')
+              | x < y                  -> meta y es' s1
+              | otherwise              -> meta x es s2
+            (MetaS x es , _          ) -> meta x es s2
+            (_          , MetaS x es ) -> meta x es s1
 
-            (Prop    , Prop    ) -> yes
-            (Type _  , Prop    ) -> no
-            (Prop    , Type _  ) -> no
+            -- diagonal cases for rigid sorts
+            (Type a     , Type b     ) -> equalLevel a b
+            (SizeUniv   , SizeUniv   ) -> yes
+            (Prop       , Prop       ) -> yes
+            (Inf        , Inf        ) -> yes
 
-            (Inf     , Inf     )             -> yes
-            (Inf     , Type (Max as@(_:_)))  -> mapM_ isInf as
-            (Type (Max as@(_:_)), Inf)       -> mapM_ isInf as
-            -- Andreas, 2014-06-27:
-            -- @Type (Max [])@ (which is Set0) falls through to error.
-            (Inf     , _       )             -> no
-            (_       , Inf     )             -> no
+            -- @UnivSort s == Set (lsuc l)@ iff @s == Set l@
+            (Type a     , UnivSort s )
+              | Just a' <- levelSucView a -> equalSort (Type a') s
+            (UnivSort s , Type b     )
+              | Just b' <- levelSucView b -> equalSort s (Type b')
 
-            -- Andreas, 2014-06-27:  Why are there special cases for Set0?
-            -- Andreas, 2015-02-14:  Probably because s ⊔ s' = Set0
-            -- entailed that both s and s' are Set0.
-            -- This is no longer true if  SizeUniv ⊔ s = s
+            -- if @PiSort a b == Set0@, then @b == Set0@
+            -- we use this fact to solve metas in @b@,
+            -- hopefully allowing the @PiSort@ to reduce.
+            (Type (Max []) , PiSort a b   ) -> piSortEqualToSet0 a b
+            (PiSort a b    , Type (Max [])) -> piSortEqualToSet0 a b
 
-            -- (DLub s1 s2, s0@(Type (Max []))) -> do
-            --   equalSort s1 s0
-            --   underAbstraction_ s2 $ \s2 -> equalSort s2 s0
-            -- (s0@(Type (Max [])), DLub s1 s2) -> do
-            --   equalSort s0 s1
-            --   underAbstraction_ s2 $ \s2 -> equalSort s0 s2
+            -- @PiSort a b == SizeUniv@ iff @b == SizeUniv@
+            (SizeUniv   , PiSort a b ) ->
+              underAbstraction_ b $ equalSort SizeUniv
+            (PiSort a b , SizeUniv   ) ->
+              underAbstraction_ b $ equalSort SizeUniv
 
-            (DLub{}  , _       )             -> postpone
-            (_       , DLub{}  )             -> postpone
+            -- Prop and SizeUniv don't contain any universes,
+            -- so they cannot be a UnivSort
+            (Prop       , UnivSort s ) -> no
+            (UnivSort s , Prop       ) -> no
+            (SizeUniv   , UnivSort s ) -> no
+            (UnivSort s , SizeUniv   ) -> no
+
+            -- PiSort and UnivSort could compute later, so we postpone
+            (PiSort{}   , _          ) -> postpone
+            (_          , PiSort{}   ) -> postpone
+            (UnivSort{} , _          ) -> postpone
+            (_          , UnivSort{} ) -> postpone
+
+            -- any other combinations of sorts are not equal
+            (_          , _          ) -> no
+
+    where
+      -- perform assignment (MetaS x es) := s
+      meta x es s = do
+        reportSLn "tc.meta.sort" 30 $ "Assigning meta sort"
+        reportSDoc "tc.meta.sort" 50 $ text "meta" <+> sep [pretty x, prettyList $ map pretty es, pretty s]
+        assignE DirEq x es (Sort s) __IMPOSSIBLE__
+
+      -- equate @piSort a b@ to @Set0@
+      -- note: this assumes @piSort a b == Set0@ implies @b == Set0@,
+      -- which becomes invalid if we add Prop.
+      piSortEqualToSet0 a b = do
+        let set0 = Type $ Max []
+        underAbstraction_ b $ equalSort set0
+        -- we may have instantiated some metas, so @a@ could reduce
+        a <- reduce a
+        case funSort' a set0 of
+          Just s  -> equalSort s set0
+          Nothing -> addConstraint $ SortCmp CmpEq (funSort a set0) set0
 
 ---------------------------------------------------------------------------
 -- * Definitions
