@@ -21,10 +21,6 @@ import Prelude hiding (null)
 import Control.Monad
 import Control.Monad.Trans ( lift )
 
-#if !MIN_VERSION_base(4,8,0)
-import Control.Applicative hiding (empty)
-#endif
-
 import Data.Either (lefts)
 import qualified Data.List as List
 import Data.Monoid (Any(..))
@@ -41,14 +37,18 @@ import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
 
 import Agda.TypeChecking.Monad
+import Agda.TypeChecking.Monad.Builtin (litType)
 
+import Agda.TypeChecking.Rules.LHS (checkSortOfSplitVar)
 import Agda.TypeChecking.Rules.LHS.Problem (allFlexVars)
 import Agda.TypeChecking.Rules.LHS.Unify
 
 import Agda.TypeChecking.Coverage.Match
 import Agda.TypeChecking.Coverage.SplitTree
 
+import Agda.TypeChecking.Conversion (tryConversion, equalType)
 import Agda.TypeChecking.Datatypes (getConForm)
+import Agda.TypeChecking.Patterns.Internal (dotPatternsToPatterns)
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Reduce
@@ -83,10 +83,10 @@ import Agda.Utils.Impossible
 data SplitClause = SClause
   { scTel    :: Telescope
     -- ^ Type of variables in @scPats@.
-  , scPats   :: [NamedArg DeBruijnPattern]
+  , scPats   :: [NamedArg SplitPattern]
     -- ^ The patterns leading to the currently considered branch of
     --   the split tree.
-  , scSubst  :: PatternSubstitution
+  , scSubst  :: Substitution' SplitPattern
     -- ^ Substitution from 'scTel' to old context.
     --   Only needed directly after split on variable:
     --   * To update 'scTarget'
@@ -99,8 +99,8 @@ data SplitClause = SClause
     --   'splitResult', which does not split on a variable,
     --   should reset it to the identity 'idS', lest it be
     --   applied to 'scTarget' again, leading to Issue 1294.
-  , scModuleParameterSub :: ModuleParamDict
-    -- ^ We need to keep track of the module parameter substitutions for the
+  , scCheckpoints :: Map CheckpointId Substitution
+    -- ^ We need to keep track of the module parameter checkpoints for the
     -- clause for the purpose of inferring missing instance clauses.
   , scTarget :: Maybe (Arg Type)
     -- ^ The type of the rhs, living in context 'scTel'.
@@ -114,8 +114,8 @@ data SplitClause = SClause
 data Covering = Covering
   { covSplitArg     :: Arg Nat
      -- ^ De Bruijn level (counting dot patterns) of argument we split on.
-  , covSplitClauses :: [(QName, SplitClause)]
-      -- ^ Covering clauses, indexed by constructor these clauses share.
+  , covSplitClauses :: [(SplitTag, SplitClause)]
+      -- ^ Covering clauses, indexed by constructor/literal these clauses share.
   }
 
 -- | Project the split clauses out of a covering.
@@ -126,9 +126,9 @@ splitClauses (Covering _ qcs) = map snd qcs
 clauseToSplitClause :: Clause -> SplitClause
 clauseToSplitClause cl = SClause
   { scTel    = clauseTel  cl
-  , scPats   = namedClausePats cl
+  , scPats   = toSplitPatterns $ namedClausePats cl
   , scSubst  = idS  -- Andreas, 2014-07-15  TODO: Is this ok?
-  , scModuleParameterSub = Map.empty
+  , scCheckpoints = Map.empty -- #2996: not __IMPOSSIBLE__ for debug printing
   , scTarget = clauseType cl
   }
 
@@ -152,9 +152,12 @@ coverageCheck f t cs = do
       -- The initial module parameter substitutions need to be weakened by the
       -- number of arguments that aren't module parameters.
   fv           <- getDefFreeVars f
-  moduleParams <- raise (n - fv) <$> use stModuleParameters
+
+  -- TODO: does this make sense? Why are we weakening by n - fv?
+  checkpoints <- applySubst (raiseS (n - fv)) <$> view eCheckpoints
+
       -- construct the initial split clause
-  let sc = SClause gamma xs idS moduleParams $ Just $ defaultArg a
+  let sc = SClause gamma xs idS checkpoints $ Just $ defaultArg a
 
   reportSDoc "tc.cover.top" 10 $ do
     let prCl cl = addContext (clauseTel cl) $
@@ -238,24 +241,21 @@ cover f cs sc@(SClause tel ps _ _ target) = do
   reportSDoc "tc.cover.cover" 10 $ inTopContext $ vcat
     [ text "checking coverage of pattern:"
     , nest 2 $ text "tel  =" <+> prettyTCM tel
-    , nest 2 $ text "ps   =" <+> do addContext tel $ prettyTCMPatternList ps
+    , nest 2 $ text "ps   =" <+> do addContext tel $ prettyTCMPatternList $ fromSplitPatterns ps
+    ]
+  reportSDoc "tc.cover.cover" 60 $ vcat
+    [ nest 2 $ text "ps   =" <+> pretty ps
     ]
   cs' <- normaliseProjP cs
+  ps <- (traverse . traverse . traverse) dotPatternsToPatterns ps
   case match cs' ps of
-    Yes (i,(mps,ls0)) -> do
+    Yes (i,mps) -> do
       exact <- allM mps isTrivialPattern
       let noExactClause = if exact || clauseCatchall (cs !! i)
                           then Set.empty
                           else Set.singleton i
       reportSLn "tc.cover.cover" 10 $ "pattern covered by clause " ++ show i
-      -- Check if any earlier clauses could match with appropriate literals
-      let lsis = mapMaybe (\(j,c) -> (,j) <$> matchLits c ps) $ zip [0..i-1] cs
-      reportSLn "tc.cover.cover"  10 $ "literal matches: " ++ show lsis
-      -- Andreas, 2016-10-08, issue #2243 (#708)
-      -- If we have several literal matches with the same literals
-      -- only take the first matching clause of these.
-      let is = Map.elems $ Map.fromListWith min $ (ls0,i) : lsis
-      return $ CoverResult (SplittingDone (size tel)) (Set.fromList is) [] noExactClause
+      return $ CoverResult (SplittingDone (size tel)) (Set.singleton i) [] noExactClause
 
     No        ->  do
       reportSLn "tc.cover" 20 $ "pattern is not covered"
@@ -268,12 +268,15 @@ cover f cs sc@(SClause tel ps _ _ target) = do
           -- type error before getting to this point.
           inferMissingClause f sc
           return $ CoverResult (SplittingDone (size tel)) Set.empty [] Set.empty
-        _ -> return $ CoverResult (SplittingDone (size tel)) Set.empty [(tel, ps)] Set.empty
+        _ -> do
+          let ps' = fromSplitPatterns ps
+          return $ CoverResult (SplittingDone (size tel)) Set.empty [(tel, ps')] Set.empty
 
     -- We need to split!
     -- If all clauses have an unsplit copattern, we try that first.
-    Block res bs -> tryIf (getAny res) splitRes $ do
-      let done = return $ CoverResult (SplittingDone (size tel)) Set.empty [(tel, ps)] Set.empty
+    Block res bs -> tryIf res splitRes $ do
+      let ps' = fromSplitPatterns ps
+          done = return $ CoverResult (SplittingDone (size tel)) Set.empty [(tel, ps')] Set.empty
       if null bs then done else do
       -- Otherwise, if there are variables to split, we try them
       -- in the order determined by a split strategy.
@@ -281,9 +284,22 @@ cover f cs sc@(SClause tel ps _ _ target) = do
       -- xs is a non-empty lists of blocking variables
       -- try splitting on one of them
       xs <- splitStrategy bs tel
-      r <- altM1 (split Inductive sc) xs
+      -- Andreas, 2017-10-08, issue #2594
+      -- First, try to find split order for complete coverage.
+      -- If this fails, try to at least carry out the splitting to the end.
+      continue xs NoAllowPartialCover $ \ _err -> do
+        continue xs YesAllowPartialCover $ \ err -> do
+          typeError $ SplitError err
+  where
+    continue
+      :: [BlockingVar]
+      -> AllowPartialCover
+      -> (SplitError -> TCM CoverResult)
+      -> TCM CoverResult
+    continue xs allowPartialCover handle = do
+      r <- altM1 (split Inductive allowPartialCover sc) xs
       case r of
-        Left err -> typeError $ SplitError err
+        Left err -> handle err
         -- If we get the empty covering, we have reached an impossible case
         -- and are done.
         Right (Covering n []) ->
@@ -309,13 +325,12 @@ cover f cs sc@(SClause tel ps _ _ target) = do
               tree   = SplitAt n trees'
           return $ CoverResult tree (Set.unions useds) (concat psss) (Set.unions noex)
 
-  where
-    tryIf :: Monad m => Bool -> m (Maybe a) -> m a -> m a
-    tryIf True  me m = fromMaybeM m me
+    tryIf :: Monad m => Bool -> m (Either err a) -> m a -> m a
+    tryIf True  me m = fromRightM (const m) me
     tryIf False me m = m
 
     -- Try to split result
-    splitRes :: TCM (Maybe CoverResult)
+    splitRes :: TCM (Either CosplitError CoverResult)
     splitRes = do
       reportSLn "tc.cover" 20 $ "blocked by projection pattern"
       -- forM is a monadic map over a Maybe here
@@ -336,41 +351,38 @@ cover f cs sc@(SClause tel ps _ _ target) = do
         return $ CoverResult tree (Set.unions useds) (concat psss) (Set.unions noex)
 
     gatherEtaSplits :: Int -> SplitClause
-                    -> [NamedArg DeBruijnPattern] -> [NamedArg DeBruijnPattern]
+                    -> [NamedArg SplitPattern] -> [NamedArg SplitPattern]
     gatherEtaSplits n sc []
        | n >= 0    = __IMPOSSIBLE__ -- we should have encountered the main
                                     -- split by now already
        | otherwise = []
     gatherEtaSplits n sc (p:ps) = case namedArg p of
-      VarP x
+      VarP _ x
        | n == 0    -> case p' of -- this is the main split
-           VarP  _      -> __IMPOSSIBLE__
-           DotP  _      -> __IMPOSSIBLE__
-           AbsurdP _    -> __IMPOSSIBLE__
+           VarP  _ _    -> p : gatherEtaSplits (-1) sc ps
+           DotP  _ _    -> __IMPOSSIBLE__
            ConP  _ _ qs -> qs ++ gatherEtaSplits (-1) sc ps
-           LitP  _      -> __IMPOSSIBLE__
+           LitP  _      -> gatherEtaSplits (-1) sc ps
            ProjP{}      -> __IMPOSSIBLE__
        | otherwise ->
            updateNamedArg (\ _ -> p') p : gatherEtaSplits (n-1) sc ps
-        where p' = lookupS (scSubst sc) $ dbPatVarIndex x
-      DotP  _      -> p : gatherEtaSplits (n-1) sc ps -- count dot patterns
-      AbsurdP _    -> p : gatherEtaSplits (n-1) sc ps
+        where p' = lookupS (scSubst sc) $ splitPatVarIndex x
+      DotP  _ _    -> p : gatherEtaSplits (n-1) sc ps -- count dot patterns
       ConP  _ _ qs -> gatherEtaSplits n sc (qs ++ ps)
       LitP  _      -> gatherEtaSplits n sc ps
       ProjP{}      -> gatherEtaSplits n sc ps
 
-    addEtaSplits :: Int -> [NamedArg DeBruijnPattern] -> SplitTree -> SplitTree
+    addEtaSplits :: Int -> [NamedArg SplitPattern] -> SplitTree -> SplitTree
     addEtaSplits k []     t = t
     addEtaSplits k (p:ps) t = case namedArg p of
-      VarP  _       -> addEtaSplits (k+1) ps t
-      DotP  _       -> addEtaSplits (k+1) ps t
-      AbsurdP _     -> addEtaSplits (k+1) ps t
-      ConP c cpi qs -> SplitAt (p $> k) [(conName c , addEtaSplits k (qs ++ ps) t)]
+      VarP  _ _     -> addEtaSplits (k+1) ps t
+      DotP  _ _     -> addEtaSplits (k+1) ps t
+      ConP c cpi qs -> SplitAt (p $> k) [(SplitCon (conName c) , addEtaSplits k (qs ++ ps) t)]
       LitP  _       -> __IMPOSSIBLE__
       ProjP{}       -> __IMPOSSIBLE__
 
-    etaRecordSplits :: Int -> [NamedArg DeBruijnPattern] -> (QName,SplitClause)
-                    -> SplitTree -> (QName,SplitTree)
+    etaRecordSplits :: Int -> [NamedArg SplitPattern] -> (SplitTag,SplitClause)
+                    -> SplitTree -> (SplitTag,SplitTree)
     etaRecordSplits n ps (q , sc) t =
       (q , addEtaSplits 0 (gatherEtaSplits n sc ps) t)
 
@@ -381,9 +393,11 @@ inferMissingClause
   -> SplitClause
        -- ^ Clause to add.  Clause hiding (in 'clauseType') must be 'Instance'.
    -> TCM ()
-inferMissingClause f (SClause tel ps _ mpsub (Just t)) = setCurrentRange f $ do
+inferMissingClause f (SClause tel ps _ cps (Just t)) = setCurrentRange f $ do
   reportSDoc "tc.cover.infer" 20 $ addContext tel $ text "Trying to infer right-hand side of type" <+> prettyTCM t
-  cl <- addContext tel $ withModuleParameters mpsub $ do
+  cl <- addContext tel
+        $ locally eCheckpoints (const cps)
+        $ checkpoint IdS $ do    -- introduce a fresh checkpoint
     (_x, rhs) <- case getHiding t of
                   Instance{} -> newIFSMeta "" (unArg t)
                   Hidden     -> __IMPOSSIBLE__
@@ -391,7 +405,7 @@ inferMissingClause f (SClause tel ps _ mpsub (Just t)) = setCurrentRange f $ do
     return $ Clause { clauseLHSRange  = noRange
                     , clauseFullRange = noRange
                     , clauseTel       = tel
-                    , namedClausePats = ps
+                    , namedClausePats = fromSplitPatterns ps
                     , clauseBody      = Just rhs
                     , clauseType      = Just t
                     , clauseCatchall  = False
@@ -401,7 +415,7 @@ inferMissingClause f (SClause tel ps _ mpsub (Just t)) = setCurrentRange f $ do
 inferMissingClause _ (SClause _ _ _ _ Nothing) = __IMPOSSIBLE__
 
 splitStrategy :: BlockingVars -> Telescope -> TCM BlockingVars
-splitStrategy bs tel = return $ updateLast clearBlockingVarCons xs
+splitStrategy bs tel = return $ updateLast setBlockingVarOverlap xs
   -- Make sure we do not insists on precomputed coverage when
   -- we make our last try to split.
   -- Otherwise, we will not get a nice error message.
@@ -427,7 +441,7 @@ isDatatype ind at = do
   let t       = unDom at
       throw f = throwError . f =<< do liftTCM $ buildClosure t
   t' <- liftTCM $ reduce t
-  case ignoreSharing $ unEl t' of
+  case unEl t' of
     Def d es -> do
       let ~(Just args) = allApplyElims es
       def <- liftTCM $ theDef <$> getConstInfo d
@@ -454,12 +468,12 @@ isDatatype ind at = do
 --   if target becomes a function type.
 --   Returns the domains of the function type (if any).
 fixTarget :: SplitClause -> TCM (Telescope, SplitClause)
-fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scModuleParameterSub = mpsub, scTarget = target } =
+fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scCheckpoints = cps, scTarget = target } =
   caseMaybe target (return (empty, sc)) $ \ a -> do
     reportSDoc "tc.cover.target" 20 $ sep
       [ text "split clause telescope: " <+> prettyTCM sctel
       , text "old patterns          : " <+> do
-          addContext sctel $ prettyTCMPatternList ps
+          addContext sctel $ prettyTCMPatternList $ fromSplitPatterns ps
       ]
     reportSDoc "tc.cover.target" 60 $ sep
       [ text "substitution          : " <+> text (show sigma)
@@ -468,7 +482,7 @@ fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scModuleParam
       [ text "target type before substitution (variables may be wrong): " <+> do
           addContext sctel $ prettyTCM a
       ]
-    TelV tel b <- telView $ applyPatSubst sigma $ unArg a
+    TelV tel b <- telView $ applySplitPSubst sigma $ unArg a
     reportSDoc "tc.cover.target" 15 $ sep
       [ text "target type telescope (after substitution): " <+> do
           addContext sctel $ prettyTCM tel
@@ -489,7 +503,7 @@ fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scModuleParam
           , scPats   = ps'
           , scSubst  = wkS n $ sigma -- Should be wkS instead of liftS since
                                      -- variables are only added to new tel.
-          , scModuleParameterSub = applySubst (raiseS n) mpsub
+          , scCheckpoints        = applySubst (raiseS n) cps
           , scTarget = newTarget
           }
     -- Separate debug printing to find cause of crash (Issue 1374)
@@ -498,7 +512,7 @@ fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scModuleParam
       ]
     reportSDoc "tc.cover.target" 30 $ sep
       [ text "new split clause patterns    : " <+> do
-          addContext sctel' $ prettyTCMPatternList ps'
+          addContext sctel' $ prettyTCMPatternList $ fromSplitPatterns ps'
       ]
     reportSDoc "tc.cover.target" 60 $ sep
       [ text "new split clause substitution: " <+> text (show $ scSubst sc')
@@ -520,7 +534,7 @@ hiddenInserted ai
   | visible ai = setOrigin UserWritten ai
   | otherwise  = setOrigin Inserted ai
 
--- | @computeNeighbourhood delta1 delta2 d pars ixs hix tel ps mpsub con@
+-- | @computeNeighbourhood delta1 delta2 d pars ixs hix tel ps con@
 --
 --   @
 --      delta1   Telescope before split point
@@ -532,7 +546,7 @@ hiddenInserted ai
 --      hix      Index of split variable
 --      tel      Telescope for patterns ps
 --      ps       Patterns before doing the split
---      mpsub    Current module parameter substitutions
+--      cps      Current module parameter checkpoints
 --      con      Constructor to fit into hole
 --   @
 --   @dtype == d pars ixs@
@@ -545,11 +559,11 @@ computeNeighbourhood
   -> Args                         -- ^ Data type indices.
   -> Nat                          -- ^ Index of split variable.
   -> Telescope                    -- ^ Telescope for the patterns.
-  -> [NamedArg DeBruijnPattern]   -- ^ Patterns before doing the split.
-  -> ModuleParamDict              -- ^ Current module parameter substitution.
+  -> [NamedArg SplitPattern]      -- ^ Patterns before doing the split.
+  -> Map CheckpointId Substitution -- ^ Current checkpoints
   -> QName                        -- ^ Constructor to fit into hole.
   -> CoverM (Maybe SplitClause)   -- ^ New split clause if successful.
-computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
+computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps cps c = do
 
   -- Get the type of the datatype
   dtype <- liftTCM $ (`piApply` pars) . defType <$> getConstInfo d
@@ -565,7 +579,7 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
   -- Lookup the type of the constructor at the given parameters
   (gamma0, cixs) <- do
     TelV gamma0 (El _ d) <- liftTCM $ telView (ctype `piApply` pars)
-    let Def _ es = ignoreSharing d
+    let Def _ es = d
         Just cixs = allApplyElims es
     return (gamma0, cixs)
 
@@ -573,7 +587,6 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
   -- of constructor
 
   let preserve (x, t@(El _ (Def d' _))) | d == d' = (n, t)
-      preserve (x, (El s (Shared p))) = preserve (x, El s $ derefPtr p)
       preserve (x, t) = (x, t)
       gammal = map (fmap preserve) . telToList $ gamma0
       gamma  = telFromList gammal
@@ -605,19 +618,20 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
       debugSubst "rho0" rho0
 
       -- We have Δ₁' ⊢ ρ₀ : Δ₁Γ, so split it into the part for Δ₁ and the part for Γ
-      let (rho1,rho2) = splitS (size gamma) rho0
+      let (rho1,rho2) = splitS (size gamma) $ toSplitPSubst rho0
 
       -- Andreas, 2015-05-01  I guess it is fine to use no @conPType@
       -- as the result of splitting is never used further down the pipeline.
       -- After splitting, Agda reloads the file.
       -- Andreas, 2017-09-03, issue #2729: remember that pattern was generated by case split.
-      let conp    = ConP con (ConPatternInfo (Just ConOSplit) Nothing) $ applySubst rho2 $
-                      map (mapArgInfo hiddenInserted) $ tele2NamedArgs gamma0 gamma
+      let cpi  = noConPatternInfo{ conPRecord = Just PatOSplit }
+          conp = ConP con cpi $ applySubst rho2 $
+                   map (mapArgInfo hiddenInserted) $ tele2NamedArgs gamma0 gamma
           -- Andreas, 2016-09-08, issue #2166: use gamma0 for correct argument names
 
       -- Compute final context and substitution
       let rho3    = consS conp rho1            -- Δ₁' ⊢ ρ₃ : Δ₁(x:D)
-          delta2' = applyPatSubst rho3 delta2  -- Δ₂' = Δ₂ρ₃
+          delta2' = applySplitPSubst rho3 delta2  -- Δ₂' = Δ₂ρ₃
           delta'  = delta1' `abstract` delta2' -- Δ'  = Δ₁'Δ₂'
           rho     = liftS (size delta2) rho3   -- Δ' ⊢ ρ : Δ₁(x:D)Δ₂
 
@@ -629,9 +643,9 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
       let ps' = applySubst rho ps
       debugPlugged delta' ps'
 
-      let mpsub' = applySubst (fromPatternSubstitution rho) mpsub
+      let cps' = applySplitPSubst rho cps
 
-      return $ Just $ SClause delta' ps' rho mpsub' Nothing -- target fixed later
+      return $ Just $ SClause delta' ps' rho cps' Nothing -- target fixed later
 
   where
     debugInit con ctype d pars ixs cixs delta1 delta2 gamma tel ps hix =
@@ -641,7 +655,7 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
           [ text "context=" <+> (inTopContext . prettyTCM =<< getContextTelescope)
           , text "con    =" <+> prettyTCM con
           , text "ctype  =" <+> prettyTCM ctype
-          , text "ps     =" <+> do inTopContext $ addContext tel $ prettyTCMPatternList ps
+          , text "ps     =" <+> do inTopContext $ addContext tel $ prettyTCMPatternList $ fromSplitPatterns ps
           , text "d      =" <+> prettyTCM d
           , text "pars   =" <+> do prettyList $ map prettyTCM pars
           , text "ixs    =" <+> do addContext delta1 $ prettyList $ map prettyTCM ixs
@@ -672,18 +686,30 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
     debugPs tel ps =
       liftTCM $ reportSDoc "tc.cover.split.con" 20 $
         inTopContext $ addContext tel $ nest 2 $ vcat
-          [ text "ps     =" <+> prettyTCMPatternList ps
+          [ text "ps     =" <+> prettyTCMPatternList (fromSplitPatterns ps)
           ]
 
     debugPlugged delta' ps' =
       liftTCM $ reportSDoc "tc.cover.split.con" 20 $
         inTopContext $ addContext delta' $ nest 2 $ vcat
-          [ text "ps'    =" <+> do prettyTCMPatternList ps'
+          [ text "ps'    =" <+> do prettyTCMPatternList $ fromSplitPatterns ps'
           ]
 
+-- | Introduce trailing pattern variables via 'fixTarget'?
+data FixTarget
+  = YesFixTarget
+  | NoFixTarget
+  deriving (Show)
+
+-- | Allow partial covering for split?
+data AllowPartialCover
+  = YesAllowPartialCover  -- To try to coverage-check incomplete splits.
+  | NoAllowPartialCover   -- Default.
+  deriving (Eq, Show)
+
 -- | Entry point from @Interaction.MakeCase@.
 splitClauseWithAbsurd :: SplitClause -> Nat -> TCM (Either SplitError (Either SplitClause Covering))
-splitClauseWithAbsurd c x = split' Inductive False c (BlockingVar x Nothing)
+splitClauseWithAbsurd c x = split' Inductive NoAllowPartialCover NoFixTarget c (BlockingVar x [] [] True)
   -- Andreas, 2016-05-03, issue 1950:
   -- Do not introduce trailing pattern vars after split,
   -- because this does not work for with-clauses.
@@ -692,8 +718,8 @@ splitClauseWithAbsurd c x = split' Inductive False c (BlockingVar x Nothing)
 --   @splitLast CoInductive@ is used in the @refine@ tactics.
 
 splitLast :: Induction -> Telescope -> [NamedArg DeBruijnPattern] -> TCM (Either SplitError Covering)
-splitLast ind tel ps = split ind sc (BlockingVar 0 Nothing)
-  where sc = SClause tel ps empty empty Nothing
+splitLast ind tel ps = split ind NoAllowPartialCover sc (BlockingVar 0 [] [] True)
+  where sc = SClause tel (toSplitPatterns ps) empty empty Nothing
 
 -- | @split ind splitClause x = return res@
 --   splits @splitClause@ at pattern var @x@ (de Bruijn index).
@@ -710,10 +736,13 @@ splitLast ind tel ps = split ind sc (BlockingVar 0 Nothing)
 split :: Induction
          -- ^ Coinductive constructors are allowed if this argument is
          -- 'CoInductive'.
+      -> AllowPartialCover
+         -- ^ Don't fail if computed 'Covering' does not cover all constructors.
       -> SplitClause
       -> BlockingVar
       -> TCM (Either SplitError Covering)
-split ind sc x = fmap blendInAbsurdClause <$> split' ind True sc x
+split ind allowPartialCover sc x =
+  fmap blendInAbsurdClause <$> split' ind allowPartialCover YesFixTarget sc x
   where
     n = lookupPatternVar sc $ blockingVarNo x
     blendInAbsurdClause :: Either SplitClause Covering -> Covering
@@ -728,7 +757,7 @@ lookupPatternVar SClause{ scTel = tel, scPats = pats } x = arg $>
   where n = if k < 0
             then __IMPOSSIBLE__
             else fromMaybe __IMPOSSIBLE__ $ permPicks perm !!! k
-        perm = fromMaybe __IMPOSSIBLE__ $ dbPatPerm pats
+        perm = fromMaybe __IMPOSSIBLE__ $ dbPatPerm $ fromSplitPatterns pats
         k = size tel - x - 1
         arg = telVars (size tel) tel !! k
 
@@ -749,15 +778,18 @@ lookupPatternVar SClause{ scTel = tel, scPats = pats } x = arg $>
 split' :: Induction
           -- ^ Coinductive constructors are allowed if this argument is
           -- 'CoInductive'.
-       -> Bool
-          -- ^ If 'True', introduce new trailing variable patterns via
+       -> AllowPartialCover
+          -- ^ Don't fail if computed 'Covering' does not cover all constructors.
+       -> FixTarget
+          -- ^ If 'YesFixTarget', introduce new trailing variable patterns via
           --   'fixTarget'.
        -> SplitClause
        -> BlockingVar
        -> TCM (Either SplitError (Either SplitClause Covering))
-split' ind fixtarget sc@(SClause tel ps _ mpsub target) (BlockingVar x mcons) = liftTCM $ runExceptT $ do
+split' ind allowPartialCover fixtarget sc@(SClause tel ps _ cps target) (BlockingVar x pcons' plits overlap) =
+ liftTCM $ runExceptT $ do
 
-  debugInit tel x ps mpsub
+  debugInit tel x ps cps
 
   -- Split the telescope at the variable
   -- t = type of the variable,  Δ₁ ⊢ t
@@ -765,34 +797,60 @@ split' ind fixtarget sc@(SClause tel ps _ mpsub target) (BlockingVar x mcons) =
     let (tel1, Dom info (n, t) : tel2) = splitAt (size tel - x - 1) $ telToList tel
     return (n, Dom info t, telFromList tel1, telFromList tel2)
 
-  -- Check that t is a datatype or a record
-  -- Andreas, 2010-09-21, isDatatype now directly throws an exception if it fails
-  -- cons = constructors of this datatype
-  (d, pars, ixs, cons) <- inContextOfT $ isDatatype ind t
-
   -- Compute the neighbourhoods for the constructors
-  ns <- catMaybes <$> do
-    forM cons $ \ con ->
-      fmap (con,) <$> do
-        msc <- computeNeighbourhood delta1 n delta2 d pars ixs x tel ps mpsub con
-        if not fixtarget then return msc else do
-        Trav.forM msc $ \ sc -> lift $ snd <$> fixTarget sc{ scTarget = target }
+  let computeNeighborhoods = do
+        -- Check that t is a datatype or a record
+        -- Andreas, 2010-09-21, isDatatype now directly throws an exception if it fails
+        -- cons = constructors of this datatype
+        (d, pars, ixs, cons) <- inContextOfT $ isDatatype ind t
+        mns <- forM cons $ \ con -> fmap (SplitCon con,) <$>
+          computeNeighbourhood delta1 n delta2 d pars ixs x tel ps cps con
+        return $ catMaybes mns
+
+      computeLitNeighborhoods = do
+        typeOk <- liftTCM $ do
+          t' <- litType $ headWithDefault {-'-} __IMPOSSIBLE__ plits
+          liftTCM $ dontAssignMetas $ tryConversion $ equalType (unDom t) t'
+        unless typeOk $ throwError . NotADatatype =<< do liftTCM $ buildClosure (unDom t)
+        ns <- forM plits $ \lit -> do
+          let delta2' = subst 0 (Lit lit) delta2
+              delta'  = delta1 `abstract` delta2'
+              rho     = liftS x $ consS (LitP lit) idS
+              ps'     = applySubst rho ps
+              cps'    = applySplitPSubst rho cps
+          return (SplitLit lit , SClause delta' ps' rho cps' Nothing)
+        ca <- do
+          let delta' = tel -- telescope is unchanged for catchall branch
+              varp   = VarP PatOSplit $ SplitPatVar
+                         { splitPatVarName   = underscore
+                         , splitPatVarIndex  = 0
+                         , splitExcludedLits = plits
+                         }
+              rho    = liftS x $ consS varp $ raiseS 1
+              ps'    = applySubst rho ps
+          return (SplitCatchall , SClause delta' ps' rho cps Nothing)
+        return $ ns ++ [ ca ]
+
+  ns <- if null pcons' && not (null plits)
+        then computeLitNeighborhoods
+        else computeNeighborhoods
+
+  ns <- case fixtarget of
+    NoFixTarget  -> return ns
+    YesFixTarget -> lift $ forM ns $ \(con,sc) ->
+      (con,) . snd <$> fixTarget sc{ scTarget = target }
 
   case ns of
     []  -> do
-      let ps' = (fmap . fmap . fmap . fmap)
-                  (\(DBPatVar name y) -> if (x==y)
-                                         then DBPatVar absurdPatternName y
-                                         else DBPatVar name y)
-                  ps
+      let absurdp = VarP PatOAbsurd $ SplitPatVar underscore 0 []
+          rho = liftS x $ consS absurdp $ raiseS 1
+          ps' = applySubst rho ps
       return $ Left $ SClause
-               { scTel  = telFromList $ telToList delta1 ++
-                                        [fmap ((,) "()") t] ++ -- add name "()"
-                                        telToList delta2
-               , scPats = ps
-               , scSubst = idS -- not used anyway
-               , scModuleParameterSub = __IMPOSSIBLE__ -- not used
-               , scTarget = Nothing -- not used
+               { scTel  = tel
+               , scPats = ps'
+               , scSubst              = __IMPOSSIBLE__ -- not used
+               , scCheckpoints        = __IMPOSSIBLE__ -- not used
+               , scTarget             = Nothing
                }
 
     -- Andreas, 2011-10-03
@@ -803,36 +861,37 @@ split' ind fixtarget sc@(SClause tel ps _ mpsub target) (BlockingVar x mcons) =
 
   -- Andreas, 2012-10-10 fail if precomputed constructor set does not cover
   -- all the data type constructors
-
-    _ | Just pcons' <- mcons,
-        let pcons = map conName pcons',
-        let cons = (map fst ns),
-        let diff = Set.fromList cons Set.\\ Set.fromList pcons,
+  -- Andreas, 2017-10-08 ... unless partial covering is explicitly allowed.
+    _ | allowPartialCover == NoAllowPartialCover,
+        overlap == False,
+        let ptags = map (SplitCon . conName) pcons' ++ map SplitLit plits,
+        let tags  = map fst ns,
+        let diff  = Set.fromList tags Set.\\ Set.fromList ptags,
         not (Set.null diff) -> do
           liftTCM $ reportSDoc "tc.cover.precomputed" 10 $ vcat
-            [ hsep $ text "pcons =" : map prettyTCM pcons
-            , hsep $ text "cons  =" : map prettyTCM cons
+            [ hsep $ text "ptags =" : map prettyTCM ptags
+            , hsep $ text "tags  =" : map prettyTCM tags
             ]
           throwError (GenericSplitError "precomputed set of constructors does not cover all cases")
 
-    _  -> return $ Right $ Covering (lookupPatternVar sc x) ns
+    _  -> do
+      liftTCM $ checkSortOfSplitVar $ unDom t
+      return $ Right $ Covering (lookupPatternVar sc x) ns
 
   where
-    inContextOfT :: MonadTCM tcm => tcm a -> tcm a
-    inContextOfT = addContext tel . escapeContext (x + 1)
-
-    inContextOfDelta2 :: MonadTCM tcm => tcm a -> tcm a
+    inContextOfT, inContextOfDelta2 :: (MonadTCM tcm, MonadDebug tcm) => tcm a -> tcm a
+    inContextOfT      = addContext tel . escapeContext (x + 1)
     inContextOfDelta2 = addContext tel . escapeContext x
 
     -- Debug printing
-    debugInit tel x ps mpsub = liftTCM $ inTopContext $ do
+    debugInit tel x ps cps = liftTCM $ inTopContext $ do
       reportSDoc "tc.cover.top" 10 $ vcat
         [ text "TypeChecking.Coverage.split': split"
         , nest 2 $ vcat
           [ text "tel     =" <+> prettyTCM tel
           , text "x       =" <+> prettyTCM x
-          , text "ps      =" <+> do addContext tel $ prettyTCMPatternList ps
-          , text "mpsub   =" <+> prettyTCM mpsub
+          , text "ps      =" <+> do addContext tel $ prettyTCMPatternList $ fromSplitPatterns ps
+          , text "cps     =" <+> prettyTCM cps
           ]
         ]
 
@@ -845,13 +904,31 @@ split' ind fixtarget sc@(SClause tel ps _ mpsub target) (BlockingVar x mcons) =
         , text "t      =" <+> inContextOfT (prettyTCM t)
         ]
 
+-- | What could go wrong if we try to split on the result?
+data CosplitError
+  = CosplitNoTarget
+      -- ^ We do not know the target type of the clause.
+  | CosplitNoRecordType Telescope Type
+      -- ^ Type living in the given telescope is not a record type.
+  | CosplitIrrelevantProjections
+      -- ^ Record has irrelevant fields, but we do not have irrelevant projections.
+
+instance PrettyTCM CosplitError where
+  prettyTCM = \case
+    CosplitNoTarget ->
+      text "target type is unknown"
+    CosplitIrrelevantProjections ->
+      text "record has irrelevant fields, but no corresponding projections"
+    CosplitNoRecordType tel t -> addContext tel $ do
+      text "target type " <+> prettyTCM t <+> text " is not a record type"
+
 -- | @splitResult f sc = return res@
 --
 --   If the target type of @sc@ is a record type, a covering set of
 --   split clauses is returned (@sc@ extended by all valid projection patterns),
 --   otherwise @res == Nothing@.
 --   Note that the empty set of split clauses is returned if the record has no fields.
-splitResult :: QName -> SplitClause -> TCM (Maybe Covering)
+splitResult :: QName -> SplitClause -> TCM (Either CosplitError Covering)
 splitResult f sc@(SClause tel ps _ _ target) = do
   reportSDoc "tc.cover.split" 10 $ vcat
     [ text "splitting result:"
@@ -859,8 +936,8 @@ splitResult f sc@(SClause tel ps _ _ target) = do
     , nest 2 $ text "target =" <+> (addContext tel $ maybe empty prettyTCM target)
     ]
   -- if we want to split projections, but have no target type, we give up
-  let done = return Nothing
-  caseMaybe target done $ \ t -> do
+  let failure = return . Left
+  caseMaybe target (failure CosplitNoTarget) $ \ t -> do
     isR <- addContext tel $ isRecordType $ unArg t
     case isR of
       Just (_r, vs, Record{ recFields = fs }) -> do
@@ -869,21 +946,24 @@ splitResult f sc@(SClause tel ps _ _ target) = do
           , text   "applied to parameters vs = " <+> (addContext tel $ prettyTCM vs)
           , text $ "and have fields       fs = " ++ prettyShow fs
           ]
-        let es = patternsToElims ps
+        -- Andreas, 2018-03-19, issue #2971, check that we have a "strong" record type,
+        -- i.e., with all the projections.  Otherwise, we may not split.
+        ifNotM (strongRecord fs) (failure CosplitIrrelevantProjections) $ {-else-} do
+        let es = patternsToElims $ fromSplitPatterns ps
         -- Note: module parameters are part of ps
         let self  = defaultArg $ Def f [] `applyE` es
             pargs = vs ++ [self]
         reportSDoc "tc.cover" 20 $ sep
           [ text   "we are              self = " <+> (addContext tel $ prettyTCM $ unArg self)
           ]
-        let n = defaultArg $ permRange $ fromMaybe __IMPOSSIBLE__ $ dbPatPerm ps
+        let n = defaultArg $ permRange $ fromMaybe __IMPOSSIBLE__ $ dbPatPerm $ fromSplitPatterns ps
             -- Andreas & James, 2013-11-19 includes the dot patterns!
             -- See test/succeed/CopatternsAndDotPatterns.agda for a case with dot patterns
             -- and copatterns which fails for @n = size tel@ with a broken case tree.
 
         -- Andreas, 2016-07-22 read the style of projections from the user's lips
         projOrigin <- ifM (optPostfixProjections <$> pragmaOptions) (return ProjPostfix) (return ProjPrefix)
-        Just . Covering n <$> do
+        Right . Covering n <$> do
           forM fs $ \ proj -> do
             -- compute the new target
             dType <- defType <$> do getConstInfo $ unArg proj -- WRONG: typeOfConst $ unArg proj
@@ -894,20 +974,28 @@ splitResult f sc@(SClause tel ps _ _ target) = do
                          , scSubst  = idS
                          , scTarget = target'
                          }
-            return (unArg proj, sc')
-      _ -> done
+            return (SplitCon (unArg proj), sc')
+      _ -> failure $ CosplitNoRecordType tel $ unArg t
+  where
+  -- A record type is strong if it has all the projections.
+  -- This is the case if --irrelevant-projections or no field is irrelevant.
+  -- TODO: what about shape irrelevance?
+  strongRecord :: [Arg QName] -> TCM Bool
+  strongRecord fs = (optIrrelevantProjections <$> pragmaOptions) `or2M`
+    (return $ not $ any isIrrelevant fs)
+
 
 -- * Boring instances
 
 -- | For debugging only.
 instance PrettyTCM SplitClause where
-  prettyTCM (SClause tel pats sigma mpsub target) = sep
+  prettyTCM (SClause tel pats sigma cps target) = sep
     [ text "SplitClause"
     , nest 2 $ vcat
       [ text "tel          =" <+> prettyTCM tel
       , text "pats         =" <+> sep (map (prettyTCM . namedArg) pats)
       , text "subst        =" <+> (text . show) sigma
-      , text "mpsub        =" <+> prettyTCM mpsub
+      , text "checkpoints  =" <+> prettyTCM cps
       , text "target       =" <+> do
           caseMaybe target empty $ \ t -> do
             addContext tel $ prettyTCM t
