@@ -34,7 +34,7 @@ getConHead c = mapRight (conSrcCon . theDef) <$> getConstInfo' c
 --   if possible.
 getConForm :: QName -> TCM (Either SigError ConHead)
 getConForm c = caseEitherM (getConHead c) (return . Left) $ \ ch -> do
-  Con con _ [] <- ignoreSharing <$> constructorForm (Con ch ConOCon [])
+  Con con _ [] <- constructorForm (Con ch ConOCon [])
   return $ Right con
 
 -- | Augment constructor with record fields (preserve constructor name).
@@ -118,7 +118,7 @@ getFullyAppliedConType c t = do
   reportSLn "tc.getConType" 35 $ List.intercalate " " $
     [ "getFullyAppliedConType", prettyShow c, prettyShow t ]
   c <- fromRight __IMPOSSIBLE__ <$> do getConHead $ conName c
-  case ignoreSharing $ unEl t of
+  case unEl t of
     -- Note that if we come e.g. from getConType,
     -- then the non-parameter arguments of @es@ might contain __IMPOSSIBLE__
     -- coming from strengthening.  (Thus, printing them is not safe.)
@@ -137,9 +137,6 @@ getFullyAppliedConType c t = do
         _ ->  return Nothing
     _ -> return Nothing
 
-data HasEta = NoEta | YesEta
-  deriving (Eq)
-
 data ConstructorInfo
   = DataCon Nat                  -- ^ Arity.
   | RecordCon HasEta [Arg QName] -- ^ List of field names.
@@ -155,7 +152,7 @@ getConstructorInfo c = do
     Constructor{ conData = d, conArity = n } -> do
       (theDef <$> getConstInfo d) >>= \case
         r@Record{ recFields = fs } ->
-           return $ RecordCon (if recEtaEquality r then YesEta else NoEta) fs
+           return $ RecordCon (recEtaEquality r) fs
         Datatype{} ->
            return $ DataCon n
         _ -> __IMPOSSIBLE__
@@ -186,7 +183,7 @@ isDataOrRecordType d = do
 -- | Precodition: 'Term' is 'reduce'd.
 isDataOrRecord :: Term -> TCM (Maybe QName)
 isDataOrRecord v = do
-  case ignoreSharing v of
+  case v of
     Def d _ -> fmap (const d) <$> isDataOrRecordType d
     _       -> return Nothing
 
@@ -201,12 +198,35 @@ getNumberOfParameters d = do
 
 -- | Precondition: Name is a data or record type.
 getConstructors :: QName -> TCM [QName]
-getConstructors d = do
-  def <- theDef <$> getConstInfo d
-  case def of
-    Datatype{dataCons = cs} -> return cs
-    Record{recConHead = h}  -> return [conName h]
-    _                       -> __IMPOSSIBLE__
+getConstructors d = fromMaybe __IMPOSSIBLE__ <$>
+  getConstructors' d
+
+-- | 'Nothing' if not data or record type name.
+getConstructors' :: QName -> TCM (Maybe [QName])
+getConstructors' d = getConstructors_ . theDef <$> getConstInfo d
+
+-- | 'Nothing' if not data or record definition.
+getConstructors_ :: Defn -> Maybe [QName]
+getConstructors_ = \case
+    Datatype{dataCons = cs} -> Just cs
+    Record{recConHead = h}  -> Just [conName h]
+    _                       -> Nothing
+
+-- | Precondition: Name is a data or record type.
+getConHeads :: QName -> TCM [ConHead]
+getConHeads d = fromMaybe __IMPOSSIBLE__ <$>
+  getConHeads' d
+
+-- | 'Nothing' if not data or record type name.
+getConHeads' :: QName -> TCM (Maybe [ConHead])
+getConHeads' d = getConHeads_ . theDef <$> getConstInfo d
+
+-- | 'Nothing' if not data or record definition.
+getConHeads_ :: Defn -> Maybe [ConHead]
+getConHeads_ = \case
+    Datatype{dataCons = cs} -> Just $ map (\ c -> ConHead c Inductive []) cs
+    Record{recConHead = h}  -> Just [h]
+    _                       -> Nothing
 
 {- UNUSED
 data DatatypeInfo = DataInfo
