@@ -1,8 +1,6 @@
 
 module Agda.TypeChecking.Empty (isEmptyType) where
 
-import Control.Applicative
-import Control.Monad
 import Control.Monad.Except
 
 import Data.Semigroup
@@ -15,6 +13,7 @@ import Agda.Syntax.Position
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Coverage
+import Agda.TypeChecking.Coverage.Match ( fromSplitPatterns )
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Reduce
@@ -41,7 +40,10 @@ instance Monoid ErrorNonEmpty where
 
 -- | Check whether a type is empty.
 --   This check may be postponed as emptiness constraint.
-isEmptyType :: Range -> Type -> TCM ()
+isEmptyType
+  :: Range   -- ^ Range of the absurd pattern.
+  -> Type    -- ^ Type that should be empty (empty data type or iterated product of such).
+  -> TCM ()
 isEmptyType r t = caseEitherM (loop t) failure return
   where
   failure DontKnow          = addConstraint $ IsEmpty r t
@@ -56,10 +58,10 @@ isEmptyType r t = caseEitherM (loop t) failure return
     case mr of
 
       -- If t is blocked or a meta, we cannot decide emptiness now.  Postpone.
-      Left Nothing -> return $ Left DontKnow
+      Left (Blocked m t) -> return $ Left DontKnow
 
       -- If t is not a record type, try to split
-      Left (Just t) -> do
+      Left (NotBlocked nb t) -> do
         -- from the current context xs:ts, create a pattern list
         -- xs _ : ts t and try to split on _ (the last variable)
         tel0 <- getContextTelescope
@@ -73,13 +75,13 @@ isEmptyType r t = caseEitherM (loop t) failure return
             Left UnificationStuck{} -> return $ Left DontKnow
             Left _                  -> return $ Left Fail
             Right cov -> do
-              let ps = map (namedArg . last . scPats) $ splitClauses cov
+              let ps = map (namedArg . last . fromSplitPatterns . scPats) $ splitClauses cov
               if (null ps) then return (Right ()) else
                 Left . FailBecause <$> do typeError_ $ ShouldBeEmpty t ps
 
       -- If t is a record type, see if any of the field types is empty
       Right (r, pars, def) -> do
-        if not $ recEtaEquality def then return $ Left Fail else do
+        if recEtaEquality def == NoEta then return $ Left Fail else do
           checkTel $ recTel def `apply` pars
 
   checkTel :: Telescope -> TCM (Either ErrorNonEmpty ())
