@@ -16,15 +16,15 @@ module Agda.TypeChecking.Errors
   , sayWhen
   ) where
 
-import Prelude hiding (null)
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ( (<>), null )
+#else
+import Prelude hiding ( null )
+#endif
 
 import Control.Monad.Reader
 import Control.Monad.State
 
-#if __GLASGOW_HASKELL__ <= 708
-import Data.Foldable (foldMap)
-#endif
-
 import Data.Function
 import Data.List (nub, sortBy, intersperse, isInfixOf)
 import Data.Maybe
@@ -46,7 +46,7 @@ import Agda.Syntax.Abstract.Views (deepUnscope)
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Translation.InternalToAbstract
 import Agda.Syntax.Translation.AbstractToConcrete
-import Agda.Syntax.Scope.Monad (isDatatypeModule)
+import Agda.Syntax.Scope.Monad (isDatatypeModule, withContextPrecedence)
 import Agda.Syntax.Scope.Base
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Closure
@@ -69,6 +69,7 @@ import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
+import Agda.Utils.NonemptyList
 import Agda.Utils.Pretty ( prettyShow )
 import qualified Agda.Utils.Pretty as P
 import Agda.Utils.Size
@@ -163,7 +164,7 @@ prettyWarning wng = liftTCM $ case wng of
       ++ [prettyTCM ocs]
 
     OldBuiltin old new -> fwords $
-      "Builtin " ++ old ++ " does no longer exist. " ++
+      "Builtin " ++ old ++ " no longer exists. " ++
       "It is now bound by BUILTIN " ++ new
 
     EmptyRewritePragma -> fsep . pwords $ "Empty REWRITE pragma"
@@ -174,6 +175,14 @@ prettyWarning wng = liftTCM $ case wng of
       pwords "It is pointless for INLINE'd function" ++ [prettyTCM q] ++
       pwords "to have a separate Haskell definition"
 
+    InversionDepthReached f -> do
+      maxDepth <- maxInversionDepth
+      fsep $ pwords "Refusing to invert pattern matching of" ++ [prettyTCM f] ++
+             pwords ("because the maximum depth (" ++ show maxDepth ++ ") has been reached.") ++
+             pwords "Most likely this means you have an unsatisfiable constraint, but it could" ++
+             pwords "also mean that you need to increase the maximum depth using the flag" ++
+             pwords "--inversion-max-depth=N"
+
     GenericWarning d -> return d
 
     GenericNonFatalError d -> return d
@@ -184,7 +193,7 @@ prettyWarning wng = liftTCM $ case wng of
     SafeFlagPragma xs ->
       let plural | length xs == 1 = ""
                  | otherwise      = "s"
-      in fsep $ [fwords ("Cannot set OPTION pragma" ++ plural)]
+      in fsep $ [fwords ("Cannot set OPTIONS pragma" ++ plural)]
                 ++ map text xs ++ [fwords "with safe flag."]
 
     SafeFlagNonTerminating -> fsep $
@@ -193,13 +202,13 @@ prettyWarning wng = liftTCM $ case wng of
     SafeFlagTerminating -> fsep $
       pwords "Cannot use TERMINATING pragma with safe flag."
 
-    SafeFlagPrimTrustMe -> fsep (pwords "Cannot use primTrustMe with safe flag")
+    SafeFlagPrimTrustMe -> fsep (pwords "Cannot use primTrustMe with safe flag.")
 
     SafeFlagNoPositivityCheck -> fsep $
       pwords "Cannot use NO_POSITIVITY_CHECK pragma with safe flag."
 
     SafeFlagPolarity -> fsep $
-      pwords "The POLARITY pragma must not be used in safe mode."
+      pwords "Cannot use POLARITY pragma with safe flag."
 
     ParseWarning pw -> pretty pw
 
@@ -207,9 +216,9 @@ prettyWarning wng = liftTCM $ case wng of
       [text old] ++ pwords "has been deprecated. Use" ++ [text new] ++ pwords
       "instead. This will be an error in Agda" ++ [text version <> text "."]
 
-    NicifierIssue ws -> vcat $ do
-      for ws $ \ w -> do
-        sayWhere (getRange w) $ pretty w
+    NicifierIssue w -> sayWhere (getRange w) $ pretty w
+
+    UserWarning str -> text str
 
 prettyTCWarnings :: [TCWarning] -> TCM String
 prettyTCWarnings = fmap (unlines . intersperse "") . prettyTCWarnings'
@@ -234,8 +243,8 @@ applyFlagsToTCWarnings ifs ws = do
   -- This is a way to collect all of them and remove duplicates.
   let pragmas w = case tcWarning w of { SafeFlagPragma ps -> ([w], ps); _ -> ([], []) }
   let sfp = case fmap nub (foldMap pragmas ws) of
-              (TCWarning r w p:_, sfp) ->
-                 [TCWarning r (SafeFlagPragma sfp) p]
+              (TCWarning r w p b:_, sfp) ->
+                 [TCWarning r (SafeFlagPragma sfp) p b]
               _                        -> []
 
 
@@ -260,6 +269,7 @@ applyFlagsToTCWarnings ifs ws = do
           UselessPublic                -> True
           ParseWarning{}               -> True
           UnreachableClauses{}         -> True
+          InversionDepthReached{}      -> True
           CoverageNoExactSplit{}       -> catchallNotOK
           UselessInline{}              -> True
           GenericWarning{}             -> True
@@ -273,6 +283,7 @@ applyFlagsToTCWarnings ifs ws = do
           SafeFlagPolarity             -> True
           DeprecationWarning{}         -> True
           NicifierIssue{}              -> True
+          UserWarning{}                -> True
 
   return $ sfp ++ filter (cleanUp . tcWarning) ws
 
@@ -293,7 +304,7 @@ panic s = fwords $ "Panic: " ++ s
 
 nameWithBinding :: QName -> TCM Doc
 nameWithBinding q =
-  sep [ prettyTCM q, text "bound at", prettyTCM r ]
+  (prettyTCM q <+> text "bound at") <?> prettyTCM r
   where
     r = nameBindingSite $ qnameName q
 
@@ -317,6 +328,7 @@ errorString err = case err of
   AmbiguousTopLevelModuleName {}           -> "AmbiguousTopLevelModuleName"
   BadArgumentsToPatternSynonym{}           -> "BadArgumentsToPatternSynonym"
   TooFewArgumentsToPatternSynonym{}        -> "TooFewArgumentsToPatternSynonym"
+  CannotResolveAmbiguousPatternSynonym{}   -> "CannotResolveAmbiguousPatternSynonym"
   BothWithAndRHS                           -> "BothWithAndRHS"
   BuiltinInParameterisedModule{}           -> "BuiltinInParameterisedModule"
   BuiltinMustBeConstructor{}               -> "BuiltinMustBeConstructor"
@@ -352,6 +364,7 @@ errorString err = case err of
   LocalVsImportedModuleClash{}             -> "LocalVsImportedModuleClash"
   MetaCannotDependOn{}                     -> "MetaCannotDependOn"
   MetaOccursInItself{}                     -> "MetaOccursInItself"
+  MetaIrrelevantSolution{}                 -> "MetaIrrelevantSolution"
   ModuleArityMismatch{}                    -> "ModuleArityMismatch"
   ModuleDefinedInOtherFile {}              -> "ModuleDefinedInOtherFile"
   ModuleDoesntExport{}                     -> "ModuleDoesntExport"
@@ -403,10 +416,10 @@ errorString err = case err of
   ImpossibleConstructor{}                  -> "ImpossibleConstructor"
   TerminationCheckFailed{}                 -> "TerminationCheckFailed"
   TooFewFields{}                           -> "TooFewFields"
-  TooManyArgumentsInLHS{}                  -> "TooManyArgumentsInLHS"
   TooManyFields{}                          -> "TooManyFields"
   TooManyPolarities{}                      -> "TooManyPolarities"
   SplitOnIrrelevant{}                      -> "SplitOnIrrelevant"
+  SplitOnNonVariable{}                     -> "SplitOnNonVariable"
   DefinitionIsIrrelevant{}                 -> "DefinitionIsIrrelevant"
   VariableIsIrrelevant{}                   -> "VariableIsIrrelevant"
   UnequalBecauseOfUniverseConflict{}       -> "UnequalBecauseOfUniverseConflict"
@@ -420,6 +433,7 @@ errorString err = case err of
   WithOnFreeVariable{}                     -> "WithOnFreeVariable"
   UnexpectedWithPatterns{}                 -> "UnexpectedWithPatterns"
   UninstantiatedDotPattern{}               -> "UninstantiatedDotPattern"
+  ForcedConstructorNotInstantiated{}       -> "ForcedConstructorNotInstantiated"
   UninstantiatedModule{}                   -> "UninstantiatedModule"
   SolvedButOpenHoles{}                     -> "SolvedButOpenHoles"
   UnusedVariableInPatternSynonym           -> "UnusedVariableInPatternSynonym"
@@ -443,7 +457,7 @@ instance PrettyTCM TCErr where
     -- Gallais, 2016-05-14
     -- Given where `NonFatalErrors` are created, we know for a
     -- fact that ̀ws` is non-empty.
-    TypeError _ (Closure _ _ _ _ (NonFatalErrors ws)) -> foldr1 ($$) $ fmap prettyTCM ws
+    TypeError _ Closure{ clValue = NonFatalErrors ws } -> foldr1 ($$) $ fmap prettyTCM ws
     -- Andreas, 2014-03-23
     -- This use of localState seems ok since we do not collect
     -- Benchmark info during printing errors.
@@ -565,6 +579,10 @@ instance PrettyTCM TypeError where
     UninstantiatedDotPattern e -> fsep $
       pwords "Failed to infer the value of dotted pattern"
 
+    ForcedConstructorNotInstantiated p -> fsep $
+      pwords "Failed to infer that constructor pattern "
+      ++ [prettyA p] ++ pwords " is forced"
+
     IlltypedPattern p a -> do
       let ho _ _ = fsep $ pwords "Cannot pattern match on functions"
       ifPiType a ho $ {- else -} \ _ -> do
@@ -583,10 +601,6 @@ instance PrettyTCM TypeError where
            pwords "with pattern" ++ prettyA p :
            pwords "(did you supply too many arguments?)"
 
-    TooManyArgumentsInLHS a -> fsep $
-      pwords "Left hand side gives too many arguments to a function of type"
-      ++ [prettyTCM a]
-
     WrongNumberOfConstructorArguments c expect given -> fsep $
       pwords "The constructor" ++ [prettyTCM c] ++
       pwords "expects" ++ [prettyTCM expect] ++
@@ -677,11 +691,14 @@ instance PrettyTCM TypeError where
       pwords "Functions may not return sizes, thus, function type " ++
       [ prettyTCM v ] ++ pwords " is illegal"
 
-    SplitOnIrrelevant p t -> fsep $
-      pwords "Cannot pattern match" ++ [prettyA p] ++
-      pwords "against" ++ [text $ verbalize $ getRelevance t] ++
+    SplitOnIrrelevant t -> fsep $
+      pwords "Cannot pattern match against" ++ [text $ verbalize $ getRelevance t] ++
       pwords "argument of type" ++ [prettyTCM t]
 
+    SplitOnNonVariable v t -> fsep $
+      pwords "Cannot pattern match because the (refined) argument " ++
+      [ prettyTCM v ] ++ pwords " is not a variable."
+
     DefinitionIsIrrelevant x -> fsep $
       text "Identifier" : prettyTCM x : pwords "is declared irrelevant, so it cannot be used here"
     VariableIsIrrelevant x -> fsep $
@@ -706,8 +723,6 @@ instance PrettyTCM TypeError where
 
     UnequalRelevance cmp a b -> fsep $
       [prettyTCM a, notCmp cmp, prettyTCM b] ++
--- Andreas 2010-09-21 to reveal Forced annotations, print also uglily
---            [text $ show a, notCmp cmp, text $ show b] ++
       pwords "because one is a relevant function type and the other is an irrelevant function type"
 
     UnequalHiding a b -> fsep $
@@ -769,6 +784,10 @@ instance PrettyTCM TypeError where
     MetaOccursInItself m -> fsep $
       pwords "Cannot construct infinite solution of metavariable" ++ [prettyTCM $ MetaV m []]
 
+    MetaIrrelevantSolution m _ -> fsep $
+      pwords "Cannot instantiate the metavariable because (part of) the" ++
+      pwords "solution was created in an irrelevant context."
+
     BuiltinMustBeConstructor s e -> fsep $
       [prettyA e] ++ pwords "must be a constructor in the binding to builtin" ++ [text s]
 
@@ -898,14 +917,14 @@ instance PrettyTCM TypeError where
     AmbiguousName x ys -> vcat
       [ fsep $ pwords "Ambiguous name" ++ [pretty x <> text "."] ++
                pwords "It could refer to any one of"
-      , nest 2 $ vcat $ map nameWithBinding ys
+      , nest 2 $ vcat $ map nameWithBinding (toList ys)
       , fwords "(hint: Use C-c C-w (in Emacs) if you want to know why)"
       ]
 
     AmbiguousModule x ys -> vcat
       [ fsep $ pwords "Ambiguous module name" ++ [pretty x <> text "."] ++
                pwords "It could refer to any one of"
-      , nest 2 $ vcat $ map help ys
+      , nest 2 $ vcat $ map help (toList ys)
       , fwords "(hint: Use C-c C-w (in Emacs) if you want to know why)"
       ]
       where
@@ -1020,10 +1039,21 @@ instance PrettyTCM TypeError where
         isPlaceholder NoPlaceholder{} = False
 
     BadArgumentsToPatternSynonym x -> fsep $
-      pwords "Bad arguments to pattern synonym " ++ [prettyTCM x]
+      pwords "Bad arguments to pattern synonym " ++ [prettyTCM $ headAmbQ x]
 
     TooFewArgumentsToPatternSynonym x -> fsep $
-      pwords "Too few arguments to pattern synonym " ++ [prettyTCM x]
+      pwords "Too few arguments to pattern synonym " ++ [prettyTCM $ headAmbQ x]
+
+    CannotResolveAmbiguousPatternSynonym defs -> vcat
+      [ fsep $ pwords "Cannot resolve overloaded pattern synonym" ++ [prettyTCM x <> comma] ++
+               pwords "since candidates have different shapes:"
+      , nest 2 $ vcat $ map prDef (toList defs)
+      , fsep $ pwords "(hint: overloaded pattern synonyms must be equal up to variable and constructor names)"
+      ]
+      where
+        (x, _) = headNe defs
+        prDef (x, (xs, p)) = prettyA (A.PatternSynDef x xs p) <?> (text "at" <+> pretty r)
+          where r = nameBindingSite $ qnameName x
 
     UnusedVariableInPatternSynonym -> fsep $
       pwords "Unused variable in pattern synonym."
@@ -1241,9 +1271,8 @@ instance PrettyTCM TypeError where
       NotHidden  -> prettyPat 1 x
 
     prettyPat :: Integer -> (I.Pattern' a) -> TCM Doc
-    prettyPat _ (I.VarP _) = text "_"
-    prettyPat _ (I.DotP _) = text "._"
-    prettyPat _ (I.AbsurdP _) = text absurdPatternName
+    prettyPat _ (I.VarP _ _) = text "_"
+    prettyPat _ (I.DotP _ _) = text "._"
     prettyPat n (I.ConP c _ args) =
       mpar n args $
         prettyTCM c <+> fsep (map (prettyArg . fmap namedThing) args)
@@ -1264,7 +1293,7 @@ prettyInEqual t1 t2 = do
      -- if printed differently, no extra explanation needed
     if P.render d1 /= P.render d2 then empty else do
       (v1, v2) <- instantiate (t1, t2)
-      case (ignoreSharing v1, ignoreSharing v2) of
+      case (v1, v2) of
         (I.Var i1 _, I.Var i2 _)
           | i1 == i2  -> generic -- possible, see issue 1826
           | otherwise -> varVar i1 i2
@@ -1364,7 +1393,7 @@ instance PrettyTCM UnificationFailure where
 
 
 instance PrettyTCM Call where
-  prettyTCM c = case c of
+  prettyTCM c = withContextPrecedence TopCtx $ case c of
     CheckClause t cl -> do
       reportSLn "error.checkclause" 60 $ "prettyTCM CheckClause: cl = " ++ show (deepUnscope cl)
       clc <- abstractToConcrete_ cl
@@ -1406,6 +1435,12 @@ instance PrettyTCM Call where
       pwords "to a function of type" ++
       [prettyTCM t0]
 
+    CheckTargetType r infTy expTy -> sep
+      [ text "when checking that the inferred type of an application"
+      , nest 2 $ prettyTCM infTy
+      , text "matches the expected type"
+      , nest 2 $ prettyTCM expTy ]
+
     CheckRecDef _ x ps cs ->
       fsep $ pwords "when checking the definition of" ++ [prettyTCM x]
 
@@ -1418,7 +1453,7 @@ instance PrettyTCM Call where
 
     CheckConstructor{} -> __IMPOSSIBLE__
 
-    CheckFunDef _ f _ ->
+    CheckFunDefCall _ f _ ->
       fsep $ pwords "when checking the definition of" ++ [prettyTCM f]
 
     CheckPragma _ p ->
@@ -1440,6 +1475,9 @@ instance PrettyTCM Call where
     CheckPatternShadowing c -> fsep $
       pwords "when checking the clause" ++ [prettyA c]
 
+    CheckNamedWhere m -> fsep $
+      pwords "when checking the named where block" ++ [prettyA m]
+
     InferVar x ->
       fsep $ pwords "when inferring the type of" ++ [prettyTCM x]
 
@@ -1479,7 +1517,7 @@ instance PrettyTCM Call where
 
     where
     hPretty :: Arg (Named_ Expr) -> TCM Doc
-    hPretty a = pretty =<< abstractToConcreteCtx (hiddenArgumentCtx (getHiding a)) a
+    hPretty a = withContextPrecedence (ArgumentCtx PreferParen) $ pretty =<< abstractToConcreteHiding a a
 
 ---------------------------------------------------------------------------
 -- * Natural language
@@ -1501,7 +1539,6 @@ instance Verbalize Relevance where
       Relevant   -> "relevant"
       Irrelevant -> "irrelevant"
       NonStrict  -> "shape-irrelevant"
-      Forced{}   -> "forced"
 
 -- | Indefinite article.
 data Indefinite a = Indefinite a
