@@ -18,15 +18,17 @@
 
 module Agda.TypeChecking.Free
     ( FreeVars(..)
+    , VarCounts(..)
     , Free
     , IsVarSet(..)
     , IgnoreSorts(..)
     , runFree , rigidVars, relevantVars, allVars
-    , allFreeVars
+    , allFreeVars, allFreeVarsWithOcc
     , allRelevantVars, allRelevantVarsIgnoring
     , freeIn, freeInIgnoringSorts, isBinderUsed
     , relevantIn, relevantInIgnoringSortAnn
     , Occurrence(..)
+    , VarOcc(..)
     , occurrence
     , closed
     , freeVars -- only for testing
@@ -38,7 +40,8 @@ import Prelude hiding (null)
 import Control.Monad.Reader
 
 import Data.Maybe
-import Data.Semigroup (Semigroup, Monoid, (<>), mempty, mappend, mconcat, Any(..), All(..))
+import Data.Monoid ( Monoid, mempty, mappend, mconcat )
+import Data.Semigroup ( Semigroup, (<>), Any(..), All(..) )
 import Data.IntSet (IntSet)
 import qualified Data.IntSet as Set
 import Data.IntMap (IntMap)
@@ -53,7 +56,7 @@ import Agda.Syntax.Internal
 
 import Agda.TypeChecking.Free.Lazy
   ( Free(..) , FreeEnv(..), initFreeEnv
-  , VarOcc(..), IgnoreSorts(..), Variable, SingleVar
+  , VarOcc(..), topVarOcc, TheVarMap, theVarMap, IgnoreSorts(..), Variable, SingleVar
   , MetaSet, IsVarSet(..), runFreeM
   )
 import qualified Agda.TypeChecking.Free.Lazy as Free
@@ -75,7 +78,7 @@ data FreeVars = FV
     --   can still become strongly rigid if put under a constructor
     --   whereas weakly rigid ones stay weakly rigid.
   , weaklyRigidVars   :: VarSet
-    -- ^ Ordinary rigid variables, e.g., in arguments of variables.
+    -- ^ Ordinary rigid variables, e.g., in arguments of variables or functions.
   , flexibleVars      :: IntMap MetaSet
     -- ^ Variables occuring in arguments of metas.
     --   These are only potentially free, depending how the meta variable is instantiated.
@@ -83,17 +86,14 @@ data FreeVars = FV
   , irrelevantVars    :: VarSet
     -- ^ Variables in irrelevant arguments and under a @DontCare@, i.e.,
     --   in irrelevant positions.
-  , unusedVars        :: VarSet
-    -- ^ Variables in 'UnusedArg'uments.
   } deriving (Eq, Show)
 
-mapSRV, mapUGV, mapWRV, mapIRV, mapUUV
+mapSRV, mapUGV, mapWRV, mapIRV
   :: (VarSet -> VarSet) -> FreeVars -> FreeVars
 mapSRV f fv = fv { stronglyRigidVars = f $ stronglyRigidVars fv }
 mapUGV f fv = fv { unguardedVars     = f $ unguardedVars     fv }
 mapWRV f fv = fv { weaklyRigidVars   = f $ weaklyRigidVars   fv }
 mapIRV f fv = fv { irrelevantVars    = f $ irrelevantVars    fv }
-mapUUV f fv = fv { unusedVars        = f $ unusedVars        fv }
 
 mapFXV :: (IntMap MetaSet -> IntMap MetaSet) -> FreeVars -> FreeVars
 mapFXV f fv = fv { flexibleVars      = f $ flexibleVars      fv }
@@ -112,7 +112,7 @@ relevantVars fv = Set.unions [rigidVars fv, Map.keysSet (flexibleVars fv)]
 
 -- | @allVars fv@ includes irrelevant variables.
 allVars :: FreeVars -> VarSet
-allVars fv = Set.unions [relevantVars fv, irrelevantVars fv, unusedVars fv]
+allVars fv = Set.unions [relevantVars fv, irrelevantVars fv]
 
 data Occurrence
   = NoOccurrence
@@ -121,7 +121,6 @@ data Occurrence
   | Unguarded         -- ^ In top position, or only under inductive record constructors.
   | WeaklyRigid       -- ^ In arguments to variables and definitions.
   | Flexible MetaSet  -- ^ In arguments of metas.
-  | Unused
   deriving (Eq,Show)
 
 -- | Compute an occurrence of a single variable in a piece of internal syntax.
@@ -136,7 +135,6 @@ occurrenceFV x fv
   | x `Set.member` weaklyRigidVars   fv = WeaklyRigid
   | Just ms <- Map.lookup x (flexibleVars fv) = Flexible ms
   | x `Set.member` irrelevantVars    fv = Irrelevantly
-  | x `Set.member` unusedVars        fv = Unused
   | otherwise                           = NoOccurrence
 
 -- | Mark variables as flexible.  Useful when traversing arguments of metas.
@@ -182,24 +180,17 @@ underConstructor (ConHead c i fs) =
 irrelevantly :: FreeVars -> FreeVars
 irrelevantly fv = empty { irrelevantVars = allVars fv }
 
--- | Mark all free variables as unused, except for irrelevant vars.
-unused :: FreeVars -> FreeVars
-unused fv = empty
-  { irrelevantVars = irrelevantVars fv
-  , unusedVars     = Set.unions [ rigidVars fv, Map.keysSet (flexibleVars fv), unusedVars fv ]
-  }
-
 -- | Pointwise union.
 union :: FreeVars -> FreeVars -> FreeVars
-union (FV sv1 gv1 rv1 fv1 iv1 uv1) (FV sv2 gv2 rv2 fv2 iv2 uv2) =
-  FV (Set.union sv1 sv2) (Set.union gv1 gv2) (Set.union rv1 rv2) (Map.unionWith mappend fv1 fv2) (Set.union iv1 iv2) (Set.union uv1 uv2)
+union (FV sv1 gv1 rv1 fv1 iv1) (FV sv2 gv2 rv2 fv2 iv2) =
+  FV (Set.union sv1 sv2) (Set.union gv1 gv2) (Set.union rv1 rv2) (Map.unionWith mappend fv1 fv2) (Set.union iv1 iv2)
 
 unions :: [FreeVars] -> FreeVars
 unions = foldr union empty
 
 instance Null FreeVars where
-  empty = FV Set.empty Set.empty Set.empty Map.empty Set.empty Set.empty
-  null (FV a b c d e f) = null a && null b && null c && null d && null e && null f
+  empty = FV Set.empty Set.empty Set.empty Map.empty Set.empty
+  null (FV a b c d e) = null a && null b && null c && null d && null e
 
 -- | Free variable sets form a monoid under 'union'.
 instance Semigroup FreeVars where
@@ -212,7 +203,7 @@ instance Monoid FreeVars where
 
 -- | @delete x fv@ deletes variable @x@ from variable set @fv@.
 delete :: Nat -> FreeVars -> FreeVars
-delete n (FV sv gv rv fv iv uv) = FV (Set.delete n sv) (Set.delete n gv) (Set.delete n rv) (Map.delete n fv) (Set.delete n iv) (Set.delete n uv)
+delete n (FV sv gv rv fv iv) = FV (Set.delete n sv) (Set.delete n gv) (Set.delete n rv) (Map.delete n fv) (Set.delete n iv)
 
 instance Singleton Variable FreeVars where
   singleton i = mapUGV (Set.insert i) mempty
@@ -227,8 +218,7 @@ instance IsVarSet FreeVars where
         Free.StronglyRigid -> strongly
       goRel r = case r of
         Relevant   -> id
-        NonStrict  -> id    -- we don't track non-strict and
-        Forced{}   -> id    -- forced in FreeVars
+        NonStrict  -> id    -- we don't track non-strict in FreeVars
         Irrelevant -> irrelevantly
 
 -- In most cases we don't care about the VarOcc.
@@ -238,6 +228,20 @@ instance IsVarSet [Int]  where withVarOcc _ = id
 instance IsVarSet Any    where withVarOcc _ = id
 instance IsVarSet All    where withVarOcc _ = id
 
+newtype VarCounts = VarCounts { varCounts :: IntMap Int }
+
+instance Semigroup VarCounts where
+  VarCounts fv1 <> VarCounts fv2 = VarCounts (Map.unionWith (+) fv1 fv2)
+
+instance Monoid VarCounts where
+  mempty = VarCounts Map.empty
+  mappend = (<>)
+
+instance IsVarSet VarCounts where withVarOcc _ = id
+
+instance Singleton Variable VarCounts where
+  singleton i = VarCounts $ Map.singleton i 1
+
 -- * Collecting free variables.
 
 bench :: a -> a
@@ -285,8 +289,8 @@ newtype RelevantIn a = RelevantIn {getRelevantIn :: a}
 
 instance IsVarSet a => IsVarSet (RelevantIn a) where
   withVarOcc o x
-    | irrelevant (varRelevance o) = mempty
-    | otherwise                   = RelevantIn $ withVarOcc o $ getRelevantIn x
+    | isIrrelevant (varRelevance o) = mempty
+    | otherwise = RelevantIn $ withVarOcc o $ getRelevantIn x
 
 relevantIn' :: Free a => IgnoreSorts -> Nat -> a -> Bool
 relevantIn' ig x t = getAny . getRelevantIn $ runFree (RelevantIn . Any . (x ==)) ig t
@@ -310,7 +314,11 @@ closed t = getAll $ runFree (const $ All False) IgnoreNot t
 allFreeVars :: Free a => a -> VarSet
 allFreeVars = runFree Set.singleton IgnoreNot
 
--- | Collect all relevant free variables, excluding the "unused" ones, possibly ignoring sorts.
+-- | Collect all free variables together with information about their occurrence.
+allFreeVarsWithOcc :: Free a => a -> TheVarMap
+allFreeVarsWithOcc = theVarMap . runFree (singleton . (,topVarOcc)) IgnoreNot
+
+-- | Collect all relevant free variables, possibly ignoring sorts.
 allRelevantVarsIgnoring :: Free a => IgnoreSorts -> a -> VarSet
 allRelevantVarsIgnoring ig = getRelevantIn . runFree (RelevantIn . Set.singleton) ig
 
