@@ -1,3 +1,4 @@
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE UndecidableInstances #-}
 
 -- | Computing the free variables of a term lazily.
@@ -33,7 +34,9 @@ import Control.Monad.Reader
 
 import Data.Foldable (foldMap)
 import Data.IntMap (IntMap)
-import Data.Semigroup (Semigroup, Monoid, (<>), mempty, mappend, mconcat)
+import qualified Data.IntMap as IntMap
+import Data.Monoid ( Monoid, mempty, mappend, mconcat )
+import Data.Semigroup ( Semigroup, (<>) )
 import Data.Set (Set)
 
 import Agda.Syntax.Common
@@ -55,6 +58,8 @@ type MetaSet = Set MetaId
 --   The constructors are listed in increasing order (wrt. information content).
 data FlexRig
   = Flexible MetaSet  -- ^ In arguments of metas.
+                      --   The set of metas is used by ''Agda.TypeChecking.Rewriting.NonLinMatch''
+                      --   to generate the right blocking information.
   | WeaklyRigid       -- ^ In arguments to variables and definitions.
   | Unguarded         -- ^ In top position, or only under inductive record constructors.
   | StronglyRigid     -- ^ Under at least one and only inductive constructors.
@@ -112,6 +117,10 @@ botVarOcc = VarOcc (Flexible mempty) Irrelevant
 composeVarOcc :: VarOcc -> VarOcc -> VarOcc
 composeVarOcc (VarOcc o r) (VarOcc o' r') = VarOcc (composeFlexRig o o') (max r r')
 
+instance LensRelevance VarOcc where
+  getRelevance = varRelevance
+  setRelevance rel (VarOcc x _) = VarOcc x rel
+
 -- | Any representation of a set of variables need to be able to be modified by
 --   a variable occurrence. This is to ensure that free variable analysis is
 --   compositional. For instance, it should be possible to compute `fv (v [u/x])`
@@ -129,10 +138,27 @@ class (Semigroup a, Monoid a) => IsVarSet a where
   --      ```
   withVarOcc :: VarOcc -> a -> a
 
-type VarMap = IntMap VarOcc
+type TheVarMap = IntMap VarOcc
+newtype VarMap = VarMap { theVarMap :: TheVarMap }
+  deriving (Show, Singleton (Variable, VarOcc))
+
+mapVarMap :: (TheVarMap -> TheVarMap) -> VarMap -> VarMap
+mapVarMap f = VarMap . f . theVarMap
+
+instance Semigroup VarMap where
+  VarMap m <> VarMap m' = VarMap $ IntMap.unionWith maxVarOcc m m'
+
+-- Andreas & Jesper, 2018-05-11, issue #3052:
+
+-- | Proper monoid instance for @VarMap@ rather than inheriting the broken one from IntMap.
+--   We combine two occurrences of a variable using 'maxVarOcc'.
+instance Monoid VarMap where
+  mempty  = VarMap IntMap.empty
+  mappend = (<>)
+  mconcat = VarMap . IntMap.unionsWith maxVarOcc . map theVarMap
 
 instance IsVarSet VarMap where
-  withVarOcc o = fmap (composeVarOcc o)
+  withVarOcc o = mapVarMap $ fmap $ composeVarOcc o
 
 -- * Collecting free variables.
 
@@ -261,7 +287,6 @@ instance Free Term where
     Level l      -> freeVars' l
     MetaV m ts   -> go (Flexible $ singleton m) $ freeVars' ts
     DontCare mt  -> goRel Irrelevant $ freeVars' mt
-    Shared p     -> freeVars' (derefPtr p)
 
 instance Free a => Free (Type' a) where
   freeVars' (El s t) =
@@ -277,7 +302,9 @@ instance Free Sort where
       Prop       -> mempty
       Inf        -> mempty
       SizeUniv   -> mempty
-      DLub s1 s2 -> go WeaklyRigid $ freeVars' (s1, s2)
+      PiSort s1 s2 -> go WeaklyRigid $ freeVars' (s1, s2)
+      UnivSort s -> go WeaklyRigid $ freeVars' s
+      MetaS x es -> go (Flexible $ singleton x) $ freeVars' es
 
 instance Free Level where
   freeVars' (Max as) = freeVars' as
