@@ -4,7 +4,6 @@
 -}
 module Agda.TypeChecking.Implicit where
 
-import Control.Applicative
 import Control.Monad
 
 import Agda.Syntax.Common
@@ -44,7 +43,7 @@ implicitNamedArgs :: Int -> (Hiding -> ArgName -> Bool) -> Type -> TCM (NamedArg
 implicitNamedArgs 0 expand t0 = return ([], t0)
 implicitNamedArgs n expand t0 = do
     t0' <- reduce t0
-    case ignoreSharing $ unEl t0' of
+    case unEl t0' of
       Pi (Dom info a) b | let x = absName b, expand (getHiding info) x -> do
           info' <- if hidden info then return info else do
             reportSDoc "tc.term.args.ifs" 15 $
