@@ -3,6 +3,10 @@
 
 module Agda.TypeChecking.InstanceArguments where
 
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ((<>))
+#endif
+
 import Control.Applicative hiding (empty)
 import Control.Monad.Reader
 import Control.Monad.State
@@ -57,13 +61,13 @@ initialIFSCandidates t = do
     getContextVars :: TCM [Candidate]
     getContextVars = do
       ctx <- getContext
-      reportSDoc "tc.instance.cands" 40 $ hang (text "Getting candidates from context") 2 (inTopContext $ prettyTCM ctx)
+      reportSDoc "tc.instance.cands" 40 $ hang (text "Getting candidates from context") 2 (inTopContext $ prettyTCM $ PrettyContext ctx)
           -- Context variables with their types lifted to live in the full context
       let varsAndRaisedTypes = [ (var i, raise (i + 1) t) | (i, t) <- zip [0..] ctx ]
           vars = [ Candidate x t ExplicitStayExplicit (isOverlappable info)
                  | (x, Dom info (_, t)) <- varsAndRaisedTypes
                  , isInstance info
-                 , not (unusableRelevance $ argInfoRelevance info)
+                 , not (unusableRelevance info)
                  ]
 
       -- {{}}-fields of variables are also candidates
@@ -85,7 +89,7 @@ initialIFSCandidates t = do
       let lets = [ Candidate v t ExplicitStayExplicit False
                  | (v, Dom info t) <- env
                  , isInstance info
-                 , not (unusableRelevance $ argInfoRelevance info)
+                 , not (unusableRelevance info)
                  ]
       return $ vars ++ fields ++ lets
 
@@ -171,7 +175,7 @@ findInScope m Nothing = do
     -- potential candidates, so we add them to the context to make
     -- initialIFSCandidates pick them up.
     TelV tel t <- telView t
-    cands <- addContext' tel $ initialIFSCandidates t
+    cands <- addContext tel $ initialIFSCandidates t
     case cands of
       Nothing -> do
         reportSLn "tc.instance" 20 "Can't figure out target of instance goal. Postponing constraint."
@@ -251,8 +255,8 @@ findInScope' m cands = ifM (isFrozen m) (do
 -- | Precondition: type is spine reduced and ends in a Def or a Var.
 insidePi :: Type -> (Type -> TCM a) -> TCM a
 insidePi t ret =
-  case ignoreSharing $ unEl t of
-    Pi a b     -> addContext' (absName b, a) $ insidePi (absBody b) ret
+  case unEl t of
+    Pi a b     -> addContext (absName b, a) $ insidePi (absBody b) ret
     Def{}      -> ret t
     Var{}      -> ret t
     Sort{}     -> __IMPOSSIBLE__
@@ -261,7 +265,6 @@ insidePi t ret =
     Lit{}      -> __IMPOSSIBLE__
     Level{}    -> __IMPOSSIBLE__
     MetaV{}    -> __IMPOSSIBLE__
-    Shared{}   -> __IMPOSSIBLE__
     DontCare{} -> __IMPOSSIBLE__
 
 -- | A meta _M is rigidly constrained if there is a constraint _M us == D vs,
@@ -274,7 +277,7 @@ rigidlyConstrainedMetas = do
   where
     isRigid v = do
       bv <- reduceB v
-      case ignoreSharing <$> bv of
+      case bv of
         Blocked{}    -> return False
         NotBlocked _ v -> case v of
           MetaV{}    -> return False
@@ -287,11 +290,10 @@ rigidlyConstrainedMetas = do
           Level{}    -> return False
           DontCare{} -> return False
           Lam{}      -> __IMPOSSIBLE__
-          Shared{}   -> __IMPOSSIBLE__
     rigidMetas c =
       case clValue $ theConstraint c of
         ValueCmp _ _ u v ->
-          case (ignoreSharing u, ignoreSharing v) of
+          case (u, v) of
             (MetaV m us, _) | isJust (allApplyElims us) -> ifM (isRigid v) (return $ Just m) (return Nothing)
             (_, MetaV m vs) | isJust (allApplyElims vs) -> ifM (isRigid u) (return $ Just m) (return Nothing)
             _              -> return Nothing
@@ -305,6 +307,9 @@ rigidlyConstrainedMetas = do
         IsEmpty{}     -> return Nothing
         CheckSizeLtSat{} -> return Nothing
         FindInScope{} -> return Nothing
+        CheckFunDef{} -> return Nothing
+        HasBiggerSort{} -> return Nothing
+        HasPTSRule{}  -> return Nothing
 
 isRigid :: MetaId -> TCM Bool
 isRigid i = do
@@ -315,7 +320,7 @@ isRigid i = do
 --   constrained. Note that level metas are never considered rigidly constrained
 --   (#1865).
 areThereNonRigidMetaArguments :: Term -> TCM (Maybe MetaId)
-areThereNonRigidMetaArguments t = case ignoreSharing t of
+areThereNonRigidMetaArguments t = case t of
     Def n args -> do
       TelV tel _ <- telView . defType =<< getConstInfo n
       let varOccs EmptyTel           = []
@@ -339,7 +344,6 @@ areThereNonRigidMetaArguments t = case ignoreSharing t of
     Level{}  -> __IMPOSSIBLE__
     MetaV{}  -> __IMPOSSIBLE__
     Pi{}     -> __IMPOSSIBLE__
-    Shared{} -> __IMPOSSIBLE__
     DontCare{} -> __IMPOSSIBLE__
   where
     areThereNonRigidMetaArgs :: Elims -> TCM (Maybe MetaId)
@@ -350,10 +354,10 @@ areThereNonRigidMetaArguments t = case ignoreSharing t of
 
     isNonRigidMeta :: Term -> TCM (Maybe MetaId)
     isNonRigidMeta v =
-      case ignoreSharing v of
+      case v of
         Def _ es  -> areThereNonRigidMetaArgs es
         Var _ es  -> areThereNonRigidMetaArgs es
-        Con _ _ vs-> areThereNonRigidMetaArgs (map Apply vs)
+        Con _ _ vs-> areThereNonRigidMetaArgs vs
         MetaV i _ -> ifM (isRigid i) (return Nothing) $ do
           -- Ignore unconstrained level and size metas (#1865)
           mlvl <- getBuiltinDefName builtinLevel
@@ -468,7 +472,7 @@ checkCandidates m t cands = disableDestructiveUpdate $
     anyMetaTypes [] = return False
     anyMetaTypes (Candidate _ a _ _ : cands) = do
       a <- instantiate a
-      case ignoreSharing $ unEl a of
+      case unEl a of
         MetaV{} -> return True
         _       -> anyMetaTypes cands
 
@@ -574,11 +578,11 @@ isIFSConstraint _             = False
 applyDroppingParameters :: Term -> Args -> TCM Term
 applyDroppingParameters t vs = do
   let fallback = return $ t `apply` vs
-  case ignoreSharing t of
+  case t of
     Con c ci [] -> do
       def <- theDef <$> getConInfo c
       case def of
-        Constructor {conPars = n} -> return $ Con c ci (drop n vs)
+        Constructor {conPars = n} -> return $ Con c ci (map Apply $ drop n vs)
         _ -> __IMPOSSIBLE__
     Def f [] -> do
       mp <- isProjection f
