@@ -1,5 +1,6 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE NondecreasingIndentation #-}
+{-# LANGUAGE GADTs #-}
 
 module Agda.TypeChecking.MetaVars where
 
@@ -8,9 +9,11 @@ import Prelude hiding (null)
 import Control.Monad.Reader
 
 import Data.Function
+import qualified Data.IntMap as IntMap
 import qualified Data.List as List
 import qualified Data.Map as Map
 import qualified Data.Foldable as Fold
+import qualified Data.Traversable as Trav
 
 import Agda.Syntax.Abstract.Name as A
 import Agda.Syntax.Common
@@ -21,6 +24,7 @@ import Agda.Syntax.Position (killRange)
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Reduce
+import Agda.TypeChecking.Sort
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Constraints
@@ -43,6 +47,7 @@ import Agda.Utils.Except
   , runExceptT
   )
 
+import Agda.Utils.Function
 import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.Maybe
@@ -136,7 +141,10 @@ assignTerm' x tel v = do
 
 -- | Create a sort meta that cannot be instantiated with 'Inf' (Setω).
 newSortMetaBelowInf :: TCM Sort
-newSortMetaBelowInf = newSortMeta' $ HasType ()
+newSortMetaBelowInf = do
+  x <- newSortMeta
+  hasBiggerSort x
+  return x
 
 -- | Create a sort meta that may be instantiated with 'Inf' (Setω).
 newSortMeta :: TCM Sort
@@ -148,9 +156,8 @@ newSortMeta' judge =
   ifM hasUniversePolymorphism (newSortMetaCtx' judge =<< getContextArgs)
   -- else (no universe polymorphism)
   $ do i   <- createMetaInfo
-       lvl <- levelType
-       x   <- newMeta i normalMetaPriority (idP 0) $ judge lvl
-       return $ Type $ Max [Plus 0 $ MetaLevel x []]
+       x   <- newMeta i normalMetaPriority (idP 0) $ judge typeDontCare
+       return $ MetaS x []
 
 -- | Create a sort meta that may be instantiated with 'Inf' (Setω).
 newSortMetaCtx :: Args -> TCM Sort
@@ -161,12 +168,11 @@ newSortMetaCtx' judge vs = do
   ifM typeInType (return $ mkType 0) $ {- else -} do
     i   <- createMetaInfo
     tel <- getContextTelescope
-    lvl <- levelType
-    let t = telePi_ tel lvl
+    let t = telePi_ tel typeDontCare
     x   <- newMeta i normalMetaPriority (idP 0) $ judge t
     reportSDoc "tc.meta.new" 50 $
       text "new sort meta" <+> prettyTCM x <+> text ":" <+> prettyTCM t
-    return $ Type $ Max [Plus 0 $ MetaLevel x $ map Apply vs]
+    return $ MetaS x $ map Apply vs
 
 newTypeMeta :: Sort -> TCM Type
 newTypeMeta s = El s . snd <$> newValueMeta RunMetaOccursCheck (sort s)
@@ -192,7 +198,8 @@ newIFSMetaCtx s t vs = do
     [ text "new ifs meta:"
     , nest 2 $ prettyTCM vs <+> text "|-"
     ]
-  i0 <- createMetaInfo
+  -- Andreas, 2017-10-04, issue #2753: no metaOccurs check for instance metas
+  i0 <- createMetaInfo' DontRunMetaOccursCheck
   let i = i0 { miNameSuggestion = s }
   TelV tel _ <- telView t
   let perm = idP (size tel)
@@ -253,7 +260,7 @@ newValueMetaCtx' b a tel perm vs = do
     ]
   etaExpandMetaSafe x
   -- Andreas, 2012-09-24: for Metas X : Size< u add constraint X+1 <= u
-  u <- shared $ MetaV x $ map Apply vs
+  let u = MetaV x $ map Apply vs
   boundedSizeMetaHook u tel a
   return (x, u)
 
@@ -280,15 +287,20 @@ newArgsMetaCtx = newArgsMetaCtx' trueCondition
 newArgsMetaCtx' :: Condition -> Type -> Telescope -> Permutation -> Args -> TCM Args
 newArgsMetaCtx' condition (El s tm) tel perm ctx = do
   tm <- reduce tm
-  case ignoreSharing tm of
+  case tm of
     Pi dom@(Dom info a) codom | condition dom codom -> do
+      let r    = getRelevance info
+          -- Issue #3031: It's not enough to applyRelevanceToContext, since most (all?)
+          -- of the context lives in tel. Don't forget the arguments in ctx.
+          tel' = telFromList . map (inverseApplyRelevance r) . telToList $ tel
+          ctx' = (map . mapRelevance) (r `inverseComposeRelevance`) ctx
       (_, u) <- applyRelevanceToContext (getRelevance info) $
                {-
                  -- Andreas, 2010-09-24 skip irrelevant record fields when eta-expanding a meta var
                  -- Andreas, 2010-10-11 this is WRONG, see Issue 347
                 if r == Irrelevant then return DontCare else
                 -}
-                 newValueMetaCtx RunMetaOccursCheck a tel perm ctx
+                 newValueMetaCtx RunMetaOccursCheck a tel' perm ctx'
       args <- newArgsMetaCtx' condition (codom `absApp` u) tel perm ctx
       return $ Arg info u : args
     _  -> return []
@@ -306,7 +318,7 @@ newRecordMetaCtx r pars tel perm ctx = do
   ftel   <- flip apply pars <$> getRecordFieldTypes r
   fields <- newArgsMetaCtx (telePi_ ftel $ sort Prop) tel perm ctx
   con    <- getRecordConstructor r
-  return $ Con con ConOSystem fields
+  return $ Con con ConOSystem (map Apply fields)
 
 newQuestionMark :: InteractionId -> Type -> TCM (MetaId, Term)
 newQuestionMark = newQuestionMark' $ newValueMeta' RunMetaOccursCheck
@@ -371,7 +383,7 @@ blockTypeOnProblem (El s a) pid = El s <$> blockTermOnProblem (El Inf $ Sort s)
 --
 --   Auxiliary function to create a postponed type checking problem.
 unblockedTester :: Type -> TCM Bool
-unblockedTester t = ifBlockedType t (\ m t -> return False) (\ t -> return True)
+unblockedTester t = ifBlockedType t (\ m t -> return False) (\ _ t -> return True)
 
 -- | Create a postponed type checking problem @e : t@ that waits for type @t@
 --   to unblock (become instantiated or its constraints resolved).
@@ -471,51 +483,53 @@ etaExpandMeta kinds m = whenM (asks envAssignMetas `and2M` isEtaExpandable kinds
           reportSDoc "tc.meta.eta" 20 $ do
             text "we do not expand meta variable" <+> prettyTCM m <+>
               text ("(requested was expansion of " ++ show kinds ++ ")")
-    meta           <- lookupMeta m
-    let HasType _ a = mvJudgement meta
-    TelV tel b     <- telView a
-    -- if the target type @b@ of @m@ is a meta variable @x@ itself
-    -- (@NonBlocked (MetaV{})@),
-    -- or it is blocked by a meta-variable @x@ (@Blocked@), we cannot
-    -- eta expand now, we have to postpone this.  Once @x@ is
-    -- instantiated, we can continue eta-expanding m.  This is realized
-    -- by adding @m@ to the listeners of @x@.
-    ifBlocked (unEl b) (\ x _ -> waitFor x) $ \ t -> case ignoreSharing t of
-      lvl@(Def r es) ->
-        ifM (isEtaRecord r) {- then -} (do
-          let ps = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
-          let expand = do
-                u <- withMetaInfo' meta $ newRecordMetaCtx r ps tel (idP $ size tel) $ teleArgs tel
-                inTopContext $ do
-                  verboseS "tc.meta.eta" 15 $ do
-                    du <- prettyTCM u
-                    reportSDoc "tc.meta.eta" 15 $ sep
-                      [ text "eta expanding: " <+> pretty m <+> text " --> "
-                      , nest 2 $ prettyTCM u
-                      ]
-                  -- Andreas, 2012-03-29: No need for occurrence check etc.
-                  -- we directly assign the solution for the meta
-                  -- 2012-05-23: We also bypass the check for frozen.
-                  noConstraints $ assignTerm' m (telToArgs tel) u  -- should never produce any constraints
-          if Records `elem` kinds then
-            expand
-           else if (SingletonRecords `elem` kinds) then do
-             singleton <- isSingletonRecord r ps
-             case singleton of
-               Left x      -> waitFor x
-               Right False -> dontExpand
-               Right True  -> expand
-            else dontExpand
-        ) $ {- else -} ifM (andM [ return $ Levels `elem` kinds
-                        , typeInType
-                        , (Just lvl ==) <$> getBuiltin' builtinLevel
-                        ]) (do
-          reportSLn "tc.meta.eta" 20 $ "Expanding level meta to 0 (type-in-type)"
-          -- Andreas, 2012-03-30: No need for occurrence check etc.
-          -- we directly assign the solution for the meta
-          noConstraints $ assignTerm m (telToArgs tel) (Level $ Max [])
-       ) $ {- else -} dontExpand
-      _ -> dontExpand
+    meta <- lookupMeta m
+    case mvJudgement meta of
+      IsSort{} -> dontExpand
+      HasType _ a -> do
+        TelV tel b <- telView a
+        -- if the target type @b@ of @m@ is a meta variable @x@ itself
+        -- (@NonBlocked (MetaV{})@),
+        -- or it is blocked by a meta-variable @x@ (@Blocked@), we cannot
+        -- eta expand now, we have to postpone this.  Once @x@ is
+        -- instantiated, we can continue eta-expanding m.  This is realized
+        -- by adding @m@ to the listeners of @x@.
+        ifBlocked (unEl b) (\ x _ -> waitFor x) $ \ _ t -> case t of
+          lvl@(Def r es) ->
+            ifM (isEtaRecord r) {- then -} (do
+              let ps = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
+              let expand = do
+                    u <- withMetaInfo' meta $ newRecordMetaCtx r ps tel (idP $ size tel) $ teleArgs tel
+                    inTopContext $ do
+                      verboseS "tc.meta.eta" 15 $ do
+                        du <- prettyTCM u
+                        reportSDoc "tc.meta.eta" 15 $ sep
+                          [ text "eta expanding: " <+> pretty m <+> text " --> "
+                          , nest 2 $ prettyTCM u
+                          ]
+                      -- Andreas, 2012-03-29: No need for occurrence check etc.
+                      -- we directly assign the solution for the meta
+                      -- 2012-05-23: We also bypass the check for frozen.
+                      noConstraints $ assignTerm' m (telToArgs tel) u  -- should never produce any constraints
+              if Records `elem` kinds then
+                expand
+               else if (SingletonRecords `elem` kinds) then do
+                 singleton <- isSingletonRecord r ps
+                 case singleton of
+                   Left x      -> waitFor x
+                   Right False -> dontExpand
+                   Right True  -> expand
+                else dontExpand
+            ) $ {- else -} ifM (andM [ return $ Levels `elem` kinds
+                            , typeInType
+                            , (Just lvl ==) <$> getBuiltin' builtinLevel
+                            ]) (do
+              reportSLn "tc.meta.eta" 20 $ "Expanding level meta to 0 (type-in-type)"
+              -- Andreas, 2012-03-30: No need for occurrence check etc.
+              -- we directly assign the solution for the meta
+              noConstraints $ assignTerm m (telToArgs tel) (Level $ Max [])
+           ) $ {- else -} dontExpand
+          _ -> dontExpand
 
 -- | Eta expand blocking metavariables of record type, and reduce the
 -- blocked thing.
@@ -589,9 +603,9 @@ assign dir x args v = do
     text "MetaVars.assign: assigning to " <+> prettyTCM v
 
   reportSLn "tc.meta.assign" 75 $
-    "MetaVars.assign: assigning to " ++ show v
+    "MetaVars.assign: assigning meta  " ++ show x ++ "  with args  " ++ show args ++ "  to  " ++ show v
 
-  case (ignoreSharing v, mvJudgement mvar) of
+  case (v, mvJudgement mvar) of
       (Sort Inf, HasType{}) -> typeError SetOmegaNotValidType
       _                     -> return ()
 
@@ -647,14 +661,16 @@ assign dir x args v = do
       -- e.g. _1 x (suc x) = suc (_2 x y)
       -- even though the lhs is not a pattern, we can prune the y from _2
 
-      (relVL, irrVL) <- do
+      (relVL, nonstrictVL, irrVL) <- do
         -- Andreas, 2016-11-03 #2211 attempt to do s.th. for unused
         if False -- irrelevant $ getMetaRelevance mvar
           then do
             reportSDoc "tc.meta.assign" 25 $ text "meta is irrelevant or unused"
-            return (Set.toList $ allFreeVars args, empty)
+            return (Set.toList $ allFreeVars args, empty, empty)
           else do
-            let relVL = Set.toList $ allRelevantVars args
+            let vars  = allFreeVarsWithOcc args
+                relVL       = IntMap.keys $ IntMap.filter isRelevant vars
+                nonstrictVL = IntMap.keys $ IntMap.filter isNonStrict vars
             -- Andreas, 2011-10-06 only irrelevant vars that are direct
             -- arguments to the meta, hence, can be abstracted over, may
             -- appear on the rhs.  (test/fail/Issue483b)
@@ -662,13 +678,12 @@ assign dir x args v = do
             let fromIrrVar (Var i [])   = return [i]
                 fromIrrVar (Con c _ vs)   =
                   ifM (isNothing <$> isRecordConstructor (conName c)) (return []) $
-                    concat <$> mapM (fromIrrVar . {- stripDontCare .-} unArg) vs
-                fromIrrVar (Shared p)   = fromIrrVar (derefPtr p)
+                    concat <$> mapM (fromIrrVar . {- stripDontCare .-} unArg) (fromMaybe __IMPOSSIBLE__ (allApplyElims vs))
                 fromIrrVar _ = return []
             irrVL <- concat <$> mapM fromIrrVar
                        [ v | Arg info v <- args, isIrrelevant info ]
                           -- irrelevant (getRelevance info) ]
-            return (relVL, irrVL)
+            return (relVL, nonstrictVL, irrVL)
       reportSDoc "tc.meta.assign" 20 $
           let pr (Var n []) = text (show n)
               pr (Def c []) = prettyTCM c
@@ -676,6 +691,7 @@ assign dir x args v = do
           in vcat
                [ text "mvar args:" <+> sep (map (pr . unArg) args)
                , text "fvars lhs (rel):" <+> sep (map (text . show) relVL)
+               , text "fvars lhs (nonstrict):" <+> sep (map (text . show) nonstrictVL)
                , text "fvars lhs (irr):" <+> sep (map (text . show) irrVL)
                ]
 
@@ -684,7 +700,7 @@ assign dir x args v = do
       -- Herein, distinguish relevant and irrelevant vars,
       -- since when abstracting irrelevant lhs vars, they may only occur
       -- irrelevantly on rhs.
-      v <- liftTCM $ occursCheck x (relVL, irrVL) v
+      v <- liftTCM $ occursCheck x (relVL, nonstrictVL, irrVL) v
 
       reportSLn "tc.meta.assign" 15 "passed occursCheck"
       verboseS "tc.meta.assign" 30 $ do
@@ -738,6 +754,12 @@ assign dir x args v = do
           m <- getContextSize
           assignMeta' m x t (length args) ids v
   where
+    -- | Try to remove meta arguments from lhs that mention variables not occurring on rhs.
+    attemptPruning
+      :: MetaId  -- ^ Meta-variable (lhs)
+      -> Args    -- ^ Meta arguments (lhs)
+      -> FVs     -- ^ Variables occuring on the rhs
+      -> TCM a
     attemptPruning x args fvs = do
       -- non-linear lhs: we cannot solve, but prune
       killResult <- prune x args $ Set.toList fvs
@@ -800,7 +822,7 @@ attemptInertRHSImprovement m args v = do
                 reportSDoc "tc.meta.inert" 30 $ nest 2 $ text "can't do projections from inert"
                 patternViolation
               Just args -> return args
-      case ignoreSharing v of
+      case v of
         Var x elims -> (, Var x . map Apply) <$> typeOfBV x
         Con c ci args  -> notInert -- (, Con c ci) <$> defType <$> getConstInfo (conName c)
         Def f elims -> do
@@ -821,7 +843,6 @@ attemptInertRHSImprovement m args v = do
         Level{}    -> notInert
         MetaV{}    -> notInert
         DontCare{} -> notInert
-        Shared{}   -> __IMPOSSIBLE__
 
     ensureNeutral :: Term -> Term -> TCM ()
     ensureNeutral rhs v = do
@@ -834,7 +855,7 @@ attemptInertRHSImprovement m args v = do
               reportSDoc "tc.meta.inert" 30 $ nest 2 $ text "argument shares head with RHS:" <+> prettyTCM arg
               patternViolation
             | otherwise  = return ()
-      case fmap ignoreSharing b of
+      case b of
         Blocked{}      -> notNeutral v
         NotBlocked r v ->                      -- Andrea(s) 2014-12-06 can r be useful?
           case v of
@@ -848,7 +869,6 @@ attemptInertRHSImprovement m args v = do
             MetaV{}    -> notNeutral v
             Con{}      -> notNeutral v
             Lam{}      -> notNeutral v
-            Shared{}   -> __IMPOSSIBLE__
 -- END UNUSED -}
 
 -- | @assignMeta m x t ids u@ solves @x ids = u@ for meta @x@ of type @t@,
@@ -951,7 +971,7 @@ subtypingForSizeLt dir   x mvar t args v cont = do
     caseMaybe mSizeLt fallback $ \ qSizeLt -> do
       -- Check whether v is a SIZELT
       v <- reduce v
-      case ignoreSharing v of
+      case v of
         Def q [Apply (Arg ai u)] | q == qSizeLt -> do
           -- Clone the meta into a new size meta @y@.
           -- To this end, we swap the target of t for Size.
@@ -974,13 +994,21 @@ subtypingForSizeLt dir   x mvar t args v cont = do
         _ -> fallback
 
 -- | Eta-expand bound variables like @z@ in @X (fst z)@.
-expandProjectedVars :: (Normalise a, TermLike a, Show a, PrettyTCM a, NoProjectedVar a,
-                        Subst Term a, PrettyTCM b, Subst Term b) =>
-  a -> b -> (a -> b -> TCM c) -> TCM c
+expandProjectedVars
+  :: ( Show a, PrettyTCM a, NoProjectedVar a
+     -- , Normalise a, TermLike a, Subst Term a
+     , ReduceAndEtaContract a
+     , PrettyTCM b, Subst Term b
+     )
+  => a  -- ^ Meta variable arguments.
+  -> b  -- ^ Right hand side.
+  -> (a -> b -> TCM c)
+  -> TCM c
 expandProjectedVars args v ret = loop (args, v) where
   loop (args, v) = do
     reportSDoc "tc.meta.assign.proj" 45 $ text "meta args: " <+> prettyTCM args
-    args <- etaContract =<< normalise args
+    args <- callByName $ reduceAndEtaContract args
+    -- args <- etaContract =<< normalise args
     reportSDoc "tc.meta.assign.proj" 45 $ text "norm args: " <+> prettyTCM args
     reportSDoc "tc.meta.assign.proj" 85 $ text "norm args: " <+> text (show args)
     let done = ret args v
@@ -1011,12 +1039,12 @@ data ProjVarExc = ProjVarExc Int [(ProjOrigin, QName)]
 
 instance NoProjectedVar Term where
   noProjectedVar t =
-    case ignoreSharing t of
+    case t of
       Var i es
         | qs@(_:_) <- takeWhileJust id $ map isProjElim es -> Left $ ProjVarExc i qs
       -- Andreas, 2015-09-12 Issue #1316:
       -- Also look in inductive record constructors
-      Con (ConHead _ Inductive (_:_)) _ vs -> noProjectedVar vs
+      Con (ConHead _ Inductive (_:_)) _ es | Just vs <- allApplyElims es -> noProjectedVar vs
       _ -> return ()
 
 instance NoProjectedVar a => NoProjectedVar (Arg a) where
@@ -1026,6 +1054,29 @@ instance NoProjectedVar a => NoProjectedVar [a] where
   noProjectedVar = Fold.mapM_ noProjectedVar
 
 
+-- | Normalize just far enough to be able to eta-contract maximally.
+class (TermLike a, Subst Term a, Reduce a) => ReduceAndEtaContract a where
+  reduceAndEtaContract :: a -> TCM a
+
+  default reduceAndEtaContract
+    :: (Traversable f, TermLike b, Subst Term b, Reduce b, ReduceAndEtaContract b, f b ~ a)
+    => a -> TCM a
+  reduceAndEtaContract = Trav.mapM reduceAndEtaContract
+
+instance ReduceAndEtaContract a => ReduceAndEtaContract [a] where
+instance ReduceAndEtaContract a => ReduceAndEtaContract (Arg a) where
+
+instance ReduceAndEtaContract Term where
+  reduceAndEtaContract u = do
+    reduce u >>= \case
+      -- In case of lambda or record constructor, it makes sense to
+      -- reduce further.
+      Lam ai (Abs x b) -> etaLam ai x =<< reduceAndEtaContract b
+      Con c ci es | Just args <- allApplyElims es -> etaCon c ci args $ \ r c ci args -> do
+        args <- reduceAndEtaContract args
+        etaContractRecord r c ci args
+      v -> return v
+
 {- UNUSED, BUT KEEP!
 -- Wrong attempt at expanding bound variables.
 -- The following code curries meta instead.
@@ -1051,7 +1102,7 @@ etaExpandProjectedVar mvar x t n qs = inTopContext $ do
 {-
   -- first, strip the leading n domains (which remain unchanged)
   TelV gamma core <- telViewUpTo n t
-  case ignoreSharing $ unEl core of
+  case unEl core of
     -- There should be at least one domain left
     Pi (Dom ai a) b -> do
       -- Eta-expand @dom@ along @qs@ into a telescope @tel@, computing a substitution.
@@ -1133,7 +1184,7 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
 
     isVarOrIrrelevant :: Res -> (Arg Term, Term) -> ExceptT InvertExcept TCM Res
     isVarOrIrrelevant vars (arg, t) =
-      case ignoreSharing <$> arg of
+      case arg of
         -- i := x
         Arg info (Var i []) -> return $ (Arg info i, t) `cons` vars
 
@@ -1143,20 +1194,25 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
 
         -- (i, j) := x  becomes  [i := fst x, j := snd x]
         -- Andreas, 2013-09-17 but only if constructor is fully applied
-        Arg info (Con c ci vs) -> do
+        Arg info (Con c ci es) -> do
           let fallback
                | isIrrelevant info = return vars
                | otherwise                              = failure
           isRC <- lift $ isRecordConstructor $ conName c
           case isRC of
             Just (_, Record{ recFields = fs })
-              | length fs == length vs -> do
+              | length fs == length es -> do
                 let aux (Arg _ v) (Arg info' f) = (Arg ai v,) $ t `applyE` [Proj ProjSystem f] where
                      ai = ArgInfo
-                       { argInfoHiding       = min (getHiding info) (getHiding info')
-                       , argInfoRelevance    = max (getRelevance info) (getRelevance info')
-                       , argInfoOrigin       = min (getOrigin info) (getOrigin info')
+                       { argInfoHiding   = min (getHiding info) (getHiding info')
+                       , argInfoModality = Modality
+                         { modRelevance  = max (getRelevance info) (getRelevance info')
+                         , modQuantity   = max (getQuantity  info) (getQuantity  info')
+                         }
+                       , argInfoOrigin   = min (getOrigin info) (getOrigin info')
+                       , argInfoFreeVariables = unknownFreeVariables
                        }
+                    vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
                 res <- loop $ zipWith aux vs fs
                 return $ res `append` vars
               | otherwise -> fallback
@@ -1181,21 +1237,17 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
         Arg _ Sort{}     -> neutralArg
         Arg _ Level{}    -> neutralArg
 
-        Arg info (Shared p) -> isVarOrIrrelevant vars (Arg info $ derefPtr p, t)
-
     -- managing an assoc list where duplicate indizes cannot be irrelevant vars
     append :: Res -> Res -> Res
     append res vars = foldr cons vars res
 
     -- adding an irrelevant entry only if not present
     cons :: (Arg Nat, Term) -> Res -> Res
-    cons a@(Arg (ArgInfo _ Irrelevant _) i, t) vars
-      | any ((i==) . unArg . fst) vars  = vars
-      | otherwise                       = a : vars
-    -- adding a relevant entry:
-    cons a@(Arg info i, t) vars = a :
-      -- filter out duplicate irrelevants
-      filter (not . (\ a@(Arg info j, t) -> isIrrelevant info && i == j)) vars
+    cons a@(Arg ai i, t) vars
+      | isIrrelevant ai = applyUnless (any ((i==) . unArg . fst) vars) (a :) vars
+      | otherwise       = a :  -- adding a relevant entry
+          -- filter out duplicate irrelevants
+          filter (not . (\ a@(Arg info j, t) -> isIrrelevant info && i == j)) vars
 
 -- UNUSED
 -- -- | Used in 'Agda.Interaction.BasicOps.giveExpr'.
