@@ -20,7 +20,6 @@ instance MentionsMeta Term where
     Level l      -> mm l
     DontCare v   -> False   -- we don't have to look inside don't cares when deciding to wake constraints
     MetaV y args -> x == y || mm args   -- TODO: we really only have to look one level deep at meta args
-    Shared p     -> mm $ derefPtr p
     where
       mm v = mentionsMeta x v
 
@@ -47,7 +46,9 @@ instance MentionsMeta Sort where
     Prop       -> False
     Inf        -> False
     SizeUniv   -> False
-    DLub s1 s2 -> mentionsMeta x (s1, s2)
+    PiSort s1 s2 -> mentionsMeta x (s1, s2)
+    UnivSort s -> mentionsMeta x s
+    MetaS m es -> x == m || mentionsMeta x es
 
 instance MentionsMeta t => MentionsMeta (Abs t) where
   mentionsMeta x = mentionsMeta x . unAbs
@@ -89,7 +90,7 @@ instance MentionsMeta ProblemConstraint where
 instance MentionsMeta Constraint where
   mentionsMeta x c = case c of
     ValueCmp _ t u v    -> mm (t, u, v)
-    ElimCmp _ t v as bs -> mm ((t, v), (as, bs))
+    ElimCmp _ _ t v as bs -> mm ((t, v), (as, bs))
     LevelCmp _ u v      -> mm (u, v)
     TypeCmp _ a b       -> mm (a, b)
     TelCmp a b _ u v    -> mm ((a, b), (u, v))
@@ -101,6 +102,9 @@ instance MentionsMeta Constraint where
     FindInScope{}       -> True   -- this needs to be woken up for any meta
     IsEmpty r t         -> mm t
     CheckSizeLtSat t    -> mm t
+    CheckFunDef{}       -> True   -- not sure what metas this depends on
+    HasBiggerSort a     -> mm a
+    HasPTSRule a b      -> mm (a, b)
     where
       mm v = mentionsMeta x v
 
