@@ -16,7 +16,6 @@
 
 module Agda.TypeChecking.MetaVars.Occurs where
 
-import Control.Applicative
 import Control.Monad
 import Control.Monad.Reader
 
@@ -134,7 +133,6 @@ data OccursCtx
   = Flex          -- ^ We are in arguments of a meta.
   | Rigid         -- ^ We are not in arguments of a meta but a bound var.
   | StronglyRigid -- ^ We are at the start or in the arguments of a constructor.
-  | Top           -- ^ We are at the term root (this turns into @StronglyRigid@).
   | Irrel         -- ^ We are in an irrelevant argument.
   deriving (Eq, Show)
 
@@ -149,14 +147,8 @@ unfold :: UnfoldStrategy -> Term -> TCM (Blocked Term)
 unfold NoUnfold  v = notBlocked <$> instantiate v
 unfold YesUnfold v = reduceB v
 
--- | Leave the top position.
-leaveTop :: OccursCtx -> OccursCtx
-leaveTop Top = StronglyRigid
-leaveTop ctx = ctx
-
 -- | Leave the strongly rigid position.
 weakly :: OccursCtx -> OccursCtx
-weakly Top           = Rigid
 weakly StronglyRigid = Rigid
 weakly ctx = ctx
 
@@ -170,26 +162,32 @@ patternViolation' n err = do
   patternViolation
 
 abort :: OccursCtx -> TypeError -> TCM a
-abort Top           err = typeError err
 abort StronglyRigid err = typeError err -- here, throw an uncatchable error (unsolvable constraint)
 abort Flex          err = patternViolation' 70 (show err) -- throws a PatternErr, which leads to delayed constraint
 abort Rigid         err = patternViolation' 70 (show err)
 abort Irrel         err = patternViolation' 70 (show err)
 
--- | Distinguish relevant and irrelevant variables in occurs check.
-type Vars = ([Nat],[Nat])
+-- | Distinguish relevant, irrelevant and nonstrict variables in occurs check.
+type Vars = ([Nat],[Nat],[Nat])
+-- TODO: refactor this into an actual datatype
 
 goIrrelevant :: Vars -> Vars
-goIrrelevant (relVs, irrVs) = (irrVs ++ relVs, [])
+goIrrelevant (relVs, nonstrictVs, irrVs) = (irrVs ++ nonstrictVs ++ relVs, [], [])
+
+goNonStrict :: Vars -> Vars
+goNonStrict (relVs, nonstrictVs, irrVs) = (nonstrictVs ++ relVs, [], irrVs)
 
 allowedVar :: Nat -> Vars -> Bool
-allowedVar i (relVs, irrVs) = i `elem` relVs
+allowedVar i (relVs, nonstrictVs, irrVs) = i `elem` relVs
 
 takeRelevant :: Vars -> [Nat]
-takeRelevant = fst
+takeRelevant (relVs, nonstrictVs, irrVs) = relVs
+
+takeAll :: Vars -> [Nat]
+takeAll (rel, nst, irr) = rel ++ nst ++ irr
 
 liftUnderAbs :: Vars -> Vars
-liftUnderAbs (relVs, irrVs) = (0 : map (1+) relVs, map (1+) irrVs)
+liftUnderAbs (relVs, nonstrictVs, irrVs) = (0 : map (1+) relVs, map (+1) nonstrictVs, map (1+) irrVs)
 
 -- | Extended occurs check.
 class Occurs t where
@@ -203,13 +201,14 @@ occursCheck
   => MetaId -> Vars -> a -> TCM a
 occursCheck m xs v = disableDestructiveUpdate $ Bench.billTo [ Bench.Typing, Bench.OccursCheck ] $ do
   mv <- lookupMeta m
+  let ctx = if isIrrelevant (getMetaRelevance mv) then Irrel else StronglyRigid
   initOccursCheck mv
       -- TODO: Can we do this in a better way?
   let redo m = m -- disableDestructiveUpdate m >> m
   -- First try without normalising the term
-  redo (occurs NoUnfold  Top m xs v) `catchError` \_ -> do
+  redo (occurs NoUnfold  ctx m xs v) `catchError` \_ -> do
     initOccursCheck mv
-    redo (occurs YesUnfold Top m xs v) `catchError` \err -> case err of
+    redo (occurs YesUnfold ctx m xs v) `catchError` \err -> case err of
                             -- Produce nicer error messages
       TypeError _ cl -> case clValue cl of
         MetaOccursInItself{} ->
@@ -233,6 +232,12 @@ occursCheck m xs v = disableDestructiveUpdate $ Bench.billTo [ Bench.Typing, Ben
                    , text $ "which is not in scope of the metavariable or irrelevant in the metavariable but relevant in the solution"
                    ]
             )
+        MetaIrrelevantSolution _ _ ->
+          typeError . GenericError . show =<<
+            fsep [ text ("Cannot instantiate the metavariable " ++ prettyShow m ++ " to solution")
+                 , prettyTCM v
+                 , text "since (part of) the solution was created in an irrelevant context."
+                 ]
         _ -> throwError err
       _ -> throwError err
 
@@ -269,15 +274,22 @@ instance Occurs Term where
                   abort (strongly ctx) $ MetaCannotDependOn m (takeRelevant xs) i
                 -- is a singleton type with unique inhabitant sv
                 Right (Just sv) -> return $ sv `applyE` es
-          Lam h f     -> Lam h <$> occ (leaveTop ctx) f
-          Level l     -> Level <$> occ ctx l  -- stay in Top
+          Lam h f     -> Lam h <$> occ ctx f
+          Level l     -> Level <$> occ ctx l
           Lit l       -> return v
-          DontCare v  -> dontCare <$> occurs red Irrel m (goIrrelevant xs) v
-          Def d es    -> Def d <$> occDef d (leaveTop ctx) es
-          Con c ci vs -> Con c ci <$> occ (leaveTop ctx) vs  -- if strongly rigid, remain so
-          Pi a b      -> uncurry Pi <$> occ (leaveTop ctx) (a,b)
-          Sort s      -> Sort <$> occ (leaveTop ctx) s
-          v@Shared{}  -> updateSharedTerm (occ ctx) v
+          DontCare v  -> if ctx == Irrel then
+                           dontCare <$> occurs red ctx m xs v
+                         else
+                           abort (strongly ctx) $ MetaIrrelevantSolution m v
+          Def d es    -> do
+            drel <- relOfConst d
+            unless (not (unusableRelevance drel) || ctx == Irrel) $ do
+              reportSDoc "tc.meta.occurs" 35 $ text ("relevance of definition: " ++ show drel)
+              abort ctx $ MetaIrrelevantSolution m $ Def d []
+            Def d <$> occDef d ctx es
+          Con c ci vs -> Con c ci <$> occ ctx vs  -- if strongly rigid, remain so
+          Pi a b      -> uncurry Pi <$> occ ctx (a,b)
+          Sort s      -> Sort <$> occurs red ctx m (goNonStrict xs) s
           MetaV m' es -> do
               -- Check for loop
               --   don't fail hard on this, since we might still be on the top-level
@@ -310,7 +322,7 @@ instance Occurs Term where
                     -- Andreas, 2014-03-02, see issue 1070:
                     -- Do not prune when meta is projected!
                     caseMaybe (allApplyElims es) (throwError err) $ \ vs -> do
-                      killResult <- prune m' vs (takeRelevant xs)
+                      killResult <- prune m' vs (takeAll xs)
                       if (killResult == PrunedEverything)
                         -- after successful pruning, restart occurs check
                         then occurs red ctx m xs =<< instantiate (MetaV m' es)
@@ -338,7 +350,6 @@ instance Occurs Term where
       Con c _ vs -> metaOccurs m vs
       Pi a b     -> metaOccurs m (a,b)
       Sort s     -> metaOccurs m s
-      Shared p   -> metaOccurs m $ derefPtr p
       MetaV m' vs | m == m' -> patternViolation' 50 $ "Found occurrence of " ++ prettyShow m
                   | otherwise -> metaOccurs m vs
 
@@ -376,9 +387,7 @@ instance Occurs Level where
 
 instance Occurs PlusLevel where
   occurs red ctx m xs l@ClosedLevel{} = return l
-  occurs red ctx m xs (Plus n l) = Plus n <$> occurs red ctx' m xs l
-    where ctx' | n == 0    = ctx
-               | otherwise = leaveTop ctx  -- we leave Top only if we encounter at least one successor
+  occurs red ctx m xs (Plus n l) = Plus n <$> occurs red ctx m xs l
   metaOccurs m ClosedLevel{} = return ()
   metaOccurs m (Plus n l)    = metaOccurs m l
 
@@ -389,7 +398,7 @@ instance Occurs LevelAtom where
            NoUnfold  -> instantiate l
     case l of
       MetaLevel m' args -> do
-        MetaV m' args <- ignoreSharing <$> occurs red ctx m xs (MetaV m' args)
+        MetaV m' args <- occurs red ctx m xs (MetaV m' args)
         return $ MetaLevel m' args
       NeutralLevel r v  -> NeutralLevel r  <$> occurs red ctx m xs v
       BlockedLevel m' v -> BlockedLevel m' <$> occurs red Flex m xs v
@@ -415,23 +424,34 @@ instance Occurs Sort where
             YesUnfold -> reduce s
             NoUnfold  -> instantiate s
     case s' of
-      DLub s1 s2 -> uncurry DLub <$> occurs red (weakly ctx) m xs (s1,s2)
+      PiSort s1 s2 -> uncurry PiSort <$> occurs red (weakly ctx) m xs (s1,s2)
       Type a     -> Type <$> occurs red ctx m xs a
       Prop       -> return s'
       Inf        -> return s'
       SizeUniv   -> return s'
+      UnivSort s -> UnivSort <$> occurs red (weakly ctx) m xs s
+      MetaS x es -> do
+        MetaV x es <- occurs red ctx m xs (MetaV x es)
+        return $ MetaS x es
 
   metaOccurs m s = do
     s <- instantiate s
     case s of
-      DLub s1 s2 -> metaOccurs m (s1,s2)
+      PiSort s1 s2 -> metaOccurs m (s1,s2)
       Type a     -> metaOccurs m a
       Prop       -> return ()
       Inf        -> return ()
       SizeUniv   -> return ()
+      UnivSort s -> metaOccurs m s
+      MetaS x es -> metaOccurs m $ MetaV x es
 
 instance Occurs a => Occurs (Elim' a) where
-  occurs red ctx m xs e@Proj{}  = return e
+  occurs red ctx m xs e@(Proj _ f) = do
+    frel <- relOfConst f
+    unless (not (unusableRelevance frel) || ctx == Irrel) $ do
+      reportSDoc "tc.meta.occurs" 35 $ text ("relevance of projection: " ++ show frel)
+      abort ctx $ MetaIrrelevantSolution m $ Def f []
+    return e
   occurs red ctx m xs (Apply a) = Apply <$> occurs red ctx m xs a
 
   metaOccurs m (Proj{} ) = return ()
@@ -447,6 +467,8 @@ instance (Occurs a, Subst t a) => Occurs (Abs a) where
 instance Occurs a => Occurs (Arg a) where
   occurs red ctx m xs (Arg info x) | isIrrelevant info = Arg info <$>
     occurs red Irrel m (goIrrelevant xs) x
+  occurs red ctx m xs (Arg info x) | isNonStrict info  = Arg info <$>
+    occurs red ctx m (goNonStrict xs) x
   occurs red ctx m xs (Arg info x) = Arg info <$>
     occurs red ctx m xs x
 
@@ -510,7 +532,7 @@ hasBadRigid xs t = do
   let failure = throwError ()
   tb <- liftTCM $ reduceB t
   let t = ignoreBlocking tb
-  case ignoreSharing t of
+  case t of
     Var x _      -> return $ notElem x xs
     -- Issue 1153: A lambda has to be considered matchable.
     -- Lam _ v    -> hasBadRigid (0 : map (+1) xs) (absBody v)
@@ -530,15 +552,15 @@ hasBadRigid xs t = do
     -- offending variables under a constructor could be removed by
     -- the right instantiation of the meta variable.
     -- Thus, they are not rigid.
-    Con c _ args -> do
+    Con c _ es | Just args <- allApplyElims es -> do
       ifM (liftTCM $ isEtaCon (conName c))
         -- in case of a record con, we can in principle prune
         -- (but not this argument; the meta could become a projection!)
         (and <$> mapM (hasBadRigid xs . unArg) args)  -- not andM, we need to force the exceptions!
         failure
+    Con c _ es | otherwise -> failure
     Lit{}        -> failure -- matchable
     MetaV{}      -> failure -- potentially matchable
-    Shared p     -> __IMPOSSIBLE__
 
 -- | Check whether a term @Def f es@ is finally stuck.
 --   Currently, we give only a crude approximation.
@@ -592,13 +614,21 @@ class FoldRigid a where
 instance FoldRigid Term where
   foldRigid f t = do
     b <- liftTCM $ reduceB t
-    case ignoreSharing $ ignoreBlocking b of
+    case ignoreBlocking b of
+      -- Upon entry, we are in rigid position, thus,
+      -- bound variables are rigid ones.
       Var i es   -> f i `mappend` fold es
       Lam _ t    -> fold t
       Lit{}      -> mempty
       Def _ es   -> case b of
+        -- If the definition is blocked by a meta, its arguments
+        -- may be in flexible positions.
         Blocked{}                   -> mempty
+        -- If the definition is incomplete, arguments might disappear
+        -- by reductions that come with more clauses, thus, these
+        -- arguments are not rigid.
         NotBlocked MissingClauses _ -> mempty
+        -- _        -> mempty -- breaks: ImproveInertRHS, Issue442, PruneRecord, PruningNonMillerPattern
         _        -> fold es
       Con _ _ ts -> fold ts
       Pi a b     -> fold (a,b)
@@ -606,7 +636,6 @@ instance FoldRigid Term where
       Level l    -> fold l
       MetaV{}    -> mempty
       DontCare{} -> mempty
-      Shared{}   -> __IMPOSSIBLE__
     where fold = foldRigid f
 
 instance FoldRigid Type where
@@ -619,7 +648,9 @@ instance FoldRigid Sort where
       Prop       -> mempty
       Inf        -> mempty
       SizeUniv   -> mempty
-      DLub s1 s2 -> fold (s1, s2)
+      PiSort s1 s2 -> fold (s1, s2)
+      UnivSort s -> fold s
+      MetaS{}    -> mempty
     where fold = foldRigid f
 
 instance FoldRigid Level where
@@ -645,6 +676,9 @@ instance (Subst t a, FoldRigid a) => FoldRigid (Abs a) where
 instance FoldRigid a => FoldRigid (Arg a) where
   foldRigid f a =
     case getRelevance a of
+      -- Irrelevant arguments are definitionally equal to
+      -- values, so the variables there are not considered
+      -- "definitely rigid".
       Irrelevant -> mempty
       _          -> foldRigid f $ unArg a
 
@@ -741,8 +775,10 @@ performKill kills m a = do
   -- (de Bruijn level order).
   let perm = Perm n
              [ i | (i, Arg _ False) <- zip [0..] kills ]
-  m' <- newMeta (mvInfo mv) (mvPriority mv) perm
-                (HasType __IMPOSSIBLE__ a)
+      judg = case mvJudgement mv of
+        HasType{} -> HasType __IMPOSSIBLE__ a
+        IsSort{}  -> IsSort  __IMPOSSIBLE__ a
+  m' <- newMeta (mvInfo mv) (mvPriority mv) perm judg
   -- Andreas, 2010-10-15 eta expand new meta variable if necessary
   etaExpandMetaSafe m'
   let -- Arguments to new meta (de Bruijn indices)
