@@ -2,7 +2,6 @@
 
 module Agda.TypeChecking.Monad.Builtin where
 
-import Control.Applicative
 import Control.Monad.State
 import Control.Monad.Trans.Maybe
 
@@ -37,6 +36,7 @@ litType l = case l of
     _ <- primZero
     when (n > 0) $ void $ primSuc
     el <$> primNat
+  LitWord64 _ _ -> el <$> primWord64
   LitFloat _ _  -> el <$> primFloat
   LitChar _ _   -> el <$> primChar
   LitString _ _ -> el <$> primString
@@ -72,8 +72,8 @@ getBuiltin x =
 getBuiltin' :: HasBuiltins m => String -> m (Maybe Term)
 getBuiltin' x = do
     builtin <- getBuiltinThing x
-    case builtin of         -- ignore sharing to make sure zero isn't reduced to Lit 0
-        Just (Builtin t) -> return $ Just $ ignoreSharing $ killRange t
+    case builtin of
+        Just (Builtin t) -> return $ Just $ killRange t
         _                -> return Nothing
 
 getPrimitive' :: HasBuiltins m => String -> m (Maybe PrimFun)
@@ -92,7 +92,7 @@ constructorForm v = constructorForm' primZero primSuc v
 
 constructorForm' :: Applicative m => m Term -> m Term -> Term -> m Term
 constructorForm' pZero pSuc v =
-  case ignoreSharing v of
+  case v of
     Lit (LitNat r n)
       | n == 0    -> pZero
       | n > 0     -> (`apply1` Lit (LitNat r $ n - 1)) <$> pSuc
@@ -108,6 +108,8 @@ primInteger, primIntegerPos, primIntegerNegSuc,
     primList, primNil, primCons, primIO, primNat, primSuc, primZero,
     primNatPlus, primNatMinus, primNatTimes, primNatDivSucAux, primNatModSucAux,
     primNatEquality, primNatLess,
+    -- Machine words
+    primWord64,
     primSizeUniv, primSize, primSizeLt, primSizeSuc, primSizeInf, primSizeMax,
     primInf, primSharp, primFlat,
     primEquality, primRefl,
@@ -124,7 +126,7 @@ primInteger, primIntegerPos, primIntegerNegSuc,
     primAssoc, primAssocLeft, primAssocRight, primAssocNon,
     primPrecedence, primPrecRelated, primPrecUnrelated,
     primFixity, primFixityFixity,
-    primAgdaLiteral, primAgdaLitNat, primAgdaLitFloat, primAgdaLitString, primAgdaLitChar, primAgdaLitQName, primAgdaLitMeta,
+    primAgdaLiteral, primAgdaLitNat, primAgdaLitWord64, primAgdaLitFloat, primAgdaLitString, primAgdaLitChar, primAgdaLitQName, primAgdaLitMeta,
     primAgdaSort, primAgdaSortSet, primAgdaSortLit, primAgdaSortUnsupported,
     primAgdaDefinition, primAgdaDefinitionFunDef, primAgdaDefinitionDataDef, primAgdaDefinitionRecordDef,
     primAgdaDefinitionPostulate, primAgdaDefinitionPrimitive, primAgdaDefinitionDataConstructor,
@@ -137,7 +139,7 @@ primInteger, primIntegerPos, primIntegerNegSuc,
     primAgdaTCMTypeError, primAgdaTCMInferType, primAgdaTCMCheckType,
     primAgdaTCMNormalise, primAgdaTCMReduce,
     primAgdaTCMCatchError, primAgdaTCMGetContext, primAgdaTCMExtendContext, primAgdaTCMInContext,
-    primAgdaTCMFreshName, primAgdaTCMDeclareDef, primAgdaTCMDefineFun,
+    primAgdaTCMFreshName, primAgdaTCMDeclareDef, primAgdaTCMDeclarePostulate, primAgdaTCMDefineFun,
     primAgdaTCMGetType, primAgdaTCMGetDefinition,
     primAgdaTCMQuoteTerm, primAgdaTCMUnquoteTerm,
     primAgdaTCMBlockOnMeta, primAgdaTCMCommit, primAgdaTCMIsMacro,
@@ -169,6 +171,7 @@ primNatDivSucAux = getBuiltin builtinNatDivSucAux
 primNatModSucAux = getBuiltin builtinNatModSucAux
 primNatEquality  = getBuiltin builtinNatEquals
 primNatLess      = getBuiltin builtinNatLess
+primWord64       = getBuiltin builtinWord64
 primSizeUniv     = getBuiltin builtinSizeUniv
 primSize         = getBuiltin builtinSize
 primSizeLt       = getBuiltin builtinSizeLt
@@ -232,6 +235,7 @@ primAgdaErrorPartTerm   = getBuiltin builtinAgdaErrorPartTerm
 primAgdaErrorPartName   = getBuiltin builtinAgdaErrorPartName
 primAgdaLiteral   = getBuiltin builtinAgdaLiteral
 primAgdaLitNat    = getBuiltin builtinAgdaLitNat
+primAgdaLitWord64 = getBuiltin builtinAgdaLitWord64
 primAgdaLitFloat  = getBuiltin builtinAgdaLitFloat
 primAgdaLitChar   = getBuiltin builtinAgdaLitChar
 primAgdaLitString = getBuiltin builtinAgdaLitString
@@ -270,6 +274,7 @@ primAgdaTCMExtendContext = getBuiltin builtinAgdaTCMExtendContext
 primAgdaTCMInContext     = getBuiltin builtinAgdaTCMInContext
 primAgdaTCMFreshName     = getBuiltin builtinAgdaTCMFreshName
 primAgdaTCMDeclareDef    = getBuiltin builtinAgdaTCMDeclareDef
+primAgdaTCMDeclarePostulate   = getBuiltin builtinAgdaTCMDeclarePostulate
 primAgdaTCMDefineFun     = getBuiltin builtinAgdaTCMDefineFun
 primAgdaTCMGetType            = getBuiltin builtinAgdaTCMGetType
 primAgdaTCMGetDefinition      = getBuiltin builtinAgdaTCMGetDefinition
@@ -284,6 +289,7 @@ primAgdaTCMDebugPrint         = getBuiltin builtinAgdaTCMDebugPrint
 builtinNat, builtinSuc, builtinZero, builtinNatPlus, builtinNatMinus,
   builtinNatTimes, builtinNatDivSucAux, builtinNatModSucAux, builtinNatEquals,
   builtinNatLess, builtinInteger, builtinIntegerPos, builtinIntegerNegSuc,
+  builtinWord64,
   builtinFloat, builtinChar, builtinString, builtinUnit, builtinUnitUnit,
   builtinBool, builtinTrue, builtinFalse,
   builtinList, builtinNil, builtinCons, builtinIO,
@@ -306,7 +312,7 @@ builtinNat, builtinSuc, builtinZero, builtinNatPlus, builtinNatMinus,
   builtinAgdaTermDef, builtinAgdaTermCon, builtinAgdaTermPi,
   builtinAgdaTermSort, builtinAgdaTermLit, builtinAgdaTermUnsupported, builtinAgdaTermMeta,
   builtinAgdaErrorPart, builtinAgdaErrorPartString, builtinAgdaErrorPartTerm, builtinAgdaErrorPartName,
-  builtinAgdaLiteral, builtinAgdaLitNat, builtinAgdaLitFloat,
+  builtinAgdaLiteral, builtinAgdaLitNat, builtinAgdaLitWord64, builtinAgdaLitFloat,
   builtinAgdaLitChar, builtinAgdaLitString, builtinAgdaLitQName, builtinAgdaLitMeta,
   builtinAgdaClause, builtinAgdaClauseClause, builtinAgdaClauseAbsurd, builtinAgdaPattern,
   builtinAgdaPatVar, builtinAgdaPatCon, builtinAgdaPatDot, builtinAgdaPatLit,
@@ -321,7 +327,7 @@ builtinNat, builtinSuc, builtinZero, builtinNatPlus, builtinNatMinus,
   builtinAgdaTCMCheckType, builtinAgdaTCMNormalise, builtinAgdaTCMReduce,
   builtinAgdaTCMCatchError,
   builtinAgdaTCMGetContext, builtinAgdaTCMExtendContext, builtinAgdaTCMInContext,
-  builtinAgdaTCMFreshName, builtinAgdaTCMDeclareDef, builtinAgdaTCMDefineFun,
+  builtinAgdaTCMFreshName, builtinAgdaTCMDeclareDef, builtinAgdaTCMDeclarePostulate, builtinAgdaTCMDefineFun,
   builtinAgdaTCMGetType, builtinAgdaTCMGetDefinition,
   builtinAgdaTCMQuoteTerm, builtinAgdaTCMUnquoteTerm,
   builtinAgdaTCMBlockOnMeta, builtinAgdaTCMCommit, builtinAgdaTCMIsMacro,
@@ -338,6 +344,7 @@ builtinNatDivSucAux                  = "NATDIVSUCAUX"
 builtinNatModSucAux                  = "NATMODSUCAUX"
 builtinNatEquals                     = "NATEQUALS"
 builtinNatLess                       = "NATLESS"
+builtinWord64                        = "WORD64"
 builtinInteger                       = "INTEGER"
 builtinIntegerPos                    = "INTEGERPOS"
 builtinIntegerNegSuc                 = "INTEGERNEGSUC"
@@ -416,6 +423,7 @@ builtinAgdaErrorPartTerm             = "AGDAERRORPARTTERM"
 builtinAgdaErrorPartName             = "AGDAERRORPARTNAME"
 builtinAgdaLiteral                   = "AGDALITERAL"
 builtinAgdaLitNat                    = "AGDALITNAT"
+builtinAgdaLitWord64                 = "AGDALITWORD64"
 builtinAgdaLitFloat                  = "AGDALITFLOAT"
 builtinAgdaLitChar                   = "AGDALITCHAR"
 builtinAgdaLitString                 = "AGDALITSTRING"
@@ -454,6 +462,7 @@ builtinAgdaTCMExtendContext = "AGDATCMEXTENDCONTEXT"
 builtinAgdaTCMInContext     = "AGDATCMINCONTEXT"
 builtinAgdaTCMFreshName     = "AGDATCMFRESHNAME"
 builtinAgdaTCMDeclareDef    = "AGDATCMDECLAREDEF"
+builtinAgdaTCMDeclarePostulate    = "AGDATCMDECLAREPOSTULATE"
 builtinAgdaTCMDefineFun     = "AGDATCMDEFINEFUN"
 builtinAgdaTCMGetType       = "AGDATCMGETTYPE"
 builtinAgdaTCMGetDefinition = "AGDATCMGETDEFINITION"
@@ -498,9 +507,9 @@ data CoinductionKit = CoinductionKit
 
 coinductionKit' :: TCM CoinductionKit
 coinductionKit' = do
-  Def inf   _ <- ignoreSharing <$> primInf
-  Def sharp _ <- ignoreSharing <$> primSharp
-  Def flat  _ <- ignoreSharing <$> primFlat
+  Def inf   _ <- primInf
+  Def sharp _ <- primSharp
+  Def flat  _ <- primFlat
   return $ CoinductionKit
     { nameOfInf   = inf
     , nameOfSharp = sharp
@@ -528,15 +537,10 @@ primEqualityName = do
   -- 2. type polymorphic only
   -- 3. monomorphic.
   let lamV (Lam i b)  = mapFst (getHiding i :) $ lamV (unAbs b)
-      lamV (Shared p) = lamV (derefPtr p)
       lamV v          = ([], v)
   return $ case lamV eq of
     (_, Def equality _) -> equality
-  -- OLD:
-  --   ([Hidden, Hidden], Def equality _) -> equality
-  --   ([Hidden],         Def equality _) -> equality
-  --   ([],               Def equality _) -> equality
-    _                                  -> __IMPOSSIBLE__
+    _                   -> __IMPOSSIBLE__
 
 -- | Check whether the type is actually an equality (lhs ≡ rhs)
 --   and extract lhs, rhs, and their type.
@@ -546,7 +550,7 @@ primEqualityName = do
 equalityView :: Type -> TCM EqualityView
 equalityView t0@(El s t) = do
   equality <- primEqualityName
-  case ignoreSharing t of
+  case t of
     Def equality' es | equality' == equality -> do
       let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
       let n = length vs
