@@ -3,10 +3,10 @@
 module Agda.TypeChecking.Monad.Constraints where
 
 import Control.Arrow ((&&&))
-import Control.Applicative
 import Control.Monad.State
 import Control.Monad.Reader
 
+import qualified Data.Foldable as Fold
 import qualified Data.List as List
 import Data.Set (Set)
 import qualified Data.Set as Set
@@ -43,13 +43,12 @@ solvingProblem pid = solvingProblems (Set.singleton pid)
 solvingProblems :: Set ProblemId -> TCM a -> TCM a
 solvingProblems pids m = verboseBracket "tc.constr.solve" 50 ("working on problems " ++ show (Set.toList pids)) $ do
   x <- local (\e -> e { envActiveProblems = pids `Set.union` envActiveProblems e }) m
-  sequence_
-    [ ifNotM (isProblemSolved pid)
+  Fold.forM_ pids $ \ pid -> do
+    ifNotM (isProblemSolved pid)
         (reportSLn "tc.constr.solve" 50 $ "problem " ++ show pid ++ " was not solved.")
       $ {- else -} do
         reportSLn "tc.constr.solve" 50 $ "problem " ++ show pid ++ " was solved!"
         wakeConstraints (return . blockedOn pid . clValue . theConstraint)
-    | pid <- Set.toList pids ]
   return x
   where
     blockedOn pid (Guarded _ pid') = pid == pid'
@@ -166,6 +165,9 @@ addConstraintTo bucket c = do
     isBlocking FindInScope{} = False
     isBlocking IsEmpty{}     = True
     isBlocking CheckSizeLtSat{} = True
+    isBlocking CheckFunDef{} = True
+    isBlocking HasBiggerSort{} = False
+    isBlocking HasPTSRule{}  = False
 
 -- | Add already awake constraints
 addAwakeConstraints :: Constraints -> TCM ()
