@@ -118,12 +118,12 @@ instance IsInstantiatedMeta MetaId where
 instance IsInstantiatedMeta Term where
   isInstantiatedMeta = loop where
    loop v =
-    case ignoreSharing v of
+    case v of
       MetaV x _  -> isInstantiatedMeta x
       DontCare v -> loop v
       Level l    -> isInstantiatedMeta l
       Lam _ b    -> isInstantiatedMeta b
-      Con _ _ vs -> isInstantiatedMeta vs
+      Con _ _ es | Just vs <- allApplyElims es -> isInstantiatedMeta vs
       _          -> __IMPOSSIBLE__
 
 instance IsInstantiatedMeta Level where
@@ -191,7 +191,7 @@ createMetaInfo' b = do
 
 setValueMetaName :: Term -> MetaNameSuggestion -> TCM ()
 setValueMetaName v s = do
-  case ignoreSharing v of
+  case v of
     MetaV mi _ -> setMetaNameSuggestion mi s
     u          -> do
       reportSLn "tc.meta.name" 70 $
@@ -226,7 +226,10 @@ modifyInteractionPoints f =
 registerInteractionPoint :: Bool -> Range -> Maybe Nat -> TCM InteractionId
 registerInteractionPoint preciseRange r maybeId = do
   m <- use stInteractionPoints
-  if not preciseRange then continue m else do
+  -- If we're given an interaction id we shouldn't look up by range.
+  -- This is important when doing 'refine', since all interaction points
+  -- created by the refine gets the same range.
+  if not preciseRange || isJust maybeId then continue m else do
     -- If the range does not come from a file, it is not
     -- precise, so ignore it.
     Strict.caseMaybe (rangeFile r) (continue m) $ \ _ -> do
@@ -331,7 +334,13 @@ newMeta' :: MetaInstantiation -> MetaInfo -> MetaPriority -> Permutation ->
 newMeta' inst mi p perm j = do
   x <- fresh
   let j' = j { jMetaId = x }  -- fill the identifier part of the judgement
-      mv = MetaVar mi p perm j' inst Set.empty Instantiable
+      mv = MetaVar{ mvInfo             = mi
+                  , mvPriority         = p
+                  , mvPermutation      = perm
+                  , mvJudgement        = j'
+                  , mvInstantiation    = inst
+                  , mvListeners        = Set.empty
+                  , mvFrozen           = Instantiable }
   -- printing not available (import cycle)
   -- reportSDoc "tc.meta.new" 50 $ text "new meta" <+> prettyTCM j'
   stMetaStore %= Map.insert x mv
@@ -455,7 +464,6 @@ instance UnFreezeMeta Type where
   unfreezeMeta (El s t) = unfreezeMeta s >> unfreezeMeta t
 
 instance UnFreezeMeta Term where
-  unfreezeMeta (Shared p)    = unfreezeMeta $ derefPtr p
   unfreezeMeta (MetaV x _)   = unfreezeMeta x
   unfreezeMeta (Sort s)      = unfreezeMeta s
   unfreezeMeta (Level l)     = unfreezeMeta l
@@ -464,7 +472,7 @@ instance UnFreezeMeta Term where
   unfreezeMeta _             = return ()
 
 instance UnFreezeMeta Sort where
-  unfreezeMeta (Type l)      = unfreezeMeta l
+  unfreezeMeta (MetaS x _)   = unfreezeMeta x
   unfreezeMeta _             = return ()
 
 instance UnFreezeMeta Level where
