@@ -9,6 +9,7 @@ import Control.Arrow (first, second, (***))
 import Control.Applicative hiding (empty)
 import Control.Monad.State
 import Control.Monad.Reader
+import Control.Monad.Writer
 import Control.Monad.Trans.Maybe
 
 import qualified Data.List as List
@@ -34,9 +35,9 @@ import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Env
 import Agda.TypeChecking.Monad.Mutual
 import Agda.TypeChecking.Monad.Open
-import Agda.TypeChecking.Monad.Local
 import Agda.TypeChecking.Monad.State
 import Agda.TypeChecking.Monad.Trace
+import Agda.TypeChecking.DropArgs
 import Agda.TypeChecking.Warnings
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Substitute
@@ -171,8 +172,8 @@ setFunctionFlag flag val q = modifyGlobalDefinition q $ set (theDefLens . funFla
 markStatic :: QName -> TCM ()
 markStatic = setFunctionFlag FunStatic True
 
-markInline :: QName -> TCM ()
-markInline = setFunctionFlag FunInline True
+markInline :: Bool -> QName -> TCM ()
+markInline b = setFunctionFlag FunInline b
 
 markInjective :: QName -> TCM ()
 markInjective q = modifyGlobalDefinition q $ \def -> def { defInjective = True }
@@ -207,12 +208,14 @@ addSection m = do
       reportSLn "impossible" 60 $ "with content " ++ prettyShow sec
       __IMPOSSIBLE__
   -- Add the new section.
-  setDefaultModuleParameters m
+  setModuleCheckpoint m
   modifySignature $ over sigSections $ Map.insert m sec
 
-setDefaultModuleParameters :: ModuleName -> TCM ()
-setDefaultModuleParameters m =
-  stModuleParameters %= Map.insert m defaultModuleParameters
+-- | Sets the checkpoint for the given module to the current checkpoint.
+setModuleCheckpoint :: ModuleName -> TCM ()
+setModuleCheckpoint m = do
+  chkpt <- view eCurrentCheckpoint
+  stModuleCheckpoints %= Map.insert m chkpt
 
 -- | Get a section.
 --
@@ -370,13 +373,22 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
     copyDef ts (x, y) = do
       def <- getConstInfo x
       np  <- argsToUse (qnameModule x)
-      copyDef' np def
+      -- Issue #3083: We need to use the hiding from the telescope of the
+      -- original module. This can be different than the hiding for the common
+      -- parent in the case of record modules.
+      hidings <- map getHiding . telToList <$> lookupSection (qnameModule x)
+      let ts' = zipWith setHiding hidings ts
+      commonTel <- lookupSection (commonParentModule old $ qnameModule x)
+      reportSLn "tc.mod.apply" 80 $ init $ unlines
+        [ "copyDef " ++ prettyShow x ++ " -> " ++ prettyShow y
+        , "ts' = " ++ prettyShow ts' ]
+      copyDef' ts' np def
       where
-        copyDef' np d = do
+        copyDef' ts np d = do
           reportSLn "tc.mod.apply" 60 $ "making new def for " ++ prettyShow y ++ " from " ++ prettyShow x ++ " with " ++ show np ++ " args " ++ show (defAbstract d)
-          reportSLn "tc.mod.apply" 80 $
-            "args = " ++ show ts' ++ "\n" ++
-            "old type = " ++ prettyShow (defType d)
+          reportSLn "tc.mod.apply" 80 $ init $ unlines
+            [ "args = " ++ show ts'
+            , "old type = " ++ prettyShow (defType d) ]
           reportSLn "tc.mod.apply" 80 $
             "new type = " ++ prettyShow t
           addConstant y =<< nd y
@@ -427,7 +439,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
             -- Even if we apply the record argument (must be @var 0@), we stay a projection.
             -- This is because we may abstract the record argument later again.
             -- See succeed/ProjectionNotNormalized.agda
-            isVar0 t = case ignoreSharing $ unArg t of Var 0 [] -> True; _ -> False
+            isVar0 t = case unArg t of Var 0 [] -> True; _ -> False
             proj   = case oldDef of
               Function{funProjection = Just p@Projection{projIndex = n}}
                 | size ts' < n || (size ts' == n && maybe True isVar0 (lastMaybe ts'))
@@ -475,13 +487,17 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                         , clauseFullRange = getRange $ defClauses d
                         , clauseTel       = EmptyTel
                         , namedClausePats = []
-                        , clauseBody      = Just $ case oldDef of
+                        , clauseBody      = Just $ dropArgs pars $ case oldDef of
                             Function{funProjection = Just p} -> projDropParsApply p ProjSystem ts'
                             _ -> Def x $ map Apply ts'
                         , clauseType      = Just $ defaultArg t
                         , clauseCatchall  = False
                         , clauseUnreachable = Just False -- definitely not unreachable
                         }
+              where
+                -- The number of remaining parameters. We need to drop the
+                -- lambdas corresponding to these from the clause body above.
+                pars = max 0 $ maybe 0 (pred . projIndex) proj
 
     {- Example
 
@@ -528,7 +544,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
 -- | Add a display form to a definition (could be in this or imported signature).
 addDisplayForm :: QName -> DisplayForm -> TCM ()
 addDisplayForm x df = do
-  d <- makeLocal df
+  d <- makeOpen df
   let add = updateDefinition x $ \ def -> def{ defDisplay = d : defDisplay def }
   ifM (isLocal x)
     {-then-} (modifySignature add)
@@ -573,7 +589,6 @@ canonicalName x = do
   where
     can body = canonicalName $ extract $ fromMaybe __IMPOSSIBLE__ body
     extract (Def x _)  = x
-    extract (Shared p) = extract $ derefPtr p
     extract _          = __IMPOSSIBLE__
 
 sameDef :: QName -> QName -> TCM (Maybe QName)
@@ -621,7 +636,7 @@ sigError f a = \case
   SigUnknown s -> f s
   SigAbstract  -> a
 
-class (Functor m, Applicative m, Monad m, HasOptions m, MonadDebug m) => HasConstInfo m where
+class (Functor m, Applicative m, Monad m, HasOptions m, MonadDebug m, MonadReader TCEnv m) => HasConstInfo m where
   -- | Lookup the definition of a name. The result is a closed thing, all free
   --   variables have been abstracted over.
   getConstInfo :: QName -> m Definition
@@ -665,7 +680,7 @@ instance HasConstInfo (TCMT IO) where
       Left SigAbstract      -> notInScope $ qnameToConcrete q
 
 defaultGetConstInfo
-  :: (HasOptions m, MonadDebug m)
+  :: (HasOptions m, MonadDebug m, MonadReader TCEnv m)
   => TCState -> TCEnv -> QName -> m (Either SigError Definition)
 defaultGetConstInfo st env q = do
     let defs  = st^.(stSignature . sigDefinitions)
@@ -701,6 +716,10 @@ instance HasConstInfo m => HasConstInfo (ExceptT err m) where
   getConstInfo' = lift . getConstInfo'
   getRewriteRulesFor = lift . getRewriteRulesFor
 
+instance (Monoid w, HasConstInfo m) => HasConstInfo (WriterT w m) where
+  getConstInfo' = lift . getConstInfo'
+  getRewriteRulesFor = lift . getRewriteRulesFor
+
 {-# INLINE getConInfo #-}
 getConInfo :: MonadTCM tcm => ConHead -> tcm Definition
 getConInfo = liftTCM . getConstInfo . conName
@@ -722,6 +741,10 @@ setPolarity q pol = do
     "Setting polarity of " ++ prettyShow q ++ " to " ++ prettyShow pol ++ "."
   modifySignature $ updateDefinition q $ updateDefPolarity $ const pol
 
+-- | Look up the forced arguments of a definition.
+getForcedArgs :: QName -> TCM [IsForced]
+getForcedArgs q = defForced <$> getConstInfo q
+
 -- | Get argument occurrence info for argument @i@ of definition @d@ (never fails).
 getArgOccurrence :: QName -> Nat -> TCM Occurrence
 getArgOccurrence d i = do
@@ -810,6 +833,15 @@ definitelyNonRecursive_ = maybe False null . getMutual_
 getCurrentModuleFreeVars :: TCM Nat
 getCurrentModuleFreeVars = size <$> (lookupSection =<< currentModule)
 
+--   For annoying reasons the qnameModule of a pattern lambda is not correct
+--   (#2883), so make sure to grab the right module for those.
+getDefModule :: HasConstInfo m => QName -> m ModuleName
+getDefModule f = do
+  def <- getConstInfo f
+  return $ case theDef def of
+    Function{ funExtLam = Just (ExtLamInfo m) } -> m
+    _                                           -> qnameModule f
+
 -- | Compute the number of free variables of a defined name. This is the sum of
 --   number of parameters shared with the current module and the number of
 --   anonymous variables (if the name comes from a let-bound module).
@@ -839,7 +871,7 @@ getModuleFreeVars m = do
 --   We have to insert the module telescope of the common prefix
 --   of the current module and the module where the definition comes from.
 --   (Properly raised to the current context.)
---y
+--
 --   Example:
 --   @
 --      module M₁ Γ where
