@@ -98,7 +98,7 @@ buildSubstitution err n vs = parallelS $ map unArg $ matchedArgs err n vs
 -- upon failure, no further matching is performed.
 
 foldMatch
-  :: forall p v . (p -> v -> ReduceM (Match Term, v))
+  :: forall p v . IsProjP p => (p -> v -> ReduceM (Match Term, v))
   -> [p] -> [v] -> ReduceM (Match Term, [v])
 foldMatch match = loop where
   loop :: [p] -> [v] -> ReduceM (Match Term, [v])
@@ -108,7 +108,21 @@ foldMatch match = loop where
       (p : ps, v : vs) -> do
         (r, v') <- match p v
         case r of
-          No         -> return (No        , v' : vs)
+          No | Just{} <- isProjP p -> return (No, v' : vs)
+          No         -> do
+            -- Issue 2964: Even when the first pattern doesn't match we should
+            -- continue to the next patterns (and potentially block on them)
+            -- because the splitting order in the case tree may not be
+            -- left-to-right.
+            (r', _vs') <- loop ps vs
+            -- Issue 2968: do not use vs' here, because it might
+            -- contain ill-typed terms due to eta-expansion at wrong
+            -- type.
+            let vs1 = v' : vs
+            case r' of
+              Yes s' us' -> return (No         , vs1)
+              No         -> return (No         , vs1)
+              DontKnow m -> return (DontKnow m , vs1)
           DontKnow m -> return (DontKnow m, v' : vs)
           Yes s us   -> do
             (r', vs') <- loop ps vs
@@ -154,8 +168,9 @@ matchCopattern pat@ProjP{} elim@(Proj _ q) = do
   q         <- getOriginalProjection q
   return $ if p == q then (Yes YesSimplification empty, elim)
                      else (No,                          elim)
-matchCopattern ProjP{} Apply{}   = __IMPOSSIBLE__
-matchCopattern _       Proj{}    = __IMPOSSIBLE__
+-- The following two cases are not impossible, see #2964
+matchCopattern ProjP{} elim@Apply{}   = return (No , elim)
+matchCopattern _       elim@Proj{}    = return (No , elim)
 matchCopattern p       (Apply v) = mapSnd Apply <$> matchPattern p v
 
 matchPatterns :: [NamedArg DeBruijnPattern]
@@ -178,14 +193,13 @@ matchPattern :: DeBruijnPattern
              -> ReduceM (Match Term, Arg Term)
 matchPattern p u = case (p, u) of
   (ProjP{}, _            ) -> __IMPOSSIBLE__
-  (VarP x , arg          ) -> return (Yes NoSimplification entry, arg)
+  (VarP _ x , arg          ) -> return (Yes NoSimplification entry, arg)
     where entry = singleton (dbPatVarIndex x, arg)
-  (DotP _ , arg@(Arg _ v)) -> return (Yes NoSimplification empty, arg)
-  (AbsurdP _ , arg@(Arg _ v)) -> return (Yes NoSimplification empty, arg)
+  (DotP _ _ , arg@(Arg _ v)) -> return (Yes NoSimplification empty, arg)
   (LitP l , arg@(Arg _ v)) -> do
     w <- reduceB' v
     let arg' = arg $> ignoreBlocking w
-    case ignoreSharing <$> w of
+    case w of
       NotBlocked _ (Lit l')
           | l == l'            -> return (Yes YesSimplification empty , arg')
           | otherwise          -> return (No                          , arg')
@@ -204,7 +218,7 @@ matchPattern p u = case (p, u) of
         -- This case is necessary if we want to use the clauses before
         -- record pattern translation (e.g., in type-checking definitions by copatterns).
         unless (size fs == size ps) __IMPOSSIBLE__
-        mapSnd (Arg info . Con c (fromConPatternInfo cpi)) <$> do
+        mapSnd (Arg info . Con c (fromConPatternInfo cpi) . map Apply) <$> do
           matchPatterns ps $ for fs $ \ (Arg ai f) -> Arg ai $ v `applyE` [Proj ProjSystem f]
     where
     isEtaRecordCon :: QName -> ReduceM (Maybe [Arg QName])
@@ -212,7 +226,7 @@ matchPattern p u = case (p, u) of
       (theDef <$> getConstInfo c) >>= \case
         Constructor{ conData = d } -> do
           (theDef <$> getConstInfo d) >>= \case
-            r@Record{ recFields = fs } | recEtaEquality r -> return $ Just fs
+            r@Record{ recFields = fs } | YesEta <- recEtaEquality r -> return $ Just fs
             _ -> return Nothing
         _ -> __IMPOSSIBLE__
 
@@ -226,7 +240,7 @@ matchPattern p u = case (p, u) of
         --    an axiom at this stage (if we are checking the
         --    projection functions for a record type).
 {-
-        w <- case ignoreSharing <$> w of
+        w <- case w of
                NotBlocked r (Def f es) ->   -- Andreas, 2014-06-12 TODO: r == ReallyNotBlocked sufficient?
                  unfoldDefinitionE True reduceB' (Def f []) f es
                    -- reduceB is used here because some constructors
@@ -241,11 +255,11 @@ matchPattern p u = case (p, u) of
                _ -> return w
         let v = ignoreBlocking w
             arg = Arg info v  -- the reduced argument
-        case ignoreSharing <$> w of
+        case w of
           NotBlocked _ (Con c' ci vs)
             | c == c'               -> do
-                (m, vs) <- yesSimplification <$> matchPatterns ps vs
-                return (m, Arg info $ Con c' ci vs)
+                (m, vs) <- yesSimplification <$> matchPatterns ps (fromMaybe __IMPOSSIBLE__ $ allApplyElims vs)
+                return (m, Arg info $ Con c' ci (map Apply vs))
             | otherwise             -> return (No                          , arg)
           NotBlocked _ (MetaV x vs) -> return (DontKnow $ Blocked x ()     , arg)
           Blocked x _               -> return (DontKnow $ Blocked x ()     , arg)
