@@ -5,7 +5,11 @@
 -- | Check that a datatype is strictly positive.
 module Agda.TypeChecking.Positivity where
 
-import Prelude hiding (null)
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ( (<>), null )
+#else
+import Prelude hiding ( null )
+#endif
 
 import Control.Applicative hiding (empty)
 import Control.DeepSeq
@@ -62,7 +66,7 @@ import Agda.Utils.Size
 #include "undefined.h"
 import Agda.Utils.Impossible
 
-type Graph n e = Graph.Graph n n e
+type Graph n e = Graph.Graph n e
 
 -- | Check that the datatypes in the mutual block containing the given
 --   declarations are strictly positive.
@@ -222,17 +226,20 @@ checkStrictlyPositive mi qset = disableDestructiveUpdate $ do
     -- Set the polarity of the arguments to a couple of definitions
     setArgOccs :: Set QName -> [QName] -> Graph Node Occurrence -> TCM ()
     setArgOccs qset qs g = do
-      -- Compute a map from each name in q to the maximal argument index
-      let maxs = Map.fromListWith max
-           [ (q, i) | ArgNode q i <- Set.toList $ Graph.sourceNodes g, q `Set.member` qset ]
+      -- Andreas, 2018-05-11, issue #3049: we need to be pessimistic about
+      -- argument polarity beyond the formal arity of the function.
+      --
+      -- -- Compute a map from each name in q to the maximal argument index
+      -- let maxs = Map.fromListWith max
+      --      [ (q, i) | ArgNode q i <- Set.toList $ Graph.nodes g, q `Set.member` qset ]
       forM_ qs $ \ q -> inConcreteOrAbstractMode q $ \ def -> do
         reportSDoc "tc.pos.args" 10 $ text "checking args of" <+> prettyTCM q
         n <- getDefArity def
         -- If there is no outgoing edge @ArgNode q i@, all @n@ arguments are @Unused@.
         -- Otherwise, we obtain the occurrences from the Graph.
         let findOcc i = fromMaybe Unused $ Graph.lookup (ArgNode q i) (DefNode q) g
-            args = caseMaybe (Map.lookup q maxs) (replicate n Unused) $ \ m ->
-              map findOcc [0 .. max m (n - 1)]
+            args = -- caseMaybe (Map.lookup q maxs) (replicate n Unused) $ \ m ->
+              map findOcc [0 .. n-1]  -- [0 .. max m (n - 1)] -- triggers issue #3049
         reportSDoc "tc.pos.args" 10 $ sep
           [ text "args of" <+> prettyTCM q <+> text "="
           , nest 2 $ prettyList $ map prettyTCM args
@@ -243,16 +250,14 @@ checkStrictlyPositive mi qset = disableDestructiveUpdate $ do
         setArgOccurrences q $!! args
 
 getDefArity :: Definition -> TCM Int
-getDefArity def = case theDef def of
-  defn@Function{} -> do
-    let dropped = projectionArgs defn
-    -- TODO: instantiateFull followed by arity could perhaps be
-    -- optimised, presumably the instantiation can be performed
-    -- lazily.
-    subtract dropped . arity <$> instantiateFull (defType def)
-  Datatype{ dataPars = n } -> return n
-  Record{ recPars = n }    -> return n
-  _                        -> return 0
+getDefArity def = do
+  let dropped = case theDef def of
+        defn@Function{} -> projectionArgs defn
+        _ -> 0
+  -- TODO: instantiateFull followed by arity could perhaps be
+  -- optimised, presumably the instantiation can be performed
+  -- lazily.
+  subtract dropped . arity <$> instantiateFull (defType def)
 
 -- Operations on occurrences -------------------------------------------
 
@@ -288,7 +293,7 @@ instance PrettyTCM OccursWhere where
                           [prettyTCM q]
         UnderInf     -> pwords "under" ++
                         [do -- this cannot fail if an 'UnderInf' has been generated
-                            Def inf _ <- ignoreSharing <$> primInf
+                            Def inf _ <- primInf
                             prettyTCM inf]
         VarArg       -> pwords "in an argument to a bound variable"
         MetaArg      -> pwords "in an argument to a metavariable"
@@ -505,7 +510,6 @@ instance ComputeOccurrences Term where
     Lit{}        -> return emptyOB
     Sort{}       -> return emptyOB
     DontCare _   -> return emptyOB -- Andreas, 2011-09-09: do we need to check for negative occurrences in irrelevant positions?
-    Shared p     -> occurrences $ derefPtr p
 
 instance ComputeOccurrences Level where
   occurrences (Max as) = occurrences as
@@ -680,10 +684,10 @@ instance StarSemiRing Edge where
 -- expensive. (See 'computeEdges' for an example.)
 buildOccurrenceGraph :: Set QName -> TCM (Graph Node Edge)
 buildOccurrenceGraph qs =
-  Graph.fromListWith oplus . concat <$>
+  Graph.fromEdgesWith oplus . concat <$>
     mapM defGraph (Set.toList qs)
   where
-    defGraph :: QName -> TCM [Graph.Edge Node Node Edge]
+    defGraph :: QName -> TCM [Graph.Edge Node Edge]
     defGraph q = inConcreteOrAbstractMode q $ \ _def -> do
       occs <- computeOccurrences' q
 
@@ -750,7 +754,7 @@ computeEdges
   -> QName
      -- ^ The current name.
   -> OccurrencesBuilder
-  -> TCM [Graph.Edge Node Node Edge]
+  -> TCM [Graph.Edge Node Edge]
 computeEdges muts q ob =
   ($ []) <$> mkEdge __IMPOSSIBLE__ StrictPos (preprocess ob)
   where
