@@ -13,7 +13,6 @@
 module Agda.TypeChecking.Primitive where
 
 import Control.Monad
-import Control.Applicative
 import Control.Monad.Reader (asks)
 
 import Data.Char
@@ -22,8 +21,7 @@ import qualified Data.Map as Map
 import Data.Maybe
 import Data.Traversable (traverse)
 import Data.Monoid (mempty)
-
-import Numeric.IEEE ( IEEE(identicalIEEE) )
+import Data.Word
 
 import Agda.Interaction.Options
 import qualified Agda.Interaction.Options.Lenses as Lens
@@ -57,6 +55,7 @@ import Agda.Utils.Monad
 import Agda.Utils.Pretty (pretty, prettyShow)
 import Agda.Utils.Size
 import Agda.Utils.String ( Str(Str), unStr )
+import Agda.Utils.Float
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -104,6 +103,7 @@ instance PrimTerm a => PrimType a where
 
 class    PrimTerm a       where primTerm :: a -> TCM Term
 instance PrimTerm Integer where primTerm _ = primInteger
+instance PrimTerm Word64  where primTerm _ = primWord64
 instance PrimTerm Bool    where primTerm _ = primBool
 instance PrimTerm Char    where primTerm _ = primChar
 instance PrimTerm Double  where primTerm _ = primFloat
@@ -131,6 +131,7 @@ class ToTerm a where
   toTermR = (pure .) <$> toTerm
 
 instance ToTerm Nat     where toTerm = return $ Lit . LitNat noRange . toInteger
+instance ToTerm Word64  where toTerm = return $ Lit . LitWord64 noRange
 instance ToTerm Lvl     where toTerm = return $ Level . Max . (:[]) . ClosedLevel . unLvl
 instance ToTerm Double  where toTerm = return $ Lit . LitFloat noRange
 instance ToTerm Char    where toTerm = return $ Lit . LitChar noRange
@@ -182,7 +183,6 @@ instance ToTerm ArgInfo where
           Relevant   -> rel
           Irrelevant -> irr
           NonStrict  -> rel
-          Forced{}   -> irr
       ]
 
 instance ToTerm Fixity' where
@@ -243,22 +243,22 @@ class FromTerm a where
 
 instance FromTerm Integer where
   fromTerm = do
-    Con pos _    [] <- ignoreSharing <$> primIntegerPos
-    Con negsuc _ [] <- ignoreSharing <$> primIntegerNegSuc
+    Con pos _    [] <- primIntegerPos
+    Con negsuc _ [] <- primIntegerNegSuc
     toNat         <- fromTerm :: TCM (FromTermFunction Nat)
     return $ \ v -> do
       b <- reduceB' v
       let v'  = ignoreBlocking b
           arg = (<$ v')
-      case ignoreSharing $ unArg (ignoreBlocking b) of
-        Con c ci [u]
+      case unArg (ignoreBlocking b) of
+        Con c ci [Apply u]
           | c == pos    ->
             redBind (toNat u)
-              (\ u' -> notReduced $ arg $ Con c ci [ignoreReduced u']) $ \ n ->
+              (\ u' -> notReduced $ arg $ Con c ci [Apply $ ignoreReduced u']) $ \ n ->
             redReturn $ fromIntegral n
           | c == negsuc ->
             redBind (toNat u)
-              (\ u' -> notReduced $ arg $ Con c ci [ignoreReduced u']) $ \ n ->
+              (\ u' -> notReduced $ arg $ Con c ci [Apply $ ignoreReduced u']) $ \ n ->
             redReturn $ fromIntegral $ -n - 1
         _ -> return $ NoReduction (reduced b)
 
@@ -267,6 +267,11 @@ instance FromTerm Nat where
     LitNat _ n -> Just $ fromInteger n
     _          -> Nothing
 
+instance FromTerm Word64 where
+  fromTerm = fromLiteral $ \ case
+    LitWord64 _ n -> Just n
+    _             -> Nothing
+
 instance FromTerm Lvl where
   fromTerm = fromReducedTerm $ \l -> case l of
     Level (Max [ClosedLevel n]) -> Just $ Lvl n
@@ -306,7 +311,7 @@ instance FromTerm Bool where
                 | t =?= false -> Just False
                 | otherwise   -> Nothing
         where
-            a =?= b = ignoreSharing a === ignoreSharing b
+            a =?= b = a === b
             Def x [] === Def y []   = x == y
             Con x _ [] === Con y _ [] = x == y
             Var n [] === Var m []   = n == m
@@ -324,23 +329,22 @@ instance (ToTerm a, FromTerm a) => FromTerm [a] where
     where
       isCon (Lam _ b)  = isCon $ absBody b
       isCon (Con c _ _)= return c
-      isCon (Shared p) = isCon (derefPtr p)
       isCon v          = __IMPOSSIBLE__
 
       mkList nil cons toA fromA t = do
         b <- reduceB' t
         let t = ignoreBlocking b
         let arg = (<$ t)
-        case ignoreSharing $ unArg t of
+        case unArg t of
           Con c ci []
             | c == nil  -> return $ YesReduction NoSimplification []
-          Con c ci [x,xs]
-            | c == cons ->
+          Con c ci es
+            | c == cons, Just [x,xs] <- allApplyElims es ->
               redBind (toA x)
-                  (\x' -> notReduced $ arg $ Con c ci [ignoreReduced x',xs]) $ \y ->
+                  (\x' -> notReduced $ arg $ Con c ci (map Apply [ignoreReduced x',xs])) $ \y ->
               redBind
                   (mkList nil cons toA fromA xs)
-                  (fmap $ \xs' -> arg $ Con c ci [defaultArg $ fromA y, xs']) $ \ys ->
+                  (fmap $ \xs' -> arg $ Con c ci (map Apply [defaultArg $ fromA y, xs'])) $ \ys ->
               redReturn (y : ys)
           _ -> return $ NoReduction (reduced b)
 
@@ -359,7 +363,7 @@ redReturn = return . YesReduction YesSimplification
 fromReducedTerm :: (Term -> Maybe a) -> TCM (FromTermFunction a)
 fromReducedTerm f = return $ \t -> do
     b <- reduceB' t
-    case f $ ignoreSharing $ unArg (ignoreBlocking b) of
+    case f $ unArg (ignoreBlocking b) of
         Just x  -> return $ YesReduction NoSimplification x
         Nothing -> return $ NoReduction (reduced b)
 
@@ -379,7 +383,7 @@ primTrustMe = do
   eqTy <- defType <$> getConstInfo eq
   -- E.g. @eqTy = eqTel → Set a@ where @eqTel = {a : Level} {A : Set a} (x y : A)@.
   TelV eqTel eqCore <- telView eqTy
-  let eqSort = case ignoreSharing $ unEl eqCore of
+  let eqSort = case unEl eqCore of
         Sort s -> s
         _      -> __IMPOSSIBLE__
 
@@ -389,10 +393,10 @@ primTrustMe = do
 
   -- BUILTIN REFL maybe a constructor with one (the principal) argument or only parameters.
   -- Get the ArgInfo of the principal argument of refl.
-  con@(Con rf ci []) <- ignoreSharing <$> primRefl
+  con@(Con rf ci []) <- primRefl
   minfo <- fmap (setOrigin Inserted) <$> getReflArgInfo rf
   let (refl :: Arg Term -> Term) = case minfo of
-        Just ai -> Con rf ci . (:[]) . setArgInfo ai
+        Just ai -> Con rf ci . (:[]) . Apply . setArgInfo ai
         Nothing -> const con
 
   -- The implementation of primTrustMe:
@@ -446,7 +450,7 @@ genPrimForce b ret = do
         u <- reduceB' u
         let isWHNF Blocked{} = return False
             isWHNF (NotBlocked _ u) =
-              case ignoreSharing $ unArg u of
+              case unArg u of
                 Lit{}      -> return True
                 Con{}      -> return True
                 Lam{}      -> return True
@@ -462,7 +466,6 @@ genPrimForce b ret = do
                     _          -> False
                 MetaV{}    -> return False
                 Var{}      -> return False
-                Shared{}   -> __IMPOSSIBLE__
 
         ifM (isWHNF u)
             (redReturn $ ret (unArg f) (ignoreBlocking u))
@@ -615,15 +618,15 @@ garr :: (Relevance -> Relevance) -> TCM Type -> TCM Type -> TCM Type
 garr f a b = do
   a' <- a
   b' <- b
-  return $ El (getSort a' `sLub` getSort b') $
+  return $ El (funSort (getSort a') (getSort b')) $
            Pi (Dom (mapRelevance f defaultArgInfo) a') (NoAbs "_" b')
 
 gpi :: ArgInfo -> String -> TCM Type -> TCM Type -> TCM Type
 gpi info name a b = do
   a <- a
-  b <- addContext' (name, Dom info a) b
+  b <- addContext (name, Dom info a) b
   let y = stringToArgName name
-  return $ El (getSort a `dLub` Abs y (getSort b))
+  return $ El (piSort (getSort a) (Abs y (getSort b)))
               (Pi (Dom info a) (Abs y b))
 
 hPi, nPi :: String -> TCM Type -> TCM Type -> TCM Type
@@ -657,9 +660,6 @@ el t = El (mkType 0) <$> t
 tset :: TCM Type
 tset = return $ sort (mkType 0)
 
-tSetOmega :: TCM Type
-tSetOmega = return $ sort Inf
-
 sSizeUniv :: Sort
 sSizeUniv = mkType 0
 -- Andreas, 2016-04-14 switching off SizeUniv, unfixing issue #1428
@@ -731,6 +731,10 @@ primitiveFunctions = Map.fromList
   , "primNatEquality"     |-> mkPrimFun2 ((==) :: Rel Nat)
   , "primNatLess"         |-> mkPrimFun2 ((<)  :: Rel Nat)
 
+  -- Machine words
+  , "primWord64ToNat"     |-> mkPrimFun1 (fromIntegral :: Word64 -> Nat)
+  , "primWord64FromNat"   |-> mkPrimFun1 (fromIntegral :: Nat -> Word64)
+
   -- Level functions
   , "primLevelZero"       |-> mkPrimLevelZero
   , "primLevelSuc"        |-> mkPrimLevelSuc
@@ -747,8 +751,9 @@ primitiveFunctions = Map.fromList
   -- because Haskell's Eq, which equates 0.0 and -0.0, allows to prove
   -- a contradiction (see Issue #2169).
   , "primFloatEquality"   |-> mkPrimFun2 (floatEq         :: Rel Double)
+  , "primFloatLess"       |-> mkPrimFun2 (floatLt         :: Rel Double)
   , "primFloatNumericalEquality" |-> mkPrimFun2 ((==)     :: Rel Double)
-  , "primFloatNumericalLess"     |-> mkPrimFun2 (floatLt  :: Rel Double)
+  , "primFloatNumericalLess"     |-> mkPrimFun2 ((<)      :: Rel Double)
   , "primFloatSqrt"       |-> mkPrimFun1 (sqrt            :: Double -> Double)
   , "primRound"           |-> mkPrimFun1 (round           :: Double -> Integer)
   , "primFloor"           |-> mkPrimFun1 (floor           :: Double -> Integer)
@@ -803,27 +808,6 @@ primitiveFunctions = Map.fromList
   where
     (|->) = (,)
 
-floatEq :: Double -> Double -> Bool
-floatEq x y = identicalIEEE x y || (isNaN x && isNaN y)
-
-floatLt :: Double -> Double -> Bool
-floatLt x y =
-  case compareFloat x y of
-    LT -> True
-    _  -> False
-  where
-    -- Also implemented in the GHC/UHC backends
-    compareFloat :: Double -> Double -> Ordering
-    compareFloat x y
-      | identicalIEEE x y          = EQ
-      | isNegInf x                 = LT
-      | isNegInf y                 = GT
-      | isNaN x && isNaN y         = EQ
-      | isNaN x                    = LT
-      | isNaN y                    = GT
-      | otherwise                  = compare x y
-    isNegInf z = z < 0 && isInfinite z
-
 lookupPrimitiveFunction :: String -> TCM PrimitiveImpl
 lookupPrimitiveFunction x =
   fromMaybe (typeError $ NoSuchPrimitiveFunction x)
@@ -842,7 +826,7 @@ getBuiltinName b = do
   where
   getName v = do
     v <- reduce v
-    case unSpine $ ignoreSharing v of
+    case unSpine $ v of
       Def x _   -> return x
       Con x _ _ -> return $ conName x
       Lam _ b   -> getName $ unAbs b
