@@ -116,7 +116,7 @@ unProjView pv =
 projView :: HasConstInfo m => Term -> m ProjectionView
 projView v = do
   let fallback = return $ NoProjection v
-  case ignoreSharing v of
+  case v of
     Def f es -> caseMaybeM (isProjection f) fallback $ \ isP -> do
       if projIndex isP <= 0 then fallback else do
         case es of
@@ -251,8 +251,9 @@ makeProjection x = -- if True then return () else do
         ifM recursive (reportSLn "tc.proj.like" 30 $ "  recursive functions are not considered for projection-likeness") $ do
           {- else -}
           case lastMaybe (filter (checkOccurs cls . snd) ps0) of
-            Nothing -> reportSLn "tc.proj.like" 50 $
-              "  occurs check failed\n    clauses = " ++ show cls
+            Nothing -> reportSDoc "tc.proj.like" 50 $ nest 2 $ vcat
+              [ text "occurs check failed"
+              , nest 2 $ text "clauses =" <?> vcat (map pretty cls) ]
             Just (d, n) -> do
               -- Yes, we are projection-like!
               reportSDoc "tc.proj.like" 10 $ sep
@@ -293,7 +294,16 @@ makeProjection x = -- if True then return () else do
       reportSLn "tc.proj.like" 30 $ "  abstract functions can't be projections"
     Function{funProjection = Just{}} ->
       reportSLn "tc.proj.like" 30 $ "  already projection like"
-    _ -> reportSLn "tc.proj.like" 30 $ "  not a function"
+    Function{funMutual = Just (_:_)} ->
+      reportSLn "tc.proj.like" 30 $ "  mutual functions can't be projections"
+    Function{funMutual = Nothing} ->
+      reportSLn "tc.proj.like" 30 $ "  mutuality check has not run yet"
+    Axiom          -> reportSLn "tc.proj.like" 30 $ "  not a function, but Axiom"
+    AbstractDefn{} -> reportSLn "tc.proj.like" 30 $ "  not a function, but AbstractDefn"
+    Constructor{}  -> reportSLn "tc.proj.like" 30 $ "  not a function, but Constructor"
+    Datatype{}     -> reportSLn "tc.proj.like" 30 $ "  not a function, but Datatype"
+    Primitive{}    -> reportSLn "tc.proj.like" 30 $ "  not a function, but Primitive"
+    Record{}       -> reportSLn "tc.proj.like" 30 $ "  not a function, but Record"
   where
     -- @validProj (d,n)@ checks whether the head @d@ of the type of the
     -- @n@th argument is injective in all args (i.d. being name of data/record/axiom).
@@ -339,7 +349,6 @@ makeProjection x = -- if True then return () else do
         noMatch ProjP{}= False
         noMatch VarP{} = True
         noMatch DotP{} = True
-        noMatch AbsurdP{} = True
 
     -- Make sure non of the parameters occurs in the body of the function.
     checkBody m n b = not . getAny $ runFree badVar IgnoreNot b
@@ -358,9 +367,9 @@ makeProjection x = -- if True then return () else do
     --
     candidateArgs :: [Term] -> Type -> [(Arg QName, Int)]
     candidateArgs vs t =
-      case ignoreSharing $ unEl t of
+      case unEl t of
         Pi a b
-          | Def d es <- ignoreSharing $ unEl $ unDom a,
+          | Def d es <- unEl $ unDom a,
             Just us  <- allApplyElims es,
             vs == map unArg us -> (d <$ argFromDom a, length vs) : candidateRec b
           | otherwise          -> candidateRec b
