@@ -2,7 +2,6 @@
 
 module Agda.TypeChecking.Quote where
 
-import Control.Applicative
 import Control.Arrow ((&&&))
 import Control.Monad
 import Control.Monad.State (runState, get, put)
@@ -41,6 +40,7 @@ import Agda.Utils.String ( Str(Str), unStr )
 import Agda.Utils.VarSet (VarSet)
 import qualified Agda.Utils.VarSet as Set
 import Agda.Utils.FileName
+import Agda.Utils.Size
 
 #include "undefined.h"
 
@@ -76,6 +76,7 @@ quotingKit = do
   meta            <- primAgdaTermMeta
   lit             <- primAgdaTermLit
   litNat          <- primAgdaLitNat
+  litWord64       <- primAgdaLitNat
   litFloat        <- primAgdaLitFloat
   litChar         <- primAgdaLitChar
   litString       <- primAgdaLitString
@@ -95,8 +96,8 @@ quotingKit = do
   sucLevel        <- primLevelSuc
   lub             <- primLevelMax
   lkit            <- requireLevels
-  Con z _ _       <- ignoreSharing <$> primZero
-  Con s _ _       <- ignoreSharing <$> primSuc
+  Con z _ _       <- primZero
+  Con s _ _       <- primSuc
   unsupported     <- primAgdaTermUnsupported
 
   agdaDefinitionFunDef          <- primAgdaDefinitionFunDef
@@ -124,13 +125,15 @@ quotingKit = do
       quoteRelevance Relevant   = pure relevant
       quoteRelevance Irrelevant = pure irrelevant
       quoteRelevance NonStrict  = pure relevant
-      quoteRelevance Forced{}   = pure relevant
 
+      -- TODO: quote Quanity
       quoteArgInfo :: ArgInfo -> ReduceM Term
-      quoteArgInfo (ArgInfo h r _) = arginfo !@ quoteHiding h @@ quoteRelevance r
+      quoteArgInfo (ArgInfo h m _ _) =
+        arginfo !@ quoteHiding h @@ quoteRelevance (getRelevance m)
 
       quoteLit :: Literal -> ReduceM Term
       quoteLit l@LitNat{}    = litNat    !@! Lit l
+      quoteLit l@LitWord64{} = litWord64 !@! Lit l
       quoteLit l@LitFloat{}  = litFloat  !@! Lit l
       quoteLit l@LitChar{}   = litChar   !@! Lit l
       quoteLit l@LitString{} = litString !@! Lit l
@@ -149,7 +152,9 @@ quotingKit = do
       quoteSort Prop     = pure unsupportedSort
       quoteSort Inf      = pure unsupportedSort
       quoteSort SizeUniv = pure unsupportedSort
-      quoteSort DLub{}   = pure unsupportedSort
+      quoteSort PiSort{} = pure unsupportedSort
+      quoteSort UnivSort{} = pure unsupportedSort
+      quoteSort (MetaS x es) = quoteTerm $ MetaV x es
 
       quoteType :: Type -> ReduceM Term
       quoteType (El _ t) = quoteTerm t
@@ -161,9 +166,10 @@ quotingKit = do
       quotePats ps = list $ map (quoteArg quotePat . fmap namedThing) ps
 
       quotePat :: DeBruijnPattern -> ReduceM Term
-      quotePat (AbsurdP x)       = pure absurdP
-      quotePat (VarP x)          = varP !@! quoteString (dbPatVarName x)
-      quotePat (DotP _)          = pure dotP
+      quotePat p
+       | patternOrigin p == Just PatOAbsurd = pure absurdP
+      quotePat (VarP o x)        = varP !@! quoteString (dbPatVarName x)
+      quotePat (DotP _ _)        = pure dotP
       quotePat (ConP c _ ps)     = conP !@ quoteQName (conName c) @@ quotePats ps
       quotePat (LitP l)          = litP !@ quoteLit l
       quotePat (ProjP _ x)       = projP !@ quoteQName x
@@ -211,31 +217,26 @@ quotingKit = do
             let
               conOrProjPars = defParameters defn
               ts = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
-              qx Function{ funExtLam = Just (ExtLamInfo h nh), funClauses = cs } = do
+              qx Function{ funExtLam = Just (ExtLamInfo m), funClauses = cs } = do
                     -- An extended lambda should not have any extra parameters!
                     unless (null conOrProjPars) __IMPOSSIBLE__
-                    -- Andreas, 2017-01-23 quoting Ulf
-                    -- "One would hope that @n@ includes the @h + nh@ parameters of the ext.lam."
-                    -- Let's see!
-                    -- unless (n >= h + nh) __IMPOSSIBLE__
-                    -- Actually, no, it does not!  ExtLam is not touched by module application.
-                    -- TODO: fixe me!  See #2404.
-                    extlam !@ list (map (quoteClause . (`apply` (take (h + nh) ts))) cs)
+                    n <- size <$> lookupSection m
+                    extlam !@ list (map (quoteClause . (`apply` (take n ts))) cs)
               qx Function{ funCompiled = Just Fail, funClauses = [cl] } =
                     extlam !@ list [quoteClause $ dropArgs (length (namedClausePats cl) - 1) cl]
               qx _ = def !@! quoteName x
             qx (theDef defn) @@ list (drop n $ conOrProjPars ++ map (quoteArg quoteTerm) ts)
-          Con x ci ts -> do
+          Con x ci es | Just ts <- allApplyElims es -> do
             cDef <- getConstInfo (conName x)
             n    <- getDefFreeVars (conName x)
             let args = list $ drop n $ defParameters cDef ++ map (quoteArg quoteTerm) ts
             con !@! quoteConName x @@ args
+          Con x ci es -> pure unsupported
           Pi t u     -> pi !@  quoteDom quoteType t
                             @@ quoteAbs quoteType u
           Level l    -> quoteTerm (unlevelWithKit lkit l)
           Lit l      -> lit !@ quoteLit l
           Sort s     -> sort !@ quoteSort s
-          Shared p   -> quoteTerm $ derefPtr p
           MetaV x es -> meta !@! quoteMeta currentFile x @@ quoteArgs vs
             where vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
           DontCare{} -> pure unsupported -- could be exposed at some point but we have to take care
