@@ -3,8 +3,12 @@
 
 module Agda.TypeChecking.Records where
 
--- import Control.Applicative
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ((<>))
+#endif
+
 import Control.Monad
+import Control.Monad.Trans.Maybe
 
 import Data.Function
 import qualified Data.List as List
@@ -42,6 +46,9 @@ import Agda.Utils.Size
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+mkCon :: ConHead -> ConInfo -> Args -> Term
+mkCon h info args = Con h info (map Apply args)
+
 -- | Order the fields of a record construction.
 --   Use the second argument for missing fields.
 orderFields :: QName -> a -> [C.Name] -> [(C.Name, a)] -> TCM [a]
@@ -147,9 +154,12 @@ getRecordFieldTypes :: QName -> TCM Telescope
 getRecordFieldTypes r = recTel <$> getRecordDef r
 
 -- | Get the field names belonging to a record type.
-getRecordTypeFields :: Type -> TCM [Arg QName]
-getRecordTypeFields t =
-  case ignoreSharing $ unEl t of
+getRecordTypeFields
+  :: Type  -- ^ Record type.  Need not be reduced.
+  -> TCM [Arg QName]
+getRecordTypeFields t = do
+  t <- reduce t  -- Andreas, 2018-03-03, fix for #2989.
+  case unEl t of
     Def r _ -> do
       rDef <- theDef <$> getConstInfo r
       case rDef of
@@ -183,10 +193,10 @@ isRecordType t = either (const Nothing) Just <$> tryRecordType t
 --   Succeeds only if type is not blocked by a meta var.
 --   If yes, return its name, parameters, and definition.
 --   If no, return the reduced type (unless it is blocked).
-tryRecordType :: Type -> TCM (Either (Maybe Type) (QName, Args, Defn))
-tryRecordType t = ifBlockedType t (\ _ _ -> return $ Left Nothing) $ \ t -> do
-  let no = return $ Left $ Just t
-  case ignoreSharing $ unEl t of
+tryRecordType :: Type -> TCM (Either (Blocked Type) (QName, Args, Defn))
+tryRecordType t = ifBlockedType t (\ m a -> return $ Left $ Blocked m a) $ \ nb t -> do
+  let no = return $ Left $ NotBlocked nb t
+  case unEl t of
     Def r es -> do
       let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
       caseMaybeM (isRecord r) no $ \ def -> return $ Right (r,vs,def)
@@ -238,7 +248,7 @@ getDefType f t = do
                 | otherwise = n - 1
       reportSLn "tc.deftype" 20 $ "projIndex    = " ++ show n
       -- we get the parameters from type @t@
-      case ignoreSharing $ unEl t of
+      case unEl t of
         Def d es -> do
           -- Andreas, 2013-10-22
           -- we need to check this @Def@ is fully reduced.
@@ -320,7 +330,7 @@ typeElims a self (e : es) = do
 {-# SPECIALIZE isEtaRecord :: QName -> TCM Bool #-}
 {-# SPECIALIZE isEtaRecord :: QName -> ReduceM Bool #-}
 isEtaRecord :: HasConstInfo m => QName -> m Bool
-isEtaRecord r = maybe False recEtaEquality <$> isRecord r
+isEtaRecord r = maybe False ((YesEta ==) . recEtaEquality) <$> isRecord r
 
 isEtaCon :: HasConstInfo m => QName -> m Bool
 isEtaCon c = getConstInfo' c >>= \case
@@ -332,11 +342,11 @@ isEtaCon c = getConstInfo' c >>= \case
 
 -- | Check if a name refers to a record which is not coinductive.  (Projections are then size-preserving)
 isInductiveRecord :: QName -> TCM Bool
-isInductiveRecord r = maybe False (\ d -> recInduction d /= Just CoInductive || not (recRecursive d)) <$> isRecord r
+isInductiveRecord r = maybe False (\ d -> recInduction d /= Just CoInductive) <$> isRecord r
 
 -- | Check if a type is an eta expandable record and return the record identifier and the parameters.
 isEtaRecordType :: Type -> TCM (Maybe (QName, Args))
-isEtaRecordType a = case ignoreSharing $ unEl a of
+isEtaRecordType a = case unEl a of
   Def d es -> do
     let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
     ifM (isEtaRecord d) (return $ Just (d, vs)) (return Nothing)
@@ -367,7 +377,7 @@ isGeneratedRecordConstructor c = ignoreAbstractMode $ do
 --   Projections do not preserve guardedness.
 unguardedRecord :: QName -> TCM ()
 unguardedRecord q = modifySignature $ updateDefinition q $ updateTheDef $ \case
-  r@Record{} -> r { recEtaEquality' = setEtaEquality (recEtaEquality' r) False }
+  r@Record{} -> r { recEtaEquality' = setEtaEquality (recEtaEquality' r) NoEta }
   _ -> __IMPOSSIBLE__
 
 -- | Turn on eta for inductive guarded recursive records.
@@ -379,7 +389,7 @@ recursiveRecord q = do
     r@Record{ recInduction = ind, recEtaEquality' = eta } ->
       r { recEtaEquality' = eta' }
       where
-      eta' | ok, eta == Inferred False, ind /= Just CoInductive = Inferred True
+      eta' | ok, eta == Inferred NoEta, ind /= Just CoInductive = Inferred YesEta
            | otherwise = eta
     _ -> __IMPOSSIBLE__
 
@@ -388,9 +398,9 @@ nonRecursiveRecord :: QName -> TCM ()
 nonRecursiveRecord q = whenM etaEnabled $ do
   -- Do nothing if eta is disabled by option.
   modifySignature $ updateDefinition q $ updateTheDef $ \case
-    r@Record{ recInduction = ind, recEtaEquality' = Inferred False }
+    r@Record{ recInduction = ind, recEtaEquality' = Inferred NoEta }
       | ind /= Just CoInductive ->
-      r { recEtaEquality' = Inferred True }
+      r { recEtaEquality' = Inferred YesEta }
     r@Record{} -> r
     _          -> __IMPOSSIBLE__
 
@@ -438,7 +448,7 @@ expandRecordVar i gamma0 = do
           text " is not a record type"
         return Nothing
   caseMaybeM (isRecordType a) failure $ \ (r, pars, def) -> do
-    if not (recEtaEquality def) then return Nothing else Just <$> do
+    if recEtaEquality def == NoEta then return Nothing else Just <$> do
       -- Get the record fields @Γ₁ ⊢ tel@ (@tel = Γ'@).
       -- TODO: compose argInfo ai with tel.
       let tel = recTel def `apply` pars
@@ -446,7 +456,7 @@ expandRecordVar i gamma0 = do
           fs  = recFields def
       -- Construct the record pattern @Γ₁, Γ' ⊢ u := c ys@.
           ys  = zipWith (\ f i -> f $> var i) fs $ downFrom m
-          u   = Con (recConHead def) ConOSystem ys
+          u   = mkCon (recConHead def) ConOSystem ys
       -- @Γ₁, Γ' ⊢ τ₀ : Γ₁, x:_@
           tau0 = consS u $ raiseS m
       -- @Γ₁, Γ', Γ₂ ⊢ τ₀ : Γ₁, x:_, Γ₂@
@@ -498,7 +508,7 @@ curryAt :: Type -> Int -> TCM (Term -> Term, Term -> Term, Type)
 curryAt t n = do
   -- first, strip the leading n domains (which remain unchanged)
   TelV gamma core <- telViewUpTo n t
-  case ignoreSharing $ unEl core of
+  case unEl core of
     -- There should be at least one domain left
     Pi (Dom ai a) b -> do
       -- Eta-expand @dom@ along @qs@ into a telescope @tel@, computing a substitution.
@@ -506,13 +516,13 @@ curryAt t n = do
       -- This might trigger another call to @etaExpandProjectedVar@ later.
       -- A more efficient version does all the eta-expansions at once here.
       (r, pars, def) <- fromMaybe __IMPOSSIBLE__ <$> isRecordType a
-      unless (recEtaEquality def) __IMPOSSIBLE__
+      when (recEtaEquality def == NoEta) __IMPOSSIBLE__
       -- TODO: compose argInfo ai with tel.
       let tel = recTel def `apply` pars
           m   = size tel
           fs  = recFields def
           ys  = zipWith (\ f i -> f $> var i) fs $ downFrom m
-          u   = Con (recConHead def) ConOSystem ys
+          u   = mkCon (recConHead def) ConOSystem ys
           b'  = raise m b `absApp` u
           t'  = gamma `telePi` (tel `telePi` b')
           gammai = map domInfo $ telToList gamma
@@ -562,13 +572,14 @@ etaExpandRecord'_ forceEta r pars def u = do
             , recFields      = xs
             , recTel         = tel
             } = def
-      eta = recEtaEquality def
       tel' = apply tel pars
-  unless (eta || forceEta) __IMPOSSIBLE__ -- make sure we do not expand non-eta records (unless forced to)
-  case ignoreSharing u of
+  -- Make sure we do not expand non-eta records (unless forced to):
+  unless (recEtaEquality def == YesEta || forceEta) __IMPOSSIBLE__
+  case u of
 
     -- Already expanded.
-    Con con_ ci args -> do
+    Con con_ ci es -> do
+      let args = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
       when (con /= con_) $ do
         reportSDoc "impossible" 10 $ vcat
           [ text "etaExpandRecord_: the following two constructors should be identical"
@@ -596,16 +607,19 @@ etaExpandAtRecordType :: Type -> Term -> TCM (Telescope, Term)
 etaExpandAtRecordType t u = do
   (r, pars, def) <- fromMaybe __IMPOSSIBLE__ <$> isRecordType t
   (tel, con, ci, args) <- etaExpandRecord_ r pars def u
-  return (tel, Con con ci args)
+  return (tel, mkCon con ci args)
 
 -- | The fields should be eta contracted already.
 --
 --   We can eta contract if all fields @f = ...@ are irrelevant
 --   or all fields @f@ are the projection @f v@ of the same value @v@,
 --   but we need at least one relevant field to find the value @v@.
---
---   TODO: this can be moved out of TCM (but only if ConHead
---   stores also the Arg-decoration of the record fields.
+
+--   TODO: this can be moved out of TCM.
+--   Andreas, 2018-01-28: attempted just that, but Auto does not
+--   put the conFields there (it does not run in TCM).
+--   If we get rid of Auto, we can do this.  (Tests not involving Auto pass.)
+
 {-# SPECIALIZE etaContractRecord :: QName -> ConHead -> ConInfo -> Args -> TCM Term #-}
 {-# SPECIALIZE etaContractRecord :: QName -> ConHead -> ConInfo -> Args -> ReduceM Term #-}
 etaContractRecord :: HasConstInfo m => QName -> ConHead -> ConInfo -> Args -> m Term
@@ -623,7 +637,7 @@ etaContractRecord r c ci args = do
           (_, Just (h, es)) | Proj _o f <- last es, unArg ax == f
                             -> Just $ Just $ h $ init es
           _                 -> Nothing
-      fallBack = return (Con c ci args)
+      fallBack = return (mkCon c ci args)
   case compare (length args) (length xs) of
     LT -> fallBack       -- Not fully applied
     GT -> __IMPOSSIBLE__ -- Too many arguments. Impossible.
@@ -655,7 +669,7 @@ isSingletonRecord' :: Bool -> QName -> Args -> TCM (Either MetaId (Maybe Term))
 isSingletonRecord' regardIrrelevance r ps = do
   reportSLn "tc.meta.eta" 30 $ "Is " ++ prettyShow r ++ " a singleton record type?"
   def <- getRecordDef r
-  emap (Con (recConHead def) ConOSystem) <$> check (recTel def `apply` ps)
+  emap (mkCon (recConHead def) ConOSystem) <$> check (recTel def `apply` ps)
   where
   check :: Telescope -> TCM (Either MetaId (Maybe [Arg Term]))
   check tel = do
@@ -691,13 +705,80 @@ isSingletonTypeModuloRelevance t = liftTCM $ do
 isSingletonType' :: Bool -> Type -> TCM (Either MetaId (Maybe Term))
 isSingletonType' regardIrrelevance t = do
     TelV tel t <- telView t
-    ifBlockedType t (\ m _ -> return $ Left m) $ \ t -> do
+    ifBlockedType t (\ m _ -> return $ Left m) $ \ _ t -> do
       res <- isRecordType t
       case res of
-        Just (r, ps, def) | recEtaEquality def -> do
+        Just (r, ps, def) | YesEta <- recEtaEquality def -> do
           emap (abstract tel) <$> isSingletonRecord' regardIrrelevance r ps
         _ -> return $ Right Nothing
 
+-- | Checks whether the given term (of the given type) is beta-eta-equivalent
+--   to a variable. Returns just the de Bruijn-index of the variable if it is,
+--   or nothing otherwise.
+isEtaVar :: Term -> Type -> TCM (Maybe Int)
+isEtaVar u a = runMaybeT $ isEtaVarG u a Nothing []
+  where
+    -- Checks whether the term u (of type a) is beta-eta-equivalent to
+    -- `Var i es`, and returns i if it is. If the argument mi is `Just i'`,
+    -- then i and i' are also required to be equal (else Nothing is returned).
+    isEtaVarG :: Term -> Type -> Maybe Int -> [Elim' Int] -> MaybeT TCM Int
+    isEtaVarG u a mi es = do
+      (u, a) <- liftTCM $ reduce (u, a)
+      liftTCM $ reportSDoc "tc.lhs" 80 $ text "isEtaVarG" <+> nest 2 (vcat
+        [ text "u  = " <+> text (show u)
+        , text "a  = " <+> prettyTCM a
+        , text "mi = " <+> text (show mi)
+        , text "es = " <+> prettyList (map (text . show) es)
+        ])
+      case (u, unEl a) of
+        (Var i' es', _) -> do
+          guard $ mi == (i' <$ mi)
+          b <- liftTCM $ typeOfBV i'
+          areEtaVarElims (var i') b es' es
+          return i'
+        (_, Def d pars) -> do
+          guard =<< do liftTCM $ isEtaRecord d
+          fs <- liftTCM $ map unArg . recFields . theDef <$> getConstInfo d
+          is <- forM fs $ \f -> do
+            let o = ProjSystem
+            (_, _, fa) <- MaybeT $ projectTyped u a o f
+            isEtaVarG (u `applyE` [Proj o f]) fa mi (es ++ [Proj o f])
+          case (mi, is) of
+            (Just i, _)     -> return i
+            (Nothing, [])   -> mzero
+            (Nothing, i:is) -> guard (all (==i) is) >> return i
+        (_, Pi dom cod) -> addContext dom $ do
+          let u'  = raise 1 u `apply` [argFromDom dom $> var 0]
+              a'  = absBody cod
+              mi' = fmap (+1) mi
+              es' = (fmap . fmap) (+1) es ++ [Apply $ argFromDom dom $> 0]
+          (-1+) <$> isEtaVarG u' a' mi' es'
+        _ -> mzero
+
+    -- `areEtaVarElims u a es es'` checks whether the given elims es (as applied
+    -- to the term u of type a) are beta-eta-equal to either projections or
+    -- variables with de Bruijn indices given by es'.
+    areEtaVarElims :: Term -> Type -> Elims -> [Elim' Int] -> MaybeT TCM ()
+    areEtaVarElims u a []    []    = return ()
+    areEtaVarElims u a []    (_:_) = mzero
+    areEtaVarElims u a (_:_) []    = mzero
+    areEtaVarElims u a (Proj o f : es) (Proj _ f' : es') = do
+      guard $ f == f'
+      a       <- liftTCM $ reduce a
+      (_, _, fa) <- MaybeT $ projectTyped u a o f
+      areEtaVarElims (u `applyE` [Proj o f]) fa es es'
+    -- These two cases can occur only when we're looking at two different
+    -- variables (i.e. one of function type and the other of record type) so
+    -- it's definitely not the variable we're looking for (or someone is playing
+    -- Jedi mind tricks on us)
+    areEtaVarElims u a (Proj{}  : _ ) (Apply _ : _  ) = mzero
+    areEtaVarElims u a (Apply _ : _ ) (Proj{}  : _  ) = mzero
+    areEtaVarElims u a (Apply v : es) (Apply i : es') = do
+      ifNotPiType a (const mzero) $ \dom cod -> do
+      _ <- isEtaVarG (unArg v) (unDom dom) (Just $ unArg i) []
+      areEtaVarElims (u `apply` [fmap var i]) (cod `absApp` var (unArg i)) es es'
+
+
 -- | Auxiliary function.
 emap :: (a -> b) -> Either c (Maybe a) -> Either c (Maybe b)
 emap = mapRight . fmap
@@ -724,7 +805,6 @@ instance NormaliseProjP a => NormaliseProjP (Named_ a) where
 instance NormaliseProjP (Pattern' x) where
   normaliseProjP p@VarP{}        = return p
   normaliseProjP p@DotP{}        = return p
-  normaliseProjP p@AbsurdP{}     = return p
   normaliseProjP (ConP c cpi ps) = ConP c cpi <$> normaliseProjP ps
   normaliseProjP p@LitP{}        = return p
   normaliseProjP (ProjP o d0)    = ProjP o <$> getOriginalProjection d0
