@@ -6,9 +6,9 @@ module Agda.TypeChecking.Reduce where
 
 import Prelude hiding (mapM)
 import Control.Monad.Reader hiding (mapM)
-import Control.Applicative
 
 import qualified Data.List as List
+import Data.List ((\\))
 import Data.Maybe
 import Data.Map (Map)
 import Data.Traversable
@@ -42,6 +42,7 @@ import {-# SOURCE #-} Agda.TypeChecking.Reduce.Fast
 
 import Agda.Utils.Function
 import Agda.Utils.Functor
+import Agda.Utils.Lens
 import Agda.Utils.Monad
 import Agda.Utils.HashMap (HashMap)
 import Agda.Utils.Size
@@ -105,9 +106,7 @@ instance Instantiate Term where
       BlockedConst _                   -> return t
       PostponedTypeCheckingProblem _ _ -> return t
   instantiate' (Level l) = levelTm <$> instantiate' l
-  instantiate' (Sort s) = sortTm <$> instantiate' s
-  instantiate' v@Shared{} =
-    updateSharedTerm instantiate' v
+  instantiate' (Sort s) = Sort <$> instantiate' s
   instantiate' t = return t
 
 instance Instantiate Level where
@@ -121,7 +120,7 @@ instance Instantiate LevelAtom where
   instantiate' l = case l of
     MetaLevel m vs -> do
       v <- instantiate' (MetaV m vs)
-      case ignoreSharing v of
+      case v of
         MetaV m vs -> return $ MetaLevel m vs
         _          -> return $ UnreducedLevel v
     UnreducedLevel l -> UnreducedLevel <$> instantiate' l
@@ -143,8 +142,11 @@ instance Instantiate Type where
 
 instance Instantiate Sort where
   instantiate' s = case s of
-    Type l -> levelSort <$> instantiate' l
-    _      -> return s
+    MetaS x es -> instantiate' (MetaV x es) >>= \case
+      Sort s'      -> return s'
+      MetaV x' es' -> return $ MetaS x' es'
+      _            -> __IMPOSSIBLE__
+    _ -> return s
 
 instance Instantiate Elim where
   instantiate' (Apply v) = Apply <$> instantiate' v
@@ -183,8 +185,8 @@ instance Instantiate Constraint where
   instantiate' (ValueCmp cmp t u v) = do
     (t,u,v) <- instantiate' (t,u,v)
     return $ ValueCmp cmp t u v
-  instantiate' (ElimCmp cmp t v as bs) =
-    ElimCmp cmp <$> instantiate' t <*> instantiate' v <*> instantiate' as <*> instantiate' bs
+  instantiate' (ElimCmp cmp fs t v as bs) =
+    ElimCmp cmp fs <$> instantiate' t <*> instantiate' v <*> instantiate' as <*> instantiate' bs
   instantiate' (LevelCmp cmp u v)   = uncurry (LevelCmp cmp) <$> instantiate' (u,v)
   instantiate' (TypeCmp cmp a b)    = uncurry (TypeCmp cmp) <$> instantiate' (a,b)
   instantiate' (TelCmp a b cmp tela telb) = uncurry (TelCmp a b cmp)  <$> instantiate' (tela,telb)
@@ -194,6 +196,9 @@ instance Instantiate Constraint where
   instantiate' (FindInScope m b args) = FindInScope m b <$> mapM instantiate' args
   instantiate' (IsEmpty r t)        = IsEmpty r <$> instantiate' t
   instantiate' (CheckSizeLtSat t)   = CheckSizeLtSat <$> instantiate' t
+  instantiate' c@CheckFunDef{}      = return c
+  instantiate' (HasBiggerSort a)    = HasBiggerSort <$> instantiate' a
+  instantiate' (HasPTSRule a b)     = uncurry HasPTSRule <$> instantiate' (a,b)
 
 instance Instantiate e => Instantiate (Map k e) where
     instantiate' = traverse instantiate'
@@ -218,18 +223,24 @@ instance Instantiate EqualityView where
 
 -- | Case on whether a term is blocked on a meta (or is a meta).
 --   That means it can change its shape when the meta is instantiated.
-ifBlocked :: MonadTCM tcm =>  Term -> (MetaId -> Term -> tcm a) -> (Term -> tcm a) -> tcm a
+ifBlocked :: MonadTCM tcm =>  Term -> (MetaId -> Term -> tcm a) -> (NotBlocked -> Term -> tcm a) -> tcm a
 ifBlocked t blocked unblocked = do
   t <- liftTCM $ reduceB t
-  case ignoreSharing <$> t of
+  case t of
     Blocked m _              -> blocked m (ignoreBlocking t)
     NotBlocked _ (MetaV m _) -> blocked m (ignoreBlocking t)
-    NotBlocked{}             -> unblocked (ignoreBlocking t)
+    NotBlocked nb _          -> unblocked nb (ignoreBlocking t)
+
+isBlocked :: MonadTCM tcm => Term -> tcm (Maybe MetaId)
+isBlocked t = ifBlocked t (\m _ -> return $ Just m) (\_ _ -> return Nothing)
 
 -- | Case on whether a type is blocked on a meta (or is a meta).
-ifBlockedType :: MonadTCM tcm => Type -> (MetaId -> Type -> tcm a) -> (Type -> tcm a) -> tcm a
+ifBlockedType :: MonadTCM tcm => Type -> (MetaId -> Type -> tcm a) -> (NotBlocked -> Type -> tcm a) -> tcm a
 ifBlockedType (El s t) blocked unblocked =
-  ifBlocked t (\ m v -> blocked m $ El s v) (\ v -> unblocked $ El s v)
+  ifBlocked t (\ m v -> blocked m $ El s v) (\ nb v -> unblocked nb $ El s v)
+
+isBlockedType :: MonadTCM tcm => Type -> tcm (Maybe MetaId)
+isBlockedType t = ifBlockedType t (\m _ -> return $ Just m) (\_ _ -> return Nothing)
 
 class Reduce t where
     reduce'  :: t -> ReduceM t
@@ -249,15 +260,17 @@ instance Reduce Sort where
         red s = do
           s <- instantiate' s
           case s of
-            DLub s1 s2 -> do
-              s <- dLub <$> reduce' s1 <*> reduce' s2
-              case s of
-                DLub{}  -> return s
-                _       -> reduce' s   -- TODO: not so nice
+            PiSort s1 s2 -> do
+              (s1,s2) <- reduce' (s1,s2)
+              maybe (return $ PiSort s1 s2) reduce' $ piSort' s1 s2
+            UnivSort s' -> do
+              s' <- reduce' s'
+              maybe (return $ UnivSort s') reduce' $ univSort' s'
             Prop       -> return s
-            Type s'    -> levelSort <$> reduce' s'
+            Type s'    -> Type <$> reduce' s'
             Inf        -> return Inf
             SizeUniv   -> return SizeUniv
+            MetaS x es -> return s
 
 instance Reduce Elim where
   reduce' (Apply v) = Apply <$> reduce' v
@@ -282,7 +295,7 @@ instance Reduce LevelAtom where
       fromTm v = do
         bv <- reduceB' v
         let v = ignoreBlocking bv
-        case ignoreSharing <$> bv of
+        case bv of
           NotBlocked r (MetaV m vs) -> return $ NotBlocked r $ MetaLevel m vs
           Blocked m _               -> return $ Blocked m    $ BlockedLevel m v
           NotBlocked r _            -> return $ NotBlocked r $ NeutralLevel r v
@@ -298,8 +311,9 @@ instance Reduce t => Reduce [t] where
 
 instance Reduce t => Reduce (Arg t) where
     reduce' a = case getRelevance a of
-                 Irrelevant -> return a             -- Don't reduce' irr. args!?
-                 _          -> traverse reduce' a
+      Irrelevant -> return a             -- Don't reduce' irr. args!?
+                                         -- Andreas, 2018-03-03, caused #2989.
+      _          -> traverse reduce' a
 
     reduceB' t = traverse id <$> traverse reduceB' t
 
@@ -317,18 +331,28 @@ instance (Reduce a, Reduce b,Reduce c) => Reduce (a,b,c) where
 instance Reduce Term where
   reduceB' = {-# SCC "reduce'<Term>" #-} maybeFastReduceTerm
 
+shouldTryFastReduce :: ReduceM Bool
+shouldTryFastReduce = do
+  allowed <- asks envAllowedReductions
+  let optionalReductions = [NonTerminatingReductions, UnconfirmedReductions]
+      requiredReductions = allReductions \\ optionalReductions
+  return $ (allowed \\ optionalReductions) == requiredReductions
+
 maybeFastReduceTerm :: Term -> ReduceM (Blocked Term)
 maybeFastReduceTerm v = do
   let tryFast = case v of
-                  Def{} -> True
-                  Con{} -> True
-                  _     -> False
+                  Def{}   -> True
+                  Con{}   -> True
+                  MetaV{} -> True
+                  _       -> False
   if not tryFast then slowReduceTerm v
-                 else do
-    s <- optSharing   <$> commandLineOptions
-    allowed <- asks envAllowedReductions
-    let notAll = List.delete NonTerminatingReductions allowed /= allReductions
-    if s || notAll then slowReduceTerm v else fastReduce (elem NonTerminatingReductions allowed) v
+                 else
+    case v of
+      MetaV x _ -> ifM (isOpen x) (return $ notBlocked v) (maybeFast v)
+      _         -> maybeFast v
+  where
+    isOpen x = isOpenMeta . mvInstantiation <$> lookupMeta x
+    maybeFast v = ifM shouldTryFastReduce (fastReduce v) (slowReduceTerm v)
 
 slowReduceTerm :: Term -> ReduceM (Blocked Term)
 slowReduceTerm v = do
@@ -340,12 +364,13 @@ slowReduceTerm v = do
 --      MetaV x args -> notBlocked . MetaV x <$> reduce' args
       MetaV x es -> done
       Def f es   -> unfoldDefinitionE False reduceB' (Def f []) f es
-      Con c ci args -> do
+      Con c ci es -> do
+          let args = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
           -- Constructors can reduce' when they come from an
           -- instantiated module.
           v <- unfoldDefinition False reduceB' (Con c ci []) (conName c) args
           traverse reduceNat v
-      Sort s   -> fmap sortTm <$> reduceB' s
+      Sort s   -> fmap Sort <$> reduceB' s
       Level l  -> ifM (elem LevelReductions <$> asks envAllowedReductions)
                     {- then -} (fmap levelTm <$> reduceB' l)
                     {- else -} done
@@ -354,24 +379,22 @@ slowReduceTerm v = do
       Var _ _  -> done
       Lam _ _  -> done
       DontCare _ -> done
-      Shared{}   -> updateSharedTermF reduceB' v
     where
       -- NOTE: reduceNat can traverse the entire term.
-      reduceNat v@Shared{} = updateSharedTerm reduceNat v
       reduceNat v@(Con c ci []) = do
         mz  <- getBuiltin' builtinZero
         case v of
           _ | Just v == mz  -> return $ Lit $ LitNat (getRange c) 0
           _                 -> return v
-      reduceNat v@(Con c ci [a]) | visible a && isRelevant a = do
-        ms  <- fmap ignoreSharing <$> getBuiltin' builtinSuc
+      reduceNat v@(Con c ci [Apply a]) | visible a && isRelevant a = do
+        ms  <- getBuiltin' builtinSuc
         case v of
           _ | Just (Con c ci []) == ms -> inc <$> reduce' (unArg a)
           _                         -> return v
           where
-            inc w = case ignoreSharing w of
+            inc w = case w of
               Lit (LitNat r n) -> Lit (LitNat (fuseRange c r) $ n + 1)
-              _                -> Con c ci [defaultArg w]
+              _                -> Con c ci [Apply $ defaultArg w]
       reduceNat v = return v
 
 -- Andreas, 2013-03-20 recursive invokations of unfoldCorecursion
@@ -384,12 +407,8 @@ unfoldCorecursionE (Apply (Arg info v)) = fmap (Apply . Arg info) <$>
 unfoldCorecursion :: Term -> ReduceM (Blocked Term)
 unfoldCorecursion v = do
   v <- instantiate' v
-  case compressPointerChain v of
+  case v of
     Def f es -> unfoldDefinitionE True unfoldCorecursion (Def f []) f es
-    v@(Shared p) ->
-      case derefPtr p of
-        Def{} -> updateSharedFM unfoldCorecursion v
-        _     -> slowReduceTerm v
     _ -> slowReduceTerm v
 
 -- | If the first argument is 'True', then a single delayed clause may
@@ -494,7 +513,7 @@ unfoldDefinitionStep unfoldDelayed v0 f es =
           debugReduce ev
           return ev
       where
-      defaultResult = noReduction $ NotBlocked AbsurdMatch vfull
+      defaultResult = noReduction $ NotBlocked ReallyNotBlocked vfull
       vfull         = v0 `applyE` map ignoreReduced es
       debugReduce ev = verboseS "tc.reduce" 90 $ do
         case ev of
@@ -542,7 +561,7 @@ reduceHead' v = do -- ignoreAbstractMode $ do
   -- first, possibly rewrite literal v to constructor form
   v <- constructorForm v
   traceSDoc "tc.inj.reduce" 30 (text "reduceHead" <+> prettyTCM v) $ do
-  case ignoreSharing v of
+  case v of
     Def f es -> do
 
       abstractMode <- envAbstractMode <$> ask
@@ -568,6 +587,24 @@ reduceHead' v = do -- ignoreAbstractMode $ do
         _                               -> return $ notBlocked v
     _ -> return $ notBlocked v
 
+-- | Unfold a single inlined function.
+unfoldInlined :: Term -> TCM Term
+unfoldInlined = runReduceM . unfoldInlined'
+
+unfoldInlined' :: Term -> ReduceM Term
+unfoldInlined' v = do
+  inTypes <- view eWorkingOnTypes
+  case v of
+    _ | inTypes -> return v -- Don't inline in types (to avoid unfolding of goals)
+    Def f es -> do
+      def <- theDef <$> getConstInfo f
+      case def of   -- Only for simple definitions with no pattern matching (TODO: maybe copatterns?)
+        Function{ funCompiled = Just Done{}, funDelayed = NotDelayed }
+          | def ^. funInline ->
+          ignoreBlocking <$> unfoldDefinitionE False (return . notBlocked) (Def f []) f es
+        _ -> return v
+    _ -> return v
+
 -- | Apply a definition using the compiled clauses, or fall back to
 --   ordinary clauses if no compiled clauses exist.
 appDef_ :: QName -> Term -> [Clause] -> Maybe CompiledClauses -> RewriteRules -> MaybeReducedArgs -> ReduceM (Reduced (Blocked Term) Term)
@@ -644,8 +681,8 @@ instance Reduce Constraint where
   reduce' (ValueCmp cmp t u v) = do
     (t,u,v) <- reduce' (t,u,v)
     return $ ValueCmp cmp t u v
-  reduce' (ElimCmp cmp t v as bs) =
-    ElimCmp cmp <$> reduce' t <*> reduce' v <*> reduce' as <*> reduce' bs
+  reduce' (ElimCmp cmp fs t v as bs) =
+    ElimCmp cmp fs <$> reduce' t <*> reduce' v <*> reduce' as <*> reduce' bs
   reduce' (LevelCmp cmp u v)    = uncurry (LevelCmp cmp) <$> reduce' (u,v)
   reduce' (TypeCmp cmp a b)     = uncurry (TypeCmp cmp) <$> reduce' (a,b)
   reduce' (TelCmp a b cmp tela telb) = uncurry (TelCmp a b cmp)  <$> reduce' (tela,telb)
@@ -655,6 +692,9 @@ instance Reduce Constraint where
   reduce' (FindInScope m b cands) = FindInScope m b <$> mapM reduce' cands
   reduce' (IsEmpty r t)         = IsEmpty r <$> reduce' t
   reduce' (CheckSizeLtSat t)    = CheckSizeLtSat <$> reduce' t
+  reduce' c@CheckFunDef{}       = return c
+  reduce' (HasBiggerSort a)     = HasBiggerSort <$> reduce' a
+  reduce' (HasPTSRule a b)      = uncurry HasPTSRule <$> reduce' (a,b)
 
 instance Reduce e => Reduce (Map k e) where
     reduce' = traverse reduce'
@@ -689,7 +729,7 @@ instance Simplify Term where
       Def f vs   -> do
         let keepGoing simp v = return (simp, notBlocked v)
         (simpl, v) <- unfoldDefinition' False keepGoing (Def f []) f vs
-        traceSDoc "tc.simplify'" 20 (
+        traceSDoc "tc.simplify'" 90 (
           text ("simplify': unfolding definition returns " ++ show simpl)
             <+> prettyTCM (ignoreBlocking v)) $ do
         case simpl of
@@ -697,14 +737,13 @@ instance Simplify Term where
           NoSimplification  -> Def f <$> simplify' vs
       MetaV x vs -> MetaV x  <$> simplify' vs
       Con c ci vs-> Con c ci <$> simplify' vs
-      Sort s     -> sortTm   <$> simplify' s
+      Sort s     -> Sort     <$> simplify' s
       Level l    -> levelTm  <$> simplify' l
       Pi a b     -> Pi       <$> simplify' a <*> simplify' b
       Lit l      -> return v
       Var i vs   -> Var i    <$> simplify' vs
       Lam h v    -> Lam h    <$> simplify' v
       DontCare v -> dontCare <$> simplify' v
-      Shared{}   -> updateSharedTerm simplify' v
 
 simplifyBlocked' :: Simplify t => Blocked t -> ReduceM t
 simplifyBlocked' (Blocked _ t) = return t
@@ -720,11 +759,13 @@ instance Simplify Elim where
 instance Simplify Sort where
     simplify' s = do
       case s of
-        DLub s1 s2 -> dLub <$> simplify' s1 <*> simplify' s2
-        Type s     -> levelSort <$> simplify' s
+        PiSort s1 s2 -> piSort <$> simplify' s1 <*> simplify' s2
+        UnivSort s -> univSort <$> simplify' s
+        Type s     -> Type <$> simplify' s
         Prop       -> return s
         Inf        -> return s
         SizeUniv   -> return s
+        MetaS x es -> MetaS x <$> simplify' es
 
 instance Simplify Level where
   simplify' (Max as) = levelMax <$> simplify' as
@@ -788,8 +829,8 @@ instance Simplify Constraint where
   simplify' (ValueCmp cmp t u v) = do
     (t,u,v) <- simplify' (t,u,v)
     return $ ValueCmp cmp t u v
-  simplify' (ElimCmp cmp t v as bs) =
-    ElimCmp cmp <$> simplify' t <*> simplify' v <*> simplify' as <*> simplify' bs
+  simplify' (ElimCmp cmp fs t v as bs) =
+    ElimCmp cmp fs <$> simplify' t <*> simplify' v <*> simplify' as <*> simplify' bs
   simplify' (LevelCmp cmp u v)    = uncurry (LevelCmp cmp) <$> simplify' (u,v)
   simplify' (TypeCmp cmp a b)     = uncurry (TypeCmp cmp) <$> simplify' (a,b)
   simplify' (TelCmp a b cmp tela telb) = uncurry (TelCmp a b cmp) <$> simplify' (tela,telb)
@@ -799,6 +840,9 @@ instance Simplify Constraint where
   simplify' (FindInScope m b cands) = FindInScope m b <$> mapM simplify' cands
   simplify' (IsEmpty r t)         = IsEmpty r <$> simplify' t
   simplify' (CheckSizeLtSat t)    = CheckSizeLtSat <$> simplify' t
+  simplify' c@CheckFunDef{}       = return c
+  simplify' (HasBiggerSort a)     = HasBiggerSort <$> simplify' a
+  simplify' (HasPTSRule a b)      = uncurry HasPTSRule <$> simplify' (a,b)
 
 instance Simplify Bool where
   simplify' = return
@@ -844,31 +888,32 @@ instance Normalise Sort where
     normalise' s = do
       s <- reduce' s
       case s of
-        DLub s1 s2 -> dLub <$> normalise' s1 <*> normalise' s2
+        PiSort s1 s2 -> piSort <$> normalise' s1 <*> normalise' s2
+        UnivSort s -> univSort <$> normalise' s
         Prop       -> return s
-        Type s     -> levelSort <$> normalise' s
+        Type s     -> Type <$> normalise' s
         Inf        -> return Inf
         SizeUniv   -> return SizeUniv
+        MetaS x es -> return s
 
 instance Normalise Type where
     normalise' (El s t) = El <$> normalise' s <*> normalise' t
 
 instance Normalise Term where
-    normalise' = ignoreBlocking <.> (reduceB' >=> traverse normaliseArgs)
-      where
-        normaliseArgs :: Term -> ReduceM Term
-        normaliseArgs v = case v of
-                Var n vs    -> Var n <$> normalise' vs
-                Con c ci vs -> Con c ci <$> normalise' vs
-                Def f vs    -> Def f <$> normalise' vs
-                MetaV x vs  -> MetaV x <$> normalise' vs
-                Lit _       -> return v
-                Level l     -> levelTm <$> normalise' l
-                Lam h b     -> Lam h <$> normalise' b
-                Sort s      -> sortTm <$> normalise' s
-                Pi a b      -> uncurry Pi <$> normalise' (a,b)
-                Shared{}    -> updateSharedTerm normalise' v
-                DontCare _  -> return v
+    normalise' v = ifM shouldTryFastReduce (fastNormalise v) (slowNormaliseArgs =<< reduce' v)
+
+slowNormaliseArgs :: Term -> ReduceM Term
+slowNormaliseArgs v = case v of
+  Var n vs    -> Var n      <$> normalise' vs
+  Con c ci vs -> Con c ci   <$> normalise' vs
+  Def f vs    -> Def f      <$> normalise' vs
+  MetaV x vs  -> MetaV x    <$> normalise' vs
+  Lit _       -> return v
+  Level l     -> levelTm    <$> normalise' l
+  Lam h b     -> Lam h      <$> normalise' b
+  Sort s      -> Sort       <$> normalise' s
+  Pi a b      -> uncurry Pi <$> normalise' (a, b)
+  DontCare _  -> return v
 
 instance Normalise Elim where
   normalise' (Apply v) = Apply <$> normalise' v
@@ -931,8 +976,8 @@ instance Normalise Constraint where
   normalise' (ValueCmp cmp t u v) = do
     (t,u,v) <- normalise' (t,u,v)
     return $ ValueCmp cmp t u v
-  normalise' (ElimCmp cmp t v as bs) =
-    ElimCmp cmp <$> normalise' t <*> normalise' v <*> normalise' as <*> normalise' bs
+  normalise' (ElimCmp cmp fs t v as bs) =
+    ElimCmp cmp fs <$> normalise' t <*> normalise' v <*> normalise' as <*> normalise' bs
   normalise' (LevelCmp cmp u v)    = uncurry (LevelCmp cmp) <$> normalise' (u,v)
   normalise' (TypeCmp cmp a b)     = uncurry (TypeCmp cmp) <$> normalise' (a,b)
   normalise' (TelCmp a b cmp tela telb) = uncurry (TelCmp a b cmp) <$> normalise' (tela,telb)
@@ -942,6 +987,9 @@ instance Normalise Constraint where
   normalise' (FindInScope m b cands) = FindInScope m b <$> mapM normalise' cands
   normalise' (IsEmpty r t)         = IsEmpty r <$> normalise' t
   normalise' (CheckSizeLtSat t)    = CheckSizeLtSat <$> normalise' t
+  normalise' c@CheckFunDef{}       = return c
+  normalise' (HasBiggerSort a)     = HasBiggerSort <$> normalise' a
+  normalise' (HasPTSRule a b)      = uncurry HasPTSRule <$> normalise' (a,b)
 
 instance Normalise Bool where
   normalise' = return
@@ -953,18 +1001,17 @@ instance Normalise Char where
   normalise' = return
 
 instance Normalise ConPatternInfo where
-  normalise' (ConPatternInfo mr mt) = ConPatternInfo mr <$> normalise' mt
+  normalise' i = normalise' (conPType i) <&> \ t -> i { conPType = t }
 
 instance Normalise DBPatVar where
   normalise' = return
 
 instance Normalise a => Normalise (Pattern' a) where
   normalise' p = case p of
-    VarP x       -> VarP <$> normalise' x
+    VarP o x     -> VarP o <$> normalise' x
     LitP _       -> return p
     ConP c mt ps -> ConP c <$> normalise' mt <*> normalise' ps
-    DotP v       -> DotP <$> normalise' v
-    AbsurdP x    -> AbsurdP <$> normalise' x
+    DotP o v     -> DotP o <$> normalise' v
     ProjP{}      -> return p
 
 instance Normalise DisplayForm where
@@ -1006,11 +1053,13 @@ instance InstantiateFull Sort where
     instantiateFull' s = do
         s <- instantiate' s
         case s of
-            Type n     -> levelSort <$> instantiateFull' n
+            Type n     -> Type <$> instantiateFull' n
             Prop       -> return s
-            DLub s1 s2 -> dLub <$> instantiateFull' s1 <*> instantiateFull' s2
+            PiSort s1 s2 -> piSort <$> instantiateFull' s1 <*> instantiateFull' s2
+            UnivSort s -> univSort <$> instantiateFull' s
             Inf        -> return s
             SizeUniv   -> return s
+            MetaS x es -> MetaS x <$> instantiateFull' es
 
 instance (InstantiateFull a) => InstantiateFull (Type' a) where
     instantiateFull' (El s t) =
@@ -1028,9 +1077,8 @@ instance InstantiateFull Term where
           Lit _       -> return v
           Level l     -> levelTm <$> instantiateFull' l
           Lam h b     -> Lam h <$> instantiateFull' b
-          Sort s      -> sortTm <$> instantiateFull' s
+          Sort s      -> Sort <$> instantiateFull' s
           Pi a b      -> uncurry Pi <$> instantiateFull' (a,b)
-          Shared{}    -> updateSharedTerm instantiateFull' v
           DontCare v  -> dontCare <$> instantiateFull' v
 
 instance InstantiateFull Level where
@@ -1044,7 +1092,7 @@ instance InstantiateFull LevelAtom where
   instantiateFull' l = case l of
     MetaLevel m vs -> do
       v <- instantiateFull' (MetaV m vs)
-      case ignoreSharing v of
+      case v of
         MetaV m vs -> return $ MetaLevel m vs
         _          -> return $ UnreducedLevel v
     NeutralLevel r v -> NeutralLevel r <$> instantiateFull' v
@@ -1072,15 +1120,14 @@ instance InstantiateFull Int where
     instantiateFull' = return
 
 instance InstantiateFull ConPatternInfo where
-  instantiateFull' (ConPatternInfo mr mt) = ConPatternInfo mr <$> instantiateFull' mt
+    instantiateFull' i = instantiateFull' (conPType i) <&> \ t -> i { conPType = t }
 
 instance InstantiateFull DBPatVar where
     instantiateFull' = return
 
 instance InstantiateFull a => InstantiateFull (Pattern' a) where
-    instantiateFull' (VarP x)       = VarP <$> instantiateFull' x
-    instantiateFull' (DotP t)       = DotP <$> instantiateFull' t
-    instantiateFull' (AbsurdP p)    = AbsurdP <$> instantiateFull' p
+    instantiateFull' (VarP o x)     = VarP o <$> instantiateFull' x
+    instantiateFull' (DotP o t)     = DotP o <$> instantiateFull' t
     instantiateFull' (ConP n mt ps) = ConP n <$> instantiateFull' mt <*> instantiateFull' ps
     instantiateFull' l@LitP{}       = return l
     instantiateFull' p@ProjP{}      = return p
@@ -1122,8 +1169,8 @@ instance InstantiateFull Constraint where
     ValueCmp cmp t u v -> do
       (t,u,v) <- instantiateFull' (t,u,v)
       return $ ValueCmp cmp t u v
-    ElimCmp cmp t v as bs ->
-      ElimCmp cmp <$> instantiateFull' t <*> instantiateFull' v <*> instantiateFull' as <*> instantiateFull' bs
+    ElimCmp cmp fs t v as bs ->
+      ElimCmp cmp fs <$> instantiateFull' t <*> instantiateFull' v <*> instantiateFull' as <*> instantiateFull' bs
     LevelCmp cmp u v    -> uncurry (LevelCmp cmp) <$> instantiateFull' (u,v)
     TypeCmp cmp a b     -> uncurry (TypeCmp cmp) <$> instantiateFull' (a,b)
     TelCmp a b cmp tela telb -> uncurry (TelCmp a b cmp) <$> instantiateFull' (tela,telb)
@@ -1133,6 +1180,9 @@ instance InstantiateFull Constraint where
     FindInScope m b cands -> FindInScope m b <$> mapM instantiateFull' cands
     IsEmpty r t         -> IsEmpty r <$> instantiateFull' t
     CheckSizeLtSat t    -> CheckSizeLtSat <$> instantiateFull' t
+    c@CheckFunDef{}     -> return c
+    HasBiggerSort a     -> HasBiggerSort <$> instantiateFull' a
+    HasPTSRule a b      -> uncurry HasPTSRule <$> instantiateFull' (a,b)
 
 instance (InstantiateFull a) => InstantiateFull (Elim' a) where
   instantiateFull' (Apply v) = Apply <$> instantiateFull' v
@@ -1194,9 +1244,6 @@ instance InstantiateFull RewriteRule where
 instance InstantiateFull a => InstantiateFull (Open a) where
   instantiateFull' (OpenThing n a) = OpenThing n <$> instantiateFull' a
 
-instance InstantiateFull a => InstantiateFull (Local a) where
-  instantiateFull' = traverseF instantiateFull'
-
 instance InstantiateFull DisplayForm where
   instantiateFull' (Display n ps v) = uncurry (Display n) <$> instantiateFull' (ps, v)
 
@@ -1235,11 +1282,13 @@ instance InstantiateFull a => InstantiateFull (WithArity a) where
   instantiateFull' (WithArity n a) = WithArity n <$> instantiateFull' a
 
 instance InstantiateFull a => InstantiateFull (Case a) where
-  instantiateFull' (Branches cop cs ls m) =
+  instantiateFull' (Branches cop cs eta ls m lz) =
     Branches cop
       <$> instantiateFull' cs
+      <*> instantiateFull' eta
       <*> instantiateFull' ls
       <*> instantiateFull' m
+      <*> pure lz
 
 instance InstantiateFull CompiledClauses where
   instantiateFull' Fail        = return Fail
@@ -1257,11 +1306,12 @@ instance InstantiateFull Clause where
 
 instance InstantiateFull Interface where
     instantiateFull' (Interface h ms mod scope inside
-                               sig display b foreignCode
+                               sig display userwarn b foreignCode
                                highlighting pragmas patsyns warnings) =
         Interface h ms mod scope inside
             <$> instantiateFull' sig
             <*> instantiateFull' display
+            <*> return userwarn
             <*> instantiateFull' b
             <*> return foreignCode
             <*> return highlighting
@@ -1276,6 +1326,9 @@ instance InstantiateFull a => InstantiateFull (Builtin a) where
 instance InstantiateFull QName where
   instantiateFull' = return
 
+instance InstantiateFull ConHead where
+  instantiateFull' = return
+
 instance InstantiateFull a => InstantiateFull (Maybe a) where
   instantiateFull' = mapM instantiateFull'
 
