@@ -77,6 +77,7 @@ import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Rewriting.NonLinMatch
 import qualified Agda.TypeChecking.Reduce.Monad as Red
+import Agda.TypeChecking.Warnings
 
 import Agda.Utils.Functor
 import qualified Agda.Utils.HashMap as HMap
@@ -112,13 +113,12 @@ verifyBuiltinRewrite v t = do
     (failure $ text "because it should accept at least two arguments") $
     \ (RelView tel delta a b core) -> do
     unless (visible a && visible b) $ failure $ text "because its two final arguments are not both visible."
-    case ignoreSharing (unEl core) of
+    case unEl core of
       Sort{}   -> return ()
       Con{}    -> __IMPOSSIBLE__
       Level{}  -> __IMPOSSIBLE__
       Lam{}    -> __IMPOSSIBLE__
       Pi{}     -> __IMPOSSIBLE__
-      Shared{} -> __IMPOSSIBLE__
       _ -> failure $ text "because its type does not end in a sort, but in "
              <+> do inTopContext $ addContext tel $ prettyTCM core
 
@@ -190,7 +190,7 @@ addRewriteRule q = do
         [ prettyTCM q , text " is not a legal rewrite rule" ]
 
   -- Check that type of q targets rel.
-  case ignoreSharing $ unEl core of
+  case unEl core of
     Def rel' es@(_:_:_) | rel == rel' -> do
       -- Because of the type of rel (Γ → sort), all es are applications.
       let vs = map unArg $ fromMaybe __IMPOSSIBLE__ $ allApplyElims es
@@ -214,13 +214,15 @@ addRewriteRule q = do
       lhs <- modifyAllowedReductions (const [InlineReductions]) $ normalise lhs
 
       -- Find head symbol f of the lhs and its arguments.
-      (f , hd , es) <- case ignoreSharing lhs of
+      (f , hd , es) <- case lhs of
         Def f es -> return (f , Def f , es)
         Con c ci vs -> do
-          let hd = Con c ci . fromMaybe __IMPOSSIBLE__ . allApplyElims
-          return (conName c , hd  , map Apply vs)
+          let hd = Con c ci
+          return (conName c , hd  , vs)
         _        -> failureNotDefOrCon
 
+      ifNotAlreadyAdded f $ do
+
       rew <- addContext gamma1 $ do
         -- Normalize lhs args: we do not want to match redexes.
         es <- etaContract =<< normalise es
@@ -271,22 +273,18 @@ addRewriteRule q = do
       unless (v == v') $ do
         reportSDoc "rewriting" 20 $ text "v  = " <+> text (show v)
         reportSDoc "rewriting" 20 $ text "v' = " <+> text (show v')
-        -- Andreas, 2016-06-01, issue 1997
-        -- A reason for a reduction of the lhs could be that
-        -- the rewrite rule has already been added.
-        -- In this case, we want a nicer error message.
-        checkNotAlreadyAdded f
         typeError . GenericDocError =<< fsep
           [ prettyTCM q <+> text " is not a legal rewrite rule, since the left-hand side "
           , prettyTCM v <+> text " reduces to " <+> prettyTCM v' ]
 
-    checkNotAlreadyAdded :: QName -> TCM ()
-    checkNotAlreadyAdded f = do
+    ifNotAlreadyAdded :: QName -> TCM () -> TCM ()
+    ifNotAlreadyAdded f cont = do
       rews <- getRewriteRulesFor f
       -- check if q is already an added rewrite rule
-      when (any ((q ==) . rewName) rews) $
-        typeError . GenericDocError =<< do
+      if any ((q ==) . rewName) rews then
+        genericWarning =<< do
           text "Rewrite rule " <+> prettyTCM q <+> text " has already been added"
+      else cont
 
     usedArgs :: [Pos.Occurrence] -> IntSet
     usedArgs occs = IntSet.fromList $ map snd $ usedIxs
@@ -303,11 +301,6 @@ addRewriteRules f rews = do
   let matchables = getMatchables rews
   reportSDoc "rewriting" 30 $ text "matchable symbols: " <+> prettyTCM matchables
   modifySignature $ addRewriteRulesFor f rews matchables
-  --rules <- getRewriteRulesFor f
-  --reportSDoc "rewriting" 20 $ vcat
-  --  [ text "rewrite rules for " <+> prettyTCM f <+> text ":"
-  --  , vcat (map prettyTCM rules)
-  --  ]
 
 -- | @rewriteWith t f es rew@
 --   tries to rewrite @f es : t@ with @rew@, returning the reduct if successful.
@@ -332,32 +325,6 @@ rewriteWith mt v rew@(RewriteRule q gamma _ ps rhs b) es = do
         ]) $ do
       return $ Right v'
 
-    {- OLD CODE:
-    -- Freeze all metas, remember which one where not frozen before.
-    -- This ensures that we do not instantiate metas while matching
-    -- on the rewrite lhs.
-    ms <- freezeMetas
-    res <- tryConversion' $ do
-
-      -- Create new metas for the lhs variables of the rewriting rule.
-      xs <- newTelMeta gamma
-      let sigma        = parallelS $ map unArg xs
-          (lhs', rhs', b') = applySubst sigma (lhs, rhs, b)
-      -- Unify type and term with type and lhs of rewrite rule.
-      whenJust mt $ \ t -> leqType t b'
-      local (\ e -> e {envCompareBlocked = True}) $ equalTerm b' lhs' v
-      -- Check that all variables have been solved for.
-      unlessM (isInstantiatedMeta xs) $ do
-        reportSDoc "rewriting" 20 $ text "lhs variables solved with: " <+> do
-          sep $ map prettyTCM xs
-        -- The following error is caught immediately by tryConversion.
-        typeError $ GenericError $ "free variables not bound by left hand side"
-      return rhs'
-
-    -- Thaw metas that were frozen by a call to this function.
-    unfreezeMetas' (`elem` ms)
-    return res-}
-
 -- | @rewrite b v rules es@ tries to rewrite @v@ applied to @es@ with the
 --   rewrite rules @rules@. @b@ is the default blocking tag.
 rewrite :: Blocked_ -> Term -> RewriteRules -> Elims -> ReduceM (Reduced (Blocked Term) Term)
