@@ -30,10 +30,6 @@ import Control.Monad.State
 import Debug.Trace
 import System.IO.Unsafe
 
-#if __GLASGOW_HASKELL__ <= 708
-import Data.Foldable ( foldMap )
-#endif
-
 import Data.Maybe
 import Data.Traversable (Traversable,traverse)
 import Data.IntMap (IntMap)
@@ -111,7 +107,9 @@ instance PatternFrom Sort NLPat where
       Prop     -> done
       Inf      -> done
       SizeUniv -> done
-      DLub _ _ -> done
+      PiSort _ _ -> done
+      UnivSort _ -> done
+      MetaS{}  -> __IMPOSSIBLE__
 
 instance PatternFrom Term NLPat where
   patternFrom r k v = do
@@ -120,7 +118,7 @@ instance PatternFrom Term NLPat where
                  return PWild
                else
                  return $ PTerm v
-    case ignoreSharing v of
+    case v of
       Var i es
        | i < k     -> PBoundVar i <$> patternFrom r k es
        | otherwise -> do
@@ -128,7 +126,7 @@ instance PatternFrom Term NLPat where
          -- in order to build a Miller pattern
          let mbvs = mfilter fastDistinct $ forM es $ \e -> do
                       e <- isApplyElim e
-                      case ignoreSharing $ unArg e of
+                      case unArg e of
                         Var j [] | j < k -> Just $ e $> j
                         _                -> Nothing
          case mbvs of
@@ -143,8 +141,8 @@ instance PatternFrom Term NLPat where
       Lit{}    -> done
       Def f es | isIrrelevant r -> done
       Def f es -> do
-        Def lsuc [] <- ignoreSharing <$> primLevelSuc
-        Def lmax [] <- ignoreSharing <$> primLevelMax
+        Def lsuc [] <- primLevelSuc
+        Def lmax [] <- primLevelMax
         case es of
           [x]     | f == lsuc -> done
           [x , y] | f == lmax -> done
@@ -152,17 +150,16 @@ instance PatternFrom Term NLPat where
       Con c ci vs | isIrrelevant r -> do
         mr <- isRecordConstructor (conName c)
         case mr of
-          Just (_, def) | recEtaEquality def ->
-            PDef (conName c) <$> patternFrom r k (Apply <$> vs)
+          Just (_, def) | YesEta <- recEtaEquality def ->
+            PDef (conName c) <$> patternFrom r k vs
           _ -> done
-      Con c ci vs -> PDef (conName c) <$> patternFrom r k (Apply <$> vs)
+      Con c ci vs -> PDef (conName c) <$> patternFrom r k vs
       Pi a b | isIrrelevant r -> done
       Pi a b   -> PPi <$> patternFrom r k a <*> patternFrom r k b
       Sort s   -> done
       Level l  -> __IMPOSSIBLE__
       DontCare{} -> return PWild
       MetaV{}    -> __IMPOSSIBLE__
-      Shared{}   -> __IMPOSSIBLE__
 
 instance (PatternFrom a b) => PatternFrom (Abs a) (Abs b) where
   patternFrom r k (Abs name x)   = Abs name   <$> patternFrom r (k+1) x
@@ -323,18 +320,18 @@ instance Match NLPat Term where
           Right (Just v) -> tellSub r (i-n) $ teleLam tel $ renameP __IMPOSSIBLE__ perm v
       PDef f ps -> do
         v <- liftRed $ constructorForm =<< unLevel v
-        case ignoreSharing v of
+        case v of
           Def f' es
             | f == f'   -> match r gamma k ps es
           Con c _ vs
-            | f == conName c -> match r gamma k ps (Apply <$> vs)
+            | f == conName c -> match r gamma k ps vs
             | otherwise -> do -- @c@ may be a record constructor
                 mr <- liftRed $ isRecordConstructor (conName c)
                 case mr of
-                  Just (_, def) | recEtaEquality def -> do
+                  Just (_, def) | YesEta <- recEtaEquality def -> do
                     let fs = recFields def
                         qs = map (fmap $ \f -> PDef f (ps ++ [Proj ProjSystem f])) fs
-                    match r gamma k qs vs
+                    match r gamma k (map Apply qs) vs
                   _ -> no (text "")
           Lam i u -> do
             let pbody = PDef f (raise 1 ps ++ [Apply $ Arg i $ PTerm (var 0)])
@@ -345,7 +342,7 @@ instance Match NLPat Term where
           v' -> do -- @f@ may be a record constructor as well
             mr <- liftRed $ isRecordConstructor f
             case mr of
-              Just (_, def) | recEtaEquality def -> do
+              Just (_, def) | YesEta <- recEtaEquality def -> do
                 let fs  = recFields def
                     ws  = map (fmap $ \f -> v `applyE` [Proj ProjSystem f]) fs
                     qs = fromMaybe __IMPOSSIBLE__ $ allApplyElims ps
@@ -354,19 +351,19 @@ instance Match NLPat Term where
       PLam i p' -> do
         let body = Abs (absName p') $ raise 1 v `apply` [Arg i (var 0)]
         match r gamma k p' body
-      PPi pa pb  -> case ignoreSharing v of
+      PPi pa pb  -> case v of
         Pi a b -> match r gamma k pa a >> match r gamma k pb b
         MetaV m es -> matchingBlocked $ Blocked m ()
         _ -> no (text "")
-      PBoundVar i ps -> case ignoreSharing v of
+      PBoundVar i ps -> case v of
         Var i' es | i == i' -> match r gamma k ps es
         Con c _ vs -> do -- @c@ may be a record constructor
           mr <- liftRed $ isRecordConstructor (conName c)
           case mr of
-            Just (_, def) | recEtaEquality def -> do
+            Just (_, def) | YesEta <- recEtaEquality def -> do
               let fs = recFields def
                   qs = map (fmap $ \f -> PBoundVar i (ps ++ [Proj ProjSystem f])) fs
-              match r gamma k qs vs
+              match r gamma k (map Apply qs) vs
             _ -> no (text "")
         Lam info u -> do
           let pbody = PBoundVar i (raise 1 ps ++ [Apply $ Arg info $ PTerm (var 0)])
