@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 
 module Agda.TypeChecking.Rules.Builtin
   ( bindBuiltin
@@ -12,13 +13,15 @@ import Control.Applicative hiding (empty)
 import Control.Monad
 import Control.Monad.Reader (ask)
 import Control.Monad.State (get)
-import Data.List (find)
+import Data.List (find, sortBy)
+import Data.Function (on)
 
 import qualified Agda.Syntax.Abstract as A
 import qualified Agda.Syntax.Abstract.Views as A
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Position
+import Agda.Syntax.Scope.Base
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
@@ -41,9 +44,11 @@ import {-# SOURCE #-} Agda.TypeChecking.Rules.Builtin.Coinduction
 import {-# SOURCE #-} Agda.TypeChecking.Rewriting
 
 import Agda.Utils.Except ( MonadError(catchError) )
+import Agda.Utils.Functor
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.NonemptyList
 import Agda.Utils.Null
 import Agda.Utils.Size
 
@@ -57,6 +62,9 @@ import Agda.Utils.Impossible
 builtinPostulate :: TCM Type -> BuiltinDescriptor
 builtinPostulate = BuiltinPostulate Relevant
 
+findBuiltinInfo :: String -> Maybe BuiltinInfo
+findBuiltinInfo b = find ((b ==) . builtinName) coreBuiltins
+
 coreBuiltins :: [BuiltinInfo]
 coreBuiltins =
   [ (builtinList               |-> BuiltinData (tset --> tset) [builtinNil, builtinCons])
@@ -66,7 +74,7 @@ coreBuiltins =
   , (builtinBool               |-> BuiltinData tset [builtinTrue, builtinFalse])
   , (builtinNat                |-> BuiltinData tset [builtinZero, builtinSuc])
   , (builtinUnit               |-> BuiltinData tset [builtinUnitUnit])  -- actually record, but they are treated the same
-  , (builtinAgdaLiteral        |-> BuiltinData tset [builtinAgdaLitNat, builtinAgdaLitFloat,
+  , (builtinAgdaLiteral        |-> BuiltinData tset [builtinAgdaLitNat, builtinAgdaLitWord64, builtinAgdaLitFloat,
                                                      builtinAgdaLitChar, builtinAgdaLitString,
                                                      builtinAgdaLitQName, builtinAgdaLitMeta])
   , (builtinAgdaPattern        |-> BuiltinData tset [builtinAgdaPatVar, builtinAgdaPatCon, builtinAgdaPatDot,
@@ -78,6 +86,7 @@ coreBuiltins =
   , (builtinAgdaPatProj        |-> BuiltinDataCons (tqname --> tpat))
   , (builtinAgdaPatAbsurd      |-> BuiltinDataCons tpat)
   , (builtinLevel              |-> builtinPostulate tset)
+  , (builtinWord64             |-> builtinPostulate tset)
   , (builtinInteger            |-> BuiltinData tset [builtinIntegerPos, builtinIntegerNegSuc])
   , (builtinIntegerPos         |-> BuiltinDataCons (tnat --> tinteger))
   , (builtinIntegerNegSuc      |-> BuiltinDataCons (tnat --> tinteger))
@@ -148,6 +157,7 @@ coreBuiltins =
   , (builtinAgdaTermMeta         |-> BuiltinDataCons (tmeta --> targs --> tterm))
   , (builtinAgdaTermUnsupported|-> BuiltinDataCons tterm)
   , (builtinAgdaLitNat    |-> BuiltinDataCons (tnat --> tliteral))
+  , (builtinAgdaLitWord64 |-> BuiltinDataCons (tword64 --> tliteral))
   , (builtinAgdaLitFloat  |-> BuiltinDataCons (tfloat --> tliteral))
   , (builtinAgdaLitChar   |-> BuiltinDataCons (tchar --> tliteral))
   , (builtinAgdaLitString |-> BuiltinDataCons (tstring --> tliteral))
@@ -213,6 +223,7 @@ coreBuiltins =
   , builtinAgdaTCMInContext     |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tlist (targ ttype) --> tTCM 1 (varM 0) --> tTCM 1 (varM 0))
   , builtinAgdaTCMFreshName     |-> builtinPostulate (tstring --> tTCM_ primQName)
   , builtinAgdaTCMDeclareDef    |-> builtinPostulate (targ tqname --> ttype --> tTCM_ primUnit)
+  , builtinAgdaTCMDeclarePostulate    |-> builtinPostulate (targ tqname --> ttype --> tTCM_ primUnit)
   , builtinAgdaTCMDefineFun     |-> builtinPostulate (tqname --> tlist tclause --> tTCM_ primUnit)
   , builtinAgdaTCMGetType            |-> builtinPostulate (tqname --> tTCM_ primAgdaTerm)
   , builtinAgdaTCMGetDefinition      |-> builtinPostulate (tqname --> tTCM_ primAgdaDefinition)
@@ -253,6 +264,7 @@ coreBuiltins =
         terrorpart = el primAgdaErrorPart
         tnat       = el primNat
         tint       = el primInteger
+        tword64    = el primWord64
         tunit      = el primUnit
         tinteger   = el primInteger
         tfloat     = el primFloat
@@ -399,7 +411,7 @@ coreBuiltins =
 inductiveCheck :: String -> Int -> Term -> TCM (QName, Definition)
 inductiveCheck b n t = do
   t <- etaContract =<< normalise t
-  case ignoreSharing t of
+  case t of
     Def q _ -> do
       def <- getConstInfo q
       let yes = return (q, def)
@@ -423,14 +435,14 @@ inductiveCheck b n t = do
         , "constructors"
         ]
 
--- | @bindPostulatedName builtin e m@ checks that @e@ is a postulated
--- name @q@, and binds the builtin @builtin@ to the term @m q def@,
+-- | @bindPostulatedName builtin q m@ checks that @q@ is a postulated
+-- name, and binds the builtin @builtin@ to the term @m q def@,
 -- where @def@ is the current 'Definition' of @q@.
 
 bindPostulatedName ::
-  String -> A.Expr -> (QName -> Definition -> TCM Term) -> TCM ()
-bindPostulatedName builtin e m = do
-  q   <- getName e
+  String -> ResolvedName -> (QName -> Definition -> TCM Term) -> TCM ()
+bindPostulatedName builtin x m = do
+  q   <- getName x
   def <- getConstInfo q
   case theDef def of
     Axiom {} -> bindBuiltinName builtin =<< m q def
@@ -439,18 +451,9 @@ bindPostulatedName builtin e m = do
   err = typeError $ GenericError $
           "The argument to BUILTIN " ++ builtin ++
           " must be a postulated name"
-
-  getName (A.Def q)          = return q
-  getName (A.ScopedExpr _ e) = getName e
-  getName _                  = err
-
--- REPLACED by TC.Functions.getDef
--- getDef :: Term -> TCM QName
--- getDef t = do
---   t <- etaContract =<< normalise t
---   case ignoreSharing t of
---     Def d _ -> return d
---     _ -> __IMPOSSIBLE__
+  getName = \case
+    DefinedName _ d -> return $ anameName d
+    _ -> err
 
 addHaskellPragma :: QName -> String -> TCM ()
 addHaskellPragma = addPragma ghcBackendName
@@ -464,28 +467,46 @@ bindAndSetHaskellCode b hs t = do
 bindBuiltinBool :: Term -> TCM ()
 bindBuiltinBool = bindAndSetHaskellCode builtinBool "= type Bool"
 
+-- | Check that we're not trying to bind true and false to the same
+-- constructor.
+checkBuiltinBool :: TCM ()
+checkBuiltinBool = do
+  true  <- getBuiltin' builtinTrue
+  false <- getBuiltin' builtinFalse
+  when (true == false) $
+    genericError "Cannot bind TRUE and FALSE to the same constructor"
+
 bindBuiltinInt :: Term -> TCM ()
 bindBuiltinInt = bindAndSetHaskellCode builtinInteger "= type Integer"
 
 bindBuiltinNat :: Term -> TCM ()
 bindBuiltinNat t = do
-  nat <- fromMaybe __IMPOSSIBLE__ <$> getDef t
-  def <- theDef <$> getConstInfo nat
-  case def of
-    Datatype { dataCons = [c1, c2] } -> do
-      bindBuiltinName builtinNat t
-      let getArity c = arity <$> (normalise . defType =<< getConstInfo c)
-      [a1, a2] <- mapM getArity [c1, c2]
-      let (zero, suc) | a2 > a1   = (c1, c2)
-                      | otherwise = (c2, c1)
-          tnat = el primNat
-          rerange = setRange (getRange nat)
-      addHaskellPragma nat "= type Integer"
-      bindBuiltinInfo (BuiltinInfo builtinZero $ BuiltinDataCons tnat)
-                      (A.Con $ AmbQ [rerange zero])
-      bindBuiltinInfo (BuiltinInfo builtinSuc  $ BuiltinDataCons (tnat --> tnat))
-                      (A.Con $ AmbQ [rerange suc])
-    _ -> __IMPOSSIBLE__
+  bindBuiltinData builtinNat t
+  name <- fromMaybe __IMPOSSIBLE__ <$> getDef t
+  addHaskellPragma name "= type Integer"
+
+-- | Only use for datatypes with distinct arities of constructors.
+--   Binds the constructors together with the datatype.
+bindBuiltinData :: String -> Term -> TCM ()
+bindBuiltinData s t = do
+  bindBuiltinName s t
+  name <- fromMaybe __IMPOSSIBLE__ <$> getDef t
+  Datatype{ dataCons = cs } <- theDef <$> getConstInfo name
+  let getArity c = do
+        Constructor{ conArity = a } <- theDef <$> getConstInfo c
+        return a
+      getBuiltinArity (BuiltinDataCons t) = arity <$> t
+      getBuiltinArity _ = __IMPOSSIBLE__
+      sortByM f xs = map fst . sortBy (compare `on` snd) . zip xs <$> mapM f xs
+  -- Order constructurs by arity
+  cs <- sortByM getArity cs
+  -- Do the same for the builtins
+  let bcis = fromMaybe __IMPOSSIBLE__ $ do
+        BuiltinData _ bcs <- builtinDesc <$> findBuiltinInfo s
+        mapM findBuiltinInfo bcs
+  bcis <- sortByM (getBuiltinArity . builtinDesc) bcis
+  unless (length cs == length bcis) __IMPOSSIBLE__  -- we already checked this
+  zipWithM_ (\ c bci -> bindBuiltinInfo bci (A.Con $ unambiguous $ setRange (getRange name) c)) cs bcis
 
 bindBuiltinUnit :: Term -> TCM ()
 bindBuiltinUnit t = do
@@ -498,12 +519,9 @@ bindBuiltinUnit t = do
     _ -> genericError "Builtin UNIT must be a singleton record type"
 
 -- | Bind BUILTIN EQUALITY and BUILTIN REFL.
-bindBuiltinEquality :: A.Expr -> TCM ()
-bindBuiltinEquality (A.ScopedExpr scope e) = do
-  setScope scope
-  bindBuiltinEquality e
-bindBuiltinEquality e = do
-  (v, _t) <- inferExpr e
+bindBuiltinEquality :: ResolvedName -> TCM ()
+bindBuiltinEquality x = do
+  (v, _t) <- inferExpr (A.nameExpr x)
 
   -- Equality needs to be a data type with 1 constructor
   (eq, def) <- inductiveCheck builtinEquality 1 v
@@ -540,7 +558,7 @@ bindBuiltinEquality e = do
       unless (all ((Just 0 ==) . deBruijnView) ts) wrongRefl
 
       -- Check the target
-      case ignoreSharing $ unEl conCore of
+      case unEl conCore of
         Def _ es -> do
           let vs = map unArg $ fromMaybe __IMPOSSIBLE__ $ allApplyElims es
           (a,b) <- reduce $ fromMaybe __IMPOSSIBLE__ $ last2 vs
@@ -553,7 +571,6 @@ bindBuiltinEquality e = do
   wrongRefl = genericError "Wrong type of constructor of BUILTIN EQUALITY"
 
 bindBuiltinInfo :: BuiltinInfo -> A.Expr -> TCM ()
-bindBuiltinInfo i (A.ScopedExpr scope e) = setScope scope >> bindBuiltinInfo i e
 bindBuiltinInfo (BuiltinInfo s d) e = do
     case d of
       BuiltinData t cs -> do
@@ -565,13 +582,13 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
            | s == builtinNat      -> bindBuiltinNat      v
            | s == builtinInteger  -> bindBuiltinInt      v
            | s == builtinUnit     -> bindBuiltinUnit     v
+           | s == builtinList     -> bindBuiltinData s   v
            | otherwise            -> bindBuiltinName s   v
 
       BuiltinDataCons t -> do
 
         let name (Lam h b)  = name (absBody b)
             name (Con c ci _) = Con c ci []
-            name (Shared p) = name $ ignoreSharing (derefPtr p)
             name _          = __IMPOSSIBLE__
 
         v0 <- checkExpr e =<< t
@@ -582,8 +599,11 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
 
         let v@(Con h _ []) = name v0
             c = conName h
+
         bindBuiltinName s v
 
+        when (s `elem` [builtinFalse, builtinTrue]) checkBuiltinBool
+
       BuiltinPrim pfname axioms -> do
         case e of
           A.Def qx -> do
@@ -597,8 +617,13 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
             let cls = defClauses info
                 a   = defAbstract info
                 mcc = defCompiled info
+                inv = defInverse info
             bindPrimitive pfname $ pf { primFunName = qx }
-            addConstant qx $ info { theDef = Primitive a pfname cls mcc }
+            addConstant qx $ info { theDef = Primitive { primAbstr    = a
+                                                       , primName     = pfname
+                                                       , primClauses  = cls
+                                                       , primInv      = inv
+                                                       , primCompiled = mcc } }
 
             -- needed? yes, for checking equations for mul
             bindBuiltinName s v
@@ -619,6 +644,7 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
                 when (s == builtinChar)   $ addHaskellPragma q "= type Char"
                 when (s == builtinString) $ addHaskellPragma q "= type Data.Text.Text"
                 when (s == builtinFloat)  $ addHaskellPragma q "= type Double"
+                when (s == builtinWord64) $ addHaskellPragma q "= type MAlonzo.RTE.Word64"
                 bindBuiltinName s v
               _        -> err
           _ -> err
@@ -631,36 +657,57 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
         bindBuiltinName s v
 
 -- | Bind a builtin thing to an expression.
-bindBuiltin :: String -> A.Expr -> TCM ()
-bindBuiltin b e = do
-  unlessM ((0 ==) <$> getContextSize) $ typeError $ BuiltinInParameterisedModule b
+bindBuiltin :: String -> ResolvedName -> TCM ()
+bindBuiltin b x = do
+  unlessM ((0 ==) <$> getContextSize) $ do
+    -- Andreas, 2017-11-01, issue #2824
+    -- Only raise an error if the name for the builtin is defined in a parametrized module.
+    let failure = typeError $ BuiltinInParameterisedModule b
+    -- Get the non-empty list of AbstractName for x
+    xs <- case x of
+      VarName{}            -> failure
+      DefinedName _ x      -> return $ x :! []
+      FieldName xs         -> return xs
+      ConstructorName xs   -> return xs
+      PatternSynResName xs -> failure
+      UnknownName          -> failure
+    -- For ambiguous names, we check all of their definitions:
+    unlessM (allM xs $ ((0 ==) . size) <.> lookupSection . qnameModule . anameName) $
+      failure
+  -- Since the name was define in a parameter-free context, we can switch to the empty context.
+  -- (And we should!)
+  inTopContext $ do
   if | b == builtinRefl  -> warning $ OldBuiltin b builtinEquality
-     | b == builtinZero  -> nowNat b
-     | b == builtinSuc   -> nowNat b
-     | b == builtinInf   -> bindBuiltinInf e
-     | b == builtinSharp -> bindBuiltinSharp e
-     | b == builtinFlat  -> bindBuiltinFlat e
-     | b == builtinEquality -> bindBuiltinEquality e
-     | Just i <- find ((b ==) . builtinName) coreBuiltins -> bindBuiltinInfo i e
+     | b == builtinZero  -> now builtinNat b
+     | b == builtinSuc   -> now builtinNat b
+     | b == builtinNil   -> now builtinList b
+     | b == builtinCons  -> now builtinList b
+     | b == builtinInf   -> bindBuiltinInf x
+     | b == builtinSharp -> bindBuiltinSharp x
+     | b == builtinFlat  -> bindBuiltinFlat x
+     | b == builtinEquality -> bindBuiltinEquality x
+     | Just i <- findBuiltinInfo b -> bindBuiltinInfo i (A.nameExpr x)
      | otherwise -> typeError $ NoSuchBuiltinName b
   where
-    nowNat b = warning $ OldBuiltin b builtinNat
+    now new b = warning $ OldBuiltin b new
 
 isUntypedBuiltin :: String -> Bool
 isUntypedBuiltin b = elem b [builtinFromNat, builtinFromNeg, builtinFromString]
 
-bindUntypedBuiltin :: String -> A.Expr -> TCM ()
-bindUntypedBuiltin b e =
-  case A.unScope e of
-    A.Def q  -> bindBuiltinName b (Def q [])
-    A.Proj _ (AmbQ [q]) -> bindBuiltinName b (Def q [])
-    e        -> genericError $ "The argument to BUILTIN " ++ b ++ " must be a defined unambiguous name"
+bindUntypedBuiltin :: String -> ResolvedName -> TCM ()
+bindUntypedBuiltin b = \case
+  DefinedName _ x -> bindBuiltinName b (Def (anameName x) [])
+  FieldName (x :! []) -> bindBuiltinName b (Def (anameName x) [])
+  _ -> genericError $ "The argument to BUILTIN " ++ b ++ " must be a defined unambiguous name"
 
 -- | Bind a builtin thing to a new name.
+--
+-- Since their type is closed, it does not matter whether we are in a
+-- parameterized module when we declare them.
+-- We simply ignore the parameters.
 bindBuiltinNoDef :: String -> A.QName -> TCM ()
-bindBuiltinNoDef b q = do
-  unlessM ((0 ==) <$> getContextSize) $ typeError $ BuiltinInParameterisedModule b
-  case lookup b $ map (\ (BuiltinInfo b i) -> (b, i)) coreBuiltins of
+bindBuiltinNoDef b q = inTopContext $ do
+  case builtinDesc <$> findBuiltinInfo b of
     Just (BuiltinPostulate rel mt) -> do
       t <- mt
       addConstant q $ defaultDefn (setRelevance rel defaultArgInfo) q t def
@@ -672,6 +719,7 @@ bindBuiltinNoDef b q = do
         def | b == builtinSizeUniv = emptyFunction
                 { funClauses = [ (empty :: Clause) { clauseBody = Just $ Sort sSizeUniv } ]
                 , funCompiled = Just (CC.Done [] $ Sort sSizeUniv)
+                , funMutual    = Just []
                 , funTerminates = Just True
                 }
             | otherwise = Axiom
