@@ -3,12 +3,10 @@
 
 module Agda.TypeChecking.Rules.Data where
 
-import Control.Applicative
 import Control.Monad
 
 import Data.List (genericTake)
 import Data.Maybe (fromMaybe)
-import qualified Data.Set as Set
 
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Views (deepUnscope)
@@ -18,7 +16,7 @@ import Agda.Syntax.Position
 import qualified Agda.Syntax.Info as Info
 
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Monad.Builtin (primLevel)
+import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.MetaVars
@@ -35,12 +33,7 @@ import Agda.Interaction.Options
 
 import Agda.Utils.Except
 import Agda.Utils.List
-import Agda.Utils.Monad
-import Agda.Utils.Permutation
-import qualified Agda.Utils.Pretty as P
 import Agda.Utils.Size
-import Agda.Utils.Tuple
-import qualified Agda.Utils.VarSet as VarSet
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -89,7 +82,9 @@ checkDataDef i name ps cs =
               -- However, it might give the dreaded "Cannot instantiate meta..."
               -- error which we replace by a more understandable error
               -- in case of a suspected dependency.
-              s <- newSortMetaBelowInf
+              -- Jesper, 2017-09-14: since Set i depends non-strictly on i,
+              -- we need to create the new meta in a nonstrict context.
+              s <- applyRelevanceToContext NonStrict newSortMetaBelowInf
               catchError_ (addContext ixTel $ equalType s0 $ raise nofIxs $ sort s) $ \ err ->
                   if any (`freeIn` s0) [0..nofIxs - 1] then typeError . GenericDocError =<<
                      fsep [ text "The sort of" <+> prettyTCM name
@@ -99,9 +94,6 @@ checkDataDef i name ps cs =
                   else throwError err
               return s
 
-            -- the small parameters are taken into consideration for --without-K
-            smallPars <- smallParams tel s
-
             reportSDoc "tc.data.sort" 20 $ vcat
               [ text "checking datatype" <+> prettyTCM name
               , nest 2 $ vcat
@@ -110,7 +102,6 @@ checkDataDef i name ps cs =
                 , text "sort:   " <+> prettyTCM s
                 , text "indices:" <+> text (show nofIxs)
                 , text "params:"  <+> text (show $ deepUnscope ps)
-                , text "small params:" <+> text (show smallPars)
                 ]
               ]
             let npars = size tel
@@ -118,8 +109,6 @@ checkDataDef i name ps cs =
             -- Change the datatype from an axiom to a datatype with no constructors.
             let dataDef = Datatype
                   { dataPars       = npars
-                  , dataSmallPars  = Perm npars smallPars
-                  , dataNonLinPars = Drop 0 $ Perm npars []
                   , dataIxs        = nofIxs
                   , dataInduction  = Inductive
                   , dataClause     = Nothing
@@ -135,19 +124,10 @@ checkDataDef i name ps cs =
                 -- polarity and argOcc.s determined by the positivity checker
 
             -- Check the types of the constructors
-            -- collect the non-linear parameters of each constructor
-            nonLins <- mapM (checkConstructor name tel' nofIxs s) cs
-            -- compute the ascending list of non-linear parameters of the data type
-            let nonLinPars0 = Set.toAscList $ Set.unions $ map Set.fromList nonLins
-                -- The constructors are analyzed in the absolute context,
-                -- but the data definition happens in the relative module context,
-                -- so we apply to the free module variables.
-                -- Unfortunately, we lose precision here, since 'abstract', which
-                -- is then performed by addConstant, cannot restore the linearity info.
-                nonLinPars  = Drop freeVars $ Perm (npars + freeVars) nonLinPars0
+            mapM_ (checkConstructor name tel' nofIxs s) cs
 
             -- Return the data definition
-            return dataDef{ dataNonLinPars = nonLinPars }
+            return dataDef
 
         let s      = dataSort dataDef
             cons   = map A.axiomName cs  -- get constructor names
@@ -170,33 +150,15 @@ checkDataDef i name ps cs =
 -- | Ensure that the type is a sort.
 --   If it is not directly a sort, compare it to a 'newSortMetaBelowInf'.
 forceSort :: Type -> TCM Sort
-forceSort t = case ignoreSharing $ unEl t of
+forceSort t = case unEl t of
   Sort s -> return s
   _      -> do
-    s <- newSortMetaBelowInf
+    -- Universes depend non-strictly on their argument
+    s <- applyRelevanceToContext NonStrict newSortMetaBelowInf
     equalType t (sort s)
     return s
 
 
--- | A parameter is small if its sort fits into the data sort.
---   @smallParams@ overapproximates the small parameters (in doubt: small).
-smallParams :: Telescope -> Sort -> TCM [Int]
-smallParams tel s = do
-  -- get the types of the parameters
-  let as = map (snd . unDom) $ telToList tel
-  -- get the big parameters
-  concat <$> do
-    forM (zip [0..] as) $ \ (i, a) -> do
-      -- A type is small if it is not Level or its sort is <= the data sort.
-      -- In doubt (unsolvable constraints), a type is small.
-      -- So, only if we have a solid error, the type is big.
-      localTCState $ do
-        ([] <$ do equalTerm topSort (unEl a) =<< primLevel)  -- NB: if primLevel fails, the next alternative is picked
-        <|> ([i] <$ (getSort a `leqSort` s))
-        <|> return []
-  where
-    (<|>) m1 m2 = m1 `catchError_` (const m2)
-
 -- | Type check a constructor declaration. Checks that the constructor targets
 --   the datatype and that it fits inside the declared sort.
 --   Returns the non-linear parameters.
@@ -206,7 +168,7 @@ checkConstructor
   -> Nat           -- ^ Number of indices of the data type.
   -> Sort          -- ^ Sort of the data type.
   -> A.Constructor -- ^ Constructor declaration (type signature).
-  -> TCM [Int]     -- ^ Non-linear parameters.
+  -> TCM ()
 checkConstructor d tel nofIxs s (A.ScopedDecl scope [con]) = do
   setScope scope
   checkConstructor d tel nofIxs s con
@@ -224,32 +186,32 @@ checkConstructor d tel nofIxs s con@(A.Axiom _ i ai Nothing c e) =
           Irrelevant -> typeError $ GenericError $ "Irrelevant constructors are not supported"
           _ -> __IMPOSSIBLE__
         -- check that the type of the constructor is well-formed
-        t <- isType_ e
+        t <- workOnTypes $ isType_ e
         -- check that the type of the constructor ends in the data type
         n <- getContextSize
         debugEndsIn t d n
-        nonLinPars <- constructs n t d
-        debugNonLinPars nonLinPars
-        -- check which constructor arguments are determined by the type ('forcing')
-        t' <- addForcingAnnotations t
+        constructs n t d
+        -- compute which constructor arguments are forced
+        forcedArgs <- computeForcingAnnotations t
         -- check that the sort (universe level) of the constructor type
         -- is contained in the sort of the data type
         -- (to avoid impredicative existential types)
         debugFitsIn s
-        arity <- t' `fitsIn` s
-        debugAdd c t'
+        arity <- fitsIn forcedArgs t s
+        debugAdd c t
 
         -- add parameters to constructor type and put into signature
         let con = ConHead c Inductive [] -- data constructors have no projectable fields and are always inductive
         escapeContext (size tel) $
           addConstant c $
-            defaultDefn defaultArgInfo c (telePi tel t') $ Constructor
+            defaultDefn defaultArgInfo c (telePi tel t) $ Constructor
               { conPars   = size tel
               , conArity  = arity
               , conSrcCon = con
               , conData   = d
               , conAbstr  = Info.defAbstract i
               , conInd    = Inductive
+              , conForced = forcedArgs
               , conErased = []  -- computed during compilation to treeless
               }
 
@@ -257,7 +219,6 @@ checkConstructor d tel nofIxs s con@(A.Axiom _ i ai Nothing c e) =
         when (Info.defInstance i == InstanceDef) $ do
           addNamedInstance c d
 
-        return nonLinPars
   where
     debugEnter c e =
       reportSDoc "tc.data.con" 5 $ vcat
@@ -271,9 +232,6 @@ checkConstructor d tel nofIxs s con@(A.Axiom _ i ai Nothing c e) =
               ]
         , nest 2 $ text "nofPars =" <+> text (show n)
         ]
-    debugNonLinPars ks =
-      reportSDoc "tc.data.con" 15 $
-        text "these constructor parameters are non-linear:" <+> text (show ks)
     debugFitsIn s =
       reportSDoc "tc.data.con" 15 $ sep
         [ text "checking that the type fits in"
@@ -322,7 +280,7 @@ bindParameters' _ (A.DomainFull (A.TypedBindings _ (Arg _ A.TLet{})) : _) _ _ =
   __IMPOSSIBLE__
 
 bindParameters' ts0 ps0@(A.DomainFree info x : ps) t ret = do
-  case ignoreSharing $ unEl t of
+  case unEl t of
     -- Andreas, 2011-04-07 ignore relevance information in binding?!
     Pi arg@(Dom info' a) b -> do
       if | info == info'                  -> do
@@ -331,7 +289,7 @@ bindParameters' ts0 ps0@(A.DomainFree info x : ps) t ret = do
             ts <- caseList ts0 (return []) $ \ t0 ts -> do
               equalType t0 a
               return ts
-            continue ts ps x
+            continue ts ps $ A.unBind x
 
          | visible info, notVisible info' ->
             continue ts0 ps0 =<< freshName_ (absName b)
@@ -340,7 +298,7 @@ bindParameters' ts0 ps0@(A.DomainFree info x : ps) t ret = do
              -- Andreas, 2016-12-30 Concrete.Definition excludes this case
       where
       continue ts ps x = do
-        addContext' (x, arg) $
+        addContext (x, arg) $
           bindParameters' (raise 1 ts) ps (absBody b) $ \ tel s ->
             ret (ExtendTel arg $ Abs (nameToArgName x) tel) s
     _ -> __IMPOSSIBLE__
@@ -351,8 +309,8 @@ bindParameters' ts0 ps0@(A.DomainFree info x : ps) t ret = do
 --
 --   As a side effect, return the arity of the constructor.
 
-fitsIn :: Type -> Sort -> TCM Int
-fitsIn t s = do
+fitsIn :: [IsForced] -> Type -> Sort -> TCM Int
+fitsIn forceds t s = do
   reportSDoc "tc.data.fits" 10 $
     sep [ text "does" <+> prettyTCM t
         , text "of sort" <+> prettyTCM (getSort t)
@@ -363,19 +321,16 @@ fitsIn t s = do
   -- s' <- instantiateFull (getSort t)
   -- noConstraints $ s' `leqSort` s
   t <- reduce t
-  case ignoreSharing $ unEl t of
+  case unEl t of
     Pi dom b -> do
       withoutK <- optWithoutK <$> pragmaOptions
-      -- Forced constructor arguments are ignored in size-checking.
-      when (withoutK || notForced (getRelevance dom)) $ do
+      let (forced,forceds') = nextIsForced forceds
+      unless (isForced forced && not withoutK) $ do
         sa <- reduce $ getSort dom
         unless (sa == SizeUniv) $ sa `leqSort` s
       addContext (absName b, dom) $ do
-        succ <$> fitsIn (absBody b) (raise 1 s)
+        succ <$> fitsIn forceds' (absBody b) (raise 1 s)
     _ -> return 0 -- getSort t `leqSort` s  -- Andreas, 2013-04-13 not necessary since constructor type ends in data type
-  where
-    notForced Forced{} = False
-    notForced _        = True
 
 -- | Return the parameters that share variables with the indices
 -- nonLinearParameters :: Int -> Type -> TCM [Int]
@@ -384,18 +339,14 @@ fitsIn t s = do
 -- | Check that a type constructs something of the given datatype. The first
 --   argument is the number of parameters to the datatype.
 --
---   As a side effect, return the parameters that occur free in indices.
---   E.g. in @data Eq (A : Set)(a : A) : A -> Set where refl : Eq A a a@
---   this would include parameter @a@, but not @A@.
---
---   TODO: what if there's a meta here?
-constructs :: Int -> Type -> QName -> TCM [Int]
+constructs :: Int -> Type -> QName -> TCM ()
 constructs nofPars t q = constrT 0 t
     where
-        constrT :: Nat -> Type -> TCM [Int]
+        -- The number n counts the proper (non-parameter) constructor arguments.
+        constrT :: Nat -> Type -> TCM ()
         constrT n t = do
             t <- reduce t
-            case ignoreSharing $ unEl t of
+            case unEl t of
                 Pi _ (NoAbs _ b)  -> constrT n b
                 Pi a b            -> underAbstraction a b $ constrT (n + 1)
                   -- OR: addCxtString (absName b) a $ constrT (n + 1) (absBody b)
@@ -404,18 +355,21 @@ constructs nofPars t q = constrT 0 t
                   (pars, ixs) <- normalise $ splitAt nofPars vs
                   -- check that the constructor parameters are the data parameters
                   checkParams n pars
-                  -- compute the non-linear parameters
-                  m <- getContextSize  -- Note: n /= m if NoAbs encountered
-                  let nl = nonLinearParams m pars ixs
-                  -- assert that these are correct indices into the parameter vector
-                  when (any (< 0) nl) __IMPOSSIBLE__
-                  when (any (>= nofPars) nl) __IMPOSSIBLE__
-                  return nl
                 MetaV{} -> do
                   def <- getConstInfo q
-                  xs <- newArgsMeta $ defType def
-                  let t' = El (dataSort $ theDef def) $ Def q $ map Apply xs
-                  equalType t t'
+                  -- Analyse the type of q (name of the data type)
+                  let td = defType def
+                  TelV tel core <- telView td
+                  -- Construct the parameter arguments
+                  -- The parameters are @n + nofPars - 1 .. n@
+                  let us = zipWith (\ arg x -> var x <$ arg ) (telToArgs tel) $
+                             take nofPars $ downFrom (nofPars + n)
+                  -- The indices are fresh metas
+                  xs <- newArgsMeta =<< piApplyM td us
+                  let t' = El (dataSort $ theDef def) $ Def q $ map Apply $ us ++ xs
+                  -- Andreas, 2017-11-07, issue #2840
+                  -- We should not postpone here, otherwise we might upset the positivity checker.
+                  noConstraints $ equalType t t'
                   constrT n t'
                 _ -> typeError $ ShouldEndInApplicationOfTheDatatype t
 
@@ -431,18 +385,6 @@ constructs nofPars t q = constrT 0 t
                     t <- typeOfBV i
                     equalTerm t (unArg arg) (var i)
 
-        -- return the parameters (numbered 0,1,...,size pars-1 from left to right)
-        -- that occur relevantly in the indices
-        nonLinearParams n pars ixs =
-          -- compute the free de Bruijn indices in the data indices
-          -- ALT: Ignore all sorts?
-          let fv = allRelevantVarsIgnoring IgnoreInAnnotations ixs
-          -- keep relevant ones, convert to de Bruijn levels
-          -- note: xs is descending list
-              xs = map ((n-1) -) $ VarSet.toList fv
-          -- keep those that correspond to parameters
-          -- in ascending list
-          in  reverse $ filter (< size pars) xs
 
 {- UNUSED, Andreas 2012-09-13
 -- | Force a type to be a specific datatype.
@@ -450,7 +392,7 @@ forceData :: QName -> Type -> TCM Type
 forceData d (El s0 t) = liftTCM $ do
     t' <- reduce t
     d  <- canonicalName d
-    case ignoreSharing t' of
+    case t' of
         Def d' _
             | d == d'   -> return $ El s0 t'
             | otherwise -> fail $ "wrong datatype " ++ show d ++ " != " ++ show d'
@@ -468,7 +410,7 @@ forceData d (El s0 t) = liftTCM $ do
 isCoinductive :: Type -> TCM (Maybe Bool)
 isCoinductive t = do
   El s t <- reduce t
-  case ignoreSharing t of
+  case t of
     Def q _ -> do
       def <- getConstInfo q
       case theDef def of
@@ -489,5 +431,4 @@ isCoinductive t = do
     Pi    {} -> return (Just False)
     Sort  {} -> return (Just False)
     MetaV {} -> return Nothing
-    Shared{} -> __IMPOSSIBLE__
     DontCare{} -> __IMPOSSIBLE__
