@@ -3,7 +3,11 @@
 
 module Agda.TypeChecking.Rules.Decl where
 
-import Prelude hiding (null)
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ( (<>), null )
+#else
+import Prelude hiding ( null )
+#endif
 
 import Control.Monad
 import Control.Monad.Reader
@@ -60,6 +64,7 @@ import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Unquote
 import Agda.TypeChecking.Warnings
 
+import Agda.TypeChecking.Rules.Application
 import Agda.TypeChecking.Rules.Term
 import Agda.TypeChecking.Rules.Data    ( checkDataDef )
 import Agda.TypeChecking.Rules.Record  ( checkRecDef )
@@ -297,7 +302,7 @@ revisitRecordPatternTranslation qs = do
   -- and the set of function definitions in the mutual block
   classify q = inConcreteOrAbstractMode q $ \ def -> do
     case theDef def of
-      Record{ recEtaEquality' = Inferred True } -> return $ Just $ Left q
+      Record{ recEtaEquality' = Inferred YesEta } -> return $ Just $ Left q
       Function
         { funProjection = Nothing
             -- Andreas, 2017-08-10, issue #2664:
@@ -390,15 +395,10 @@ highlight_ d = do
       -- all that remains is the module declaration.
     A.RecSig{}               -> highlight d
     A.RecDef i x ind eta c ps tel cs ->
-      highlight (A.RecDef i x ind eta c [] dummy (fields cs))
-      -- The telescope and all record module declarations except
-      -- for the fields have already been highlighted.
+      highlight (A.RecDef i x ind eta c [] dummy cs)
+      -- The telescope has already been highlighted.
       where
-      fields (A.ScopedDecl _ ds1 : ds2) = fields ds1 ++ fields ds2
-      fields (d@A.Field{}        : ds)  = d : fields ds
-      fields (_                  : ds)  = fields ds
-      fields []                         = []
-      -- Andreas, 2016-01-22, issue 1791
+      -- Andreas, 2016-01-22, issue 1790
       -- The expression denoting the record constructor type
       -- is replace by a dummy expression in order to /not/
       -- generate highlighting from it.
@@ -602,7 +602,11 @@ checkPrimitive i x e =
     bindPrimitive s pf
     addConstant x $
       defaultDefn defaultArgInfo x t $
-        Primitive (Info.defAbstract i) s [] Nothing
+        Primitive { primAbstr    = Info.defAbstract i
+                  , primName     = s
+                  , primClauses  = []
+                  , primInv      = NotInjective
+                  , primCompiled = Nothing }
 
 assertCurrentModule :: QName -> String -> TCM ()
 assertCurrentModule x err =
@@ -683,11 +687,12 @@ checkPragma r p =
             Function{} -> markStatic x
             _          -> typeError $ GenericError "STATIC directive only works on functions"
         A.InjectivePragma x -> markInjective x
-        A.InlinePragma x -> do
+        A.InlinePragma b x -> do
           def <- getConstInfo x
           case theDef def of
-            Function{} -> markInline x
-            _          -> typeError $ GenericError "INLINE directive only works on functions"
+            Function{} -> markInline b x
+            _          -> typeError $ GenericError $ sINLINE ++ " directive only works on functions"
+              where sINLINE = if b then "INLINE" else "NOINLINE"
         A.OptionsPragma{} -> typeError $ GenericError $ "OPTIONS pragma only allowed at beginning of file, before top module declaration"
         A.DisplayPragma f ps e -> checkDisplayPragma f ps e
         A.EtaPragma r -> do
@@ -696,11 +701,11 @@ checkPragma r p =
           caseMaybeM (isRecord r) noRecord $ \case
             Record{ recInduction = ind, recEtaEquality' = eta } -> do
               unless (ind == Just CoInductive) $ noRecord
-              when (eta == Specified False) $ typeError $ GenericError $
-                "ETA pragram conflicts with no-eta-equality declaration"
+              when (eta == Specified NoEta) $ typeError $ GenericError $
+                "ETA pragma conflicts with no-eta-equality declaration"
             _ -> __IMPOSSIBLE__
           modifySignature $ updateDefinition r $ updateTheDef $ \case
-            def@Record{} -> def { recEtaEquality' = Specified True }
+            def@Record{} -> def { recEtaEquality' = Specified YesEta }
             _ -> __IMPOSSIBLE__
 
 -- | Type check a bunch of mutual inductive recursive definitions.
@@ -847,7 +852,8 @@ checkSectionApplication' i m1 (A.SectionApp ptel m2 args) copyInfo = do
       ]
     -- Now, type check arguments.
     ts <- (noConstraints $ checkArguments_ DontExpandLast (getRange i) args tel') >>= \case
-      (ts, etaTel') | (size etaTel == size etaTel') -> return ts
+      (ts', etaTel') | (size etaTel == size etaTel')
+                     , Just ts <- allApplyElims ts' -> return ts
       _ -> __IMPOSSIBLE__
     -- Perform the application of the module parameters.
     let aTel = tel' `apply` ts
@@ -868,7 +874,7 @@ checkSectionApplication' i m1 (A.SectionApp ptel m2 args) copyInfo = do
     args <- instantiateFull $ vs ++ ts
     let n = size aTel
     etaArgs <- inTopContext $ addContext aTel getContextArgs
-    addContext' (KeepNames aTel) $
+    addContext (KeepNames aTel) $
       applySection m1 (ptel `abstract` aTel) m2 (raise n args ++ etaArgs) copyInfo
 
 checkSectionApplication' i m1 (A.RecordModuleIFS x) copyInfo = do
@@ -911,8 +917,8 @@ checkSectionApplication' i m1 (A.RecordModuleIFS x) copyInfo = do
   when (tel == EmptyTel) $
     typeError $ GenericError $ prettyShow (qnameToConcrete name) ++ " is not a parameterised section"
 
-  addContext' telInst $ do
-    vs <- moduleParamsToApply $ qnameModule name
+  addContext telInst $ do
+    vs <- moduleParamsToApply x
     reportSDoc "tc.mod.apply" 20 $ vcat
       [ nest 2 $ text "vs      =" <+> sep (map prettyTCM vs)
       , nest 2 $ text "args    =" <+> sep (map (parens . prettyTCM) args)
