@@ -3,22 +3,27 @@
 
 module Agda.TypeChecking.Rules.Def where
 
-import Prelude hiding (mapM)
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ( (<>), mapM, null )
+#else
+import Prelude hiding ( mapM, null )
+#endif
+
 import Control.Arrow ((***))
-import Control.Applicative
 import Control.Monad.State hiding (forM, mapM)
 import Control.Monad.Reader hiding (forM, mapM)
 
 import Data.Function
 import Data.Maybe
-import Data.Traversable
+import Data.Traversable (Traversable, traverse, forM, mapM)
+import qualified Data.Map as Map
 import qualified Data.Set as Set
 
 import Agda.Syntax.Common
 import qualified Agda.Syntax.Concrete as C
 import Agda.Syntax.Concrete (exprFieldA)
 import Agda.Syntax.Position
-import Agda.Syntax.Abstract.Pattern ( containsAbsurdPattern )
+import Agda.Syntax.Abstract.Pattern as A
 import qualified Agda.Syntax.Abstract as A
 import qualified Agda.Syntax.Abstract.Views as A
 import Agda.Syntax.Internal as I
@@ -34,6 +39,7 @@ import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Conversion
+import Agda.TypeChecking.Inlining
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Patterns.Abstract (expandPatternSynonyms)
@@ -47,11 +53,11 @@ import Agda.TypeChecking.Injectivity
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.SizedTypes.Solve
 import Agda.TypeChecking.RecordPatterns
-import Agda.TypeChecking.CompiledClause (CompiledClauses'(..))
+import Agda.TypeChecking.CompiledClause (CompiledClauses'(..), hasProjectionPatterns)
 import Agda.TypeChecking.CompiledClause.Compile
 import Agda.TypeChecking.Primitive hiding (Nat)
 
-import Agda.TypeChecking.Rules.Term                ( checkExpr, inferExpr, inferExprForWith, checkDontExpandLast, checkTelescope )
+import Agda.TypeChecking.Rules.Term                ( checkExpr, inferExpr, inferExprForWith, checkDontExpandLast, checkTelescope, catchIlltypedPatternBlockedOnMeta )
 import Agda.TypeChecking.Rules.LHS                 ( checkLeftHandSide, LHSResult(..), bindAsPatterns )
 import Agda.TypeChecking.Rules.LHS.Problem         ( AsBinding(..) )
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Decl ( checkDecls )
@@ -61,6 +67,7 @@ import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.Maybe ( whenNothing )
 import Agda.Utils.Monad
+import Agda.Utils.Null
 import Agda.Utils.Permutation
 import Agda.Utils.Pretty ( prettyShow )
 import qualified Agda.Utils.Pretty as P
@@ -80,7 +87,7 @@ checkFunDef delayed i name cs = do
         info  <- flip setRelevance defaultArgInfo <$> relOfConst name
         case isAlias cs t of
           Just (e, mc, x) ->
-            traceCall (CheckFunDef (getRange i) (qnameName name) cs) $ do
+            traceCall (CheckFunDefCall (getRange i) (qnameName name) cs) $ do
               -- Andreas, 2012-11-22: if the alias is in an abstract block
               -- it has been frozen.  We unfreeze it to enable type inference.
               -- See issue 729.
@@ -91,6 +98,10 @@ checkFunDef delayed i name cs = do
         -- If it's a macro check that it ends in Term → TC ⊤
         ismacro <- isMacro . theDef <$> getConstInfo name
         when (ismacro || Info.defMacro i == MacroDef) $ checkMacroType t
+    `catchIlltypedPatternBlockedOnMeta` \ (err, x) -> do
+        reportSDoc "tc.def" 20 $ vcat $
+          [ text "checking function definition got stuck on meta: " <+> text (show x) ]
+        addConstraint $ CheckFunDef delayed i name cs
 
 checkMacroType :: Type -> TCM ()
 checkMacroType t = do
@@ -111,12 +122,12 @@ isAlias cs t =
           -- if we have just one clause without pattern matching and
           -- without a type signature, then infer, to allow
           -- "aliases" for things starting with hidden abstractions
-          Just (e, mc) | Just x <- isMeta (ignoreSharing $ unEl t) -> Just (e, mc, x)
+          Just (e, mc) | Just x <- isMeta (unEl t) -> Just (e, mc, x)
           _ -> Nothing
   where
     isMeta (MetaV x _) = Just x
     isMeta _           = Nothing
-    trivialClause [A.Clause (A.LHS i (A.LHSHead f []) []) _ _ (A.RHS e mc) [] _] = Just (e, mc)
+    trivialClause [A.Clause (A.LHS i (A.LHSHead f [])) _ (A.RHS e mc) (A.WhereDecls _ []) _] = Just (e, mc)
     trivialClause _ = Nothing
 
 -- | Check a trivial definition of the form @f = e@
@@ -129,7 +140,7 @@ checkAlias t' ai delayed i name e mc = atClause name 0 (A.RHS e mc) $ do
 
 {-
   -- Infer the type of the rhs
-  (v, t) <- applyRelevanceToContext (argInfoRelevance ai) $
+  (v, t) <- applyRelevanceToContext (getRelevance ai) $
                                     inferOrCheck e (Just t')
   -- v <- coerce v t t'
 -}
@@ -207,7 +218,7 @@ checkFunDefS :: Type             -- ^ the type we expect the function to have
              -> TCM ()
 checkFunDefS t ai delayed extlam with i name withSub cs = do
 
-    traceCall (CheckFunDef (getRange i) (qnameName name) cs) $ do   -- TODO!! (qnameName)
+    traceCall (CheckFunDefCall (getRange i) (qnameName name) cs) $ do   -- TODO!! (qnameName)
         reportSDoc "tc.def.fun" 10 $
           sep [ text "checking body of" <+> prettyTCM name
               , nest 2 $ text ":" <+> prettyTCM t
@@ -233,7 +244,7 @@ checkFunDefS t ai delayed extlam with i name withSub cs = do
         cs <- traceCall NoHighlighting $ do -- To avoid flicker.
           forM (zip cs [0..]) $ \ (c, clauseNo) -> do
             atClause name clauseNo (A.clauseRHS c) $ do
-              c <- applyRelevanceToContext (argInfoRelevance ai) $ do
+              c <- applyRelevanceToContext (getRelevance ai) $ do
                 checkClause t withSub c
               -- Andreas, 2013-11-23 do not solve size constraints here yet
               -- in case we are checking the body of an extended lambda.
@@ -292,9 +303,13 @@ checkFunDefS t ai delayed extlam with i name withSub cs = do
         cc <- Bench.billTo [Bench.Coverage] $
           inTopContext $ compileClauses (Just (name, fullType)) cs
 
+        -- Clause compilation runs the coverage checker, which might add
+        -- some extra clauses.
+        cs <- defClauses <$> getConstInfo name
+
         reportSDoc "tc.cc" 60 $ inTopContext $ do
           sep [ text "compiled clauses of" <+> prettyTCM name
-              , nest 2 $ text (show cc)
+              , nest 2 $ pretty cc
               ]
 
         -- The macro tag might be on the type signature
@@ -304,7 +319,7 @@ checkFunDefS t ai delayed extlam with i name withSub cs = do
         inTopContext $ addConstant name =<< do
           -- If there was a pragma for this definition, we can set the
           -- funTerminates field directly.
-          useTerPragma $ defaultDefn ai name fullType $
+          defn <- autoInline $
              set funMacro (ismacro || Info.defMacro i == MacroDef) $
              emptyFunction
              { funClauses        = cs
@@ -314,8 +329,9 @@ checkFunDefS t ai delayed extlam with i name withSub cs = do
              , funAbstr          = Info.defAbstract i
              , funExtLam         = extlam
              , funWith           = with
-             , funCopatternLHS   = isCopatternLHS cs
+             , funCopatternLHS   = hasProjectionPatterns cc
              }
+          useTerPragma $ defaultDefn ai name fullType defn
 
         reportSDoc "tc.def.fun" 10 $ do
           sep [ text "added " <+> prettyTCM name <+> text ":"
@@ -339,13 +355,23 @@ useTerPragma def@Defn{ defName = name, theDef = fun@Function{}} = do
 useTerPragma def = return def
 
 
--- | Insert some patterns in the in with-clauses LHS of the given RHS
+-- | Insert some with-patterns into the with-clauses LHS of the given RHS.
+-- (Used for @rewrite@.)
 insertPatterns :: [A.Pattern] -> A.RHS -> A.RHS
-insertPatterns pats (A.WithRHS aux es cs) = A.WithRHS aux es (map insertToClause cs)
-    where insertToClause (A.Clause (A.LHS i lhscore ps) dots sdots rhs ds catchall)
-              = A.Clause (A.LHS i lhscore (pats ++ ps)) dots sdots (insertPatterns pats rhs) ds catchall
-insertPatterns pats (A.RewriteRHS qes rhs wh) = A.RewriteRHS qes (insertPatterns pats rhs) wh
-insertPatterns pats rhs = rhs
+insertPatterns pats = \case
+  A.WithRHS aux es cs -> A.WithRHS aux es $ for cs $
+    \ (A.Clause (A.LHS info core)                              spats rhs                       ds catchall) ->
+       A.Clause (A.LHS info (insertPatternsLHSCore pats core)) spats (insertPatterns pats rhs) ds catchall
+  A.RewriteRHS qes spats rhs wh -> A.RewriteRHS qes spats (insertPatterns pats rhs) wh
+  rhs@A.AbsurdRHS -> rhs
+  rhs@A.RHS{}     -> rhs
+
+-- | Insert with-patterns before the trailing with patterns.
+-- If there are none, append the with-patterns.
+insertPatternsLHSCore :: [A.Pattern] -> A.LHSCore -> A.LHSCore
+insertPatternsLHSCore pats = \case
+  A.LHSWith core wps [] -> A.LHSWith core (pats ++ wps) []
+  core                  -> A.LHSWith core pats []
 
 -- | Parameters for creating a @with@-function.
 data WithFunctionProblem
@@ -376,18 +402,16 @@ checkClause
   -> A.SpineClause -- ^ Clause.
   -> TCM Clause    -- ^ Type-checked clause.
 
-checkClause t withSub c@(A.Clause (A.SpineLHS i x aps withPats) namedDots strippedDots rhs0 wh catchall) = do
+checkClause t withSub c@(A.Clause (A.SpineLHS i x aps) strippedPats rhs0 wh catchall) = do
     reportSDoc "tc.lhs.top" 30 $ text "Checking clause" $$ prettyA c
-    unless (null withPats) $
-      typeError $ UnexpectedWithPatterns withPats
+    unlessNull (trailingWithPatterns aps) $ \ withPats -> do
+      typeError $ UnexpectedWithPatterns $ map namedArg withPats
     traceCall (CheckClause t c) $ do
       aps <- expandPatternSynonyms aps
       cxtNames <- reverse . map (fst . unDom) <$> getContext
-      when (not $ null namedDots) $ reportSDoc "tc.lhs.top" 50 $
-        text "namedDots:" <+> vcat [ prettyTCM x <+> text "=" <+> prettyTCM v <+> text ":" <+> prettyTCM a | A.NamedDot x v a <- namedDots ]
-      -- Not really an as-pattern, but this does the right thing.
-      bindAsPatterns [ AsB x v a | A.NamedDot x v a <- namedDots ] $
-        checkLeftHandSide (CheckPatternShadowing c) (Just x) aps t withSub strippedDots $ \ lhsResult@(LHSResult npars delta ps trhs patSubst asb) -> do
+      when (not $ null strippedPats) $ reportSDoc "tc.lhs.top" 50 $
+        text "strippedPats:" <+> vcat [ prettyA p <+> text "=" <+> prettyTCM v <+> text ":" <+> prettyTCM a | A.ProblemEq p v a <- strippedPats ]
+      checkLeftHandSide (CheckPatternShadowing c) (Just x) aps t withSub strippedPats $ \ lhsResult@(LHSResult npars delta ps absurdPat trhs patSubst asb) -> do
         -- Note that we might now be in irrelevant context,
         -- in case checkLeftHandSide walked over an irrelevant projection pattern.
 
@@ -410,10 +434,11 @@ checkClause t withSub c@(A.Clause (A.SpineLHS i x aps withPats) namedDots stripp
             updateRHS rhs@A.RHS{}               = rhs
             updateRHS rhs@A.AbsurdRHS{}         = rhs
             updateRHS (A.WithRHS q es cs)       = A.WithRHS q es (map updateClause cs)
-            updateRHS (A.RewriteRHS qes rhs wh) = A.RewriteRHS qes (updateRHS rhs) wh
+            updateRHS (A.RewriteRHS qes spats rhs wh) =
+              A.RewriteRHS qes (applySubst patSubst spats) (updateRHS rhs) wh
 
-            updateClause (A.Clause f dots sdots rhs wh ca) =
-              A.Clause f (applySubst patSubst dots) (applySubst patSubst sdots) (updateRHS rhs) wh ca
+            updateClause (A.Clause f spats rhs wh ca) =
+              A.Clause f (applySubst patSubst spats) (updateRHS rhs) wh ca
 
         (body, with) <- bindAsPatterns asb $ checkWhere wh $ checkRHS i x aps t' lhsResult rhs
 
@@ -471,9 +496,8 @@ checkRHS
   -> A.RHS                   -- ^ Rhs to check.
   -> TCM (Maybe Term, WithFunctionProblem)
                                               -- Note: the as-bindings are already bound (in checkClause)
-checkRHS i x aps t lhsResult@(LHSResult _ delta ps trhs _ _asb) rhs0 = handleRHS rhs0
+checkRHS i x aps t lhsResult@(LHSResult _ delta ps absurdPat trhs _ _asb) rhs0 = handleRHS rhs0
   where
-  absurdPat = containsAbsurdPattern aps
   handleRHS rhs =
     case rhs of
 
@@ -492,8 +516,8 @@ checkRHS i x aps t lhsResult@(LHSResult _ delta ps trhs _ _asb) rhs0 = handleRHS
       -- Andreas, 2014-01-17, Issue 1402:
       -- If the rewrites are discarded since lhs=rhs, then
       -- we can actually have where clauses.
-      A.RewriteRHS [] rhs wh -> checkWhere wh $ handleRHS rhs
-      A.RewriteRHS ((qname,eq):qes) rhs wh -> do
+      A.RewriteRHS [] strippedPats rhs wh -> checkWhere wh $ handleRHS rhs
+      A.RewriteRHS ((qname,eq):qes) strippedPats rhs wh -> do
 
         -- Action for skipping this rewrite.
         -- We do not want to create unsolved metas in case of
@@ -510,7 +534,7 @@ checkRHS i x aps t lhsResult@(LHSResult _ delta ps trhs _ _asb) rhs0 = handleRHS
              -- 3. and a large overall number of ?s.
              let sameIP = (==) `on` (^.stInteractionPoints)
              when (sameIP st st') $ put st
-             handleRHS $ A.RewriteRHS qes rhs wh
+             handleRHS $ A.RewriteRHS qes strippedPats rhs wh
 
         -- Get value and type of rewrite-expression.
 
@@ -535,7 +559,7 @@ checkRHS i x aps t lhsResult@(LHSResult _ delta ps trhs _ _asb) rhs0 = handleRHS
 
         -- Get the name of builtin REFL.
 
-        Con reflCon _ [] <- ignoreSharing <$> primRefl
+        Con reflCon _ [] <- primRefl
         reflInfo <- fmap (setOrigin Inserted) <$> getReflArgInfo reflCon
 
         -- Andreas, 2017-01-11:
@@ -547,7 +571,7 @@ checkRHS i x aps t lhsResult@(LHSResult _ delta ps trhs _ _asb) rhs0 = handleRHS
         -- -- rewriting with @refl@ is attempted.
         -- let isReflProof = do
         --      v <- reduce proof
-        --      case ignoreSharing v of
+        --      case v of
         --        Con c _ [] | c == reflCon -> return True
         --        _ -> return False
         -- ifM isReflProof recurse $ {- else -} do
@@ -555,7 +579,7 @@ checkRHS i x aps t lhsResult@(LHSResult _ delta ps trhs _ _asb) rhs0 = handleRHS
         -- Process 'rewrite' clause like a suitable 'with' clause.
 
         -- The REFL constructor might have an argument
-        let reflPat  = A.ConP (ConPatInfo ConOCon patNoRange) (AmbQ [conName reflCon]) $
+        let reflPat  = A.ConP (ConPatInfo ConOCon patNoRange False) (unambiguous $ conName reflCon) $
               maybeToList $ fmap (\ ai -> Arg ai $ unnamed $ A.WildP patNoRange) reflInfo
 
         -- Andreas, 2015-12-25  Issue #1740:
@@ -572,12 +596,16 @@ checkRHS i x aps t lhsResult@(LHSResult _ delta ps trhs _ _asb) rhs0 = handleRHS
         let rhs'     = insertPatterns pats rhs
             (rhs'', outerWhere) -- the where clauses should go on the inner-most with
               | null qes  = (rhs', wh)
-              | otherwise = (A.RewriteRHS qes rhs' wh, [])
+              | otherwise = (A.RewriteRHS qes strippedPats rhs' wh, A.noWhereDecls)
             -- Andreas, 2014-03-05 kill range of copied patterns
             -- since they really do not have a source location.
-            cs       = [A.Clause (A.LHS i (A.LHSHead x (killRange aps)) pats) [] [] rhs'' outerWhere False]
-
-        checkWithRHS x qname t lhsResult [withExpr] [withType] cs
+            cl = A.Clause (A.LHS i $ insertPatternsLHSCore pats $ A.LHSHead x $ killRange aps)
+                   strippedPats rhs'' outerWhere False
+        reportSDoc "tc.rewrite" 60 $ vcat
+          [ text "rewrite"
+          , text "  rhs' = " <> (text . show) rhs'
+          ]
+        checkWithRHS x qname t lhsResult [withExpr] [withType] [cl]
 
       -- Case: @with@
       A.WithRHS aux es cs -> do
@@ -613,7 +641,7 @@ checkWithRHS
   -> [A.Clause]              -- ^ With-clauses to check.
   -> TCM (Maybe Term, WithFunctionProblem)
                                 -- Note: as-bindings already bound (in checkClause)
-checkWithRHS x aux t (LHSResult npars delta ps trhs _ _asb) vs0 as cs = Bench.billTo [Bench.Typing, Bench.With] $ do
+checkWithRHS x aux t (LHSResult npars delta ps _absurdPat trhs _ _asb) vs0 as cs = Bench.billTo [Bench.Typing, Bench.With] $ do
         let withArgs = withArguments vs0 as
             perm = fromMaybe __IMPOSSIBLE__ $ dbPatPerm ps
         (vs, as)  <- normalise (vs0, as)
@@ -725,18 +753,16 @@ checkWithFunction cxtNames (WithFunction f aux t delta delta1 delta2 vs as b qs
   -- Check generated type directly in internal syntax.
   setCurrentRange cs
     (traceCall NoHighlighting $   -- To avoid flicker.
-      checkType withFunType)
+     checkType withFunType)
     `catchError` \err -> case err of
       TypeError s e -> do
         put s
         wt <- reify withFunType
-        enterClosure e $ do
-          traceCall (CheckWithFunctionType wt) . typeError
+        enterClosure e $ traceCall (CheckWithFunctionType wt) . typeError
       err           -> throwError err
 
   -- With display forms are closed
-
-  df <- makeGlobal =<< withDisplayForm f aux delta1 delta2 n qs perm' perm
+  df <- safeInTopContext $ makeOpen =<< withDisplayForm f aux delta1 delta2 n qs perm' perm
 
   reportSLn "tc.with.top" 20 "created with display form"
 
@@ -775,10 +801,12 @@ checkWithFunction cxtNames (WithFunction f aux t delta delta1 delta2 vs as b qs
 
 -- | Type check a where clause.
 checkWhere
-  :: [A.Declaration] -- ^ Where-declarations to check.
-  -> TCM a           -- ^ Continuation.
+  :: A.WhereDeclarations -- ^ Where-declarations to check.
+  -> TCM a               -- ^ Continuation.
   -> TCM a
-checkWhere ds ret = loop ds
+checkWhere wh@(A.WhereDecls whmod ds) ret = do
+  ensureNoNamedWhereInRefinedContext whmod
+  loop ds
   where
     loop ds = case ds of
       [] -> ret
@@ -789,6 +817,32 @@ checkWhere ds ret = loop ds
             ret
       _ -> __IMPOSSIBLE__
 
+    -- #2897: We can't handle named where-modules in refined contexts.
+    ensureNoNamedWhereInRefinedContext Nothing = return ()
+    ensureNoNamedWhereInRefinedContext (Just m) = traceCall (CheckNamedWhere m) $ do
+      args <- map unArg <$> (moduleParamsToApply =<< currentModule)
+      unless (isWeakening args) $ -- weakened contexts are fine
+        genericDocError =<< do
+          names <- map (argNameToString . fst . unDom) . telToList <$>
+                    (lookupSection =<< currentModule)
+          let pr x v = text (x ++ " =") <+> prettyTCM v
+          vcat
+            [ fsep (pwords $ "Named where-modules are not allowed when module parameters have been refined by pattern matching. " ++
+                             "See https://github.com/agda/agda/issues/2897.")
+            , text $ "In this case the module parameter" ++
+                     (if length args > 0 then "s have" else " has") ++
+                     " been refined to"
+            , nest 2 $ vcat (zipWith pr names args) ]
+      where
+        isWeakening [] = True
+        isWeakening (Var i [] : args) = isWk (i - 1) args
+          where
+            isWk i []                = True
+            isWk i (Var j [] : args) = i == j && isWk (i - 1) args
+            isWk _ _ = False
+        isWeakening _ = False
+
+
 -- | Enter a new section during type-checking.
 
 newSection :: ModuleName -> A.Telescope -> TCM a -> TCM a
