@@ -3,7 +3,6 @@
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE NondecreasingIndentation   #-}
 {-# LANGUAGE UndecidableInstances       #-}
-{-# LANGUAGE ViewPatterns               #-}
 
 -- | Unification algorithm for specializing datatype indices, as described in
 --     \"Unifiers as Equivalences: Proof-Relevant Unification of Dependently
@@ -111,7 +110,6 @@ import Prelude hiding (null)
 import Control.Arrow ((***))
 import Control.Applicative hiding (empty)
 import Control.Monad
-import Control.Monad.Plus
 import Control.Monad.State
 import Control.Monad.Trans.Maybe
 import Control.Monad.Reader
@@ -122,7 +120,6 @@ import qualified Data.Map as Map
 import Data.Semigroup hiding (Arg)
 import qualified Data.List as List
 
-import Data.Typeable (Typeable)
 import Data.Foldable (Foldable)
 import Data.Traversable (Traversable,traverse)
 import qualified Data.Traversable as Trav
@@ -142,6 +139,7 @@ import Agda.TypeChecking.Conversion -- equalTerm
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.DropArgs
+import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.Level (reallyUnLevelView)
 import Agda.TypeChecking.Reduce
 import qualified Agda.TypeChecking.Patterns.Match as Match
@@ -155,7 +153,7 @@ import Agda.TypeChecking.MetaVars (assignV, newArgsMetaCtx)
 import Agda.TypeChecking.EtaContract
 import Agda.Interaction.Options (optInjectiveTypeConstructors, optWithoutK)
 
-import Agda.TypeChecking.Rules.LHS.Problem hiding (Substitution)
+import Agda.TypeChecking.Rules.LHS.Problem
 -- import Agda.TypeChecking.SyntacticEquality
 
 import Agda.Utils.Except ( MonadError(catchError, throwError) )
@@ -183,7 +181,7 @@ data UnificationResult' a
   = Unifies  a                    -- ^ Unification succeeded.
   | NoUnify  NegativeUnification  -- ^ Terms are not unifiable.
   | DontKnow [UnificationFailure] -- ^ Some other error happened, unification got stuck.
-  deriving (Typeable, Show, Functor, Foldable, Traversable)
+  deriving (Show, Functor, Foldable, Traversable)
 
 -- | Unify indices.
 --
@@ -225,7 +223,7 @@ unifyIndices tel flex a us vs = liftTCM $ Bench.billTo [Bench.Typing, Bench.Chec
 ----------------------------------------------------
 
 data Equality = Equal
-  { eqType  :: Type
+  { eqType  :: Dom Type
   , eqLeft  :: Term
   , eqRight :: Term
   }
@@ -317,11 +315,11 @@ varCount :: UnifyState -> Int
 varCount = size . varTel
 
 -- | Get the type of the i'th variable in the given state
-getVarType :: Int -> UnifyState -> Type
-getVarType i s = if i < 0 then __IMPOSSIBLE__ else unDom $ (flattenTel $ varTel s) !! i
+getVarType :: Int -> UnifyState -> Dom Type
+getVarType i s = if i < 0 then __IMPOSSIBLE__ else (flattenTel $ varTel s) !! i
 
-getVarTypeUnraised :: Int -> UnifyState -> Type
-getVarTypeUnraised i s = if i < 0 then __IMPOSSIBLE__ else snd . unDom $ (telToList $ varTel s) !! i
+getVarTypeUnraised :: Int -> UnifyState -> Dom Type
+getVarTypeUnraised i s = if i < 0 then __IMPOSSIBLE__ else snd <$> (telToList $ varTel s) !! i
 
 eqCount :: UnifyState -> Int
 eqCount = size . eqTel
@@ -332,13 +330,13 @@ eqCount = size . eqTel
 getEquality :: Int -> UnifyState -> Equality
 getEquality k UState { eqTel = eqs, eqLHS = lhs, eqRHS = rhs } =
   if k < 0 then __IMPOSSIBLE__ else
-    Equal (unDom $ (flattenTel eqs) !! k) (unArg $ lhs !! k) (unArg $ rhs !! k)
+    Equal (flattenTel eqs !! k) (unArg $ lhs !! k) (unArg $ rhs !! k)
 
 -- | As getEquality, but with the unraised type
 getEqualityUnraised :: Int -> UnifyState -> Equality
 getEqualityUnraised k UState { eqTel = eqs, eqLHS = lhs, eqRHS = rhs } =
   if k < 0 then __IMPOSSIBLE__ else
-    Equal (snd . unDom $ (telToList eqs) !! k)
+    Equal (snd <$> (telToList eqs) !! k)
           (unArg $ lhs !! k)
           (unArg $ rhs !! k)
 
@@ -362,7 +360,9 @@ addEq a u v = addEqs (ExtendTel (defaultDom a) (Abs underscore EmptyTel)) [u] [v
 
 -- | Instantiate the k'th variable with the given value.
 --   Returns Nothing if there is a cycle.
-solveVar :: Int -> Term -> UnifyState -> Maybe (UnifyState, PatternSubstitution)
+solveVar :: Int    -- ^ Index @k@
+         -> Term   -- ^ Solution @u@
+         -> UnifyState -> Maybe (UnifyState, PatternSubstitution)
 solveVar k u s = case instantiateTelescope (varTel s) k u of
   Nothing -> Nothing
   Just (tel' , sigma , rho) -> Just $ (,sigma) $ UState
@@ -405,7 +405,7 @@ solveEq k u s = (,sigma) $ s
   where
     u'    = raise k u
     n     = eqCount s
-    sigma = liftS (n-k-1) $ consS (DotP u') idS
+    sigma = liftS (n-k-1) $ consS (dotP u') idS
 
 -- | Simplify the k'th equation with the given value (which can depend on other
 --   equation variables). Returns Nothing if there is a cycle.
@@ -433,8 +433,8 @@ data UnifyStep
     }
   | Solution
     { solutionAt         :: Int
-    , solutionType       :: Type
-    , solutionVar        :: Int
+    , solutionType       :: Dom Type
+    , solutionVar        :: FlexibleVar Int
     , solutionTerm       :: Term
     }
   | Injectivity
@@ -531,7 +531,7 @@ instance PrettyTCM UnifyStep where
       , text "record type:" <+> prettyTCM r
       , text "parameters: " <+> prettyTCM pars
       ])
-    EtaExpandEquation k r pars -> text "EtaExpandVar" $$ nest 2 (vcat $
+    EtaExpandEquation k r pars -> text "EtaExpandEquation" $$ nest 2 (vcat $
       [ text "position:   " <+> text (show k)
       , text "record type:" <+> prettyTCM r
       , text "parameters: " <+> prettyTCM pars
@@ -589,72 +589,6 @@ isHom n x = do
   guard $ getAll $ runFree (All . (>= n)) IgnoreNot x
   return $ raise (-n) x
 
--- | Checks whether the given term (of the given type) is beta-eta-equivalent
---   to a variable. Returns just the de Bruijn-index of the variable if it is,
---   or nothing otherwise.
-isEtaVar :: Term -> Type -> TCM (Maybe Int)
-isEtaVar u a = runMaybeT $ isEtaVarG u a Nothing []
-  where
-    -- Checks whether the term u (of type a) is beta-eta-equivalent to
-    -- `Var i es`, and returns i if it is. If the argument mi is `Just i'`,
-    -- then i and i' are also required to be equal (else Nothing is returned).
-    isEtaVarG :: Term -> Type -> Maybe Int -> [Elim' Int] -> MaybeT TCM Int
-    isEtaVarG u a mi es = do
-      (u, a) <- liftTCM $ reduce (u, a)
-      liftTCM $ reportSDoc "tc.lhs.unify" 80 $ text "isEtaVarG" <+> nest 2 (sep
-        [ text "u  = " <+> text (show u)
-        , text "a  = " <+> prettyTCM a
-        , text "mi = " <+> text (show mi)
-        , text "es = " <+> prettyList (map (text . show) es)
-        ])
-      case (ignoreSharing u, ignoreSharing $ unEl a) of
-        (Var i' es', _) -> do
-          guard $ mi == (i' <$ mi)
-          b <- liftTCM $ typeOfBV i'
-          areEtaVarElims (var i') b es' es
-          return i'
-        (_, Def d pars) -> do
-          guard =<< do liftTCM $ isEtaRecord d
-          fs <- liftTCM $ map unArg . recFields . theDef <$> getConstInfo d
-          is <- forM fs $ \f -> do
-            let o = ProjSystem
-            (_, _, fa) <- MaybeT $ projectTyped u a o f
-            isEtaVarG (u `applyE` [Proj o f]) fa mi (es ++ [Proj o f])
-          case (mi, is) of
-            (Just i, _)     -> return i
-            (Nothing, [])   -> mzero
-            (Nothing, i:is) -> guard (all (==i) is) >> return i
-        (_, Pi dom cod) -> addContext dom $ do
-          let u'  = raise 1 u `apply` [argFromDom dom $> var 0]
-              a'  = absBody cod
-              mi' = fmap (+1) mi
-              es' = (fmap . fmap) (+1) es ++ [Apply $ argFromDom dom $> 0]
-          (-1+) <$> isEtaVarG u' a' mi' es'
-        _ -> mzero
-
-    -- `areEtaVarElims u a es es'` checks whether the given elims es (as applied
-    -- to the term u of type a) are beta-eta-equal to either projections or
-    -- variables with de Bruijn indices given by es'.
-    areEtaVarElims :: Term -> Type -> Elims -> [Elim' Int] -> MaybeT TCM ()
-    areEtaVarElims u a []    []    = return ()
-    areEtaVarElims u a []    (_:_) = mzero
-    areEtaVarElims u a (_:_) []    = mzero
-    areEtaVarElims u a (Proj o f : es) (Proj _ f' : es') = do
-      guard $ f == f'
-      a       <- liftTCM $ reduce a
-      (_, _, fa) <- MaybeT $ projectTyped u a o f
-      areEtaVarElims (u `applyE` [Proj o f]) fa es es'
-    -- These two cases can occur only when we're looking at two different
-    -- variables (i.e. one of function type and the other of record type) so
-    -- it's definitely not the variable we're looking for (or someone is playing
-    -- Jedi mind tricks on us)
-    areEtaVarElims u a (Proj{}  : _ ) (Apply _ : _  ) = mzero
-    areEtaVarElims u a (Apply _ : _ ) (Proj{}  : _  ) = mzero
-    areEtaVarElims u a (Apply v : es) (Apply i : es') = do
-      ifNotPiType a (const mzero) $ \dom cod -> do
-      _ <- isEtaVarG (unArg v) (unDom dom) (Just $ unArg i) []
-      areEtaVarElims (u `apply` [fmap var i]) (cod `absApp` var (unArg i)) es es'
-
 findFlexible :: Int -> FlexibleVars -> Maybe (FlexibleVar Nat)
 findFlexible i flex =
   let flex'      = map flexVar flex
@@ -663,8 +597,8 @@ findFlexible i flex =
 
 basicUnifyStrategy :: Int -> UnifyStrategy
 basicUnifyStrategy k s = do
-  Equal a u v <- liftTCM $ eqUnLevel (getEquality k s)
-  ha <- mfromMaybe $ isHom n a
+  Equal (Dom info a) u v <- liftTCM $ eqUnLevel (getEquality k s)
+  ha <- fromMaybeMP $ isHom n a
   (mi, mj) <- liftTCM $ addContext (varTel s) $ (,) <$> isEtaVar u ha <*> isEtaVar v ha
   liftTCM $ reportSDoc "tc.lhs.unify" 30 $ text "isEtaVar results: " <+> text (show [mi,mj])
   case (mi, mj) of
@@ -674,10 +608,10 @@ basicUnifyStrategy k s = do
      | Just fi <- findFlexible i flex
      , Just fj <- findFlexible j flex -> do
        let choice = chooseFlex fi fj
-           firstTryLeft  = msum [ return (Solution k ha i v)
-                                , return (Solution k ha j u)]
-           firstTryRight = msum [ return (Solution k ha j u)
-                                , return (Solution k ha i v)]
+           firstTryLeft  = msum [ return (Solution k (Dom info ha) fi v)
+                                , return (Solution k (Dom info ha) fj u)]
+           firstTryRight = msum [ return (Solution k (Dom info ha) fj u)
+                                , return (Solution k (Dom info ha) fi v)]
        liftTCM $ reportSDoc "tc.lhs.unify" 40 $ text "fi = " <+> text (show fi)
        liftTCM $ reportSDoc "tc.lhs.unify" 40 $ text "fj = " <+> text (show fj)
        liftTCM $ reportSDoc "tc.lhs.unify" 40 $ text "chooseFlex: " <+> text (show choice)
@@ -687,9 +621,9 @@ basicUnifyStrategy k s = do
          ExpandBoth   -> mzero -- This should be taken care of by etaExpandEquationStrategy
          ChooseEither -> firstTryRight
     (Just i, _)
-     | Just _ <- findFlexible i flex -> return $ Solution k ha i v
+     | Just fi <- findFlexible i flex -> return $ Solution k (Dom info ha) fi v
     (_, Just j)
-     | Just _ <- findFlexible j flex -> return $ Solution k ha j u
+     | Just fj <- findFlexible j flex -> return $ Solution k (Dom info ha) fj u
     _ -> mzero
   where
     flex = flexVars s
@@ -697,16 +631,16 @@ basicUnifyStrategy k s = do
 
 dataStrategy :: Int -> UnifyStrategy
 dataStrategy k s = do
-  Equal a u v <- liftTCM $ eqConstructorForm =<< eqUnLevel (getEqualityUnraised k s)
-  case ignoreSharing $ unEl a of
-    Def d es -> do
-      npars <- mcatMaybes $ liftTCM $ getNumberOfParameters d
+  Equal (Dom _ a) u v <- liftTCM $ eqConstructorForm =<< eqUnLevel (getEqualityUnraised k s)
+  case unEl a of
+    Def d es | Type{} <- getSort a -> do
+      npars <- catMaybesMP $ liftTCM $ getNumberOfParameters d
       let (pars,ixs) = splitAt npars $ fromMaybe __IMPOSSIBLE__ $ allApplyElims es
-      hpars <- mfromMaybe $ isHom k pars
+      hpars <- fromMaybeMP $ isHom k pars
       liftTCM $ reportSDoc "tc.lhs.unify" 40 $ addContext (varTel s) $
         text "Found equation at datatype " <+> prettyTCM d
          <+> text " with (homogeneous) parameters " <+> prettyTCM hpars
-      case (ignoreSharing u, ignoreSharing v) of
+      case (u, v) of
         (Con c _ _   , Con c' _ _  ) | c == c' -> return $ Injectivity k a d hpars ixs c
         (Con c _ _   , Con c' _ _  ) -> return $ Conflict k d hpars u v
         (Var i []  , v         ) -> ifOccursStronglyRigid i v $ return $ Cycle k d hpars i v
@@ -725,16 +659,16 @@ dataStrategy k s = do
 
 checkEqualityStrategy :: Int -> UnifyStrategy
 checkEqualityStrategy k s = do
-  let Equal a u v = getEquality k s
+  let Equal (Dom _ a) u v = getEquality k s
       n = eqCount s
-  ha <- mfromMaybe $ isHom n a
+  ha <- fromMaybeMP $ isHom n a
   return $ Deletion k ha u v
 
 literalStrategy :: Int -> UnifyStrategy
 literalStrategy k s = do
   let n = eqCount s
-  Equal a u v <- liftTCM $ eqUnLevel $ getEquality k s
-  ha <- mfromMaybe $ isHom n a
+  Equal (Dom _ a) u v <- liftTCM $ eqUnLevel $ getEquality k s
+  ha <- fromMaybeMP $ isHom n a
   case (u , v) of
     (Lit l1 , Lit l2)
      | l1 == l2  -> return $ Deletion k ha u v
@@ -743,30 +677,39 @@ literalStrategy k s = do
 
 etaExpandVarStrategy :: Int -> UnifyStrategy
 etaExpandVarStrategy k s = do
-  Equal a u v <- liftTCM $ eqUnLevel (getEquality k s)
-  shouldEtaExpand u a s `mplus` shouldEtaExpand v a s
+  Equal (Dom _ a) u v <- liftTCM $ eqUnLevel (getEquality k s)
+  shouldEtaExpand u v a s `mplus` shouldEtaExpand v u a s
   where
     -- TODO: use IsEtaVar to check if the term is a variable
-    shouldEtaExpand :: Term -> Type -> UnifyStrategy
-    shouldEtaExpand (Var i es) a s = do
-      fi       <- mfromMaybe $ findFlexible i (flexVars s)
+    shouldEtaExpand :: Term -> Term -> Type -> UnifyStrategy
+    shouldEtaExpand (Var i es) v a s = do
+      fi       <- fromMaybeMP $ findFlexible i (flexVars s)
       liftTCM $ reportSDoc "tc.lhs.unify" 50 $
         text "Found flexible variable " <+> text (show i)
-      ps       <- mfromMaybe $ allProjElims es
-      guard $ not $ null ps
+      -- Issue 2888: Do this if there are projections or if it's a singleton
+      -- record or if it's unified against a record constructor term. Basically
+      -- we need to avoid EtaExpandEquation if EtaExpandVar is possible, or the
+      -- forcing translation is unhappy.
+      let Dom _ b = getVarTypeUnraised (varCount s - 1 - i) s
+      (d, pars) <- catMaybesMP $ liftTCM $ isEtaRecordType b
+      ps        <- fromMaybeMP $ allProjElims es
+      sing      <- liftTCM $ (Right True ==) <$> isSingletonRecord d pars
+      con       <- liftTCM $ isRecCon v  -- is the other term a record constructor?
+      guard $ not (null ps) || sing || con
       liftTCM $ reportSDoc "tc.lhs.unify" 50 $
         text "with projections " <+> prettyTCM (map snd ps)
-      let b = getVarTypeUnraised (varCount s - 1 - i) s
-      (d, pars) <- mcatMaybes $ liftTCM $ isEtaRecordType b
       liftTCM $ reportSDoc "tc.lhs.unify" 50 $
         text "at record type " <+> prettyTCM d
       return $ EtaExpandVar fi d pars
-    shouldEtaExpand _ _ _ = mzero
+    shouldEtaExpand _ _ _ _ = mzero
+
+    isRecCon (Con c _ _) = isJust <$> isRecordConstructor (conName c)
+    isRecCon _           = return False
 
 etaExpandEquationStrategy :: Int -> UnifyStrategy
 etaExpandEquationStrategy k s = do
-  let Equal a u v = getEqualityUnraised k s
-  (d, pars) <- mcatMaybes $ liftTCM $ addContext tel $ isEtaRecordType a
+  let Equal (Dom _ a) u v = getEqualityUnraised k s
+  (d, pars) <- catMaybesMP $ liftTCM $ addContext tel $ isEtaRecordType a
   sing <- liftTCM $ (Right True ==) <$> isSingletonRecord d pars
   projLeft <- liftTCM $ shouldProject u
   projRight <- liftTCM $ shouldProject v
@@ -774,7 +717,7 @@ etaExpandEquationStrategy k s = do
   return $ EtaExpandEquation k d pars
   where
     shouldProject :: Term -> TCM Bool
-    shouldProject u = case ignoreSharing u of
+    shouldProject u = case u of
       Def f es   -> usesCopatterns f
       Con c _ _  -> isJust <$> isRecordConstructor (conName c)
 
@@ -786,15 +729,14 @@ etaExpandEquationStrategy k s = do
       Level _    -> __IMPOSSIBLE__
       MetaV _ _  -> return False
       DontCare _ -> return False
-      Shared _   -> __IMPOSSIBLE__
 
     tel = varTel s `abstract` telFromList (take k $ telToList $ eqTel s)
 
 simplifySizesStrategy :: Int -> UnifyStrategy
 simplifySizesStrategy k s = do
   isSizeName <- liftTCM isSizeNameTest
-  let Equal a u v = getEquality k s
-  case ignoreSharing $ unEl a of
+  let Equal (Dom _ a) u v = getEquality k s
+  case unEl a of
     Def d _ -> do
       guard $ isSizeName d
       su <- liftTCM $ sizeView u
@@ -812,7 +754,7 @@ injectiveTypeConStrategy k s = do
   guard injTyCon
   eq <- liftTCM $ eqUnLevel $ getEquality k s
   case eq of
-    Equal a u@(ignoreSharing -> Def d es) v@(ignoreSharing -> Def d' es') | d == d' -> do
+    Equal a u@(Def d es) v@(Def d' es') | d == d' -> do
       -- d must be a data, record or axiom
       def <- liftTCM $ getConstInfo d
       guard $ case theDef def of
@@ -832,7 +774,7 @@ injectivePragmaStrategy :: Int -> UnifyStrategy
 injectivePragmaStrategy k s = do
   eq <- liftTCM $ eqUnLevel $ getEquality k s
   case eq of
-    Equal a u@(ignoreSharing -> Def d es) v@(ignoreSharing -> Def d' es') | d == d' -> do
+    Equal a u@(Def d es) v@(Def d' es') | d == d' -> do
       -- d must have an injective pragma
       def <- liftTCM $ getConstInfo d
       guard $ defInjective def
@@ -907,12 +849,15 @@ unifyStep s Deletion{ deleteAt = k , deleteType = a , deleteLeft = u , deleteRig
         tellUnifyProof sigma
         Unifies <$> liftTCM (reduceEqTel s')
 
-unifyStep s Solution{ solutionAt = k , solutionType = a , solutionVar = i , solutionTerm = u } = do
+unifyStep s Solution{ solutionAt   = k
+                    , solutionType = Dom info a
+                    , solutionVar  = fi@FlexibleVar{ flexVar = i }
+                    , solutionTerm = u } = do
   let m = varCount s
 
   -- Check that the type of the variable is equal to the type of the equation
   -- (not just a subtype), otherwise we cannot instantiate (see Issue 2407).
-  let a' = getVarType (m-1-i) s
+  let Dom info' a' = getVarType (m-1-i) s
   equalTypes <- liftTCM $ addContext (varTel s) $ do
     reportSDoc "tc.lhs.unify" 45 $ text "Equation type: " <+> prettyTCM a
     reportSDoc "tc.lhs.unify" 45 $ text "Variable type: " <+> prettyTCM a'
@@ -920,15 +865,30 @@ unifyStep s Solution{ solutionAt = k , solutionType = a , solutionVar = i , solu
       equalType a a'
     return Nothing
     `catchError` \err -> return $ Just err
+  -- The conditions on the relevances are as follows (see #2640):
+  -- - If the type of the equation is relevant, then the solution must be
+  --   usable in a relevant position.
+  -- - If the type of the equation is (shape-)irrelevant, then the solution
+  --   must be usable in a μ-relevant position where μ is the relevance
+  --   of the variable being solved.
+  let eqrel  = getRelevance info
+      varrel = getRelevance info'
+      rel    = if NonStrict `moreRelevant` eqrel then varrel else Relevant
+  reportSDoc "tc.lhs.unify" 65 $ text $ "Equation relevance: " ++ show eqrel
+  reportSDoc "tc.lhs.unify" 65 $ text $ "Variable relevance: " ++ show varrel
+  reportSDoc "tc.lhs.unify" 65 $ text $ "Solution must be usable in a " ++ show rel ++ " position."
+  usable <- liftTCM $ addContext (varTel s) $ usableRel rel u
+  reportSDoc "tc.lhs.unify" 45 $ text "Relevance ok: " <+> prettyTCM usable
   case equalTypes of
     Just err -> return $ DontKnow []
-    Nothing  -> caseMaybeM (trySolveVar (m-1-i) u s)
+    Nothing | usable -> caseMaybeM (trySolveVar (m-1-i) u s)
       (return $ DontKnow [UnifyRecursiveEq (varTel s) a i u])
       (\(s',sub) -> do
         tellUnifySubst sub
         let (s'', sigma) = solveEq k (applyPatSubst sub u) s'
         tellUnifyProof sigma
         Unifies <$> liftTCM (reduce s''))
+    Nothing -> return $ DontKnow []
   where
     trySolveVar i u s = case solveVar i u s of
       Just x  -> return $ Just x
@@ -946,7 +906,7 @@ unifyStep s (Injectivity k a d pars ixs c) = do
   addContext (varTel s) $ reportSDoc "tc.lhs.unify" 40 $
     text "Constructor type: " <+> prettyTCM ctype
   TelV ctel ctarget <- liftTCM $ telView ctype
-  let cixs = case ignoreSharing $ unEl ctarget of
+  let cixs = case unEl ctarget of
                Def d' es | d == d' ->
                  let args = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
                  in  drop (length pars) args
@@ -980,10 +940,39 @@ unifyStep s (Injectivity k a d pars ixs c) = do
     -- same type for distinct constructors c1 and c2.
     NoUnify _ -> __IMPOSSIBLE__
 
-    -- TODO: we could still make progress here if not --without-K,
-    -- but I'm not sure if it's necessary.
+    -- Higher-dimensional unification has failed. If not --without-K,
+    -- we can simply ignore the higher-dimensional equations and
+    -- simplify the equation as in the non-indexed case.
+    DontKnow _ | not withoutK -> do
+      -- using the same variable names as in the case where hdu succeeds.
+      let eqTel1' = eqTel1 `abstract` raise (size eqTel1) ctel
+          rho1    = raiseS (size ctel)
+          ceq     = ConP c noConPatternInfo $ teleNamedArgs ctel
+          rho3    = consS ceq rho1
+          eqTel2' = applyPatSubst rho3 eqTel2
+          eqTel'  = eqTel1' `abstract` eqTel2'
+          rho     = liftS (size eqTel2) rho3
+
+      tellUnifyProof rho
+
+      eqTel' <- liftTCM $ reduce eqTel'
+
+      -- Compute new lhs and rhs by matching the old ones against rho
+      (lhs', rhs') <- liftTCM . reduce =<< do
+        let ps = applySubst rho $ teleNamedArgs $ eqTel s
+        (lhsMatch, _) <- liftTCM $ runReduceM $ Match.matchPatterns ps $ eqLHS s
+        (rhsMatch, _) <- liftTCM $ runReduceM $ Match.matchPatterns ps $ eqRHS s
+        case (lhsMatch, rhsMatch) of
+          (Match.Yes _ lhs', Match.Yes _ rhs') -> return
+            (reverse $ Match.matchedArgs __IMPOSSIBLE__ (size eqTel') lhs',
+             reverse $ Match.matchedArgs __IMPOSSIBLE__ (size eqTel') rhs')
+          _ -> __IMPOSSIBLE__
+
+      return $ Unifies $ s { eqTel = eqTel' , eqLHS = lhs' , eqRHS = rhs' }
+
+
     DontKnow _ -> let n           = eqCount s
-                      Equal a u v = getEquality k s
+                      Equal (Dom _ a) u v = getEquality k s
                   in return $ DontKnow [UnifyIndicesNotVars
                        (varTel s `abstract` eqTel s) a
                        (raise n u) (raise n v) (raise (n-k) ixs)]
