@@ -26,6 +26,7 @@ import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Polarity
 import Agda.TypeChecking.Irrelevance
+import Agda.TypeChecking.CompiledClause (hasProjectionPatterns)
 import Agda.TypeChecking.CompiledClause.Compile
 
 import Agda.TypeChecking.Rules.Data ( bindParameters, fitsIn, forceSort)
@@ -62,7 +63,7 @@ checkRecDef
   :: Info.DefInfo              -- ^ Position and other info.
   -> QName                     -- ^ Record type identifier.
   -> Maybe (Ranged Induction)  -- ^ Optional: (co)inductive declaration.
-  -> Maybe Bool                -- ^ Optional: user specified eta/no-eta
+  -> Maybe HasEta              -- ^ Optional: user specified eta/no-eta
   -> Maybe QName               -- ^ Optional: constructor name.
   -> [A.LamBinding]            -- ^ Record parameters.
   -> A.Expr                    -- ^ Approximate type of constructor (@fields@ -> Set).
@@ -86,8 +87,7 @@ checkRecDef i name ind eta con ps contel fields =
 
       -- Check and evaluate field types.
       reportSDoc "tc.rec" 15 $ text "checking fields"
-      -- WRONG: contype <- workOnTypes $ killRange <$> (instantiateFull =<< isType_ contel)
-      contype <- instantiateFull =<< isType_ contel
+      contype <- workOnTypes $ instantiateFull =<< isType_ contel
       reportSDoc "tc.rec" 20 $ vcat
         [ text "contype = " <+> prettyTCM contype ]
 
@@ -141,7 +141,7 @@ checkRecDef i name ind eta con ps contel fields =
           -- Andreas, 2016-09-20, issue #2197.
           -- Eta is inferred by the positivity checker.
           -- We should turn it off until it is proven to be safe.
-          haveEta      = maybe (Inferred False) Specified eta
+          haveEta      = maybe (Inferred NoEta) Specified eta
           -- haveEta      = maybe (Inferred $ conInduction == Inductive && etaenabled) Specified eta
           con = ConHead conName conInduction $ map unArg fs
 
@@ -150,13 +150,13 @@ checkRecDef i name ind eta con ps contel fields =
           -- See issue 392.
           -- Unless it's been declared coinductive or no-eta-equality (#2607).
           recordRelevance
-            | eta          == Just False  = Relevant
+            | eta          == Just NoEta  = Relevant
             | conInduction == CoInductive = Relevant
             | otherwise                   = minimum $ Irrelevant : (map getRelevance $ telToList ftel)
 
       -- Andreas, 2017-01-26, issue #2436
       -- Disallow coinductive records with eta-equality
-      when (conInduction == CoInductive && etaEqualityToBool haveEta == True) $ do
+      when (conInduction == CoInductive && theEtaEquality haveEta == YesEta) $ do
         typeError . GenericDocError =<< do
           sep [ text "Agda doesn't like coinductive records with eta-equality."
               , text "If you must, use pragma"
@@ -202,6 +202,7 @@ checkRecDef i name ind eta con ps contel fields =
               , conData   = name
               , conAbstr  = Info.defAbstract conInfo
               , conInd    = conInduction
+              , conForced = []
               , conErased = []
               }
 
@@ -210,7 +211,7 @@ checkRecDef i name ind eta con ps contel fields =
         addNamedInstance conName name
 
       -- Check that the fields fit inside the sort
-      _ <- contype `fitsIn` s
+      _ <- fitsIn [] contype s
 
       {- Andreas, 2011-04-27 WRONG because field types are checked again
          and then non-stricts should not yet be irrelevant
@@ -247,7 +248,7 @@ checkRecDef i name ind eta con ps contel fields =
 -}
 
       let info = setRelevance recordRelevance defaultArgInfo
-          addRecordVar = addContext' ("", Dom info rect)
+          addRecordVar = addContext ("", Dom info rect)
           -- the record variable has the empty name by intention, see issue 208
 
       let m = qnameToMName name  -- Name of record module.
@@ -256,7 +257,7 @@ checkRecDef i name ind eta con ps contel fields =
       -- section telescope changes the semantics, see e.g.
       -- test/Succeed/RecordInParModule.
       -- Ulf, 2016-03-02 but it's the right thing to do (#1759)
-      modifyContext (modifyContextEntries hideOrKeepInstance) $ addRecordVar $ do
+      modifyContext (map hideOrKeepInstance) $ addRecordVar $ do
 
         -- Add the record section.
 
@@ -275,7 +276,7 @@ checkRecDef i name ind eta con ps contel fields =
       -- Andreas, 2016-02-09, Issue 1815 (see also issue 1759).
       -- For checking the record declarations, hide the record parameters
       -- and the parameters of the parent modules.
-      modifyContext (modifyContextEntries hideOrKeepInstance) $ addRecordVar $ do
+      modifyContext (map hideOrKeepInstance) $ addRecordVar $ do
 
         -- Check the types of the fields and the other record declarations.
         withCurrentModule m $ do
@@ -286,7 +287,7 @@ checkRecDef i name ind eta con ps contel fields =
           -- record type.
           -- See test/Succeed/ProjectionsTakeModuleTelAsParameters.agda.
           tel' <- getContextTelescope
-          setDefaultModuleParameters m
+          setModuleCheckpoint m
           checkRecordProjections m name hasNamedCon con tel' (raise 1 ftel) fields
 
       return ()
@@ -319,6 +320,7 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
     checkProjs ftel1 ftel2 (A.ScopedDecl scope fs' : fs) =
       setScope scope >> checkProjs ftel1 ftel2 (fs' ++ fs)
 
+    -- Case: projection.
     checkProjs ftel1 (ExtendTel (Dom ai t) ftel2) (A.Field info x _ : fs) =
       traceCall (CheckProjection (getRange info) x t) $ do
       -- Andreas, 2012-06-07:
@@ -392,15 +394,16 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
               Relevant   -> id
               NonStrict  -> id
               Irrelevant -> DontCare
-              _          -> __IMPOSSIBLE__
 
         let -- Andreas, 2010-09-09: comment for existing code
             -- split the telescope into parameters (ptel) and the type or the record
             -- (rt) which should be  R ptel
             telList = telToList tel
             (_ptel,[rt]) = splitAt (size tel - 1) telList
-            cpo    = if hasNamedCon then ConOCon else ConORec
-            cpi    = ConPatternInfo (Just cpo) (Just $ argFromDom $ fmap snd rt)
+            cpo    = if hasNamedCon then PatOCon else PatORec
+            cpi    = ConPatternInfo { conPRecord = Just cpo
+                                    , conPType   = Just $ argFromDom $ fmap snd rt
+                                    , conPLazy   = True }
             conp   = defaultArg $ ConP con cpi $
                      [ Arg ai' $ unnamed $ varP "x" | Dom ai' _ <- telToList ftel ]
             body   = Just $ bodyMod $ var (size ftel2)
@@ -458,8 +461,9 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
                 { funClauses        = [clause]
                 , funCompiled       = Just cc
                 , funProjection     = Just projection
+                , funMutual         = Just []  -- Projections are not mutually recursive with anything
                 , funTerminates     = Just True
-                , funCopatternLHS   = isCopatternLHS [clause]
+                , funCopatternLHS   = hasProjectionPatterns cc
                 })
               { defArgOccurrences = [StrictPos] }
           computePolarity [projname]
@@ -468,6 +472,7 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
 
         recurse
 
+    -- Case: definition.
     checkProjs ftel1 ftel2 (d : fs) = do
       checkDecl d
       checkProjs ftel1 ftel2 fs
