@@ -22,7 +22,6 @@ module Agda.TypeChecking.Serialise
   )
   where
 
-import Control.Applicative
 import Control.Arrow (second)
 import Control.DeepSeq
 import qualified Control.Exception as E
@@ -61,7 +60,7 @@ import Agda.Utils.Except
 -- 32-bit machines). Word64 does not have these problems.
 
 currentInterfaceVersion :: Word64
-currentInterfaceVersion = 20170907 * 10 + 0
+currentInterfaceVersion = 20180518 * 10 + 0
 
 -- | Encodes something. To ensure relocatability file paths in
 -- positions are replaced with module names.
@@ -152,8 +151,6 @@ decode s = do
   -- input is malformed. The decoder is (intended to be) strict enough
   -- to ensure that all such errors can be caught by the handler here.
 
-  shared <- sharedFun
-
   (mf, r) <- liftIO $ E.handle (\(E.ErrorCall s) -> noResult s) $ do
 
     (ver, s, _) <- return $ runGetState B.get s 0
@@ -172,7 +169,6 @@ decode s = do
         st <- St (ar nL) (ar sL) (ar bL) (ar iL) (ar dL)
                 <$> liftIO H.new
                 <*> return mf <*> return incs
-                <*> return shared
         (r, st) <- runStateT (runExceptT (value r)) st
         return (Just (modFile st), r)
 
