@@ -4,13 +4,8 @@
 {-# LANGUAGE TypeOperators        #-}
 {-# LANGUAGE UndecidableInstances #-}
 
-#if __GLASGOW_HASKELL__ <= 708
-{-# OPTIONS_GHC -fcontext-stack=30 #-}
-#endif
-
 module Agda.TypeChecking.Serialise.Base where
 
-import Control.Applicative
 import Control.Monad
 import Control.Monad.Reader
 import Control.Monad.State.Strict (StateT, gets)
@@ -156,7 +151,6 @@ data St = St
     -- ^ Maps module names to file names. Constructed by the decoder.
   , includes  :: [AbsolutePath]
     -- ^ The include directories.
-  , mkShared  :: Term -> Term
   }
 
 -- | Monad used by the encoder.
