@@ -4,7 +4,6 @@
 
 module Agda.TypeChecking.Serialise.Instances.Abstract where
 
-import Control.Applicative
 import qualified Data.Map as Map
 import qualified Data.Set as Set
 
@@ -25,6 +24,10 @@ import Agda.Utils.Except
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+instance EmbPrj A.BindName where
+  icod_ (A.BindName n) = icodeN' A.BindName n
+  value = valueN A.BindName
+
 instance EmbPrj Scope where
   icod_ (Scope a b c d e) = icodeN' Scope a b c d e
 
@@ -99,14 +102,26 @@ instance EmbPrj KindOfName where
     valu [5] = valuN MacroName
     valu _   = malformed
 
+instance EmbPrj Binder where
+  icod_ LambdaBound   = icodeN' LambdaBound
+  icod_ PatternBound  = icodeN 1 PatternBound
+  icod_ LetBound      = icodeN 2 LetBound
+
+  value = vcase valu where
+    valu []  = valuN LambdaBound
+    valu [1] = valuN PatternBound
+    valu [2] = valuN LetBound
+    valu _   = malformed
+
 instance EmbPrj LocalVar where
   icod_ (LocalVar a b c)  = icodeN' LocalVar a b c
 
   value = valueN LocalVar
 
 instance EmbPrj ConPatInfo where
-  icod_ (ConPatInfo a _) = icod_ a
-  value a                = flip ConPatInfo patNoRange <$> value a
+  icod_ (ConPatInfo a _ b) = icodeN' (\a b -> ConPatInfo a patNoRange b) a b
+
+  value = valueN $ \a b -> ConPatInfo a patNoRange b
 
 -- Only for pattern synonyms (where a is Void)
 instance EmbPrj a => EmbPrj (A.Pattern' a) where
@@ -115,12 +130,13 @@ instance EmbPrj a => EmbPrj (A.Pattern' a) where
   icod_ (A.DefP p a b)        = icodeN 2 (A.DefP p) a b
   icod_ t@(A.WildP p)         = icodeN 3 t
   icod_ (A.AsP p a b)         = icodeN 4 (A.AsP p) a b
-  icod_ (A.DotP p a b)        = icodeN 5 (A.DotP p) a b
+  icod_ (A.DotP p a)          = icodeN 5 (A.DotP p) a
   icod_ t@(A.AbsurdP _)       = icodeN 6 t
   icod_ (A.LitP a)            = icodeN 7 A.LitP a
   icod_ (A.ProjP p a b)       = icodeN 8 (A.ProjP p) a b
   icod_ (A.PatternSynP p a b) = icodeN 9 (A.PatternSynP p) a b
   icod_ (A.RecP p a)          = icodeN 10 (A.RecP p) a
+  icod_ (A.WithP i a)         = icodeN 11 (A.WithP i) a
 
   value = vcase valu where
     valu [0, a]       = valuN A.VarP a
@@ -128,42 +144,51 @@ instance EmbPrj a => EmbPrj (A.Pattern' a) where
     valu [2, a, b]    = valuN (A.DefP i) a b
     valu [3]          = valuN (A.WildP i)
     valu [4, a, b]    = valuN (A.AsP i) a b
-    valu [5, a, b]    = valuN (A.DotP i) a b
+    valu [5, a]       = valuN (A.DotP i) a
     valu [6]          = valuN (A.AbsurdP i)
     valu [7, a]       = valuN (A.LitP) a
     valu [8, a, b]    = valuN (A.ProjP i) a b
     valu [9, a, b]    = valuN (A.PatternSynP i) a b
     valu [10, a]      = valuN (A.RecP i) a
+    valu [11, a]      = valuN (A.WithP i) a
     valu _            = malformed
 
     i = patNoRange
 
+instance EmbPrj ParenPreference where
+  icod_ PreferParen     = icodeN' PreferParen
+  icod_ PreferParenless = icodeN 1 PreferParenless
+  value = vcase valu where
+    valu []  = valuN PreferParen
+    valu [1] = valuN PreferParenless
+    valu _   = malformed
+
 instance EmbPrj Precedence where
   icod_ TopCtx                 = icodeN' TopCtx
   icod_ FunctionSpaceDomainCtx = icodeN 1 FunctionSpaceDomainCtx
   icod_ (LeftOperandCtx a)     = icodeN 2 LeftOperandCtx a
-  icod_ (RightOperandCtx a)    = icodeN 3 RightOperandCtx a
+  icod_ (RightOperandCtx a b)  = icodeN 3 RightOperandCtx a b
   icod_ FunctionCtx            = icodeN 4 FunctionCtx
-  icod_ ArgumentCtx            = icodeN 5 ArgumentCtx
+  icod_ (ArgumentCtx a)        = icodeN 5 ArgumentCtx a
   icod_ InsideOperandCtx       = icodeN 6 InsideOperandCtx
   icod_ WithFunCtx             = icodeN 7 WithFunCtx
   icod_ WithArgCtx             = icodeN 8 WithArgCtx
   icod_ DotPatternCtx          = icodeN 9 DotPatternCtx
 
   value = vcase valu where
-    valu []     = valuN TopCtx
-    valu [1]    = valuN FunctionSpaceDomainCtx
-    valu [2, a] = valuN LeftOperandCtx a
-    valu [3, a] = valuN RightOperandCtx a
-    valu [4]    = valuN FunctionCtx
-    valu [5]    = valuN ArgumentCtx
-    valu [6]    = valuN InsideOperandCtx
-    valu [7]    = valuN WithFunCtx
-    valu [8]    = valuN WithArgCtx
-    valu [9]    = valuN DotPatternCtx
-    valu _      = malformed
+    valu []        = valuN TopCtx
+    valu [1]       = valuN FunctionSpaceDomainCtx
+    valu [2, a]    = valuN LeftOperandCtx a
+    valu [3, a, b] = valuN RightOperandCtx a b
+    valu [4]       = valuN FunctionCtx
+    valu [5, a]    = valuN ArgumentCtx a
+    valu [6]       = valuN InsideOperandCtx
+    valu [7]       = valuN WithFunCtx
+    valu [8]       = valuN WithArgCtx
+    valu [9]       = valuN DotPatternCtx
+    valu _         = malformed
 
 instance EmbPrj ScopeInfo where
-  icod_ (ScopeInfo a b c d e f g) = icodeN' (\ a b c d -> ScopeInfo a b c d e f g) a b c d
+  icod_ (ScopeInfo a b c d e f g h) = icodeN' (\ a b c d e -> ScopeInfo a b c d e f g h) a b c d e
 
-  value = valueN (\ a b c d -> ScopeInfo a b c d Map.empty Map.empty Set.empty)
+  value = valueN (\ a b c d e -> ScopeInfo a b c d e Map.empty Map.empty Set.empty)
