@@ -1,18 +1,12 @@
 {-# LANGUAGE CPP                #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 
-#if __GLASGOW_HASKELL__ <= 708
-{-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE OverlappingInstances       #-}
-#endif
-
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 module Agda.TypeChecking.Serialise.Instances.Common (SerialisedRange(..)) where
 
 import Prelude hiding (mapM)
 
-import Control.Applicative
 import Control.Monad.Reader hiding (mapM)
 import Control.Monad.State.Strict (gets, modify)
 import Control.Exception
@@ -28,15 +22,13 @@ import Data.Maybe
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Set (Set)
+import qualified Data.IntSet as IntSet
+import Data.IntSet (IntSet)
 import qualified Data.Set as Set
 import Data.Sequence (Seq)
 import qualified Data.Sequence as Seq
 import Data.Traversable ( mapM )
 
-#if __GLASGOW_HASKELL__ <= 708
-import Data.Typeable ( Typeable )
-#endif
-
 import Data.Void
 
 import Agda.Syntax.Common
@@ -57,8 +49,10 @@ import Agda.Utils.HashMap (HashMap)
 import qualified Agda.Utils.HashMap as HMap
 import Agda.Utils.FileName
 import Agda.Utils.Maybe
+import Agda.Utils.NonemptyList
 import qualified Agda.Utils.Maybe.Strict as Strict
-import Agda.Utils.Trie
+import Agda.Utils.Trie (Trie(..))
+import qualified Agda.Utils.Trie as Trie
 
 import Agda.Utils.Except
 
@@ -68,11 +62,7 @@ import qualified Agda.Utils.Empty as Empty
 #include "undefined.h"
 import Agda.Utils.Impossible
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPING #-} EmbPrj String where
-#else
-instance EmbPrj String where
-#endif
   icod_   = icodeString
   value i = (! i) `fmap` gets stringE
 
@@ -208,11 +198,7 @@ instance EmbPrj TopLevelModuleName where
 
   value = valueN TopLevelModuleName
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPABLE #-} EmbPrj a => EmbPrj [a] where
-#else
-instance EmbPrj a => EmbPrj [a] where
-#endif
   icod_ xs = icodeNode =<< mapM icode xs
   value    = vcase (mapM value)
 --   icode []       = icode0'
@@ -221,6 +207,10 @@ instance EmbPrj a => EmbPrj [a] where
 --                            valu [x, xs] = valu2 (:) x xs
 --                            valu _       = malformed
 
+instance EmbPrj a => EmbPrj (NonemptyList a) where
+  icod_ = icod_ . toList
+  value = listCaseNe malformed return <=< value
+
 instance (Ord a, Ord b, EmbPrj a, EmbPrj b) => EmbPrj (BiMap a b) where
   icod_ m = icode (BiMap.toList m)
   value m = BiMap.fromList <$> value m
@@ -233,6 +223,10 @@ instance (Ord a, EmbPrj a) => EmbPrj (Set a) where
   icod_ s = icode (Set.toList s)
   value s = Set.fromList `fmap` value s
 
+instance EmbPrj IntSet where
+  icod_ s = icode (IntSet.toList s)
+  value s = IntSet.fromList <$> value s
+
 instance (Ord a, EmbPrj a, EmbPrj b) => EmbPrj (Trie a b) where
   icod_ (Trie a b)= icodeN' Trie a b
 
@@ -256,9 +250,6 @@ instance EmbPrj Range where
 -- | Ranges that should be serialised properly.
 
 newtype SerialisedRange = SerialisedRange { underlyingRange :: Range }
-#if __GLASGOW_HASKELL__ <= 708
-  deriving (Typeable)
-#endif
 
 instance EmbPrj SerialisedRange where
   icod_ (SerialisedRange r) =
@@ -378,7 +369,7 @@ instance EmbPrj a => EmbPrj (Ranged a) where
   value = valueN Ranged
 
 instance EmbPrj ArgInfo where
-  icod_ (ArgInfo h r o) = icodeN' ArgInfo h r o
+  icod_ (ArgInfo h r o fv) = icodeN' ArgInfo h r o fv
 
   value = valueN ArgInfo
 
@@ -406,6 +397,15 @@ instance EmbPrj a => EmbPrj (Dom a) where
 
   value = valueN Dom
 
+instance EmbPrj HasEta where
+  icod_ YesEta = icodeN' YesEta
+  icod_ NoEta  = icodeN 1 NoEta
+
+  value = vcase valu where
+    valu []  = valuN YesEta
+    valu [1] = valuN NoEta
+    valu _   = malformed
+
 instance EmbPrj Induction where
   icod_ Inductive   = icodeN' Inductive
   icod_ CoInductive = icodeN 1 CoInductive
@@ -427,18 +427,29 @@ instance EmbPrj Hiding where
   value 3 = return (Instance YesOverlap)
   value _ = malformed
 
+instance EmbPrj Quantity where
+  icod_ Quantity0 = return 0
+  icod_ Quantityω = return 1
+
+  value 0 = return Quantity0
+  value 1 = return Quantityω
+  value _ = malformed
+
+instance EmbPrj Modality where
+  icod_ (Modality a b) = icodeN' Modality a b
+
+  value = vcase $ \case
+    [a, b] -> valuN Modality a b
+    _ -> malformed
+
 instance EmbPrj Relevance where
   icod_ Relevant       = return 0
   icod_ Irrelevant     = return 1
-  icod_ (Forced Small) = return 2
-  icod_ (Forced Big)   = return 3
-  icod_ NonStrict      = return 4
+  icod_ NonStrict      = return 2
 
   value 0 = return Relevant
   value 1 = return Irrelevant
-  value 2 = return (Forced Small)
-  value 3 = return (Forced Big)
-  value 4 = return NonStrict
+  value 2 = return NonStrict
   value _ = malformed
 
 instance EmbPrj Origin where
@@ -453,6 +464,15 @@ instance EmbPrj Origin where
   value 3 = return CaseSplit
   value _ = malformed
 
+instance EmbPrj FreeVariables where
+  icod_ UnknownFVs   = icodeN' UnknownFVs
+  icod_ (KnownFVs a) = icodeN' KnownFVs a
+
+  value = vcase valu where
+    valu []  = valuN UnknownFVs
+    valu [a] = valuN KnownFVs a
+    valu _   = malformed
+
 instance EmbPrj ConOrigin where
   icod_ ConOSystem = return 0
   icod_ ConOCon    = return 1
@@ -482,6 +502,7 @@ instance EmbPrj Agda.Syntax.Literal.Literal where
   icod_ (LitChar   a b)   = icodeN 3 LitChar a b
   icod_ (LitQName  a b)   = icodeN 5 LitQName a b
   icod_ (LitMeta   a b c) = icodeN 6 LitMeta a b c
+  icod_ (LitWord64 a b)   = icodeN 7 LitWord64 a b
 
   value = vcase valu where
     valu [a, b]       = valuN LitNat    a b
@@ -490,6 +511,7 @@ instance EmbPrj Agda.Syntax.Literal.Literal where
     valu [3, a, b]    = valuN LitChar   a b
     valu [5, a, b]    = valuN LitQName  a b
     valu [6, a, b, c] = valuN LitMeta   a b c
+    valu [7, a, b]    = valuN LitWord64 a b
     valu _            = malformed
 
 instance EmbPrj IsAbstract where
