@@ -4,7 +4,6 @@
 
 module Agda.TypeChecking.Serialise.Instances.Internal where
 
-import Control.Applicative
 import Control.Monad.State.Strict
 
 import Agda.Syntax.Internal as I
@@ -92,10 +91,8 @@ instance EmbPrj I.Term where
   icod_ (MetaV    a b) = __IMPOSSIBLE__
   icod_ (DontCare a  ) = icodeN 8 DontCare a
   icod_ (Level    a  ) = icodeN 9 Level a
-  icod_ (Shared p)     = icodeMemo termD termC p $ icode (derefPtr p)
 
-  value r = vcase valu' r where
-    valu' xs       = gets mkShared <*> valu xs
+  value = vcase valu where
     valu [a]       = valuN var   a
     valu [0, a, b] = valuN Var   a b
     valu [1, a, b] = valuN Lam   a b
@@ -140,14 +137,17 @@ instance EmbPrj I.Sort where
   icod_ Prop        = icodeN' Prop
   icod_ SizeUniv    = icodeN 1 SizeUniv
   icod_ Inf         = icodeN 2 Inf
-  icod_ (DLub a b)  = icodeN 3 DLub a b -- Andreas, 2017-01-18: not __IMPOSSIBLE__ see #2408
+  icod_ (PiSort a b) = icodeN 3 PiSort a b
+  icod_ (UnivSort a) = icodeN 4 UnivSort a
+  icod_ (MetaS a b)  = __IMPOSSIBLE__
 
   value = vcase valu where
     valu []        = valuN Prop
     valu [0, a]    = valuN Type  a
     valu [1]       = valuN SizeUniv
     valu [2]       = valuN Inf
-    valu [3, a, b] = valuN DLub a b
+    valu [3, a, b] = valuN PiSort a b
+    valu [4, a]    = valuN UnivSort a
     valu _         = malformed
 
 instance EmbPrj DisplayForm where
@@ -160,18 +160,9 @@ instance EmbPrj a => EmbPrj (Open a) where
 
   value = valueN OpenThing
 
-instance EmbPrj a => EmbPrj (Local a) where
-  icod_ (Local a b) = icodeN' Local a b
-  icod_ (Global a)  = icodeN' Global a
-
-  value = vcase valu where
-    valu [a, b] = valuN Local a b
-    valu [a]    = valuN Global a
-    valu _      = malformed
-
-instance EmbPrj CtxId where
-  icod_ (CtxId a) = icode a
-  value n         = CtxId `fmap` value n
+instance EmbPrj CheckpointId where
+  icod_ (CheckpointId a) = icode a
+  value n                = CheckpointId `fmap` value n
 
 instance EmbPrj DisplayTerm where
   icod_ (DTerm    a  )   = icodeN' DTerm a
@@ -238,7 +229,7 @@ instance EmbPrj ProjLams where
   value = valueN ProjLams
 
 instance EmbPrj ExtLamInfo where
-  icod_ (ExtLamInfo a b) = icodeN' ExtLamInfo a b
+  icod_ (ExtLamInfo a) = icodeN' ExtLamInfo a
 
   value = valueN ExtLamInfo
 
@@ -254,6 +245,14 @@ instance EmbPrj Polarity where
   value 3 = return Nonvariant
   value _ = malformed
 
+instance EmbPrj IsForced where
+  icod_ Forced    = return 0
+  icod_ NotForced = return 1
+
+  value 0 = return Forced
+  value 1 = return NotForced
+  value _ = malformed
+
 instance EmbPrj Occurrence where
   icod_ StrictPos = return 0
   icod_ Mixed     = return 1
@@ -283,19 +282,19 @@ instance EmbPrj Defn where
   icod_ Axiom                                   = icodeN 0 Axiom
   icod_ (Function    a b t c d e f g h i j k m) =
     icodeN 1 (\ a b -> Function a b t) a b c d e f g h i j k m
-  icod_ (Datatype    a b c d e f g h i j)       = icodeN 2 Datatype a b c d e f g h i j
+  icod_ (Datatype    a b c d e f g h)           = icodeN 2 Datatype a b c d e f g h
   icod_ (Record      a b c d e f g h i j)       = icodeN 3 Record a b c d e f g h i j
-  icod_ (Constructor a b c d e f g)             = icodeN 4 Constructor a b c d e f g
-  icod_ (Primitive   a b c d)                   = icodeN 5 Primitive a b c d
+  icod_ (Constructor a b c d e f g h)           = icodeN 4 Constructor a b c d e f g h
+  icod_ (Primitive   a b c d e)                 = icodeN 5 Primitive a b c d e
   icod_ AbstractDefn{}                          = __IMPOSSIBLE__
 
   value = vcase valu where
     valu [0]                                     = valuN Axiom
     valu [1, a, b, c, d, e, f, g, h, i, j, k, m] = valuN (\ a b -> Function a b Nothing) a b c d e f g h i j k m
-    valu [2, a, b, c, d, e, f, g, h, i, j]       = valuN Datatype a b c d e f g h i j
+    valu [2, a, b, c, d, e, f, g, h]             = valuN Datatype a b c d e f g h
     valu [3, a, b, c, d, e, f, g, h, i, j]       = valuN Record  a b c d e f g h i j
-    valu [4, a, b, c, d, e, f, g]                = valuN Constructor a b c d e f g
-    valu [5, a, b, c, d]                         = valuN Primitive   a b c d
+    valu [4, a, b, c, d, e, f, g, h]             = valuN Constructor a b c d e f g h
+    valu [5, a, b, c, d, e]                      = valuN Primitive   a b c d e
     valu _                                       = malformed
 
 instance EmbPrj FunctionFlag where
@@ -315,7 +314,7 @@ instance EmbPrj a => EmbPrj (WithArity a) where
   value = valueN WithArity
 
 instance EmbPrj a => EmbPrj (Case a) where
-  icod_ (Branches a b c d) = icodeN' Branches a b c d
+  icod_ (Branches a b c d e f) = icodeN' Branches a b c d e f
 
   value = valueN Branches
 
@@ -343,11 +342,15 @@ instance EmbPrj TermHead where
   icod_ SortHead     = icodeN' SortHead
   icod_ PiHead       = icodeN 1 PiHead
   icod_ (ConsHead a) = icodeN 2 ConsHead a
+  icod_ (VarHead a)  = icodeN 3 VarHead a
+  icod_ UnknownHead  = icodeN 4 UnknownHead
 
   value = vcase valu where
     valu []     = valuN SortHead
     valu [1]    = valuN PiHead
     valu [2, a] = valuN ConsHead a
+    valu [3, a] = valuN VarHead a
+    valu [4]    = valuN UnknownHead
     valu _      = malformed
 
 instance EmbPrj I.Clause where
@@ -356,7 +359,7 @@ instance EmbPrj I.Clause where
   value = valueN Clause
 
 instance EmbPrj I.ConPatternInfo where
-  icod_ (ConPatternInfo a b) = icodeN' ConPatternInfo a b
+  icod_ (ConPatternInfo a b c) = icodeN' ConPatternInfo a b c
 
   value = valueN ConPatternInfo
 
@@ -365,21 +368,42 @@ instance EmbPrj I.DBPatVar where
 
   value = valueN DBPatVar
 
-instance EmbPrj a => EmbPrj (I.Pattern' a) where
-  icod_ (VarP a    ) = icodeN' VarP a
-  icod_ (ConP a b c) = icodeN 1 ConP a b c
-  icod_ (LitP a    ) = icodeN 2 LitP a
-  icod_ (DotP a    ) = icodeN 3 DotP a
-  icod_ (ProjP a b ) = icodeN 4 ProjP a b
-  icod_ (AbsurdP a ) = icodeN 5 AbsurdP a
+instance EmbPrj I.PatOrigin where
+  icod_ PatOSystem  = icodeN' PatOSystem
+  icod_ PatOSplit   = icodeN 1 PatOSplit
+  icod_ (PatOVar a) = icodeN 2 PatOVar a
+  icod_ PatODot     = icodeN 3 PatODot
+  icod_ PatOWild    = icodeN 4 PatOWild
+  icod_ PatOCon     = icodeN 5 PatOCon
+  icod_ PatORec     = icodeN 6 PatORec
+  icod_ PatOLit     = icodeN 7 PatOLit
+  icod_ PatOAbsurd  = icodeN 8 PatOAbsurd
 
   value = vcase valu where
-    valu [a]       = valuN VarP a
+    valu []     = valuN PatOSystem
+    valu [1]    = valuN PatOSplit
+    valu [2, a] = valuN PatOVar a
+    valu [3]    = valuN PatODot
+    valu [4]    = valuN PatOWild
+    valu [5]    = valuN PatOCon
+    valu [6]    = valuN PatORec
+    valu [7]    = valuN PatOLit
+    valu [8]    = valuN PatOAbsurd
+    valu _      = malformed
+
+instance EmbPrj a => EmbPrj (I.Pattern' a) where
+  icod_ (VarP a b  ) = icodeN 0 VarP a b
+  icod_ (ConP a b c) = icodeN 1 ConP a b c
+  icod_ (LitP a    ) = icodeN 2 LitP a
+  icod_ (DotP a b  ) = icodeN 3 DotP a b
+  icod_ (ProjP a b ) = icodeN 4 ProjP a b
+
+  value = vcase valu where
+    valu [0, a, b] = valuN VarP a b
     valu [1, a, b, c] = valuN ConP a b c
     valu [2, a]    = valuN LitP a
-    valu [3, a]    = valuN DotP a
+    valu [3, a, b] = valuN DotP a b
     valu [4, a, b] = valuN ProjP a b
-    valu [5, a]    = valuN AbsurdP a
     valu _         = malformed
 
 instance EmbPrj a => EmbPrj (Builtin a) where
