@@ -42,7 +42,7 @@ import Agda.Utils.Impossible
 
 -- | Check whether a type is either not a SIZELT or a SIZELT that is non-empty.
 checkSizeLtSat :: Term -> TCM ()
-checkSizeLtSat t = do
+checkSizeLtSat t = whenM haveSizeLt $ do
   reportSDoc "tc.size" 10 $ do
     tel <- getContextTelescope
     sep
@@ -58,7 +58,7 @@ checkSizeLtSat t = do
         addConstraint $ CheckSizeLtSat t
   let ok :: TCM ()
       ok = reportSLn "tc.size.lt" 20 $ "- succeeded: not an empty type of sizes"
-  ifBlocked t (const postpone) $ \ t -> do
+  ifBlocked t (const postpone) $ \ _ t -> do
     reportSLn "tc.size.lt" 20 $ "- type is not blocked"
     caseMaybeM (isSizeType t) ok $ \ b -> do
       reportSLn "tc.size.lt" 20 $ " - type is a size type"
@@ -66,7 +66,7 @@ checkSizeLtSat t = do
         BoundedNo -> ok
         BoundedLt b -> do
           reportSDoc "tc.size.lt" 20 $ text " - type is SIZELT" <+> prettyTCM b
-          ifBlocked b (\ _ _ -> postpone t) $ \ b -> do
+          ifBlocked b (\ _ _ -> postpone t) $ \ _ b -> do
             reportSLn "tc.size.lt" 20 $ " - size bound is not blocked"
             catchConstraint (CheckSizeLtSat t) $ do
               unlessM (checkSizeNeverZero b) $ do
@@ -82,7 +82,7 @@ checkSizeNeverZero u = do
     SizeInf     -> return True  -- OK, infty is never 0.
     SizeSuc{}   -> return True  -- OK, a + 1 is never 0.
     OtherSize u ->
-      case ignoreSharing u of
+      case u of
         Var i [] -> checkSizeVarNeverZero i
         -- neutral sizes cannot be guaranteed > 0
         _ -> return False
@@ -115,7 +115,7 @@ checkSizeNeverZero u = do
 --         case b of
 --           BoundedNo -> no
 --           BoundedLt u -> ifBlocked u (\ _ _ -> perhaps) $ \ u -> do
---             case ignoreSharing u of
+--             case u of
 --                Var i' [] | i == i' -> yes
 --                _ -> no
 
@@ -151,11 +151,11 @@ checkSizeVarNeverZero i = do
         perhaps = tell (Any True) >> cont
     -- If we encounter a blocked type in the context, we cannot
     -- give a definite answer.
-    ifBlockedType t (\ _ _ -> perhaps) $ \ t -> do
+    ifBlockedType t (\ _ _ -> perhaps) $ \ _ t -> do
       caseMaybeM (liftTCM $ isSizeType t) cont $ \ b -> do
         case b of
           BoundedNo -> cont
-          BoundedLt u -> ifBlocked u (\ _ _ -> perhaps) $ \ u -> do
+          BoundedLt u -> ifBlocked u (\ _ _ -> perhaps) $ \ _ u -> do
             reportSLn "tc.size" 60 $ "minSizeVal upper bound u = " ++ show u
             v <- liftTCM $ deepSizeView u
             case v of
@@ -175,7 +175,7 @@ checkSizeVarNeverZero i = do
 isBounded :: MonadTCM tcm => Nat -> tcm BoundedSize
 isBounded i = liftTCM $ do
   t <- reduce =<< typeOfBV i
-  case ignoreSharing $ unEl t of
+  case unEl t of
     Def x [Apply u] -> do
       sizelt <- getBuiltin' builtinSizeLt
       return $ if (Just (Def x []) == sizelt) then BoundedLt $ unArg u else BoundedNo
@@ -192,7 +192,7 @@ boundedSizeMetaHook v tel0 a = do
       n <- getContextSize
       let tel | n > 0     = telFromList $ drop n $ telToList tel0
               | otherwise = tel0
-      addContext' tel $ do
+      addContext tel $ do
         v <- sizeSuc 1 $ raise (size tel) v `apply` teleArgs tel
         -- compareSizes CmpLeq v u
         size <- sizeType
@@ -215,8 +215,8 @@ trySizeUniv cmp t m n x els1 y els2 = do
       forceInfty u = compareSizes CmpEq (unArg u) =<< primSizeInf
   -- Get the SIZE built-ins.
   (size, sizelt) <- flip catchError (const failure) $ do
-     Def size   _ <- ignoreSharing <$> primSize
-     Def sizelt _ <- ignoreSharing <$> primSizeLt
+     Def size   _ <- primSize
+     Def sizelt _ <- primSizeLt
      return (size, sizelt)
   case (cmp, els1, els2) of
      -- Case @Size< _ <= Size@: true.
@@ -236,11 +236,11 @@ trySizeUniv cmp t m n x els1 y els2 = do
 --   Precondition: sized types are enabled.
 deepSizeView :: Term -> TCM DeepSizeView
 deepSizeView v = do
-  Def inf [] <- ignoreSharing <$> primSizeInf
-  Def suc [] <- ignoreSharing <$> primSizeSuc
+  Def inf [] <- primSizeInf
+  Def suc [] <- primSizeSuc
   let loop v = do
         v <- reduce v
-        case ignoreSharing v of
+        case v of
           Def x []        | x == inf -> return $ DSizeInf
           Def x [Apply u] | x == suc -> sizeViewSuc_ suc <$> loop (unArg u)
           Var i []                   -> return $ DSizeVar i 0
@@ -255,7 +255,7 @@ sizeMaxView v = do
   max <- getBuiltinDefName builtinSizeMax
   let loop v = do
         v <- reduce v
-        case ignoreSharing v of
+        case v of
           Def x []                   | Just x == inf -> return $ [DSizeInf]
           Def x [Apply u]            | Just x == suc -> maxViewSuc_ (fromJust suc) <$> loop (unArg u)
           Def x [Apply u1, Apply u2] | Just x == max -> maxViewMax <$> loop (unArg u1) <*> loop (unArg u2)
@@ -416,6 +416,7 @@ getSizeMetas interactionMetas = do
         let no = return Nothing
         mi <- lookupMeta m
         case mvJudgement mi of
+          _ | BlockedConst{} <- mvInstantiation mi -> no  -- Blocked terms should not be touched (#2637, #2881)
           HasType _ a -> do
             TelV tel b <- telView a
             -- b is reduced
@@ -530,14 +531,14 @@ oldSizeExpr u = do
   case s of
     SizeInf     -> patternViolation
     SizeSuc u   -> mapSnd (+1) <$> oldSizeExpr u
-    OtherSize u -> case ignoreSharing u of
+    OtherSize u -> case u of
       Var i []  -> return (Rigid i, 0)
       MetaV m es | Just xs <- mapM isVar es, fastDistinct xs
                 -> return (SizeMeta m xs, 0)
       _ -> patternViolation
   where
     isVar (Proj{})  = Nothing
-    isVar (Apply v) = case ignoreSharing $ unArg v of
+    isVar (Apply v) = case unArg v of
       Var i [] -> Just i
       _        -> Nothing
 
