@@ -93,11 +93,7 @@ import Agda.Utils.Function
 import Agda.Utils.Functor
 import Agda.Utils.Lens
 
-#if MIN_VERSION_base(4,8,0)
 import qualified Agda.Utils.List as List hiding ( uncons )
-#else
-import qualified Agda.Utils.List as List
-#endif
 
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
@@ -323,34 +319,33 @@ castConstraintToCurrentContext' cl = do
 
 castConstraintToCurrentContext :: Closure TCM.Constraint -> MaybeT TCM TCM.Constraint
 castConstraintToCurrentContext cl = do
-  -- The target context
-  gamma <- asks envContext
-  -- The context where the constraint lives.
-  let delta = envContext $ clEnv cl
-  -- The constraint
-  let c = clValue cl
-  let findInGamma (Ctx cid (Dom _ (x, t))) =
-        -- try to find same CtxId (safe)
-        case List.findIndex ((cid ==) . ctxId) gamma of
-          Just i -> Just i
-          Nothing ->
-            -- match by name (hazardous)
-            -- This is one of the seven deadly sins (not respecting alpha).
-            List.findIndex ((x ==) . fst . unDom . ctxEntry) gamma
-  let cand = map findInGamma delta
-  -- The domain of our substitution
-  let coveredVars = VarSet.fromList $ catMaybes $ zipWith ($>) cand [0..]
-  -- Check that all the free variables of the constraint are contained in
-  -- coveredVars.
-  -- We ignore the free variables occurring in sorts.
-  guard $ getAll $ runFree (All . (`VarSet.member` coveredVars)) IgnoreAll c
-  -- Turn cand into a substitution.
-  -- Since we ignored the free variables in sorts, we better patch up
-  -- the substitution with some dummy term (Sort Prop) rather than __IMPOSSIBLE__.
-  let dummy = Sort Prop
-  let sigma = parallelS $ map (maybe dummy var) cand
+  -- The checkpoint of the contraint
+  let cp = envCurrentCheckpoint $ clEnv cl
+  sigma <- caseMaybeM (view $ eCheckpoints . key cp)
+          (do
+            -- We are not in a descendant of the constraint checkpoint.
+            -- Here be dragons!!
+            gamma <- asks envContext -- The target context
+            let findInGamma (Dom _ (x, t)) =
+                  -- match by name (hazardous)
+                  -- This is one of the seven deadly sins (not respecting alpha).
+                  List.findIndex ((x ==) . fst . unDom) gamma
+            let delta = envContext $ clEnv cl
+                cand  = map findInGamma delta
+            -- The domain of our substitution
+            let coveredVars = VarSet.fromList $ catMaybes $ zipWith ($>) cand [0..]
+            -- Check that all the free variables of the constraint are contained in
+            -- coveredVars.
+            -- We ignore the free variables occurring in sorts.
+            guard $ getAll $ runFree (All . (`VarSet.member` coveredVars)) IgnoreAll (clValue cl)
+            -- Turn cand into a substitution.
+            -- Since we ignored the free variables in sorts, we better patch up
+            -- the substitution with some dummy term (Sort Prop) rather than __IMPOSSIBLE__.
+            let dummy = Sort Prop
+            return $ parallelS $ map (maybe dummy var) cand
+          ) return -- Phew, we've got the checkpoint! All is well.
   -- Apply substitution to constraint and pray that the Gods are merciful on us.
-  return $ applySubst sigma c
+  return $ applySubst sigma (clValue cl)
   -- Note: the resulting constraint may not well-typed.
   -- Even if it is, it may map variables to their wrong counterpart.
 
@@ -498,7 +493,7 @@ solveCluster flag ccs = do
     iterateSolver Map.empty hg csF emptySolution
 
   -- Convert solution to meta instantiation.
-  forM_ (Map.assocs $ theSolution sol) $ \ (m, a) -> do
+  solved <- fmap Set.unions $ forM (Map.assocs $ theSolution sol) $ \ (m, a) -> do
     unless (validOffset a) __IMPOSSIBLE__
     -- Solution does not contain metas
     u <- unSizeExpr $ fmap __IMPOSSIBLE__ a
@@ -518,8 +513,9 @@ solveCluster flag ccs = do
       [ text $ "  xs = " ++ show xs
       , text $ "  u  = " ++ show u
       ]
-    unlessM (isFrozen x) $
+    ifM (isFrozen x) (return Set.empty) $ do
       assignMeta n x t xs u
+      return $ Set.singleton x
     -- WRONG:
     -- let partialSubst = List.sort $ zip xs $ map var $ downFrom n
     -- assignMeta' n x t (length xs) partialSubst u
@@ -531,8 +527,7 @@ solveCluster flag ccs = do
   ims <- Set.fromList <$> getInteractionMetas
 
   --  ms = unsolved size metas from cluster
-  let ms = Set.fromList (map sizeMetaId metas) Set.\\  -- Some CPP or ghc does not like trailing backslash, thus, this comment!
-             Set.mapMonotonic MetaId (Map.keysSet $ theSolution sol)
+  let ms = Set.fromList (map sizeMetaId metas) Set.\\ solved
   --  Make sure they do not contain an interaction point
   let noIP = Set.null $ Set.intersection ims ms
 
@@ -576,7 +571,7 @@ solveCluster flag ccs = do
         forM_ cs0 $ \ cl -> enterClosure cl solveConstraint
 
 -- | Collect constraints from a typing context, looking for SIZELT hypotheses.
-getSizeHypotheses :: Context -> TCM [(CtxId, SizeConstraint)]
+getSizeHypotheses :: Context -> TCM [(Nat, SizeConstraint)]
 getSizeHypotheses gamma = inTopContext $ modifyContext (const gamma) $ do
   (_, msizelt) <- getBuiltinSize
   caseMaybe msizelt (return []) $ \ sizelt -> do
@@ -584,13 +579,13 @@ getSizeHypotheses gamma = inTopContext $ modifyContext (const gamma) $ do
     catMaybes <$> do
       forM (zip [0..] gamma) $ \ (i, ce) -> do
         -- Get name and type of variable i.
-        let xt = unDom $ ctxEntry ce
-            x  = prettyShow $ fst xt
-        t <- reduce . raise (1 + i) . unEl . snd $ xt
-        case ignoreSharing t of
+        let (x, t) = unDom ce
+            s      = prettyShow x
+        t <- reduce . raise (1 + i) . unEl $ t
+        case t of
           Def d [Apply u] | d == sizelt -> do
             caseMaybeM (sizeExpr $ unArg u) (return Nothing) $ \ a ->
-              return $ Just $ (ctxId ce, Constraint (Rigid (NamedRigid x i) 0) Lt a)
+              return $ Just $ (i, Constraint (Rigid (NamedRigid s i) 0) Lt a)
           _ -> return Nothing
 
 -- | Convert size constraint into form where each meta is applied
@@ -729,7 +724,7 @@ instance PrettyTCM (SizeConstraint) where
 -- | Size constraint with de Bruijn indices.
 data HypSizeConstraint = HypSizeConstraint
   { sizeContext    :: Context
-  , sizeHypIds     :: [CtxId]
+  , sizeHypIds     :: [Nat] -- ^ DeBruijn indices
   , sizeHypotheses :: [SizeConstraint]  -- ^ Living in @Context@.
   , sizeConstraint :: SizeConstraint    -- ^ Living in @Context@.
   }
@@ -740,7 +735,7 @@ instance Flexs SizeMeta HypSizeConstraint where
 instance PrettyTCM HypSizeConstraint where
   prettyTCM (HypSizeConstraint cxt _ hs c) =
     inTopContext $ modifyContext (const cxt) $ do
-      let cxtNames = reverse $ map (fst . unDom . ctxEntry) cxt
+      let cxtNames = reverse $ map (fst . unDom) cxt
       -- text ("[#cxt=" ++ show (size cxt) ++ "]") <+> do
       prettyList (map prettyTCM cxtNames) <+> do
       applyUnless (null hs)
@@ -780,7 +775,7 @@ sizeExpr u = do
   case s of
     SizeInf     -> return $ Just Infty
     SizeSuc u   -> fmap (`plus` (1 :: Offset)) <$> sizeExpr u
-    OtherSize u -> case ignoreSharing u of
+    OtherSize u -> case u of
       Var i []    -> (\ x -> Just $ Rigid (NamedRigid x i) 0) . prettyShow <$> nameOfBV i
 --      MetaV m es  -> return $ Just $ Flex (SizeMeta m es) 0
       MetaV m es | Just xs <- mapM isVar es, List.fastDistinct xs
@@ -788,7 +783,7 @@ sizeExpr u = do
       _           -> return Nothing
   where
     isVar (Proj{})  = Nothing
-    isVar (Apply v) = case ignoreSharing $ unArg v of
+    isVar (Apply v) = case unArg v of
       Var i [] -> Just i
       _        -> Nothing
 
