@@ -2,10 +2,6 @@
 {-# LANGUAGE DeriveDataTypeable   #-}
 {-# LANGUAGE UndecidableInstances #-}
 
-#if __GLASGOW_HASKELL__ <= 708
-{-# LANGUAGE OverlappingInstances #-}
-#endif
-
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 -- | This module contains the definition of hereditary substitution
@@ -22,21 +18,22 @@ module Agda.TypeChecking.Substitute
   , Substitution'(..), Substitution
   ) where
 
-import Control.Applicative
+import Control.Arrow (first, second)
 import Data.Function
 import Data.Functor
 import qualified Data.List as List
 import Data.Map (Map)
 import Data.Maybe
 import Data.Monoid
-import Data.Typeable (Typeable)
 
 import Debug.Trace (trace)
+import Language.Haskell.TH.Syntax (thenCmp) -- lexicographic combination of Ordering
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
 import qualified Agda.Syntax.Abstract as A
+import Agda.Syntax.Position (Range)
 
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Free as Free
@@ -69,7 +66,6 @@ instance Apply Term where
           Apply a : es0 -> lazyAbsApp b (unArg a) `applyE` es0
           _             -> __IMPOSSIBLE__
       MetaV x es' -> MetaV x (es' ++ es)
-      Shared p    -> Shared $ applyE p es
       Lit{}       -> __IMPOSSIBLE__
       Level{}     -> __IMPOSSIBLE__
       Pi _ _      -> __IMPOSSIBLE__
@@ -81,23 +77,24 @@ instance Apply Term where
 --   returns its field @f@.
 canProject :: QName -> Term -> Maybe (Arg Term)
 canProject f v =
-  case ignoreSharing v of
+  case v of
     (Con (ConHead _ _ fs) _ vs) -> do
       i <- List.elemIndex f fs
-      headMaybe (drop i vs)
+      isApplyElim =<< headMaybe (drop i vs)
     _ -> Nothing
 
 -- | Eliminate a constructed term.
-conApp :: ConHead -> ConInfo -> Args -> Elims -> Term
+conApp :: ConHead -> ConInfo -> Elims -> Elims -> Term
 conApp ch                  ci args []             = Con ch ci args
-conApp ch                  ci args (Apply a : es) = conApp ch ci (args ++ [a]) es
+conApp ch                  ci args (a@Apply{} : es) = conApp ch ci (args ++ [a]) es
 conApp ch@(ConHead c _ fs) ci args (Proj o f : es) =
   let failure = flip trace __IMPOSSIBLE__ $
         "conApp: constructor " ++ show c ++
         " with fields " ++ show fs ++
         " projected by " ++ show f
+      isApply e = fromMaybe __IMPOSSIBLE__ $ isApplyElim e
       i = maybe failure id            $ List.elemIndex f fs
-      v = maybe failure argToDontCare $ headMaybe $ drop i args
+      v = maybe failure (argToDontCare . isApply)  $ headMaybe $ drop i args
   in  applyE v es
 
   -- -- Andreas, 2016-07-20 futile attempt to magically fix ProjOrigin
@@ -172,9 +169,6 @@ instance Apply Sort where
   applyE s [] = s
   applyE s _  = __IMPOSSIBLE__
 
-instance Apply a => Apply (Ptr a) where
-  applyE p xs = fmap (`applyE` xs) p
-
 -- @applyE@ does not make sense for telecopes, definitions, clauses etc.
 
 instance Subst Term a => Apply (Tele a) where
@@ -200,26 +194,14 @@ instance Apply RewriteRule where
        , rewType    = applyNLPatSubst sub (rewType r)
        }
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPING #-} Apply [Occ.Occurrence] where
-#else
-instance Apply [Occ.Occurrence] where
-#endif
   apply occ args = List.drop (length args) occ
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPING #-} Apply [Polarity] where
-#else
-instance Apply [Polarity] where
-#endif
   apply pol args = List.drop (length args) pol
 
 -- | Make sure we only drop variable patterns.
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPING #-} Apply [NamedArg (Pattern' a)] where
-#else
-instance Apply [NamedArg (Pattern' a)] where
-#endif
   apply ps args = loop (length args) ps
     where
     loop 0 ps = ps
@@ -229,7 +211,6 @@ instance Apply [NamedArg (Pattern' a)] where
       in  case namedArg p of
             VarP{}  -> recurse
             DotP{}  -> __IMPOSSIBLE__
-            AbsurdP{} -> __IMPOSSIBLE__
             LitP{}  -> __IMPOSSIBLE__
             ConP{}  -> __IMPOSSIBLE__
             ProjP{} -> __IMPOSSIBLE__
@@ -272,7 +253,7 @@ instance Apply Defn where
               where
                 larg  = last args -- the record value
                 args' = [larg]
-                isVar0 = case ignoreSharing $ unArg larg of Var 0 [] -> True; _ -> False
+                isVar0 = case unArg larg of Var 0 [] -> True; _ -> False
 {-
     Function{ funClauses = cs, funCompiled = cc, funInv = inv
             , funProjection = Just p@Projection{ projIndex = n } }
@@ -287,13 +268,9 @@ instance Apply Defn where
           }
       where args' = [last args]  -- the record value
 -}
-    Datatype{ dataPars = np, dataSmallPars = sps, dataNonLinPars = nlps, dataClause = cl
-            {-, dataArgOccurrences = occ-} } ->
+    Datatype{ dataPars = np, dataClause = cl } ->
       d { dataPars = np - size args
-        , dataSmallPars  = apply sps args
-        , dataNonLinPars = apply nlps args
         , dataClause     = apply cl args
---        , dataArgOccurrences = List.drop (length args) occ
         }
     Record{ recPars = np, recClause = cl, recTel = tel
           {-, recArgOccurrences = occ-} } ->
@@ -340,11 +317,11 @@ instance Apply Clause where
         -- We then have to create a substitution from the old telescope to the
         -- new telescope that we can apply to dot patterns and the clause body.
         rhoP :: PatternSubstitution
-        rhoP = mkSub DotP n rps rargs
+        rhoP = mkSub dotP n rps rargs
         rho  = mkSub id   n rps rargs
 
         substP :: Nat -> Term -> [NamedArg DeBruijnPattern] -> [NamedArg DeBruijnPattern]
-        substP i v = subst i (DotP v)
+        substP i v = subst i (dotP v)
 
         -- Building the substitution from the old telescope to the new. The
         -- interesting case is when we have a variable pattern:
@@ -361,10 +338,9 @@ instance Apply Clause where
         mkSub _ _ [] [] = idS
         mkSub tm n (p : ps) (v : vs) =
           case namedArg p of
-            VarP (DBPatVar _ i) -> mkSub tm (n - 1) (substP i v' ps) vs `composeS` singletonS i (tm v')
+            VarP _ (DBPatVar _ i) -> mkSub tm (n - 1) (substP i v' ps) vs `composeS` singletonS i (tm v')
               where v' = raise (n - 1) v
             DotP{}  -> mkSub tm n ps vs
-            AbsurdP p' -> mkSub tm n (setNamedArg p p' : ps) (v : vs)
             ConP c _ ps' -> mkSub tm n (ps' ++ ps) (projections c v ++ vs)
             LitP{}  -> __IMPOSSIBLE__
             ProjP{} -> __IMPOSSIBLE__
@@ -383,9 +359,8 @@ instance Apply Clause where
         newTel n tel [] [] = tel
         newTel n tel (p : ps) (v : vs) =
           case namedArg p of
-            VarP (DBPatVar _ i) -> newTel (n - 1) (subTel (size tel - 1 - i) v tel) (substP i (raise (n - 1) v) ps) vs
+            VarP _ (DBPatVar _ i) -> newTel (n - 1) (subTel (size tel - 1 - i) v tel) (substP i (raise (n - 1) v) ps) vs
             DotP{}              -> newTel n tel ps vs
-            AbsurdP{}           -> __IMPOSSIBLE__
             ConP c _ ps'        -> newTel n tel (ps' ++ ps) (projections c v ++ vs)
             LitP{}              -> __IMPOSSIBLE__
             ProjP{}             -> __IMPOSSIBLE__
@@ -417,10 +392,10 @@ instance Apply a => Apply (WithArity a) where
   applyE (WithArity n a) es   = WithArity n $ applyE a es
 
 instance Apply a => Apply (Case a) where
-  apply (Branches cop cs ls m) args =
-    Branches cop (apply cs args) (apply ls args) (apply m args)
-  applyE (Branches cop cs ls m) es =
-    Branches cop (applyE cs es) (applyE ls es) (applyE m es)
+  apply (Branches cop cs eta ls m lz) args =
+    Branches cop (apply cs args) (second (`apply` args) <$> eta) (apply ls args) (apply m args) lz
+  applyE (Branches cop cs eta ls m lz) es =
+    Branches cop (applyE cs es) (second (`applyE` es) <$> eta)(applyE ls es) (applyE m es) lz
 
 instance Apply FunctionInverse where
   apply NotInjective  args = NotInjective
@@ -440,11 +415,7 @@ instance Apply DisplayTerm where
   applyE (DDef c es')        es = DDef c $ es' ++ map (fmap DTerm) es
   applyE (DWithApp v ws es') es = DWithApp v ws $ es' ++ es
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPABLE #-} Apply t => Apply [t] where
-#else
-instance Apply t => Apply [t] where
-#endif
   apply  ts args = map (`apply` args) ts
   applyE ts es   = map (`applyE` es) ts
 
@@ -499,7 +470,6 @@ instance Abstract Permutation where
 piApply :: Type -> Args -> Type
 piApply t []                      = t
 piApply (El _ (Pi  _ b)) (a:args) = lazyAbsApp b (unArg a) `piApply` args
-piApply (El s (Shared p)) args    = piApply (El s $ derefPtr p) args
 piApply t args                    =
   trace ("piApply t = " ++ show t ++ "\n  args = " ++ show args) __IMPOSSIBLE__
 
@@ -532,19 +502,11 @@ instance Abstract RewriteRule where
   abstract tel (RewriteRule q gamma f ps rhs t) =
     RewriteRule q (abstract tel gamma) f ps rhs t
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPING #-} Abstract [Occ.Occurrence] where
-#else
-instance Abstract [Occ.Occurrence] where
-#endif
   abstract tel []  = []
   abstract tel occ = replicate (size tel) Mixed ++ occ -- TODO: check occurrence
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPING #-} Abstract [Polarity] where
-#else
-instance Abstract [Polarity] where
-#endif
   abstract tel []  = []
   abstract tel pol = replicate (size tel) Invariant ++ pol -- TODO: check polarity
 
@@ -581,10 +543,8 @@ instance Abstract Defn where
           d' = d { funProjection = Just $ abstract tel p }
           tel1 = telFromList $ drop (size tel - 1) $ telToList tel
 
-    Datatype{ dataPars = np, dataSmallPars = sps, dataNonLinPars = nlps, dataClause = cl } ->
+    Datatype{ dataPars = np, dataClause = cl } ->
       d { dataPars       = np + size tel
-        , dataSmallPars  = abstract tel sps
-        , dataNonLinPars = abstract tel nlps
         , dataClause     = abstract tel cl
         }
     Record{ recPars = np, recClause = cl, recTel = tel' } ->
@@ -621,8 +581,9 @@ instance Abstract a => Abstract (WithArity a) where
   abstract tel (WithArity n a) = WithArity n $ abstract tel a
 
 instance Abstract a => Abstract (Case a) where
-  abstract tel (Branches cop cs ls m) =
-    Branches cop (abstract tel cs) (abstract tel ls) (abstract tel m)
+  abstract tel (Branches cop cs eta ls m lz) =
+    Branches cop (abstract tel cs) (second (abstract tel) <$> eta)
+                 (abstract tel ls) (abstract tel m) lz
 
 telVars :: Int -> Telescope -> [Arg DeBruijnPattern]
 telVars m = map (fmap namedThing) . (namedTelVars m)
@@ -637,11 +598,7 @@ instance Abstract FunctionInverse where
   abstract tel NotInjective  = NotInjective
   abstract tel (Inverse inv) = Inverse $ abstract tel inv
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPABLE #-} Abstract t => Abstract [t] where
-#else
-instance Abstract t => Abstract [t] where
-#endif
   abstract tel = map (abstract tel)
 
 instance Abstract t => Abstract (Maybe t) where
@@ -695,23 +652,21 @@ instance Subst Term Term where
     Lit l       -> Lit l
     Level l     -> levelTm $ applySubst rho l
     Pi a b      -> uncurry Pi $ applySubst rho (a,b)
-    Sort s      -> sortTm $ applySubst rho s
-    Shared p    -> Shared $ applySubst rho p
+    Sort s      -> Sort $ applySubst rho s
     DontCare mv -> dontCare $ applySubst rho mv
 
-instance Subst t a => Subst t (Ptr a) where
-  applySubst rho = fmap (applySubst rho)
-
 instance Subst Term a => Subst Term (Type' a) where
   applySubst rho (El s t) = applySubst rho s `El` applySubst rho t
 
 instance Subst Term Sort where
   applySubst rho s = case s of
-    Type n     -> levelSort $ sub n
+    Type n     -> Type $ sub n
     Prop       -> Prop
     Inf        -> Inf
     SizeUniv   -> SizeUniv
-    DLub s1 s2 -> DLub (sub s1) (sub s2)
+    PiSort s1 s2 -> piSort (sub s1) (sub s2)
+    UnivSort s -> univSort $ sub s
+    MetaS x es -> MetaS x $ sub es
     where sub x = applySubst rho x
 
 instance Subst Term Level where
@@ -730,25 +685,25 @@ instance Subst Term LevelAtom where
 instance Subst Term Name where
   applySubst rho = id
 
-#if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPING #-} Subst Term String where
-#else
-instance Subst Term String where
-#endif
   applySubst rho = id
 
 instance Subst Term ConPatternInfo where
-  applySubst rho (ConPatternInfo mr mt) = ConPatternInfo mr $ applySubst rho mt
+  applySubst rho i = i{ conPType = applySubst rho $ conPType i }
 
 instance Subst Term Pattern where
   applySubst rho p = case p of
     ConP c mt ps -> ConP c (applySubst rho mt) $ applySubst rho ps
-    DotP t       -> DotP $ applySubst rho t
-    VarP s       -> p
-    AbsurdP p    -> AbsurdP $ applySubst rho p
+    DotP o t     -> DotP o $ applySubst rho t
+    VarP o s     -> p
     LitP l       -> p
     ProjP{}      -> p
 
+instance Subst Term A.ProblemEq where
+  applySubst rho (A.ProblemEq p v a) =
+    uncurry (A.ProblemEq p) $ applySubst rho (v,a)
+
+
 instance DeBruijn NLPat where
   deBruijnVar i = PVar i []
   deBruijnView p = case p of
@@ -828,7 +783,7 @@ instance Subst t a => Subst t (Tele a) where
 instance Subst Term Constraint where
   applySubst rho c = case c of
     ValueCmp cmp a u v       -> ValueCmp cmp (rf a) (rf u) (rf v)
-    ElimCmp ps a v e1 e2     -> ElimCmp ps (rf a) (rf v) (rf e1) (rf e2)
+    ElimCmp ps fs a v e1 e2  -> ElimCmp ps fs (rf a) (rf v) (rf e1) (rf e2)
     TypeCmp cmp a b          -> TypeCmp cmp (rf a) (rf b)
     TelCmp a b cmp tel1 tel2 -> TelCmp (rf a) (rf b) cmp (rf tel1) (rf tel2)
     SortCmp cmp s1 s2        -> SortCmp cmp (rf s1) (rf s2)
@@ -838,18 +793,12 @@ instance Subst Term Constraint where
     CheckSizeLtSat t         -> CheckSizeLtSat (rf t)
     FindInScope m b cands    -> FindInScope m b (rf cands)
     UnBlock{}                -> c
+    CheckFunDef{}            -> c
+    HasBiggerSort s          -> HasBiggerSort (rf s)
+    HasPTSRule s1 s2         -> HasPTSRule (rf s1) (rf s2)
     where
       rf x = applySubst rho x
 
-instance Subst Term ModuleParameters where
-  applySubst rho mp = mp { mpSubstitution = applySubst rho $ mpSubstitution mp }
-
-instance Subst Term A.NamedDotPattern where
-  applySubst rho (A.NamedDot x v a) = A.NamedDot x (applySubst rho v) (applySubst rho a)
-
-instance Subst Term A.StrippedDotPattern where
-  applySubst rho (A.StrippedDot e v a) = A.StrippedDot e (applySubst rho v) (applySubst rho a)
-
 instance Subst t a => Subst t (Elim' a) where
   applySubst rho e = case e of
     Apply v -> Apply $ applySubst rho v
@@ -860,13 +809,15 @@ instance Subst t a => Subst t (Abs a) where
   applySubst rho (NoAbs x a) = NoAbs x $ applySubst rho a
 
 instance Subst t a => Subst t (Arg a) where
-  applySubst rho = fmap (applySubst rho)
+  applySubst IdS arg = arg
+  applySubst rho arg = setFreeVariables unknownFreeVariables $ fmap (applySubst rho) arg
 
 instance Subst t a => Subst t (Named name a) where
   applySubst rho = fmap (applySubst rho)
 
 instance Subst t a => Subst t (Dom a) where
-  applySubst rho = fmap (applySubst rho)
+  applySubst IdS dom = dom
+  applySubst rho dom = setFreeVariables unknownFreeVariables $ fmap (applySubst rho) dom
 
 instance Subst t a => Subst t (Maybe a) where
   applySubst rho = fmap (applySubst rho)
@@ -904,9 +855,10 @@ instance Subst Term EqualityView where
     (applySubst rho b)
 
 instance DeBruijn DeBruijnPattern where
-  debruijnNamedVar n i  = VarP $ DBPatVar n i
-  deBruijnView (VarP x) = Just $ dbPatVarIndex x
-  deBruijnView _        = Nothing
+  debruijnNamedVar n i             = varP $ DBPatVar n i
+  -- deBruijnView returns Nothing, to prevent consS and the like
+  -- from dropping the names and origins when building a substitution.
+  deBruijnView _                   = Nothing
 
 fromPatternSubstitution :: PatternSubstitution -> Substitution
 fromPatternSubstitution = fmap patternToTerm
@@ -914,20 +866,42 @@ fromPatternSubstitution = fmap patternToTerm
 applyPatSubst :: (Subst Term a) => PatternSubstitution -> a -> a
 applyPatSubst = applySubst . fromPatternSubstitution
 
+
+usePatOrigin :: PatOrigin -> Pattern' a -> Pattern' a
+usePatOrigin o p = case patternOrigin p of
+  Nothing         -> p
+  Just PatOSplit  -> p
+  Just PatOAbsurd -> p
+  Just _          -> case p of
+    (VarP _ x) -> VarP o x
+    (DotP _ u) -> DotP o u
+    (ConP c (ConPatternInfo (Just _) b l) ps)
+      -> ConP c (ConPatternInfo (Just o) b l) ps
+    ConP{}  -> __IMPOSSIBLE__
+    LitP{}  -> __IMPOSSIBLE__
+    ProjP{} -> __IMPOSSIBLE__
+
 instance Subst DeBruijnPattern DeBruijnPattern where
   applySubst IdS p = p
   applySubst rho p = case p of
-    VarP x       -> useName (dbPatVarName x) $ lookupS rho $ dbPatVarIndex x
-    DotP u       -> DotP $ applyPatSubst rho u
+    VarP o x     ->
+      usePatOrigin o $
+      useName (dbPatVarName x) $
+      lookupS rho $ dbPatVarIndex x
+    DotP o u     -> DotP o $ applyPatSubst rho u
     ConP c ci ps -> ConP c ci $ applySubst rho ps
-    AbsurdP p    -> AbsurdP $ applySubst rho p
     LitP x       -> p
     ProjP{}      -> p
     where
       useName :: PatVarName -> DeBruijnPattern -> DeBruijnPattern
-      useName n (VarP x) | isUnderscore (dbPatVarName x) = debruijnNamedVar n (dbPatVarIndex x)
+      useName n (VarP o x)
+        | isUnderscore (dbPatVarName x)
+        = VarP o $ x { dbPatVarName = n }
       useName _ x = x
 
+instance Subst Term Range where
+  applySubst _ = id
+
 ---------------------------------------------------------------------------
 -- * Projections
 ---------------------------------------------------------------------------
@@ -960,7 +934,7 @@ projDropParsApply (Projection prop d r _ lams) o args =
 
 type TelView = TelV Type
 data TelV a  = TelV { theTel :: Tele (Dom a), theCore :: a }
-  deriving (Typeable, Show, Functor)
+  deriving (Show, Functor)
 
 deriving instance (Subst t a, Eq  a) => Eq  (TelV a)
 deriving instance (Subst t a, Ord a) => Ord (TelV a)
@@ -974,7 +948,7 @@ telView' = telView'UpTo (-1)
 -- Takes off all (exposed ones) if @n < 0@.
 telView'UpTo :: Int -> Type -> TelView
 telView'UpTo 0 t = TelV EmptyTel t
-telView'UpTo n t = case ignoreSharing $ unEl t of
+telView'UpTo n t = case unEl t of
   Pi a b  -> absV a (absName b) $ telView'UpTo (n - 1) (absBody b)
   _       -> TelV EmptyTel t
   where
@@ -1007,7 +981,7 @@ bindsWithHidingToTel = bindsWithHidingToTel' nameToArgName
 mkPi :: Dom (ArgName, Type) -> Type -> Type
 mkPi (Dom info (x, a)) b = el $ Pi (Dom info a) (mkAbs x b)
   where
-    el = El $ dLub (getSort a) (Abs x (getSort b)) -- dLub checks x freeIn
+    el = El $ piSort (getSort a) (Abs x (getSort b)) -- piSort checks x freeIn
 
 mkLam :: Arg ArgName -> Term -> Term
 mkLam a v = Lam (argInfo a) (Abs (unArg a) v)
@@ -1018,9 +992,7 @@ telePi' reAbs = telePi where
   telePi (ExtendTel u tel) t = el $ Pi u $ reAbs b
     where
       b  = (`telePi` t) <$> tel
-      s1 = getSort u
-      s2 = getSort <$> b
-      el = El $ dLub s1 s2
+      el = El $ piSort (getSort u) (getSort <$> b)
 
 -- | Uses free variable analysis to introduce 'NoAbs' bindings.
 telePi :: Telescope -> Type -> Type
@@ -1030,17 +1002,6 @@ telePi = telePi' reAbs
 telePi_ :: Telescope -> Type -> Type
 telePi_ = telePi' id
 
-{- OLD
--- | Everything will be a pi.
-telePi_  EmptyTel        t = t
-telePi_ (ExtendTel u tel) t = el $ Pi u b
-  where
-    el = El (dLub s1 s2)
-    b  = fmap (flip telePi_ t) tel
-    s1 = getSort $ unDom u
-    s2 = fmap getSort b
--}
-
 -- | Abstract over a telescope in a term, producing lambdas.
 --   Dumb abstraction: Always produces 'Abs', never 'NoAbs'.
 --
@@ -1094,11 +1055,8 @@ deriving instance Ord Sort
 deriving instance Eq Level
 deriving instance Ord Level
 deriving instance Eq PlusLevel
-deriving instance Ord LevelAtom
 deriving instance Eq NotBlocked
-deriving instance Ord NotBlocked
 deriving instance Eq t => Eq (Blocked t)
-deriving instance Ord t => Ord (Blocked t)
 deriving instance Eq Candidate
 
 deriving instance (Subst t a, Eq a)  => Eq  (Tele a)
@@ -1117,6 +1075,9 @@ instance Ord PlusLevel where
 instance Eq LevelAtom where
   (==) = (==) `on` unLevelAtom
 
+instance Ord LevelAtom where
+  compare = compare `on` unLevelAtom
+
 -- | Syntactic 'Type' equality, ignores sort annotations.
 instance Eq a => Eq (Type' a) where
   (==) = (==) `on` unEl
@@ -1136,16 +1097,10 @@ instance Eq Term where
   Level l    == Level l'     = l == l'
   MetaV m vs == MetaV m' vs' = m == m' && vs == vs'
   DontCare _ == DontCare _   = True
-  Shared p   == Shared q     = p == q || derefPtr p == derefPtr q
-  Shared p   == b            = derefPtr p == b
-  a          == Shared q     = a == derefPtr q
   _          == _            = False
 
 instance Ord Term where
-  Shared a   `compare` Shared x | a == x = EQ
-  Shared a   `compare` x          = compare (derefPtr a) x
-  a          `compare` Shared x   = compare a (derefPtr x)
-  Var a b    `compare` Var x y    = compare (a, b) (x, y)
+  Var a b    `compare` Var x y    = compare x a `thenCmp` compare b y -- sort de Bruijn indices down (#2765)
   Var{}      `compare` _          = LT
   _          `compare` Var{}      = GT
   Def a b    `compare` Def x y    = compare (a, b) (x, y)
@@ -1174,9 +1129,29 @@ instance Ord Term where
   _          `compare` MetaV{}    = GT
   DontCare{} `compare` DontCare{} = EQ
 
+-- Andreas, 2017-10-04, issue #2775, ignore irrelevant arguments during with-abstraction.
+--
+-- For reasons beyond my comprehension, the following Eq instances are not employed
+-- by with-abstraction in TypeChecking.Abstract.isPrefixOf.
+-- Instead, I modified the general Eq instance for Arg to ignore the argument
+-- if irrelevant.
+
+-- -- | Ignore irrelevant arguments in equality check.
+-- --   Also ignore origin.
+-- instance {-# OVERLAPPING #-} Eq (Arg Term) where
+--   a@(Arg (ArgInfo h r _o) t) == a'@(Arg (ArgInfo h' r' _o') t') = trace ("Eq (Arg Term) on " ++ show a ++ " and " ++ show a') $
+--     h == h' && ((r == Irrelevant) || (r' == Irrelevant) || (t == t'))
+--     -- Andreas, 2017-10-04: According to Syntax.Common, equality on Arg ignores Relevance and Origin.
+
+-- instance {-# OVERLAPPING #-} Eq Args where
+--   us == vs = length us == length vs && and (zipWith (==) us vs)
+
+-- instance {-# OVERLAPPING #-} Eq Elims where
+--   us == vs = length us == length vs && and (zipWith (==) us vs)
+
 -- | Equality of binders relies on weakening
---   which is a specical case of renaming
---   which is a specical case of substitution.
+--   which is a special case of renaming
+--   which is a special case of substitution.
 instance (Subst t a, Eq a) => Eq (Abs a) where
   NoAbs _ a == NoAbs _ b = a == b
   Abs   _ a == Abs   _ b = a == b
@@ -1199,67 +1174,66 @@ instance (Subst t a, Ord a) => Ord (Elim' a) where
   Proj{}   `compare` Apply{}  = GT
 
 ---------------------------------------------------------------------------
--- * Level stuff
+-- * Sort stuff
 ---------------------------------------------------------------------------
 
--- | The ``rule'', if Agda is considered as a functional
---   pure type system (pts).
---
---   TODO: This needs to be properly implemented, requiring
---   refactoring of Agda's handling of levels.
---   Without impredicativity or 'SizeUniv', Agda's pts rule is
---   just the least upper bound, which is total and commutative.
---   The handling of levels relies on this simplification.
-pts :: Sort -> Sort -> Sort
-pts = sLub
+-- | Get the next higher sort.
+univSort' :: Sort -> Maybe Sort
+univSort' (Type l) = Just $ Type $ levelSuc l
+univSort' s        = Nothing
 
-sLub :: Sort -> Sort -> Sort
-sLub s Prop = s
-sLub Prop s = s
-sLub Inf _ = Inf
-sLub _ Inf = Inf
-sLub SizeUniv s = s         -- one can freely quantify over sizes in any Set
-sLub _ SizeUniv = SizeUniv  -- but everything resulting in a size lives in the SizeUniv
-sLub (Type (Max as)) (Type (Max bs)) = Type $ levelMax (as ++ bs)
--- sLub (DLub a b) c = DLub (sLub a c) b -- no longer commutative!
-sLub (DLub a NoAbs{}) c = __IMPOSSIBLE__
-sLub (DLub a (Abs x b)) c = DLub a $ Abs x $ sLub b $ raise 1 c
-sLub a (DLub b c) = DLub (sLub a b) c
+univSort :: Sort -> Sort
+univSort s = fromMaybe (UnivSort s) $ univSort' s
 
--- | Dependent least upper bound, to assign a level to expressions
---   like @forall i -> Set i@.
---
---   @dLub s1 \i.s2 = \omega@ if @i@ appears in the rigid variables of @s2@.
-dLub :: Sort -> Abs Sort -> Sort
-dLub s1 (NoAbs _ s2) = sLub s1 s2
-dLub s1 b@(Abs _ s2) = case occurrence 0 s2 of
-  Flexible _    -> DLub s1 b
-  -- Andreas, 2017-01-18, issue #2408:
-  -- The sort of @.(a : A) → Set (f a)@ in context @f : .A → Level@
-  -- is @dLub Set λ a → Set (lsuc (f a))@, but @DLub@s are not serialized.
-  -- Alternatives:
-  -- 1. -- Irrelevantly -> sLub s1 (absApp b $ DontCare $ Sort Prop)
-  --    We cheat here by simplifying the sort to @Set (lsuc (f *))@
-  --    where * is a dummy value.  The rationale is that @f * = f a@ (irrelevance!)
-  --    and that if we already have a neutral level @f a@
-  --    it should not hurt to have @f *@ even if type @A@ is empty.
-  --    However: sorts are printed in error messages when sorts do not match.
-  --    Also, sorts with a dummy like Prop would be ill-typed.
-  -- 2. We keep the DLub, and serialize it.
-  --    That's clean and principled, even though DLubs make level solving harder.
-  Irrelevantly  -> DLub s1 b
-  NoOccurrence  -> sLub s1 (noabsApp __IMPOSSIBLE__ b)
---  Free.Unused   -> sLub s1 (absApp b __IMPOSSIBLE__) -- triggers Issue784
-  Free.Unused   -> DLub s1 b
-  StronglyRigid -> Inf
-  Unguarded     -> Inf
-  WeaklyRigid   -> Inf
+-- | Compute the sort of a function type from the sorts of its
+--   domain and codomain.
+funSort' :: Sort -> Sort -> Maybe Sort
+funSort' a b = case (a, b) of
+  (Inf           , _            ) -> Just Inf
+  (_             , Inf          ) -> Just Inf
+  (Type (Max as) , Type (Max bs)) -> Just $ Type $ levelMax $ as ++ bs
+  (SizeUniv      , b            ) -> Just b
+  (_             , SizeUniv     ) -> Just SizeUniv
+  (a             , b            ) -> Nothing
 
-lvlView :: Term -> Level
-lvlView v = case ignoreSharing v of
-  Level l       -> l
-  Sort (Type l) -> l
-  _             -> Max [Plus 0 $ UnreducedLevel v]
+funSort :: Sort -> Sort -> Sort
+funSort a b = fromMaybe (PiSort a (NoAbs underscore b)) $ funSort' a b
+
+-- | Compute the sort of a pi type from the sorts of its domain
+--   and codomain.
+piSort' :: Sort -> Abs Sort -> Maybe Sort
+piSort' a      (NoAbs _ b) = funSort' a b
+piSort' a bAbs@(Abs   _ b) = case occurrence 0 b of
+    NoOccurrence  -> funSort' a $ noabsApp {-'-} __IMPOSSIBLE__ bAbs
+    -- Andreas, 2017-01-18, issue #2408:
+    -- The sort of @.(a : A) → Set (f a)@ in context @f : .A → Level@
+    -- is @dLub Set λ a → Set (lsuc (f a))@, but @DLub@s are not serialized.
+    -- Alternatives:
+    -- 1. -- Irrelevantly -> sLub s1 (absApp b $ DontCare $ Sort Prop)
+    --    We cheat here by simplifying the sort to @Set (lsuc (f *))@
+    --    where * is a dummy value.  The rationale is that @f * = f a@ (irrelevance!)
+    --    and that if we already have a neutral level @f a@
+    --    it should not hurt to have @f *@ even if type @A@ is empty.
+    --    However: sorts are printed in error messages when sorts do not match.
+    --    Also, sorts with a dummy like Prop would be ill-typed.
+    -- 2. We keep the DLub, and serialize it.
+    --    That's clean and principled, even though DLubs make level solving harder.
+    -- Jesper, 2018-04-20: another alternative:
+    -- 3. Return @Inf@ as in the relevant case. This is conservative and might result
+    --    in more occurrences of @Setω@ than desired, but at least it doesn't pollute
+    --    the sort system with new 'exotic' sorts.
+    Irrelevantly  -> Just Inf
+    StronglyRigid -> Just Inf
+    Unguarded     -> Just Inf
+    WeaklyRigid   -> Just Inf
+    Flexible _    -> Nothing
+
+piSort :: Sort -> Abs Sort -> Sort
+piSort a b = fromMaybe (PiSort a b) $ piSort' a b
+
+---------------------------------------------------------------------------
+-- * Level stuff
+---------------------------------------------------------------------------
 
 levelMax :: [PlusLevel] -> Level
 levelMax as0 = Max $ ns ++ List.sort bs
@@ -1285,7 +1259,7 @@ levelMax as0 = Max $ ns ++ List.sort bs
       UnreducedLevel v -> expandTm v
       MetaLevel{}      -> [Plus 0 l]
       where
-        expandTm v = case ignoreSharing v of
+        expandTm v = case v of
           Level (Max as)       -> as
           Sort (Type (Max as)) -> as
           _                    -> [Plus 0 l]
@@ -1301,29 +1275,6 @@ levelMax as0 = Max $ ns ++ List.sort bs
       where
         ns = [ m | Plus m a'  <- bs, a == a', m > n ]
 
-sortTm :: Sort -> Term
-sortTm (Type l) = Sort $ levelSort l
-sortTm s        = Sort s
-
-levelSort :: Level -> Sort
-levelSort (Max as)
-  | List.any (levelIs Inf     ) as = Inf
-  | List.any (levelIs SizeUniv) as = SizeUniv
-  where
-    levelIs s ClosedLevel{}     = False
-    levelIs s (Plus _ l)        = atomIs s l
-    atomIs s (NeutralLevel _ a) = tmIs s a
-    atomIs s (UnreducedLevel a) = tmIs s a
-    atomIs s MetaLevel{}        = False
-    atomIs s BlockedLevel{}     = False
-    tmIs s (Sort s')            = s == s'
-    tmIs s (Shared p)           = tmIs s $ derefPtr p
-    tmIs s _                    = False
-levelSort l =
-  case ignoreSharing $ levelTm l of
-    Sort s -> s
-    _      -> Type l
-
 levelTm :: Level -> Term
 levelTm l =
   case l of
@@ -1335,3 +1286,15 @@ unLevelAtom (MetaLevel x es)   = MetaV x es
 unLevelAtom (NeutralLevel _ v) = v
 unLevelAtom (UnreducedLevel v) = v
 unLevelAtom (BlockedLevel _ v) = v
+
+levelSucView :: Level -> Maybe Level
+levelSucView (Max []) = Nothing
+levelSucView (Max as) = Max <$> traverse atomPred as
+  where
+    atomPred :: PlusLevel -> Maybe PlusLevel
+    atomPred (ClosedLevel n)
+      | n > 0     = Just $ ClosedLevel (n-1)
+      | otherwise = Nothing
+    atomPred (Plus n l)
+      | n > 0     = Just $ Plus (n-1) l
+      | otherwise = Nothing
