@@ -68,22 +68,6 @@ pure2 a = pure (a,a)
 (<**>) :: Applicative f => f (a -> b, a -> b) -> f (a,a) -> f (b,b)
 ff <**> xx = pure (uncurry (***)) <*> ff <*> xx
 
-{-
-updateSharedM2 :: Monad m =>  (Term -> Term -> m (Term, Term)) -> Term -> Term -> m (Term, Term)
-updateSharedM2 f v0@(Shared p) = do
-  v <- f (derefPtr p)
-  case derefPtr (setPtr v p) of
-    Var _ [] -> return v
-    _        -> compressPointerChain v0 `pseq` return v0
-updateSharedM2 f v = f v
-
-updateSharedTerm2 :: MonadTCM tcm => (Term -> Term -> tcm (Term, Term)) -> Term -> Term -> tcm (Term, Term)
-updateSharedTerm f v v' =
-  ifM (liftTCM $ asks envAllowDestructiveUpdate)
-      (updateSharedM2 f v v')
-      (f (ignoreSharing v) (ignoreSharing v'))
--}
-
 -- | Instantiate full as long as things are equal
 class SynEq a where
   synEq  :: a -> a -> SynEqM (a,a)
@@ -94,9 +78,7 @@ class SynEq a where
 instance SynEq Term where
   synEq v v' = do
     (v, v') <- lift $ instantiate' (v, v')
-    -- currently destroys sharing
-    -- TODO: preserve sharing!
-    case (ignoreSharing v, ignoreSharing v') of
+    case (v, v') of
       (Var   i vs, Var   i' vs') | i == i' -> Var i   <$$> synEq vs vs'
       (Con c ci vs,Con c' ci' vs') | c == c' -> Con c (bestConInfo ci ci') <$$> synEq vs vs'
       (Def   f vs, Def   f' vs') | f == f' -> Def f   <$$> synEq vs vs'
@@ -104,13 +86,11 @@ instance SynEq Term where
       (Lit   l   , Lit   l'    ) | l == l' -> pure2 $ v
       (Lam   h b , Lam   h' b' ) | h == h' -> Lam h   <$$> synEq b b'
       (Level l   , Level l'    )           -> levelTm <$$> synEq l l'
-      (Sort  s   , Sort  s'    )           -> sortTm  <$$> synEq s s'
+      (Sort  s   , Sort  s'    )           -> Sort    <$$> synEq s s'
       (Pi    a b , Pi    a' b' )           -> Pi      <$$> synEq a a' <**> synEq' b b'
       (DontCare _, DontCare _  )           -> pure (v, v')
          -- Irrelevant things are syntactically equal. ALT:
          -- DontCare <$$> synEq v v'
-      (Shared{}  , _           )           -> __IMPOSSIBLE__
-      (_         , Shared{}    )           -> __IMPOSSIBLE__
       _                                    -> inequal (v, v')
 
 instance SynEq Level where
@@ -146,10 +126,13 @@ instance SynEq Sort where
   synEq s s' = do
     (s, s') <- lift $ instantiate' (s, s')
     case (s, s') of
-      (Type l  , Type l'   ) -> levelSort <$$> synEq l l'
-      (DLub a b, DLub a' b') -> dLub <$$> synEq a a' <**> synEq' b b'
+      (Type l  , Type l'   ) -> Type <$$> synEq l l'
+      (PiSort a b, PiSort a' b') -> piSort <$$> synEq a a' <**> synEq' b b'
+      (UnivSort a, UnivSort a') -> univSort <$$> synEq a a'
+      (SizeUniv, SizeUniv  ) -> pure2 s
       (Prop    , Prop      ) -> pure2 s
       (Inf     , Inf       ) -> pure2 s
+      (MetaS x es , MetaS x' es') | x == x' -> MetaS x <$$> synEq es es'
       _ -> inequal (s, s')
 
 -- | Syntactic equality ignores sorts.
@@ -193,6 +176,6 @@ instance SynEq a => SynEq (Dom a) where
   synEq (Dom ai a) (Dom ai' a') = Dom <$$> synEq ai ai' <**> synEq a a'
 
 instance SynEq ArgInfo where
-  synEq ai@(ArgInfo h r o) ai'@(ArgInfo h' r' o')
+  synEq ai@(ArgInfo h r o _) ai'@(ArgInfo h' r' o' _)
     | h == h', r == r' = pure2 ai
     | otherwise        = inequal (ai, ai')
