@@ -77,13 +77,19 @@ teleArgNames = map (argFromDom . fmap fst) . telToList
 
 teleArgs :: (DeBruijn a) => Telescope -> [Arg a]
 teleArgs tel =
-  [ Arg info (debruijnNamedVar n i)
+  [ Arg info (deBruijnVar i)
   | (i, Dom info (n,_)) <- zip (downFrom $ size l) l ]
   where l = telToList tel
 
+withNamedArgsFromTel :: [a] -> Telescope -> [NamedArg a]
+xs `withNamedArgsFromTel` tel =
+  [ Arg info (Named (Just $ Ranged noRange $ argNameToString name) x)
+  | (x, Dom info (name,_)) <- zip xs l ]
+  where l = telToList tel
+
 teleNamedArgs :: (DeBruijn a) => Telescope -> [NamedArg a]
 teleNamedArgs tel =
-  [ Arg info (Named (Just $ Ranged noRange $ argNameToString name) (debruijnNamedVar name i))
+  [ Arg info (Named (Just $ Ranged noRange $ argNameToString name) (deBruijnVar i))
   | (i, Dom info (name,_)) <- zip (downFrom $ size l) l ]
   where l = telToList tel
 
@@ -99,6 +105,18 @@ tele2NamedArgs tel0 tel =
   l  = telToList tel
   l0 = telToList tel0
 
+-- | Split the telescope at the specified position.
+splitTelescopeAt :: Int -> Telescope -> (Telescope,Telescope)
+splitTelescopeAt n tel
+  | n <= 0    = (EmptyTel, tel)
+  | otherwise = splitTelescopeAt' n tel
+    where
+      splitTelescopeAt' _ EmptyTel          = (EmptyTel,EmptyTel)
+      splitTelescopeAt' 1 (ExtendTel a tel) = (ExtendTel a (tel $> EmptyTel), absBody tel)
+      splitTelescopeAt' m (ExtendTel a tel) = (ExtendTel a (tel $> tel'), tel'')
+        where
+          (tel', tel'') = splitTelescopeAt (m - 1) $ absBody tel
+
 -- | Permute telescope: permutes or drops the types in the telescope according
 --   to the given permutation. Assumes that the permutation preserves the
 --   dependencies in the telescope.
@@ -242,7 +260,7 @@ instantiateTelescope tel k u = guard ok $> (tel', sigma, rho)
     rho   = reverseP perm  -- works on de Bruijn levels
 
     u1    = renameP __IMPOSSIBLE__ perm u -- Γ' ⊢ u1 : A'
-    us    = map (\i -> fromMaybe (DotP u1) (deBruijnVar <$> List.findIndex (i ==) is)) [ 0 .. n-1 ]
+    us    = map (\i -> fromMaybe (dotP u1) (deBruijnVar <$> List.findIndex (i ==) is)) [ 0 .. n-1 ]
     sigma = us ++# raiseS (n-1)
 
     ts1   = permute rho $ applyPatSubst sigma ts0
@@ -263,8 +281,9 @@ expandTelescopeVar gamma k delta c = (tel', rho)
                     splitExactlyAt k $ telToList gamma
 
     cpi         = ConPatternInfo
-      { conPRecord = Just ConOSystem
+      { conPRecord = Just PatOSystem
       , conPType   = Just $ snd <$> argFromDom a
+      , conPLazy   = True
       }
     cargs       = map (setOrigin Inserted) $ teleNamedArgs delta
     cdelta      = ConP c cpi cargs                    -- Γ₁Δ ⊢ c Δ : D pars
@@ -292,7 +311,7 @@ telViewUpTo' :: Int -> (Dom Type -> Bool) -> Type -> TCM TelView
 telViewUpTo' 0 p t = return $ TelV EmptyTel t
 telViewUpTo' n p t = do
   t <- reduce t
-  case ignoreSharing $ unEl t of
+  case unEl t of
     Pi a b | p a -> absV a (absName b) <$> telViewUpTo' (n - 1) p (absBody b)
     _            -> return $ TelV EmptyTel t
   where
@@ -308,7 +327,7 @@ mustBePi t = ifNotPiType t __IMPOSSIBLE__ $ \ a b -> return (a,b)
 ifPi :: MonadTCM tcm => Term -> (Dom Type -> Abs Type -> tcm a) -> (Term -> tcm a) -> tcm a
 ifPi t yes no = do
   t <- liftTCM $ reduce t
-  case ignoreSharing t of
+  case t of
     Pi a b -> yes a b
     _      -> no t
 
@@ -360,8 +379,8 @@ data OutputTypeName
 getOutputTypeName :: Type -> TCM OutputTypeName
 getOutputTypeName t = do
   TelV tel t' <- telView t
-  ifBlocked (unEl t') (\ _ _ -> return OutputTypeNameNotYetKnown) $ \ v ->
-    case ignoreSharing v of
+  ifBlocked (unEl t') (\ _ _ -> return OutputTypeNameNotYetKnown) $ \ _ v ->
+    case v of
       -- Possible base types:
       Def n _  -> return $ OutputTypeName n
       Sort{}   -> return NoOutputTypeName
@@ -373,7 +392,6 @@ getOutputTypeName t = do
       Level{}  -> __IMPOSSIBLE__
       MetaV{}  -> __IMPOSSIBLE__
       Pi{}     -> __IMPOSSIBLE__
-      Shared{} -> __IMPOSSIBLE__
       DontCare{} -> __IMPOSSIBLE__
 
 addTypedInstance :: QName -> Type -> TCM ()
