@@ -2,19 +2,20 @@
 
 module Agda.TypeChecking.Unquote where
 
-import Control.Arrow ((&&&), (***), first, second)
-import Control.Applicative
-import Control.Monad.State (StateT(..), evalStateT, get, gets, put, modify)
-import Control.Monad.Reader (ReaderT(..), ask, asks)
-import Control.Monad.Writer (WriterT(..), execWriterT, runWriterT, tell)
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ((<>))
+#endif
+
+import Control.Arrow (first, second)
+import Control.Monad.State
+import Control.Monad.Reader
+import Control.Monad.Writer hiding ((<>))
 import Control.Monad.Trans (lift)
-import Control.Monad
 
 import Data.Char
 import Data.Maybe (fromMaybe)
 import Data.Traversable (traverse)
-import Data.Map (Map)
-import qualified Data.Map as Map
+import Data.Word
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal as I
@@ -23,23 +24,16 @@ import Agda.Syntax.Literal
 import Agda.Syntax.Position
 import Agda.Syntax.Fixity
 import Agda.Syntax.Info
-import Agda.Syntax.Translation.InternalToAbstract
 import Agda.Syntax.Translation.ReflectedToAbstract
 
-import Agda.TypeChecking.CompiledClause
-import Agda.TypeChecking.DropArgs
-import Agda.TypeChecking.Free
-import Agda.TypeChecking.Level
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce
-import Agda.TypeChecking.Reduce.Monad
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Quote
 import Agda.TypeChecking.Conversion
-import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.EtaContract
 import Agda.TypeChecking.Primitive
 
@@ -55,14 +49,10 @@ import Agda.Utils.Except
 import Agda.Utils.Either
 import Agda.Utils.FileName
 import Agda.Utils.Lens
-import Agda.Utils.Maybe
-import Agda.Utils.Maybe.Strict (toLazy)
 import Agda.Utils.Monad
-import Agda.Utils.Permutation ( Permutation(Perm), compactP )
 import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.String ( Str(Str), unStr )
-import Agda.Utils.VarSet (VarSet)
-import qualified Agda.Utils.VarSet as Set
+import qualified Agda.Interaction.Options.Lenses as Lens
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -124,21 +114,21 @@ inOriginalContext m =
 
 isCon :: ConHead -> TCM Term -> UnquoteM Bool
 isCon con tm = do t <- liftU tm
-                  case ignoreSharing t of
+                  case t of
                     Con con' _ _ -> return (con == con')
                     _ -> return False
 
 isDef :: QName -> TCM Term -> UnquoteM Bool
 isDef f tm = do
   t <- liftU (etaContract =<< normalise =<< tm)
-  case ignoreSharing t of
+  case t of
     Def g _ -> return (f == g)
     _       -> return False
 
 reduceQuotedTerm :: Term -> UnquoteM Term
 reduceQuotedTerm t = do
   b <- liftU $ ifBlocked t (\ m _ -> pure $ Left  m)
-                           (\ t   -> pure $ Right t)
+                           (\ _ t -> pure $ Right t)
   case b of
     Left m  -> do s <- gets snd; throwError $ BlockedOnMeta s m
     Right t -> return t
@@ -184,13 +174,18 @@ pickName a =
               isAlpha c -> [toLower c]
     _        -> "_"
 
+-- TODO: reflect Quantity
+instance Unquote Modality where
+  unquote t = (`Modality` defaultQuantity) <$> unquote t
+
 instance Unquote ArgInfo where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
-      Con c _ [h,r] -> do
+    case t of
+      Con c _ es | Just [h,r] <- allApplyElims es -> do
         choice
-          [(c `isCon` primArgArgInfo, ArgInfo <$> unquoteN h <*> unquoteN r <*> pure Reflected)]
+          [(c `isCon` primArgArgInfo,
+              ArgInfo <$> unquoteN h <*> unquoteN r <*> pure Reflected <*> pure unknownFreeVariables)]
           __IMPOSSIBLE__
       Con c _ _ -> __IMPOSSIBLE__
       _ -> throwError $ NonCanonical "arg info" t
@@ -198,8 +193,8 @@ instance Unquote ArgInfo where
 instance Unquote a => Unquote (Arg a) where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
-      Con c _ [info,x] -> do
+    case t of
+      Con c _ es | Just [info,x] <- allApplyElims es -> do
         choice
           [(c `isCon` primArgArg, Arg <$> unquoteN info <*> unquoteN x)]
           __IMPOSSIBLE__
@@ -214,7 +209,7 @@ instance Unquote R.Elim where
 instance Unquote Bool where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
+    case t of
       Con c _ [] ->
         choice [ (c `isCon` primTrue,  pure True)
                , (c `isCon` primFalse, pure False) ]
@@ -224,28 +219,35 @@ instance Unquote Bool where
 instance Unquote Integer where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
+    case t of
       Lit (LitNat _ n) -> return n
       _ -> throwError $ NonCanonical "integer" t
 
+instance Unquote Word64 where
+  unquote t = do
+    t <- reduceQuotedTerm t
+    case t of
+      Lit (LitWord64 _ n) -> return n
+      _ -> throwError $ NonCanonical "word64" t
+
 instance Unquote Double where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
+    case t of
       Lit (LitFloat _ x) -> return x
       _ -> throwError $ NonCanonical "float" t
 
 instance Unquote Char where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
+    case t of
       Lit (LitChar _ x) -> return x
       _ -> throwError $ NonCanonical "char" t
 
 instance Unquote Str where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
+    case t of
       Lit (LitString _ x) -> return (Str x)
       _ -> throwError $ NonCanonical "string" t
 
@@ -265,8 +267,8 @@ instance PrettyTCM ErrorPart where
 instance Unquote ErrorPart where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
-      Con c _ [x] ->
+    case t of
+      Con c _ es | Just [x] <- allApplyElims es ->
         choice [ (c `isCon` primAgdaErrorPartString, StrPart  <$> unquoteNString x)
                , (c `isCon` primAgdaErrorPartTerm,   TermPart <$> unquoteN x)
                , (c `isCon` primAgdaErrorPartName,   NamePart <$> unquoteN x) ]
@@ -276,8 +278,8 @@ instance Unquote ErrorPart where
 instance Unquote a => Unquote [a] where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
-      Con c _ [x,xs] -> do
+    case t of
+      Con c _ es | Just [x,xs] <- allApplyElims es -> do
         choice
           [(c `isCon` primCons, (:) <$> unquoteN x <*> unquoteN xs)]
           __IMPOSSIBLE__
@@ -291,7 +293,7 @@ instance Unquote a => Unquote [a] where
 instance Unquote Hiding where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
+    case t of
       Con c _ [] -> do
         choice
           [(c `isCon` primHidden,  return Hidden)
@@ -304,7 +306,7 @@ instance Unquote Hiding where
 instance Unquote Relevance where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
+    case t of
       Con c _ [] -> do
         choice
           [(c `isCon` primRelevant,   return Relevant)
@@ -316,15 +318,15 @@ instance Unquote Relevance where
 instance Unquote QName where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
+    case t of
       Lit (LitQName _ x) -> return x
       _                  -> throwError $ NonCanonical "name" t
 
 instance Unquote a => Unquote (R.Abs a) where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
-      Con c _ [x,y] -> do
+    case t of
+      Con c _ es | Just [x,y] <- allApplyElims es -> do
         choice
           [(c `isCon` primAbsAbs, R.Abs <$> (hint <$> unquoteNString x) <*> unquoteN y)]
           __IMPOSSIBLE__
@@ -340,7 +342,7 @@ getCurrentPath = fromMaybe __IMPOSSIBLE__ <$> asks envCurrentPath
 instance Unquote MetaId where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
+    case t of
       Lit (LitMeta r f x) -> liftU $ do
         live <- (f ==) <$> getCurrentPath
         unless live $ do
@@ -357,12 +359,12 @@ instance Unquote a => Unquote (Dom a) where
 instance Unquote R.Sort where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
+    case t of
       Con c _ [] -> do
         choice
           [(c `isCon` primAgdaSortUnsupported, return R.UnknownS)]
           __IMPOSSIBLE__
-      Con c _ [u] -> do
+      Con c _ es | Just [u] <- allApplyElims es -> do
         choice
           [(c `isCon` primAgdaSortSet, R.SetS <$> unquoteN u)
           ,(c `isCon` primAgdaSortLit, R.LitS <$> unquoteN u)]
@@ -376,8 +378,8 @@ instance Unquote Literal where
     let litMeta r x = do
           file <- liftU getCurrentPath
           return $ LitMeta r file x
-    case ignoreSharing t of
-      Con c _ [x] ->
+    case t of
+      Con c _ es | Just [x] <- allApplyElims es ->
         choice
           [ (c `isCon` primAgdaLitNat,    LitNat    noRange <$> unquoteN x)
           , (c `isCon` primAgdaLitFloat,  LitFloat  noRange <$> unquoteN x)
@@ -392,19 +394,19 @@ instance Unquote Literal where
 instance Unquote R.Term where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
+    case t of
       Con c _ [] ->
         choice
           [ (c `isCon` primAgdaTermUnsupported, return R.Unknown) ]
           __IMPOSSIBLE__
 
-      Con c _ [x] -> do
+      Con c _ es | Just [x] <- allApplyElims es -> do
         choice
           [ (c `isCon` primAgdaTermSort,      R.Sort      <$> unquoteN x)
           , (c `isCon` primAgdaTermLit,       R.Lit       <$> unquoteN x) ]
           __IMPOSSIBLE__
 
-      Con c _ [x, y] ->
+      Con c _ es | Just [x, y] <- allApplyElims es ->
         choice
           [ (c `isCon` primAgdaTermVar,     R.Var     <$> (fromInteger <$> unquoteN x) <*> unquoteN y)
           , (c `isCon` primAgdaTermCon,     R.Con     <$> (ensureCon =<< unquoteN x) <*> unquoteN y)
@@ -430,19 +432,19 @@ instance Unquote R.Term where
 instance Unquote R.Pattern where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
+    case t of
       Con c _ [] -> do
         choice
           [ (c `isCon` primAgdaPatAbsurd, return R.AbsurdP)
           , (c `isCon` primAgdaPatDot,    return R.DotP)
           ] __IMPOSSIBLE__
-      Con c _ [x] -> do
+      Con c _ es | Just [x] <- allApplyElims es -> do
         choice
           [ (c `isCon` primAgdaPatVar,  R.VarP  <$> unquoteNString x)
           , (c `isCon` primAgdaPatProj, R.ProjP <$> unquoteN x)
           , (c `isCon` primAgdaPatLit,  R.LitP  <$> unquoteN x) ]
           __IMPOSSIBLE__
-      Con c _ [x, y] -> do
+      Con c _ es | Just [x, y] <- allApplyElims es -> do
         choice
           [ (c `isCon` primAgdaPatCon, R.ConP <$> unquoteN x <*> unquoteN y) ]
           __IMPOSSIBLE__
@@ -452,12 +454,12 @@ instance Unquote R.Pattern where
 instance Unquote R.Clause where
   unquote t = do
     t <- reduceQuotedTerm t
-    case ignoreSharing t of
-      Con c _ [x] -> do
+    case t of
+      Con c _ es | Just [x] <- allApplyElims es -> do
         choice
           [ (c `isCon` primAgdaClauseAbsurd, R.AbsurdClause <$> unquoteN x) ]
           __IMPOSSIBLE__
-      Con c _ [x, y] -> do
+      Con c _ es | Just [x, y] <- allApplyElims es -> do
         choice
           [ (c `isCon` primAgdaClauseClause, R.Clause <$> unquoteN x <*> unquoteN y) ]
           __IMPOSSIBLE__
@@ -481,7 +483,7 @@ evalTCM v = do
   liftU $ reportSDoc "tc.unquote.eval" 90 $ text "evalTCM" <+> prettyTCM v
   let failEval = throwError $ NonCanonical "type checking computation" v
 
-  case ignoreSharing v of
+  case v of
     I.Def f [] ->
       choice [ (f `isDef` primAgdaTCMGetContext, tcGetContext)
              , (f `isDef` primAgdaTCMCommit,     tcCommit) ]
@@ -499,6 +501,7 @@ evalTCM v = do
       choice [ (f `isDef` primAgdaTCMUnify,      tcFun2 tcUnify      u v)
              , (f `isDef` primAgdaTCMCheckType,  tcFun2 tcCheckType  u v)
              , (f `isDef` primAgdaTCMDeclareDef, uqFun2 tcDeclareDef u v)
+             , (f `isDef` primAgdaTCMDeclarePostulate, uqFun2 tcDeclarePostulate u v)
              , (f `isDef` primAgdaTCMDefineFun,  uqFun2 tcDefineFun  u v) ]
              failEval
     I.Def f [l, a, u] ->
@@ -644,7 +647,7 @@ evalTCM v = do
     extendCxt :: Arg R.Type -> UnquoteM a -> UnquoteM a
     extendCxt a m = do
       a <- liftU $ traverse (isType_ <=< toAbstract_) a
-      liftU1 (addContext' (domFromArg a :: Dom Type)) m
+      liftU1 (addContext (domFromArg a :: Dom Type)) m
 
     tcExtendContext :: Term -> Term -> UnquoteM Term
     tcExtendContext a m = do
@@ -654,11 +657,11 @@ evalTCM v = do
     tcInContext :: Term -> Term -> UnquoteM Term
     tcInContext c m = do
       c <- unquote c
-      liftU1 inTopContext $ go c m
+      liftU1 inTopContext $ go c (evalTCM m)
       where
-        go :: [Arg R.Type] -> Term -> UnquoteM Term
-        go []       m = evalTCM m
-        go (a : as) m = extendCxt a $ go as m
+        go :: [Arg R.Type] -> UnquoteM Term -> UnquoteM Term
+        go []       m = m
+        go (a : as) m = go as (extendCxt a m)
 
     constInfo :: QName -> TCM Definition
     constInfo x = either err return =<< getConstInfo' x
@@ -700,6 +703,28 @@ evalTCM v = do
         when (isInstance i) $ addTypedInstance x a
         primUnitUnit
 
+    tcDeclarePostulate :: Arg QName -> R.Type -> UnquoteM Term
+    tcDeclarePostulate (Arg i x) a = inOriginalContext $ do
+      clo <- commandLineOptions
+      when (Lens.getSafeMode clo) $ liftU $ typeError . GenericDocError =<<
+        text "Cannot postulate '" <+> prettyTCM x <+> text ":" <+> prettyTCM a <+> text "' with safe flag"
+      setDirty
+      let r = getRelevance i
+      when (hidden i) $ liftU $ typeError . GenericDocError =<<
+        text "Cannot declare hidden function" <+> prettyTCM x
+      tell [x]
+      liftU $ do
+        reportSDoc "tc.unquote.decl" 10 $ sep
+          [ text "declare Postulate" <+> prettyTCM x <+> text ":"
+          , nest 2 $ prettyTCM a
+          ]
+        a <- isType_ =<< toAbstract_ a
+        alreadyDefined <- isRight <$> getConstInfo' x
+        when alreadyDefined $ genericError $ "Multiple declarations of " ++ prettyShow x
+        addConstant x $ defaultDefn i x a Axiom
+        when (isInstance i) $ addTypedInstance x a
+        primUnitUnit
+
     tcDefineFun :: QName -> [R.Clause] -> UnquoteM Term
     tcDefineFun x cs = inOriginalContext $ (setDirty >>) $ liftU $ do
       whenM (isLeft <$> getConstInfo' x) $
