@@ -15,9 +15,11 @@ import Data.Monoid
 import Data.Traversable (traverse)
 
 import Agda.Syntax.Common
+import Agda.Syntax.Concrete.Pattern (IsWithP(..))
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Internal.Pattern
 import qualified Agda.Syntax.Abstract as A
+import Agda.Syntax.Abstract.Pattern as A
 import Agda.Syntax.Abstract.Views
 import Agda.Syntax.Info
 import Agda.Syntax.Position
@@ -38,12 +40,13 @@ import Agda.TypeChecking.Rules.Term
 
 import Agda.TypeChecking.Abstract
 import Agda.TypeChecking.Rules.LHS.Implicit
-import Agda.TypeChecking.Rules.LHS (isFlexiblePattern)
+import Agda.TypeChecking.Rules.LHS.Problem (ProblemEq(..))
 
 import Agda.Utils.Functor
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.NonemptyList
 import Agda.Utils.Null (empty)
 import Agda.Utils.Permutation
 import Agda.Utils.Pretty (prettyShow)
@@ -194,7 +197,6 @@ withArguments vs as = concat $ for (zip vs as) $ \case
   (v, OtherType a) -> [v]
   (prf, eqt@(EqualityType s _eq _pars _t v _v')) -> [unArg v, prf]
 
-
 -- | Compute the clauses for the with-function given the original patterns.
 buildWithFunction
   :: [Name]               -- ^ Names of the module parameters of the parent function.
@@ -213,78 +215,81 @@ buildWithFunction
 buildWithFunction cxtNames f aux t delta qs npars withSub perm n1 n cs = mapM buildWithClause cs
   where
     -- Nested with-functions will iterate this function once for each parent clause.
-    buildWithClause (A.Clause (A.SpineLHS i _ ps wps) inheritedDots inhStrippedDots rhs wh catchall) = do
-      let (wps0, wps1) = splitAt n wps
-          ps0          = map defaultNamedArg wps0
-      reportSDoc "tc.with" 50 $ text "inheritedDots:" <+> vcat [ prettyTCM x <+> text "=" <+> prettyTCM v <+> text ":" <+> prettyTCM a
-                                                               | A.NamedDot x v a <- inheritedDots ]
-      rhs <- buildRHS rhs
-      (namedDots, strippedDots, ps') <- stripWithClausePatterns cxtNames f aux t delta qs npars perm ps
-      reportSDoc "tc.with" 50 $ hang (text "strippedDots:") 2 $
-                                  vcat [ prettyTCM e <+> text "==" <+> prettyTCM v <+> (text ":" <+> prettyTCM t)
-                                       | A.StrippedDot e v t <- strippedDots ]
+    buildWithClause (A.Clause (A.SpineLHS i _ allPs) inheritedPats rhs wh catchall) = do
+      let (ps, wps)    = splitOffTrailingWithPatterns allPs
+          (wps0, wps1) = splitAt n wps
+          ps0          = map (updateNamedArg fromWithP) wps0
+            where
+            fromWithP (A.WithP _ p) = p
+            fromWithP _ = __IMPOSSIBLE__
+      reportSDoc "tc.with" 50 $ text "inheritedPats:" <+> vcat [ prettyA p <+> text "=" <+> prettyTCM v <+> text ":" <+> prettyTCM a
+                                                               | A.ProblemEq p v a <- inheritedPats ]
+      (strippedPats, ps') <- stripWithClausePatterns cxtNames f aux t delta qs npars perm ps
+      reportSDoc "tc.with" 50 $ hang (text "strippedPats:") 2 $
+                                  vcat [ prettyA p <+> text "==" <+> prettyTCM v <+> (text ":" <+> prettyTCM t)
+                                       | A.ProblemEq p v t <- strippedPats ]
+      rhs <- buildRHS strippedPats rhs
       let (ps1, ps2) = splitAt n1 ps'
-      let result = A.Clause (A.SpineLHS i aux (ps1 ++ ps0 ++ ps2) wps1) (inheritedDots ++ namedDots)
-                                                                        (inhStrippedDots ++ strippedDots) rhs wh catchall
+      let result = A.Clause (A.SpineLHS i aux $ ps1 ++ ps0 ++ ps2 ++ wps1)
+                     (inheritedPats ++ strippedPats)
+                     rhs wh catchall
       reportSDoc "tc.with" 20 $ vcat
         [ text "buildWithClause returns" <+> prettyA result
         ]
       return result
 
-    buildRHS rhs@A.RHS{}                 = return rhs
-    buildRHS rhs@A.AbsurdRHS             = return rhs
-    buildRHS (A.WithRHS q es cs)         = A.WithRHS q es <$>
+    buildRHS _ rhs@A.RHS{}                 = return rhs
+    buildRHS _ rhs@A.AbsurdRHS             = return rhs
+    buildRHS _ (A.WithRHS q es cs)         = A.WithRHS q es <$>
       mapM ((A.spineToLhs . permuteNamedDots) <.> buildWithClause . A.lhsToSpine) cs
-    buildRHS (A.RewriteRHS qes rhs wh) = flip (A.RewriteRHS qes) wh <$> buildRHS rhs
+    buildRHS strippedPats1 (A.RewriteRHS qes strippedPats2 rhs wh) =
+      flip (A.RewriteRHS qes (applySubst withSub $ strippedPats1 ++ strippedPats2)) wh <$> buildRHS [] rhs
 
-    -- The named dot patterns computed by buildWithClause lives in the context
+    -- The stripped patterns computed by buildWithClause lives in the context
     -- of the top with-clause (of the current call to buildWithFunction). When
-    -- we recurse we expect inherited named dot patterns to live in the context
+    -- we recurse we expect inherited patterns to live in the context
     -- of the innermost parent clause. Note that this makes them live in the
     -- context of the with-function arguments before any pattern matching. We
     -- need to update again once the with-clause patterns have been checked.
-    -- This happens in Rules.Def.checkClause before calling checkRHS. The same
-    -- goes for stripped dots.
+    -- This happens in Rules.Def.checkClause before calling checkRHS.
     permuteNamedDots :: A.SpineClause -> A.SpineClause
-    permuteNamedDots (A.Clause lhs dots sdots rhs wh catchall) =
-      A.Clause lhs (applySubst withSub dots) (applySubst withSub sdots) rhs wh catchall
+    permuteNamedDots (A.Clause lhs strippedPats rhs wh catchall) =
+      A.Clause lhs (applySubst withSub strippedPats) rhs wh catchall
+
+
+-- The arguments of @stripWithClausePatterns@ are documented
+-- at its type signature.
+-- The following is duplicate information, but may help reading the examples below.
+--
+-- [@Δ@]   context bound by lhs of original function.
+-- [@f@]   name of @with@-function.
+-- [@t@]   type of the original function.
+-- [@qs@]  internal patterns for original function.
+-- [@np@]  number of module parameters in @qs@
+-- [@π@]   permutation taking @vars(qs)@ to @support(Δ)@.
+-- [@ps@]  patterns in with clause (eliminating type @t@).
+-- [@ps'@] patterns for with function (presumably of type @Δ@).
 
 {-| @stripWithClausePatterns cxtNames parent f t Δ qs np π ps = ps'@
 
-[@Δ@]   context bound by lhs of original function.
-
-[@f@]   name of @with@-function.
-
-[@t@]   type of the original function.
-
-[@qs@]  internal patterns for original function.
-
-[@np@]  number of module parameters in @qs@
-
-[@π@]   permutation taking @vars(qs)@ to @support(Δ)@.
-
-[@ps@]  patterns in with clause (eliminating type @t@).
-
-[@ps'@] patterns for with function (presumably of type @Δ@).
-
 Example:
 
 @
-record Stream (A : Set) : Set where
-  coinductive
-  constructor delay
-  field       force : A × Stream A
+  record Stream (A : Set) : Set where
+    coinductive
+    constructor delay
+    field       force : A × Stream A
 
-record SEq (s t : Stream A) : Set where
-  coinductive
-  field
-    ~force : let a , as = force s
-                 b , bs = force t
-             in  a ≡ b × SEq as bs
+  record SEq (s t : Stream A) : Set where
+    coinductive
+    field
+      ~force : let a , as = force s
+                   b , bs = force t
+               in  a ≡ b × SEq as bs
 
-test : (s : Nat × Stream Nat) (t : Stream Nat) → SEq (delay s) t → SEq t (delay s)
-~force (test (a     , as) t p) with force t
-~force (test (suc n , as) t p) | b , bs = {!!}
+  test : (s : Nat × Stream Nat) (t : Stream Nat) → SEq (delay s) t → SEq t (delay s)
+  ~force (test (a     , as) t p) with force t
+  ~force (test (suc n , as) t p) | b , bs = {!!}
 @
 
 With function:
@@ -306,27 +311,27 @@ Resulting with-function clause is:
   f t (b , bs) (suc n) as t p
 @
 
-Note: stripWithClausePatterns factors @ps@ through @qs@, thus
+Note: stripWithClausePatterns factors __@ps@__ through __@qs@__, thus
 
 @
   ps = qs[ps']
 @
 
 where @[..]@ is to be understood as substitution.
-The projection patterns have vanished from @ps'@ (as they are already in @qs@).
+The projection patterns have vanished from __@ps'@__ (as they are already in __@qs@__).
 -}
 
 stripWithClausePatterns
-  :: [Name]                   -- ^ Names of the module parameters of the parent function
-  -> QName                    -- ^ Name of the parent function.
-  -> QName                    -- ^ Name of with-function.
+  :: [Name]                   -- ^ __@cxtNames@__ names of the module parameters of the parent function
+  -> QName                    -- ^ __@parent@__ name of the parent function.
+  -> QName                    -- ^ __@f@__   name of with-function.
   -> Type                     -- ^ __@t@__   top-level type of the original function.
   -> Telescope                -- ^ __@Δ@__   context of patterns of parent function.
   -> [NamedArg DeBruijnPattern] -- ^ __@qs@__  internal patterns for original function.
-  -> Nat                      -- ^ __@npars@__ number of module parameters is @qs@.
+  -> Nat                      -- ^ __@npars@__ number of module parameters in @qs@.
   -> Permutation              -- ^ __@π@__   permutation taking @vars(qs)@ to @support(Δ)@.
   -> [NamedArg A.Pattern]     -- ^ __@ps@__  patterns in with clause (eliminating type @t@).
-  -> TCM ([A.NamedDotPattern], [A.StrippedDotPattern], [NamedArg A.Pattern]) -- ^ __@ps'@__ patterns for with function (presumably of type @Δ@).
+  -> TCM ([A.ProblemEq], [NamedArg A.Pattern]) -- ^ __@ps'@__ patterns for with function (presumably of type @Δ@).
 stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
   -- Andreas, 2014-03-05 expand away pattern synoyms (issue 1074)
   ps <- expandPatternSynonyms ps
@@ -334,7 +339,7 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
   -- instantiations from qs, so we make sure
   -- that t is the top-level type of the parent function and add patterns for
   -- the module parameters to ps before stripping.
-  let paramPat i _ = A.VarP (cxtNames !! i)
+  let paramPat i _ = A.VarP $ A.BindName (cxtNames !! i)
       ps' = zipWith (fmap . fmap . paramPat) [0..] (take npars qs) ++ ps
   psi <- insertImplicitPatternsT ExpandLast ps' t
   reportSDoc "tc.with.strip" 10 $ vcat
@@ -348,31 +353,15 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
     ]
 
   -- Andreas, 2015-11-09 Issue 1710: self starts with parent-function, not with-function!
-  (ps', out) <- runWriterT $ strip (Def parent []) t psi qs
-  let (strippedDots, namedDots) = partitionEithers out
+  (ps', strippedPats) <- runWriterT $ strip (Def parent []) t psi qs
   reportSDoc "tc.with.strip" 50 $ nest 2 $
-    text "namedDots:" <+> vcat [ prettyTCM x <+> text "=" <+> prettyTCM v <+> text ":" <+> prettyTCM a | A.NamedDot x v a <- namedDots ]
+    text "strippedPats:" <+> vcat [ prettyA p <+> text "=" <+> prettyTCM v <+> text ":" <+> prettyTCM a | A.ProblemEq p v a <- strippedPats ]
   let psp = permute perm ps'
   reportSDoc "tc.with.strip" 10 $ vcat
     [ nest 2 $ text "ps' = " <+> fsep (punctuate comma $ map prettyA ps')
     , nest 2 $ text "psp = " <+> fsep (punctuate comma $ map prettyA $ psp)
     ]
-  -- Andreas, 2014-03-05 Issue 142:
-  -- In some cases, permute throws away some dot patterns of ps'
-  -- which are then never checked.
-  if True then return (namedDots, strippedDots, psp) else do
-    -- Andreas, 2014-03-05 Disabled the fix for issue 142, the following is dead code:
-    forM_ (permute (droppedP perm) ps') $ \ p -> setCurrentRange p $ do
-      reportSDoc "tc.with.strip" 10 $ text "warning: dropped pattern " <+> prettyA p
-      reportSDoc "tc.with.strip" 60 $ text $ show p
-      case namedArg p of
-        A.DotP info o e -> case unScope e of
-          A.Underscore{} -> return ()
-          e | o == UserWritten -> typeError $ GenericError $
-            "This inaccessible pattern is never checked, so only _ allowed here"
-          _ -> return ()
-        _ -> return ()
-    return (namedDots, strippedDots, psp)
+  return (strippedPats, psp)
   where
 
     strip
@@ -380,7 +369,7 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
       -> Type                         -- ^ The type to be eliminated.
       -> [NamedArg A.Pattern]       -- ^ With-clause patterns.
       -> [NamedArg DeBruijnPattern] -- ^ Parent-clause patterns with de Bruijn indices relative to Δ.
-      -> WriterT [Either A.StrippedDotPattern A.NamedDotPattern] TCM [NamedArg A.Pattern]
+      -> WriterT [ProblemEq] TCM [NamedArg A.Pattern]
             -- ^ With-clause patterns decomposed by parent-clause patterns.
             --   Also outputs named dot patterns from the parent clause that
             --   we need to add let-bindings for.
@@ -418,10 +407,10 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
       | A.AsP _ x p <- namedArg p0 = do
         (a, _) <- mustBePi t
         let v = patternToTerm (namedArg q)
-        tell [Right $ A.NamedDot x v (unDom a)]
+        tell [ProblemEq (A.VarP x) v a]
         strip self t (fmap (p <$) p0 : ps) qs
     strip self t ps0@(p0 : ps) qs0@(q : qs) = do
-      p <- liftTCM $ expandLitPattern p0
+      p <- liftTCM $ (traverse . traverse) expandLitPattern p0
       reportSDoc "tc.with.strip" 15 $ vcat
         [ text "strip"
         , nest 2 $ text "ps0 =" <+> fsep (punctuate comma $ map prettyA ps0)
@@ -430,13 +419,6 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
         , nest 2 $ text "self=" <+> prettyTCM self
         , nest 2 $ text "t   =" <+> prettyTCM t
         ]
-      let failDotPat :: Monoid w => WriterT w TCM a
-          failDotPat = do
-            d <- liftTCM $ prettyA p
-            typeError $ GenericError $
-                "Inaccessible (dotted) patterns from the parent clause must " ++
-                "also be inaccessible in the with clause, when checking the " ++
-                "pattern " ++ show d ++ ","
       case namedArg q of
         ProjP o d -> case A.maybePostfixProjP p of
           Just (o', AmbQ ds) -> do
@@ -459,48 +441,27 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
               strip self1 t1 ps qs
           Nothing -> mismatch
 
-        VarP x  -> (p :) <$> recurse (var (dbPatVarIndex x))
+        -- If a variable pattern in the parent clause was written as a dot
+        -- pattern by the user both in the parent clause and the with clause,
+        -- we can strip the dot from the with clause.
+        VarP PatODot x | A.DotP _ u <- namedArg p
+                       , A.Var y <- unScope u ->
+          (setNamedArg p (A.VarP (A.BindName y)) :) <$>
+            recurse (var (dbPatVarIndex x))
 
-        AbsurdP p -> __IMPOSSIBLE__
+        VarP _ x  -> (p :) <$> recurse (var (dbPatVarIndex x))
 
-        DotP v  -> case namedArg p of
-          A.DotP r o e  -> do
-            (a, _) <- mustBePi t
-            tell [Left $ A.StrippedDot e v (unDom a)]
-            ok p
-          A.WildP _     -> ok p
-          -- Ulf, 2016-05-30: dot patterns are no longer mandatory so a parent
-          -- dot pattern can appear as a variable in the child clause. Indeed
-          -- this happens if you use a variable in the parent and '...' in the
-          -- child. In this case we need to remember the the binding, so we can
-          -- insert a let for it.
-          A.VarP x -> do
-            (a, _) <- mustBePi t
-            tell [Right $ A.NamedDot x v (unDom a)]
-            ok p
-          -- Andreas, 2013-03-21 in case the implicit A.pattern has already been eta-expanded
-          -- we just fold it back.  This fixes issues 665 and 824.
-          A.ConP ci _ _ | patOrigin ci == ConOSystem -> okFlex p
-          -- Andreas, 2015-07-07 issue 1606: Same for flexible record patterns.
-          -- Agda might have replaced a record of dot patterns (A.ConP) by a dot pattern (I.DotP).
-          p'@A.ConP{} -> ifM (liftTCM $ isFlexiblePattern p') (okFlex p) {-else-} failDotPat
-
-          p@(A.PatternSynP pi' c' [ps']) -> do
-             reportSDoc "impossible" 10 $
-               text "stripWithClausePatterns: encountered pattern synonym " <+> prettyA p
-             __IMPOSSIBLE__
-
-          _ -> failDotPat
-          where
-            okFlex = ok . makeImplicitP
-            ok p   = (p :) <$> recurse v
+        DotP o v  -> do
+          (a, _) <- mustBePi t
+          tell [ProblemEq (namedArg p) v a]
+          (makeImplicitP p :) <$> recurse v
 
         q'@(ConP c ci qs') -> do
          reportSDoc "tc.with.strip" 60 $
            text "parent pattern is constructor " <+> prettyTCM c
          (a, b) <- mustBePi t
          -- The type of the current pattern is a datatype.
-         Def d es <- liftTCM $ ignoreSharing <$> normalise (unEl $ unDom a)
+         Def d es <- liftTCM $ normalise (unEl $ unDom a)
          let us = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
          -- Get the original constructor and field names.
          c <- either __IMPOSSIBLE__ (`withRangeOf` c) <$> do liftTCM $ getConForm $ conName c
@@ -510,21 +471,35 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
           -- Andreas, 2015-07-07 Issue 1606.
           -- Agda sometimes changes a record of dot patterns into a dot pattern,
           -- so the user should be allowed to do likewise.
-          A.DotP{} -> ifNotM (liftTCM $ isFlexiblePattern q') mismatch $ {-else-} do
-            maybe __IMPOSSIBLE__ (\ p -> strip self t (p : ps) qs0) =<< do
-              liftTCM $ expandImplicitPattern' (unDom a) $ makeImplicitP p
+          -- Jesper, 2017-11-16. This is now also allowed for data constructors.
+          A.DotP r e -> do
+            tell [ProblemEq (A.DotP r e) (patternToTerm q') a]
+            let ps' = map (unnamed (A.WildP empty) <$) qs'
+            stripConP d us b c ConOCon qs' ps'
 
           -- Andreas, 2016-12-29, issue #2363.
           -- Allow _ to stand for the corresponding parent pattern.
           A.WildP{} -> do
-            let ps' = map (updateNamedArg $ const $ A.WildP empty) qs'
+            -- Andreas, 2017-10-13, issue #2803:
+            -- Delete the name, since it can confuse insertImplicitPattern.
+            let ps' = map (unnamed (A.WildP empty) <$) qs'
+            stripConP d us b c ConOCon qs' ps'
+
+          -- Jesper, 2018-05-13, issue #2998.
+          -- We also allow turning a constructor pattern into a variable.
+          -- In general this is not type-safe since the types of some variables
+          -- in the constructor pattern may have changed, so we have to
+          -- re-check these solutions when checking the with clause (see LHS.hs)
+          A.VarP x -> do
+            tell [ProblemEq (A.VarP x) (patternToTerm q') a]
+            let ps' = map (unnamed (A.WildP empty) <$) qs'
             stripConP d us b c ConOCon qs' ps'
 
           A.ConP _ (A.AmbQ cs') ps' -> do
             -- Check whether the with-clause constructor can be (possibly trivially)
             -- disambiguated to be equal to the parent-clause constructor.
             -- Andreas, 2017-08-13, herein, ignore abstract constructors.
-            cs' <- liftTCM $ do snd . partitionEithers <$> mapM getConForm cs'
+            cs' <- liftTCM $ do snd . partitionEithers <$> mapM getConForm (toList cs')
             unless (elem c cs') mismatch
             -- Strip the subpatterns ps' and then continue.
             stripConP d us b c ConOCon qs' ps'
@@ -592,7 +567,7 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
              -- ^ Argument patterns (parent clause).
           -> [NamedArg A.Pattern]
              -- ^ Argument patterns (with clause).
-          -> WriterT [Either A.StrippedDotPattern A.NamedDotPattern] TCM [NamedArg A.Pattern]
+          -> WriterT [ProblemEq] TCM [NamedArg A.Pattern]
              -- ^ Stripped patterns.
         stripConP d us b c ci qs' ps' = do
 
@@ -611,7 +586,7 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
                  ]
 
           -- Compute the new type
-          let v  = Con c ci [ Arg info (var i) | (i, Arg info _) <- zip (downFrom $ size qs') qs' ]
+          let v  = Con c ci [ Apply $ Arg info (var i) | (i, Arg info _) <- zip (downFrom $ size qs') qs' ]
               t' = tel' `abstract` absApp (raise (size tel') b) v
               self' = tel' `abstract` apply1 (raise (size tel') self) v  -- Issue 1546
 
@@ -728,18 +703,6 @@ withDisplayForm f aux delta1 delta2 n qs perm@(Perm m _) lhsPerm = do
     sub top ys wild = parallelS $ map term [0 .. m + top - 1]
       where
         term i = maybe wild var $ List.findIndex (Just i ==) ys
-    -- -- OLD
-    -- sub top rho wild = parallelS $ map term [0 .. m - 1] ++ topTerms
-    --   where
-    --     -- Ulf, 2014-02-19: We need to rename the module parameters as well! (issue1035)
-    --     newVars  = length qs
-    --     topTerms = [ var (i + newVars) | i <- [0..top - 1] ]
-    --     -- thinking required.. but ignored
-    --     -- dropping the reverse seems to work better
-    --     -- Andreas, 2010-09-09: I DISAGREE.
-    --     -- Ulf, 2011-09-02: Thinking done. Neither was correct.
-    --     -- We had the wrong permutation and we used it incorrectly. Should work now.
-    --     term i = maybe wild var $ findIndex (Just i ==) rho
 
 -- Andreas, 2014-12-05 refactored using numberPatVars
 -- Andreas, 2013-02-28 modeled after Coverage/Match/buildMPatterns
@@ -757,8 +720,9 @@ patsToElims = map $ toElim . fmap namedThing
     toTerm :: DeBruijnPattern -> DisplayTerm
     toTerm p = case p of
       ProjP _ d   -> DDef d [] -- WRONG. TODO: convert spine to non-spine ... DDef d . defaultArg
-      VarP x      -> DTerm  $ var $ dbPatVarIndex x
-      DotP t      -> DDot   $ t
-      AbsurdP p   -> toTerm p
+      VarP PatODot x -> DDot  $ var $ dbPatVarIndex x
+      VarP o x      -> DTerm  $ var $ dbPatVarIndex x
+      DotP PatOVar{} t@(Var i []) -> DTerm t
+      DotP o t    -> DDot   $ t
       ConP c cpi ps -> DCon c (fromConPatternInfo cpi) $ toTerms ps
       LitP l      -> DTerm  $ Lit l
