@@ -1,10 +1,6 @@
 {-# LANGUAGE CPP                #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 
-#if __GLASGOW_HASKELL__ <= 708
-{-# LANGUAGE DeriveDataTypeable #-}
-#endif
-
 -- | An empty type with some useful instances.
 module Agda.Utils.Empty where
 
@@ -13,18 +9,11 @@ import Control.Exception (evaluate)
 import Data.Functor ((<$))
 import Data.Data (Data)
 
-#if __GLASGOW_HASKELL__ <= 708
-import Data.Typeable ( Typeable )
-#endif
-
 import Agda.Utils.Impossible
 
 #include "undefined.h"
 
 data Empty
-#if __GLASGOW_HASKELL__ <= 708
-  deriving Typeable
-#endif
 deriving instance Data Empty
 
 instance Eq Empty where
