@@ -6,44 +6,50 @@
 --
 --   Represented as adjacency maps in direction from source to target.
 --
---   Each source node maps to a adjacency map of outgoing edges,
+--   Each source node maps to an adjacency map of outgoing edges,
 --   which is a map from target nodes to edges.
 --
---   This allows to get outgoing edges in O(log n) time where
---   @n@ is the number of nodes in the graph.
---
---   However, the set of incoming edges can only be obtained in
---   @O(n log n)@ or @O(e)@ where @e@ is the total number of edges.
+--   Listed time complexities are for the worst case (and possibly
+--   amortised), with /n/ standing for the number of nodes in the
+--   graph and /e/ standing for the number of edges. Comparisons,
+--   predicates etc. are assumed to take constant time (unless
+--   otherwise stated).
 
 module Agda.Utils.Graph.AdjacencyMap.Unidirectional
-  ( Graph(..)
+  ( -- * Graphs and edges
+    Graph(..)
+  , invariant
   , Edge(..)
-  , transposeEdge
+    -- * Queries
+  , lookup
   , edges
+  , neighbours, neighboursMap
   , edgesFrom
   , edgesTo
   , diagonal
-  , lookup
-  , neighbours, neighboursMap
-  , sourceNodes, targetNodes
-  , Nodes(..)
-  , computeNodes, nodes
-  , fromNodes
-  , fromList, fromListWith
-  , toList
+  , nodes, sourceNodes, targetNodes, isolatedNodes
+  , Nodes(..), computeNodes
   , discrete
-  , clean
+  , acyclic
+    -- * Construction
+  , fromNodes, fromNodeSet
+  , fromEdges, fromEdgesWith
   , empty
   , singleton
   , insert, insertWith
   , insertEdge, insertEdgeWith
-  , union , unionWith
+  , union, unionWith
   , unions, unionsWith
-  , removeNode
+    -- * Transformation
+  , mapWithEdge
+  , transposeEdge, transpose
+  , clean
+  , removeNode, removeNodes
   , removeEdge
   , filterEdges
   , unzip
-  , mapWithEdge
+  , composeWith
+    -- * Strongly connected components
   , sccs'
   , sccs
   , DAG(..)
@@ -52,17 +58,21 @@ module Agda.Utils.Graph.AdjacencyMap.Unidirectional
   , reachable
   , sccDAG'
   , sccDAG
-  , acyclic
-  , reachableFrom
+    -- * Reachability
+  , reachableFrom, reachableFromSet
   , walkSatisfying
-  , composeWith
-  , complete
-  , gaussJordanFloydWarshallMcNaughtonYamadaReference
+    -- * Transitive closure
   , gaussJordanFloydWarshallMcNaughtonYamada
+  , gaussJordanFloydWarshallMcNaughtonYamadaReference
+  , complete, completeIter
   )
   where
 
-import Prelude hiding (lookup, unzip, null)
+#if MIN_VERSION_base(4,11,0)
+import Prelude hiding ( (<>), lookup, null, unzip )
+#else
+import Prelude hiding ( lookup, null, unzip )
+#endif
 
 import Control.Applicative hiding (empty)
 import Control.Monad
@@ -98,60 +108,109 @@ import Agda.Utils.Tuple
 #include "undefined.h"
 import Agda.Utils.Impossible
 
--- | @Graph s t e@ is a directed graph with
---   source nodes in @s@
---   target nodes in @t@
---   and edges in @e@.
---
---   Admits at most one edge between any two nodes.
---   Several edges can be modeled by using a collection type for @e@.
---
---   Represented as "adjacency list", or rather, adjacency map.
---   This allows to get all outgoing edges for a node
---   in @O(log n)@ time where @n@ is the number of nodes of the graph.
---
---   Incoming edges can only be computed in @O(n + e)@ time where
---   @e@ is the number of edges.
+------------------------------------------------------------------------
+-- Graphs and edges
 
-newtype Graph s t e = Graph
-  { graph :: Map s (Map t e) -- ^ Forward edges.
+-- | @Graph n e@ is a type of directed graphs with nodes in @n@ and
+--   edges in @e@.
+--
+--   At most one edge is allowed between any two nodes. Multigraphs
+--   can be simulated by letting the edge type @e@ be a collection
+--   type.
+--
+--   The graphs are represented as adjacency maps (adjacency lists,
+--   but using finite maps instead of arrays and lists). This makes it
+--   possible to compute a node's outgoing edges in logarithmic time
+--   (/O(log n)/). However, computing the incoming edges may be more
+--   expensive.
+--
+--   Note that neither the number of nodes nor the number of edges may
+--   exceed @'maxBound' :: 'Int'@.
+
+newtype Graph n e = Graph
+  { graph :: Map n (Map n e) -- ^ Forward edges.
   }
-  deriving (Eq, Functor, Show)
+  deriving Eq
 
-instance (Pretty s, Pretty t, Pretty e) => Pretty (Graph s t e) where
-  pretty = vcat . map pretty . edges
+-- The Functor instance for strict maps is the one for lazy maps, so a
+-- custom Functor instance using strict map functions is used here.
 
-data Edge s t e = Edge
-  { source :: s  -- ^ Outgoing node.
-  , target :: t  -- ^ Incoming node.
+instance Functor (Graph n) where
+  fmap f = Graph . Map.map (Map.map f) . graph
+
+-- | Internal invariant.
+
+invariant :: Ord n => Graph n e -> Bool
+invariant g =
+  -- Every target node must be present in the graph as a source node,
+  -- possibly without outgoing edges.
+  Set.isSubsetOf (targetNodes g) (nodes g)
+
+instance (Ord n, Pretty n, Pretty e) => Pretty (Graph n e) where
+  pretty g = vcat (concat $ map pretty' (Set.toAscList (nodes g)))
+    where
+    pretty' n = case edgesFrom g [n] of
+      [] -> [pretty n]
+      es -> map pretty es
+
+instance (Ord n, Show n, Show e) => Show (Graph n e) where
+  showsPrec _ g =
+    showString "union (fromEdges " .
+    shows (edges g) .
+    showString ") (fromNodes " .
+    shows (Set.toList (isolatedNodes g)) .
+    showString ")"
+
+-- | Edges.
+
+data Edge n e = Edge
+  { source :: n  -- ^ Outgoing node.
+  , target :: n  -- ^ Incoming node.
   , label  :: e  -- ^ Edge label (weight).
   } deriving (Eq, Ord, Functor, Show)
 
-instance (Pretty s, Pretty t, Pretty e) => Pretty (Edge s t e) where
-  pretty (Edge s t e) = pretty s <+> text "--(" <> pretty t <> text ")-->" <+> pretty t
+instance (Pretty n, Pretty e) => Pretty (Edge n e) where
+  pretty (Edge s t e) =
+    pretty s <+> text "--(" <> pretty e <> text ")-->" <+> pretty t
 
--- | Reverse an edge.
+------------------------------------------------------------------------
+-- Queries
 
-transposeEdge :: Edge s t e -> Edge t s e
-transposeEdge (Edge s t e) = Edge t s e
+-- | If there is an edge from @s@ to @t@, then @lookup s t g@ is
+-- @'Just' e@, where @e@ is the edge's label. /O(log n)/.
 
--- * Edge queries
+lookup :: Ord n => n -> n -> Graph n e -> Maybe e
+lookup s t (Graph g) = Map.lookup t =<< Map.lookup s g
 
--- | Turn a graph into a list of edges.  @O(n + e)@
+-- | The graph's edges. /O(n + e)/.
 
-edges :: Graph s t e -> [Edge s t e]
+edges :: Graph n e -> [Edge n e]
 edges (Graph g) =
   [ Edge s t e
   | (s, tes) <- Map.assocs g
   , (t, e)   <- Map.assocs tes
   ]
 
--- | All edges originating in the given nodes.
---   (I.e., all outgoing edges for the given nodes.)
---
---   Roughly linear in the length of the result list @O(result)@.
+-- | @neighbours u g@ consists of all nodes @v@ for which there is an
+-- edge from @u@ to @v@ in @g@, along with the corresponding edge
+-- labels. /O(log n + |@neighbours u g@|)/.
 
-edgesFrom :: Ord s => Graph s t e -> [s] -> [Edge s t e]
+neighbours :: Ord n => n -> Graph n e -> [(n, e)]
+neighbours s = Map.toList . neighboursMap s
+
+-- | @neighboursMap u g@ consists of all nodes @v@ for which there is
+-- an edge from @u@ to @v@ in @g@, along with the corresponding edge
+-- labels. /O(log n)/.
+
+neighboursMap :: Ord n => n -> Graph n e -> Map n e
+neighboursMap s (Graph g) = fromMaybe Map.empty $ Map.lookup s g
+
+-- | @edgesFrom g ns@ is a list containing all edges originating in
+-- the given nodes (i.e., all outgoing edges for the given nodes). If
+-- @ns@ does not contain duplicates, then the resulting list does not
+-- contain duplicates. /O(|@ns@| log |@n@| + |@edgesFrom g ns@|)/.
+
+edgesFrom :: Ord n => Graph n e -> [n] -> [Edge n e]
 edgesFrom (Graph g) ss =
   [ Edge s t e
   | s <- ss
@@ -159,13 +218,12 @@ edgesFrom (Graph g) ss =
   , (t, e) <- Map.assocs m
   ]
 
+-- | @edgesTo g ns@ is a list containing all edges ending in the given
+-- nodes (i.e., all incoming edges for the given nodes). If @ns@ does
+-- not contain duplicates, then the resulting list does not contain
+-- duplicates. /O(|@ns@| n log n)/.
 
--- | All edges ending in the given nodes.
---   (I.e., all incoming edges for the given nodes.)
---
---   Expensive: @O(n * |ts| * log n)@.
-
-edgesTo :: Ord t => Graph s t e -> [t] -> [Edge s t e]
+edgesTo :: Ord n => Graph n e -> [n] -> [Edge n e]
 edgesTo (Graph g) ts =
   [ Edge s t e
   | (s, m) <- Map.assocs g
@@ -173,219 +231,280 @@ edgesTo (Graph g) ts =
   , e <- maybeToList $ Map.lookup t m
   ]
 
--- | Get all self-loops.
+-- | All self-loops. /O(n log n)/.
 
-diagonal :: (Ord n) => Graph n n e -> [Edge n n e]
+diagonal :: Ord n => Graph n e -> [Edge n e]
 diagonal (Graph g) =
   [ Edge s s e
   | (s, m) <- Map.assocs g
   , e      <- maybeToList $ Map.lookup s m
   ]
 
--- | Lookup label of an edge.
+-- | All nodes. /O(n)/.
 
-lookup :: (Ord s, Ord t) => s -> t -> Graph s t e -> Maybe e
-lookup s t (Graph g) = Map.lookup t =<< Map.lookup s g
+nodes :: Graph n e -> Set n
+nodes = Map.keysSet . graph
 
--- | Get a list of outgoing edges with target.
+-- | Nodes with outgoing edges. /O(n)/.
 
-neighbours :: Ord s => s -> Graph s t e -> [(t, e)]
-neighbours s (Graph g) = maybe [] Map.assocs $ Map.lookup s g
+sourceNodes :: Graph n e -> Set n
+sourceNodes = Map.keysSet . Map.filter (not . Map.null) . graph
 
--- | Get a list of outgoing edges with target.
+-- | Nodes with incoming edges. /O(n + e log n)/.
 
-neighboursMap :: Ord s => s -> Graph s t e -> Map t e
-neighboursMap s (Graph g) = fromMaybe Map.empty $ Map.lookup s g
-
--- * Node queries
-
--- | Returns all the nodes with outgoing edges.  @O(n)@.
-
-sourceNodes :: Graph s t e -> Set s
-sourceNodes = Map.keysSet . graph
-
--- | Returns all the nodes with incoming edges.  Expensive! @O(e)@.
-
-targetNodes :: Ord t => Graph s t e -> Set t
+targetNodes :: Ord n => Graph n e -> Set n
 targetNodes = Set.fromList . map target . edges
 
--- | For homogeneous graphs, @(s = t)@ we can compute a set
---   of all nodes.
---
---   Structure @Nodes@ is for computing all nodes but also
---   remembering which were incoming and which outgoing.
---   This is mostly for efficiency reasons, to avoid recomputation
---   when all three sets are needed.
+-- | Various kinds of nodes.
 
 data Nodes n = Nodes
   { srcNodes :: Set n
+    -- ^ Nodes with outgoing edges.
   , tgtNodes :: Set n
+    -- ^ Nodes with incoming edges.
   , allNodes :: Set n
+    -- ^ All nodes, with or without edges.
   }
 
-computeNodes :: (Ord n) => Graph n n e -> Nodes n
-computeNodes g = Nodes srcs tgts (srcs `Set.union` tgts)
-  where srcs = sourceNodes g
-        tgts = targetNodes g
+-- | Constructs a 'Nodes' structure. /O(n + e log n)/.
 
--- | The set of all nodes (outgoing and incoming).
+computeNodes :: Ord n => Graph n e -> Nodes n
+computeNodes g =
+  Nodes { srcNodes = Set.filter (not . null . flip neighbours g) ns
+        , tgtNodes = targetNodes g
+        , allNodes = ns
+        }
+  where
+  ns = nodes g
 
-nodes :: (Ord n) => Graph n n e -> Set n
-nodes = allNodes . computeNodes
+-- | Nodes without incoming or outgoing edges. /O(n + e log n)/.
 
--- * Graph construction.
+isolatedNodes :: Ord n => Graph n e -> Set n
+isolatedNodes g =
+  Set.difference (allNodes ns) (Set.union (srcNodes ns) (tgtNodes ns))
+  where
+  ns = computeNodes g
 
--- | Constructs a completely disconnected graph containing the given
---   nodes. @O(n)@.
+-- | Checks whether the graph is discrete (containing no edges other
+-- than 'null' edges). /O(n + e)/.
 
-fromNodes :: Ord n => [n] -> Graph n n e
-fromNodes ns = Graph $ Map.fromList $ map (, Map.empty) ns
-
--- | Constructs a graph from a list of edges.  O(e log n)
---
---   Later edges overwrite earlier edges.
-
-fromList :: (Ord s, Ord t) => [Edge s t e] -> Graph s t e
-fromList = fromListWith $ \ new old -> new
-
--- | Constructs a graph from a list of edges.  O(e log n)
---
---   Later edges are combined with earlier edges using the supplied function.
-
-fromListWith :: (Ord s, Ord t) => (e -> e -> e) -> [Edge s t e] -> Graph s t e
-fromListWith f = List.foldl' (flip (insertEdgeWith f)) empty
-
--- | Convert a graph into a list of edges. O(e)
-
-toList :: Graph s t e -> [Edge s t e]
-toList (Graph g) = [ Edge s t a | (s,m) <- Map.assocs g, (t,a) <- Map.assocs m ]
-
--- | Check whether the graph is discrete (no edges).
---   This could be seen as an empty graph.
---   Worst-case (is discrete): @O(e)@.
-discrete :: Null e => Graph s t e -> Bool
+discrete :: Null e => Graph n e -> Bool
 discrete = all' (all' null) . graph
   where all' p = List.all p . Map.elems
 
--- | Removes 'Null' edges (and empty 'Map's).
-clean :: Null e => Graph s t e -> Graph s t e
-clean = Graph . filt . fmap filt . graph
+-- | Returns @True@ iff the graph is acyclic.
+
+acyclic :: Ord n => Graph n e -> Bool
+acyclic = all isAcyclic . sccs'
   where
-    filt :: Null a => Map k a -> Map k a
-    filt = Map.filter (not . null)
+  isAcyclic Graph.AcyclicSCC{} = True
+  isAcyclic Graph.CyclicSCC{}  = False
 
--- | Empty graph (no nodes, no edges).
+------------------------------------------------------------------------
+-- Construction
 
-empty :: Graph s t e
+-- | Constructs a completely disconnected graph containing the given
+--   nodes. /O(n log n)/.
+
+fromNodes :: Ord n => [n] -> Graph n e
+fromNodes ns = Graph $ Map.fromList $ map (, Map.empty) ns
+
+-- | Constructs a completely disconnected graph containing the given
+--   nodes. /O(n)/.
+
+fromNodeSet :: Ord n => Set n -> Graph n e
+fromNodeSet ns = Graph $ Map.fromSet (\_ -> Map.empty) ns
+
+-- | @fromEdges es@ is a graph containing the edges in @es@, with the
+-- caveat that later edges overwrite earlier edges. /O(|@es@| log n)/.
+
+fromEdges :: Ord n => [Edge n e] -> Graph n e
+fromEdges = fromEdgesWith $ \ new old -> new
+
+-- | @fromEdgesWith f es@ is a graph containing the edges in @es@.
+-- Later edges are combined with earlier edges using the supplied
+-- function. /O(|@es@| log n)/.
+
+fromEdgesWith :: Ord n => (e -> e -> e) -> [Edge n e] -> Graph n e
+fromEdgesWith f = List.foldl' (flip (insertEdgeWith f)) empty
+
+-- | Empty graph (no nodes, no edges). /O(1)/.
+
+empty :: Graph n e
 empty = Graph Map.empty
 
--- | A graph with two nodes and a single connecting edge.
+-- | A graph with two nodes and a single connecting edge. /O(1)/.
 
-singleton :: s -> t -> e -> Graph s t e
-singleton s t e = Graph $ Map.singleton s (Map.singleton t e)
+singleton :: Ord n => n -> n -> e -> Graph n e
+singleton s t e = insert s t e empty
 
--- | Insert an edge into the graph.
+-- | Inserts an edge into the graph. /O(log n)/.
 
-insert :: (Ord s, Ord t) => s -> t -> e -> Graph s t e -> Graph s t e
+insert :: Ord n => n -> n -> e -> Graph n e -> Graph n e
 insert = insertWith $ \ new old -> new
 
-insertEdge :: (Ord s, Ord t) => Edge s t e -> Graph s t e -> Graph s t e
+-- | Inserts an edge into the graph. /O(log n)/.
+
+insertEdge :: Ord n => Edge n e -> Graph n e -> Graph n e
 insertEdge (Edge s t e) = insert s t e
 
--- | Insert an edge, possibly combining @old@ edge weight with @new@ weight by
---   given function @f@ into @f new old@.
+-- | @insertWith f s t new@ inserts an edge from @s@ to @t@ into the
+-- graph. If there is already an edge from @s@ to @t@ with label @old@,
+-- then this edge gets replaced by an edge with label @f new old@, and
+-- otherwise the edge's label is @new@. /O(log n)/.
 
-insertWith :: (Ord s, Ord t) =>
-              (e -> e -> e) -> s -> t -> e -> Graph s t e -> Graph s t e
-insertWith f s t e (Graph g) = Graph (Map.alter (Just . ins) s g)
-  where ins Nothing  = Map.singleton t e
-        ins (Just m) = Map.insertWith f t e m
+insertWith ::
+  Ord n => (e -> e -> e) -> n -> n -> e -> Graph n e -> Graph n e
+insertWith f s t e (Graph g) =
+  Graph (Map.alter (Just . insNode) t $ Map.alter (Just . insEdge) s g)
+  where
+  insEdge Nothing  = Map.singleton t e
+  insEdge (Just m) = Map.insertWith f t e m
 
-insertEdgeWith :: (Ord s, Ord t) =>
-                  (e -> e -> e) -> Edge s t e -> Graph s t e -> Graph s t e
+  insNode Nothing  = Map.empty
+  insNode (Just m) = m
+
+-- | A variant of 'insertWith'. /O(log n)/.
+
+insertEdgeWith ::
+  Ord n => (e -> e -> e) -> Edge n e -> Graph n e -> Graph n e
 insertEdgeWith f (Edge s t e) = insertWith f s t e
 
 -- | Left-biased union.
+--
+-- Time complexity: See 'unionWith'.
 
-union :: (Ord s, Ord t) => Graph s t e -> Graph s t e -> Graph s t e
+union :: Ord n => Graph n e -> Graph n e -> Graph n e
 union = unionWith $ \ left right -> left
 
-unionWith :: (Ord s, Ord t) =>
-             (e -> e -> e) -> Graph s t e -> Graph s t e -> Graph s t e
-unionWith f (Graph g) (Graph g') = Graph $ Map.unionWith (Map.unionWith f) g g'
+-- | Union. The function is used to combine edge labels for edges that
+-- occur in both graphs (labels from the first graph are given as the
+-- first argument to the function).
+--
+-- Time complexity: /O(n₁ log (n₂/n₁ + 1) + e₁ log e₂/, where /n₁/ is
+-- the number of nodes in the graph with the smallest number of nodes
+-- and /n₂/ is the number of nodes in the other graph, and /e₁/ is the
+-- number of edges in the graph with the smallest number of edges and
+-- /e₂/ is the number of edges in the other graph.
+--
+-- Less complicated time complexity: /O((n + e) log n/ (where /n/ and
+-- /e/ refer to the resulting graph).
 
-unions ::(Ord s, Ord t) => [Graph s t e] -> Graph s t e
+unionWith ::
+  Ord n => (e -> e -> e) -> Graph n e -> Graph n e -> Graph n e
+unionWith f (Graph g) (Graph g') =
+  Graph $ Map.unionWith (Map.unionWith f) g g'
+
+-- | Union. /O((n + e) log n/ (where /n/ and /e/ refer to the
+-- resulting graph).
+
+unions :: Ord n => [Graph n e] -> Graph n e
 unions = unionsWith $ \ left right -> left
 
-unionsWith :: (Ord s, Ord t) => (e -> e -> e) -> [Graph s t e] -> Graph s t e
+-- | Union. The function is used to combine edge labels for edges that
+-- occur in several graphs. /O((n + e) log n/ (where /n/ and /e/ refer
+-- to the resulting graph).
+
+unionsWith :: Ord n => (e -> e -> e) -> [Graph n e] -> Graph n e
 unionsWith f = List.foldl' (unionWith f) empty
 
--- * Graph reversal
+------------------------------------------------------------------------
+-- Transformation
 
--- | The opposite graph (with all edges reversed).
+-- | A variant of 'fmap' that provides extra information to the
+-- function argument. /O(n + e)/.
 
-transpose :: (Ord s, Ord t) => Graph s t e -> Graph t s e
-transpose = fromList . map transposeEdge . edges
-
--- * Graph deconstruction.
-
--- | Auxiliary function to turn empty map into @Nothing@.
-
-discardEmpty :: Map k v -> Maybe (Map k v)
-discardEmpty m = if Map.null m then Nothing else Just m
-
--- | Removes the given source node, and all corresponding edges, from the graph.
---
---   O(log n).
-removeSourceNode :: Ord s => s -> Graph s t e -> Graph s t e
-removeSourceNode s (Graph g) = Graph $ Map.delete s g
-
--- | Removes the given target node, and all corresponding edges, from the graph.
---
---   Expensive!  @O(n log n)@.
-
-removeTargetNode :: Ord t => t -> Graph s t e -> Graph s t e
-removeTargetNode t (Graph g) = Graph $ Map.mapMaybe rem g
-  where rem = discardEmpty . Map.delete t
-
--- | Removes the given node, be it source or target,
---   and all corresponding edges, from the graph.
---
---   Expensive!  @O(n log n)@.
-
-removeNode :: Ord n => n -> Graph n n e -> Graph n n e
-removeNode n = removeTargetNode n . removeSourceNode n
-
--- | @removeEdge s t g@ removes the edge going from @s@ to @t@, if any.
---
---   @O((log n)^2)@.
-
-removeEdge :: (Ord s, Ord t) => s -> t -> Graph s t e -> Graph s t e
-removeEdge s t (Graph g) = Graph $ Map.adjust (Map.delete t) s g
-
--- | Keep only the edges that satisfy the predicate.  @O(e).@
-
-filterEdges :: (e -> Bool) -> Graph s t e -> Graph s t e
-filterEdges f (Graph g) = Graph $ Map.mapMaybe (discardEmpty . Map.filter f) g
-
--- | Unzipping a graph (naive implementation using fmap).
-
-unzip :: Graph s t (e, e') -> (Graph s t e, Graph s t e')
-unzip g = (fst <$> g, snd <$> g)
-
--- | Maps over a graph under availability of positional information,
---   like 'Map.mapWithKey'.
-
-mapWithEdge :: (Edge s t e -> e') -> Graph s t e -> Graph s t e'
+mapWithEdge :: (Edge n e -> e') -> Graph n e -> Graph n e'
 mapWithEdge f (Graph g) = Graph $ flip Map.mapWithKey g $ \ s m ->
   flip Map.mapWithKey m $ \ t e -> f (Edge s t e)
 
--- * Strongly connected components.
+-- | Reverses an edge. /O(1)/.
+
+transposeEdge :: Edge n e -> Edge n e
+transposeEdge (Edge s t e) = Edge t s e
+
+-- | The opposite graph (with all edges reversed). /O((n + e) log n)/.
+
+transpose :: Ord n => Graph n e -> Graph n e
+transpose g =
+  fromEdges (map transposeEdge (edges g))
+    `union`
+  fromNodeSet (isolatedNodes g)
+
+-- | Removes 'null' edges. /O(n + e)/.
+
+clean :: Null e => Graph n e -> Graph n e
+clean = Graph . Map.map (Map.filter (not . null)) . graph
+
+-- | @removeNodes ns g@ removes the nodes in @ns@ (and all
+-- corresponding edges) from @g@. /O((n + e) log |@ns@|)/.
+
+removeNodes :: Ord n => Set n -> Graph n e -> Graph n e
+removeNodes ns (Graph g) = Graph (Map.mapMaybeWithKey remSrc g)
+  where
+  remSrc s m
+    | Set.member s ns = Nothing
+    | otherwise       =
+        Just (Map.filterWithKey (\t _ -> not (Set.member t ns)) m)
+
+-- | @removeNode n g@ removes the node @n@ (and all corresponding
+-- edges) from @g@. /O(n + e)/.
+
+removeNode :: Ord n => n -> Graph n e -> Graph n e
+removeNode = removeNodes . Set.singleton
+
+-- | @removeEdge s t g@ removes the edge going from @s@ to @t@, if any.
+--   /O(log n)/.
+
+removeEdge :: Ord n => n -> n -> Graph n e -> Graph n e
+removeEdge s t (Graph g) = Graph $ Map.adjust (Map.delete t) s g
+
+-- | Keep only the edges that satisfy the predicate. /O(n + e)/.
+
+filterEdges :: (Edge n e -> Bool) -> Graph n e -> Graph n e
+filterEdges f =
+  Graph .
+  Map.mapWithKey (\s ->
+    Map.filterWithKey (\t l ->
+      f (Edge { source = s, target = t, label = l }))) .
+  graph
+
+-- | Unzips the graph. /O(n + e)/.
+
+-- This is a naive implementation that uses fmap.
+
+unzip :: Graph n (e, e') -> (Graph n e, Graph n e')
+unzip g = (fst <$> g, snd <$> g)
+
+-- | @composeWith times plus g g'@ finds all edges
+--   @s --c_i--> t_i --d_i--> u@ and constructs the
+--   result graph from @edge(s,u) = sum_i (c_i times d_i)@.
+--
+--   Complexity:  For each edge @s --> t@ in @g@ we look up
+--   all edges starting with @t@ in @g'@.
+--
+--   Precondition: The two graphs must have exactly the same nodes.
+
+composeWith ::
+  Ord n =>
+  (c -> d -> e) -> (e -> e -> e) ->
+  Graph n c -> Graph n d -> Graph n e
+composeWith times plus (Graph g) (Graph g') = Graph (Map.map comp g)
+  where
+  comp m = Map.fromListWith plus
+    [ (u, c `times` d)
+    | (t, c) <- Map.assocs m
+    , m'     <- maybeToList (Map.lookup t g')
+    , (u, d) <- Map.assocs m'
+    ]
+
+------------------------------------------------------------------------
+-- Strongly connected components
 
 -- | The graph's strongly connected components, in reverse topological
 -- order.
 
-sccs' :: Ord n => Graph n n e -> [Graph.SCC n]
+sccs' :: Ord n => Graph n e -> [Graph.SCC n]
 sccs' g =
   Graph.stronglyConnComp
     [ (n, n, map target (edgesFrom g [n]))
@@ -395,7 +514,7 @@ sccs' g =
 -- | The graph's strongly connected components, in reverse topological
 -- order.
 
-sccs :: Ord n => Graph n n e -> [[n]]
+sccs :: Ord n => Graph n e -> [[n]]
 sccs = map Graph.flattenSCC . sccs'
 
 -- | SCC DAGs.
@@ -470,7 +589,7 @@ reachable g scc = case scc of
 
 sccDAG' ::
   forall n e. Ord n
-  => Graph n n e
+  => Graph n e
   -> [Graph.SCC n]
      -- ^ The graph's strongly connected components.
   -> DAG n
@@ -512,21 +631,16 @@ sccDAG' g sccs = DAG theDAG componentMap secondNodeMap
   componentMap = IntMap.fromList (map (mapFst convertInt) components)
 
   secondNodeMap :: Map n Int
-  secondNodeMap = fmap convertInt firstNodeMap
+  secondNodeMap = Map.map convertInt firstNodeMap
 
 -- | Constructs a DAG containing the graph's strongly connected
 -- components.
 
-sccDAG :: Ord n => Graph n n e -> DAG n
+sccDAG :: Ord n => Graph n e -> DAG n
 sccDAG g = sccDAG' g (sccs' g)
 
--- | Returns @True@ iff the graph is acyclic.
-
-acyclic :: Ord n => Graph n n e -> Bool
-acyclic = all isAcyclic . sccs'
-  where
-  isAcyclic Graph.AcyclicSCC{} = True
-  isAcyclic Graph.CyclicSCC{}  = False
+------------------------------------------------------------------------
+-- Reachability
 
 -- | @reachableFrom g n@ is a map containing all nodes reachable from
 -- @n@ in @g@. For each node a simple path to the node is given, along
@@ -540,8 +654,28 @@ acyclic = all isAcyclic . sccs'
 -- time): /O(e log n)/, if the lists are not inspected. Inspection of
 -- a prefix of a list is linear in the length of the prefix.
 
-reachableFrom :: Ord n => Graph n n e -> n -> Map n (Int, [Edge n n e])
-reachableFrom g n = bfs (SQ.singleton (n, BQ.empty)) Map.empty
+reachableFrom :: Ord n => Graph n e -> n -> Map n (Int, [Edge n e])
+reachableFrom g n = reachableFromInternal g (Set.singleton n)
+
+-- | @reachableFromSet g ns@ is a set containing all nodes reachable
+-- from @ns@ in @g@.
+--
+-- Precondition: Every node in @ns@ must be a node in @g@. The number
+-- of nodes in the graph must not be larger than @'maxBound' ::
+-- 'Int'@.
+--
+-- Amortised time complexity (assuming that comparisons take constant
+-- time): /O((|@ns@| + e) log n)/.
+
+reachableFromSet :: Ord n => Graph n e -> Set n -> Set n
+reachableFromSet g ns = Map.keysSet (reachableFromInternal g ns)
+
+-- | Used to implement 'reachableFrom' and 'reachableFromSet'.
+
+reachableFromInternal ::
+  Ord n => Graph n e -> Set n -> Map n (Int, [Edge n e])
+reachableFromInternal g ns =
+  bfs (SQ.fromList (map (, BQ.empty) (Set.toList ns))) Map.empty
   where
   bfs !q !map = case SQ.lview q of
     Nothing          -> map
@@ -565,55 +699,38 @@ reachableFrom g n = bfs (SQ.singleton (n, BQ.empty)) Map.empty
 -- nodes in the graph must not be larger than @'maxBound' :: 'Int'@.
 --
 -- Amortised time complexity (assuming that comparisons and the
--- predicates take constant time to compute): /O(e log n)/.
+-- predicates take constant time to compute): /O(n + e log n)/.
 
 walkSatisfying ::
   Ord n =>
-  (e -> Bool) -> (e -> Bool) ->
-  Graph n n e -> n -> n -> Maybe [Edge n n e]
+  (Edge n e -> Bool) -> (Edge n e -> Bool) ->
+  Graph n e -> n -> n -> Maybe [Edge n e]
 walkSatisfying every some g from to =
   case
     [ (l1 + l2, p1 ++ [e] ++ map transposeEdge (reverse p2))
     | e <- everyEdges
-    , some (label e)
+    , some e
     , (l1, p1) <- maybeToList (Map.lookup (source e) fromReaches)
     , (l2, p2) <- maybeToList (Map.lookup (target e) reachesTo)
     ] of
     []  -> Nothing
     ess -> Just $ snd $ List.minimumBy (compare `on` fst) ess
   where
-  everyEdges = [ e | e <- toList g, every (label e) ]
+  everyEdges = [ e | e <- edges g, every e ]
 
-  fromReaches = reachableFrom (fromList everyEdges) from
+  fromReaches = reachableFrom (fromEdges everyEdges) from
 
   reachesTo =
-    reachableFrom (fromList (map transposeEdge everyEdges)) to
+    reachableFrom (fromEdges (map transposeEdge everyEdges)) to
 
--- * Graph composition
-
--- | @composeWith times plus g g'@ finds all edges
---   @s --c_i--> t_i --d_i--> u@ and constructs the
---   result graph from @edge(s,u) = sum_i (c_i times d_i)@.
---
---   Complexity:  for each edge @s --> t@ in @g@ we lookup up
---   all edges starting in with @t@ in @g'@.
---
-composeWith :: (Ord t, Ord u) => (c -> d -> e) -> (e -> e -> e) ->
-  Graph s t c -> Graph t u d -> Graph s u e
-composeWith times plus (Graph g) (Graph g') = Graph $
-  Map.mapMaybe (discardEmpty . comp) g where
-    comp m = Map.fromListWith plus
-      [ (u, c `times` d)
-      | (t, c) <- Map.assocs m
-      , m'     <- maybeToList (Map.lookup t g')
-      , (u, d) <- Map.assocs m'
-      ]
+------------------------------------------------------------------------
+-- Transitive closure
 
 -- | Transitive closure ported from "Agda.Termination.CallGraph".
 --
 --   Relatively efficient, see Issue 1560.
 
-complete :: (Eq e, Null e, SemiRing e, Ord n) => Graph n n e -> Graph n n e
+complete :: (Eq e, Null e, SemiRing e, Ord n) => Graph n e -> Graph n e
 complete g = repeatWhile (mapFst (not . discrete) . combineNewOld' g) g
   where
     combineNewOld' new old = unzip $ unionWith comb new' old'
@@ -637,7 +754,7 @@ complete g = repeatWhile (mapFst (not . discrete) . combineNewOld' g) g
 --
 --   @complete g = snd $ last $ completeIter g@
 
-completeIter :: (Eq e, Null e, SemiRing e, Ord n) => Graph n n e -> [(Graph n n e, Graph n n e)]
+completeIter :: (Eq e, Null e, SemiRing e, Ord n) => Graph n e -> [(Graph n e, Graph n e)]
 completeIter g = iterWhile (not . discrete) (combineNewOld' g) g
   where
     combineNewOld' new old = unzip $ unionWith comb new' old'
@@ -670,7 +787,7 @@ completeIter g = iterWhile (not . discrete) (combineNewOld' g) g
 
 gaussJordanFloydWarshallMcNaughtonYamadaReference ::
   forall n e. (Ord n, Eq e, StarSemiRing e) =>
-  Graph n n e -> Graph n n e
+  Graph n e -> Graph n e
 gaussJordanFloydWarshallMcNaughtonYamadaReference g =
   toGraph (foldr step initialMatrix nodeIndices)
   where
@@ -706,10 +823,12 @@ gaussJordanFloydWarshallMcNaughtonYamadaReference g =
       ]
 
   toGraph m =
-    fromList [ Edge (indexMap Map.! i) (indexMap Map.! j) e
-             | ((i, j), e) <- Array.assocs m
-             , e /= ozero
-             ]
+    fromEdges [ Edge (indexMap Map.! i) (indexMap Map.! j) e
+              | ((i, j), e) <- Array.assocs m
+              , e /= ozero
+              ]
+      `union`
+    fromNodeSet (nodes g)
 
 -- | Computes the transitive closure of the graph.
 --
@@ -730,7 +849,7 @@ gaussJordanFloydWarshallMcNaughtonYamadaReference g =
 
 gaussJordanFloydWarshallMcNaughtonYamada ::
   forall n e. (Ord n, Eq e, StarSemiRing e) =>
-  Graph n n e -> (Graph n n e, [Graph.SCC n])
+  Graph n e -> (Graph n e, [Graph.SCC n])
 gaussJordanFloydWarshallMcNaughtonYamada g =
   (loop components g, components)
   where
@@ -738,7 +857,7 @@ gaussJordanFloydWarshallMcNaughtonYamada g =
   forwardDAG = sccDAG' g components
   reverseDAG = oppositeDAG forwardDAG
 
-  loop :: [Graph.SCC n] -> Graph n n e -> Graph n n e
+  loop :: [Graph.SCC n] -> Graph n e -> Graph n e
   loop []           !g = g
   loop (scc : sccs)  g =
     loop sccs (foldr step g (Graph.flattenSCC scc))
@@ -748,7 +867,7 @@ gaussJordanFloydWarshallMcNaughtonYamada g =
     -- All nodes that can reach the SCC.
     canReach     = reachable reverseDAG scc
 
-    step :: n -> Graph n n e -> Graph n n e
+    step :: n -> Graph n e -> Graph n e
     step k !g =
       foldr (insertEdgeWith oplus) g
         [ Edge i j e
