@@ -12,7 +12,6 @@ import Prelude             hiding (concat)
 import Control.Monad       hiding (mapM, forM)
 import Control.Monad.State
 import Control.Monad.Writer
-import Control.Applicative
 import Data.Traversable as Trav hiding (for, sequence)
 import Data.Foldable as Fold
 import Data.Maybe
@@ -124,23 +123,11 @@ forM' = flip mapM'
 -- Lists and monads -------------------------------------------------------
 
 -- | A monadic version of @'mapMaybe' :: (a -> Maybe b) -> [a] -> [b]@.
-mapMaybeM
-#if __GLASGOW_HASKELL__ <= 708
-  :: (Functor m, Monad m)
-#else
-  :: Monad m
-#endif
-  => (a -> m (Maybe b)) -> [a] -> m [b]
+mapMaybeM :: Monad m => (a -> m (Maybe b)) -> [a] -> m [b]
 mapMaybeM f xs = catMaybes <$> Trav.mapM f xs
 
 -- | The @for@ version of 'mapMaybeM'.
-forMaybeM
-#if __GLASGOW_HASKELL__ <= 708
-  :: (Functor m, Monad m)
-#else
-  :: Monad m
-#endif
-  => [a] -> (a -> m (Maybe b)) -> m [b]
+forMaybeM :: Monad m => [a] -> (a -> m (Maybe b)) -> m [b]
 forMaybeM = flip mapMaybeM
 
 -- | A monadic version of @'dropWhile' :: (a -> Bool) -> [a] -> [a]@.
@@ -155,6 +142,17 @@ partitionM f [] =
 partitionM f (x:xs) =
   (\ b (l, r) -> if b then (x:l, r) else (l, x:r)) <$> f x <*> partitionM f xs
 
+-- MonadPlus -----------------------------------------------------------------
+
+-- | Translates 'Maybe' to 'MonadPlus'.
+fromMaybeMP :: MonadPlus m => Maybe a -> m a
+fromMaybeMP = maybe mzero return
+
+-- | Generalises the 'catMaybes' function from lists to an arbitrary
+-- 'MonadPlus'.
+catMaybesMP :: MonadPlus m => m (Maybe a) -> m a
+catMaybesMP = (>>= fromMaybeMP)
+
 -- Error monad ------------------------------------------------------------
 
 -- | Finally for the 'Error' class. Errors in the finally part take
