@@ -35,14 +35,6 @@ import Text.PrettyPrint (Doc, render)
 import Agda.Utils.Bag (Bag)
 import qualified Agda.Utils.Bag as Bag
 
--- Andreas, 2015-06-24 orphan instance has to go here
--- to be able to define instance Null Doc
-
-#if !MIN_VERSION_pretty(1,1,2)
-instance Eq Doc where
-  (==) = (==) `on` render
-#endif
-
 class Null a where
   empty :: a
   null  :: a -> Bool
